/*
**	Desktop Calculator.
**
**	V2.6	Jim Cathey	02/09/95
**	 	Tuned up video copy logic.
**
**	V2.5	Jim Cathey	12/15/93
**	 	Added automatic video copy when WINDOWS_ON contains a list.
**
**	V2.4	Jim Cathey	4/6/93
**		Fixed int/frac code to use modf() function, which lets
**		it work properly with larger numbers.  Added in scientific
**		and engineering display modes.  Added in miscellaneous
**		unit conversions.  Added ability to re-enter numbers from
**		the tape log with the mouse.  Fixed up input parsing to
**		handle scientific notation from the command line and the
**		tape log.  Split in 2 source modules.
**
**	V2.3	Jim Cathey	9/1/92
**		Changed keytop drawing to use an included bitmap.  Added
**		scientific notation input when in floating-point mode.
**		Added many scientific calculator functions.
**
**	V2.2	Jim Cathey	7/18/91
**		Added Cut/Paste logic, patched around bug in C runtime
**		library that core-dumps when converting a floating-point
**		value to 0x80000000.
**
**	V2.1	Jim Cathey	12/12/89
**		Changed to support new TextSize call and friends.
**
**	V2.0	Jim Cathey	11/15/89
**		Terminate digit entry on change from hex to octal or back.
**
**	V1.9	Jim Cathey	11/2/89
**		Fixed octal digit entry so that it works.
**
**	V1.8	Jim Cathey	9/20/89
**		Fixed so mouse-scrolling didn't flash the arrows of the
**		scrollbar (avoid extra HiliteControls of 0 hilite state).
**		Also changed scrollbar logic a bit to use new call.
**
**	V1.7	Jim Cathey	8/14/89
**		Fixed double button flash on mouse activity.
**
**	V1.6	Jim Cathey	8/7/89
**		Fixed bug in MoveWindow call (last parm s.b. 0).
**
**	V1.5	Jim Cathey	5/19/89
**		Cleaned up excessive display flashes on data entry.
**
**	V1.4	Jim Cathey	5/17/89
**		Fixed fixed-place 3 to actually work.  Picked up 1-6 as well.
**		Fixed (again, perhaps correctly) persistent stack lift bug.
**
**	V1.3	Jim Cathey	5/9/89
**		Fixed flashing of keytops to not take so long on typeahead.
**
**	V1.2	Jim Cathey	5/9/89
**		Fixed memory conversion bug in hex<->dex mode switch.  Also
**		fixed stack lift bug for STO/RCL.
**
**	V1.1	Jim Cathey	4/21/89
**		Fixed sign change bug in hex mode when _not_ entering digits.
**
**	V1.0	Jim Cathey	4/21/89
**
**	Usage:	calc [-xpos] [-ypos] [-idigits] [-e[=][t]] [-p] [-mX] [-f]
**		The command "calc -x20 -y30 -i44 -e=t" places calc's window
**		at screen coordinates (20, 30), preloads 44 into the X
**		register, and exports X upon an = request, and upon exit, and
**		exports the tape log upon exit.  The -p option shuts off
**		the 'programmer' features (hex and octal modes).  The -mX
**		sets the initial mode to one of: F-floating, <2|3|4>-fixed
**		place, H-hex, O-octal, T-tape.  The -f option shuts off key
**		flash on keystrokes.
*/

#include "calc.h"

struct {			/* Key data structure initializer. */
    short x, y;			/* Rectangle starting position in window. */
    short bx, by;		/* Rectangle position in keytops bitmap. */
    char width, height;		/* Rectangle size. */
    char flag;			/* Mode-specific key flag, >0:sci, <0:prog. */
    char arg;			/* Argument to function. */
    int (*func)();		/* Function to call. */
} keys[] = {
    KC2, KR8, BC2, BR8, KWID, KHEI, 4, '0', dodigit, /* 0 */
    KC2, KR7, BC2, BR7, KWID, KHEI, 4, '1', dodigit, /* 1 */
    KC3, KR7, BC3, BR7, KWID, KHEI, 4, '2', dodigit, /* 2 */
    KC4, KR7, BC4, BR7, KWID, KHEI, 4, '3', dodigit, /* 3 */
    KC2, KR6, BC2, BR6, KWID, KHEI, 4, '4', dodigit, /* 4 */
    KC3, KR6, BC3, BR6, KWID, KHEI, 4, '5', dodigit, /* 5 */
    KC4, KR6, BC4, BR6, KWID, KHEI, 4, '6', dodigit, /* 6 */
    KC2, KR5, BC2, BR5, KWID, KHEI, 4, '7', dodigit, /* 7 */
    KC3, KR5, BC3, BR5, KWID, KHEI, 4, '8', dodigit, /* 8 */
    KC4, KR5, BC4, BR5, KWID, KHEI, 4, '9', dodigit, /* 9 */
    KC3, KR8, BC3, BR8, KWID, KHEI, 4, '0'|0x80, dodigit, /* 10 */
    KC4, KR8, BC4, BR8, KWID, KHEI, 4, '.', dodigit, /* 11 */
    KC2, KR4, BC2, BR4, KWID, KHEI, 4, 0, dochs, /* 12 +/- */
    KC3, KR4, BC3, BR4, KWID, KHEI, 4, 0, dodiv, /* 13 / */
    KC4, KR4, BC4, BR4, KWID, KHEI, 4, 0, domul, /* 14 */
    KC5, KR4, BC5, BR4, KWID, KHEI, 4, 0, dosub, /* 15 */
    KC5, KR5, BC5, BR5, KWID, KHEI2, 0, 0, doadd, /* 16 */
    KC2, KR3, BC2, BR3, KWID, KHEI, -1, 'A', dodigit, /* 17 */
    KC3, KR3, BC3, BR3, KWID, KHEI, -1, 'B', dodigit, /* 18 */
    KC4, KR3, BC4, BR3, KWID, KHEI, -1, 'C', dodigit, /* 19 */
    KC2, KR2, BC2, BR2, KWID, KHEI, -1, 'D', dodigit, /* 20 */
    KC3, KR2, BC3, BR2, KWID, KHEI, -1, 'E', dodigit, /* 21 */
    KC4, KR2, BC4, BR2, KWID, KHEI, -1, 'F', dodigit, /* 22 */
    KC5, KR7, BC10, BR1, KWID, KHEI2, 0, 0, doenter, /* 23 */
    KC1, KR8, BC1, BR8, KWID, KHEI, 0, 0, dolastx, /* 24 */
    KC5, KR1, BC5, BR1, KWID, KHEI, -1, 0, donot, /* 25 */
    KC2, KR1, BC2, BR1, KWID, KHEI, -1, 0, doand, /* 26 */
    KC3, KR1, BC3, BR1, KWID, KHEI, -1, 0, door, /* 27 */
    KC4, KR1, BC4, BR1, KWID, KHEI, -1, 0, doxor, /* 28 */
    KC1, KR1, BC1, BR1, KWID, KHEI, 0, '=', logxkey, /* 29 */
    KC5, KR2, BC5, BR2, KWID, KHEI, -1, 0, doshiftl, /* 30 << */
    KC5, KR3, BC5, BR3, KWID, KHEI, -1, 0, doshiftr, /* 31 >> */
    KC1, KR2, BC1, BR2, KWID, KHEI, 0, 0, dorollup, /* 32 */
    KC1, KR3, BC1, BR3, KWID, KHEI, 0, 0, dorolldown, /* 33 */
    KC1, KR4, BC1, BR4, KWID, KHEI, 0, 0, dobs, /* 34 */
    KC1, KR7, BC1, BR7, KWID, KHEI, 0, 0, dosto, /* 35 */
    KC1, KR6, BC1, BR6, KWID, KHEI, 0, 0, dorcl, /* 36 */
    KC1, KR5, BC1, BR5, KWID, KHEI, 0, 0, doexg, /* 37 */

				/* Decimal-mode scientific functions. */
    KC2, KR1, BC6, BR1, KWID, KHEI, 1, 38, doprefix, /* 38 */
    KC3, KR1, BC7, BR1, KWID, KHEI, 1, 39, dosin, /* 39, sin key */
    KC4, KR1, BC8, BR1, KWID, KHEI, 1, 40, docos, /* 40, cos key */
    KC5, KR1, BC9, BR1, KWID, KHEI, 1, 41, dotan, /* 41, tan key */
    KC2, KR2, BC6, BR2, KWID, KHEI, 1, 42, doexp, /* 42, exp key */
    KC3, KR2, BC7, BR2, KWID, KHEI, 1, 43, doln, /* 43, ln key */
    KC4, KR2, BC8, BR2, KWID, KHEI, 1, 44, dodeltapercent, /* 44, d% key */
    KC5, KR2, BC9, BR2, KWID, KHEI, 1, 45, dopercent, /* 45, % key */
    KC2, KR3, BC6, BR3, KWID, KHEI, 1, 'E', dodigit, /* 46, EEX key */
    KC3, KR3, BC7, BR3, KWID, KHEI, 1, 47, dosqrt, /* 47, Sqrt key */
    KC4, KR3, BC8, BR3, KWID, KHEI, 1, 48, dopow, /* 48, Power key */
    KC5, KR3, BC9, BR3, KWID, KHEI, 1, 49, dorecip, /* 49, Reciprocal key */

				/* Function prefix stuff. */
    KC2, KR1, BC6, BR4, KWID, KHEI, 2, 50, doprefix, /* 50 */
    KC3, KR1, BC7, BR4, KWID, KHEI, 2, 51, doasin, /* 51 */
    KC4, KR1, BC8, BR4, KWID, KHEI, 2, 52, doacos, /* 52 */
    KC5, KR1, BC9, BR4, KWID, KHEI, 2, 53, doatan, /* 53 */
    KC2, KR2, BC6, BR5, KWID, KHEI, 2, 54, doexp10, /* 54, pwr10 key */
    KC3, KR2, BC7, BR5, KWID, KHEI, 2, 55, dolog10, /* 55, log10 key */
    KC4, KR2, BC8, BR5, KWID, KHEI, 2, 56, donothing, /* 56 */
    KC5, KR2, BC9, BR5, KWID, KHEI, 2, 57, dopi, /* 57 */
    KC2, KR3, BC6, BR6, KWID, KHEI, 2, 58, dofact, /* 58 */
    KC3, KR3, BC7, BR6, KWID, KHEI, 2, 59, dosquare, /* 59 */
    KC4, KR3, BC8, BR6, KWID, KHEI, 2, 60, doint, /* 60 */
    KC5, KR3, BC9, BR6, KWID, KHEI, 2, 61, dofrac, /* 61 */

				/* Function prefix of digit keys. */
    KC2, KR5, BC10, BR5, KWID, KHEI, 3, 62, dodegtorad, /* 62 */
    KC3, KR5, BC11, BR5, KWID, KHEI, 3, 63, dointocm, /* 63 */
    KC4, KR5, BC12, BR5, KWID, KHEI, 3, 64, dodegftodegc, /* 64 */
    KC2, KR6, BC10, BR6, KWID, KHEI, 3, 65, doradtodeg, /* 65 */
    KC3, KR6, BC11, BR6, KWID, KHEI, 3, 66, docmtoin, /* 66 */
    KC4, KR6, BC12, BR6, KWID, KHEI, 3, 67, dodegctodegf, /* 67 */
    KC2, KR7, BC10, BR7, KWID, KHEI, 3, 68, dogaltolit, /* 68 */
    KC3, KR7, BC11, BR7, KWID, KHEI, 3, 69, dolbtokg, /* 69 */
    KC4, KR7, BC12, BR7, KWID, KHEI, 3, 70, domitokm, /* 70 */
    KC2, KR8, BC10, BR8, KWID, KHEI, 3, 71, dolittogal, /* 71 */
    KC3, KR8, BC11, BR8, KWID, KHEI, 3, 72, dokgtolb, /* 72 */
    KC4, KR8, BC12, BR8, KWID, KHEI, 3, 73, dokmtomi, /* 73 */

				/* Function prefix of arithmetic keys. */
    KC2, KR4, BC10, BR4, KWID, KHEI, 3, 74, dooztog, /* 74 */
    KC3, KR4, BC11, BR4, KWID, KHEI, 3, 75, dogtooz, /* 75 */
    KC4, KR4, BC12, BR4, KWID, KHEI, 3, 76, doydtom, /* 76 */
    KC5, KR4, BC5, BR8, KWID, KHEI, 3, 77, domtoyd /* 77 */
};

struct vidcopy {
    struct vidcopy *next;
    Rect winrect;
    char connectid;
    char windowid;
} *vchead;

main(argc, argv)
int argc;
char *argv[];
{
    int i, j, xpos, ypos;
    int trapqfd, wmfd;
    Rect button;
    char *argp, *Title, *dispsystem, *inparg, *progname, *getenv();

    signal(SIGFPE, SIG_IGN);	/* Don't core dump on overflows, etc. */
    xpos = ypos = 100;
    inparg = 0;
    dispsystem = getenv("WINDOWS_ON");
    export = 0;
    keyflash = progfuncts = 1;
    calcmode = 0;		/* Assume floating point at first. */
	tapeon = 0;
    radix = 10;
    progname = argv[0];
    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'e':		/* Export result to stdout. */
		export |= 1;
		switch (*argp++) {
		  case '=':
		    export |= 2;
		    break;
		  case 't':
		    export |= 4;
		    break;
		  case 0:
		    break;
		  default:
		    usage(progname);
		    exit(1);
		    break;
		}
		break;

	      case 'm':		/* Initial calculator mode. */
		switch (toupper(*argp)) {
		  case 'E':
		    engnot = 1;
		    calcmode = 6;
		    break;

		  case 'F':
		    break;

		  case 'H':
		    radix = 16;
		    calcmode = -1;
		    break;

		  case 'O':
		    radix = 8;
		    calcmode = -2;
		    break;

		  case 'S':
		    scinot = 1;
		    calcmode = 6;
		    break;

		  case 'T':
		    tapeon = 1;
		    break;

		  default:
		    calcmode = *argp - '0';
		    break;
		}
		break;

	      case 'p':		/* Remove programmer options. */
		progfuncts = 0;
		break;

	      case 'f':		/* Remove key flash. */
		keyflash = 0;
		break;

	      case 'x':		/* X position of window. */
		xpos = atoi(argp);
		break;

	      case 'y':		/* Y position of window. */
		ypos = atoi(argp);
		break;

	      case 'i':		/* Input argument to X register. */
		inparg = argp;
		break;

	      default:
		usage(progname);
		exit(1);
		break;
	    }
	}
	else {
	    dispsystem = argp;
	}
    }
    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
	fprintf(stderr, "%s: Error %d in InitGraf().\n", argv[0], -trapqfd);
	exit(2);
    }
    if ((wmfd = ConnectGraf(dispsystem)) < 0) {
	fprintf(stderr, "%s: Unable to connect to Window Manager (Error #%d in ConnectGraf).\n", 
		argv[0], -wmfd);
	exit(2);
    }
    SetRect(&winrect, xpos, ypos, xpos+(tapeon ? WIDTH2 : WIDTH), ypos+HEIGHT);
    Title = "Calculator";
    window = NewWindow(&winrect, Title, hideWindow,  noGrowDocProc,
		       frontWindow, hasDrag|hasGoAway|hasZoom|hasSendToBack,
		       wmfd);

    SetPort(window);
    BackPat((Pattern *) 45);
    ShowWindow(window);

    Event(updateEvt, updateevent);
    Event(menuEvt, menuevent);
    Event(mouseEvt, mouseevent);
    Event(keyEvt, keyboardevent);
    Event(activateEvt, activateevent);
    Event(windowEvt, windowevent);
    Event(scrapEvt, scrapevent);

    stacklift = 0;		/* Clear out the calculator. */
    entering = oneshot = tapeditems = 0;
    tapeview = 0;
    tapeheader.next = 0;
    tapeheader.text = 0;
    lasttaped = &tapeheader;
    lastx.dex = x.dex = y.dex = z.dex = t.dex = 0;
    for (i=0; i<sizeof(memory)/sizeof(memory[0]); i++)
	memory[i].dex = 0;
    entrysign = 0;		/* Numeric entry is positive. */
    entryeex = entryeexsign = 0; /* Not entering power-of-ten exponent. */
    entryptr = entrybuf;
    if (inparg) {		/* Force-feed X register. */
	i = StringLen(inparg);
	memcpy(entrybuf, inparg, i);
	entryptr += i;
	reparsex();
	stacklift = 1;
    }
    NewMenu(OPTIONMENU, "Options");
    AppendMenu(OPTIONMENU, "/FFloating Point;/22 Place Fixed;/33 Place Fixed;"
	       "/44 Place Fixed;-;/HHexadecimal;/OOctal;/SSci. Notation;"
	       "/EEng. Notation;/TShow Tape;-;/QQuit");
    SetItem(OPTIONMENU, TAPEMODE, tapeon ? "Hide Tape" : "Show Tape");
    i = (calcmode<0) ? -calcmode + (HEXMODE - 1) : calcmode;
    if (i<=0)
	i = 1;
    setone(OPTIONMENU, i, OCTMODE); /* Default to Floating Point mode. */
    if (!progfuncts) {
	DisableItem(OPTIONMENU, HEXMODE);
	DisableItem(OPTIONMENU, OCTMODE);
    }
    CheckItem(OPTIONMENU, ENGMODE, engnot);
    CheckItem(OPTIONMENU, SCIMODE, scinot);
    NewMenu(EDITMENU, "Edit");
    AppendMenu(EDITMENU, "/XCut;/CCopy;(/VPaste");
    LocalGetFontInfo(0, normalText, 12, &finfo);
    LocalGetFontInfo(1, normalText, 12, &finfo1);
    TextFont(3);
    TextSize(9);
    GetFontInfo(&finfo3);
    finfo3.leading += finfo3.ascent + finfo3.descent;

    SetRect(&regwindow, 10, 10, 188, 80);
    SetRect(&taperect, WIDTH + 1, 10, WIDTH2 - 7, HEIGHT - 10);
    tapewinsize = (taperect.edge.bottom - taperect.edge.top) / finfo3.leading;

    button = taperect;
    button.edge.left = button.edge.right - 17;
    button.edge.top--;
    button.edge.bottom++;
    taperect.edge.right -= 17;

    if (argp = dispsystem) {	/* Handle copying video to other WS. */
	char *cp, buf[80]; int fd, wid; long winds; struct vidcopy *vcp;

	winds = 0;
	while (argp = strchr(argp, '\n')) {
	    argp++;
	    cp = strchr(argp, '\n');
	    if (!cp)
		cp = argp + strlen(argp);
	    if (cp - argp) {
		strncpy(buf, argp, cp - argp);
		buf[cp - argp] = 0;
		if ((fd = ConnectGraf(buf)) >= 0) {
		    strcat(strcpy(buf, "Copy of "), Title);
		    if ((wid = NewWindow(&winrect, buf, hideWindow,
					 documentProc, frontWindow,
					 hasGoAway|hasSendToBack|hasDrag,
					 fd)) >= 0) {
			SetPort(wid);
			BackPat((Pattern *) 45);
			ShowWindow(wid);
			winds |= 1 | 1 << wid;
			if (vcp = (struct vidcopy *) malloc(sizeof *vcp)) {
			    vcp->next = vchead;
			    vchead = vcp;
			    vcp->winrect = winrect;
			    vcp->connectid = fd;
			    vcp->windowid = wid;
			}
		    } else {
			DisconnectGraf(fd);
		    }
		}
	    }
	}
	if (winds) {
	    SetPorts(winds);
	    TextFont(3);
	    TextSize(9);
	}
    }
    InitControls();
    scrollbar = NewControl(window, &button, "V", 1,
			     0, 0, 0, scrollBarProc, 0);
    SetCtlAction(scrollbar, MyAction); /* For continuous scrolling while */
				/* the buttons of scroll bar are pressed. */
    setscrollbar();		/* Init scrollbar limits. */
    while (1) {
	if (GetEvent(anyWindow, noEvent)) {
	    if ((i = (EventError() >> 16)) != wmfd) {
		deletecopy(i, -1); /* Video copy died, take out & continue. */
		continue;
	    }
	    exit(3);
	}
    }
}

/*
** Delete a video copy window, specified either by connection ID or
** by window ID.
*/
deletecopy(fd, wid)
int fd, wid;
{
    long port;
    struct vidcopy *vcp, *vcp2;

    vcp = vchead;
    vcp2 = 0;
    while (vcp) {
	if (fd >= 0 && vcp->connectid == fd
	    || wid >= 0 && vcp->windowid == wid) {
	    DisconnectGraf(vcp->connectid);
	    if (vcp2)
		vcp2->next = vcp->next;
	    else
		vchead = vcp->next;
	    GetPort(&port);
	    if (port & 0x80000000L)
		port &= ~(1 << vcp->windowid);
	    SetPort(port);
	    free(vcp);
	    break;
	}
	vcp2 = vcp;
	vcp = vcp->next;
    }
}

short mywindowisinfront;        /* TRUE if one of my windows is active. */
short doimportscrap = 1;        /* New incoming scrap. */
short doexportscrap;            /* New outgoing scrap. */

static void
activateevent(ep)
EventRecord *ep;
{
    short i;
    long port;

    if (ep->who == window) {
	if (ep->message & 1) {	/* Activating? */
	    mywindowisinfront = 1;
	    HiliteControl(scrollbar, tapeditems < tapewinsize ? 255 : 0);
	    if (doimportscrap) {
		GetPort(&port);
		if (port != ep->who)
		    SetPort(ep->who);
		fetchscrap();
		if (port != ep->who)
		    SetPort(port);
	    }
	} else {		/* Deactivating. */
	    mywindowisinfront = 0;
	    HiliteControl(scrollbar, 255); /* Mark scroll bar inactive. */
	    if (doexportscrap) {
		GetPort(&port);
		if (port != ep->who)
		    SetPort(ep->who);
		writescrap();
		if (port != ep->who)
		    SetPort(port);
	    }
	}
    }
}

static void
keyboardevent(ep)
EventRecord *ep;
{
    short key;
    long save;
    struct vidcopy *vcp;
    struct Rect *r;

    key = ep->message & F_KEYMASK;
    if (ep->message & F_KEYUP)
	return;
    r = &winrect;
    if (ep->who != window) {	/* Don't let copy windows do much. */
	if (key == K_ESC || key == K_RESET)
	    goto processkey; /* Let copy windows be deleted. */
	if (ep->message & F_ALT) {
	    if (key == K_UP || key == K_DOWN
		|| key == K_LEFT || key == K_RIGHT) {
		vcp = vchead;
		while (vcp) {
		    if (vcp->windowid == ep->who) {
			r = &vcp->winrect;
			break;
		    }
		    vcp = vcp->next;
		}
		goto processkey; /* Let copy windows move around. */
	    }
	}
dobell:
	if (ep->message & F_KEYDOWN) {
	    GetPort(&save);
	    SetPort(ep->who);
	    Bell();
	    SetPort(save);
	}
	return;
    }
processkey:
    if (!(ep->message & F_KBDIII) && key == K_ESC && ep->message & F_SHIFT) {
	key = '~';		/* Handle ISC keyboard's twiddle. */
	ep->message &= ~F_FUNC;
    }
    if (ep->message & F_FUNC) {
	if (key >= '0' && key <= '9') {
	    if (oneshot) {	/* STO or RCL in progress? */
		if (oneshot == RCLPEND) { /* RCL */
		    enterval(&memory[key - '0']);
		    flashkey(key - '0');
		    hilitekey(36);
		} else if (oneshot == STOPEND) { /* STO */
		    memory[key - '0'].dex = x.dex;
		    terminateinput();
		    flashkey(key - '0');
		    hilitekey(35);
		} else {	/* FUNC */
		    oneshot = 0;
		    drawkeys();
		    goto dobell;
		}
		oneshot = 0;
	    } else
		dodigit(key);
	} else if (oneshot) {	/* STO or RCL followed by illegal key. */
	    if (oneshot == STOPEND)
		hilitekey(35);
	    else if (oneshot == RCLPEND)
		hilitekey(36);
	    else {
		oneshot = 0;
		drawkeys();
		goto dobell;
	    }
	    oneshot = 0;
	    goto dobell;
	} else switch (key) {
	  case K_ESC:		/* Quit. */
	  case K_RESET:
	    if (ep->who == window)
		bailout();
	    else
		deletecopy(-1, ep->who);
	    break;

	  case K_BACKSP:	/* Backspace keys. */
	  case K_BCKSP2:
	    dobs();
	    break;

	  case K_NPENTER:
	  case K_ENTER:
	  case K_SEND:
	    doenter();
	    break;

	  case K_PF1:		/* CHS (Num Lock key). */
	    dochs();
	    break;

	  case K_PF2:		/* '/' key. */
	    dodiv();
	    break;

	  case K_PF3:		/* '*' key. */
	    domul();
	    break;

	  case K_PF4:		/* '-' key. */
	    dosub();
	    break;

	  case '+':		/* '+' key. */
	  case K_NMINUS:	/* (There's already a -, but no + on ISC */
	    doadd();		/*  keyboard -- that one's ENTER.) */
	    break;

	  case K_UP:		/* Roll-up Key. */
	    if (ep->message & F_ALT) {
		OffsetRect(r, 0, -18);
		MoveWindow(ep->who, r->edge.left, r->edge.top, 0);
	    } else 
		dorollup();
	    break;

	  case K_DOWN:		/* Roll-down Key. */
	    if (ep->message & F_ALT) {
		OffsetRect(r, 0, 18);
		MoveWindow(ep->who, r->edge.left, r->edge.top, 0);
	    } else 
		dorolldown();
	    break;

	  case K_LEFT:		/* X<->Y. */
	    if (ep->message & F_ALT) {
		OffsetRect(r, -18, 0);
		MoveWindow(ep->who, r->edge.left, r->edge.top, 0);
	    } else 
		doexg();
	    break;

	  case K_RIGHT:		/* X<->Y */
	    if (ep->message & F_ALT) {
		OffsetRect(r, 18, 0);
		MoveWindow(ep->who, r->edge.left, r->edge.top, 0);
	    } else 
		doexg();
	    break;

	  case K_PGUP:		/* Scroll tape up. */
	    if (tapeon && tapeditems > tapewinsize)
		MyAction(0, inPageUp);
	    else
		goto dobell;
	    break;

	  case K_PGDN:		/* Scroll tape down. */
	    if (tapeon && tapeditems > tapewinsize)
		MyAction(0, inPageDown);
	    else
		goto dobell;
	    break;

	  case K_HOME:		/* Scroll tape to beginning. */
	    if (tapeon && tapeditems > tapewinsize)
		scrollto(0);
	    else
		goto dobell;
	    break;

	  case K_END:		/* Scroll tape to end. */
	    if (tapeon && tapeditems > tapewinsize)
		scrollto(tapeditems - tapewinsize);
	    else
		goto dobell;
	    break;

	  case '.':
	    if (calcmode < 0)
		goto dobell;
	    else
		dodigit(key);
	    break;

	  case K_NPAD00:	/* Double-zero key. */
	    dodigit('0'|0x80);
	    break;

	  default:		/* Unknown function key. */
	    goto dobell;
	    break;
	}
    } else {			/* Not a function. */
	if (key >= '0' && key <= '9') {
	    if (oneshot) {	/* STO or RCL in progress? */
		if (oneshot == RCLPEND) { /* RCL */
		    enterval(&memory[key - '0']);
		    flashkey(key - '0');
		    hilitekey(36);
		} else if (oneshot == STOPEND) { /* STO */
		    memory[key - '0'].dex = x.dex;
		    terminateinput();
		    flashkey(key - '0');
		    hilitekey(35);
		} else {
		    oneshot = 0;
		    goto dobell;
		    drawkeys();
		}
		oneshot = 0;
	    } else
		dodigit(key);
	} else if (oneshot) {	/* STO or RCL followed by illegal key. */
	    if (oneshot == STOPEND)
		hilitekey(35);
	    else if (oneshot == RCLPEND)
		hilitekey(36);
	    else {
		oneshot = 0;
		drawkeys();
	    }
	    oneshot = 0;
	    goto dobell;
	} else switch (key) {
	  case '.':
	    dodigit(key);
	    break;

	  case '+':		/* Basic math functions. */
	    doadd();
	    break;

	  case '-':
	    dosub();
	    break;

	  case '*':
	    domul();
	    break;

	  case '/':
	    dodiv();
	    break;

	  case '%':
	    dopercent();
	    break;

	  case '&':		/* Basic Hex-mode operators. */
	    doand();
	    break;

	  case '|':
	    door();
	    break;

	  case '^':
	    doxor();
	    break;

	  case '~':
	    donot();
	    break;

	  case '<':
	    doshiftl();
	    break;

	  case '>':
	    doshiftr();
	    break;

	  case '=':
	    logxkey('=');
	    if (export & 2)
		exportx();
	    break;

	  case 'F':
	  case 'f':
	    if (calcmode >= 0) { /* Decimal mode? */
		if (oneshot == FUNCPEND) {
		    oneshot = 0;
		    drawkeys();
		} else {
		    doprefix();
		}
		break;
	    }			/* Fall through! */
	  case 'A':
	  case 'a':
	  case 'B':
	  case 'b':
	  case 'C':
	  case 'c':
	  case 'D':
	  case 'd':
	  case 'E':
	  case 'e':
	    key = toupper(key);
	    dodigit(key);
	    break;

	  case 'r':		/* Recall. */
	  case 'R':
	    dorcl();
	    break;

	  case 's':		/* Store. */
	  case 'S':
	    dosto();
	    break;

	  case 'q':		/* Quit. */
	  case 'Q':
	    bailout();
	    break;

	  case 'l':		/* Last X. */
	  case 'L':
	    dolastx();
	    break;

	  case 'y':		/* Y^X key. */
	  case 'Y':
	    if (calcmode >= 0)
		dopow();
	    else
		goto dobell;
	    break;

	  default:		/* Unknown non-function key. */
	    goto dobell;
	    break;
	}
    }
}

bailout()
{
    long port;

    if (export & 4)
	exporttape();
    if (export & 1)
	exportx();
    if (doexportscrap) {
	GetPort(&port);
	if (port != window)
	    SetPort(window);
	writescrap();
	FlushTwice();
    }
    exit(0);
}

static void
updateevent(ep)
EventRecord *ep;
{
    Rect scratch;
    int i, j;
    long save;

    GetPort(&save);
    SetPort(ep->who);
    BeginUpdate(ep->who);
    scratch = regwindow;
    FillRect(&scratch, whitePattern);
    InsetRect(&scratch, -1, -1);
    FrameRect(&scratch);
    i = regwindow.edge.top + finfo.ascent + 1;
    j = finfo.ascent + finfo.descent + finfo.leading;
    TextFont(3);
    TextSize(9);
    MoveTo(regwindow.edge.left + 1, i);
    DrawString("T:");
    i += j;
    MoveTo(regwindow.edge.left + 1, i);
    DrawString("Z:");
    i += j;
    MoveTo(regwindow.edge.left + 1, i);
    DrawString("Y:");
    i += j;
    MoveTo(regwindow.edge.left + 1, i);
    DrawString("X:");
    i += j;
    MoveTo(regwindow.edge.left + 1, i);
    DrawString("Last X:");
    updateregs();
    drawkeys();
    if (oneshot) {
	if (oneshot == STOPEND)
	    hilitekey(35);
	else if (oneshot == RCLPEND)
	    hilitekey(36);
	else
	    hilitekey(38);
    }
    if (tapeon) {
	scratch = taperect;
	FillRect(&scratch, whitePattern); /* Erase tape area. */
	InsetRect(&scratch, -1, -1); /* Frame tape area. */
	FrameRect(&scratch);
	drawlog(1);		/* Re-draw entire visible tape log. */
    }
    DrawControls(window);	/* Re-draw all the controls. */
    EndUpdate(ep->who);
    SetPort(save);
}

static void
mouseevent(ep)
EventRecord *ep;
{
    Rect scratch;
    ControlID p;
    short i, j, found;

    if (ep->who != window) {
	long save;
dobell:
	if (ep->message & F_KEYDOWN) {
	    GetPort(&save);
	    SetPort(ep->who);
	    Bell();
	    SetPort(save);
	}
	return;
    }
    if (ep->message & F_KEYDOWN) { /* Mouse-down event? */
	if (PtInRect(&ep->where.corner.topLeft, &taperect)) {
	    copyfrom(ep->where.edge.top - taperect.edge.top);
	} else if (FindControl(&ep->where.corner.topLeft, ep->who, &p)) {
	    i = TrackControl(p, &ep->where.corner.topLeft, (void (*)()) -1);
	    if (i == inThumb) {
		i = GetCtlValue(scrollbar);
		i = ScaleRS(i, (tapeditems - tapewinsize), tapeditems);
		scrollto(i);	/* Acknowledge thumb drags. */
	    }
	} else {		/* Not a control, button? */
	    for (found=i=0; i<(sizeof(keys)/sizeof(keys[0])); i++) {
		SetRect(&scratch, keys[i].x, keys[i].y, 
			keys[i].x + keys[i].width, keys[i].y + keys[i].height);
		j = keys[i].flag;
		if ((j == -1 && calcmode >= 0) || (j == 1 && calcmode < 0)
		    || (j == 2 && (calcmode < 0 || oneshot != FUNCPEND))
		    || (j == 3 && (calcmode < 0 || oneshot != FUNCPEND))
		    || (j == 1 && calcmode >= 0 && oneshot == FUNCPEND)
		    || (j == 4 && oneshot == FUNCPEND))
		    continue;
		if (PtInRect(&ep->where.corner.topLeft, &scratch)) {
		    if (oneshot) { /* STO or RCL or FUNC in progress? */
			if (oneshot != FUNCPEND) {
			    if (i >= 0 && i <= 9) { /* Digit. */
				if (oneshot == RCLPEND) { /* RCL */
				    enterval(&memory[i]);
				    flashkey(i);
				    hilitekey(36);
				} else if (oneshot == STOPEND) { /* STO */
				    memory[i].dex = x.dex;
				    terminateinput();
				    flashkey(i);
				    hilitekey(35);
				}
			    } else { /* Invalid STO or RCL. */
				if (oneshot == STOPEND)
				    hilitekey(35);
				else if (oneshot == RCLPEND)
				    hilitekey(36);
				goto dobell;
			    }
			} else { /* FUNC was pending. */
			    if (i != 50)
				(*keys[i].func)(keys[i].arg);
			    oneshot = 0;
			    drawkeys();
			    found = 1;
			}
			oneshot = 0;
		    } else	/* No STO or RCL in progress. */
			(*keys[i].func)(keys[i].arg); /* Found btn, exec it. */
		    found = 1;
		    break;
		}
	    }
	    if (!found)
		goto dobell;
	}
    }
}

static void
windowevent(ep)
EventRecord *ep;
{
    struct vidcopy *vcp;
    struct Rect *r;

    switch (ep->message & F_KEYMASK) {
      case windowGoaway:
	if (ep->who == window)
	    bailout();
	else
	    deletecopy(-1, ep->who);
	break;

      case windowMove:
	if (ep->who == window)
	    r = &winrect;
	else {
	    vcp = vchead;
	    while (vcp) {
		if (ep->who == vcp->windowid) {
		    r = &vcp->winrect;
		    break;
		}
		vcp = vcp->next;
	    }
	}
	OffsetRect(r, ep->where.edge.left, ep->where.edge.top);
	MoveWindow(ep->who, r->edge.left, r->edge.top, 0);
	break;

      case windowZoom:
	tapeon = !tapeon;
	SetItem(OPTIONMENU, TAPEMODE, tapeon ? "Hide Tape" : "Show Tape");
	SizeWindow(window, tapeon ? WIDTH2 : WIDTH, 0, 1);
	vcp = vchead;
	while (vcp) {
	    SizeWindow(vcp->windowid, tapeon ? WIDTH2 : WIDTH, 0, 1);
	    vcp = vcp->next;
	}
	break;
    }
}

static void
menuevent(ep)
EventRecord *ep;
{
    short menu, item;
    struct vidcopy *vcp;

    BeginMenu(ep->message);
    menu = (ep->message >> 16) & 0xFFFF;
    item = ep->message & 0xFFFF;
    switch (menu) {
      case OPTIONMENU:		/* Options menu. */
	switch (item) {
	  case FLOAT:		/* Floating point. */
	    engnot = scinot = 0;
	    setone(menu, item, ENGMODE);
	    if (calcmode < 0)
		cvt2dex();
	    calcmode = 0;
	    radix = 10;
	    updateregs();
	    drawkeys();
	    break;

	  case FIX2:		/* Fixed point, 2 place. */
	    setone(menu, item, OCTMODE);
	    if (calcmode < 0)
		cvt2dex();
	    calcmode = 2;
	    radix = 10;
	    updateregs();
	    drawkeys();
	    break;

	  case FIX3:		/* Fixed point, 3 place. */
	    setone(menu, item, OCTMODE);
	    if (calcmode < 0)
		cvt2dex();
	    calcmode = 3;
	    radix = 10;
	    updateregs();
	    drawkeys();
	    break;

	  case FIX4:		/* Fixed point, 4 place. */
	    setone(menu, item, OCTMODE);
	    if (calcmode < 0)
		cvt2dex();
	    calcmode = 4;
	    radix = 10;
	    updateregs();
	    drawkeys();
	    break;

	  case HEXMODE:		/* Fixed point, 0 place, hex. */
	    setone(menu, item, OCTMODE);
	    if (calcmode >= 0)
		cvt2hex();
	    else if (calcmode == -2) /* Changing radix? */
		terminateinput(); /* Mark as pushupable. */
	    calcmode = -1;
	    radix = 16;
	    updateregs();
	    drawkeys();
	    break;

	  case OCTMODE:		/* Fixed point, 0 place, octal. */
	    setone(menu, item, OCTMODE);
	    if (calcmode >= 0)
		cvt2hex();
	    else if (calcmode == -1) /* Changing radix? */
		terminateinput(); /* Mark as pushupable. */
	    calcmode = -2;
	    radix = 8;
	    updateregs();
	    drawkeys();
	    break;

	  case SCIMODE:
	    if (scinot = !scinot)
		engnot = 0;
	    goto engmode;
	    break;

	  case ENGMODE:
	    if (engnot = !engnot)
		scinot = 0;
	  engmode:
	    CheckItem(OPTIONMENU, SCIMODE, scinot);
	    CheckItem(OPTIONMENU, ENGMODE, engnot);
	    if (!calcmode)
		calcmode = 6;
	    updateregs();
	    break;

	  case TAPEMODE:	/* Tape toggle. */
	    tapeon = !tapeon;
	    SetItem(OPTIONMENU, TAPEMODE, tapeon ? "Hide Tape" : "Show Tape");
	    SizeWindow(window, tapeon ? WIDTH2 : WIDTH, 0, 1);
	    vcp = vchead;
	    while (vcp) {
		SizeWindow(vcp->windowid, tapeon ? WIDTH2 : WIDTH, 0, 1);
		vcp = vcp->next;
	    }
	    break;

	  case QUIT:		/* Quit. */
	    bailout();
	    break;
	}
	break;

      case EDITMENU:		/* Edit Menu */
	switch (item) {
	  case CUT:		/* Cut. */
	    cut();
	    break;

	  case COPY:		/* Copy. */
	    copy();
	    break;

	  case PASTE:		/* Paste. */
	    paste();
	    break;
	}
	break;
    }
    EndMenu(ep->message);
}

void
killscrap()
{
    scrapsize = 0;		/* Delete private scrap. */
    doimportscrap = 0;
}

void
scrapevent(ep)
EventRecord *ep;
{
    long port;

    if (ep->who == window) {
	GetPort(&port);
	if (port != ep->who)
	    SetPort(ep->who);
	if (!ep->message) {	/* Someone published new scrap. */
	    doimportscrap = 1;
	    if (mywindowisinfront)
		fetchscrap();
	} else {		/* Someone wants our (?) scrap. */
	    if (doexportscrap && mywindowisinfront)
		writescrap();
	}
	if (port != ep->who)
	    SetPort(port);
    }
}

void
fetchscrap()
{
    long size;

    killscrap();		/* Create private scrap. */
    scrapsize = GetScrap(scrapbuf, 'TEXT', sizeof(scrapbuf));
    if (scrapsize)
	EnableItem(EDITMENU, PASTE);
    else
	DisableItem(EDITMENU, PASTE);
}

void
writescrap()
{
    if (doexportscrap) {
	ZeroScrap();		/* Export private scrap. */
	PutScrap(scrapsize, 'TEXT', scrapbuf);
	PublishScrap(0);
	doexportscrap = 0;
    }
}

void
cut()				/* User-activated CUT command. */
{
    copy();
    terminateinput();		/* Also clear X. */
    x.dex = 0;
    stacklift = 0;
    updatex();
}

void
copy()				/* User-activated COPY command. */
{
    killscrap();
    calctoa(scrapbuf, &x);	/* Create private scrap. */
    scrapsize = StringLen(scrapbuf);
    doexportscrap = 1;		/* Set export-on-deactivate flag. */
    if (scrapsize)
	EnableItem(EDITMENU, PASTE);
    else
	DisableItem(EDITMENU, PASTE);
}

short pasting;

void
paste()				/* User-activated PASTE command. */
{
    EventRecord er;
    short i, minusflag, minusstate;
    char *p, ch;

    minusstate = 0;		/* Assume '-' is a CHS code at first. */
    pasting = 1;
    for (i = scrapsize, p = scrapbuf; i--; ) {
	ch = *p++;
	if (ch == ',') {
	    continue;
	} else if (ch == ' ' || ch == '\n') {
	    minusstate = 0;
	    doenter();
	} else if (ch == '-' && !minusstate) {
	    minusstate = minusflag = 1;
	    continue;
	} else {
	    minusstate = 1;
	    er.message = F_KEYDOWN | ch;
	    keyboardevent(&er);
	    if (minusflag) {
		dochs();
		minusflag = 0;
	    }
	}
    }
    terminateinput();
    pasting = 0;
}

static
setone(menu, item, lastone)	/* Check only 1 of the menu items. */
{
    short i;

    for (i=1; i<=lastone; i++)
	CheckItem(menu, i, i == item);
}

int
matherr(x)
struct exception *x;
{
    long port;

    GetPort(&port);
    if (port != window)
	SetPort(window);
    Bell();
    if (port != window)
	SetPort(port);
}

updateregs()			/* Re-draw all registers. */
{
    Rect scratch;
    char buffer[80];
    short i, j, wid;

    scratch = regwindow;
    scratch.edge.top++;
    TextMode(srcCopy);		/* For overwriting text. */
    i = scratch.edge.top + finfo.ascent;
    j = finfo.ascent + finfo.descent + finfo.leading;
    scratch.edge.bottom = scratch.edge.top + finfo.ascent + finfo.descent;
    scratch.edge.left = regwindow.edge.left
	+ LocalStringWidth(3, normalText, 9, "T:");
    calctoa(buffer, &t);
    wid = LocalStringWidth(1, normalText, 12, buffer);
    scratch.edge.right = regwindow.edge.right - wid - 1;
    FillRect(&scratch, whitePattern);
    OffsetRect(&scratch, 0, j);
    MoveTo(regwindow.edge.right - 1 - wid, i);
    TextFont(1);
    TextSize(12);
    DrawString(buffer);

    i += j;
    scratch.edge.left = regwindow.edge.left
	+ LocalStringWidth(3, normalText, 9, "Z:");
    calctoa(buffer, &z);
    wid = LocalStringWidth(1, normalText, 12, buffer);
    scratch.edge.right = regwindow.edge.right - wid - 1;
    FillRect(&scratch, whitePattern);
    OffsetRect(&scratch, 0, j);
    MoveTo(regwindow.edge.right - 1 - wid, i);
    DrawString(buffer);

    i += j;
    scratch.edge.left = regwindow.edge.left
	+ LocalStringWidth(3, normalText, 9, "Y:");
    calctoa(buffer, &y);
    wid = LocalStringWidth(1, normalText, 12, buffer);
    scratch.edge.right = regwindow.edge.right - wid - 1;
    FillRect(&scratch, whitePattern);
    OffsetRect(&scratch, 0, j);
    MoveTo(regwindow.edge.right - 1 - wid, i);
    DrawString(buffer);

    i += j;
    scratch.edge.left = regwindow.edge.left
	+ LocalStringWidth(3, normalText, 9, "X:");
    calctoa(buffer, &x);
    wid = LocalStringWidth(0, normalText, 12, buffer);
    scratch.edge.right = regwindow.edge.right - wid - 1;
    FillRect(&scratch, whitePattern);
    OffsetRect(&scratch, 0, j);
    MoveTo(regwindow.edge.right - 1 - wid, i);
    TextFont(0);
    TextSize(12);
    DrawString(buffer);

    i += j;
    scratch.edge.left = regwindow.edge.left
	+ LocalStringWidth(3, normalText, 9, "Last X:") + 1;
    calctoa(buffer, &lastx);
    wid = LocalStringWidth(1, normalText, 12, buffer);
    scratch.edge.right = regwindow.edge.right - wid - 1;
    FillRect(&scratch, whitePattern);
    MoveTo(regwindow.edge.right - 1 - wid, i);
    TextFont(1);
    DrawString(buffer);
    TextMode(srcOr);		/* Restore text mode. */
}

updatex()			/* Re-draw X register. */
{
    Rect scratch;
    char buffer[80];
    short i, j, wid;

    scratch = regwindow;
    i = regwindow.edge.top + finfo.ascent + 1;
    j = finfo.ascent + finfo.descent + finfo.leading;
    scratch.edge.bottom = scratch.edge.top + finfo.ascent + finfo.descent;
    OffsetRect(&scratch, 0, 3 * j);
    i += 3 * j;
    scratch.edge.left = regwindow.edge.left
	+ LocalStringWidth(3, normalText, 9, "X:") + 1;
    calctoa(buffer, &x);
    wid = LocalStringWidth(0, normalText, 12, buffer);
    scratch.edge.right = regwindow.edge.right - wid - 1;
    FillRect(&scratch, whitePattern);
    MoveTo(regwindow.edge.right - 1 - wid, i);
    TextFont(0);
    TextSize(12);
    TextMode(srcCopy);
    DrawString(buffer);
    TextMode(srcOr);
}

static
exportx()			/* Export X register. */
{
    char buffer[80];

    calctoa(buffer, &x);
    printf("%s\n", buffer);
}

static
exporttape()			/* Export tape log. */
{
    struct taperec *p;

    p = tapeheader.next;
    while (p) {
	printf("%*s\n", p->textsize, p->text);
	p = p->next;
    }
}

static
logxkey(chr)
char chr;
{
    logx(chr);
    flashkey(29);
}


logx(chr)			/* Log the X register to `tape'. */
char chr;
{
    short i;
    char buffer[80], *p;

    lasttaped->next = (struct taperec *) malloc(sizeof(struct taperec));
    lasttaped = lasttaped->next; /* Add a tape record. */
    calctoa(buffer, &x);
    p = malloc(i = (StringLen(buffer) + 2));
    sprintf(p, "%s %c", buffer, chr);
    tapeditems++;
    lasttaped->text = p;
    lasttaped->textsize = i;
    lasttaped->next = 0;

    if (tapeditems > tapewinsize) { /* Handle display of new object. */
	i = tapeditems - tapewinsize - 1;
	if (i != tapeview)	/* If we weren't displaying end of list, */
	    scrollto(i);	/* scroll so we are. */
	tapeview++;
	BackPat(whitePattern);	/* Now scroll up one to make room for new. */
	ScrollRect(&taperect, 0, -finfo3.leading, 0);
	BackPat((Pattern *) 45);
    }
    drawlog(0);			/* Display new object. */
    setscrollbar();
}

static
copyfrom(y)			/* Re-enter a number from the tape log. */
int y;
{
    Rect scratch;
    struct taperec *p;
    char *cp;
    short i;

    scratch = taperect;
    y -= y % finfo3.leading;
    scratch.edge.top += y;
    scratch.edge.bottom = scratch.edge.top + finfo3.leading;
    i = tapeview + (y / finfo3.leading); /* Which tape item was clicked? */
    if (i >= tapeditems)
	return;
    InvertRect(&scratch);
    FlushWindow();
    p = tapeheader.next;
    while (p) {
	if (!i) {		/* Find selected tape item. */
	    if (stacklift || entering)
		liftstack();
	    updateregs();
	    i = strchr(p->text, ' ') - p->text;
	    memcpy(entrybuf, p->text, i);
	    entryptr = entrybuf + i;
	    reparsex();
	    termnupdatex();
	    break;
	}
	i--;
	p = p->next;
    }
    nap(60);
    InvertRect(&scratch);
    logx('#');
}

static
drawlog(flag)			/* (Re)draw tape log. */
int flag;			/* TRUE to draw whole log. */
{
    Rect scratch;
    struct taperec *p;
    short i, j;

    scratch = taperect;
    ClipRect(&scratch);
    scratch.edge.top++;
    i = tapeview;
    j = 1;
    TextFont(3);
    TextSize(9);
    p = tapeheader.next;
    while (p) {
	if (i)			/* Skip over undisplayed stuff. */
	    i--;
	else {
	    if (j > tapewinsize)
		break;
	    if (flag || (j + tapeview) == tapeditems) {
		DrawJustText(p->text, p->textsize, &scratch,
			     rightJustify, leftJustify);
	    }
	    OffsetRect(&scratch, 0, finfo3.leading);
	    j++;
	}
	p = p->next;
    }
    SetRect(&scratch, -32000, -32000, 32000, 32000);
    ClipRect(&scratch);
}

static
setscrollbar()			/* Re-draw scrollbar limits. */
{
    static barstate;		/* So we don't keep doing hilite of 0 which */
				/* messes up the arrows when mouse scrolling */
    if (tapeditems <= tapewinsize) {
	HiliteControl(scrollbar, 255);
	barstate = 1;
    } else {
	if (barstate) {
	    HiliteControl(scrollbar, 0);
	    barstate = 0;
	}
	Set4CtlVals(scrollbar, 0, tapeditems,
		    ScaleRS(tapeview, tapeditems, (tapeditems - tapewinsize)),
		    tapewinsize);
    }
}

/*
** This routine is called repeatedly if the operator presses in the controls
** of scroll bars.  It is optional (installed via SetCtlAction).  Without it
** the scroll bars only accept a single click in any part of them.  Generally
** users will expect scrolling to be continuous, so you want one of these.
**
** This routine is _not_ called when dragging the thumb.
*/

void
MyAction(theControl, partCode)
ControlID theControl;
short partCode;
{
    short i, j;

    j = (tapeditems - tapewinsize); /* Max limit for setting tapeview. */
    if (partCode == inUpButton) {
	if (tapeview) {
	    scrollto(tapeview - 1);
	}
    } else if (partCode == inDownButton) {
	if (tapeview < j) {
	    scrollto(tapeview + 1);
	}
    } else if (partCode == inPageUp) {
	i = (tapeview < tapewinsize) ? tapeview : tapewinsize;
	scrollto(tapeview - i);
    } else if (partCode == inPageDown) {
	i = (tapeview < (j - tapewinsize)) ? tapewinsize : (j - tapeview);
	scrollto(tapeview + i);
    }
}

scrollto(item)			/* Scroll log to item #. */
int item;
{
    short sign, i;

    if (item == tapeview)
	return;
    BackPat(whitePattern);
    i = tapeview - item;
    tapeview = item;
    ScrollRect(&taperect, 0, finfo3.leading * i, 0);
    BackPat((Pattern *) 45);
    setscrollbar();
    drawlog(1);
}

static
calctoa(buffer, var)		/* Convert number to ASCII form. */
char *buffer;
CalcType *var;
{
    char buf[80], *p, *q;
    short i, temp, bump;

    if (calcmode < 0)
	sprintf(buffer, (calcmode == -1) ? PRTYPE2 : PRTYPE4, *var);
    else {
	if (engnot || scinot)
	    sprintf(buf, PRTYPE5, calcmode, *var);
	else if (calcmode > 0)
	    sprintf(buf, PRTYPE3, calcmode, *var);
	else
	    sprintf(buf, PRTYPE, *var);
	
	if (buf[0] == '(') {	/* An (INF) message? */
	    strcpy(buffer, buf); /* Pass it out unchanged. */
	    return;
	}
	if (engnot) {		/* Engineering notation? */
	    if (p = strchr(buf, 'E')) {
		i = atoi(&p[2]);
		if (i % 3) {	/* Exponent multiple of 3? */
		    if (p[1] == '-') { /* No, modify it, and */
			bump = 1;
			sprintf(&p[2], "%02d", i + 3 - (i % 3));
		    } else {
			bump = -1;
			sprintf(&p[2], "%02d", i - (i % 3));
		    }
		    if (p = strchr(buf, '.')) {
			while (i % 3) { /* shift dec point to compensate. */
			    temp = *p;
			    *p = p[1];
			    p[1] = temp;
			    p++;
			    i += bump;
			}
		    }
		}
	    }
	}
	for (i=0, p=buf; *p; p++) { /* Find out how many places before */
	    if (*p == '.' || *p == 'E') /* the decimal point. */
		break;
	    if (*p == '-')
		continue;
	    i++;
	}
	p = buf;
	q = buffer;
	while (i > 3) {
	    *q++ = *p;
	    if (*p++ == '-')
		continue;
	    if (!(--i % 3))	/* Insert commas every three chars. */
		*q++ = ',';
	}
	while (*q++ = *p++)
	    ;
    }
}

drawkeys()
{
    Rect dest, src;
    short i, j;
    BitMap keytops;

    keytops.rowBytes = keytop[0].rowBytes;
    SetRect(&keytops.bounds, 0, 0, keytop[0].width, keytop[0].height);
    keytops.baseAddr = (char *) &keytop[1];

    for (i=0; i<(sizeof(keys)/sizeof(keys[0])); i++) {
	SetRect(&dest, keys[i].x, keys[i].y, 
		keys[i].x + keys[i].width, keys[i].y + keys[i].height);
	SetRect(&src, keys[i].bx, keys[i].by, 
		keys[i].bx + keys[i].width, keys[i].by + keys[i].height);
	j = keys[i].flag;
	if (!j || (j == 1 && calcmode >= 0) || (j == -1 && calcmode < 0)
	    || (j == 2 && calcmode >= 0 && oneshot == FUNCPEND)
	    || (j == 3 && calcmode >= 0 && oneshot == FUNCPEND)
	    || (j == 4 && oneshot != FUNCPEND))
	    CopyBits(&keytops, 0, &src, &dest, srcCopy, 0);
    }
    dest.edge.top = keys[12].y - 3; /* Frame the keypad area. */
    dest.edge.left = keys[12].x - 3;
    dest.edge.bottom = keys[23].y + keys[23].height + 3;
    dest.edge.right = keys[23].x + keys[23].width + 3;
    FrameRect(&dest);
}

flashkey(keynum)
int keynum;
{
    Rect scratch;
    short i;

    if (keyflash) {
	SetRect(&scratch, keys[keynum].x, keys[keynum].y, 
		keys[keynum].x + keys[keynum].width,
		keys[keynum].y + keys[keynum].height);
	InvertRect(&scratch);
	if (!CheckNextEvent(anyWindow, 1<<keyEvt) && !pasting) {
				/* No typeahead, then */
	    FlushWindow();	/* do a visible flash.  Otherwise skip it. */
	    nap(60);		/* Nap to make sure flash is seen. */
	}
	InvertRect(&scratch);
    }
}

hilitekey(keynum)
int keynum;
{
    Rect scratch;
    short i;

    SetRect(&scratch, keys[keynum].x, keys[keynum].y, 
	    keys[keynum].x + keys[keynum].width,
	    keys[keynum].y + keys[keynum].height);
    InvertRect(&scratch);
}

usage(name)
char *name;
{
    fprintf(stderr, "%s Version 2.6\n", name);
    fprintf(stderr, "Usage:	%s [-xpos] [-ypos] [-idigits] [-e[=][t]] [-p] [-mX] [-f]\n", name);
    fprintf(stderr, "   	The command \"%s -x20 -y30 -i44 -e=t\" would place the window\n", name);
    fprintf(stderr, "	at screen coordinates (20, 30), would preload 44 into the X\n");
    fprintf(stderr, "	register, and export X upon an = request, and upon exit, and\n");
    fprintf(stderr, "	would export the tape log upon exit.  The -p option shuts off\n");
    fprintf(stderr, "	the 'programmer' features (hex and octal modes).  The -mX\n");
    fprintf(stderr, "	sets the initial mode to one of: F-floating, <2|3|4>-fixed\n");
    fprintf(stderr, "	place, H-hex, S-Scientific, E-Engineering, O-octal, T-tape.\n");
    fprintf(stderr, "	The -f option turns off key flashing.\n");
}
