/*
** Calculator's definitions.
*/

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <signal.h>
#include <string.h>
#include <isc/window.h>
#include <isc/keycodes.h>
#include <isc/controls.h>

#define WIDTH (200)		/* Window definitions. */
#define WIDTH2 (340)		/* For tape logging area; */
#define HEIGHT (300)

void updateevent();
void menuevent();
void mouseevent();
void keyboardevent();
void activateevent();
void windowevent();
void scrapevent();
void MyAction(ControlID theControl, short partCode);
void fetchscrap(), writescrap(), cut(), copy(), paste();

/* Menu constants. */
#define OPTIONMENU 1
#define FLOAT 1
#define FIX2 2
#define FIX3 3
#define FIX4 4
#define HEXMODE 6
#define OCTMODE 7
#define SCIMODE 8
#define ENGMODE 9
#define TAPEMODE 10
#define QUIT 12
#define EDITMENU 2
#define CUT 1
#define COPY 2
#define PASTE 3

typedef union {double dex; long hex} CalcType; /* Calculator's data type. */
#define PRTYPE "%.18G"		/* Printf string for type of above. */
#define PRTYPE2 "0x%08lX"	/* Hex mode printf string. */
#define PRTYPE3 "%.*f"		/* Fixed point printf string. */
#define PRTYPE4 "@%011lo"	/* Octal mode printf string. */
#define PRTYPE5 "%.*E"		/* Eng mode printf string. */

struct taperec {		/* Tape record. */
    struct taperec *next;
    char *text;
    short textsize;
} tapeheader, *lasttaped;
short tapeditems;		/* Count of taped items. */
short tapeview;			/* Starting item to display in tape. */

CalcType x, y, z, t, lastx;	/* Calculator's data registers. */
CalcType memory[10];
short stacklift;
short entering;
short radix;			/* I/O radix. */
short calcmode;			/* 0 = fixed, # = places, -1=hex, -2=oct. */
short engnot;			/* TRUE for Eng. notation. */
short scinot;			/* TRUE for Sci. notation. */
short entrysign;
short entryeex;
short entryeexsign;
short export;			/* TRUE to export result to stdout. */
#define STOPEND 1
#define RCLPEND 2
#define FUNCPEND 3
short oneshot;			/* 0 == normal, 1 = STO, 2 = RCL pending. */
short tapeon;			/* TRUE for adding-machine tape logging on. */
short progfuncts;		/* TRUE for hex & octal options. */
short keyflash;			/* TRUE for keyboard key flash. */
char entrybuf[80];		/* Number entry buffer. */
char *entryptr;

char scrapbuf[80];		/* Cut/Paste buffer. */
int scrapsize;			/* Amount in buffer. */

Rect winrect;			/* Display data structures. */
Rect regwindow;
Rect taperect;
short tapewinsize;		/* Number of items that shows on tape. */
int window;
FontInfo finfo, finfo1, finfo3;
ControlID scrollbar;

/*
** Key function forward declarations.
*/

int dodigit(), dochs(), dodiv(), domul(), dosub(), doadd(), doenter();
int dolastx(), donot(), doand(), door(), doxor(), logx(), logxkey();
int doshiftl(), doshiftr(), dorollup(), dorolldown(), dobs(), dosto();
int dorcl(), doexg(), donothing(), dosqrt(), dopow(), dorecip(), dopercent();
int dodeltapercent(), doln(), doexp(), dosin(), docos(), dotan(), doprefix();
int dolog10(), doexp10(), doasin(), doacos(), doatan(), dosquare(), dofact();
int doint(), dofrac(), dopi(), dodegtorad(), dointocm(), dodegftodegc();
int doradtodeg(), docmtoin(), dodegctodegf(), dogaltolit(), dolbtokg();
int domitokm(), dolittogal(), dokgtolb(), dokmtomi(), dooztog(), dogtooz();
int doydtom(), domtoyd();

/* For positioning the keycaps on the screen. */

#define RSPC 22			/* Row-to-row spacing. */
#define KR1 100			/* Row 1 starts here in window. */
#define KR2 KR1+RSPC
#define KR3 KR2+RSPC
#define KR4 KR3+RSPC
#define KR5 KR4+RSPC
#define KR6 KR5+RSPC
#define KR7 KR6+RSPC
#define KR8 KR7+RSPC

#define CSPC 34			/* Column-to-column spacing. */
#define KC1 16			/* Column 1 starts here in window. */
#define KC2 KC1+CSPC
#define KC3 KC2+CSPC
#define KC4 KC3+CSPC
#define KC5 KC4+CSPC

#define KWID 24
#define KHEI 16
#define KHEI2 KHEI+RSPC

#define BR1 0			/* Row 1 starts here in offscreen bitmap. */
#define BR2 BR1+KHEI
#define BR3 BR2+KHEI
#define BR4 BR3+KHEI
#define BR5 BR4+KHEI
#define BR6 BR5+KHEI
#define BR7 BR6+KHEI
#define BR8 BR7+KHEI

#define BC1 0			/* Column 1 starts here in offscreen bitmap. */
#define BC2 BC1+KWID
#define BC3 BC2+KWID
#define BC4 BC3+KWID
#define BC5 BC4+KWID
#define BC6 BC5+KWID
#define BC7 BC6+KWID
#define BC8 BC7+KWID
#define BC9 BC8+KWID
#define BC10 BC9+KWID
#define BC11 BC10+KWID
#define BC12 BC11+KWID

extern struct BITOBJ {		/* Header for included Data Dictionary */
    short version;		/* bitmap object that has the keytops */
    short rowBytes;		/* in it.  This must be an uncompressed */
    short width;		/* monochrome image for the code to work. */
    short height;
    short hres;
    short vres;
    short flags;
    short compress;
} keytop[2];

