/* Basic math routines for the calculator. */

#include "calc.h"

termnupdate()
{
    terminateinput();
    updateregs();
}

termnupdatex()
{
    terminateinput();
    updatex();
}

doadd()				/* Addition. */
{
    logx('+');
    lastx.dex = x.dex;
    if (calcmode < 0)
	x.hex = y.hex + x.hex;
    else
	x.dex = y.dex + x.dex;
    dropstack();
    updateregs();
    flashkey(16);
}


dosub()				/* Subtraction. */
{
    logx('-');
    lastx.dex = x.dex;
    if (calcmode < 0)
	x.hex = y.hex - x.hex;
    else
	x.dex = y.dex - x.dex;
    dropstack();
    updateregs();
    flashkey(15);
}

domul()				/* Multiplication. */
{
    logx('*');
    lastx.dex = x.dex;
    if (calcmode < 0)
	x.hex = y.hex * x.hex;
    else
	x.dex = y.dex * x.dex;
    dropstack();
    updateregs();
    flashkey(14);
}

dodiv()				/* Division. */
{
    logx('/');
    if (x.dex == 0.0)		/* Works for Hex, too! */
	matherr(0);
    else {
	lastx.dex = x.dex;
	if (calcmode < 0)
	    x.hex = y.hex / x.hex;
	else
	    x.dex = y.dex / x.dex;
	dropstack();
	updateregs();
    }
    flashkey(13);
}

dopercent()			/* Percentage/modulo. */
{
    logx('%');
    lastx.dex = x.dex;
    if (calcmode < 0) {
	if (x.hex) {
	    x.hex = y.hex % x.hex;
	    dropstack();
	} else {
	    matherr(0);
	}
    } else {
	flashkey(45);
	x.dex = y.dex * x.dex / 100.0;
	terminateinput();
    }
    updateregs();
}

dodeltapercent()		/* Delta Percentage. */
{
    flashkey(44);
    if (y.dex == 0.0)
	matherr(0);
    else {
	logx('\306');
	lastx.dex = x.dex;
	x.dex = (x.dex - y.dex) / y.dex * 100.0;
	termnupdate();
    }
}

doand()				/* Logical AND. */
{
    if (calcmode >= 0) {	/* Not in decimal modes. */
	matherr(0);
    } else {
	logx('&');
	lastx.hex = x.hex;
	x.hex = y.hex & x.hex;
	dropstack();
	updateregs();
    }
    flashkey(26);
}

door()				/* Logical OR. */
{
    if (calcmode >= 0) {	/* Not in decimal modes. */
	matherr(0);
    } else {
	logx('|');
	lastx.hex = x.hex;
	x.hex = y.hex | x.hex;
	dropstack();
	updateregs();
    }
    flashkey(27);
}

doxor()				/* Logical XOR. */
{
    if (calcmode >= 0) {	/* Not in decimal modes. */
	matherr(0);
    } else {
	logx('^');
	lastx.hex = x.hex;
	x.hex = y.hex ^ x.hex;
	dropstack();
	updateregs();
    }
    flashkey(28);
}

donot()				/* Logical NOT. */
{
    if (calcmode >= 0) {	/* Not in decimal modes. */
	matherr(0);
    } else {
	logx('~');
	lastx.hex = x.hex;
	x.hex = ~x.hex;
	termnupdatex();
    }
    flashkey(25);
}

doshiftl()			/* Logical shift. */
{
    if (calcmode >= 0) {	/* Not in decimal modes. */
	matherr(0);
    } else {
	lastx.hex = x.hex;
	logx('<');
	x.hex = x.hex << 1;
	termnupdatex();
    }
    flashkey(30);
}

doshiftr()			/* Logical shift right. */
{
    if (calcmode >= 0) {	/* Not in decimal modes. */
	matherr(0);
    } else {
	lastx.hex = x.hex;
	logx('>');
	x.hex = x.hex >> 1;
	termnupdatex();
    }
    flashkey(31);
}


dobs()				/* Handle backspace. */
{
    if (entering) {
	if (entryptr > entrybuf) {
	    entryptr--;
	    reparsex();
	} else {
	    matherr(0);
	}
	if (entryptr == entrybuf) {
	    entering = 0;
	    stacklift = 0;
	}
    } else {
	x.dex = 0;
	stacklift = 0;
    }
    updatex();
    flashkey(34);
}

doenter()			/* Handle ENTER key. */
{
    logx('#');
    liftstack();
    updateregs();
    flashkey(23);
}


dochs()				/* Handle Change Sign key. */
{
    if (entering) {
	if (entryeex)
	    entryeexsign = !entryeexsign;
	else
	    entrysign = !entrysign;
	reparsex();
    } else if (calcmode >= 0)
	x.dex = -x.dex;
    else
	x.hex = -x.hex;
    updatex();
    flashkey(12);
}

dorollup()			/* Handle Roll-up key. */
{
    CalcType temp;

    temp.dex = t.dex;
    t.dex = z.dex;
    z.dex = y.dex;
    y.dex = x.dex;
    x.dex = temp.dex;
    termnupdate();
    flashkey(32);
}


dorolldown()			/* Handle Roll-down key. */
{
    CalcType temp;

    temp.dex = x.dex;
    x.dex = y.dex;
    y.dex = z.dex;
    z.dex = t.dex;
    t.dex = temp.dex;
    termnupdate();
    flashkey(33);
}

doexg()				/* Handle X<->Y key. */
{
    CalcType temp;

    temp.dex = y.dex;
    y.dex = x.dex;
    x.dex = temp.dex;
    termnupdate();
    flashkey(37);
}


dosto()				/* Do Store operation. */
{
    oneshot = STOPEND;
    hilitekey(35);
}


dorcl()				/* Do Recall operation. */
{
    oneshot = RCLPEND;
    hilitekey(36);
}

doprefix()			/* Do function-prefix operation. */
{
    oneshot = FUNCPEND;
    drawkeys();
    hilitekey(38);
}

dolastx()			/* Do Last X operation. */
{
    enterval(&lastx);
    flashkey(24);
}

donothing(arg)			/* Do nothing (yet). */
short arg;
{
    flashkey(arg);
    matherr(0);
}

dosqrt()			/* Handle square root key. */
{
    lastx.dex = x.dex;
    x.dex = sqrt(x.dex);
    termnupdate();
    flashkey(47);
}

dopow()				/* Handle power key. */
{
    lastx.dex = x.dex;
    x.dex = pow(y.dex, x.dex);
    dropstack();
    updateregs();
    flashkey(48);
}

dorecip()			/* Handle reciprocal key. */
{
    flashkey(49);
    if (x.dex == 0.0)		/* Works for Hex, too! */
	matherr(0);
    else {
	lastx.dex = x.dex;
	x.dex = 1.0 / x.dex;
	termnupdate();
    }
}

doln()				/* Handle ln key. */
{
    flashkey(43);
    if (x.dex <= 0.0)
	matherr(0);
    else {
	lastx.dex = x.dex;
	x.dex = log(x.dex);
	termnupdate();
    }
}

doexp()				/* Handle exp key. */
{
    flashkey(42);
    lastx.dex = x.dex;
    x.dex = exp(x.dex);
    termnupdate();
}

dolog10()			/* Handle log10 key. */
{
    flashkey(55);
    if (x.dex <= 0.0)
	matherr(0);
    else {
	lastx.dex = x.dex;
	x.dex = log10(x.dex);
	termnupdate();
    }
}

doexp10()			/* Handle exp10 key. */
{
    flashkey(54);
    lastx.dex = x.dex;
    x.dex = pow(10.0, x.dex);
    termnupdate();
}

dosin()				/* Handle sin key. */
{
    flashkey(39);
    lastx.dex = x.dex;
    x.dex = sin(x.dex);
    termnupdate();
}

docos()				/* Handle cos key. */
{
    flashkey(40);
    lastx.dex = x.dex;
    x.dex = cos(x.dex);
    termnupdate();
}

dotan()				/* Handle tan key. */
{
    flashkey(41);
    lastx.dex = x.dex;
    x.dex = tan(x.dex);
    termnupdate();
}

doasin()			/* Handle arcsin key. */
{
    flashkey(51);
    lastx.dex = x.dex;
    x.dex = asin(x.dex);
    termnupdate();
}

doacos()			/* Handle arccos key. */
{
    flashkey(52);
    lastx.dex = x.dex;
    x.dex = acos(x.dex);
    termnupdate();
}

doatan()			/* Handle arctan key. */
{
    flashkey(53);
    lastx.dex = x.dex;
    x.dex = atan(x.dex);
    termnupdate();
}

dofact()			/* Handle factorial/gamma key. */
{
    flashkey(58);
    lastx.dex = x.dex;
    x.dex += 1.0;		/* Gamma/factorial (ala HP calculators). */
    x.dex = gamma(x.dex);
    x.dex = signgam * exp(x.dex);
    termnupdate();
}

dosquare()			/* Handle squaring key. */
{
    flashkey(59);
    lastx.dex = x.dex;
    x.dex = x.dex * x.dex;
    termnupdate();
}

doint()				/* Handle integer-part key. */
{
    flashkey(60);
    lastx.dex = x.dex;
    modf(lastx.dex, &x.dex);
    termnupdate();
}

dofrac()			/* Handle fractional-part key. */
{
    double scratch;

    flashkey(61);
    lastx.dex = x.dex;
    x.dex = modf(lastx.dex, &scratch);
    termnupdate();
}

dopi()				/* Handle Pi key. */
{
    CalcType temp;

    temp.dex = 4.0 * atan(1.0);
    flashkey(57);
    enterval(&temp);
    termnupdate();
}

dodegtorad()			/* Handle degrees to radians key. */
{
    flashkey(62);
    lastx.dex = x.dex;
    x.dex = x.dex * atan(1.0) / 45.0;
    termnupdate();
}

doradtodeg()			/* Handle radians to degrees key. */
{
    flashkey(65);
    lastx.dex = x.dex;
    x.dex = x.dex * 45.0 / atan(1.0);
    termnupdate();
}

dointocm()			/* Inches to centimeters. */
{
    flashkey(63);
    lastx.dex = x.dex;
    x.dex = x.dex * 2.54;
    termnupdate();
}

dodegftodegc()			/* Degrees Fahrenheit to degrees Celcius. */
{
    flashkey(64);
    lastx.dex = x.dex;
    x.dex = (x.dex - 32.0) * 5.0 / 9.0;
    termnupdate();
}

docmtoin()			/* Centimeters to inches. */
{
    flashkey(66);
    lastx.dex = x.dex;
    x.dex = x.dex / 2.54;
    termnupdate();
}

dodegctodegf()			/* Degrees Celcius to degrees Fahrenheit. */
{
    flashkey(67);
    lastx.dex = x.dex;
    x.dex = x.dex * 9.0 / 5.0 + 32.0;
    termnupdate();
}

dogaltolit()			/* Gallons to Litres. */
{
    flashkey(68);
    lastx.dex = x.dex;
    x.dex = x.dex * 3.78541;
    termnupdate();
}

dolbtokg()			/* Pounds to Kilograms. */
{
    flashkey(69);
    lastx.dex = x.dex;
    x.dex = x.dex * 0.45359237;
    termnupdate();
}

domitokm()			/* Miles to kilometers. */
{
    flashkey(73);
    lastx.dex = x.dex;
    x.dex = x.dex * 1.6;
    termnupdate();
}

dolittogal()			/* Litres to gallons. */
{
    flashkey(68);
    lastx.dex = x.dex;
    x.dex = x.dex / 3.78541;
    termnupdate();
}

dokgtolb()			/* Kilograms to pounds. */
{
    flashkey(72);
    lastx.dex = x.dex;
    x.dex = x.dex / 0.45359237;
    termnupdate();
}

dokmtomi()			/* Kilometers to miles. */
{
    flashkey(73);
    lastx.dex = x.dex;
    x.dex = x.dex / 1.6;
    termnupdate();
}

dooztog()			/* Ounces to grams. */
{
    flashkey(74);
    lastx.dex = x.dex;
    x.dex = x.dex * 28.3495;
    termnupdate();
}

dogtooz()			/* Grams to ounces. */
{
    flashkey(75);
    lastx.dex = x.dex;
    x.dex = x.dex / 28.3495;
    termnupdate();
}

doydtom()			/* Yards to meters. */
{
    flashkey(76);
    lastx.dex = x.dex;
    x.dex = x.dex * 0.9144;
    termnupdate();
}

domtoyd()			/* Meters to yards. */
{
    flashkey(77);
    lastx.dex = x.dex;
    x.dex = x.dex / 0.9144;
    termnupdate();
}

dodigit(key)			/* Accumulate a digit (or decimal point). */
short key;
{
    short flag;

    if (key & 0x80) {		/* Used for double-zero. */
	flag = 1;
	key &= 0x7F;
	flashkey(10);
    } else {
	flag = 0;
	if (key != '.' && key != 'E') {
	    if (calcmode < 0 || key <= '9')
		flashkey(key - '0');
	} else if (key == 'E') {
	    if (calcmode >= 0)
		flashkey(46);	/* EEX key. */
	    else
		flashkey(key - '0');
	} else			/* Is a decimal point. */
	    flashkey(11);
    }

    if (key == '.' && calcmode < 0) { /* No floating point in Hex/Octal. */
	matherr(0);
	return;
    }
    if ((key >= '8' && calcmode == -2) || /* For octal. */
	(key >= 'A' && calcmode != -1)) { /* No Hex in float point or octal. */
	if (calcmode < 0 || key != 'E') {
	    matherr(0);
	    return;
	}
    }
    if (stacklift) {		/* Lift the stack if this needed. */
	liftstack();
	updateregs();
    }
    if (!entering) {		/* Also clear X if this is the first char. */
	entryptr = entrybuf;
	entrysign = entryeex = entryeexsign = 0;
    }

    *entryptr++ = key;		/* Accumulate the character. */
    if (flag)
	*entryptr++ = key;	/* Accumulate the character again. */
    entering = 1;
    reparsex();
    updatex();
}

reparsex()			/* Re-calculate X from the input buffer. */
{
    CalcType temp, temp2;
    int pwroften;
    char *p;

    switch (calcmode) {
      case 0:			/* Floating-point decimal parsing. */
      case 1:
      case 2:
      case 3:
      case 4:
      case 5:
      case 6:
	temp.dex = 0;
	temp2.dex = 0.1;
	entryeex = 0;
	for (p=entrybuf; p<entryptr; p++) {
	    if (*p == '-') {	/* For parsing from command line, since '-' */
		entrysign = !entrysign;	/* won't show up interactively. */
		continue;
	    }
	    if (*p == 'x' || *p == '@')
		continue;	/* Skip radix indicators. */
	    if (*p == '.' || *p == 'E' || *p == 'e')
		break;
	    temp.dex *= radix;
	    temp.dex += *p - '0';
	}
	if (*p == '.') {
	    for (p++; p<entryptr; p++) {
		if (*p == '.')
		    continue;
		if (*p == 'E' || *p == 'e')
		    break;
		temp.dex += (*p - '0') * temp2.dex;
		temp2.dex /= 10;
	    }
	}
	if (*p == 'E' || *p == 'e') { /* Handle scientific notation. */
	    entryeex = 1;
	    pwroften = 0;
	    for (; p<entryptr; p++) {
		if (*p == '-') { /* For parsing from command line, since '-' */
		    entryeexsign = !entryeexsign;
		    continue;	/* won't show up interactively. */
		}
		if (*p == '.' || *p == 'E' || *p == 'e' || *p == '+')
		    continue;
		pwroften *= 10;
		pwroften += *p - '0';
	    }
	    if (pwroften && temp.dex == 0.0)
		temp.dex = 1.0;
	    if (entryeexsign)
		pwroften = -pwroften;
	    temp.dex *= pow(10.0, (double) pwroften);
	}
	if (entrysign)
	    x.dex = -temp.dex;
	else
	    x.dex = temp.dex;
	break;

      case -2:			/* Octal parsing. */
      case -1:			/* Hexadecimal parsing. */
	temp.hex = 0;
	for (p=entrybuf; p<entryptr; p++) {
	    if (*p == 'x' || *p == '@')
		continue;	/* Skip radix indicators. */
	    temp.hex *= radix;
	    temp.hex += *p - '0';
	    if (*p > '9')
		temp.hex -= 'A' - '9' - 1;
	}
	if (entrysign)
	    x.hex = -temp.hex;
	else
	    x.hex = temp.hex;
	break;
    }
}

enterval(val)			/* Set X to the given value. */
CalcType *val;
{
    if (stacklift || entering)
	liftstack();
    updateregs();
    x = *val;
    termnupdatex();
}

liftstack()			/* Push up the stack one. */
{
    t.dex = z.dex;
    z.dex = y.dex;
    y.dex = x.dex;
    stacklift = 0;		/* Mark as non-pushupable. */
    entering = entrysign = 0;	/* Re-start entry. */
}

dropstack()			/* Drop the stack down one (after math). */
{
    y.dex = z.dex;
    z.dex = t.dex;
    terminateinput();
}

terminateinput()		/* Terminate input process. */
{
    stacklift = 1;
    entering = entrysign = 0;
    entryptr = entrybuf;
}

cvt2dex()			/* Convert Hex registers to decimal. */
{
    short i;

    t.dex = t.hex;
    z.dex = z.hex;
    y.dex = y.hex;
    x.dex = x.hex;
    lastx.dex = lastx.hex;
    for (i=0; i<sizeof(memory)/sizeof(memory[0]); i++)
	memory[i].dex = memory[i].hex;
    terminateinput();		/* Mark as pushupable. */
}

static void			/* Stdlib screws up convert to $80000000! */
bugfix(convert)
CalcType *convert;
{
    if (convert->dex == -2147483648.0)
	convert->hex = 0x80000000;
    else
	convert->hex = convert->dex;
}

cvt2hex()			/* Convert decimal registers to Hex. */
{
    short i;

    bugfix(&t);
    bugfix(&z);
    bugfix(&y);
    bugfix(&x);
    bugfix(&lastx);
    for (i=0; i<sizeof(memory)/sizeof(memory[0]); i++)
	bugfix(&memory[i]);
    terminateinput();		/* Mark as pushupable. */
}

