/* 	Program to create character generator PROMS.
	Also used to make Document Printer III fonts.

		Modifier - Doug Hockin

		3.2	11/7/91
		  The mouse can be used to toggle bits, to select
		  a character from the chart and to copy one character
		  in the chart to another (just drag the char).

		  Also exchanged the left/right and up/down shift keys so
		  the key that moves the char left is to the left of the
		  key that moves the char right. Same thing for up/down.
		  They were backwards!

		Author - Jim Cathey

		3.1	4/4/89
		  Ported to window system.  What a grotesque hack this program
		  has ended up!  Don't show it to your kids!

		3.0	3/7/88
		  Ported to DNIX and the WPV clunkdraw graphics library.

		2.3	4/9/87
		  Added Single character read, fixed small bugs.

		2.2	4/8/87
		  Ported to WPIII graphics environment,
		  Aztec C.  A>cc chargen
			    A>as chargen
			    A>ln chargen.o -lg -lc

		2.1	10/11/84
		  Ported to Z-80 BDS C.
		    A>cc chargen -r12 -o -e4000
		    A>clink chargen -e4000

		2.0 	10/10/84
		  Adds CR, BS, Double size & Halfdot shift
		  support, sped up plotting.

		1.0 	4/21/84
		  Original release, CP/M-68K.

		Based on a program by Joe Shannon

*/

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <isc/window.h>
#include <isc/keycodes.h>

#define NCHARS 256		/* Number of characters supported */
#define NWORDS 16		/* Number of words per char */

#define CELLWIDTH 16		/* Number of bits/char row. */
#define CELLHEIGHT NWORDS	/* Number bits/char column. */

#define BLOCKWIDTH 8		/* Size of graphics blocks for fatbits. */
#define BLOCKHEIGHT 8

#define TITLEBARH 4		/* Horizontal position of Title Bar. */
#define TITLEBARV 4		/* Vertical position of Title Bar. */
#define TITLEBARWID 436		/* Width of Title Bar. */
#define TITLEBARHEI 19		/* Height of Title Bar. */

#define STATSWINH 314		/* Horizontal position of Stats display. */
#define STATSWINV 234		/* Vertical position of Stats display. */
#define STATSWINWID 121		/* Width of Stats display. */
#define STATSWINHEI 60		/* Height of Stats display. */

#define	WABASEX	310		/* Base pixel offset of working Matrix */
#define WABASEY	100

#define DCHBASEY 60		/* Base pixel offset for dbl size plot */
#define DCHBASEX 358

#define	CHBASEY	46		/* Base pixel offset for chart of chars */
#define	CHBASEX	8

#define CHLABELX 12		/* Width of Chart label box. */
#define CHLABELY 14		/* Height of Chart label box. */

#define HBASEY	50		/* Base pixel offsets for help page. */
#define HBASEX	462

#define PROMPTH 40		/* Horizontal position of Prompt Box. */
#define PROMPTV 350		/* Vertical position of Prompt Box. */
#define PROMPTWID 300		/* Width of Prompt Box. */
#define PROMPTHEI 19		/* Height of Prompt Box. */

#define PAD2 5

#define WAWIDTH	16*BLOCKWIDTH	/* Working area size in pixels. */
#define WAHEIGHT NWORDS*BLOCKHEIGHT

#define CR 13		/* ASCII's */
#define BS 8
#define REPAINT 12

#define ERROR (-1)
#define FALSE (0)
#define TRUE (-1)

int cellwid, cellheit, markerrow;
char *base1, *base2;		/* Font working storage pointers. */
int modfied;			/* Font modified since last save flag. */
char filnam1[13], filnam2[13], fnames[40];
int mywindow;			/* Active graphics drawing window. */
Rect promptrect, statswindow, chart, fatbits, dfatbits; /* Active 'windows'. */
Rect winrect;
int trapqfd, wmfd;
void doupdate(), do_mouse();
int curchr;
int downchr;
int curx, cury;

main(argc, argv)
int argc;
char *argv[];
{
    char *malloc();
    int err;

    null("Written by Jim Cathey, 4/4/89");
    cellwid = CELLWIDTH;
    cellheit = CELLHEIGHT;
    base1 = malloc(NCHARS*NWORDS); /* Grab arrays for chargen */
    base2 = malloc(NCHARS*NWORDS);
    if (!base1 || !base2) {
	fprintf(stderr, "%s: Out of memory.\n", argv[0]);
	exit(ENOMEM);
    }
    /*
    **	Initialize the graphics toolbox.
    */

    if ((trapqfd = InitGraf()) < 0) {
	fprintf(stderr, "%s: Error %d in InitGraf().\n", 
		argv[0], -trapqfd);
	exit(1);
    }

    /*
    **	Establish a connection to a window manager.
    */

    if ((wmfd = ConnectGraf(0)) < 0) {
	fprintf(stderr, "%s: Error %d in ConnectGraf().\n", 
		argv[0], -wmfd);
	exit(1);
    }

    SetRect(&winrect, 0, 20, 640, 480);
    mywindow = NewWindow(&winrect, "Font Editor",
			       hideWindow, noBordDBox, frontWindow, 
			       noOptions, wmfd);
    OffsetRect(&winrect, 0, -20);
    SetPort(mywindow);
    if (mywindow < 0) {
	fprintf(stderr, "Window Create failed, errno=%d.\n", errno);
	exit(3);
    }
    BackPat(greyPattern);
    ShowWindow(mywindow);
    BeginUpdate(mywindow);
    EndUpdate(mywindow);

    wipemout();			/* Clear all chars for a fresh start. */
    modfied = FALSE;		/* Clean slate. */
    paintup(0);			/* Draw the pretty display. */

    if (argc == 2) {		/* Were we given a filenam base? */
	getnam(argv[1]);
	if (lodfil(0, NCHARS)) {
	    drawdisp(TRUE);	/* Display the char set */
	    status("Files loaded.");
	}
	shostat(0);
	plotmat(0);		/* Display char from memory */
    } else {
	fnames[0] = '\0';
	filnam1[0] = '\0';
	filnam2[0] = '\0';
    }
    ValidRect(&winrect);
    MouseMode(F_KEYDOWN | F_KEYUP);
    Event(mouseEvt, do_mouse);
    Event(updateEvt, doupdate);
    modify();			/* Handle user input events. */
    exit(0);
}

void
do_mouse(ep)
EventRecord *ep;
{
    int upchr;

    SetPort(ep->who);
    if ((ep->message & F_KEYDOWN) == F_KEYDOWN) { /* A mouse down event. */
	if (PtInRect(&ep->where.corner.topLeft, &fatbits)) { /* In fatbits. */
	    togcur(curx, cury);	/* Turn off cursor in current spot. */
	    curx = (ep->where.corner.topLeft.h - fatbits.corner.topLeft.h) /
		BLOCKWIDTH;
	    cury = (ep->where.corner.topLeft.v - fatbits.corner.topLeft.v) /
		BLOCKHEIGHT;
	    compl(curchr, curx, cury); /* Toggle block where they clicked. */
	    togcur(curx, cury);	/* Turn on cursor in new spot. */
	}
	if (PtInRect(&ep->where.corner.topLeft, &chart)) { /* In all chars
							      chart. */
	    downchr = 
		(((ep->where.corner.topLeft.h - chart.corner.topLeft.h) /
		  (CELLWIDTH + 1)) * 0x10) +
		      ((ep->where.corner.topLeft.v - chart.corner.topLeft.v) /
		       (CELLHEIGHT + 1));
	}
    } else {			/* A mouse up event. */
	if (PtInRect(&ep->where.corner.topLeft, &chart)) { /* In all chars
							      chart. */
	    if (downchr != -1) { /* We went down in the chart. */
		upchr =		/* Figure out which char we came up on. */
		    (((ep->where.corner.topLeft.h - chart.corner.topLeft.h) /
		      (CELLWIDTH + 1)) * 0x10) +
		      ((ep->where.corner.topLeft.v - chart.corner.topLeft.v) /
		       (CELLHEIGHT + 1));
		if (upchr != downchr) { /* We went down on one and
					   up on another. */
		    copychr(downchr, upchr); /* Copy the down char to here. */
		}
		curchr = upchr;	/* Set new current char. */
		shostat(curchr); /* Show hex value of selected char. */
		togcur(curx, cury); /* Turn cursor off for a bit. */
		plotmat(curchr); /* Plot the new char in the blowups. */
		togcur(curx, cury); /* Turn cursor back on. */
	    }
	}
	downchr = -1;		/* Get rid of the down char indication so we
				   always have to go down and up inside
				   of the chart. Going down outside and
				   up inside doesn't count. */
	      
    }
}

void
doupdate(ep)
EventRecord *ep;
{
    SetPort(ep->who);
    BeginUpdate(ep->who);
    paintup(curchr);
    EndUpdate(ep->who);
}

paintup(chr)
int chr;
{
    Rect titlebar;
    int i;

    EraseRect(&winrect);
    SetRect(&titlebar, TITLEBARH, TITLEBARV, /* Title bar window. */
	    TITLEBARH + TITLEBARWID, TITLEBARV + TITLEBARHEI);
    newwindow(&titlebar, TRUE);
    centerstr(&titlebar, "Video Character Generator Utility 3.2.2  11/14/91");

    SetRect(&dfatbits, DCHBASEX, DCHBASEY, /* Double character size window. */
	    DCHBASEX + (CELLWIDTH*2), DCHBASEY + (CELLHEIGHT*2));
    InsetRect(&dfatbits, -1, -2);
    newwindow(&dfatbits, TRUE);
    InsetRect(&dfatbits, 1, 2);

    SetRect(&statswindow, STATSWINH, STATSWINV,	/* Status window. */
	    STATSWINH + STATSWINWID, STATSWINV + STATSWINHEI);
    newwindow(&statswindow, TRUE);

    SetRect(&promptrect, PROMPTH, PROMPTV, /* Prompt window. */
	    PROMPTH + PROMPTWID, PROMPTV + PROMPTHEI);
    newwindow(&promptrect, TRUE);
    InsetRect(&promptrect, 1, 1);

    SetRect(&chart, CHBASEX + CHLABELX, CHBASEY, /* Top label. */
	    CHBASEX + CHLABELX + (CELLWIDTH+1)*16 -1, CHBASEY + CHLABELY);
    FillRect(&chart, whitePattern);
    FrameRect(&chart);
    TextFont(1);		/* Use the scrawny font. */
    TextSize(12);
    for (i=0; i<16; i++) {	/* Put up top label characters. */
	MoveTo(chart.edge.left + 4 + (CELLWIDTH+1) * i, chart.edge.top + 11);
	DrawChar((i > 9) ? i+'7' : i+'0');
    }
    i = chart.edge.right - chart.edge.left - StringWidth("High Digit") - 10;
    chart.edge.left += i/2;
    chart.edge.right -= i/2;
    OffsetRect(&chart, 0, -(CHLABELY+1));
    chart.edge.bottom += 2;
    FillRect(&chart, whitePattern);
    FrameRect(&chart);
    InsetRect(&chart, 8, 0);
    MoveTo(chart.edge.left, chart.edge.top+10);
    DrawString("High Digit");

    SetRect(&chart, CHBASEX, CHBASEY + CHLABELY, /* Side label. */
	    CHBASEX + CHLABELX,
	    CHBASEY + CHLABELY + (CELLHEIGHT+1)*16 -1);
    FillRect(&chart, whitePattern);
    FrameRect(&chart);
    for (i=0; i<16; i++) {	/* Put up side label characters. */
	MoveTo(chart.edge.left + 3, chart.edge.top + 11 + (CELLHEIGHT+1) * i);
	DrawChar((i > 9) ? i+'7' : i+'0');
    }

    TextFont(0);		/* Use the double-stroke font. */
    TextSize(12);
    SetRect(&chart, CHBASEX + CHLABELX, CHBASEY + CHLABELY, 
	    CHBASEX + CHLABELX + (CELLWIDTH+1)*16 -1, 
	    CHBASEY + CHLABELY + (CELLHEIGHT+1)*16 -1);
    InsetRect(&chart, -1, -1);
    newwindow(&chart, FALSE);
    InsetRect(&chart, 1, 1);

    help();
    SetRect(&fatbits, WABASEX-1, WABASEY-1, /* Fatbits working area. */
	    WABASEX+WAWIDTH+2, WABASEY+WAHEIGHT+2);
    newwindow(&fatbits, TRUE);
    InsetRect(&fatbits, 1, 1);
    fatbits.edge.right--;
    shofname();
    shostat(chr);
    plotmat(chr);		/* Display char from memory */
    drawdisp(FALSE);		/* Put up the empty character cells. */
}

newwindow(rect, flag)		/* Put up a new rectangular window. */
Rect *rect;
int flag;			/* Whiteout insides. */
{
    PenMode(srcOr);
    if (flag) {
	OffsetRect(rect, 1, 1);	/* Put up the drop shadow */
	FrameRect(rect);
	OffsetRect(rect, -1, -1);	/* Put up the window. */
	FillRect(rect, whitePattern);
    } else {
	MoveTo(rect->edge.left + 1, rect->edge.bottom);
	LineTo(rect->edge.right + 1, rect->edge.bottom);
	MoveTo(rect->edge.right, rect->edge.top + 1);
	LineTo(rect->edge.right, rect->edge.bottom);
    }
    FrameRect(rect);		/* Frame it. */
}

centerstr(rect, string)
Rect *rect;
char *string;
{
    DrawJustText(string, StringLen(string), rect,
		 centerJustify, centerJustify);
}

char *helps[] = {
    "  --- Commands ---",
    "",
    "F \321 File Names",
    "L \321 Load Files",
    "R \321 Read single char",
    "S \321 Save Files",
    "Q \321 Quit",
    "",
    "B \321 Cell Boundary",
    "M \321 Baseline Marker",
    "A \321 Access Char",
    "<CR> \321 Next Char",
    "<BS> \321 Prev Char",
    "C \321 Clear char",
    "= \321 Copy from",
    "     character",
    "/ \321 Shift left",
    "* \321 Shift right",
    "- \321 Shift up",
    "+ \321 Shift down",
    "",
    "0 \321 Complement char",
    "5 \321 Complement Point",
    "Numeric pad keys move.",
    0
};

help()
{
    char *p;
    int i, row, col, maxwidth, maxheight, width;
    Rect helprect;

    maxwidth = 0;
    maxheight = 16;
    i = 0;
    p = helps[0];
    while (p = helps[i++]) {	/* Calculate the size of help window needed. */
	width = StringWidth(p);
	if (width > maxwidth)
	    maxwidth = width;
	maxheight += 14;
    }
    maxwidth += 8;

    SetRect(&helprect, HBASEX, HBASEY, HBASEX + maxwidth, HBASEY + maxheight);
    newwindow(&helprect, TRUE);

    row = helprect.edge.top+16;
    col = helprect.edge.left+5;
    i = 0;
    p = helps[0];
    while (p = helps[i++]) {
	MoveTo(col, row);
	    DrawString(p);
	    row += 14;
    }
}

wipemout()
{
    char *p, *q;
    int chrs;

    p = base1;
    q = base2;
    for (chrs = 0; chrs < NCHARS*NWORDS; chrs++)
	*p++ = *q++ = 0;
}

/* main program loop */
modify()
{
    int c;

    curchr = curx = cury = 0;
    downchr = -1;
    shofname();
    togcur(curx, cury);		/* turn on cursor */

    do {
	c = getkey(1);
	c = tolower(c);
	togcur(curx, cury);	/* turn off cursor */
	switch (c) {
	  case 'q':
	    if (modfied) {
		status("OK to quit without saving? ");
		if (yesno())
		    c = -1;
		else
		    clrprompt();
	    } else
		c = -1;
	    break;
	  case 'f':
	    getnam(0L);
	    break;
	  case 'l':
	    if (modfied) {
		status("OK to lose changes? ");
		if (!yesno()) {
		    clrprompt();
		    break;
		}
	    }
	    if (lodfil(0, NCHARS)) {
		drawdisp(TRUE);	/* Display the char set */
		status("Files loaded.");
	    }
	    shostat(curchr);
	    plotmat(curchr);	/* Display char from memory */
	    break;
	  case 'r':
	    if (lodfil(curchr, 1))
		status("Character read.");
	    showchr(curchr);	/* Display char from memory */
	    break;
	  case 's':
	    savfil();
	    shostat(curchr);
	    break;
	  case 'a':
	    c = asciichr(0);
	    if (c >= 0) {
		curchr = c;
		shostat(curchr);
		plotmat(curchr);
	    } else
		c = 0;
	    break;
	  case CR:
	    if (curchr != NCHARS-1) {
		curchr++;
		shostat(curchr);
		plotmat(curchr);
	    }
	    break;
	  case BS:
	    if (curchr) {
		curchr--;
		shostat(curchr);
		plotmat(curchr);
	    }
	    break;
	  case REPAINT:
	    paintup(curchr);
	    break;
	  case 'b':		/* Cell boundary. */
	    curx = cury = markerrow = 0;
	    getsize();
	    plotmat(curchr);
	    shostat(curchr);
	    break;
	  case 'm':		/* Baseline Marker. */
	    if (markerrow == (cury+1)) /* Same place as it already is? */
		markerrow = 0;	/* Take it down. */
	    else
		markerrow = cury+1; /* Put it up. */
	    plotmat(curchr);
	    break;
	  case 'c':
	    clrchr(curchr, TRUE);
	    break;
	  case '=':
	    copychr(asciichr(1), curchr);
	    break;
	  case '+':
	    shiftdwn(curchr);
	    break;
	  case '-':
	    shiftup(curchr);
	    break;
	  case '/':
	    shiftl(curchr);
	    break;
	  case '*':
	    shiftr(curchr);
	    break;
	  case '0':
	    compcell(curchr);
	    break;
	  case '1':
	    if (cury < cellheit - 1)
		cury++;
	    if (curx > 0)
		curx--;
	    break;
	  case '2':
	    if (cury < cellheit - 1)
		cury++;
	    break;
	  case '3':
	    if (cury < cellheit - 1)
		cury++;
	    if (curx < cellwid - 1)
		curx++;
	    break;
	  case '4':
	    if (curx > 0)
		curx--;
	    break;
	  case '5':
	    compl(curchr, curx, cury);
	    break;
	  case '6':
	    if (curx < cellwid - 1)
		curx++;
	    break;
	  case '7':
	    if (cury > 0)
		cury--;
	    if (curx > 0)
		curx--;
	    break;
	  case '8':
	    if (cury > 0)
		cury--;
	    break;
	  case '9':
	    if (curx < cellwid - 1)
		curx++;
	    if (cury > 0)
		cury--;
	    break;
	  case 0:
	    break;
	  default:
	    Bell();		/* Ring the bell. */
	    break;
	}
	togcur(curx, cury);	/* Turn on cursor again. */
    } while (c != -1);
    togcur(curx, cury);		/* Turn off the cursor. */
    status("Exiting...");
    EndGraf();
}

/* Draw the alpha set */
drawdisp(flag)
int flag;
{
    int row, col;

    if (flag)
	status("Files loaded \321 plotting...");
    for (row = 0; row < 16; row++) {	/* Do 16 rows of chars */
	for (col = 0; col < 16; col++){	/* of 16 columns each */
	    plotachr((col * 16) + row,
		     chart.edge.top + row * (CELLWIDTH + 1),
		     chart.edge.left + col * (CELLHEIGHT + 1));
	}
    }
}

/* clear a whole char for fresh start */
clrchr(chr, show)
int chr, show;
{
    int row;

    for (row = 0; row < NWORDS; row++)
	depword(0, chr, row);
    if (show)
	showchr(chr);
}

/* Copy from a char */
copychr(fromchr, tochr)
int fromchr, tochr;
{
    int row;
    unsigned short word;
    unsigned getword();

    if (fromchr >= 0) {
	for (row = 0; row < NWORDS; row++) {
	    word = getword(fromchr, row);
	    depword(word, tochr, row);
	}
	showchr(tochr);
    }
}

/* Complement a whole char, but not halfdot row. */
compcell(chr)
int chr;
{
    int row;
    unsigned short word;
    unsigned getword();

    for (row = 0; row < NWORDS; row++) {
	word = getword(chr, row);
	depword(word ^ ~1, chr, row);
    }
    showchr(chr);
}

/* Shift a whole char one row up. */
shiftup(chr)
int chr;
{
    int row;
    unsigned short word;
    unsigned getword();

    for (row=1; row<NWORDS; row++) {
	word = getword(chr, row);
	depword(word, chr, row - 1);
    }
    depword(0, chr, NWORDS-1);
    showchr(chr);
}

/* Shift a whole char one row down. */
shiftdwn(chr)
int chr;
{
    int row;
    unsigned short word;
    unsigned getword();

    for (row=15; row; row--) {
	word = getword(chr, row - 1);
	depword(word, chr, row);
    }
    depword(0, chr, 0);
    showchr(chr);
}

/* Shift a whole char one dot left, don't move halfdot bits. */
shiftl(chr)
int chr;
{
    int row;
    unsigned short word;
    unsigned getword();

    for (row=0; row<NWORDS; row++) {
	word = getword(chr, row);
	depword(((word & ~1) << 1) | (word & 1), chr, row);
    }
    showchr(chr);
}

/* Shift a whole char one row right, don't move halfdot bits. */
shiftr(chr)
int chr;
{
    int row;
    unsigned short word;
    unsigned getword();

    for (row=0; row<NWORDS; row++) {
	word = getword(chr, row);
	depword(((word >> 1) & ~1) | (word & 1), chr, row);
    }
    showchr(chr);
}


/* Complement the current point and adjust the rest, */
compl(chr, curx, cury)
int chr;
int curx, cury;
{
    int grow, gcol;
    unsigned getword();
    unsigned short word;
    Rect scratch;

    markcmp(curx, cury);
    word = getword(chr, cury);
    word ^= (1 << (15 - curx));
    depword(word, chr, cury);
    grow = (dfatbits.edge.top + cury*2);
    SetRect(&scratch, dfatbits.edge.left, grow, dfatbits.edge.right, grow+2);
    FillRect(&scratch, whitePattern);
    gcol = (word & 1) ? dfatbits.edge.left : dfatbits.edge.left+1;
    pltdrow(word, grow, gcol);
    shostat(chr);
    grow = chart.edge.top + (chr % 16) * (CELLHEIGHT + 1);
    gcol = chart.edge.left + (chr / 16) * (CELLWIDTH + 1);
    plotarow(word, grow + cury, gcol);
}

/* Show an entire char -- Both matrices & chart char. */
showchr(chr)
int chr;
{
    int grow, gcol;

    shostat(chr);
    plotmat(chr);		/* Show working cell. */
    grow = chart.edge.top + (chr % 16) * (CELLHEIGHT + 1);
    gcol = chart.edge.left + (chr / 16) * (CELLWIDTH + 1);
    plotachr(chr, grow, gcol);	/* Show chart entry. */
}

/* Draws working matrix & double sized character. */
plotmat(chr)
int chr;
{
    int i, j, k;

    FillRect(&fatbits, whitePattern); /* Wipe out fatbits view. */
    i = fatbits.edge.left + 1 + (cellwid * BLOCKWIDTH);
    j = fatbits.edge.top + 1 + (cellheit * BLOCKHEIGHT);
    k = fatbits.edge.top + 1 + (markerrow * BLOCKHEIGHT);
    PenMode(srcOr);		/* Draw cell boundary lines. */
    MoveTo(i, fatbits.edge.top);
    LineTo(i, j);
    MoveTo(fatbits.edge.left, j);
    LineTo(i, j);
    if (markerrow) {		/* Draw baseline marker. */
	MoveTo(fatbits.edge.left, k);
	LineTo(i-1, k);
    }
    for (i = 0; i < cellheit; i++)
	plotmrow(chr, i);
    pltdbl(chr);
}

/* Draws one row (word) of the matrix. */
plotmrow(chr, i)
int i;
int chr;
{
    register unsigned bit, k;
    register int j;
    unsigned getword();

    k = getword(chr, i);
    bit = 1 << 15;
    for (j = 0; j < cellwid; j++){
	if ( k & bit)
	    markcmp(j, i);
	bit >>= 1;
    }
}

/* Complements the mark defined by base coordinates  y=0..15 x=0..15 */
markcmp(x, y)
int x, y;
{
    int baddx, baddy;
    Rect fatbit;

    baddy = (fatbits.edge.top + 1) + (y * BLOCKHEIGHT);
    baddx = (fatbits.edge.left + 1) + (x * BLOCKWIDTH);
    SetRect(&fatbit, baddx, baddy, 
	    baddx + BLOCKWIDTH - 1, baddy + BLOCKHEIGHT - 1);
    InvertRect(&fatbit);
}


/* Plot a full char at the specified graphics position. */
plotachr(chr, grow, gcol)
int chr, grow, gcol;
{
    BitMap srcbits;
    register short row;
    Rect chartntry;
    unsigned getword();
    unsigned short bitmap[NWORDS], *p;

    /* Put up the chart entry. */
    srcbits.rowBytes = sizeof(unsigned short);
    SetRect(&srcbits.bounds, 0, 0, CELLWIDTH, NWORDS);
    p = bitmap;
    srcbits.baseAddr = (char *) bitmap;
    SetRect(&chartntry, gcol, grow, gcol+CELLWIDTH, grow+NWORDS);
    for (row=0; row<NWORDS; row++)
	*p++ = getword(chr, row);
    CopyBits(&srcbits, 0, &srcbits.bounds, &chartntry, srcCopy, 0);
}


/* Plot a row of char at the specified graphics position. */
plotarow(pattern, grow, gcol)
unsigned pattern;
int grow, gcol;
{
    register int col;
    register unsigned mask;

    mask = 1 << 15;
    for (col=0; col<cellwid; col++) {
	if (pattern & mask)
	    PenMode(srcOr);
	else
	    PenMode(srcBic);
	SetPixel(gcol++, grow);
	mask >>= 1;
    }
}


/* Put up the 'modified' indicator, also shows the char number. */
shostat(c)
int c;
{
    char scratch[20];
    Rect wipeout;

    sprintf(scratch, modfied ? "* ASCII: $%02X" : "  ASCII: $%02X", c);
    SetRect(&wipeout, statswindow.edge.left + 3, statswindow.edge.top + 38 - 9,
	    statswindow.edge.right - 1, statswindow.edge.top + 38 + 3);
    FillRect(&wipeout, whitePattern); /* Clear screen's string. */
    MoveTo(statswindow.edge.left + 3, statswindow.edge.top + 38);
    DrawString(scratch);
}

/* Plot the full double size matrix of chr. */
pltdbl(chr)
int chr;
{
    register int row, tcol;
    register unsigned pattern, mask;
    unsigned getword();
    int trow;

    FillRect(&dfatbits, whitePattern);
    for (trow = dfatbits.edge.top, row=0; row<cellheit; row++) {
	if (pattern = getword(chr, row)) {
	    tcol = dfatbits.edge.left;
	    if (!(pattern & 1))
		tcol += 1;
	    pltdrow(pattern, trow, tcol);
	    tcol += 2;
	}
	trow += 2;
    }
}

/* Plot a row of double size matrix. */
pltdrow(pattern, trow, tcol)
unsigned pattern;
int trow, tcol;
{
    register unsigned locpat, mask;
    register int col;
    int cols;
    Rect quadbit;

    locpat = pattern;
    mask = 1 << 15;
    cols = (cellwid == CELLWIDTH) ? CELLWIDTH-1 : cellwid;
    for (col=0; col<cols; col++) {
	if (locpat & mask) {
	    SetRect(&quadbit, tcol, trow, tcol+2, trow+2);
	    FillRect(&quadbit, blackPattern);
	}
	tcol += 2;
	mask >>= 1;
    }
}


/* Gets a row of the cell. */
unsigned
getword(chr, yloc)	/* Get word from buffer */
int yloc;		/* yloc=row of char */
int chr;		/* chr=character number */
{
    unsigned temp;
    temp = *(base1 + (chr * NWORDS) + yloc) << 8;
    return temp | (*(base2 + (chr * NWORDS) + yloc) & 0xff);
}


/* Puts a row of the cell. */
depword(w, chr, yloc)	/* Deposit word to buffer */
unsigned w;
int yloc;		/* w=word, yloc=row of char */
int chr;		/* chr=character number */
{
    *(base1 + (chr * NWORDS) + yloc) = w >> 8;
    *(base2 + (chr * NWORDS) + yloc) = w;
    modfied = TRUE;
}


/* Toggle the graphic cursor ON/OFF. */
togcur(curx, cury)
int curx, cury;
{
    int addrx, addry;

    addrx = fatbits.edge.left + (curx * BLOCKWIDTH);
    addry = fatbits.edge.top + (cury * BLOCKHEIGHT);
    PenMode(srcXor);
    MoveTo(addrx + BLOCKWIDTH/2, addry);	/* Vertical. */
    LineTo(addrx + BLOCKWIDTH/2, addry + BLOCKHEIGHT);
    MoveTo(addrx, addry + BLOCKHEIGHT/2);	/* Horizontal. */
    LineTo(addrx + BLOCKWIDTH, addry + BLOCKHEIGHT/2);
}

status(frmt)			/* Prompt with String 'frmt' */
char *frmt;
{
    clrprompt();
    centerstr(&promptrect, frmt);
}

prompt(frmt)			/* Prompt with String 'frmt' */
char *frmt;
{
    int height;

    clrprompt();
    height = promptrect.edge.bottom - promptrect.edge.top;
    MoveTo(promptrect.edge.left + 10, promptrect.edge.top + height/2 + 4);
    DrawString(frmt);
}

char *
dogets(count)
int count;			/* How many characters to accept. */
{
    char *ptr;
    static char scratch[30];
    int i, width, cursor;
    Rect wiper;			/* Box to type into. */
    Point save;

    GetPen(&save);		/* Save the current pen location. */
    SetRect(&wiper, save.h, promptrect.edge.top, 
	    promptrect.edge.right, promptrect.edge.bottom);
    FillRect(&wiper, whitePattern);
    PenMode(srcXor);
    cursor = 3;			/* Cursor is on. */
    ptr = scratch;
    i = 0;
    while (1) {
	do {
	    *ptr = 0;
	    width = StringWidth(scratch);
	    if (cursor>0) {	/* Put up 'cursor'? */
		MoveTo(wiper.edge.left + width + PAD2, wiper.edge.top + 1);
		Line(0, wiper.edge.bottom - wiper.edge.top - 2);
	    }
	    *ptr = getkey(0);
	    if (cursor>0) {	/* Remove 'cursor'? */
		MoveTo(wiper.edge.left + width + PAD2, wiper.edge.top + 1);
		Line(0, wiper.edge.bottom - wiper.edge.top - 2);
	    }
	    if (*ptr)		/* If there was input, cursor is to be on. */
		cursor = 3;
	    else {		/* Timed out, wink the cursor. */
		if (cursor<0)
		    cursor = 3;
		else
		    cursor--;
		continue;	/* Go for more input. */
	    }
	    *(ptr+1) = '\0';
	    if (*ptr == BS) {	/* Handle BS */
		*ptr = 0;
		if (ptr != scratch) {
		    *--ptr = 0;
		    i--;
		    FillRect(&wiper, whitePattern); /* Wipe out any string that's there. */
		    vcenterstr(&wiper, scratch, wiper.edge.left + PAD2);
		} else {
		    Bell();
		}
	    }
	    if (*ptr == CR) {	/* Hit CR? */
		if (i == 0)	/* Nothing entered?  Then quit. */
		    return 0;
		else
		    break;	/* Else finish up. */
	    }
	} while (!isgraph(*ptr));
	if (*ptr == CR) {
	    *ptr = 0;
	    break;
	}
	if (i++ == count) {	/* Too many chars, don't take it. */
	    i--;
	    *ptr = 0;
	    Bell();
	    continue;
	}
	FillRect(&wiper, whitePattern);	/* Wipe out any string that's there. */
	vcenterstr(&wiper, scratch, wiper.edge.left + PAD2); /* Redraw string. */
	ptr++;
    }
    *ptr = '\0';
    return scratch;
}

int
yesno()
{
    char *p;

    if (p = dogets(1)) {	/* Prompt for Yes/No. */
	*p = tolower(*p);
	if (*p == 'y')
	    return 1;
    }
    return 0;
}

getnam(passedname)		/* Get Filenames. */
char *passedname;
{
    char *temp;

    if (passedname)
	temp = passedname;
    else {
	prompt("Enter File Name: ");
	temp = dogets(12);
    }
    if (temp) {
	strncpy(filnam1, temp, sizeof(filnam1));
	strcpy(filnam2, filnam1);
	strcat(filnam1, ".1");
	strcat(filnam2, ".2");
	strcpy(fnames, filnam1);
	strcat(fnames, " & ");
	strcat(fnames, filnam2);
	if (StringWidth(fnames) >= 
	    ((statswindow.edge.right - statswindow.edge.left) - 3)) {
	    strcpy(fnames, filnam1);
	    strcat(fnames, " & .2");
	}
    }
    clrprompt();
    shofname();
}

shofname()
{
    char scratch[20];
    Rect wipeout;

    MoveTo(statswindow.edge.left + 3, statswindow.edge.top + 10);
    DrawString("Files:");
    SetRect(&wipeout, statswindow.edge.left + 3, statswindow.edge.top + 22 - 9,
	    statswindow.edge.right - 1, statswindow.edge.top + 22 + 4);
    FillRect(&wipeout, whitePattern); /* Clear screen's string. */
    MoveTo(statswindow.edge.left + 3, statswindow.edge.top + 22);
    DrawString(fnames);

    sprintf(scratch, "Cell: %02d x %02d", cellwid, cellheit);
    SetRect(&wipeout, statswindow.edge.left + 3, statswindow.edge.top + 53 - 9,
	    statswindow.edge.right-1, statswindow.edge.top + 53 + 4);
    FillRect(&wipeout, whitePattern);
    MoveTo(statswindow.edge.left + 3, statswindow.edge.top + 53);
    DrawString(scratch);
}

asciichr(flag)			/* Get ASCII char in HEX */
int flag;			/* 0 == 'A' command, 1 == '=' command */
{
    char scratch[20], *temp;
    int ch, i;

    do {
	if (flag) {
	    prompt("Character to copy (HEX): ");
	    temp = dogets(2);
	    if (!temp) {
		clrprompt();
		return -1;
	    }
	    strcpy(scratch, temp);
	}
	else {
	    prompt("Edit character (HEX): ");
	    temp = dogets(2);
	    if (!temp) {
		clrprompt();
		return -1;
	    }
	    strcpy(scratch, temp);
	}
	i = sscanf(scratch, "%x", &ch);
    } while (i != 1 || ch <0 || ch > 256);
    clrprompt();
    return(ch);
}

getsize()			/* Prompt for cell size */
{
    int ch, i;
    char scratch[30], *temp;

    do {
	prompt("New cell width (1..16): ");
	temp = dogets(2);
	if (!temp)
	    break;
	strcpy(scratch, temp);
	i = sscanf(scratch, "%d", &ch);
	cellwid = ch;
    } while (i != 1 || ch <0 || ch > 16);

    do {
	sprintf(scratch, "Width: %d, New cell height (1..16): ", cellwid);
	prompt(scratch);
	temp = dogets(2);
	if (!temp)
	    break;
	strcpy(scratch, temp);
	i = sscanf(scratch, "%d", &ch);
	cellheit = ch;
    } while (i != 1 || ch <0 || ch > 16);
    shofname();
    clrprompt();
}

clrprompt()
{
    FillRect(&promptrect, whitePattern);
}

savfil()	/* Save files */
{
    int i, fd;
    if (!(*filnam1)) {
	status("No file names specified!");
	return;
    }
    pstatus("Creating file '%s'...", filnam1);
    if ((fd = creat(filnam1, 0644)) == ERROR) {
	pstatus("Can't create file '%s'!", filnam1);
	return;
    }

    pstatus("Writing file '%s'...", filnam1);
    i = NWORDS * NCHARS;
    if (i != write(fd, base1, i)) {
	status("Write error - disk probably full.");
	return;
    }
    if (close(fd) == ERROR) {
	pstatus("Error closing file '%s'!", filnam1);
	return;
    }

    pstatus("Creating file '%s'...", filnam2);
    if ((fd = creat(filnam2, 0644)) == ERROR) {
	pstatus("Can't create file '%s'!", filnam2);
	return;
    }

    pstatus("Writing file '%s'...", filnam2);
    if (i != write(fd, base2, i)) {
	status("Write error - disk probably full.");
	return;
    }
    if (close(fd) == ERROR)
	pstatus("Error closing file '%s'!", filnam2);
    else {
	status("Files written.");
	modfied = FALSE;
    }
}

pstatus(cstring, arg)
char *cstring, *arg;
{
    char scratch[50];

    sprintf(scratch, cstring, arg);
    status(scratch);
}

lodfil(which, howmany)	/* Load char(s) from files. */
int which, howmany;
{
    int i, fd;

    if (!(*filnam1)) {
	status("No file names specified.");
	return FALSE;
    }
    if ((fd = open(filnam1, 0)) == ERROR) {
	pstatus("File '%s' doesn't exist!", filnam1);
	return FALSE;
    }
    modfied = TRUE;
    pstatus(" Reading file '%s'...", filnam1);

    i = NWORDS * howmany;
    if (which != 0)
	lseek(fd, (long) which*NWORDS, 0);
    if (i != read(fd, base1 + (which * NWORDS), i)) {
	pstatus("Error reading file '%s'!", filnam1);
	return FALSE;
    }
    close(fd);

    if ((fd = open(filnam2, 0)) == ERROR) {
	pstatus("File '%s' doesn't exist!", filnam2);
	return FALSE;
    }
    pstatus(" Reading file '%s'...", filnam2);

    if (which != 0)
	lseek(fd, (long) which*NWORDS, 0);
    if (i != read(fd, base2 + (which * NWORDS), i)) {
	pstatus("Error reading file '%s'!", filnam2);
	return FALSE;
    }
    close(fd);
    modfied = (howmany == 1) ? TRUE : FALSE;
    return TRUE;
}

null(string)
char *string;
{
}

int
getkey(flag)
int flag;			/* Non-zero if no timeout. */
{
    long ch;

    ch = 0;
    do {
	ch = GetKey(mywindow);
    } while (ch & F_KEYUP);
    ObscureCursor();
    if ((ch & F_KEYMASK) == 0x7F) /* Handle DEL->BS mapping. */
	ch = BS;
    if (ch & F_FUNC) {
	if ((ch & F_KEYMASK) >= '0' && (ch & F_KEYMASK) <= '9')
	    ch &= ~F_FUNC;
	else if ((ch & F_KEYMASK) == K_ENTER)
	    ch = CR;
	else if ((ch & F_KEYMASK) == K_NPENTER)
	    ch = CR;
	else if ((ch & F_KEYMASK) == K_PF1)
	    ch = BS;
	else if ((ch & F_KEYMASK) == K_PF2)
	    ch = '/';
	else if ((ch & F_KEYMASK) == K_PF3)
	    ch = '*';
	else if ((ch & F_KEYMASK) == K_PF4)
	    ch = '-';
	else if ((ch & F_KEYMASK) == '+')
	    ch &= ~F_FUNC;
	else if ((ch & F_KEYMASK) == K_UP)
	    ch = '8';
	else if ((ch & F_KEYMASK) == K_DOWN)
	    ch = '2';
	else if ((ch & F_KEYMASK) == K_RIGHT)
	    ch = '6';
	else if ((ch & F_KEYMASK) == K_LEFT)
	    ch = '4';
	else
	    ch = -1;
    }
    return ch & 0xFF;
}

vcenterstr(rect, string, hpos)
Rect *rect;
char *string;
short hpos;
{
    Rect scratch;

    scratch = *rect;
    scratch.edge.left = hpos;
    DrawJustText(string, StringLen(string), &scratch, 
		 leftJustify, centerJustify);
}
