/*
**	Font Converter,	Jim Cathey.
**
**	Program to convert a binary font file to a clunkdraw font source.
**
**	Version 1.5	6/14/90		JEC
**		Fixed widMax field of font to be correct (+1 for space).
**
**	Version 1.4	4/12/90		JEC
**		Added bit-reversed output option for TI34010.
**
**	Version 1.3	12/21/88	JEC
**		Fixed non-breaking space's width.
**
**	Version 1.2	10/19/88	JEC
**		Added error messages on file write errors.
**
**	Version 1.1	4/6/88		JEC
**		Added raw binary output.
**
**	Version 1.0	3/9/88		JEC
*/

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include "isc/clunkdraw.h"

#define NCHARS 256		/* Number of characters supported */
#define NWORDS 16		/* Number of words per char */
#define TRUE (-1)
#define FALSE 0
#define ERROR (-1)

char *base1, *base2;		/* Font working storage pointers. */

char *text1, *text2, *text3, *text4, *text5; /* Assy language pointers. */

char filename[100], *wfilename, *malloc();
char fontname[32];

int lowest_row;			/* Font metrics variables. */
int widest_col;
int first_char;
int last_char;
int baseline;
int leading;

int binary;			/* Global flag for binary output. */
int bitrev;			/*   "      " for bit-reversed output. */
int outfd;			/* Output file descriptor. */

main(argc, argv)
int argc;
char *argv[];
{
    int fd, i, infarg, source, outfarg, vaxflag;
    int header, loctab, offwidtab;

    if (argc < 2) {
	usage(argv[0]);
	exit(1);
    }

    (void) strcpy(fontname, "__font");
    source = first_char = last_char = 0;
    infarg = outfarg = vaxflag = baseline = leading = binary = bitrev = 0;
    header = loctab = offwidtab = 1;
    for (i=1; i<argc; i++) {	/* Search out arguments. */
	if (!strcmp(argv[i], "-S")) { /* Output source. */
	    source++;
	    continue;
	}
	if (!strcmp(argv[i], "-V")) { /* Polluted 68000 format. */
	    vaxflag++;
	    continue;
	}
	if (!strncmp(argv[i], "-T",2)) { /* Bit-reversed output for TI34010. */
	    bitrev++;
	    continue;
	}
	if (!strcmp(argv[i], "-o")) { /* Override output file. */
	    outfarg = ++i;
	    continue;
	}
	if (!strcmp(argv[i], "-n")) { /* Font name. */
	    (void) strcpy(fontname, argv[++i]);
	    continue;
	}
	if (!strcmp(argv[i], "-base")) { /* Set font baseline. */
	    (void) strcpy(filename, argv[++i]);
	    baseline = getnum(filename);
	    continue;
	}
	if (!strcmp(argv[i], "-first")) { /* Set first font character. */
	    (void) strcpy(filename, argv[++i]);
	    first_char = getnum(filename);
	    continue;
	}
	if (!strcmp(argv[i], "-last")) { /* Set last font character. */
	    (void) strcpy(filename, argv[++i]);
	    last_char = getnum(filename)+1;
	    continue;
	}
	if (!strcmp(argv[i], "-lead")) { /* Set font leading. */
	    (void) strcpy(filename, argv[++i]);
	    leading = getnum(filename);
	    continue;
	}
	if (!strcmp(argv[i], "-header")) { /* No header? */
	    header = 0;
	    continue;
	}
	if (!strcmp(argv[i], "-loctab")) { /* No location table? */
	    loctab = 0;
	    continue;
	}
	if (!strcmp(argv[i], "-offwidtab")) { /* No offset/width table? */
	    offwidtab = 0;
	    continue;
	}
	if (!strcmp(argv[i], "-binary")) { /* No a.out header? */
	    binary = 1;
	    continue;
	}
	infarg = i;		/* Only if it ain't one of the others. */
    }

    if (binary)
	(void) strcpy(fontname, "font");
    if (!source && !binary) {
	fprintf(stderr, "%s: Can't output COFF yet.\n", argv[0]);
	exit(1);
    }
    base1 = malloc(NCHARS*NWORDS); /* Grab arrays for chargen */
    base2 = malloc(NCHARS*NWORDS);
    if (!base1 || !base2) {
	fprintf(stderr, "%s: Out of memory.\n", argv[0]);
	exit(ENOMEM);
    }

    strcpy(filename, argv[infarg]);
    strcat(filename, ".1");
    if ((fd = open(filename, 0)) == ERROR) {
	fprintf(stderr, "%s: File '%s' doesn't exist!\n", argv[0], filename);
	exit(2);
    }
    i = NWORDS * NCHARS;
    if (i != read(fd, base1, i)) {
	fprintf(stderr, "%s: Error reading file '%s'!\n", argv[0], filename);
	exit(3);
    }
    close(fd);

    strcpy(filename, argv[infarg]);
    strcat(filename, ".2");
    if ((fd = open(filename, 0)) == ERROR) {
	fprintf(stderr, "%s: File '%s' doesn't exist!\n", argv[0], filename);
	exit(2);
    }
    if (i != read(fd, base2, i)) {
	fprintf(stderr, "%s: Error reading file '%s'!\n", argv[0], filename);
	exit(3);
    }
    close(fd);

    calcmetrics();		/* Calculate font metrics. */
    if (vaxflag) {		/* Polluted assembler? */
	text1 = "\t.globl\t";
	text2 = "\t.word\t";
	text3 = "\t.byte\t";
	text4 = bitrev ? "\t; " : "\t| ";
	text5 = ":";
    } else {			/* Normal assembler. */
	text1 = "\t.xdef\t";
	text2 = "\tdc.w\t";
	text3 = "\tdc.b\t";
	text4 = "\t; ";
	text5 = "";
    }	
    if (binary) {
	if (outfarg)
	    outfd = creat(wfilename = argv[outfarg], 0644);
	else
	    outfd = creat(wfilename = fontname, 0644);
	if (outfd < 0) {
	    fprintf(stderr, "%s: Error creating file '%s'!\n", argv[0],
		    wfilename);
	    exit(4);
	}
    }
    doheader(argv, infarg, offwidtab, loctab, header); /* Put out header, */
    bitimage();			/* the bit image,  */
    if (header) {
	if (loctab)
	    loctable();		/* the location table,  */
	if (offwidtab)
	    offwidtable();	/* and the offset/width table. */
    }
    if (!binary)
	printf("\t.end\n");
    else
	close(outfd);
    exit(0);
}

doheader(argv, infarg, offwidtab, loctab, header) /* Output the font header. */
char *argv[];
int infarg, offwidtab, loctab, header;
{
    int offset, offset2;
    int i;
    FontRec tempfont;

    i = (widest_col >= 8) ? 16 : 8; /* Calculate some font record stuff. */
    offset = (widest_col >= 8) ? 2 : 1;	/* Bytewide or wordwide image store? */
    offset *= (1+ last_char - first_char) * (lowest_row+1); /* Image size. */
    offset += sizeof(FontRec);	/* Add size of header. */
    offset2 = offset;
    if (!loctab)		/* Properly handle suppressed tables. */
	offset = 0;
    else
	offset2 += ((1+last_char - first_char) * 2);
    if (!offwidtab)
	offset2 = 0;
    if (offset & 1)		/* Word-align it. */
	offset++;
    if (offset2 & 1)		/* Word-align it. */
	offset2++;

    tempfont.fontType = 1;	/* Fill font record. */
    tempfont.firstChar = first_char;
    tempfont.lastChar = last_char-1;
    tempfont.widMax = widest_col+2;
    tempfont.kernMax = 0;
    tempfont.nDescent = baseline-lowest_row;
    tempfont.fRectWidth = i;
    tempfont.fRectHeight = lowest_row+1;
    tempfont.owTLoc = offset2;
    tempfont.ascent = baseline+1;
    tempfont.descent = lowest_row - baseline;
    tempfont.leading = leading;
    tempfont.rowWords = offset;

    if (!binary) {
	printf("%sWARNING!!!  This font file was automatically created by the\n", text4);
	printf("%s            '%s' program from the files '%s.1' & '%s.2'.\n", text4, argv[0], argv[infarg], argv[infarg]);
	printf("%s            Do not edit this file.  Make changes to the original\n", text4);
	printf("%s            binary files.\n\n", text4);
	printf("%s%s\n\n%s%s", text1, fontname, fontname, text5);
	if (!header)
	    return;
	printf("%s%d%sFont type.\n", text2, tempfont.fontType, text4);
	printf("%s%d%sFirst Character.\n", text2, tempfont.firstChar, text4);
	printf("%s%d%sLast Character.\n", text2, tempfont.lastChar, text4);
	printf("%s%d%sMaximum character width.\n", text2, tempfont.widMax, text4);
	printf("%s%d%sKern maximum.\n", text2, tempfont.kernMax, text4);
	printf("%s%d%sNegative of descent value.\n", text2, tempfont.nDescent, text4);
	printf("%s%d%sWidth of font rectangle.\n", text2, tempfont.fRectWidth, text4);
	printf("%s%d%sHeight of font rectangle.\n", text2, tempfont.fRectHeight, text4);
	printf("%s%d%sOffset to offset/width table.\n", text2, tempfont.owTLoc, text4);
	printf("%s%d%sAscent from baseline.\n", text2, tempfont.ascent, text4);
	printf("%s%d%sDescent from baseline.\n", text2, tempfont.descent, text4);
	printf("%s%d%sLeading (bottom-to-next-baseline).\n", text2, tempfont.leading, text4);
	printf("%s%d%sOffset (bytes) to loc. table.\n", text2, tempfont.rowWords, text4);
    } else {
	if (write(outfd, &tempfont, sizeof(FontRec)) != sizeof(FontRec)) {
	    fprintf(stderr, "%s: Error writing file '%s'!\n", argv[0],
		    wfilename);
	    exit(5);
	}
    }
}

bitimage()			/* Output the bit image. */
{
    register short row, bit, word, chr, shortloop, last, first, index;
    unsigned short words[NCHARS*NWORDS];
    unsigned char bytes[NCHARS*NWORDS];
    unsigned getword();

    index = 0;
    last = last_char;
    first = first_char;
    if (!binary)
	printf("\t\t%sFont's bit image follows:\n\n", text4);
    for (shortloop = 2; shortloop; shortloop--) {
	for (chr = first; chr < last; chr++) {
	    if (!binary)
		printf("char%02X%s", chr, text5);
	    for (row = 0; row <= lowest_row; row++) {
		word = getword(chr, row);
		if (widest_col >= 8) {
		    if (!binary) {
			printf("%s%%", text2);
			if (bitrev) {
			    for (bit = 0; bit < 16; bit++) 
				putchar((word & (1<<bit)) ? '1' : '0');
			} else {
			    for (bit = 0; bit < 16; bit++) 
				putchar((word & (1<<(15-bit))) ? '1' : '0');
			}
		    } else
			words[index++] = word;
		} else {	/* 8 bits wide. */
		    if (!binary) {
			printf("%s%%", text3);
			if (bitrev) {
			    for (bit = 8; bit < 16; bit++) 
				putchar((word & (1<<bit)) ? '1' : '0');
			} else {
			    for (bit = 0; bit < 8; bit++) 
				putchar((word & (1<<(15-bit))) ? '1' : '0');
			}
		    } else
			bytes[index++] = word >> 8;
		}
		if (!binary) {
		    if (row == 0 && isgraph(chr))
			printf("%s%c", text4, chr);
		    putchar('\n');
		}
	    }
	    if (!binary)
		puts("");
	}
	if (first) {		/* Go 'round again & put out char #0. */
	    first = 0;
	    last = 1;
	} else
	    break;
    }
    if (binary) {		/* Binary file. */
	if (widest_col >= 8) {
	    write(outfd, words, index*2);
	} else {
	    if (index & 1)	/* Word-align it. */
		index++;
	    write(outfd, bytes, index);
	}
    }
}

loctable()			/* Output the location table. */
{
    short offset, chr, index;
    unsigned short locary[512];

    if (!binary) {
	puts("");
	printf("\t\t%sFont's location table follows:\n\n", text4);
    }
    index = 0;
    offset = sizeof(FontRec);
    for (chr = first_char; chr < last_char; chr++) {
	if (!binary) {
	    printf("%s%d", text2, offset);
	    if (isgraph(chr))
		printf("%s%c", text4, chr);
	    puts("");
	}
	locary[index++] = offset;
	offset += (lowest_row+1);
	if (widest_col >= 8)
	    offset += (lowest_row+1);
    }
    locary[index++] = offset;	/* Also do the missing symbol. */
    if (!binary) {
	printf("%s%d", text2, offset);
	if (isgraph(0))
	    printf("%s%c", text4, 0);
	puts("");
    } else {
	write(outfd, locary, index*2);
    }
}

offwidtable()			/* Output offset/width table. */
{
    unsigned short word, mask;
    short chr, row, col, bit, shortloop, last, first, spacewid, index;
    unsigned short offtab[256];

    index = 0;
    last = last_char;
    first = first_char;
    spacewid = 0;		/* Find width of 'n' for spaces widths. */
    for (row = 0; row <= lowest_row; row++) {
	word = getword('n', row);
	mask = 0x8000;
	for (bit = 1; bit < 17; bit++) {
	    if (word & mask) {
		if (bit > spacewid)
		    spacewid = bit;
	    }
	    mask >>= 1;
	}
    }
    if (!binary)
	printf("\n\t\t%sFont's offset/width table follows:\n\n", text4);
    for (shortloop = 2; shortloop; shortloop--) {
	for (chr = first; chr < last; chr++) {
	    col = 0;
	    for (row = 0; row <= lowest_row; row++) {
		word = getword(chr, row);
		mask = 0x8000;
		for (bit = 1; bit < 17; bit++) {
		    if (word & mask) {
			if (bit > col)
			    col = bit;
		    }
		    mask >>= 1;
		}
	    }
	    if (!col && chr == ' ') /* So spaces have width. */
		col = spacewid;
	    if (!col && chr == 0xCA) /* Non-breaking space. */
		col = spacewid;
	    if (!binary) {
		if (bitrev)
		    printf("%s%d,%d", text3, col+1, 0);
		else
		    printf("%s%d,%d", text3, 0, col+1);
		if (isgraph(chr))
		    printf("%s%c", text4, chr);
		puts("");
	    } else
		offtab[index++] = col+1;
	}
	if (first) {		/* Go 'round again & put out char #0. */
	    first = 0;
	    last = 1;
	} else
	    break;
    }
    if (binary)
	write(outfd, offtab, index*2);
}

/* Gets a row of the cell. */
unsigned
getword(chr, yloc)	/* Get word from buffer */
int yloc;		/* yloc=row of char */
int chr;		/* chr=character number */
{
    unsigned temp;
    temp = *(base1 + (chr * NWORDS) + yloc) << 8;
    return temp | (*(base2 + (chr * NWORDS) + yloc) & 0xFF);
}

calcmetrics()			/* Measure the read-in font. */
{
    unsigned short word, mask, emptyflag;
    int chr, row, bit, temphighest, tempbase;

    temphighest = tempbase = 0;	/* Find the extent of the font. */
    lowest_row = widest_col = 0; /* Find the largest char(s). */
    for (chr = 1; chr < 256; chr++) {
	emptyflag = 0;
	for (row = 0; row < 16; row++) {
	    word = getword(chr, row);
	    emptyflag |= word;
	    if (word) {
		if (row > lowest_row)
		    lowest_row = row;
		if (chr == 0x30) /* Sample baseline on the '0' char. */
		    tempbase = row;
	    }
	    mask = 0x8000;
	    for (bit = 0; bit < 16; bit++) {
		if (word & mask) {
		    if (bit > widest_col)
			widest_col = bit;
		}
		mask >>= 1;
	    }
	}
	if (emptyflag) {	/* We have a live one? */
	    if (!first_char)	/* First_char not set yet? */
		first_char = chr;
	    temphighest = chr;
	}
    }
    if (!last_char)
	last_char = temphighest+1;
    if (!baseline)
	baseline = tempbase;
}

int
getnum(string)			/* Scan a (possibly hex) number. */
char *string;
{
    int i;

    if (!strncmp(string, "0x", 2))
	sscanf(string+2, "%x", &i);
    else
	i = atoi(string);
    return i;
}

usage(fn)			/* Complain about crap invocation. */
char *fn;
{
    fprintf(stderr, "Usage: %s fontname [-S [-V] | -o outputname] [-T] [-binary] [-n name] [-b baseline] [-f nn] [-l nn] [-header] [-loctab] [-offwidtab]\n\n", fn);
    fprintf(stderr, "\twhere fontname is the base name of a font.1 & font.2 pair,\n");
    fprintf(stderr, "\t      -S means to spit out an ASCII source to stdout instead\n");
    fprintf(stderr, "\t         of converting to a COFF <fontname.o> file,\n");
    fprintf(stderr, "\t      -V means to spit out polluted 68000 source instead\n");
    fprintf(stderr, "\t         of Motorola standard syntax,\n");
    fprintf(stderr, "\t      -T means to spit out bit-reversed data for the TI34010,\n");
    fprintf(stderr, "\t      -binary forces raw binary font output,\n");
    fprintf(stderr, "\t      -o forces output to the file <outputname>,\n");
    fprintf(stderr, "\t      -n name sets the output font symbol to <name>,\n");
    fprintf(stderr, "\t      -base baseline means the baseline of the characters is\n");
    fprintf(stderr, "\t            the given row rather than some default,\n");
    fprintf(stderr, "\t      -lead sets the leading of the font,\n");
    fprintf(stderr, "\t      -first nn means to output character nn as the first character\n");
    fprintf(stderr, "\t             instead of the first non-blank character,\n");
    fprintf(stderr, "\t      -last nn means to output character nn as the last character\n");
    fprintf(stderr, "\t            instead of the last non-blank character,\n");
    fprintf(stderr, "\t      -header suppresses the font header and all tables,\n");
    fprintf(stderr, "\t      -loctab suppresses the font's location table,\n");
    fprintf(stderr, "\t      -offwidtab suppresses the font's offset/width table.\n\n");

    fprintf(stderr, "\tThe files <fontname.1> and <fontname.2> are assumed to be\n");
    fprintf(stderr, "\tin 16-bit x 16-row video font PROM format.  Character #0\n");
    fprintf(stderr, "\tis treated as the font's 'missing symbol', so it doesn't\n");
    fprintf(stderr, "\tcount as the first character.  Numbers like 0xNN are hex.\n\n");
}
