
#include <time.h>
#include <isc/window.h>
#include "clock.h"

extern struct timesetup	clockdata;
extern struct tm *ltime;
extern User_polys hour_hand;
extern User_polys min_hand;
extern Rect icnrect;				/* Window rectangles	*/
extern FontInfo finfo;
extern int iconified;
extern short width;
extern short height;

Rect	watch_cntr;

short	ctop, cleft;
short	shand;						/*	second hand length	*/
short	mhand;						/*	minute hand length	*/
short	hhand;						/*	hour hand length	*/
short	mshand, hshand;

/*
** Given a time value and a hand length, calculate the appropriate hand
** x and y coordinates
*/
calc_hand_coord(tick, handlen)
short handlen, tick;
{
    short xlen, ylen;
	short table[61] = {
			 0000,  1045,  2079,  3090,   4067,  5000,
			 5877,  6691,  7431,  8090,   8660,  9135,
			 9510,  9781,  9945,  10000,  9945,  9781,
			 9510,  9135,  8660,  8090,   7431,  6691,
			 5877,  5000,  4067,  3090,   2079,  1045,
			 0000, -1045, -2079, -3090,  -4067, -5000,
			-5877, -6691, -7431, -8090,  -8660, -9135,
			-9510, -9781, -9945, -10000, -9945, -9781,
			-9510, -9135, -8660, -8090,  -7431, -6691,
			-5877, -5000, -4067, -3090,  -2079, -1045,
			 0000 };

	/*
	** value tick is 6 degrees - smallest increment of movement on clock
	** is one second so table has 60 elements + 1.
	*/
	tick += 45;				/* first offset everything so 12 is at top */
	if (tick > 59) {
		tick -= 60;
	}
#ifdef	DEBUG
	fprintf(stderr, "tick = %d, handlen = %d, ", tick, handlen);
#endif

	/*
	** now scale the hands and calculate the x and y positions
	*/
    ylen = ScaleS(handlen, table[tick], 10000);
    tick += 15;
    if (tick > 59) {
		tick -= 60;
	}
    xlen = ScaleS(handlen, table[tick], 10000);
	clockdata.h_xcord = xlen;				/* get x co-ordinate */
	clockdata.h_ycord = ylen;				/* get y co-ordinate */
#ifdef	DEBUG
	fprintf(stderr, "xlen = %d, ylen = %d\n", xlen, ylen);
#endif
}

/*
**	Draw the watch arms
*/
draw_analog_arms()
{
	int		hr, min, temptime;
	PolyHandle	segment_1, dupseg1;

	if (clockdata.old_sec != ltime->tm_sec) {
		if (!iconified) {
			/*	erase old second hand	*/
			calc_hand_coord(clockdata.old_sec, shand);
			PenMode(srcBic);
			PenSize( 1, 1 );
			MoveTo( cleft + clockdata.center, ctop + clockdata.center );
			LineTo( cleft + clockdata.center + clockdata.h_xcord, ctop + clockdata.center + clockdata.h_ycord );
		}
	}

	if (clockdata.old_min != ltime->tm_min) {
		if (!iconified) {
		/*	erase the old minute hand	*/
			clear_this_poly(&min_hand);
		}
		else {
			calc_hand_coord(clockdata.old_min, mhand);
			PenMode(srcBic);
			PenSize( 1, 1 );
			MoveTo( cleft + clockdata.center, ctop + clockdata.center );
			LineTo( cleft + clockdata.center + clockdata.h_xcord, ctop + clockdata.center + clockdata.h_ycord );
		}
	}

	if (clockdata.old_hour != ltime->tm_hour || (clockdata.old_min / 12) != (ltime->tm_min / 12)) {
		if (!iconified) {
			/*	erase the old hour hand	*/
			clear_this_poly(&hour_hand);
		}
		else {
			if (clockdata.old_hour >= 13) {
				hr = clockdata.old_hour - 12;
			}
			else {
				hr = clockdata.old_hour;
			}
			min = clockdata.old_min / 12;
			calc_hand_coord((hr * 5) + min, hhand);
			PenMode(srcBic);
			PenSize( 1, 1 );
			MoveTo( cleft + clockdata.center, ctop + clockdata.center );
			LineTo( cleft + clockdata.center + clockdata.h_xcord, ctop + clockdata.center + clockdata.h_ycord );
		}
	}

	if (clockdata.old_sec != ltime->tm_sec) {
		if (!iconified) {
			/*	draw the new second hand	*/
			calc_hand_coord(ltime->tm_sec, shand);
			PenMode(srcOr);
			PenSize( 1, 1 );
			MoveTo( cleft + clockdata.center, ctop + clockdata.center );
			LineTo( cleft + clockdata.center + clockdata.h_xcord, ctop + clockdata.center + clockdata.h_ycord );
		}
	}

	/*	draw the new minute hand	*/
	if( !iconified ) {
		draw_hand_poly(&min_hand, ltime->tm_min, mhand, mshand);
	}
	else {
		if (clockdata.old_min != ltime->tm_min) {
			calc_hand_coord(ltime->tm_min, mhand);
			PenMode(srcOr);
			PenSize( 1, 1 );
			MoveTo( cleft + clockdata.center, ctop + clockdata.center );
			LineTo( cleft + clockdata.center + clockdata.h_xcord, ctop + clockdata.center + clockdata.h_ycord );
		}
	}

	/*	draw the new hour hand	*/
	if (ltime->tm_hour>= 13 ) {
		hr = ltime->tm_hour - 12;
	}
	else {
		hr = ltime->tm_hour;
	}

	min = ltime->tm_min / 12;
	if (!iconified) {
		draw_hand_poly(&hour_hand, (hr * 5) + min, hhand, hshand);
	}
	else {
		if (clockdata.old_hour != ltime->tm_hour || (clockdata.old_min / 12) != (ltime->tm_min / 12)) {
			calc_hand_coord((hr * 5) + min, hhand);
			PenMode(srcOr);
			PenSize( 2, 2 );
			MoveTo( cleft + clockdata.center, ctop + clockdata.center );
			LineTo( cleft + clockdata.center + clockdata.h_xcord, ctop + clockdata.center + clockdata.h_ycord );
		}
	}

	if (iconified) {
		if (clockdata.old_min != ltime->tm_min || clockdata.old_hour != ltime->tm_hour || (clockdata.old_min / 12) != (ltime->tm_min / 12)) {
			PenMode(srcOr);
			watch_cntr.edge.top =  icnrect.edge.top + ((icnrect.edge.right - icnrect.edge.left) / 2) - 1;
			watch_cntr.edge.left = icnrect.edge.left + ((icnrect.edge.right - icnrect.edge.left) / 2) - 1;
			watch_cntr.edge.bottom = watch_cntr.edge.top + 2;
			watch_cntr.edge.right = watch_cntr.edge.left + 2;
			FrameRect(&watch_cntr);
		}
	}

	clockdata.old_sec = ltime->tm_sec;
	clockdata.old_min = ltime->tm_min;
	clockdata.old_hour = ltime->tm_hour;
	clockdata.old_day = ltime->tm_mday;
	clockdata.old_mon = ltime->tm_mon + 1;
	clockdata.old_year = ltime->tm_year;
}

/*
** Draw the four major axis or time marks on the analog clock face.  They are
** boxes at 12, 3, 6, and 9.
*/
draw_hand_poly(hand, time, handlen, shandlen)
User_polys *hand;
int	time;
short handlen, shandlen;
{
	int	temptime_plus;
	int	temptime_minus;
	int	temptime_half;

	if (hand->user_poly == 0) {
		temptime_plus = time + 15;
		if (temptime_plus > 59) {
			temptime_plus -= 60;
		}
		temptime_half = time + 30;
		if (temptime_half > 59) {
			temptime_half -= 60;
		}
		temptime_minus = time - 15;
		if (temptime_minus < 0) {
			temptime_minus += 60;
		}

		hand->temp_poly = OpenPoly();
		calc_hand_coord(temptime_half, shandlen);
		MoveTo(cleft + clockdata.center + clockdata.h_xcord, ctop + clockdata.center + clockdata.h_ycord);
		calc_hand_coord(temptime_minus, shandlen / 2);
		LineTo(cleft + clockdata.center + clockdata.h_xcord, ctop + clockdata.center + clockdata.h_ycord);
		calc_hand_coord(time, handlen);
		LineTo(cleft + clockdata.center + clockdata.h_xcord, ctop + clockdata.center + clockdata.h_ycord);
		calc_hand_coord(temptime_plus, shandlen / 2);
		LineTo(cleft + clockdata.center + clockdata.h_xcord, ctop + clockdata.center + clockdata.h_ycord);
		calc_hand_coord(temptime_half, shandlen);
		LineTo(cleft + clockdata.center + clockdata.h_xcord, ctop + clockdata.center + clockdata.h_ycord);
		ClosePoly();
		hand->user_poly = DupePoly(hand->temp_poly);
	}
	FillPoly(hand->user_poly, blackPattern);
	hand->present_state = ON;
}
/*eject*/
/*
**	Draw the clock face
*/
draw_clk_face()
{
	char	str[]="0123456789101112";
	short	xpos, ypos;
	int		i;
	int		twidth;
	Rect	box;
	
	PenSize(clockdata.pwidth, clockdata.pheight);
	if (clockdata.time_mode == ANALOG) {
		FrameOval(&clockdata.watch_box);
		for (i=1; i<=12; i++) {

		calc_hand_coord(i * 5, clockdata.radius);
		xpos = clockdata.watch_box.edge.left + clockdata.center + clockdata.h_xcord;
		ypos = clockdata.watch_box.edge.top + clockdata.center + clockdata.h_ycord;

			twidth = TextWidth(str, (i>9) ? i+(i%10) : i, (i>9) ? 2 : 1);
			if (i > 3 && i < 9) {
				MoveTo( xpos - (twidth / 2), ypos + (finfo.ascent / 2) - (3 * clockdata.fudge_factor));
			}
			else if (i < 2 && i > 10) {
				MoveTo( xpos - (twidth / 2), ypos + (finfo.ascent / 2) + (5 * clockdata.fudge_factor));
			}
			else if (i == 3 || i == 9) {
				MoveTo( xpos - (twidth / 2), ypos + (finfo.ascent / 3));
			}
			else {
				MoveTo( xpos - (twidth / 2), ypos + (finfo.ascent / 2));
			}
			DrawText(str, (i>9) ? i+(i%10) : i, (i>9) ? 2 : 1);
		}
	}
	else if (clockdata.time_mode == ANABOX) {
		for (i=1; i<=60; i++) {
			calc_hand_coord(i, clockdata.radius);
			xpos = clockdata.watch_box.edge.left + clockdata.center + clockdata.h_xcord;
			ypos = clockdata.watch_box.edge.top + clockdata.center + clockdata.h_ycord;

			if (i % 5 == 0) {
				MoveTo(xpos, ypos);
				box.edge.left = xpos - (2 * clockdata.pwidth);
				box.edge.top = ypos - (2 * clockdata.pheight);
				box.edge.right = xpos + (2 * clockdata.pwidth);
				box.edge.bottom = ypos + (2 * clockdata.pheight);
				FillRect(&box, blackPattern);
			}
			else {
				if (clockdata.pwidth >= 2) {
					box.edge.left = xpos - 2;
					box.edge.top = ypos - 2;
					box.edge.right = xpos + 2;
					box.edge.bottom = ypos + 2;
				}
				else {
					box.edge.left = xpos - 1;
					box.edge.top = ypos - 1;
					box.edge.right = xpos + 1;
					box.edge.bottom = ypos + 1;
				}
				FillOval(&box, blackPattern);
			}
		}
	}
	else if (clockdata.time_mode != OFF) {
		FrameRect(&clockdata.watch_box);
		box.edge.top = clockdata.watch_box.edge.top + clockdata.pheight;
		box.edge.left = clockdata.watch_box.edge.left + clockdata.pwidth;
		box.edge.bottom = clockdata.watch_box.edge.bottom - clockdata.pheight;
		box.edge.right = clockdata.watch_box.edge.right - clockdata.pwidth;
		HiliteRect(&box, greyPattern, clockdata.color_mode);
	}
}

