
#include <time.h>
#include <isc/window.h>
#include "clock.h"

extern struct timesetup	clockdata;
extern struct tm *ltime;
extern Clock digital;
extern Date ddigital;

/*
** The changed value indicates which digits have changed.  Using that value,
** either erase or draw the appropriate element. ** save || digit->led[segment].present_state == INITIALIZE
*/
do_leds(digit, changed, xcoord, ycoord, state)
Digits *digit;
int	changed;
int xcoord, ycoord;
int state;
{
	int	i, segment;
	int	mask;

	for (i = 0x01, segment = 0; i < 0x80; i <<= 1, segment++) {
		mask = changed & i;
		switch (mask) {
			case 0x01:
				if (digit->led[segment].present_state == ON) {
					erase_this_poly(&digit->led[segment]);
					digit->digit_mask &= ~mask;
				}
				else {
					draw_aseg_poly(&digit->led[segment], xcoord + (2 * clockdata.scale_factor), ycoord, clockdata.scale_factor, state);
					digit->digit_mask |= mask;
				}
				break;
			case 0x02:
				if (digit->led[segment].present_state == ON) {
					erase_this_poly(&digit->led[segment]);
					digit->digit_mask &= ~mask;
				}
				else {
					draw_bseg_poly(&digit->led[segment], xcoord + (11 * clockdata.scale_factor), ycoord + (1 * clockdata.scale_factor), clockdata.scale_factor, state);
					digit->digit_mask |= mask;
				}
				break;
			case 0x04:
				if (digit->led[segment].present_state == ON) {
					erase_this_poly(&digit->led[segment]);
					digit->digit_mask &= ~mask;
				}
				else {
					draw_cseg_poly(&digit->led[segment], xcoord + (11 * clockdata.scale_factor), ycoord + (12 * clockdata.scale_factor), clockdata.scale_factor, state);
					digit->digit_mask |= mask;
				}
				break;
			case 0x08:
				if (digit->led[segment].present_state == ON) {
					erase_this_poly(&digit->led[segment]);
					digit->digit_mask &= ~mask;
				}
				else {
					draw_dseg_poly(&digit->led[segment], xcoord + (2 * clockdata.scale_factor), ycoord + (22 * clockdata.scale_factor), clockdata.scale_factor, state);
					digit->digit_mask |= mask;
				}
				break;
			case 0x10:
				if (digit->led[segment].present_state == ON) {
					erase_this_poly(&digit->led[segment]);
					digit->digit_mask &= ~mask;
				}
				else {
					draw_eseg_poly(&digit->led[segment], xcoord, ycoord + (12 * clockdata.scale_factor), clockdata.scale_factor, state);
					digit->digit_mask |= mask;
				}
				break;
			case 0x20:
				if (digit->led[segment].present_state == ON) {
					erase_this_poly(&digit->led[segment]);
					digit->digit_mask &= ~mask;
				}
				else {
					draw_fseg_poly(&digit->led[segment], xcoord, ycoord + (1 * clockdata.scale_factor), clockdata.scale_factor, state);
					digit->digit_mask |= mask;
				}
				break;
			case 0x40:
				if (digit->led[segment].present_state == ON) {
					erase_this_poly(&digit->led[segment]);
					digit->digit_mask &= ~mask;
				}
				else {
					draw_gseg_poly(&digit->led[segment], xcoord + (2 * clockdata.scale_factor), ycoord + (11 * clockdata.scale_factor), clockdata.scale_factor, state);
					digit->digit_mask |= mask;
				}
				break;
		}
	}
}

/*
** draw the colons only if they haven't already been drawn
*/
do_colons(digit, xcoord, ycoord, state)
Colons *digit;
int xcoord, ycoord;
int state;
{
	if (digit->top.present_state == OFF) {
		draw_colon_poly(&digit->top, xcoord, ycoord + (7 * clockdata.scale_factor), clockdata.scale_factor, state);
	}
	if (digit->bot.present_state == OFF) {
		draw_colon_poly(&digit->bot, xcoord, ycoord + (15 * clockdata.scale_factor), clockdata.scale_factor, state);
	}
}

/*
** draw the slash only if it hasn't already been drawn
*/
do_slash(digit, xcoord, ycoord, state)
Slash *digit;
int xcoord, ycoord;
int state;
{
	if (digit->slant.present_state == OFF) {
		draw_slash_poly(&digit->slant, xcoord, ycoord, clockdata.scale_factor, state);
	}
}

/*
** Given the Element type (hour, minute or second) and the x and y coordinates,
** determine which segments to change and change them.
*/
do_segments(hms, xcoord, ycoord, state)
int	hms;
int	xcoord, ycoord;
int state;
{
	/*
	** The following table defines the active elements for a give decimal
	** number from 0 to 9.  This value is used as a mask to find out which
	** segments have changed from one time to the next
	**
	**				| Active Segment | Value
	**				| g f e d c b a  |
	**			 ---+----------------+-----
	**			  0 | 0 1 1 1 1 1 1  |  3F
	**			  1 | 0 0 0 0 1 1 0  |  06
	**			N 2 | 1 0 1 1 0 1 1  |  5B
	**			u 3 | 1 0 0 1 1 1 1  |  4F
	**			m 4 | 1 1 0 0 1 1 0  |  66
	**			b 5 | 1 1 0 1 1 0 1  |  6D
	**			e 6 | 1 1 1 1 1 0 1  |  7D
	**			r 7 | 0 0 0 0 1 1 1  |  07
	**			  8 | 1 1 1 1 1 1 1  |  7F
	**			  9 | 1 1 0 1 1 1 1  |  6F
	**
	**
	*/
	int	d_table[] = {0x3f, 0x06, 0x5b, 0x4f, 0x66,	/* active segment elements */
					 0x6d, 0x7d, 0x07, 0x7f, 0x6f};	/* for values 0-9 */

	int diffl, diffm, most, most_mask, least, least_mask;
	int	time;

	switch(hms) {
		case SEC:
			time = ltime->tm_sec;
			break;
		case MIN:
			time = ltime->tm_min;
			break;
		case HOUR:
			switch (clockdata.time_mode) {
				case HMS2:
				case HM2:
					time = ltime->tm_hour;
					break;
				case HMS:
				case HM:
				case AMPM:
					if (ltime->tm_hour>= 13 ) {
						time = ltime->tm_hour - 12;
					}
					else {
						time = ltime->tm_hour;
					}
					break;
			}
			break;
		case DAY:
			time = ltime->tm_mday;
			break;
		case MONTH:
			time = ltime->tm_mon;
			time = time + 1;
			break;
		case YEAR:
			time = ltime->tm_year;
			break;
	}

	/*
	** The time value divided by 10 gives the tens digit or most.
	** The time value modulo 10 gives the ones digit or least.
	** Get the appropriate mask value for the most and least digits.
	*/
	most = time / 10;
	most_mask = d_table[most];
	least = time % 10;
	least_mask = d_table[least];

	/*
	** Now calculate which segments have changed (diffl & diffm) and go update
	** the display accordingly
	*/
	switch(hms) {
		case SEC:
			if (state == INITIALIZE) {
				diffl= 0x7f;
				diffm = 0x7f;
			}
			else {
				diffl = least_mask ^ digital.seconds.least.digit_mask;
				diffm = most_mask ^ digital.seconds.most.digit_mask;
			}
			switch (clockdata.time_mode) {
				case HMS2:
				case HMS:
				do_leds(&digital.seconds.least, diffl, xcoord + (20 * clockdata.scale_factor), ycoord, state);
				do_leds(&digital.seconds.most, diffm, xcoord + (5 * clockdata.scale_factor), ycoord, state);
				do_colons(&digital.seconds.colon, xcoord, ycoord, state);
				break;
			}
			break;

		case MIN:
			if (state == INITIALIZE) {
				diffl= 0x7f;
				diffm = 0x7f;
			}
			else {
				diffl = least_mask ^ digital.minutes.least.digit_mask;
				diffm = most_mask ^ digital.minutes.most.digit_mask;
			}
			do_leds(&digital.minutes.least, diffl, xcoord + (20 * clockdata.scale_factor), ycoord, state);
			do_leds(&digital.minutes.most, diffm, xcoord + (5 * clockdata.scale_factor), ycoord, state);
			do_colons(&digital.minutes.colon, xcoord, ycoord, state);
			break;

		case HOUR:
			if (state == INITIALIZE) {
				diffl= 0x7f;
				diffm = 0x7f;
			}
			else {
				diffl = least_mask ^ digital.hours.least.digit_mask;
				diffm = most_mask ^ digital.hours.most.digit_mask;
			}
			do_leds(&digital.hours.least, diffl, xcoord + (15 * clockdata.scale_factor), ycoord, state);
			do_leds(&digital.hours.most, diffm, xcoord, ycoord, state);
			break;

		case DAY:
			if (state == INITIALIZE) {
				diffl= 0x7f;
				diffm = 0x7f;
			}
			else {
				diffl = least_mask ^ ddigital.day.least.digit_mask;
				diffm = most_mask ^ ddigital.day.most.digit_mask;
			}
			if (clockdata.date_mode == EUR) {
				do_leds(&ddigital.day.least, diffl, xcoord + (15 * clockdata.scale_factor), ycoord, state);
				do_leds(&ddigital.day.most, diffm, xcoord , ycoord, state);
			}
			else {
				do_leds(&ddigital.day.least, diffl, xcoord + (20 * clockdata.scale_factor), ycoord, state);
				do_leds(&ddigital.day.most, diffm, xcoord + (5 * clockdata.scale_factor), ycoord, state);
				do_slash(&ddigital.day.slash, xcoord, ycoord, state);
			}
			break;

		case MONTH:
			if (state == INITIALIZE) {
				diffl= 0x7f;
				diffm = 0x7f;
			}
			else {
				diffl = least_mask ^ ddigital.month.least.digit_mask;
				diffm = most_mask ^ ddigital.month.most.digit_mask;
			}
			if (clockdata.date_mode == USA) {
				do_leds(&ddigital.month.least, diffl, xcoord + (15 * clockdata.scale_factor), ycoord, state);
				do_leds(&ddigital.month.most, diffm, xcoord, ycoord, state);
			}
			else {
				do_leds(&ddigital.month.least, diffl, xcoord + (20 * clockdata.scale_factor), ycoord, state);
				do_leds(&ddigital.month.most, diffm, xcoord + (5 * clockdata.scale_factor), ycoord, state);
				do_slash(&ddigital.month.slash, xcoord, ycoord, state);
			}
			break;
		case YEAR:
			if (state == INITIALIZE) {
				diffl= 0x7f;
				diffm = 0x7f;
			}
			else {
				diffl = least_mask ^ ddigital.year.least.digit_mask;
				diffm = most_mask ^ ddigital.year.most.digit_mask;
			}
			if (clockdata.date_mode == USA ||
				clockdata.date_mode == EUR) {
				do_leds(&ddigital.year.least, diffl, xcoord + (20 * clockdata.scale_factor), ycoord, state);
				do_leds(&ddigital.year.most, diffm, xcoord + (5 * clockdata.scale_factor), ycoord, state);
				do_slash(&ddigital.year.slash, xcoord, ycoord, state);
			}
			else {
				do_leds(&ddigital.year.least, diffl, xcoord + (15 * clockdata.scale_factor), ycoord, state);
				do_leds(&ddigital.year.most, diffm, xcoord, ycoord, state);
			}
			break;
	}
}
