
#include <isc/window.h>
#include "clock.h"

extern struct timesetup	clockdata;

/*
** The poly clearing routines are defined first
*/

erase_this_poly(poly)
User_polys *poly;
{
	if (poly->user_poly > 0) {
		ErasePoly(poly->user_poly);
	}
	poly->present_state = OFF;
}

clear_this_poly(poly, state)
User_polys *poly;
int	state;
{
	if (poly->user_poly > 0) {
		ErasePoly(poly->user_poly);
		KillPoly(poly->user_poly);
		KillPoly(poly->temp_poly);
	}
	poly->temp_poly = 0;
	poly->user_poly = 0;
	poly->present_state = OFF;
}

clear_all_digits(digit, state)
Digits *digit;
int	state;
{
	int	i;

	for (i = 0; i < 7; i++) {
		clear_this_poly(&digit->led[i], state);
	}
	digit->digit_mask = 0;
}

clear_all_colons(colon, state)
Colons *colon;
int	state;
{
	clear_this_poly(&colon->top, state);
	clear_this_poly(&colon->bot, state);
}

clear_a_slash(slash, state)
Slash *slash;
int state;
{
	clear_this_poly(&slash->slant, state);
}

clear_clk_segments(segment, state)
Segments *segment;
int	state;
{
	clear_all_digits(&segment->least, state);
	clear_all_digits(&segment->most, state);
	clear_all_colons(&segment->colon, state);
}

clear_dat_segments(segment, state)
DSegments *segment;
int	state;
{
	clear_all_digits(&segment->least, state);
	clear_all_digits(&segment->most, state);
	clear_a_slash(&segment->slash, state);
}

clear_clock(aclock, state)
Clock *aclock;
int	state;
{
	clear_clk_segments(&aclock->seconds, state);
	clear_clk_segments(&aclock->minutes, state);
	clear_clk_segments(&aclock->hours, state);
}

clear_date(adate, state)
Date *adate;
int	state;
{
	clear_dat_segments(&adate->day, state);
	clear_dat_segments(&adate->month, state);
	clear_dat_segments(&adate->year, state);
}

/*eject*/
/*
** The poly drawing routines are next
**
** Draw the led segments for the digital clock.  Segments are selected by the
** segment pointer, the coordinates and a drawing scale.  If a segment has
** not been created, it is and then filled in, otherwise it is just filled in.
*/

/*
** The a segment at a scale of one looks like this
**
**           ********** 
**          *          *
**           **********  
**
*/
draw_aseg_poly(segment, xcord, ycord, scale, state)
User_polys *segment;
short xcord, ycord;
short scale;
int state;
{
	if (segment->user_poly == 0) {
		segment->temp_poly = OpenPoly();
		MoveTo(xcord + ( 1 * scale), ycord + (0 * scale));
		LineTo(xcord + ( 8 * scale), ycord + (0 * scale));
		LineTo(xcord + ( 9 * scale), ycord + (1 * scale));
		LineTo(xcord + ( 8 * scale), ycord + (2 * scale));
		LineTo(xcord + ( 1 * scale), ycord + (2 * scale));
		LineTo(xcord + ( 0 * scale), ycord + (1 * scale));
		LineTo(xcord + ( 1 * scale), ycord + (0 * scale));
		ClosePoly();
		segment->user_poly = DupePoly(segment->temp_poly);
		segment->present_state = INITIALIZE;
	}
	if (state == ON) {
		FillPoly(segment->user_poly, greyPattern);
		segment->present_state = ON;
	}
}

/*
** The b segment at a scale of one looks like this
**
**           * 
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**           *
**
*/
draw_bseg_poly(segment, xcord, ycord, scale, state)
User_polys *segment;
short xcord, ycord;
short scale;
int state;
{
	if (segment->user_poly == 0) {
		segment->temp_poly = OpenPoly();
		MoveTo(xcord + ( 1 * scale), ycord + ( 0 * scale));
		LineTo(xcord + ( 2 * scale), ycord + ( 1 * scale));
		LineTo(xcord + ( 2 * scale), ycord + (10 * scale));
		LineTo(xcord + ( 1 * scale), ycord + (11 * scale));
		LineTo(xcord + ( 0 * scale), ycord + (10 * scale));
		LineTo(xcord + ( 0 * scale), ycord + ( 1 * scale));
		LineTo(xcord + ( 1 * scale), ycord + ( 0 * scale));
		ClosePoly();
		segment->user_poly = DupePoly(segment->temp_poly);
		segment->present_state = INITIALIZE;
	}
	if (state == ON) {
		FillPoly(segment->user_poly, greyPattern);
		segment->present_state = ON;
	}
}

/*
** The c segment at a scale of one looks like this
**
**           * 
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**           * 
**
*/
draw_cseg_poly(segment, xcord, ycord, scale, state)
User_polys *segment;
short xcord, ycord;
short scale;
int state;
{
	if (segment->user_poly == 0) {
		segment->temp_poly = OpenPoly();
		MoveTo(xcord + ( 2 * scale), ycord + ( 1 * scale));
		LineTo(xcord + ( 2 * scale), ycord + (10 * scale));
		LineTo(xcord + ( 1 * scale), ycord + (11 * scale));
		LineTo(xcord + ( 0 * scale), ycord + (10 * scale));
		LineTo(xcord + ( 0 * scale), ycord + ( 1 * scale));
		LineTo(xcord + ( 1 * scale), ycord + ( 0 * scale));
		LineTo(xcord + ( 2 * scale), ycord + ( 1 * scale));
		ClosePoly();
		segment->user_poly = DupePoly(segment->temp_poly);
		segment->present_state = INITIALIZE;
	}
	if (state == ON) {
		FillPoly(segment->user_poly, greyPattern);
		segment->present_state = ON;
	}
}

/*
** The d segment at a scale of one looks like this
**
**           **********
**          *          *
**           **********
**
*/
draw_dseg_poly(segment, xcord, ycord, scale, state)
User_polys *segment;
short xcord, ycord;
short scale;
int state;
{
	if (segment->user_poly == 0) {
		segment->temp_poly = OpenPoly();
		MoveTo(xcord + ( 1 * scale), ycord + ( 0 * scale));
		LineTo(xcord + ( 8 * scale), ycord + ( 0 * scale));
		LineTo(xcord + ( 9 * scale), ycord + ( 1 * scale));
		LineTo(xcord + ( 8 * scale), ycord + ( 2 * scale));
		LineTo(xcord + ( 1 * scale), ycord + ( 2 * scale));
		LineTo(xcord + ( 0 * scale), ycord + ( 1 * scale));
		LineTo(xcord + ( 1 * scale), ycord + ( 0 * scale));
		ClosePoly();
		segment->user_poly = DupePoly(segment->temp_poly);
		segment->present_state = INITIALIZE;
	}
	if (state == ON) {
		FillPoly(segment->user_poly, greyPattern);
		segment->present_state = ON;
	}
}

/*
** The e segment at a scale of one looks like this
**
**           *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**           *
**
*/
draw_eseg_poly(segment, xcord, ycord, scale, state)
User_polys *segment;
short xcord, ycord;
short scale;
int state;
{
	if (segment->user_poly == 0) {
		segment->temp_poly = OpenPoly();
		MoveTo(xcord + ( 1 * scale), ycord + ( 0 * scale));
		LineTo(xcord + ( 2 * scale), ycord + ( 1 * scale));
		LineTo(xcord + ( 2 * scale), ycord + (10 * scale));
		LineTo(xcord + ( 1 * scale), ycord + (11 * scale));
		LineTo(xcord + ( 0 * scale), ycord + (10 * scale));
		LineTo(xcord + ( 0 * scale), ycord + ( 1 * scale));
		LineTo(xcord + ( 1 * scale), ycord + ( 0 * scale));
		ClosePoly();
		segment->user_poly = DupePoly(segment->temp_poly);
		segment->present_state = INITIALIZE;
	}
	if (state == ON) {
		FillPoly(segment->user_poly, greyPattern);
		segment->present_state = ON;
	}
}

/*
** The f segment at a scale of one looks like this
**
**           *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**          * *
**           *
**
*/
draw_fseg_poly(segment, xcord, ycord, scale, state)
User_polys *segment;
short xcord, ycord;
short scale;
int state;
{
	if (segment->user_poly == 0) {
		segment->temp_poly = OpenPoly();
		MoveTo(xcord + ( 1 * scale), ycord + ( 0 * scale));
		LineTo(xcord + ( 2 * scale), ycord + ( 1 * scale));
		LineTo(xcord + ( 2 * scale), ycord + (10 * scale));
		LineTo(xcord + ( 1 * scale), ycord + (11 * scale));
		LineTo(xcord + ( 0 * scale), ycord + (10 * scale));
		LineTo(xcord + ( 0 * scale), ycord + ( 1 * scale));
		LineTo(xcord + ( 1 * scale), ycord + ( 0 * scale));
		ClosePoly();
		segment->user_poly = DupePoly(segment->temp_poly);
		segment->present_state = INITIALIZE;
	}
	if (state == ON) {
		FillPoly(segment->user_poly, greyPattern);
		segment->present_state = ON;
	}
}


/*
** The g segment at a scale of one looks like this
**
**           ********  
**          *        * 
**           ********   
**
*/
draw_gseg_poly(segment, xcord, ycord, scale, state)
User_polys *segment;
short xcord, ycord;
short scale;
int state;
{
	if (segment->user_poly == 0) {
		segment->temp_poly = OpenPoly();
		MoveTo(xcord + ( 1 * scale), ycord + ( 0 * scale));
		LineTo(xcord + ( 8 * scale), ycord + ( 0 * scale));
		LineTo(xcord + ( 9 * scale), ycord + ( 1 * scale));
		LineTo(xcord + ( 8 * scale), ycord + ( 2 * scale));
		LineTo(xcord + ( 1 * scale), ycord + ( 2 * scale));
		LineTo(xcord + ( 0 * scale), ycord + ( 1 * scale));
		LineTo(xcord + ( 1 * scale), ycord + ( 0 * scale));
		ClosePoly();
		segment->user_poly = DupePoly(segment->temp_poly);
		segment->present_state = INITIALIZE;
	}
	if (state == ON) {
		FillPoly(segment->user_poly, greyPattern);
		segment->present_state = ON;
	}
}


/*
** The colon segments at a scale of one looks like this
**
**          ****
**          *  *
**          *  *
**          ****
**
*/
draw_colon_poly(colon, xcord, ycord, scale, state)
User_polys *colon;
short xcord, ycord;
short scale;
int state;
{
	if (colon->user_poly == 0) {
		colon->temp_poly = OpenPoly();
		MoveTo(xcord, ycord);
		LineTo(xcord + ( 3 * scale), ycord + ( 0 * scale));
		LineTo(xcord + ( 3 * scale), ycord + ( 3 * scale));
		LineTo(xcord + ( 0 * scale), ycord + ( 3 * scale));
		LineTo(xcord + ( 0 * scale), ycord + ( 0 * scale));
		ClosePoly();
		colon->user_poly = DupePoly(colon->temp_poly);
		colon->present_state = INITIALIZE;
	}
	if (state == ON) {
		FillPoly(colon->user_poly, greyPattern);
		colon->present_state = ON;
	}
}


/*
** The slash segments at a scale of one looks like this
**
**       ***
**       * *
**       * *
**       * *
**       * *
**       * *
**       * *
**       * *
**      * *
**      * *
**      * *
**      * *
**      * *
**      * *
**      * *
**     * *
**     * *
**     * *
**     * *
**     * *
**     * *
**     * *
**     ***
**
*/
draw_slash_poly(slash, xcord, ycord, scale, state)
User_polys *slash;
short xcord, ycord;
short scale;
int state;
{
	if (slash->user_poly == 0) {
		slash->temp_poly = OpenPoly();
		MoveTo(xcord + ( 2 * scale), ycord + ( 0 * scale));
		LineTo(xcord + ( 4 * scale), ycord + ( 0 * scale));
		LineTo(xcord + ( 2 * scale), ycord + (22 * scale));
		LineTo(xcord + ( 0 * scale), ycord + (22 * scale));
		LineTo(xcord + ( 2 * scale), ycord + ( 0 * scale));
		ClosePoly();
		slash->user_poly = DupePoly(slash->temp_poly);
		slash->present_state = INITIALIZE;
	}
	if (state == ON) {
		FillPoly(slash->user_poly, greyPattern);
		slash->present_state = ON;
	}
}
