/*
**	I N C L U D E S
#include <string.h>
*/
#include <stdio.h>
#include <time.h>
#include <fcntl.h>
#include <dnix/fcodes.h>
#include <isc/keycodes.h>
#include <isc/window.h>
#include <isc/controls.h>
#include <isc/textedit.h>
#include "clock.h"

/*
**	E X T E R N S
*/
extern	int			errno;
extern	int			iconified;
extern	Time_setup	clockdata;
extern	short		width;
extern	short		height;
extern	FontInfo	finfo;
extern	Rect		icnrect;		/* Window rectangles	*/

short	ctop, cleft;
short	shand;						/*	second hand length	*/
short	mhand;						/*	minute hand length	*/
short	hhand;						/*	hour hand length	*/
short	mshand, hshand;

initial_window_size()
{
	if (clockdata.time_mode != ANALOG ||
		clockdata.time_mode != ANABOX) {
		width = MINDWIDTH;
		height = MINDHEIGHT;
	}
	else {
		width = MINAWIDTH;
		height = MINAHEIGHT;
	}
}

calc_window_size()
{
#ifdef DEBUG
	fprintf(stderr, "enter calc_window_size\n");
#endif
	if (clockdata.time_mode == ANALOG ||
		clockdata.time_mode == ANABOX) {
		if(width < height) {
			height = width;
		}
		if(width > height) {
			width = height;
		}
		if (width < MINAWIDTH) {
			width = MINAWIDTH;
		}
		else if (width > MAXWIDTH) {
			width = MAXWIDTH;
		}
		if (height < MINAHEIGHT) {
			height = MINAHEIGHT;
		}
		else if (height > MAXHEIGHT) {
			height = MAXHEIGHT;
		}
	}
	else {
		if (width < MINDWIDTH) {
			width = MINDWIDTH;
		}
		else if (width > MAXWIDTH) {
			width = MAXWIDTH;
		}
		if (height < MINDHEIGHT) {
			height = MINDHEIGHT;
		}
		else if (height > MAXHEIGHT) {
			height = MAXHEIGHT;
		}
	}
	do_size_calc();
	if (clockdata.time_mode != ANALOG &&
		clockdata.time_mode != ANABOX) {
		if (width < (98 * clockdata.scale_factor) + 30) {
			width = (98 * clockdata.scale_factor) + 40;
		}
	}
#ifdef DEBUG
	fprintf(stderr, "exit calc_window_size\n");
#endif
}

/*
** This routine does the size calculations for the clock display.
*/
do_size_calc()
{
	short	value;
	int		font_size;

	/*
	**	set smallest dimension to work with
	*/
	if (!iconified &&
		(clockdata.time_mode != ANALOG &&
		 clockdata.time_mode != ANABOX)) {
		value = height;
		clockdata.pwidth = clockdata.pheight = (value + 48) / 112;	/* four sizes of pen */
		clockdata.fudge_factor = (value + 26) / 90;		/* five sizes of fonts */
		clockdata.scale_factor = (value + 7) / 71;		/* six sizes of scaling */

		clockdata.watch_box.edge.top = (height - ((22 * clockdata.scale_factor) + 30)) / 2;
		clockdata.watch_box.edge.bottom = clockdata.watch_box.edge.top + (22 * clockdata.scale_factor) + 30;
		clockdata.watch_box.edge.left = (width - ((98 * clockdata.scale_factor) + 30)) / 2;
		clockdata.watch_box.edge.right = clockdata.watch_box.edge.left + (98 * clockdata.scale_factor) + 30;
	}
	else {
		value = height - 20;
		if (height > width) {
			value = width - 20;
		}
		clockdata.pwidth = clockdata.pheight = (value + 20) / 112;	/* four sizes of pen */
		clockdata.fudge_factor = (value + 20) / 90;		/* five sizes of fonts */
		clockdata.scale_factor = (value + 20) / 75;		/* six sizes of scaling */

		clockdata.watch_box.edge.top =  (height - value) / 2;
		clockdata.watch_box.edge.bottom = clockdata.watch_box.edge.top + value;
		clockdata.watch_box.edge.left = (width - value) / 2;
		clockdata.watch_box.edge.right = clockdata.watch_box.edge.left + value;

		/*
		**	calculate correct font size to use
		*/
		font_size = 12 * clockdata.fudge_factor;
		TextFont(8);
		TextSize(font_size);
		GetFontInfo(&finfo);
	}

	if (iconified) {
		clockdata.center = (icnrect.edge.right - icnrect.edge.left) / 2;
		ctop = icnrect.edge.top;
		cleft= icnrect.edge.left;
	}
	else {
		clockdata.center = (clockdata.watch_box.edge.right - clockdata.watch_box.edge.left) / 2;
		ctop = clockdata.watch_box.edge.top;
		cleft= clockdata.watch_box.edge.left;
	}

	if (iconified || clockdata.time_mode == ANALOG) {
		clockdata.radius = clockdata.center - (finfo.ascent / 2);
		shand = (clockdata.center * 11) / 16;				/* second hand length */
		mhand = (clockdata.center * 10) / 16;				/* minute hand length */
		hhand = (clockdata.center * 7) / 16;				/* hour hand length */
		mshand = mhand / 5;						/* minute hand segment lengths */ 
		hshand = hhand / 5;						/* hour hand segment length */
	}
	else if (clockdata.time_mode == ANABOX) {
		clockdata.radius = clockdata.center;
		shand = (clockdata.center * 15) / 16;				/* second hand length */
		mhand = (clockdata.center * 14) / 16;				/* minute hand length */
		hhand = (clockdata.center * 11) / 16;				/* hour hand length */
		mshand = mhand / 5;						/* minute hand segment lengths */ 
		hshand = hhand / 5;						/* hour hand segment length */
	}
}
