/*
** Color Picker program.  Allows you to set the color palette to RGB or HSB
** colors picked visually.
**
** For use by the ColorPicker() and NWColorPicker() libwindow calls.
**
**	V1.1	Jim Cathey	8/13/91
**		Fixed bug in HSV conversion.
**
**	V1.0	Jim Cathey	3/4/91
**		Made out of Windowsetup program.  It's kind of a hack.
**
** Color transformations (HSB <-> RGB) from _Color Considerations_ 
** by Lee Baldwin, September 1984 Byte, pp 227.
*/

#include <stdio.h>
#include <isc/window.h>
#include <isc/keycodes.h>
#include <isc/controls.h>

#define VERSION ("V 1.1")

void RGBxHSV(double R, double G, double B, double *H, double *S, double *V);
void HSVxRGB(double h, double s, double v, double *R, double *G, double *B);
char *mysprintf();

#define WIDTH (445)		/* Window definitions. */
#define HEIGHT (144)

static int xpos = 96;
static int ypos = 144;
static int width = WIDTH;
static int height = HEIGHT;
static int firstactive = 1;

#define SETUPMENU 1

struct mycontrols {
    ControlID cid;
    short hilval;
} controls[14];
Rect swatches[8];		/* Color swatches. */
char names[8][4];
short colormodel;		/* 0 = HSB, 1 = RGB. */
short colornum;
struct palette {
    unsigned char index, red, green, blue;
} colors[8];


struct iostuff {		/* For Binary I/O. */
    long size;
    short ncolors;
    struct iopalette {
	unsigned short index, red, green, blue;
    } colors[8];
} instruct, outstruct;

HdwrType theHdwrRecord;
int window;
int wp6;			/* TRUE if WP6 color model. */
int newweb;			/* TRUE if WEB with the WP6 color model. */
int maxval;			/* 63 for wp6. */
int hueval;			/* 72 for wp6. */
int satval;			/* 50 for wp6. */

int binary;			/* TRUE if binary I/O vs Human. */
char *format = "%2d";
char *version = VERSION;

void updateevent();
void windowevent();
void menuevent();
void mouseevent();
void keyevent();
void activatevent();
void MyAction(ControlID theControl, short partCode);

main(argc, argv)
int argc;
char *argv[];
{
    int i, j, bnum;
    int trapqfd, fd;
    Rect winrect, button;
    char *argp, *progname;
    char buf[80];
    struct palette *p;
    struct mycontrols *mcp;

    progname = *argv;
    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'x':
		xpos = atoi(argp);
		break;
	      case 'y':
		ypos = atoi(argp);
		break;
	      case 'b':
		binary++;
		break;
	      case 'V':
		mysprintf(buf, "%s Version %s\n", progname, &version[2]);
		write(1, buf, StringLen(buf));
		exit(0);
		break;
	    }
	} else {
	    colornum = atoi(argp);
	}
    }
    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
	exit(1);
    }
    if ((fd = ConnectGraf(0)) < 0) {	/* Initialize graphics toolbox. */
	exit(2);
    }
    if (binary) {
	read(0, &i, sizeof i);	/* Read the input record. */
	i -= sizeof i;
	if ((sizeof instruct - sizeof instruct.size) < i)
	    i = sizeof instruct - sizeof instruct.size;
	read(0, &instruct.ncolors, i);
	instruct.size = i + sizeof i;
    } else {
	instruct.size = sizeof instruct;
	instruct.ncolors = 8;
    }
    defaultsetup();		/* Convert the input record. */

    Event(updateEvt, updateevent);
    Event(windowEvt, windowevent);
    Event(menuEvt, menuevent);
    Event(mouseEvt, mouseevent);
    Event(keyEvt, keyevent);
    Event(activateEvt, activatevent);

    SetRect(&winrect, xpos, ypos, xpos+WIDTH, ypos+HEIGHT);
    window = NewWindow(&winrect, "Color Picker", showWindow, documentProc,
		       frontWindow, hasGoAway|hasSendToBack|hasDrag, fd);
    SetPort(window);
    dosetup();			/* Send colors to WM. */
    GetHdwrType(&theHdwrRecord);
#if 0
    theHdwrRecord.colormodel = 3; /* For testing only. */
    theHdwrRecord.nplanes = 4;
    theHdwrRecord.montype = 2;
#endif
    wp6 = newweb = 0;
    switch (theHdwrRecord.colormodel) {
      case 4:			/* Are we a WEB w/WP6 colors? */
	newweb = 1;
      case 3:			/* Are we a WP6? */
	wp6 = 1;
    }

    maxval = 63;
    hueval = 72;		/* 5 deg/step for hue. */
    satval = 50;		/* 2% per step for sat/bri. */
    if (!wp6) {
	firstmenu(1);
    } else {
	j = 0;			/* Properly handle monochrome units, and */
	if (theHdwrRecord.nplanes > 1) /* mono monitors on color units. */
	    j = 2 | (theHdwrRecord.montype > 1); /* (Montype > 1 for color.) */

	InitControls();
	colormodel = 0;		/* Default to HSB color model. */
	if (j < 3)
	    colormodel = 1;	/* Use RGB model for mono board or monitor. */
	i = 0;
	mcp = &controls[0];
	SetRect(&button, 10, 40, 60, 60);
	mcp->cid = NewControl(window, &button, "HSB", showControl,
			      !colormodel, 0, 1, radioButProc, i); /* #0 */
	if (j < 3)
	    HiliteControl(mcp->cid, mcp->hilval = 255);
	i++; mcp++;
	OffsetRect(&button, 0, 20);
	mcp->cid = NewControl(window, &button, "RGB", showControl,
			      colormodel, 0, 1, radioButProc, i); /* #1 */
	if (j < 2)
	    HiliteControl(mcp->cid, mcp->hilval = 255);
	i++; mcp++;
	/* Scroll bars should be 17 pixels in the skinny direction. */
	SetRect(&button, 82, 20, 99, 110);
	mcp->cid = NewControl(window, &button, "H/R", showControl,
			      0, 0, maxval, scrollBarProc, i); /* #2 */
	SetCtlAction(mcp->cid, MyAction);
	SetCtlScrollRate(mcp->cid, 5);
	if (j < 3)
	    HiliteControl(mcp->cid, mcp->hilval = 255);
	i++; mcp++;
	OffsetRect(&button, 30, 0);
	mcp->cid = NewControl(window, &button, "S/G", showControl,
			      0, 0, maxval, scrollBarProc, i); /* #3 */
	SetCtlAction(mcp->cid, MyAction);
	SetCtlScrollRate(mcp->cid, 5);
	if (j < 2)
	    HiliteControl(mcp->cid, mcp->hilval = 255);
	i++; mcp++;
	OffsetRect(&button, 30, 0);
	mcp->cid = NewControl(window, &button, "B/B", showControl,
			      0, 0, maxval, scrollBarProc, i); /* #4 */
	SetCtlAction(mcp->cid, MyAction);
	SetCtlScrollRate(mcp->cid, 5);
	if (j < 3)
	    HiliteControl(mcp->cid, mcp->hilval = 255);
	i++; mcp++;

	SetRect(&button, 204, 65, 240, 77);
	p = colors;
	bnum = 0;
	mysprintf(names[bnum], format, (p++)->index);
	mcp->cid = NewControl(window, &button, names[bnum], showControl,
			      0, 0, 1, radioButProc, i); /* #5 */
	if (j < 2)
	    HiliteControl(mcp->cid, mcp->hilval = 255);
	i++; mcp++;
	OffsetRect(&button, 60, 0);
	mysprintf(names[++bnum], format, (p++)->index);
	mcp->cid = NewControl(window, &button, names[bnum], showControl,
			      0, 0, 1, radioButProc, i); /* #6 */
	if (j < 2)
	    HiliteControl(mcp->cid, mcp->hilval = 255);
	i++; mcp++;
	OffsetRect(&button, 60, 0);
	mysprintf(names[++bnum], format, (p++)->index);
	mcp->cid = NewControl(window, &button, names[bnum], showControl,
			      0, 0, 1, radioButProc, i); /* #7 */
	if (j < 2)
	    HiliteControl(mcp->cid, mcp->hilval = 255);
	i++; mcp++;
	OffsetRect(&button, 60, 0);
	mysprintf(names[++bnum], format, (p++)->index);
	mcp->cid = NewControl(window, &button, names[bnum], showControl,
			      0, 0, 1, radioButProc, i); /* #8 */
	if (j < 2)
	    HiliteControl(mcp->cid, mcp->hilval = 255);
	i++; mcp++;
	SetRect(&button, 204, 120, 240, 132);
	mysprintf(names[++bnum], format, (p++)->index);
	mcp->cid = NewControl(window, &button, names[bnum], showControl,
			      0, 0, 1, radioButProc, i); /* #9 */
	if (j < 2)
	    HiliteControl(mcp->cid, mcp->hilval = 255);
	i++; mcp++;
	OffsetRect(&button, 60, 0);
	mysprintf(names[++bnum], format, (p++)->index);
	mcp->cid = NewControl(window, &button, names[bnum], showControl,
			      0, 0, 1, radioButProc, i); /* #10 */
	if (j < 2)
	    HiliteControl(mcp->cid, mcp->hilval = 255);
	i++; mcp++;
	OffsetRect(&button, 60, 0);
	mysprintf(names[++bnum], format, (p++)->index);
	mcp->cid = NewControl(window, &button, names[bnum], showControl,
			      0, 0, 1, radioButProc, i); /* #11 */
	if (j < 2)
	    HiliteControl(mcp->cid, mcp->hilval = 255);
	i++; mcp++;
	OffsetRect(&button, 60, 0);
	mysprintf(names[++bnum], format, (p++)->index);
	mcp->cid = NewControl(window, &button, names[bnum], showControl,
			      0, 0, 1, radioButProc, i); /* #12 */
	if (j < 2)
	    HiliteControl(mcp->cid, mcp->hilval = 255);
	mcp++;

	mcp->cid = 0;		/* End marker. */
	SetRect(&swatches[0], 200, 30, 240, 60);
	swatches[4] = swatches[0];
	swatches[1] = swatches[0];
	OffsetRect(&swatches[4], 0, 55);
	OffsetRect(&swatches[1], 60, 0);
	swatches[5] = swatches[1];
	swatches[2] = swatches[1];
	OffsetRect(&swatches[5], 0, 55);
	OffsetRect(&swatches[2], 60, 0);
	swatches[6] = swatches[2];
	swatches[3] = swatches[2];
	OffsetRect(&swatches[6], 0, 55);
	OffsetRect(&swatches[3], 60, 0);
	swatches[7] = swatches[3];
	OffsetRect(&swatches[7], 0, 55);

	newmenus();
	p = colors;		/* Find starting color number. */
	j = 0;
	for (i=0; i<instruct.ncolors; i++) {
	    if (colornum == p->index) {
		j = i;
		break;
	    }
	    p++;
	}
	setcolindex(5 + j, 0 + j);
    }
    if (GetEvent(anyWindow, noEvent)) /* Process user interaction. */
	return 1;		/* Continues after a ReleaseEvent (rets 0). */
				/* GetEvent returns EOF on connection error. */

    FlushTwice();		/* We need to make sure that all of our */
    savesetup();		/* data is actually delivered.  If we quit */
				/* too soon, Netman might not deliver our */
				/* last packet. */
    return !wp6;		/* Only return 0 if everything's OK. */
}

defaultsetup()
{
    short i;
    struct palette *p;
    struct iopalette *ip;

    p = colors;			/* Setup default color palette. */
    ip = &instruct.colors[0];
    for (i=0; i<instruct.ncolors; i++) {
	if (binary) {
	    p->index = ip->index;
	    p->red = ip->red >> 8;
	    p->green = ip->green >> 8;
	    p->blue = ip->blue >> 8;
	} else {
	    p->index = i + 8;
	    p->red = p->green = p->blue = 0;
	}
	p++;
	ip++;
    }
}

setsliders()
{
    redocontrols();
}

firstmenu(flag)
int flag;
{
    DisposeMenu(SETUPMENU);	/* Then build the menus themselves. */
    NewMenu(SETUPMENU, "Setup");
    AppendMenu(SETUPMENU, "/DDefault;-;/SSave;-;/QQuit");
    if (flag) {
	DisableItem(SETUPMENU, 1);
	DisableItem(SETUPMENU, 3);
    }
    if (binary)			/* So we don't confuse the pipes by doing */
	DisableItem(SETUPMENU, 3); /* extraneous I/O to them. */
}

newmenus()
{				/* Build the menus themselves. */
    firstmenu(0);
}

unsigned short			/* Convert unsigned char to unsigned short. */
mash(val)			/* This is because we work in uchars, but */
unsigned char val;		/* the system wants ushorts. */
{
    unsigned short retval;

    retval = val << 8;
    if (val == 0xFC)		/* Make sure we max out at the top. */
	retval = 0xFFFF;
    return retval;
}

savesetup()
{
    int i;
    char buf[800], *tp;
    struct palette *p;
    struct iopalette *ip;

    if (!wp6)
	return;
    p = colors;
    ip = &outstruct.colors[0];
    if (binary) {
	outstruct.size = instruct.size;
	outstruct.ncolors = instruct.ncolors;
	for (i=0; i < sizeof(colors)/sizeof(struct palette); i++) {
	    ip->index = p->index;
	    ip->red = mash(p->red);
	    ip->green = mash(p->green);
	    ip->blue = mash(p->blue);
	    p++;
	    ip++;
	}
	write(1, &outstruct, outstruct.size);
    }
    tp = buf;
    p = colors;
    for (i=0; i < sizeof(colors)/sizeof(struct palette); i++) {
	tp = mysprintf(tp, "Color %2d = [$%04X, $%04X, $%04X]\n",
		       p->index, mash(p->red), mash(p->green), mash(p->blue));
	p++;
    }
    *tp++ = '\n';
    *tp++ = 0;
    write(binary ? 2 : 1, buf, StringLen(buf));
}

/*
** Actually send the current color palette to the WM.
*/

dosetup()
{
    int i;
    struct palette *p;

    p = colors;
    for (i=0; i < sizeof(colors)/sizeof(struct palette); i++) {
	SetRGBColor(p->index, mash(p->red), mash(p->green), mash(p->blue));
	p++;
    }
}

void
updateevent(ep)
EventRecord *ep;
{
    SetPort(ep->who);
    BeginUpdate(ep->who);
    if (!wp6) {
	TextFont(3);
	TextSize(9);
	MoveTo(416, 9);
	DrawString(version);
	TextFont(0);
	TextSize(12);
	MoveTo(20, 70);
	DrawString("You can only pick colors on a Color WP6.  "
		   "I can\325t help you.");
    } else {
	if (!firstactive)	/* Skip first redraw of controls because */
	    DrawControls(ep->who); /* all controls were created drawn. */
	TextFace(underlineText);
	MoveTo(12, 12);
	DrawString("Color");
	MoveTo(12, 26);
	DrawString("Model");
	TextFace(normalText);
	TextFont(3);
	TextSize(9);
	MoveTo(55, 17);
	DrawString("F1");
	MoveTo(85, 14);
	DrawString("F2");
	MoveTo(115, 14);
	DrawString("F3");
	MoveTo(145, 14);
	DrawString("F4");
	MoveTo(5, 141);
	DrawString("Shifted F-key runs slider the other way.");
	MoveTo(416, 9);
	DrawString(version);
	TextFont(0);
	TextSize(12);
	drawlabels();
	drawswatches();
    }
    EndUpdate(ep->who);
}

/*
** Draw the color swatches.
*/

drawswatches()
{
    Rect scratch;
    short i;
    char *label = "Color Number";
    struct palette *p;

    scratch.edge.top = swatches[0].edge.top - 26;
    scratch.edge.bottom = swatches[0].edge.top - 10;
    scratch.edge.left = swatches[0].edge.left;
    scratch.edge.right = swatches[3].edge.right;
    TextFace(underlineText);
    DrawJustText(label, StringLen(label), &scratch,
		 centerJustify, centerJustify);
    TextFace(normalText);
    p = colors;
    for (i=0; i<8; i++) {
	scratch = swatches[i];
	BackColor(p->index);
	EraseRect(&scratch);
	FrameRect(&scratch);
	p++;
    }
    BackColor(0);
    PenNormal();
}

/*
** Label the sliders with the right labels for the input color model in use.
*/

drawlabels()
{
    Rect scratch;

    SetRect(&scratch, 79, 110, 166, 122);
    EraseRect(&scratch);
    MoveTo(80, 122);
    DrawText("HueRed", colormodel * 3, 3);
    MoveTo(110, 122);
    DrawText("SatGrn", colormodel * 3, 3);
    MoveTo(141, 122);
    DrawText("BriBlu", colormodel * 3, 3);
}

void
windowevent(ep)
EventRecord *ep;
{
    Rect rect;

    SetPort(ep->who);
    switch (ep->message & F_KEYMASK) {
      case windowGoaway:
	ReleaseEvent();
	break;

      case windowMove:
	xpos += ep->where.edge.left;
	ypos += ep->where.edge.top;
	MoveWindow(ep->who, xpos, ypos, 0);
	break;
    }
}

void
menuevent(ep)
EventRecord *ep;
{
    int menuid, menuitem, devnum;
    Rect rect;

    SetPort(ep->who);
    BeginMenu(ep->message);
    menuid = ep->message >> 16;
    menuitem = ep->message & 0xFFFF;
    switch (menuid) {
      case SETUPMENU:
	switch (menuitem) {
	  case 1:
	    defaultsetup();
	    setsliders();
	    dosetup();
	    if (newweb)
		drawswatches();
	    break;

	  case 3:
	    savesetup();
	    break;

	  case 5:
	    ReleaseEvent();
	    break;
	}
	break;
    }
    EndMenu(ep->message);
}

void
mouseevent(ep)
EventRecord *ep;
{
    ControlID p;
    short i;

    SetPort(ep->who);
    if (ep->message & F_KEYDOWN) { /* Mouse-down event? */
	if (FindControl(&ep->where.corner.topLeft, ep->who, &p)) {
	    i = TrackControl(p, &ep->where.corner.topLeft, (void (*)()) -1);
	    if (i == inCheckBox) {
		SetCtlValue(p, 1); /* Select Radio button. */
		if (p == controls[0].cid) {	/* Set HSB Model. */
		    SetCtlValue(controls[1].cid, 0);
		    colormodel = 0;
		    redocontrols();
		    drawlabels();
		}
		if (p == controls[1].cid) {	/* Set RGB Model. */
		    SetCtlValue(controls[0].cid, 0);
		    colormodel = 1;
		    redocontrols();
		    drawlabels();
		}
		if (p == controls[5].cid) /* Set Color #0. */
		    setcolindex(5, 0);
		else if (p == controls[6].cid) /* Set Color #1. */
		    setcolindex(6, 1);
		else if (p == controls[7].cid) /* Set Color #2. */
		    setcolindex(7, 2);
		else if (p == controls[8].cid) /* Set Color #3. */
		    setcolindex(8, 3);
		else if (p == controls[9].cid) /* Set Color #4. */
		    setcolindex(9, 4);
		else if (p == controls[10].cid) /* Set Color #5. */
		    setcolindex(10, 5);
		else if (p == controls[11].cid) /* Set Color #6. */
		    setcolindex(11, 6);
		else if (p == controls[12].cid) /* Set Color #7. */
		    setcolindex(12, 7);
	    }
	    if (i == inThumb) {
		readsliders();
		dosetup();
		if (newweb)
		    drawswatches();
	    }
	} else
	    Bell();		/* Invalid mouse hit. */
    }
}

void
bumpindex()			/* Tab to next color swatch. */
{
    colornum++;
    if (colornum > 7)
	colornum = 0;
    setcolindex(5 + colornum, colornum);
}

void
debumpindex()			/* Tab to previous color swatch. */
{
    colornum--;
    if (colornum < 0)
	colornum = 7;
    setcolindex(5 + colornum, colornum);
}

setcolindex(cntl, wp6palnum)	/* Call when changing active palette entry. */
short cntl, wp6palnum;
{
    short i;

    for (i=5; i<13; i++) {	/* Select the proper button. */
	SetCtlValue(controls[i].cid, cntl == i);
    }
    colornum = wp6palnum;
    redocontrols();
}

void
keyevent(ep)
EventRecord *ep;
{
    short key, i;

    if (ep->message & F_KEYUP)
	return;
    SetPort(ep->who);
    key = ep->message & F_KEYMASK;
    i = (ep->message & F_SHIFT) ? inPageDown : inPageUp;
    if (ep->message & F_FUNC) {
	switch (key) {
	  case K_ESC:
	  case K_RESET:
	    ReleaseEvent();
	    break;

	  case K_UP:
	    if (ep->message & F_ALT) {
		ypos -= 18;
		MoveWindow(window, xpos, ypos, 0);
	    } else
		Bell();
	    break;

	  case K_DOWN:
	    if (ep->message & F_ALT) {
		ypos += 18;
		MoveWindow(window, xpos, ypos, 0);
	    } else
		Bell();
	    break;

	  case K_LEFT:
	    if (ep->message & F_ALT) {
		xpos -= 18;
		MoveWindow(window, xpos, ypos, 0);
	    } else
		Bell();
	    break;

	  case K_RIGHT:
	    if (ep->message & F_ALT) {
		xpos += 18;
		MoveWindow(window, xpos, ypos, 0);
	    } else
		Bell();
	    break;

	  case K_TAB:
	    if (ep->message & F_SHIFT)
		debumpindex();
	    else
		bumpindex();
	    break;

	  default:
	    break;
	}
	if (!wp6)
	    key = 0;
	switch (key) {
	  case K_F1:		/* Toggle Color Model. */
	    SetCtlValue(controls[0].cid, colormodel);
	    colormodel ^= 1;
	    SetCtlValue(controls[1].cid, colormodel);
	    redocontrols();
	    drawlabels();
	    break;

	  case K_F2:
	    MyAction(controls[2].cid, i);
	    break;

	  case K_F3:
	    MyAction(controls[3].cid, i);
	    break;

	  case K_F4:
	    MyAction(controls[4].cid, i);
	    break;
	}
    } else {			/* Non-function key. */
	if (!wp6)
	    key = 0;
	switch (key) {
	  case '0':		/* Set Color #0. */
	    setcolindex(5, 0);
	    break;

	  case '1':		/* Set Color #1. */
	    setcolindex(6, 1);
	    break;

	  case '2':		/* Set Color #2. */
	    setcolindex(7, 2);
	    break;

	  case '3':		/* Set Color #3. */
	    setcolindex(8, 3);
	    break;

	  case '4':		/* Set Color #4. */
	    setcolindex(9, 4);
	    break;

	  case '5':		/* Set Color #5. */
	    setcolindex(10, 5);
	    break;

	  case '6':		/* Set Color #6. */
	    setcolindex(11, 6);
	    break;

	  case '7':		/* Set Color #7. */
	    setcolindex(12, 7);
	    break;

	  default:
	    Bell();
	    break;
	}
    }
}

/*
** On an activate event we always appropriately activate/deactivate our
** Controls.  This is primarily done to inactivate scroll bars on windows
** that can't take events.
*/

void
activatevent(ep)
EventRecord *ep;
{
    struct mycontrols *mcp;

    if (firstactive | !wp6) {	/* Skip first activate event because all the */
	firstactive = 0;	/* controls were created active.  Saves on */
	return;			/* startup time. */
    }

    SetPort(ep->who);
    mcp = &controls[0];
    if (ep->message & 1) {	/* Activating? */
	while (mcp->cid) {
	    HiliteControl(mcp->cid, mcp->hilval);
	    mcp++;		/* Mark all of them active */
	}
    } else {			/* Deactivating. */
	while (mcp->cid)
	    HiliteControl((mcp++)->cid, 255); /* Mark all inactive. */
    }
}

/*
** This routine is called repeatedly if the operator presses in the controls
** of scroll bars.  It is optional (installed via SetCtlAction).  Without it
** the scroll bars only accept a single click in any part of them.  Generally
** users will expect scrolling to be continuous, so you want one of these.
**
** This routine is _not_ called when dragging the thumb.
*/

void
MyAction(theControl, partCode)
ControlID theControl;
short partCode;
{
    if (partCode == inUpButton || partCode == inPageUp) {
	SetCtlValue(theControl, GetCtlValue(theControl)-1);
    } else if (partCode == inDownButton || partCode == inPageDown) {
	SetCtlValue(theControl, GetCtlValue(theControl)+1);
    }
    readsliders();
    dosetup();
    if (newweb)
	drawswatches();
}


redocontrols()			/* Call when switching between HSB and  */
{				/* RGB color models. */
    struct palette *p;
    double h, s, v;
    double R, G, B;

    p = &colors[colornum];
    SetCtlMax(controls[2].cid, colormodel ? maxval : hueval);
    SetCtlMax(controls[3].cid, colormodel ? maxval : satval);
    SetCtlMax(controls[4].cid, colormodel ? maxval : satval);
    if (colormodel) {		/* RGB Model? */
	SetCtlValue(controls[2].cid, maxval - (p->red >> 2));
	SetCtlValue(controls[3].cid, maxval - (p->green >> 2));
	SetCtlValue(controls[4].cid, maxval - (p->blue >> 2));
    } else {			/* HSB Model. */
	R = (double) (p->red >> 2) / maxval;
	G = (double) (p->green >> 2) / maxval;
	B = (double) (p->blue >> 2) / maxval;
	RGBxHSV(R, G, B, &h, &s, &v);
	SetCtlValue(controls[2].cid, hueval - (int) (h * hueval));
	SetCtlValue(controls[3].cid, satval - (int) (s * satval));
	SetCtlValue(controls[4].cid, satval - (int) (v * satval));
    }
}


readsliders()			/* Call after changing a slider. */
{
    struct palette *p;
    double h, s, v;
    double R, G, B;

    p = &colors[colornum];	/* Assume RGB color model. */
    p->red = (maxval - GetCtlValue(controls[2].cid)) << 2;
    p->green = (maxval - GetCtlValue(controls[3].cid)) << 2;
    p->blue = (maxval - GetCtlValue(controls[4].cid)) << 2;
    if (!colormodel) {		/* HSB Color Model? */
				/* Yes, convert to RGB. */
	h = (double) (hueval - GetCtlValue(controls[2].cid)) / hueval;
	s = (double) (satval - GetCtlValue(controls[3].cid)) / satval;
	v = (double) (satval - GetCtlValue(controls[4].cid)) / satval;
	HSVxRGB(h, s, v, &R, &G, &B);
	p->red = (int) (R * maxval) << 2;
	p->green = (int) (G * maxval) << 2;
	p->blue = (int) (B * maxval) << 2;
    }
}

/*
** Given [h,s,v] on [0 - 1.0], return RGB on [0 - 1.0].
*/

void
HSVxRGB(h, s, v, R, G, B)
double h, s, v;
double *R, *G, *B;
{
    double p, q, t;
    int sextant;
    double fract;

    sextant = (int) (fract = (h * 6.0));
    fract = fract - (int) fract;
    p = (1.0 - s) * v;
    q = (1.0 - s * fract) * v;
    t = (1.0 - (s * (1.0 - fract))) * v;
    switch (sextant) {
      case 0:
      case 6: *R = v; *G = t; *B = p; break;
      case 1: *R = q; *G = v; *B = p; break;
      case 2: *R = p; *G = v; *B = t; break;
      case 3: *R = p; *G = q; *B = v; break;
      case 4: *R = t; *G = p; *B = v; break;
      case 5: *R = v; *G = p; *B = q; break;
    }
}

/*
** Given RGB on [0 - 1.0], return [h,s,v] on [0 - 1.0].
**
** For an HSL conversion rather than HSV (HSB), 
** replace V = MAX(R,G,B)
** with    L = (MIN(R,G,B) + MAX(R,G,B) / 2
*/

void
RGBxHSV(R, G, B, H, S, V)
double R, G, B;
double *H, *S, *V;
{
    double MAX, MIN, d, h, r, g, b;

    MAX = MIN = R;
    if (G > MAX) MAX = G;
    if (G < MIN) MIN = G;
    if (B > MAX) MAX = B;
    if (B < MIN) MIN = B;
    d = MAX - MIN;
    *V = MAX;
    if (MAX == 0.0) {
	*S = *H = *V = 0.0;
	return;
    }
    *S = d/MAX;
    if (d == 0.0) {
	*H = 0.0;
	return;
    }
    r = (MAX - R)/d;
    g = (MAX - G)/d;
    b = (MAX - B)/d;
    if (R == MAX) h = b-g;
    if (G == MAX) h = r-b + 2.0;
    if (B == MAX) h = g-r + 4.0;
    if (h < 0.0) *H = h / 6.0 + 1.0;
    if (h >= 0.0) *H = h / 6.0;
}

/*
** Sprintf replacement.
**
** Handles format strings containing:
**
**	 %s, %[0][1..9]d, %[0][1..9]x, %[0][1..9]X
**
** It's much smaller than the standard one.
*/

char *
mysprintf(buffer, fmt, arglist)
char *buffer, *fmt;
char *arglist;
{
    register char *p, *q, *p2;
    short ndigits;
    int leadfill;
    char *itoa();
    union arg {
	long lng;
	char *ptr;
    } *argptr;
	
    argptr = (union arg *) &arglist;
    p = fmt;
    q = buffer;
    while (*p) {		/* While there's some format string. */
	if (*p == '%') {	/* Is it a format control? */
	    p++;		/* Yes, which one (skip over % char)? */
	    if (*p == '0')
		leadfill = *p++;
	    else
		leadfill = ' ';
	    if (*p >= '1' && *p <= '9')	/* Format output length? */
		ndigits = *p++ - '0';
	    else
		ndigits = 0;
	    if (*p == 's') {
		p2 = argptr++->ptr; /* String. */
	    } else if (*p == 'd') {
		p2 = itoa(argptr++->lng, 10, leadfill, ndigits);
	    } else if (*p == 'x' || *p == 'X') {
		p2 = itoa(argptr++->lng, 16, leadfill, ndigits);
	    }
	    p++;		/* Skip over format control. */
	    while (*p2)		/* Copy the inclusion. */
		*q++ = *p2++;
	} else			/* No, it's a literal -- copy it. */
	    *q++ = *p++;
    }
    *q = 0;			/* Null-terminate it. */
    return q;			/* Return pointer to NULL. */
}

char *
itoa(val, radix, leadfill, ndigits)
int val, radix, leadfill, ndigits;
{
    short i;
    char *p, *p2, out;
    static char buf[20];
    unsigned long test, remainder;
	
    p = buf;
    remainder = val;
    i = ndigits;		/* Make sure there're at least the number */
    while (i--)			/* of leading whatevers that we've asked for */
	*p++ = leadfill;	/* as the field width. */
    p2 = p;
    if (radix == 16) {		/* Hex output? */
	test = 0x10000000L;
    } else {
	test = 1000000000L;	/* Decimal output? */
	if (val < 0) {
	    remainder = -val;
	    *p++ = '-';
	}
    }
    while (test > remainder)	/* Scale down to match number. */
	test /= radix;
    if (remainder)
	while (test) {
	    out = remainder / test;
	    *p++ = out + ((out >= 10) ? '7' : '0');
	    remainder %= test;
	    test /= radix;
	}
    else
	*p++ = '0';
    *p = 0;
    if (ndigits) {		/* A field width set? */
	if ((p2 + ndigits) >= p) /* Make sure we never truncate anything. */
	    return p - ndigits;
	else
	    return p2;
    }
    return buf;			/* No field width, return minimum number. */
}
