/*
**
**	Crabs program.
**
*/

#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <isc/clunkdraw.h>

char *errs[] = {"",		/* No error for 0. */
		    "(Must setuid root?)",
		    "Must be run from a TTY.",
		    "Cannot run graphics remotely."};

BitMap	grayMap, crabMap;
BitMap	*screenPtr;		/* Pointer to screen bitmap */
GrafPort crabPort;
Rect crabRect, topleftRect, drawRect;
char *grayPtr, *crabPtr;

unsigned long grayBits[] = {
		    0xAAAA5555,
		    0xAAAA5555,
		    0xAAAA5555,
		    0xAAAA5555,
		    0xAAAA5555,
		    0xAAAA5555,
		    0xAAAA5555,
		    0xAAAA5555};

unsigned long crabBits[] = {
		    0xBEAA7755,
		    0xAAAA7F55,
		    0xBEAA7F55,
		    0xAAAA5555,
		    0xAAAA5555,
		    0xAAAA5555,
		    0xAAAA5555,
		    0xAAAA5555}; 

short crabCount, maxCrabs, started;
Point crabWhere[40];
int runflag;

main(argc, argv)
int argc;
char *argv[];
{
    int err, doQuit();

    started = 0;
    if (getppid() != 1)		/* Detatch ourself. */
	if (fork() != 0)
	    exit(0);
    if (errno = InitGraf(0x4E415354L)) { /* Open up the graphics library. */
	err = errno;
	if (errno>3 || errno<0)
	    err=1;
	fprintf(stderr, "%s: Can't initialize Graphics system.  %s\n",
		argv[0], errs[err]);
	exit(errno);
    }

    for (err = 0; err < _NFILE; err++)
	close(err);		/* Do the rest of the 'daemonization'. */
    setpgrp();
    chdir("/");
    umask(0);

    runflag = 1;
    signal(SIGTERM, doQuit);	/* For killing our little beasts. */

    nice(4);			/* Make ourself a little nicer. */
    if (argc == 2) {		/* We have a snooze value. */
	if (!strcmp("random", argv[1]))
	    sleep(((time() & 0x0F) + 1) * 60);
	else
	    sleep(atoi(argv[1]));
    }

    doOpen();
    while (runflag) {
	doRun();
	nap(40);
    }
    doClose();
    EndGraf();
    exit(0);
}

doOpen()			/* Open routine */
{
    register short whichCrab, i;

    OpenPort(&crabPort);
    screenPtr = &crabPort.portBits; /* Screen bitmap. */

    /*  set up bitmaps for the screen, the crab, and the gray stuff the crabs
	leave behind.  The latter 2 bitmaps are bigger than they really
	need to be because of the constraint that rowBytes must be even. */
	
    SetRect(&crabRect,0,0,16,16);
    SetRect(&topleftRect,0,0,8,8);
    grayMap.baseAddr = (char *) grayBits;
    grayMap.rowBytes = 2;
    grayMap.bounds = crabRect;
    crabMap.baseAddr = (char *) crabBits;
    crabMap.rowBytes = 2;
    crabMap.bounds = crabRect;
	
	/*allocate a random # of crabs to starting positions down the left 
	  edge of the screen */
	
    maxCrabs = (screenPtr->bounds.edge.bottom - screenPtr->bounds.edge.top) / 8;
    for (i = 0, whichCrab = 0; i < maxCrabs; i++)
	if (rand() & 0x0001) {
	    crabWhere[whichCrab].h = screenPtr->bounds.edge.left;
	    crabWhere[whichCrab].v = i * 8;
	    whichCrab++;
	}
    crabCount = whichCrab;	/* Keep track of how many crabs were allocd. */
	
    /* paste the top left corner of the crab bitmap ( i.e., the crab itself)
       in each crab's starting position on the screen */
	
	started = 1;
	for (whichCrab = 0; whichCrab < crabCount; whichCrab++) {
	    SetRect(&drawRect,
		    crabWhere[whichCrab].h,
		    crabWhere[whichCrab].v,
		    crabWhere[whichCrab].h + 8,
		    crabWhere[whichCrab].v + 8);
	    CopyBits(&crabMap, screenPtr,
		     &topleftRect, &drawRect, srcCopy, 0L);
	}
}

doClose()
{
    register short whichCrab;
	
    /* Replace all the current crabs with gray. */
    if (started) {
	for (whichCrab = 0; whichCrab < crabCount; whichCrab++) {
	    SetRect(&drawRect,
		    crabWhere[whichCrab].h,
		    crabWhere[whichCrab].v,
		    crabWhere[whichCrab].h + 8,
		    crabWhere[whichCrab].v + 8);
	    CopyBits(&grayMap, screenPtr, &topleftRect, &drawRect, srcCopy, 0L);
	}
    }
    ClosePort(&crabPort);
}

doRun()				/* Run routine */
{
    short offset, voffset, sign;
    register short whichCrab;
	
    /*first, replace all the current crabs with gray*/
    for (whichCrab = 0; whichCrab < crabCount; whichCrab++) {
	SetRect(&drawRect,
		crabWhere[whichCrab].h,
		crabWhere[whichCrab].v,
		crabWhere[whichCrab].h + 8,
		crabWhere[whichCrab].v + 8);
	CopyBits(&grayMap, screenPtr,
		 &topleftRect, &drawRect, srcCopy, 0L);
    }
	
    /* change each crab's position a little bit vertically,
       somewhat more horizontally */
    for (whichCrab = 0; whichCrab < crabCount; whichCrab++) {
	offset = (rand() & 0x0003);		/* offset = 0,1,2, or 3*/
	sign = ((rand() & 0x0001) ? 1 : -1);	/* + or - */
	/* +/- 2, 4, 6, or 8 horiz */
	crabWhere[whichCrab].h = 
	    (crabWhere[whichCrab].h + sign * (2+2*offset));
	/* Wraparound */
	if (crabWhere[whichCrab].h + 8 >= screenPtr->bounds.edge.right)
	    crabWhere[whichCrab].h = screenPtr->bounds.edge.left;
	else if (crabWhere[whichCrab].h < screenPtr->bounds.edge.left)
	    crabWhere[whichCrab].h = screenPtr->bounds.edge.right - 8;
	voffset = (rand() & 0x0003);
	switch (voffset) {
	  case 0:
	    crabWhere[whichCrab].v = (crabWhere[whichCrab].v - 2);
	    /* Wraparound */
	    if (crabWhere[whichCrab].v + 8 >= screenPtr->bounds.edge.bottom)
		crabWhere[whichCrab].v = screenPtr->bounds.edge.top;
	    else if (crabWhere[whichCrab].v < screenPtr->bounds.edge.top)
		crabWhere[whichCrab].v = screenPtr->bounds.edge.bottom - 8;
	    break;
		
	  case 1:
	  case 2:
	    break;
		
	  case 3:
	    crabWhere[whichCrab].v = (crabWhere[whichCrab].v + 2)
		% (screenPtr->bounds.edge.bottom - screenPtr->bounds.edge.top);
	    break;
		
	  default:
	    break;
	}
    }
				
    /* redraw the little buggers */
    for (whichCrab=0; whichCrab<crabCount; whichCrab++) {
	SetRect(&drawRect,
		crabWhere[whichCrab].h,
		crabWhere[whichCrab].v,
		crabWhere[whichCrab].h + 8,
		crabWhere[whichCrab].v + 8);
	CopyBits(&crabMap, screenPtr, &topleftRect, &drawRect, srcCopy, 0L);
    }
}

doQuit()
{
	runflag = 0;
}
