/*
** cyclewatch
**
** This program displays bar graphs of process time usage.  Made out of
** the carcass of winmem.
**
** Revision History:
**
** Version 1.2		03/01/94	Jim Cathey
**	 Include process 0 (the page swapper) in the display.
**
** Version 1.1		12/15/93	Jim Cathey
**	 Added automatic video copy when WINDOWS_ON contains a list.
**
** Version 1.0		07/14/92	Jim Cathey
**	 Initial Release
*/

#include <malloc.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/times.h>
#include <dnix/pstat.h>
#include <dnix/dutsname.h>
#include <dnix/fcodes.h>
#include <dnix/errno.h>
#include <dnix/sysinfo.h>
#include <dnix/osconfig.h>
#include <isc/window.h>
#include <isc/keycodes.h>
#include <isc/controls.h>

#define VERSION "V1.2"

#define TRUE 1
#define FALSE 0
#define SLIDE_OFFSET 16
#define BARHORZ 200

#define FONT 5
#define FONTSIZE 9

/*
** Menus
*/
#define M_SAMPLE 1
#define M_AVERAGE 2
#define M_MODE 3

extern int optind, opterr;
extern char *optarg, *getenv(), *mysprintf();

void mouseevent(), keyevent(), menuevent(), updateevent();
void windowevent(), activatevent(), napevent();
void MyAction(ControlID theControl, short partCode);

void (*completion) () = napevent;
int napreq;

Rect world, screen, playpen, heading, detail, slider;
FontInfo fontrec;
ControlID slideID;

short rate_array[8] = {0, 1, 2, 5, 10, 15, 30, 60};
short sample_item = 1;
short sample_rate = 1;
short avg_array[8] = {0, 1, 2, 5, 10, 20, 50, 100};
short avg_item = 3;
short ntoavg = 5;

short window, trapq, wmfd, width, height, weRicon, busymode;
short title_length, paused, coarse, fine;
short start_proc, win_procs, last_proc;
struct dutsname duts;		/* For our system name et. al. */
struct tms timebuf;		/* Scratch. */

struct timeuse {
    short changed;		/* True if display update needed. */
    short pid;
    char name[14];
    long ticksused;		/* Running sum of used ticks. */
    long utime;			/* Ticks used since start of process. */
} *tu;				/* Array (1 per process slot). */
short tusize;			/* Number of TU's available. */
unsigned short *buckets;	/* 2-D array of moving average buckets. */
long globalticks;		/* Ticks available for use over all samples. */
long busyticks;			/* Ticks used over all samples. */
long globaltime;		/* For getting system ticker. */

main(argc, argv)
int argc;
char *argv[];

{
    struct Sysinfo sys;
    struct osconfig config;
    int ch, i;
    short horz, vert, x, y, voffset;
    char Title[80];
    char Domain[64];
    char *dispsystem, *cp, *argp;

    x = 5;
    y = 45;
 /*
 ** Parse our arg list.
 */
    dispsystem = getenv("WINDOWS_ON");
    opterr = FALSE;
    while ((ch = getopt(argc, argv, "Vs:a:x:y:")) != EOF) {
	switch (ch) {
	  case 'V':
	    mysprintf(Title, "%s %s\n", argv[0], VERSION);
	    write(1, Title, StringLen(Title));
	    exit(0);
	    break;

	  case 's':
	    sample_item = myatoi(optarg);
	    if (sample_item < 1)
		sample_item = 1;
	    else if (sample_item > sizeof(rate_array) - 2)
		sample_item = sizeof(rate_array) - 2;
	    sample_rate = rate_array[sample_item];
	    break;

	  case 'a':
	    avg_item = myatoi(optarg);
	    if (avg_item < 1)
		avg_item = 1;
	    else if (avg_item > sizeof(avg_array) - 2)
		avg_item = sizeof(avg_array) - 2;
	    ntoavg = avg_array[avg_item];
	    break;

	  case 'x':
	    x = myatoi(optarg);
	    break;

	  case 'y':
	    y = myatoi(optarg);
	    break;

	  case '?':
	  default:
	    mysprintf(Title, "usage: %s [-V] [-s Sample-rate] [-a average] "
		      "[-x offset] [-y offset] [system-name]\n", argv[0]);
	    write(1, Title, StringLen(Title));
	    exit(1);
	    break;
	}
    }

 /*
 ** Setup window.
 */
    trapq = InitGraf();
    if (trapq < 0)
	exit(1);
    dnix(F_UNAME, &duts, sizeof(struct dutsname));
    strcpy(Title, duts.nodename);
    if (optind < argc)
	dispsystem = argv[optind];
    if (dispsystem && strchr(dispsystem, '.')) {
	if ((wmfd = open("/etc/domainid", O_RDONLY|O_NOCTTY)) >= 0) {
	    if (read(wmfd, Domain, sizeof(Domain) - 1) > 0) {
		cp = strchr(Domain, '\n');
		if (cp)		/* Knock off newline. */
		    *cp = 0;
		strcat(Title, Domain);
	    }
	    close(wmfd);
	}
    }
    wmfd = ConnectGraf(dispsystem);
    if (wmfd < 0)
	exit(2);
    strcat(Title, " : Process Time Display");
    SetRect(&screen, 0, 0, 0, 0);
    world = screen;
    OffsetRect(&world, x, y);
    window = NewWindow(&world, Title, hideWindow, documentProc, frontWindow,
		       hasDrag | hasGoAway | hasZoom | hasSendToBack | hasGrow,
		       wmfd);
    SetPort(window);
 /*
 ** Create our menus.
 */
    NewMenu(M_SAMPLE, "Rate");
    AppendMenu(M_SAMPLE, "1;2;5;10;15;30;60");
    CheckItem(M_SAMPLE, sample_item, TRUE);
    NewMenu(M_AVERAGE, "Average");
    AppendMenu(M_AVERAGE, "1;2;5;10;20;50;100");
    CheckItem(M_AVERAGE, avg_item, TRUE);
    NewMenu(M_MODE, "Graph");
    AppendMenu(M_MODE, "% of CPU Used;% of Busy Cycles");
    CheckItem(M_MODE, busymode + 1, TRUE);

    PenMode(srcCopy);
    TextMode(srcCopy);
    TextFont(FONT);
    TextSize(FONTSIZE);
    LocalGetFontInfo(FONT, normalText, FONTSIZE, &fontrec);
    width = fontrec.widMax;
    height = fontrec.ascent + fontrec.descent;
    title_length = width * 42;

 /*
 ** Now size the window and display it.
 */
    horz = width + title_length + BARHORZ + SLIDE_OFFSET;
    voffset = height * 5;
    vert = voffset * 2 + (height * 18);
    world.edge.right += horz;
    world.edge.bottom += vert;
    screen.edge.right += horz;
    screen.edge.bottom += vert;
    playpen = screen;
    playpen.edge.top += voffset - fontrec.ascent;
    playpen.edge.right -= SLIDE_OFFSET;
    SetRect(&detail, playpen.edge.left, screen.edge.bottom - height * 8,
	    playpen.edge.right, screen.edge.bottom);
    playpen.edge.bottom = playpen.edge.top +
	((detail.edge.bottom - playpen.edge.top) / height * height);
    win_procs = (playpen.edge.bottom - playpen.edge.top) / height;
    SetRect(&heading, screen.edge.left, screen.edge.top, playpen.edge.right,
	    playpen.edge.top);
    SizeWindow(window, (world.edge.right - world.edge.left),
	       (world.edge.bottom - world.edge.top), TRUE);
    ShowWindow(window);

    if (argp = dispsystem) {	/* Handle copying video to other WS. */
	char *cp, buf[80]; int fd; long winds;

	winds = 0;
	while (argp = strchr(argp, '\n')) {
	    argp++;
	    cp = strchr(argp, '\n');
	    if (!cp)
		cp = argp + strlen(argp);
	    if (cp - argp) {
		strncpy(buf, argp, cp - argp);
		buf[cp - argp] = 0;
		if ((fd = ConnectGraf(buf)) >= 0) {
		    winds |= 1 | 1 << NewWindow(&world, Title, showWindow,
						documentProc, frontWindow,
						hasSendToBack, fd);
		}
	    }
	}
	if (winds) {
	    SetPorts(winds);
	    PenMode(srcCopy);
	    TextMode(srcCopy);
	    TextFont(FONT);
	    TextSize(FONTSIZE);
	}
    }
 /*
 ** Find out how many procs to allow for.
 */
    dnix(F_SYSINFO, F_SYS_GETOSCFIG, &config, sizeof (config));
    last_proc = config.oc_ntnod - 1;
    tusize = config.oc_ntnod;
    tu = (struct timeuse *) calloc(tusize, sizeof(struct timeuse));
    buckets = (unsigned short *) calloc((tusize + 1), ntoavg * sizeof(short));

 /*
 ** Create the slider.
 */
    InitControls();
    SetRect(&slider, (screen.edge.right - SLIDE_OFFSET), (screen.edge.top - 1),
	    (screen.edge.right + 1), (screen.edge.bottom - SLIDE_OFFSET + 1));
    slideID = NewControl(window, &slider, "", showControl, start_proc, 0,
			 last_proc, scrollBarProc, 0);
    SetCtlAction(slideID, MyAction); /* For continuous scrolling while */
                                /* the buttons of scroll bar are pressed. */
    SetCtlScrollRate(slideID, 5);
    SetCtlValues(slideID, start_proc, win_procs);

 /*
 ** Bind our event handlers.
 */
    Event(keyEvt, keyevent);
    Event(menuEvt, menuevent);
    Event(mouseEvt, mouseevent);
    Event(updateEvt, updateevent);
    Event(windowEvt, windowevent);
    Event(activateEvt, activatevent);
    globaltime = times(&timebuf);
    napevent(0, 0, 0);		/* Start timer & take first samples. */

 /*
 ** Begin handling events.
 */
  more:
    if (GetEvent(anyWindow, noEvent)) {
	if ((i = (EventError() >> 16)) != wmfd) {
	    DisconnectGraf(i);	/* Video copy died, take it out & continue. */
	    goto more;
	}
	exit(3);
    }
    exit(0);
}

percent(amount, total, max)
short amount, total, max;
{

    if (total)
	return ((amount * max) / total);
    return 0;
}

/*
** Display the time-use information for the processes.
*/

ShowProcs()
{
    struct timeuse *tup;
    short task, vert, slot;
    int clipped;		/* TRUE if we've drawn anything. */
    int ticks;			/* Ticks since last sampling. */
    Rect scratch;

    clipped = FALSE;
    vert = playpen.edge.top + fontrec.ascent;
    ticks = times(&timebuf);
    for (task = 0; task < win_procs; task++) {
	slot = task + start_proc;
	tup = &tu[slot];
	if (((vert-fontrec.ascent) >= playpen.edge.bottom) /* No procs */
	    || (slot >= last_proc)) /* or window space left --- done. */
	    break;
	if (tup->changed || tup->ticksused) {
	    if (!clipped) {	/* Avoid sending drawing commands if */
		ClipRect(&playpen); /* nothing is changing. */
		TextMode(srcCopy);
		clipped = TRUE;
	    }
	    ShowProc(tup, vert);
	    tup->changed = 0;
	}
	vert += height;
    }
    globaltime = ticks;
    if (clipped) {
	scratch = playpen;
	scratch.edge.top = vert + fontrec.descent;
	EraseRect(&scratch);
	ClipRect(&screen);
    }
}

/*
** Display the time information for one process.
*/

ShowProc(tup, vert)
struct timeuse *tup;
short vert;
{
    Rect temprect;
#define TR temprect.edge
    char buffer[40];
    short horz, perbusy, percpu, max;

    if (tup->pid >= 0) {	/* Process exists. */
	MoveTo(width, vert+1);
	perbusy = tup->ticksused ? (tup->ticksused * 100 / busyticks) : 0;
	percpu = tup->ticksused ? (tup->ticksused * 100 / globalticks) : 0;
	mysprintf(buffer, "%5d %14s %6d %3d %3d", tup->pid, tup->name,
		  tup->utime, perbusy, percpu);
	DrawString(buffer);

	horz = title_length;
	SetRect(&temprect, horz, (vert - fontrec.ascent),
		(horz + BARHORZ - 8), (vert + fontrec.descent));
	FrameRect(&temprect);
	InsetRect(&temprect, 1, 1);
	max = TR.right - width;
	TR.right = TR.left + percent(busymode ? perbusy : percpu,
				     100, (BARHORZ - 10));
	if (TR.right > max)
	    TR.right = max;
	ForeColor(2);
	FillRect(&temprect, blackPattern); /* Percent of CPU used. */

	TR.left = TR.right;
	TR.right = horz + BARHORZ - 9;
	if (TR.right > max)
	    TR.right = max;
	ForeColor(1);
	BackColor(0);
	FillRect(&temprect, whitePattern);
    } else {			/* Process doesn't exist. */
	SetRect(&temprect, playpen.edge.left, (vert-fontrec.ascent),
		playpen.edge.right,(vert+fontrec.descent));
	EraseRect(&temprect);
    }
}

drawkey()
{
    Rect key;
    char *cp;
    int horz, vert, wid;

    vert = height;
    cp = VERSION;
    horz = playpen.edge.right - 2
	- LocalStringWidth(FONT, normalText, FONTSIZE, cp);
    MoveTo(horz, vert);
    DrawString(cp);
}

drawlabels()
{
    char buffer[40];

    TextMode(srcCopy);
    drawkey();
    MoveTo(3 * width, height);
    mysprintf(buffer, "Tick rate: %d Hz", duts.clocktics);
    DrawString(buffer);
    Move(5*width, 0);
    mysprintf(buffer, "Averaging %d %s sample%s ", ntoavg,
	      (coarse > fine) ? "coarse" : "fine",
	      ntoavg > 1 ? "s" : "");
    DrawString(buffer);
    mysprintf(buffer, "at %d second%s each.  ", sample_rate,
	      sample_rate > 1 ? "s" : "");
    DrawString(buffer);
    TextFace(underlineText);
    MoveTo(3*width, height * 3);
    DrawString("PID");
    Move(2*width, 0);
    DrawString("Process");
    Move(8*width, 0);
    DrawString("Ticks");
    Move(width, 0);
    DrawString("%Bsy");
    Move(width, 0);
    DrawString("%CPU");
    Move(4*width, 0);
    TextFace(normalText);
    DrawChar('0');
    Move(2*width, 0);
    TextFace(underlineText);
    if (busymode) {
	DrawString("Percentage of Busy Cycles");
	Move(2*width, 0);
    } else {
	DrawString("Percentage of CPU");
	TextFace(normalText);
	DrawString("        ");
	Move(2*width, 0);
    }
    TextFace(normalText);
    DrawString("100");
}

#ifdef DEBUG
abort()
{
    kill(getpid(), 6);
}

/*
** Check the sampling data structures for consistency.  For debugging.
*/

checkitout()
{
    short task, tasksum, i;
    long sumticks;

    sumticks = 0;
    for (task = 0; task < tusize; task++) {
	tasksum = 0;
	for (i=0; i<ntoavg; i++) {
	    tasksum += buckets[task * ntoavg + i];
	}
	sumticks += tasksum;
	if (tu[task].ticksused != tasksum)
	    abort();
    }
    if (sumticks != busyticks)
	abort();
    tasksum = 0;
    for (i=0; i<ntoavg; i++) {
	tasksum += buckets[(short) tusize * ntoavg + i];
    }
    if (tasksum != globalticks)
	abort();
}
#endif

/*
** Remove all traces of a task.
*/

cleanout(task, running)
short task;
int running;
{
    struct timeuse *tup;
    int i;

    tup = &tu[task];
    if (tup->pid >= 0) {	/* Process died, delete it. */
	tup->pid = -1;
	if (running)
	    busyticks -= tup->ticksused;
	tup->ticksused = 0;
	tup->name[0] = 0;
	tup->changed = 1;
	for (i=0; i<ntoavg; i++) /* Empty sample buckets. */
	    buckets[(short) task * ntoavg + i] = 0;
    }
}

/*
** Event handlers.
*/

void
napevent(routine, rpar, rn)
int (*routine) ();
int rpar, rn;
{
    struct psinfo tcbinfo;	/* Holds info for each process. */
    struct timeuse *tup;
    int i, task, ticks;
    short status;
    char *cp;

    napreq = dnix(F_NOWAIT | F_NAP, (sample_rate * 1000), 0,0,0,0, trapq,
		  &completion);
    if (!paused) {
	if (rpar < 0)		/* Cancelled timers 'start over'. */
	    routine = 0;
	if (routine) {		/* Don't do this the first time. */
	    for (task = 0; task < tusize; task++) { /* Delete oldest sample. */
		ticks = buckets[(short) task * ntoavg];
		buckets[(short) task * ntoavg] = 0;
		tup = &tu[task];
		tup->ticksused -= ticks;
#ifdef DEBUG
		if (tup->ticksused < 0)
		    abort();
#endif
		if (ticks)
		    tup->changed = 1;
		busyticks -= ticks;

#ifdef DEBUG
		checkitout();
#endif
	    }
	    globalticks -= buckets[(short) tusize * ntoavg];
	    buckets[(short) tusize * ntoavg] = 0;
	    if (ntoavg > 1) {	/* Move samples for moving average. */
		BlockMove(&buckets[1], &buckets[0],
			  ((short) (tusize + 1) * ntoavg - 1) * sizeof(short));
#ifdef DEBUG
		for (task = 0; task <= tusize; task++)
		    buckets[(short) (task + 1) * ntoavg - 1] = 0;
#endif
	    }
	}
	for (task = 0; task < tusize; task++) { /* Take new samples. */
	    tup = &tu[task];
	    status = dnix(F_PSTAT, &tcbinfo, task, 0, 0, 0, 0, 0);
	    if (status < 0) {
		if (status == ESRCH) { /* No such process - continue */
		    cleanout(task, routine);
		    continue;
		}
		while (task < tusize) /* Out of valid tasks, clean out */
		    cleanout(task++, routine); /* any remainder. */
		break;
	    } else {		/* A process exists. */
		if (!routine) {	/* First time through? */
		    if (tcbinfo.ps_swtim/duts.clocktics == tcbinfo.ps_utime) {
			ticks = tcbinfo.ps_swtim; /* Newest OS! */
			fine++;
		    } else {	/* Older OS, not as granular. */
			ticks = tcbinfo.ps_utime * duts.clocktics;
			coarse++;
		    }
		} else {	/* Normal path, 2nd and successive passes. */
		    if (coarse > fine) /* Coarse or fine sampling? */
			ticks = tcbinfo.ps_utime * duts.clocktics;
		    else
			ticks = tcbinfo.ps_swtim; /* Newest OS! */
		}
		if (cp = strrchr(tcbinfo.ps_spar, '/'))	/* Find base name */
		    cp++;	/* of the process. */
		else
		    cp = tcbinfo.ps_spar;
		if (tup->pid != tcbinfo.ps_pid || /* Same process there? */
		    strncmp(cp, tup->name, sizeof tup->name)) {
		    if (routine)
			busyticks -= tup->ticksused; /* No, delete old one. */
		    tup->ticksused = 0;
		    for (i=0; i<ntoavg; i++) /* Empty sample buckets. */
			buckets[(short) task * ntoavg + i] = 0;
		    tup->utime = 0;
		    tup->pid = tcbinfo.ps_pid; /* Install new one. */
		    BlockMove(cp, tup->name, sizeof tup->name);
		}
		if (status = ticks - tup->utime) {
		    if (routine) {
			buckets[(short)(task + 1) * ntoavg - 1] = status;
			tup->ticksused += status;
#ifdef DEBUG
			if (tup->ticksused < 0 || status < 0)
			    abort();
#endif
			busyticks += status;
		    }
		    tup->utime = ticks;
		    tup->changed = 1;
		} else {
		    buckets[(short)(task + 1) * ntoavg - 1] = 0;
		}
	    }
#ifdef DEBUG
	    if (routine)
		checkitout();
#endif
	}
	ticks = times(&timebuf);
	if (routine)
	    globalticks += (buckets[(short)(tusize + 1) * ntoavg - 1]
			    = ticks - globaltime);
	globaltime = ticks;
#ifdef DEBUG
	if (routine)
	    checkitout();
#endif
	if (routine)
	    ShowProcs();
    }
}

void
updateevent(ep)
EventRecord *ep;
{
    int i;
    long save;

    GetPort(&save);
    SetPort(ep->who);
    BeginUpdate(ep->who);
    if (!weRicon || ep->who != window) {
	DrawControls(ep->who);
	DrawGrowIcon(ep->who);

	drawlabels();
	for (i=0; i<tusize; i++)
	    tu[i].changed = 1;
	ShowProcs();
    }
    EndUpdate(ep->who);
    SetPort(save);
}

void
mouseevent(ep)
EventRecord *ep;
{
    ControlID cid;
    int i;

    if (ep->who != window) {
	long save;

	if (ep->message & F_KEYDOWN) {
	    GetPort(&save);
	    SetPort(ep->who);
	    Bell();
	    SetPort(save);
	}
	return;
    }
    paused = TRUE;		/* No timer updates while refreshing. */
    if (FindControl(&ep->where.corner.topLeft, ep->who, &cid)) {
	switch (TrackControl(cid, &ep->where.corner.topLeft, (void (*)())-1)) {
	  case inThumb:
	    MyAction(cid, inThumb);
	    break;
	}
    }
    paused = FALSE;
}

/*
** This routine is called repeatedly if the operator presses in the controls
** of scroll bars.  It is optional (installed via SetCtlAction).  Without it
** the scroll bars only accept a single click in any part of them.  Generally
** users will expect scrolling to be continuous, so you want one of these.
**
** This routine is _not_ automatically called when dragging the thumb.
*/

void
MyAction(theControl, partCode)
ControlID theControl;
short partCode;
{
    short i, j;

    i = start_proc;
    j = win_procs / 2;
    if (!j)
	j = 1;
    switch (partCode) {
      case inPageUp:
	if (start_proc > 0) {
	    start_proc -= j;
	    if (start_proc < 0)
		start_proc = 0;
	    i = (i - start_proc) * height;
	}
	break;

      case inUpButton:
	if (start_proc > 0) {
	    start_proc--;
	    i = height;
	}
	break;

      case inPageDown:
	if (start_proc < last_proc) {
	    start_proc += j;
	    if (start_proc > last_proc)
		start_proc = last_proc;
	    i = -(start_proc - i) * height;
	}
	break;

      case inDownButton:
	if (start_proc < last_proc) {
	    start_proc++;
	    i = -height;
	}
	break;

      case inThumb:
	i = start_proc;
	start_proc = GetCtlValue(theControl);
	i = (i - start_proc) * height;
	break;

      default:
	i = 0;
    }
    if (i) {
	SetCtlValues(theControl, start_proc, win_procs);
	ScrollRect(&playpen, 0, i, -1);
	BeginUpdateRect(window, &playpen); /* Avoid delay (update event). */
	for (i=0; i<tusize; i++) /* Force a full refresh. */
	    tu[i].changed = 1;
	ShowProcs();
	EndUpdate(window);
    }
}

void
keyevent(ep)
EventRecord *ep;
{
    int i;

    if (ep->message & F_KEYUP)
	return;

    if (ep->who != window) {
	long save;

	if (ep->message & F_KEYDOWN) {
	    GetPort(&save);
	    SetPort(ep->who);
	    Bell();
	    SetPort(save);
	}
	return;
    }
    if (ep->message & F_FUNC) {
	switch (ep->message & F_KEYMASK) {
	  case K_ESC:		/* Time to quit. */
	  case K_RESET:
	    ReleaseEvent();
	    break;

          case K_UP:		/* Scroll list up. */
            if (ep->message & F_ALT) {
                OffsetRect(&world, 0, -18);
                MoveWindow(window, world.edge.left, world.edge.top, 0);
	    } else {
		if (start_proc > 0)
		    MyAction(slideID, inUpButton);
		else
		    Bell();
	    }
            break;

          case K_DOWN:		/* Scroll list down. */
            if (ep->message & F_ALT) {
                OffsetRect(&world, 0, 18);
                MoveWindow(window, world.edge.left, world.edge.top, 0);
            } else {
		if (start_proc < last_proc)
		    MyAction(slideID, inDownButton);
		else
		    Bell();
	    }
            break;

          case K_LEFT:
            if (ep->message & F_ALT) {
                OffsetRect(&world, -18, 0);
                MoveWindow(window, world.edge.left, world.edge.top, 0);
            } else
		Bell();
            break;

          case K_RIGHT:
            if (ep->message & F_ALT) {
                OffsetRect(&world, 18, 0);
                MoveWindow(window, world.edge.left, world.edge.top, 0);
            } else
		Bell();
            break;

          case K_PGUP:          /* Scroll list up a half screen. */
            if (start_proc > 0)
                MyAction(slideID, inPageUp);
            else
                Bell();
            break;

          case K_PGDN:          /* Scroll list down a half screen. */
            if (start_proc < last_proc)
                MyAction(slideID, inPageDown);
            else
                Bell();
            break;

          case K_HOME:          /* Scroll list to beginning. */
            if (start_proc) {
		SetCtlValue(slideID, 0);
                MyAction(slideID, inThumb);
            } else
                Bell();
            break;

	  default:
	    Bell();
	    break;
	}
    }
}

void
menuevent(ep)
EventRecord *ep;
{
    Rect eraser;
    short mid, item;

    paused = TRUE;		/* No timer updates while changing. */
    BeginMenu(ep->message);
    mid = ep->message >> 16;
    item = ep->message & 0xFFFF;
    switch (mid) {
      case M_SAMPLE:
	if (item != sample_item) {
	    CheckItem(M_SAMPLE, sample_item, FALSE);
	    sample_item = item;
	    sample_rate = rate_array[sample_item];
	    CheckItem(M_SAMPLE, sample_item, TRUE);
	    dnix(F_CANCEL, napreq); /* Re-start timer. */
	    drawlabels();
	}
	break;

      case M_AVERAGE:
	if (item != avg_item) {
	    CheckItem(M_AVERAGE, avg_item, FALSE);
	    ntoavg = avg_array[item];
	    CheckItem(M_AVERAGE, item, TRUE);
	    avg_item = item;
	    free(buckets);
	    buckets = (unsigned short *) calloc((tusize + 1),
						ntoavg * sizeof(short));
	    for (item=0; item<tusize; item++) {
		tu[item].ticksused = 0;
		tu[item].changed = 1;
	    }
	    busyticks = 0;
	    globalticks = 0;
	    drawlabels();
	}
	break;

      case M_MODE:
	if (item != busymode + 1) {
	    CheckItem(M_MODE, busymode + 1, FALSE);
	    busymode = !busymode;
	    CheckItem(M_MODE, item, TRUE);
	    for (item=0; item<tusize; item++) {
		tu[item].changed = 1;
	    }
	    drawlabels();
	}
	break;
    }
    EndMenu(ep->message);
    paused = FALSE;
}

void
windowevent(ep)
EventRecord *ep;
{
    int i, j;

    if (ep->who != window) {
	return;
    }
    switch (ep->message & F_KEYMASK) {
      case windowMove:
	OffsetRect(&world, ep->where.edge.left, ep->where.edge.top);
	MoveWindow(ep->who, world.edge.left, world.edge.top, 0);
	break;

      case windowZoom:
	if (weRicon) {
	    UnIconifyWindow(window);
	    weRicon = FALSE;
	} else {
	    IconifyWindow(window);
	    weRicon = TRUE;
	}
	break;

      case windowGoaway:
	ReleaseEvent();
	break;

      case windowResize:
	if (!ep->where.edge.top && !ep->where.edge.left)
	    break;
	i = screen.edge.bottom;
	if ((i + ep->where.edge.top) < (j=playpen.edge.top + detail.edge.bottom
					- detail.edge.top + height + 2))
	    ep->where.edge.top = j - i; /* Leave room for >= 1 process. */
	world.edge.right += ep->where.edge.left;
	world.edge.bottom += ep->where.edge.top;
	screen.edge.right += ep->where.edge.left;
	screen.edge.bottom += ep->where.edge.top;
	detail.edge.right += ep->where.edge.left;
	detail.edge.bottom += ep->where.edge.top;
	detail.edge.top += ep->where.edge.top;
	playpen.edge.right += ep->where.edge.left;
	playpen.edge.bottom = playpen.edge.top +
	    ((detail.edge.bottom - playpen.edge.top) / height * height);
	win_procs = (playpen.edge.bottom - playpen.edge.top) / height;
	slider.edge.bottom += ep->where.edge.top;
	SizeWindow(ep->who, (world.edge.right - world.edge.left),
		   (world.edge.bottom - world.edge.top), TRUE);
	ClipRect(&screen);
	HideControlNoInval(slideID);
	win_procs = (playpen.edge.bottom - playpen.edge.top) / height;
	if (win_procs < 0)
	    win_procs = 0;
	SizeControl(slideID, (slider.edge.right - slider.edge.left),
		    (slider.edge.bottom - slider.edge.top));
	MoveControl(slideID, (screen.edge.right - SLIDE_OFFSET),
		    (screen.edge.top - 1));
	SetCtlValues(slideID, start_proc, win_procs);
	ShowControl(slideID);
	EraseRect(&screen);
	InvalRect(&screen);
	break;
    }
}

void
activatevent(ep)
EventRecord *ep;
{
    if (!weRicon && ep->who == window) {
	if (ep->message & F_KEYMASK)
	    HiliteControl(slideID, 0); /* Activating. */
	else
	    HiliteControl(slideID, 255); /* Deactivating. */
    }
}

/*
** Sprintf replacement.
**
** Handles format strings containing:
**
**   %[.][0..9][0..9]s, %[-][0][1..9]d, %[-][0][1..9]x, %[-][0][1..9]X, %c
**
** It's much smaller than the standard one.
*/

char *
mysprintf(buffer, fmt, arglist)
char *buffer, *fmt;
char *arglist;
{
    register char *p, *q, *p2;
    short ndigits;
    int leadfill;
    int negflag;
    char *itoa();
    union arg {
	long lng;
	char *ptr;
    } *argptr;

    argptr = (union arg *) &arglist;
    p = fmt;
    q = buffer;
    while (*p) {		/* While there's some format string. */
	if (*p == '%') {	/* Is it a format control? */
	    p++;		/* Yes, which one (skip over % char)? */
	    negflag = 1;
	    if (*p == '-') {
		p++;
		negflag = -1;
	    }
	    if (*p == '.')
		p++;
	    if (*p == '0')
		leadfill = *p++;
	    else
		leadfill = ' ';
	    if (*p >= '1' && *p <= '9') { /* Format output length? */
		ndigits = *p++ - '0';
		if (*p >= '0' && *p <= '9')
		    ndigits = ndigits * 10 + (*p++ - '0');
	    } else
		ndigits = 0;
	    if (*p == 'c')
		*q++ = argptr++->lng;
	    else if (*p == 's') {
		p2 = argptr++->ptr; /* String. */
		if (ndigits) {	/* String width specified? */
		    while (*p2 && ndigits--) /* Yes, copy until we */
			*q++ = *p2++; /* have copied all, or run out. */
		    while (ndigits-- > 0) /* Any field width left? */
			*q++ = ' '; /* Yes, fill with spaces. */
		} else {
		    while (*p2)	/* Copy the inclusion. */
			*q++ = *p2++;
		}
	    } else {
		if (*p == 'd') {
		    p2 = itoa(argptr++->lng, 10, leadfill, ndigits * negflag);
		} else if (*p == 'x' || *p == 'X') {
		    p2 = itoa(argptr++->lng, 16, leadfill, ndigits * negflag);
		}
		while (*p2)	/* Copy the inclusion. */
		    *q++ = *p2++;
	    }
	    p++;		/* Skip over format control. */
	} else			/* No, it's a literal -- copy it. */
	    *q++ = *p++;
    }
    *q = 0;			/* Null-terminate it. */
    return q;
}

char *
itoa(val, radix, leadfill, ndigits)
int val, radix, leadfill, ndigits;
{
    short i, negflag;
    char *p, *p2, out;
    static char buf[20];
    unsigned long test, remainder;

    p = buf;
    remainder = val;
    negflag = (ndigits >= 0) ? 1 : -1;
    ndigits *= negflag;
    i = ndigits;		/* Make sure there're at least the number */
    while (i--)			/* of leading whatevers that we've asked for */
	*p++ = leadfill;	/* as the field width. */
    p2 = p;
    if (negflag < 0)
	p2 = p = buf;
    if (radix == 16) {		/* Hex output? */
	test = 0x10000000L;
    } else {
	test = 1000000000L;	/* Decimal output? */
	if (val < 0) {
	    remainder = -val;
	    *p++ = '-';
	}
    }
    while (test > remainder)	/* Scale down to match number. */
	test /= radix;
    if (remainder)
	while (test) {
	    out = remainder / test;
	    *p++ = out + ((out >= 10) ? '7' : '0');
	    remainder %= test;
	    test /= radix;
	}
    else
	*p++ = '0';
    if (negflag < 0) {
	if ((p - buf) < ndigits)
	    p = buf + ndigits;
    }
    *p = 0;
    if (negflag >= 0 && ndigits) { /* A field width set? */
	if ((p2 + ndigits) >= p) /* Make sure we never truncate. */
	    return p - ndigits;
	else
	    return p2;
    }
    return buf;			/* No field width, return minimum number. */
}

/*
** Smaller than the standard atoi() function.
*/

myatoi(p)			/* Parse out a number. */
char *p;
{
    int val;
    char chr;

    val = 0;
    while (1) {
	chr = *p++ - '0';
	if (chr < 0 || chr > 9)
	    break;
	val = (val * 10) + chr;
    }
    return val;
}
