/*
** Item definition dialogs.
*/

#include <stdio.h>
#include <isc/keycodes.h>
#include <isc/window.h>
#include <isc/controls.h>
#include <isc/textedit.h>
#include <isc/dlgman.h>
#include "dlged.h"
#include "Dbox.h"
#include "Dline.h"
#include "Dcontrol.h"
#include "Dpicklist.h"
#include "Dpat.h"
#include "Dsbar.h"
#include "Dtext.h"
#include "Dtextbox.h"
#include "Dtextedit.h"
#include "Duseritem.h"

DefineObj(rp) 
DedItem *rp;
{
    DefPtr = rp;
    switch (rp->type) {
      case DMCHECKBOX:
      case DMPUSHBUT:
      case DMDPUSHBUT:
      case DMRADIOBUT:
	DefineControl(rp);
	break;

      case DMSCROLLBAR:
	DefineScrollbar(rp);
	break;

      case DMPATBOX:
	DefinePat(rp);
	break;

      case DMTEXTEDIT:
	DefineTE(rp);
	break;

      case DMTEXT:
	DefineTextObj(rp);
	break;

      case DMBOX:
	DefineBoxObj(rp);
	break;

      case DMLINE:
	DefineLine(rp);
	break;

      case DMPICKLIST:
	DefinePicklist(rp);
	break;

      case DMTEXTBOX:
	DefineTBox(rp);
	break;

      case DMUSERITEM:
	DefineUserItem(rp);
	break;
    }
    display(rp);
    DefPtr = 0;
}

DefinePicklist(rp)
DedItem	*rp;
{
    DMDcb *cb, *prevcb;
    int group, tmp, i;
    char buf[20];
    
    cb = DMNew(Dpicklist, windowid, 0);
    incollect = 1;
    prevcb = RfshDlg;
    RfshDlg = cb;
    group = DMCollect(cb);
    if (group == DMok) {
	dirty = 1;
	DMGet(cb, DPICKLIST_name, rp->name);
	DMGet(cb, DPICKLIST_validate, &rp->validate);
	DMGet(cb, DPICKLIST_type, buf);
	sscanf(buf, "%x", &tmp);
	rp->type = (byte) tmp;

	DMGet(cb, DPICKLIST_group, buf);
	sscanf(buf, "%x", &tmp);
	rp->group = (byte) tmp;

	DMGet(cb, DPICKLIST_key, buf);
	sscanf(buf, "%x", &rp->keycode);

	DMGet(cb, DPICKLIST_prefill, &rp->prefill);

	DMGet(cb, DPICKLIST_usercode, buf);
	sscanf(buf, "%x", &rp->usercode);

	i = rp->attr & ~(DDIhideinit | DDIinitially);
	if (DMGet(cb, DPICKLIST_inithidden, &tmp), tmp)
	    i |= DDIhideinit;
	if (DMGet(cb, DPICKLIST_initdisabled, &tmp), tmp)
	    i |= DDIinitially;
	rp->attr = i;

	i = rp->when & ~(DMdlgexit | DMWhileBefore | DMbeforeDrawing);
	if (DMGet(cb, DPICKLIST_valrexit, &tmp), tmp)
	    i |= DMdlgexit;
	if (DMGet(cb, DPICKLIST_valduring, &tmp), tmp)
	    i |= DMWhileBefore;
	if (DMGet(cb, DPICKLIST_valrdraw, &tmp), tmp)
	    i |= DMbeforeDrawing;
	rp->when = (byte) i;

	DMGet(cb, DPICKLIST_layer, buf);
	sscanf(buf, "%x", &tmp);
	rp->layer = tmp;
    }
    DMDispose(cb);
    incollect = 0;
    RfshDlg = prevcb;
}

DefineTextObj(rp)
DedItem	*rp;
{
    DMDcb *cb, *prevcb;
    int group, tmp, i;
    char buf[20];
    
    cb = DMNew(Dtext, windowid, 0);
    incollect = 1;
    prevcb = RfshDlg;
    RfshDlg = cb;
    group = DMCollect(cb);
    if (group == DMok) {
	dirty = 1;
	DMGet(cb, DTEXT_name, rp->name);
	DMGet(cb, DTEXT_prefill, rp->prefill);
	DMGet(cb, DTEXT_text, rp->text);
	i = rp->TXTflags & ~Jmask;
	if (DMGet(cb,DTEXT_left,&tmp), tmp)
	    i |= leftJust;
	if (DMGet(cb, DTEXT_right, &tmp), tmp)
	    i |= rightJust;
	if (DMGet(cb, DTEXT_center, &tmp), tmp)
	    i |= centerJust;
	if (DMGet(cb, DTEXT_top, &tmp), tmp)
	    i |= topJust;
	if (DMGet(cb, DTEXT_vcenter, &tmp), tmp)
	    i |= vcenterJust;
	if (DMGet(cb, DTEXT_bottom, &tmp), tmp)
	    i |= bottomJust;
	rp->TXTflags = i;
	erase(rp);

	DMGet(cb, DTEXT_type, buf);
	sscanf(buf, "%x", &tmp);
	rp->type = (byte) tmp;

	DMGet(cb, DTEXT_group, buf);
	sscanf(buf, "%x", &tmp);
	rp->group = (byte) tmp;

	DMGet(cb, DTEXT_usercode, buf);
	sscanf(buf, "%x", &rp->usercode);

	DMGet(cb, DTEXT_fgcolor, buf);
	sscanf(buf, "%x", &tmp);
	rp->fgColor = (byte) tmp;

	i = rp->attr & ~(DDIhideinit | DDIinitially | DDIinvisible);
	if (DMGet(cb, DTEXT_inithidden, &tmp), tmp)
	    i |= DDIhideinit;
	if (DMGet(cb, DTEXT_initdisabled, &tmp), tmp)
	    i |= DDIinitially;
	if (DMGet(cb, DTEXT_disabledhide, &tmp), tmp)
	    i |= DDIinvisible;
	rp->attr = i;

	DMGet(cb, DTEXT_bkcolor, buf);
	sscanf(buf, "%x", &tmp);
	rp->bkColor = (byte) tmp;

	DMGet(cb, DTEXT_layer, buf);
	sscanf(buf, "%x", &tmp);
	rp->layer = tmp;

	display(rp);
    }
    DMDispose(cb);
    incollect = 0;
    RfshDlg = prevcb;
}

DefineBoxObj(rp)
DedItem *rp;
{
    DMDcb *cb;
    DMDcb *prevcb;
    int group;
    int tmp;
    char buf[20];
    
    cb = DMNew(Dbox, windowid, 0);
    incollect = 1;
    prevcb = RfshDlg;
    RfshDlg = cb;
    group = DMCollect(cb);
    if (group == DMok) {
	dirty = 1;
	DMGet(cb, DBOX_name, rp->name);

	DMGet(cb, DBOX_fgcolor, buf);
	sscanf(buf, "%x", &tmp);
	rp->fgColor = (byte) tmp;

	DMGet(cb, DBOX_bkcolor, buf);
	sscanf(buf, "%x", &tmp);
	rp->bkColor = (byte) tmp;
	rp->pattern = (short) valpat;
	group = rp->attr;
	DMGet(cb, DBOX_fillnone, &tmp);
	if (tmp)
	    group &= ~BFILLED;
	else
	    group |= BFILLED;
	group &= ~(BORDEDMASK | PENMASK);
	if (DMGet(cb, DBOX_none, &tmp), tmp)
	    group |= BNONE;
	else if (DMGet(cb, DBOX_shadow, &tmp), tmp)
	    group |= BSHADOW;
	else if (DMGet(cb, DBOX_single, &tmp), tmp)
	    group |= BSINGLE;
	else if (DMGet(cb, DBOX_double, &tmp), tmp)
	    group |= BDOUBLE;
	else if (DMGet(cb, DBOX_triple, &tmp), tmp)
	    group |= BTRIPLE;

	if (DMGet(cb, DBOX_pen1x1, &tmp), tmp)
	    group |= BPENWIDTH1;
	else if (DMGet(cb, DBOX_pen2x2, &tmp), tmp)
	    group |= BPENWIDTH2;
	else if (DMGet(cb, DBOX_pen3x3, &tmp), tmp)
	    group |= BPENWIDTH3;
	DMGet(cb, DBOX_inithidden, &tmp);
	if (tmp)
	    group |= DDIhideinit;
	else
	    group &= ~DDIhideinit;
	rp->attr = group;

	DMGet(cb, DBOX_group, buf);
	sscanf(buf, "%x", &tmp);
	rp->group = (byte) tmp;
	DMGet(cb, DBOX_usercode, buf);
	sscanf(buf, "%x", &rp->usercode);
    }
    DMDispose(cb);
    incollect = 0;
    RfshDlg = prevcb;
}

DefineLine(rp)
DedItem *rp;
{
    DMDcb *cb;
    DMDcb *prevcb;
    int group;
    int tmp, size;
    char buf[20];
    Rect tr;
    
    cb = DMNew(Dline, windowid, 0);
    incollect = 1;
    prevcb = RfshDlg;
    RfshDlg = cb;
    group = DMCollect(cb);
    if (group == DMok) {
	dirty = 1;
	DMGet(cb, DLINE_name, rp->name);

	DMGet(cb, DLINE_fgcolor, buf);
	sscanf(buf, "%x", &tmp);
	rp->fgColor = (byte) tmp;

	DMGet(cb, DLINE_bkcolor, buf);
	sscanf(buf, "%x", &tmp);
	rp->bkColor = (byte) tmp;
	rp->pattern = (short) valpat;
	group = rp->attr;
	group &= ~PENMASK;
	size = 1;
	if (DMGet(cb, DLINE_pen2x2, &tmp), tmp) {
	    group |= BPENWIDTH2;
	    size = 2;
	} else if (DMGet(cb, DLINE_pen3x3, &tmp), tmp) {
	    group |= BPENWIDTH3;
	    size = 3;
	}
	if (DMGet(cb, DLINE_otherway, &tmp), tmp)
	    group |= LPENTRTOBL;
	DMGet(cb, DLINE_inithidden, &tmp);
	if (tmp)
	    group |= DDIhideinit;
	else
	    group &= ~DDIhideinit;
	rp->attr = group;
	tr = rp->rect;
	if (DMGet(cb, DLINE_forcehorz, &tmp), tmp) {
	    tr.edge.bottom = tr.edge.top + size;
	} else if (DMGet(cb, DLINE_forcevert, &tmp), tmp) {
	    tr.edge.right = tr.edge.left + size;
	}
	
	rp->rect = tr;
	DMGet(cb, DLINE_group, buf);
	sscanf(buf, "%x", &tmp);
	rp->group = (byte) tmp;
	DMGet(cb, DLINE_usercode, buf);
	sscanf(buf, "%x", &rp->usercode);
    }
    DMDispose(cb);
    incollect = 0;
    RfshDlg = prevcb;
}

DefineControl(rp)
DedItem	*rp;
{
    DMDcb *cb, *prevcb;
    int group, tmp, i;
    char buf[20];
    
    cb = DMNew(Dcontrol, windowid, 0);
    incollect = 1;
    prevcb = RfshDlg;
    RfshDlg = cb;
    group = DMCollect(cb);
    if (group == DMok) {
	dirty = 1;
	DMGet(cb, DCONTROL_text, rp->text);
	DMGet(cb, DCONTROL_name, rp->name);

	DMGet(cb, DCONTROL_fgcolor, buf);
	sscanf(buf, "%x", &tmp);
	rp->fgColor = (byte) tmp;

	DMGet(cb, DCONTROL_bkcolor, buf);
	sscanf(buf, "%x", &tmp);
	rp->bkColor = (byte) tmp;

	DMGet(cb, DCONTROL_validate, &rp->validate);
	DMGet(cb, DCONTROL_type, buf);
	sscanf(buf, "%x", &tmp);
	rp->type = (byte) tmp;

	DMGet(cb, DCONTROL_group, buf);
	sscanf(buf, "%x", &tmp);
	rp->group = (byte) tmp;

	DMGet(cb, DCONTROL_key, buf);
	sscanf(buf, "%x", &rp->keycode);

	DMGet(cb, DCONTROL_value, buf);
	sscanf(buf, "%x", &tmp);
	rp->value = (short) tmp;

	DMGet(cb, DCONTROL_minvalue, buf);
	sscanf(buf, "%x", &tmp);
	rp->min = (short) tmp;

	DMGet(cb, DCONTROL_maxvalue, buf);
	sscanf(buf, "%x", &tmp);
	rp->max = (short) tmp;

	i = rp->attr & ~(DDIinvisible | DDIinitially | DDIhideinit);
	if (DMGet(cb, DCONTROL_invisible, &tmp), tmp)
	    i |= DDIinvisible;
	if (DMGet(cb, DCONTROL_initdisabled, &tmp), tmp)
	    i |= DDIinitially;
	if (DMGet(cb, DCONTROL_inithidden, &tmp), tmp)
	    i |= DDIhideinit;
	rp->attr = i;

	DMGet(cb, DCONTROL_prefill, &rp->prefill);
	DMGet(cb, DCONTROL_usercode, buf);
	sscanf(buf, "%x", &rp->usercode);

	i = rp->when & ~(DMdlgexit | DMWhileBefore | DMWhileAfter);
	if (DMGet(cb, DCONTROL_valrexit, &tmp), tmp)
	    i |= DMdlgexit;
	if (DMGet(cb, DCONTROL_valduring, &tmp), tmp)
	    i |= DMWhileBefore;
	if (DMGet(cb, DCONTROL_valafter, &tmp), tmp)
	    i |= DMWhileAfter;
	rp->when = i;

	DMGet(cb, DCONTROL_layer, buf);
	sscanf(buf, "%x", &tmp);
	rp->layer = tmp;
    }
    DMDispose(cb);
    incollect = 0;
    RfshDlg = prevcb;

    DisposeControl(rp->id);
    newitem(rp);
}

DefineScrollbar(rp)
DedItem	*rp;
{
    DMDcb *cb, *prevcb;
    int group, tmp, i;
    char buf[20];
    
    cb = DMNew(Dsbar, windowid, 0);
    incollect = 1;
    prevcb = RfshDlg;
    RfshDlg = cb;
    group = DMCollect(cb);
    if (group == DMok) {
	dirty = 1;
	DMGet(cb, DSBAR_name, rp->name);
	DMGet(cb, DSBAR_validate, &rp->validate);
	DMGet(cb, DSBAR_type, buf);
	sscanf(buf, "%x", &tmp);
	rp->type = (byte) tmp;

	DMGet(cb, DSBAR_group, buf);
	sscanf(buf, "%x", &tmp);
	rp->group = (byte) tmp;

	DMGet(cb, DSBAR_key, buf);
	sscanf(buf, "%x", &rp->keycode);
	
	DMGet(cb, DSBAR_value, buf);
	sscanf(buf, "%x", &tmp);
	rp->value = (short) tmp;

	DMGet(cb, DSBAR_minvalue, buf);
	sscanf(buf, "%x", &tmp);
	rp->min = (short) tmp;

	DMGet(cb, DSBAR_maxvalue, buf);
	sscanf(buf, "%x", &tmp);
	rp->max = (short) tmp;

	DMGet(cb, DSBAR_butinc, buf);
	sscanf(buf, "%x", &tmp);
	rp->font = (short) tmp;

	DMGet(cb, DSBAR_pageinc, buf);
	sscanf(buf, "%x", &tmp);
	rp->fontsize = (short) tmp;

	i = rp->attr & ~(DDIinvisible | DDIinitially | DDIhideinit);
	if (DMGet(cb, DSBAR_invisible, &tmp), tmp)
	    i |= DDIinvisible;
	if (DMGet(cb, DSBAR_initdisabled, &tmp), tmp)
	    i |= DDIinitially;
	if (DMGet(cb, DSBAR_inithidden, &tmp), tmp)
	    i |= DDIhideinit;
	rp->attr = i;

	DMGet(cb, DSBAR_prefill, &rp->prefill);
	DMGet(cb, DSBAR_usercode, buf);
	sscanf(buf, "%x", &rp->usercode);

	i = rp->when & ~(DMdlgexit | DMWhileBefore | DMWhileAfter);
	if (DMGet(cb, DSBAR_valrexit, &tmp), tmp)
	    i |= DMdlgexit;
	if (DMGet(cb, DSBAR_valduring, &tmp), tmp)
	    i |= DMWhileBefore;
	if (DMGet(cb, DSBAR_valafter, &tmp), tmp)
	    i |= DMWhileAfter;
	rp->when = (byte) i;
    }
    DMDispose(cb);
    incollect = 0;
    RfshDlg = prevcb;
}

DefineTE(rp)
DedItem	*rp;
{
    DMDcb *cb, *prevcb;
    int group, tmp, i;
    char buf[20];

    cb = DMNew(Dtextedit, windowid, 0);
    incollect = 1;
    prevcb = RfshDlg;
    RfshDlg = cb;
    group = DMCollect(cb);
    if (group == DMok) {
	dirty = 1;
	DMGet(cb, DTEXTEDIT_text, rp->text);
	DMGet(cb, DTEXTEDIT_name, rp->name);
	DMGet(cb, DTEXTEDIT_tabto, rp->tabto);
	DMGet(cb, DTEXTEDIT_validate, &rp->validate);

	DMGet(cb, DTEXTEDIT_type, buf);
	sscanf(buf, "%x", &tmp);
	rp->type = (byte) tmp;

	DMGet(cb, DTEXTEDIT_group, buf);
	sscanf(buf, "%x", &tmp);
	rp->group = (byte) tmp;

	DMGet(cb, DTEXTEDIT_key, buf);
	sscanf(buf, "%x", &rp->keycode);

	DMGet(cb, DTEXTEDIT_prefill, &rp->prefill);

	DMGet(cb, DTEXTEDIT_usercode, buf);
	sscanf(buf, "%x", &rp->usercode);

	DMGet(cb, DTEXTEDIT_maxdigits, buf);
	sscanf(buf, "%x", &tmp);
	rp->max = (short) tmp;

	i = rp->when & ~(DMdlgexit | DMWhileBefore | DMTEexit | DMTEenter);
	if (DMGet(cb, DTEXTEDIT_valrexit, &tmp), tmp)
	    i |= DMdlgexit;
	if (DMGet(cb, DTEXTEDIT_valduring, &tmp), tmp)
	    i |= DMWhileBefore;
	if (DMGet(cb,DTEXTEDIT_valexit,&tmp), tmp)
	    i |= DMTEexit;
	if (DMGet(cb,DTEXTEDIT_valenter,&tmp), tmp)
	    i |= DMTEenter;
	rp->when = (byte) i;
	
	i = rp->value & ~(DItebox | DItemultiline | DItewordwrap | DItescroll
			  | DItecenterJust | DIterightJust);
	if (DMGet(cb, DTEXTEDIT_tebox, &tmp), tmp)
	    i |= DItebox;
	if (DMGet(cb, DTEXTEDIT_temultiline, &tmp), tmp)
	    i |= DItemultiline;
	if (DMGet(cb, DTEXTEDIT_tewordwrap, &tmp), tmp)
	    i |= DItewordwrap;
	if (DMGet(cb, DTEXTEDIT_tescroll, &tmp), tmp)
	    i |= DItescroll;
	TESetJust(teJustLeft, rp->id);
	if (DMGet(cb, DTEXTEDIT_justcenter, &tmp), tmp) {
	    i |= DItecenterJust;
	    TESetJust(teJustCenter, rp->id);
	}
	if (DMGet(cb, DTEXTEDIT_justright, &tmp), tmp) {
	    i |= DIterightJust;
	    TESetJust(teJustRight, rp->id);
	}
	rp->value = i;
	drawte(rp);

	i = rp->attr & ~(DDIinvisible | DDIghostText | DDIghostFrame
			 | DDIhideinit | DDIinitially);
	if (DMGet(cb, DTEXTEDIT_invisible, &tmp), tmp)
	    i |= DDIinvisible;
	if (DMGet(cb, DTEXTEDIT_ghosttext, &tmp), tmp)
	    i |= DDIghostText;
	if (DMGet(cb, DTEXTEDIT_ghostframe, &tmp), tmp)
	    i |= DDIghostFrame;
	if (DMGet(cb, DTEXTEDIT_inithidden, &tmp), tmp)
	    i |= DDIhideinit;
	if (DMGet(cb, DTEXTEDIT_initdisabled, &tmp), tmp)
	    i |= DDIinitially;
	rp->attr = i;

	DMGet(cb, DTEXTEDIT_layer, buf);
	sscanf(buf, "%x", &tmp);
	rp->layer = tmp;
    }
    DMDispose(cb);
    incollect = 0;
    RfshDlg = prevcb;
}

DefineTBox(rp)
DedItem	*rp;
{
    DMDcb *cb, *prevcb;
    int group, tmp, i;
    char buf[20];

    cb = DMNew(Dtextbox, windowid, 0);
    incollect = 1;
    prevcb = RfshDlg;
    RfshDlg = cb;
    group = DMCollect(cb);
    if (group == DMok) {
	dirty = 1;
	DMGet(cb, DTEXTBOX_text,rp->text);
	DMGet(cb, DTEXTBOX_name,rp->name);
	DMGet(cb, DTEXTBOX_validate,&rp->validate);

	DMGet(cb, DTEXTBOX_type, buf);
	sscanf(buf, "%x", &tmp);
	rp->type = (byte) tmp;

	DMGet(cb, DTEXTBOX_group, buf);
	sscanf(buf,"%x", &tmp);
	rp->group = (byte) tmp;

	DMGet(cb, DTEXTBOX_prefill, &rp->prefill);

	DMGet(cb, DTEXTBOX_usercode, buf);
	sscanf(buf, "%x", &rp->usercode);

	i = rp->value & ~(DItebox | DItewordwrap | DItecenterJust
			  | DIterightJust);
	if (DMGet(cb, DTEXTBOX_tebox, &tmp), tmp)
	    i |= DItebox;
	if (DMGet(cb, DTEXTBOX_tewordwrap, &tmp), tmp)
	    i |= DItewordwrap;
	if (DMGet(cb, DTEXTBOX_justcenter, &tmp), tmp)
	    i |= DItecenterJust;
	if (DMGet(cb, DTEXTBOX_justright, &tmp), tmp)
	    i |= DIterightJust;
	rp->value = i;
	display(rp);

	i = rp->when & ~(DMbeforeDrawing);
	if (DMGet(cb, DTEXTBOX_valrdraw, &tmp), tmp)
	    i |= DMbeforeDrawing;
	rp->when = (byte) i;

	i = rp->attr & ~(DDIinvisible | DDIghostText | DDIghostFrame
			 | DDIhideinit);
	if (DMGet(cb, DTEXTBOX_invisible, &tmp), tmp)
	    i |= DDIinvisible;
	if (DMGet(cb, DTEXTBOX_ghosttext, &tmp), tmp)
	    i |= DDIghostText;
	if (DMGet(cb, DTEXTBOX_ghostframe, &tmp), tmp)
	    i |= DDIghostFrame;
	if (DMGet(cb, DTEXTBOX_inithidden, &tmp), tmp)
	    i |= DDIhideinit;
	rp->attr = i;

	DMGet(cb, DTEXTBOX_layer, buf);
	sscanf(buf, "%x", &tmp);
	rp->layer = tmp;
    }
    DMDispose(cb);
    incollect = 0;
    RfshDlg = prevcb;
}

DefinePat(rp)
DedItem	*rp;
{
    DMDcb *cb, *prevcb;
    int group, tmp, i;
    char buf[20];
    
    cb = DMNew(Dpat, windowid, 0);
    incollect = 1;
    prevcb = RfshDlg;
    RfshDlg = cb;
    group = DMCollect(cb);
    if (group == DMok) {
	dirty = 1;
	DMGet(cb, DPAT_name, rp->name);
	DMGet(cb, DPAT_validate, &rp->validate);
	DMGet(cb, DPAT_type, buf);
	sscanf(buf, "%x", &tmp);
	rp->type = (byte) tmp;

	DMGet(cb, DPAT_group, buf);
	sscanf(buf, "%x", &tmp);
	rp->group = (byte) tmp;

	DMGet(cb, DPAT_value, buf);
	sscanf(buf, "%x", &tmp);
	rp->value = (short) tmp;

	DMGet(cb, DPAT_usercode, buf);
	sscanf(buf, "%x", &rp->usercode);

	i = rp->when & ~(DMdlgexit | DMWhileBefore);
	if (DMGet(cb, DPAT_valrexit, &tmp), tmp)
	    i |= DMdlgexit;
	if (DMGet(cb, DPAT_valduring, &tmp), tmp)
	    i |= DMWhileBefore;
	rp->when = (byte) i;

	i = rp->attr & ~(DDIhideinit | DDIinitially);
	if (DMGet(cb, DPAT_inithidden, &tmp), tmp)
	    i |= DDIhideinit;
	if (DMGet(cb, DPAT_initdisabled, &tmp), tmp)
	    i |= DDIinitially;
	rp->attr = i;

	rp->pattern = (short) valpat;
	SetCtlValues(rp->id, rp->selected, rp->pattern);
	drawcontrol(rp);
    }
    DMDispose(cb);
    incollect = 0;
    RfshDlg = prevcb;
}

DefineUserItem(rp)
DedItem	*rp;
{
    DMDcb *cb, *prevcb;
    int group, tmp, i;
    char buf[20];
    
    cb = DMNew(Duseritem, windowid, 0);
    incollect = 1;
    prevcb = RfshDlg;
    RfshDlg = cb;
    group = DMCollect(cb);
    if (group == DMok) {
	dirty = 1;
	DMGet(cb, DUSERITEM_name, rp->name);
	DMGet(cb, DUSERITEM_validate, &rp->validate);
	DMGet(cb, DUSERITEM_type, buf);
	sscanf(buf, "%x", &tmp);
	rp->type = (byte) tmp;
	
	DMGet(cb, DUSERITEM_group, buf);
	sscanf(buf, "%x", &tmp);
	rp->group = (byte) tmp;
	
	DMGet(cb, DUSERITEM_prefill, &rp->prefill);
	
	DMGet(cb, DUSERITEM_usercode, buf);
	sscanf(buf, "%x", &rp->usercode);
	
	i = rp->when & ~(DMdlgexit | DMWhileBefore | DMbeforeDrawing);
	if (DMGet(cb, DUSERITEM_valrexit, &tmp), tmp)
	    i |= DMdlgexit;
	if (DMGet(cb, DUSERITEM_valduring, &tmp), tmp)
	    i |= DMWhileBefore;
	if (DMGet(cb, DUSERITEM_valrdraw, &tmp), tmp)
	    i |= DMbeforeDrawing;
	rp->when = (byte) i;

	i = rp->attr & ~(DDIhideinit | DDIinitially);
	if (DMGet(cb, DUSERITEM_inithidden, &tmp), tmp)
	    i |= DDIhideinit;
	if (DMGet(cb, DUSERITEM_initdisabled, &tmp), tmp)
	    i |= DDIinitially;
	rp->attr = i;

	DMGet(cb, DUSERITEM_layer, buf);
	sscanf(buf, "%x", &tmp);
	rp->layer = tmp;
    }
    DMDispose(cb);
    incollect = 0;
    RfshDlg = prevcb;
}

