/*
** Miscellaneous directory-reading/caching routines and other
** mostly non-GUI support.
*/

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <dirent.h>
#include <isc/window.h>
#include "filepicker.h"

#define S_IFMPB		0070000	/* Stat type that's reserved for Netman. */

/*
** Convert the current directory's path to a pick list.
*/

makepathlist()
{
    int nitems;
    struct adirent *dip;
    char *cp, *cp2, *cp3;

    if (curdir && (cp = curdir->name)) {
	if (dip = curpath.adir) {
	    for (nitems = curpath.nitems; nitems; nitems--)
		free(dip++->d_name);
	    free(curpath.adir);
	}
	nitems = 0;
	while (cp && cp[1]) {
	    cp = strchr(&cp[1], '/');
	    nitems++;
	}
	nitems++;
	dip = (struct adirent *) malloc(sizeof(struct adirent) * (nitems+1));
	memset(dip, 0, sizeof(struct adirent) * (nitems+1));
	curpath.adir = dip;
	curpath.nitems = curpath.displayable = curpath.selectable = 
	    curpath.lastvis = nitems;
	curpath.curbar = nitems - 1;
	curpath.viewitems = dirviewitems;
	if (nitems > dirviewitems)
	    curpath.curview = nitems - dirviewitems;
	else 
	    curpath.curview = 0;
	cp = curdir->name;
	while (*cp) {
	    if (cp2 = strchr(cp, '/')) {
		cp2++;
		dip->d_ino = DIR;
		cp3 = malloc(dip->d_size = 1 + cp2 - cp);
		dip->d_name = cp3;
		strncpy(cp3, cp, cp2 - cp);
		cp3[cp2 - cp] = 0;
		cp = cp2;
		dip++;
	    }
	}
    }
}

/*
** Filter out directory entries whose names are not suitable.
*/

void
namecheck(dip)
struct adirent *dip;
{
    if (!(options & SF_DOTFILES)) {
	if (dip->d_name[0] == '.')
	    dip->d_ino = EMPTY;
    }
    if (dip->d_ino & FILE) {
	if (shmatchstr) {
	    if (!shellmatch(dip->d_name, shmatchstr))
		dip->d_ino = EMPTY;
	} else if (regmatch) {
	    if (!regexec(regmatch, dip->d_name))
		dip->d_ino = EMPTY;
	}
    }
}

/*
** Replacement regular-expression error handler.
*/

void
regerror()
{				/* Prevents printf from being linked in. */
}

/*
** Filter out directory entries whose permissions are not suitable.
*/

void
permcheck(dip, tstat)
struct adirent *dip;
struct stat *tstat;
{
    int mask, match;

    if (dip->d_ino != EMPTY) {
	if (mode == doopen)
	    mask = match = 0400; /* Look for owner-readables. */
	else if (mode == dosave)
	    mask = match = 0200; /* Look for owner-writables. */
	if (S_ISDIR(tstat->st_mode)) {
	    mask = match = 0500; /* Look for owner-usable directories */
#if 0
	    if (mode == dodirsave || mode == dosave)
		mask = match = 0300;
#endif
	}
	if (uid && uid != tstat->st_uid) { /* If we don't own the item, */
	    mask >>= 3;		/* maybe we're in its group? */
	    match >>= 3;
	    if (gid != tstat->st_gid) {	/* If we're not in the group, */
		mask >>= 3;	/* use world permission bits. */
		match >>= 3;
	    }
	}
	if (((tstat->st_mode & mask) != match)
	    || (((mode == dosave) || (mode == dodirsave))
		&& !S_ISDIR(tstat->st_mode))
	    || (tstat->st_mode & S_IFMT) == S_IFMPB) {
	    if (options & SF_GRAYBAD)
		dip->d_ino |= PERMGRAY;
	    else if (options & SF_HIDEBAD)
		dip->d_ino |= PERMHIDE;
	}
    }
}

/*
** Resolve the fate of the given directory entry.
*/

finishoff(dp, dip)
struct dircache *dp;
struct adirent *dip;
{
    struct stat tstat;
    char *pathbuf, *alloca();
    int stlen;

    stlen = dp->namelen + 1 + strlen(dip->d_name);
    pathbuf = alloca(stlen);
    strcpy(pathbuf, dp->name);
    strcat(pathbuf, dip->d_name);
    if (stat(pathbuf, &tstat)) {
	dip->d_ino = EMPTY;
    } else {
	if (S_ISDIR(tstat.st_mode))
	    dip->d_ino |= DIR;
	else {
	    dip->d_ino |= FILE;
	}
	namecheck(dip);
	permcheck(dip, &tstat);
    }
    dp->unfinished--;
    if (!(dip->d_ino & (EMPTY | NAMEHIDE | PERMHIDE))) {
	dp->displayable++;
	if ((dip - dp->adir) > dp->lastvis)
	    dp->lastvis = dip - dp->adir;
	if (!(dip->d_ino & (NAMEGRAY | PERMGRAY)))
	    dp->selectable++;
	else if (&dp->adir[dp->curbar] == dip)
	    dp->curbar++;
    }
}

/*
** In the background finish off statting the unknown directory entries.
** This is so the user interface can start as soon as it can display
** a little bit of the list.
*/

int
busywork()
{
    struct dircache *dp;
    struct adirent *dip;

    if (dp = curdir) {
	if (dp->unfinished) {
	    dip = dp->adir;
	    while (dip->d_ino != UNKNOWN)
		dip++;
	    finishoff(dp, dip);
	    setfilebar();
	} else {
	    Event(noEvent, 0);
	    compact(dp);	/* Delete empty slots for simplicity. */
	}
	return 1;
    }
    return 0;
}

/*
** Remove non-displayable entries, to simplify the job of the
** display code (which is _still_ slightly flakey with respect
** to sparse lists of filenames.
*/

compact(dp)
struct dircache *dp;
{
    int i, index, limit;
    struct adirent *dip;

    index = 0;
    limit = dp->nitems;
    dip = dp->adir;
    for (i=0; i<limit; i++) {
	if (dip->d_ino & (EMPTY | NAMEHIDE | PERMHIDE)) {
	    dp->nitems--;
	    free(dip->d_name);
	    BlockMove(&dip[1], dip, sizeof(*dip) * (dp->nitems - index));
	    if (index <= dp->curbar)
		dp->curbar--;
	    if (index <= dp->curview)
		dp->curview--;
	    if (index <= dp->curviewend)
		dp->curviewend--;
	    if (index <= dp->firstvis)
		dp->firstvis--;
	    if (index <= dp->lastvis)
		dp->lastvis--;
	} else {
	    dip++;
	    index++;
	}
    }
    dp->adir[dp->nitems].d_ino = 0; /* End terminator for safety. */
}

/*
** Qsort directory entry comparison function.
*/

int
dircmp(dip1, dip2)
struct adirent *dip1, *dip2;
{
    return strcmp(dip1->d_name, dip2->d_name);
}

/*
** Find the given directory and bring it in to memory
** if it isn't already or it has changed since last we
** brought it in.  Returns 1 if all's OK and the directory
** already existed, 2 if all's OK and it is new (or updated).
*/

int
cachedir(path)
char *path;
{
    struct dircache *dp, *ldp;
    struct adirent *dip;
    struct dirent *dep;
    struct stat tstat;
    int statted, fd, unfinished, i, cnt, maxdents;
    char dirbuf[2048];

    statted = 0;
    ldp = 0;
    dp = dirchain;
    while (dp) {
	if (!strcmp(path, dp->name)) { /* Name matches, check the date. */
	    if (stat(path, &tstat)) { /* Cached entry now is invalid, */
		free(dp->name);	/* delete it from the cache. */
		dip = dp->adir;
		for (i = dp->nitems; i; i--)
		    free(dip++->d_name);
		free(dp->adir);
		if (ldp)
		    ldp->next = dp->next;
		else
		    dirchain = dp->next;
		free(dp);
		curdir = 0;
		return 0;
	    }
	    if (tstat.st_mtime == dp->dstat.st_mtime) { /* Unchanged? */
		curdir = dp;
		return 1;	/* All's cool then. */
	    }
	    statted = 1;
	    break;
	}
	ldp = dp;
	dp = dp->next;
    }
    if (dp) {			/* If we get here it's time to (re)cache */
	dip = dp->adir;		/* the directory. */
	for (i = dp->nitems; i; i--)
	    free(dip++->d_name);
	free(dp->adir);
    } else {
	dp = (struct dircache *) malloc(sizeof *dp);
	i = strlen(path);	/* Make sure cached name ends in '/'. */
	if (path[i - 1] != '/')
	    i++;
	dp->namelen = i;
	dp->name = malloc(i + 1);
	strcpy(dp->name, path);
	dp->name[i - 1] = '/';
	dp->name[i] = 0;
	dp->next = dirchain;
	dirchain = dp;
    }
    if (statted)		/* Avoid double-statting (for speed). */
	dp->dstat = tstat;
    else {
	if (stat(path, &dp->dstat) || !S_ISDIR(dp->dstat.st_mode)) {
	  takeitout:
	    free(dp->name);	/* Stat failed, delete it from the cache. */
	    if (ldp && dp != dirchain)
		ldp->next = dp->next;
	    else
		dirchain = dp->next;
	    free(dp);
	    curdir = 0;
	    return 0;
	}
    }
    if ((fd = open(path, 0)) < 0)
	goto takeitout;
    maxdents = 128;
    dp->adir = (struct adirent *) malloc(maxdents * sizeof(struct adirent));
    cnt = i = 0;
    while (1) {			/* Read a directory. */
	if (!cnt) {
	    if ((cnt = getdents(fd, dirbuf, sizeof(dirbuf))) <= 0) {
		break;
	    }
	    dep = (struct dirent *) dirbuf;
	    cnt -= dep->d_reclen;
	} else {
	    dep = (struct dirent *) ((int) dep + dep->d_reclen);
	    cnt -= dep->d_reclen;
	}
	if (i >= maxdents) {	/* Chunk full, expand it. */
	    maxdents += 128;
	    dp->adir = (struct adirent *) realloc(dp->adir, maxdents
						  * sizeof(struct adirent));
	}
	dip = &dp->adir[i++];
	dip->d_name = (char *) malloc(dip->d_size = strlen(dep->d_name) + 1);
	strcpy(dip->d_name, dep->d_name);
    }
    close(fd);
    dip = dp->adir;
    dp->nitems = unfinished = i;
    qsort(dip, i, sizeof *dip, dircmp); /* Sort directory by name. */
    while (i--) {		/* Type of most entries is as yet unknown. */
	if (!dip->d_name[0]
	    || !strcmp(dip->d_name, "..systemfiles")
	    || (dip->d_name[0] == '.' && dip->d_size > 2 && !dip->d_name[2])) {
	    if (!dip->d_name[1] || dip->d_name[1] == '.') {
		dip->d_ino = EMPTY;
		unfinished--;
	    }
	} else {
	    dip->d_ino = UNKNOWN;
	}
	dip++;
    }
    dp->unfinished = unfinished;
    dp->displayable = dp->selectable = 0;
    dp->curview = dp->curviewend = dp->curbar = dp->firstvis = dp->lastvis = 0;
    dp->viewitems = fileviewitems;
    curdir = dp;
    return 2;
}
