/*
** File Picker program.
**
** For use by FilePicker() and NWFilePicker() libwindow calls.  I'm
** afraid this isn't written very well...  In fact, the more I work
** with/on it the crappier it seems.  Bleagh.  I ain't using _this_
** one on any resumes, that's for sure.
**
** This program uses the dialog editor to lay out the collection dialog(s),
** but it does _not_ use the Dialog Manager, largely because the pick list
** handling of DM won't handle sparse lists (ones whose information isn't
** fully known before display time).
**
** It tries very hard to present an operating user interface as soon as
** possible, which means it ends up stat-ing files that don't fit into
** the displayable part of the pick list in the background.  Also, it
** caches each directory it visits so that re-visiting these directories
** while browsing is very fast (unless they've been modified).  This makes
** it seem to run much faster than it would without these tricks, which is
** especially important on a workstation.
**
** Has many options.  Files are checked for permission, and failures
** can either be greyed out or not shown in the list.  Files can also
** be filtered by name, using either shell or regular expression pattern
** matching.  Directories (with adequate permissions) are always shown
** in the list.  Names starting with a `.` may be optionally included
** in the list(s).  The two basic modes are "Open" and "Save As", with
** slightly different behavior in the two modes.
**
** File names (full or partial) can be typed in literally.  The entry
** method understands standard Korn-shell keyboard editing sequences and
** the `~` shorthand notation.  The dialog is fully controllable from the
** keyboard without requiring a mouse (though using one certainly works).
**
** Files in the current pick list have a keyboard-driven incremental search
** applied to typed-in names.  After a short timeout the existing search
** string is deleted so that new searches may be initiated.
**
** File overwrites (on save) are checked for confirmation.
**
**	V1.2	Jim Cathey	12/9/93
**		Converted it to use dirent library, so it'll work with
**		filesystems other than classic SVR2.  Made it interpret
**		an environment of "WINDOWS_ON=ws1\nws2\nws3" (etc.) as
**		a request to throw copies of its window to ws2 and ws3.
**		(Useful for training sessions, etc.)  Made it remember
**		where it was in every directory it has visited, unless you
**		go directly to a directory via the TE box in which case it
**		presents the directory list from the beginning.  This
**		makes browsing 'feel' better.  Also, directories that have
**		changed since the last visit will do the same (reset to
**		the beginning).
**
**	V1.1	Jim Cathey	12/8/93
**		Fixed bug where the wrong file could end up selected
**		initially, on some directories.
**
**	V1.0	Jim Cathey	8/12/92
*/

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <Regexp.h>
#include <isc/window.h>
#include <isc/keycodes.h>
#include <isc/controls.h>
#include <isc/textedit.h>
#include <isc/dlgman.h>
#include "Dialog.h"
#include "Dialog2.h"
#include "Dialog3.h"
#include "filepicker.h"

#define VERSION ("V 1.2")

void fileaction(ControlID theControl, short partCode);
void diraction(ControlID theControl, short partCode);

static int xpos;		/* Default window position. */
static int ypos;

#define FILEMENU 1
#define EDITMENU 2

int dblclktime;			/* Double-click time. */
short debugmode;		/* TRUE for slow drawing. */
int window;
char *Title;
char *version = VERSION;
char *slash = "/";
char *newline = "\n";

ControlID okpush, cancelpush, dirscroll, filescroll;
ControlID okpush2, okpush3, cancelpush3;
TERec *fullpath;
#define FONT 0			/* Font to use in the dialog. */
#define SIZE 12
FontInfo finfo;
short barslot = -1;		/* For selection bar tracking. */
Rect barrect;
short dirslot = -1;		/* For selection bar tracking. */
Rect dirrect;
short teactive;
short permerror;		/* Flag for permissions error dialog. */
short overwrite;		/* Flag for overwrite query dialog. */
short binary;			/* Flag for waiting on pipe to quit. */

char scrapbuf[240];		/* Cut/Paste buffer. */
int scrapsize;			/* Amount in buffer. */
short mywindowisinfront;        /* TRUE if one of my windows is active. */
short doimportscrap = 1;        /* New incoming scrap. */
short doexportscrap;            /* New outgoing scrap. */


void updateevent();		/* Event-handling routines. */
void windowevent();
void menuevent();
void mouseevent();
void keyevent();
void activatevent();
int busywork();
short wordbrk(), wordbrk2();

main(argc, argv)
int argc;
char *argv[];
{
    int i, j, bnum, width, height;
    int trapqfd, wmfd;
    Rect winrect, button;
    char *argp, *progname, *getenv();
    char buf[80];
    HdwrType hwtype;
    void writescrap();

    mode = doopen;
    options = SF_GRAYBAD;	/* By default, gray out unsuitable files. */
    uid = geteuid();		/* "    ", match permissions to these. */
    gid = getegid();
    progname = *argv;
    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'a':		/* Show all files. */
		options |= SF_DOTFILES;
		break;
	      case 'b':		/* Binary (programmatic) mode. */
		binary = 1;
		break;
	      case 'd':		/* Directory pick mode. */
		switch (*argp++) {
		  case 'o':
		    mode = dodiropen;
		    break;
		  case 's':
		    mode = dodirsave;
		    break;
		}
		break;
	      case 'f':		/* Flush debug mode. */
		FlushDebug(debugmode = !debugmode);
		break;
	      case 'h':		/* Hide unsuitable files. */
		options &= ~SF_GRAYBAD;
		options |= SF_HIDEBAD;
		break;
	      case 'g':		/* Different GID than getegid(). */
		if (!*argp)
		    --argc, argp = *++argv;
		gid = atoi(argp);
		break;
	      case 'm':		/* Shell expression name filter. */
		if (!*argp)
		    --argc, argp = *++argv;
		shmatchstr = argp;
		break;
	      case 'o':		/* Open mode. */
		mode = doopen;
		break;
	      case 'r':		/* Regular expression name filter. */
		if (!*argp)
		    --argc, argp = *++argv;
		regmatch = regcomp(argp);
		break;
	      case 's':		/* Save mode. */
		mode = dosave;
		break;
	      case 't':		/* Different window title. */
		if (!*argp)
		    --argc, argp = *++argv;
		Title = argp;
		break;
	      case 'u':		/* Different UID than geteuid(). */
		if (!*argp)
		    --argc, argp = *++argv;
		uid = atoi(argp);
		break;
	      case 'x':		/* X position of window. */
		if (!*argp)
		    --argc, argp = *++argv;
		xpos = atoi(argp);
		break;
	      case 'y':		/* Y position of window. */
		if (!*argp)
		    --argc, argp = *++argv;
		ypos = atoi(argp);
		break;
	      case 'V':		/* Version printout. */
		strcpy(buf, progname);
		strcat(buf, " Version ");
		strcat(buf, &version[2]);
		strcat(buf, newline);
		write(1, buf, StringLen(buf));
		exit(0);
		break;
	    }
	} else {
	    initialpath = argp;
	}
    }
    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
	exit(2);
    }
    if ((wmfd = ConnectGraf(0)) < 0) {	/* Initialize graphics toolbox. */
	exit(3);
    }

    if (!Title) {
	switch (mode) {
	  case doopen:
	    Title = "Open File\311";
	    break;
	  case dosave:
	    Title = "Save As\311";
	    break;
	  case dodiropen:
	  case dodirsave:
	    Title = "Select Directory\311";
	    break;
	}
    }
    Event(updateEvt, updateevent);
    Event(windowEvt, windowevent);
    Event(menuEvt, menuevent);
    Event(mouseEvt, mouseevent);
    Event(keyEvt, keyevent);
    Event(activateEvt, activatevent);

    width = Dialog[DIALOG_Frame].rect.edge.right
	- Dialog[DIALOG_Frame].rect.edge.left;
    height = Dialog[DIALOG_Frame].rect.edge.bottom
	- Dialog[DIALOG_Frame].rect.edge.top;
    GetHdwrType(&hwtype);
    if (!xpos)
	xpos = (hwtype.hdots - width) / 2; /* Auto-center the window. */
    if (!ypos)
	ypos = (hwtype.vdots - height) / 3;
    SetRect(&winrect, xpos, ypos, xpos+width, ypos+height);
    if (xpos < 0)		/* Keep window on screen. */
	OffsetRect(&winrect, -xpos, 0);
    if (ypos < 0)
	OffsetRect(&winrect, 0, -ypos);
    if (winrect.edge.right > hwtype.hdots)
	OffsetRect(&winrect, hwtype.hdots - winrect.edge.right, 0);
    if (winrect.edge.bottom > hwtype.vdots)
	OffsetRect(&winrect, hwtype.vdots - winrect.edge.bottom, 0);
    xpos = winrect.edge.left;	/* For window movement later. */
    ypos = winrect.edge.top;
    window = NewWindow(&winrect, Title, showWindow, documentProc,
		       frontWindow, hasGoAway|hasSendToBack|hasDrag, wmfd);
    SetPort(window);
    newmenus();
    LocalGetFontInfo(FONT, normalText, SIZE, &finfo);
    dblclktime = 80;		/* GetDblClickTime(); */
    finfo.leading += finfo.ascent + finfo.descent;
    fileviewitems = (Dialog[DIALOG_files].rect.edge.bottom
		     - Dialog[DIALOG_files].rect.edge.top) / finfo.leading;
    dirviewitems = (Dialog[DIALOG_directories].rect.edge.bottom
		     - Dialog[DIALOG_directories].rect.edge.top)/finfo.leading;
    InitControls();
    TEInit();
    initdialog();
    BeginUpdate(window);	/* Avoid initial update event. */
    drawdialog();
    EndUpdate(window);

    if (argp = getenv("WINDOWS_ON")) { /* Handle copying video to other WS. */
	char *cp; int fd; long winds;

	winds = 0;
	while (argp = strchr(argp, '\n')) {
	    argp++;
	    cp = strchr(argp, '\n');
	    if (!cp)
		cp = argp + strlen(argp);
	    if (cp - argp) {
		strncpy(buf, argp, cp - argp);
		buf[cp - argp] = 0;
		if ((fd = ConnectGraf(buf)) >= 0) {
		    winds |= 1 | 1 << NewWindow(&winrect, Title, showWindow,
						documentProc, frontWindow,
						hasSendToBack, fd);
		}
	    }
	}
	if (winds)
	    SetPorts(winds);
    }
    FlushWindow();		/* Get GUI stuff going while we mungle some */
    parsepath();		/* directories for awhile. */
  more:
    if (GetEvent(anyWindow, noEvent)) { /* Process user interaction. */
	if ((i = (EventError() >> 16)) != wmfd) {
	    DisconnectGraf(i);	/* Video copy died, take it out & continue. */
	    goto more;
	}
	return 4;		/* Continues after a ReleaseEvent (rets 0). */
    }				/* GetEvent returns EOF on connection error. */
    if (doexportscrap) {
	writescrap();
	FlushTwice();
    }
    return 0;
}

/*
** Figure out where we should be from the given full path name.  It
** may be the name of a directory or a file.  If a file, it will be
** the thing initially selected by the dialog.  If a directory, the
** first element in it will be selected.  If the path is bad, as much
** of the path as is good will be used, with the closest name match
** of the bad part selected.
*/

parsepath()
{
    int item, retval, i;
    struct stat stbuf;
    char *cp;

    if (!initialpath) {
	if (!(initialpath = getcwd(0, 1024)))
	    exit(5);
    } else {
	initialpath = twiddle(initialpath, uid); /* Make sure we've */
	if (initialpath) {
	    if (*initialpath != '/') { /* an absolute path. */
		char *cwd, *newwd;
		int len;
		cwd = getcwd(0, 1024);
		len = StringLen(cwd) + 2;
		len += StringLen(initialpath);
		newwd = malloc(len);
		strcpy(newwd, cwd);
		strcat(newwd, slash);
		strcat(newwd, initialpath);
		free(cwd);
		initialpath = newwd;
	    }
	} else {
	    return 0;
	}
    }
    canon(initialpath);		/* Clean up the pathname. */
    retval = 0;
    if (mode == dosave) {	/* Save the suggested base name (if any). */
	if (cp = strrchr(initialpath, '/')) {
	    basename.d_ino = StringLen(&cp[1]);
	    basename.d_name = malloc(basename.d_size = basename.d_ino + 1);
	    strcpy(basename.d_name, &cp[1]);
	    retval = 1;
	}
    }
    newsearchstr("\001");	/* Will 'match' the first entry. */
    if (!stat(initialpath, &stbuf)) {
	if (S_ISDIR(stbuf.st_mode)) {
	    retval = 0;
	    basename.d_ino = 0;	/* Was a dir only, no base name. */
	    if (initialpath[StringLen(initialpath) - 1] != '/')
		strcat(initialpath, slash);
	} else if (S_ISREG(stbuf.st_mode)) {
	    retval = 1;
	}
    }
    while (1) {
	cachedir(initialpath);
	if (curdir && !strncmp(initialpath, curdir->name,
			       StringLen(initialpath)))
	    break;
	item = StringLen(initialpath);
	if (initialpath[item - 1] == '/') /* Lop off trailing slash, if any. */
	    initialpath[item - 1] = 0;
	if (!(cp = strrchr(initialpath, '/')))
	    break;
	cp++;
	newsearchstr(cp);
	*cp = 0;		/* Lop off the last part of name & repeat. */
    }
    if (curdir && curdir->unfinished) /* Any background work to do? */
	Event(noEvent, (void (*)()) busywork); /* Yes, schedule it. */
    else
	Event(noEvent, 0);
    showfiles();
    putupfilebar();
    setfilebar();
    item = doisearch();
    i = item - curdir->viewitems / 2; /* Fake the current view position */
    if (i < curdir->firstvis)	/* if necessary to avoid statting everything */
	i = curdir->firstvis;	/* at first. */
    if (curdir->unfinished && item > curdir->curviewend) { /* Don't do it if */
	curdir->curview = i;	/* the selected item's already visible, or */
	curdir->curbar = curdir->curview; /* we screw up big time! */
    }
    selectfile(item);
    makepathlist();
    removedirbar();
    erasepathlist();
    showdirs();
    putupdirbar();
    updateTE();
    return retval;
}

newsearchstr(str)
char *str;
{
    if (searchstring.d_name) {
	free(searchstring.d_name);
    }
    searchstring.d_ino = StringLen(str);
    searchstring.d_size = searchstring.d_ino + 1;
    searchstring.d_name = malloc(searchstring.d_size);
    strcpy(searchstring.d_name, str);
}

/*
** Search for a matching (or nearly so) directory entry.  Used for
** the keyboard-driven incremental search.
*/

doisearch()
{
    int nitems, sitem, matchval;
    struct adirent *dip;

    dip = curdir->adir;
    nitems = curdir->nitems;
    if (nitems && searchstring.d_ino) {
	for (sitem = 0; sitem < nitems; sitem++) {
	    matchval = strncmp(searchstring.d_name, dip->d_name,
			       searchstring.d_ino);
	    if (matchval <= 0)
		return sitem;
	    dip++;
	}
    }
    return curdir->nitems - 1;
}

/*
** Move the file selection bar to the given item, and scroll
** the list if necessary to make the bar visible.  Skips over
** unselectable items.
*/

selectfile(item)
int item;
{
    int delta, newview;
    struct dircache *dp;
    struct adirent *dip;

    dp = curdir;
    if (item < dp->firstvis || item >= dp->nitems)
	return;			/* Sanity check. */
    delta = item - dp->curbar;
    dip = &dp->adir[item];
    while (1) {			/* Make sure the target is selectable. */
	if (dp->unfinished && dip->d_ino == UNKNOWN)
	    finishoff(dp, dip);
	if (!(dip->d_ino & (EMPTY | NAMEHIDE|PERMHIDE | NAMEGRAY|PERMGRAY)))
	    break;		/* Found a good one to start on, use it. */
	if (delta > 0) {	/* Moving forward. */
	    item++;
	    dip++;
	    if (item >= dp->nitems) { /* Nothing left to select, */
		delta = -1;	/* go the other way. */
		item--;
		dip--;
	    }
	} else {		/* Moving backward. */
	    item--;
	    dip--;
	    if (item < dp->firstvis) /* Nothing left to select. */
		return;
	}
    }
    if (item - dp->curbar) {	/* Moved? */
	removefilebar();
	dp->curbar = item;
	newview = item;
	if (item < dp->curview || item > dp->curviewend) {
	    if (item < dp->curview) /* Need to move the view down? */
		delta = 3;
	    else		/* Need to move the view up. */
		delta = dp->viewitems - 3;
	    while (delta && newview > dp->firstvis) {
		dip--;
		newview--;
		if (dp->unfinished && dip->d_ino == UNKNOWN)
		    finishoff(dp, dip);
		if (!(dip->d_ino & (EMPTY | NAMEHIDE | PERMHIDE)))
		    delta--;
	    }
	    scrollfileto(newview);
	} else {
	    putupfilebar();
	}
    }
}

/*
** Display the currently visible part of the file pick list,
** stat-ing any unfinished entries that might be needed.
*/

showfiles()
{
    int len, face, x, y, xlimit, ylimit, ystart, nitems, slot;
    DMItem *itemp;
    struct dircache *dp;
    struct adirent *dip;
    Rect scratch;

    if (dp = curdir) {
	itemp = &Dialog[DIALOG_files];
	x = itemp->rect.edge.left + 3;
	y = ystart = itemp->rect.edge.top + finfo.ascent + 2;
	ylimit = itemp->rect.edge.bottom - 3;
	xlimit = itemp->rect.edge.right - 17;
	scratch = itemp->rect;
	scratch.edge.right = xlimit;
	ClipRect(&scratch);
	dip = &dp->adir[dp->curview];
	nitems = dp->nitems - dp->curview;
	TextMode(srcCopy);
	slot = 0;
	while (y < ylimit && nitems--) {
	    MoveTo(x, y);
	    if (dp->unfinished && dip->d_ino == UNKNOWN)
		finishoff(dp, dip);
	    if (dip->d_ino & (FILE | DIR)) {
		if (!(dip->d_ino & (NAMEHIDE | PERMHIDE))) {
		    if (dip->d_ino & (NAMEGRAY | PERMGRAY))
			face = ghostText;
		    else
			face = normalText;
		    TextFace(face);
		    DrawString(dip->d_name);
		    dp->curviewend = dip - dp->adir;
		    len = LocalStringWidth(FONT, face, SIZE, dip->d_name);
		    if (dip->d_ino & DIR) {
			DrawChar('/');
			len += LocalCharWidth(FONT, face, SIZE, '/');
		    }
		    scratch.edge.top = y - finfo.ascent - 1;
		    scratch.edge.bottom = scratch.edge.top
			+ finfo.leading;
		    scratch.edge.right = xlimit;
		    scratch.edge.left = x + len;
		    EraseRect(&scratch);
		    y += finfo.leading;
		    slot++;
		}
	    } else if (y == ystart) { /* Started on an un-good entry? */
		if (dp->curview == dp->curbar)
		    dp->curbar++;
		if (dp->curview == dp->firstvis)
		    dp->firstvis++;
		dp->curview++;	/* Move to a good one. */
	    }
	    dip++;
	}
	scratch.edge.right = itemp->rect.edge.right - 17; /* Erase the */
	scratch.edge.left = itemp->rect.edge.left + 1; /* unused bottom. */
	scratch.edge.top = y - finfo.ascent;
	scratch.edge.bottom = itemp->rect.edge.bottom - 1;
	EraseRect(&scratch);
	ClipRect(&Dialog[0].rect);
	TextMode(srcOr);
	TextFace(normalText);
    }
}

/*
** Put up file selection bar.
*/

putupfilebar()
{
    int slot, y;
    struct dircache *dp;
    struct adirent *dip;
    DMItem *itemp;
    Rect scratch;

    if ((dp = curdir) && dp->curbar >= dp->curview
	&& dp->curbar <= dp->curviewend && !teactive) {
	slot = 0;
	dip = &dp->adir[y = dp->curview];
	while (y < dp->curbar) { /* Find position in visible list. */
	    if (!(dip->d_ino & (EMPTY | NAMEHIDE | PERMHIDE))) {
		slot++;
	    }
	    y++;
	    dip++;
	}
	if (dp->selectable && slot >= 0 && slot < dp->viewitems) {
	    itemp = &Dialog[DIALOG_files];
	    y = itemp->rect.edge.top + finfo.ascent + 2;
	    y += slot * finfo.leading;
	    scratch.edge.top = y - finfo.ascent - 1;
	    scratch.edge.bottom = scratch.edge.top
		+ finfo.leading;
	    scratch.edge.right = itemp->rect.edge.right - 17;
	    scratch.edge.left = itemp->rect.edge.left + 1;
	    barslot = slot;
	    barrect = scratch;
	    InvertRect(&scratch);
	}
    }
}

/*
** Remove file selection bar if present.
*/

removefilebar()
{
    if (barslot >= 0) {
	InvertRect(&barrect);
	barslot = -1;
    }
}

/*
** Put up directory selection bar.
*/

putupdirbar()
{
    int slot, y;
    struct dircache *dp;
    DMItem *itemp;
    Rect scratch;

    dp = &curpath;
    slot = dp->curbar - dp->curview; /* Position in visible list. */
    if (slot >= 0 && slot < dp->viewitems) {
	itemp = &Dialog[DIALOG_directories];
	y = itemp->rect.edge.top + finfo.ascent + 2;
	y += slot * finfo.leading;
	scratch.edge.top = y - finfo.ascent - 1;
	scratch.edge.bottom = scratch.edge.top
	    + finfo.leading;
	scratch.edge.right = itemp->rect.edge.right - 17;
	scratch.edge.left = itemp->rect.edge.left + 1;
	dirslot = slot;
	dirrect = scratch;
	InvertRect(&scratch);
    }
}

/*
** Remove directory selection bar if present.
*/

removedirbar()
{
    if (dirslot >= 0) {
	InvertRect(&dirrect);
	dirslot = -1;
    }
}

/*
** Move the selection bar in the path list.
*/

selectpath(item)
int item;
{
    if (item < 0 || item > curpath.nitems)
	return;
    removedirbar();
    curpath.curbar = item;
    putupdirbar();
}

/*
** Make the TE contain the current full path.
*/

updateTE()
{
    DMItem *itemp;
    int selstart;

    itemp = &Dialog[DIALOG_fullname];
    TESetText(curdir->name, 1024, fullpath);
    TEUpdateErase(&itemp->rect, fullpath);
    selstart = StringLen(curdir->name);
    if (mode == dosave) {
	if (basename.d_ino)
	    TEInsert(basename.d_name, basename.d_ino, fullpath);
	TESetSelect(selstart, selstart + basename.d_ino, fullpath);
	TEActivate(fullpath);
	removefilebar();
	teactive = 1;
    }
}

/*
** Move the current directory up in the current path.
*/

newpath(item)
int item;
{
    struct dircache *dp;
    struct adirent *dip;
    char *cp, buf[1024];

    if (item < 0 || item > curpath.nitems)
	return;
    dp = &curpath;
    dip = curpath.adir;
    buf[0] = 0;			/* Build the new path string. */
    do {
	strcat(buf, dip->d_name);
	cp = buf + StringLen(buf) - 1;
	if (*cp++ != '/') {
	    *cp++ = '/';
	    *cp = 0;
	}
	dip++;
    } while (item--);
    removefilebar();
    removedirbar();
    newsearchstr(dip->d_name);	/* Move selection bar to the entry for */
    if (cp = strchr(searchstring.d_name, '/'))
	*cp = 0;		/* the next directory (if any). */
    cachedir(buf);
    showfiles();
    if (!(mode == dosave && teactive))
	putupfilebar();
    setfilebar();
    item = doisearch();
    if (item >= 0)
	selectfile(item);
    makepathlist();
    Event(noEvent, 0);
    if (curdir && curdir->unfinished) /* Any background work to do? */
	Event(noEvent, (void (*)()) busywork); /* Yes, schedule it. */
    erasepathlist();
    removedirbar();
    showdirs();
    putupdirbar();
    updateTE();
}

/*
** Draw the directories pick list.
*/

showdirs()
{
    int x, y, ylimit, nitems, len;
    DMItem *itemp;
    struct adirent *dip;
    Rect scratch;

    if (curdir) {
	itemp = &Dialog[DIALOG_directories];
	scratch = itemp->rect;
	scratch.edge.right -= 17;
	ClipRect(&scratch);
	x = itemp->rect.edge.left + 3;
	y = itemp->rect.edge.top + finfo.ascent + 2;
	ylimit = itemp->rect.edge.bottom - 3;
	TextMode(srcCopy);
	nitems = curpath.nitems - curpath.curview;
	dip = curpath.adir + curpath.curview;
	while (y < ylimit && nitems--) {
	    MoveTo(x, y);
	    len = StringLen(dip->d_name);
	    DrawText(dip->d_name, 0, len);
	    if (dip->d_name[len - 1] != '/')
		DrawChar('/');
	    dip++;
	    y += finfo.leading;
	}
	TextMode(srcOr);
	scratch.edge.right = itemp->rect.edge.right - 17; /* Erase the */
	scratch.edge.left = itemp->rect.edge.left + 1; /* unused bottom. */
	scratch.edge.top = y - finfo.ascent;
	scratch.edge.bottom = itemp->rect.edge.bottom - 1;
	EraseRect(&scratch);
	ClipRect(&Dialog[0].rect);
	setdirbar();
    }
}

erasepathlist()
{
    DMItem *itemp;
    Rect scratch;

    itemp = &Dialog[DIALOG_directories];
    scratch.edge.top = itemp->rect.edge.top + 1;
    scratch.edge.right = itemp->rect.edge.right - 17; /* Erase the */
    scratch.edge.left = itemp->rect.edge.left + 1; /* unused bottom. */
    scratch.edge.bottom = itemp->rect.edge.bottom - 1;
    EraseRect(&scratch);
}

newmenus()
{
    NewMenu(FILEMENU, "File");
    AppendMenu(FILEMENU, "/QQuit");
    NewMenu(EDITMENU, "Edit");
    AppendMenu(EDITMENU, "(/ZUndo;-;/XCut;/CCopy;/VPaste");
}

/*
** Handle a provisional selection, returning the full path name if it's
** a selectable target, or moving the current directory as necessary.
*/

finishup(reset)
int reset;			/* TRUE to reset list to its start. */
{
    char buf[1024];
    struct adirent *dip;
    int new;

    removefilebar();
    removedirbar();
    if (curdir) {
	dip = &curdir->adir[curdir->curbar];
	strcpy(buf, curdir->name);
	strcat(buf, dip->d_name);
	if (dip->d_ino & DIR) {
	    if (buf[StringLen(buf) - 1] != '/')
		strcat(buf, slash);
	    if (new = cachedir(buf)) {
		erasepathlist();
		if (curdir && curdir->unfinished) /* Any bkgnd work to do? */
		    Event(noEvent, (void (*)()) busywork); /* Yes, schedule. */
		else
		    Event(noEvent, 0);
		showfiles();
		putupfilebar();
		setfilebar();
		if (reset || new == 2) { /* Select first entry? */
		    newsearchstr("\001"); /* Will 'match' the first */
		    selectfile(doisearch()); /* selectable entry. */
		}
		makepathlist();
		showdirs();
		putupdirbar();
		updateTE();
		return 0;
	    }
	} else {
	    spitout(buf);
	}
	return 1;			/* Time to exit. */
    }
    return 0;
}

spitout(buf)
char *buf;
{
    short i;
    char *p, *alloca();

    i = StringLen(buf);
    p = alloca(i + 1 + sizeof i);
    if (binary) {		/* Programmatic interface? */
	*(short *) p = i;
	BlockMove(buf, &p[sizeof i], i);
	write(1, p, i + sizeof i); /* Send client length & the string. */
	read(0, &i, sizeof i);	/* Synchronize, wait for client finished. */
    } else {			/* No, was manually run from a shell. */
	BlockMove(buf, p, i);
	p[i] = '\n';
	write(1, p, i + 1);	/* Send client the string. */
    }
}

void
updateevent(ep)
EventRecord *ep;
{
    long save;

    GetPort(&save);
    SetPort(ep->who);
    BeginUpdate(ep->who);
    drawdialogs();
    EndUpdate(ep->who);
    SetPort(save);
}

void
windowevent(ep)
EventRecord *ep;
{
    if (ep->who != window) {
	return;
    }
    switch (ep->message & F_KEYMASK) {
      case windowGoaway:
	ReleaseEvent();
	break;

      case windowMove:
	xpos += ep->where.edge.left;
	ypos += ep->where.edge.top;
	MoveWindow(ep->who, xpos, ypos, 0);
	break;
    }
}

void
menuevent(ep)
EventRecord *ep;
{
    void cut(), copy(), paste();
    short menu, item;

    BeginMenu(ep->message);
    menu = (ep->message >> 16) & 0xFFFF;
    item = ep->message & 0xFFFF;
    switch (menu) {
      case FILEMENU:
	ReleaseEvent();
	break;

      case EDITMENU:		/* Edit Menu */
	switch (item) {
	  case 3:		/* Cut. */
	    cut();
	    break;

	  case 4:		/* Copy. */
	    copy();
	    break;

	  case 5:		/* Paste. */
	    paste();
	    break;
        }
        break;
    }
    EndMenu(ep->message);
}


void
killscrap()
{
    scrapsize = 0;              /* Delete private scrap. */
    doimportscrap = 0;
}

void
fetchscrap()
{
    long size;

    killscrap();		/* Create private scrap. */
    scrapsize = GetScrap(scrapbuf, 'TEXT', sizeof(scrapbuf));
    TESetScrapLen(scrapsize);
    if (scrapsize) {
	BlockMove(scrapbuf, TEGetScrapAddr(), scrapsize);
	EnableItem(EDITMENU, 5);
    } else
	DisableItem(EDITMENU, 5);
}

void
writescrap()
{
    if (doexportscrap) {
	ZeroScrap();		/* Export private scrap. */
	PutScrap(scrapsize, 'TEXT', scrapbuf);
	PublishScrap(0);
	doexportscrap = 0;
    }
}

void
cut()				/* User-activated CUT command. */
{
    killscrap();
    if (teactive) {
	TECut(fullpath);
	if (scrapsize = TEGetScrapLen()) {
	    BlockMove(TEGetScrapAddr(), scrapbuf, scrapsize);
	    EnableItem(EDITMENU, 5);
	} else
	    DisableItem(EDITMENU, 5);
	doexportscrap = 1;	/* Set export-on-deactivate flag. */
    } else
	copy();
}

void
copy()				/* User-activated COPY command. */
{
    killscrap();
    if (teactive) {
	TECopy(fullpath);
	if (scrapsize = TEGetScrapLen()) {
	    BlockMove(TEGetScrapAddr(), scrapbuf, scrapsize);
	    EnableItem(EDITMENU, 5);
	} else
	    DisableItem(EDITMENU, 5);
    } else {
	scrapsize = StringLen(curdir->name);
	BlockMove(curdir->name, scrapbuf, scrapsize);
    }
    doexportscrap = 1;		/* Set export-on-deactivate flag. */
    if (scrapsize)
	EnableItem(EDITMENU, 5);
    else
	DisableItem(EDITMENU, 5);
}

void
paste()				/* User-activated PASTE command. */
{
    if (teactive)
	TEPaste(fullpath);
    else
	Bell();
}

void
scrapevent(ep)
EventRecord *ep;
{
    if (!ep->message) {		/* Someone published new scrap. */
	doimportscrap = 1;
	if (mywindowisinfront)
	    fetchscrap();
    } else {			/* Someone wants our (?) scrap. */
	if (doexportscrap && mywindowisinfront)
	    writescrap();
    }
}

/*
** Handle mousies from the operator.
*/

void
mouseevent(ep)
EventRecord *ep;
{
    ControlID cid;
    short i, j;
    int typed;
    static long lasttime;
    Rect hit, hit2, hit3;
    struct adirent *dip;

    if (ep->who != window) {
	long save;

	if (ep->message & F_KEYDOWN) {
	    GetPort(&save);
	    SetPort(ep->who);
	    Bell();
	    SetPort(save);
	}
	return;
    }
    if (ep->message & F_KEYUP) {
	lasttime = ep->when;
	return;
    }
    if (ep->message & F_KEYDOWN) { /* Mouse-down event? */
	if (FindControl(&ep->where.corner.topLeft, ep->who, &cid)) {
	    i = TrackControl(cid, &ep->where.corner.topLeft, (void (*)()) -1);
	    if (i == inButton) {
		if (cid == cancelpush) {
		    ReleaseEvent();
		} else if (cid == okpush) {
		    typed = 0;
		    i = 1;
		    if (teactive) {
			typed = 1;
			if (mode != dosave) {
			    teactive = 0;
			    TEDeactivate(fullpath);
			}
			i = readtepath();
			if (mode == dosave && i) {
			    ReleaseEvent();
			}
		    }
		    if (i && finishup(typed))
			ReleaseEvent();
		} else if (cid == okpush2) {
		    takedownalert();
		} else if (cid == okpush3) { /* Approved overwrite. */
		    char buf[1024];
		    i = TEGetText(buf, fullpath);
		    buf[i] = 0;
		    spitout(buf);
		    ReleaseEvent();
		} else if (cid == cancelpush3) {
		    takedownalert();
		}
	    } else if (i == inThumb) {
		if (cid == filescroll) {
		    i = GetCtlValue(cid) + curdir->firstvis;
		    i = ScaleRS(i, (curdir->nitems - curdir->viewitems),
				curdir->nitems);
		    if (i < curdir->firstvis)
			i = curdir->firstvis;
		    if (i > curdir->nitems - curdir->viewitems)
			i = curdir->nitems - curdir->viewitems;
		    scrollfileto(i); /* Acknowledge thumb drags. */
		} else if (cid == dirscroll) {
		    i = GetCtlValue(cid);
		    i = ScaleRS(i, (curpath.nitems - curpath.viewitems),
				curpath.nitems);
		    scrolldirto(i); /* Acknowledge thumb drags. */
		}
	    }
	} else if (!(permerror | overwrite)) { /* Pick list hit? */
	    hit = Dialog[DIALOG_files].rect;
	    hit.edge.right -= 16;
	    hit2 = Dialog[DIALOG_directories].rect;
	    hit2.edge.right -= 16;
	    hit3 = Dialog[DIALOG_fullname].rect;
	    if (PtInRect(&ep->where.corner.topLeft, &hit)) { /* File picker? */
		if (teactive) {
		    teactive = 0;
		    TEDeactivate(fullpath);
		    putupfilebar();
		}
		j = (ep->where.edge.top - Dialog[DIALOG_files].rect.edge.top)
		    / finfo.leading;
		i = curdir->curview;
		while (i < curdir->nitems) { /* Skip over */
		    dip = &curdir->adir[i]; /* non-entities. */
		    while (i < curdir->nitems && 
			   dip->d_ino & (EMPTY | NAMEHIDE | PERMHIDE)) {
			i++;
			dip++;
		    }
		    if (!j--)
			break;
		    i++;
		}
		if (i < curdir->nitems) {
		    if (!(curdir->adir[i].d_ino & (NAMEGRAY | PERMGRAY))) {
			if (ep->when < (lasttime + dblclktime) /* Dbl-click? */
			    && i == curdir->curbar) {
			    if (finishup(0))
				ReleaseEvent();
			} else {
			    selectfile(i);
			}
		    }
		} else {
		    Bell();
		}
	    } else if (PtInRect(&ep->where.corner.topLeft, &hit2)) {
		i = (ep->where.edge.top /* Path picker? */
		     - Dialog[DIALOG_directories].rect.edge.top)
		    / finfo.leading;
		i += curpath.curview;
		if (i < curpath.nitems) {
		    if (ep->when < (lasttime + dblclktime) /* Double-click? */
			&& i == curpath.curbar) {
			newpath(i);
		    } else {
			selectpath(i);
		    }
		} else {
		    Bell();
		}
	    } else if (PtInRect(&ep->where.corner.topLeft, &hit3)) {
		if (!teactive) {
		    teactive = 1;
		    removefilebar();
		    TEActivate(fullpath);
		}
		TEClick(&ep->where.corner.topLeft, !!(ep->message & F_SHIFT),
			fullpath);
	    } else {
		Bell();
	    }
	} else {
	    Bell();
	}
    }
}

/*
** Read a pathname out of the TE, and either take it or move to it
** depending on whether it's a file or a directory.
*/

readtepath()
{
    char buf[1024];
    struct stat stbuf;
    short mask;
    int i;

    i = TEGetText(buf, fullpath);
    buf[i] = 0;
    initialpath = buf;
    removefilebar();
    removedirbar();
    if (!parsepath())
	return 0;
    if (curdir && mode == dosave && basename.d_ino) {
	strcpy(buf, curdir->name);
	strncat(buf, basename.d_name, basename.d_ino);
	mask = 0200;
	if (stat(buf, &stbuf)) { /* File doesn't exist, can we write it? */
	    if (uid && uid != curdir->dstat.st_uid) {
		mask >>= 3;
		if (gid != curdir->dstat.st_gid) {
		    mask >>= 3;
		}
	    }
	    if (curdir->dstat.st_mode & mask) {
		spitout(buf);
		return 1;	/* Can create it, allow us to exit. */
	    }
	} else {		/* File exists, check writeability. */
	    if (uid && uid != stbuf.st_uid) {
		mask >>= 3;
		if (gid != stbuf.st_gid) {
		    mask >>= 3;
		}
	    }
	    if (stbuf.st_mode & mask) {
		overwrite = 1;
		Bell();
		removedirbar();
		removefilebar();
		drawdialogs();
		return 0;	/* Ask if we should write it. */
	    }
	}
	permerror = 1;
	Bell();
	removedirbar();
	removefilebar();
	drawdialogs();
	return 0;		/* Bad name, don't quit. */
    }
    return 1;			/* It's time to quit. */
}

/*
** Remove the alert from the window.
*/

takedownalert()
{
    HiliteControl(dirscroll, 0);
    HiliteControl(filescroll, 0);
    if (permerror) {
	permerror = 0;
	HideControlNoInval(okpush2);
	InvalRect(&Dialog2[DIALOG2_Frame2].rect);
	EraseRect(&Dialog2[DIALOG2_Frame2].rect);
    } else if (overwrite) {
	overwrite = 0;
	HideControlNoInval(okpush3);
	HideControlNoInval(cancelpush3);
	InvalRect(&Dialog3[DIALOG3_Frame2].rect);
	EraseRect(&Dialog3[DIALOG3_Frame2].rect);
    }
    BeginUpdate(window);
    drawdialog();
    EndUpdate(window);
    putupfilebar();
    putupdirbar();
}

short
wordbrk(text, charPos)		/* Custom Text Edit word-break routine. */
char *text;
short charPos;
{
    if (text[charPos] <= ' ' || text[charPos] == '/')
	return 1;
    return 0;
}

short
wordbrk2(text, charPos)		/* Custom Text Edit word-break routine. */
char *text;
short charPos;
{
    if (text[charPos] <= ' ' || (charPos && text[charPos - 1] == '/'))
	return 1;
    return 0;
}

/*
** Handle keystrokes from the operator.
*/

void
keyevent(ep)
EventRecord *ep;
{
    short key, i;
    int typed;
    static long lasttime;
    ControlID cid;

    if (ep->who != window) {
	long save;

	if (ep->message & F_KEYDOWN) {
	    GetPort(&save);
	    SetPort(ep->who);
	    Bell();
	    SetPort(save);
	}
	return;
    }
    if (ep->message & F_KEYUP) {
	lasttime = (ep->message & F_FUNC) ? 0 : ep->when;
	return;
    }
    SetWordBreak((ep->message & F_ALT) ? wordbrk2 : wordbrk, fullpath);
    key = ep->message & F_KEYMASK;
    i = (ep->message & F_SHIFT) ? inPageDown : inPageUp;
    if (ep->message & F_FUNC) {
	lasttime = 0;		/* Function keys terminate the i-search. */
	switch (key) {
	  case K_ESC:		/* Quitting time! */
	  case K_RESET:
	  takedown2:
	    cid = cancelpush;
	    if (permerror)
		cid = okpush2;
	    else if (overwrite)
		cid = cancelpush3;
	    HiliteControl(cid, inButton);
	    FlushWindow();	/* Flash the button. */
	    nap(100);
	    HiliteControl(cid, 0);
	    if (!(permerror | overwrite))
		ReleaseEvent();
	    else {
		takedownalert();
	    }
	    break;

	  case K_ENTER:
	  case K_NPENTER:
	  case K_SEND:
	    cid = okpush;
	    if (permerror)
		goto takedown2;
	    else if (overwrite)
		cid = okpush3;
	    HiliteControl(cid, inButton);
	    FlushWindow();	/* Flash the button. */
	    nap(100);
	    HiliteControl(cid, 0);
	    if (overwrite) {	/* Approved overwrite, time to quit. */
		char buf[1024];
		i = TEGetText(buf, fullpath);
		buf[i] = 0;
		spitout(buf);
		ReleaseEvent();
	    } else {
		i = 1;
		typed = 0;
		if (teactive) {
		    typed = 1;
		    if (mode != dosave) {
			teactive = 0;
			TEDeactivate(fullpath);
		    }
		    i = readtepath();
		    if (mode == dosave && i) {
			ReleaseEvent();
			break;
		    }
		}
		if (i && finishup(typed))
		    ReleaseEvent();
	    }
	    break;

	  case K_BACKSP:
	  case K_BCKSP2:
	    if (teactive)
		TEKey(ep->message, fullpath);
	    else
		Bell();
	    break;

	  case K_TAB:
	    if (teactive = !teactive) {
		TESetSelect(0, 1024, fullpath);
		removefilebar();
		TEActivate(fullpath);
	    } else {
		TEDeactivate(fullpath);
		putupfilebar();
	    }
	    break;

	  case K_UP:
	    if (ep->message & F_ALT) {
		ypos -= 18;
		MoveWindow(ep->who, xpos, ypos, 0);
	    } else if (teactive) {
		TEKey(ep->message, fullpath);
	    } else if (curdir->curbar > curdir->firstvis) {
		selectfile(curdir->curbar - 1);
	    } else
		Bell();
	    break;

	  case K_DOWN:
	    if (ep->message & F_ALT) {
		ypos += 18;
		MoveWindow(ep->who, xpos, ypos, 0);
	    } else if (teactive) {
		TEKey(ep->message, fullpath);
	    } else if (curdir->curbar < curdir->lastvis) {
		selectfile(curdir->curbar + 1);
	    } else
		Bell();
	    break;

	  case K_LEFT:
	    if (ep->message & F_ALT) {
		xpos -= 18;
		MoveWindow(ep->who, xpos, ypos, 0);
	    } else if (teactive) {
		TEKey(ep->message, fullpath);
	    } else if (curdir->curview > curdir->firstvis) {
		scrollfileto(curdir->curview - 1);
	    } else
		Bell();
	    break;

	  case K_RIGHT:
	    if (ep->message & F_ALT) {
		xpos += 18;
		MoveWindow(ep->who, xpos, ypos, 0);
	    } else if (teactive) {
		TEKey(ep->message, fullpath);
	    } else if (curdir->curview < curdir->nitems) {
		scrollfileto(curdir->curview + 1);
	    } else
		Bell();
	    break;

	  case K_PGUP:		/* Scroll file list  up. */
	    if (teactive) {
		TEKey(ep->message, fullpath);
	    } else if (curdir->displayable > curdir->viewitems)
		fileaction(0, inPageUp);
	    else
		Bell();
	    break;

	  case K_PGDN:		/* Scroll file list down. */
	    if (teactive) {
		TEKey(ep->message, fullpath);
	    } else if (curdir->displayable > curdir->viewitems)
		fileaction(0, inPageDown);
	    else
		Bell();
	    break;

	  case K_HOME:		/* Scroll file list to beginning. */
	    if (teactive) {
		TEKey(ep->message, fullpath);
	    } else if (curdir->displayable > curdir->viewitems)
		scrollfileto(curdir->firstvis);
	    else
		Bell();
	    break;

	  case K_END:		/* Scroll file list to end. */
	    if (teactive) {
		TEKey(ep->message, fullpath);
	    } else if (curdir->displayable > curdir->viewitems)
		scrollfileto(curdir->nitems);
	    else
		Bell();
	    break;

	  case K_DELETE:	/* Move up one in the heirarchy. */
	    if (teactive)
		TEKey(ep->message, fullpath);
	    else
		newpath(curpath.curbar - 1);
	    break;

	  case K_INSERT:	/* Debug mode toggle. */
	    if (ep->message & F_CTRL)
		FlushDebug(debugmode = !debugmode);
	    break;

	  default:
	    if (teactive) {
		TEKey(ep->message, fullpath);
	    }
 	    break;
	}
    } else if (!(permerror | overwrite)) { /* Normal keys, normal mode? */
	if (teactive) {
	    TEKey(ep->message, fullpath);
	} else {		/* Do an i-search on the file picklist. */
	    if (ep->when > (lasttime + dblclktime)) {
		searchstring.d_ino = 0; /* Too late, */
		memset(searchstring.d_name, 0, searchstring.d_size);
	    }			/* clear out any search string. */
	    i = searchstring.d_ino;
	    if (i > searchstring.d_size) {
		searchstring.d_name = realloc(searchstring.d_name,
					      searchstring.d_size += 16);
	    }
	    searchstring.d_name[i++] = key;
	    searchstring.d_ino = i;
	    i = doisearch();	/* Find the current search element. */
	    if (i >= 0)
		selectfile(i);
	    lasttime = ep->when;
	}
    } else {			/* No typing keys when error alert is up. */
	Bell();
    }
    SetWordBreak(wordbrk, fullpath);
}

void
activatevent(ep)
EventRecord *ep;
{
    short i;

    if (ep->message & 1) {	/* Activating? */
	mywindowisinfront = 1;
	HiliteControl(filescroll,
		      (curdir->displayable < curdir->viewitems
		       || permerror || overwrite) ? 255 : 0);
	HiliteControl(dirscroll,
		      (curpath.displayable < curpath.viewitems 
		       || permerror || overwrite) ? 255 : 0);
	HiliteControl(cancelpush, 0);
	HiliteControl(okpush, 0);
	HiliteControl(okpush2, 0);
	HiliteControl(cancelpush3, 0);
	HiliteControl(okpush3, 0);
	if (teactive)
	    TEActivate(fullpath);
	if (doimportscrap)
	    fetchscrap();
    } else {			/* Deactivating. */
	mywindowisinfront = 0;
	HiliteControl(filescroll, 255); /* Mark scroll bar inactive. */
	HiliteControl(dirscroll, 255); /* Mark scroll bar inactive. */
	HiliteControl(cancelpush, 255);
	HiliteControl(okpush, 255);
	HiliteControl(okpush2, 255);
	HiliteControl(cancelpush3, 255);
	HiliteControl(okpush3, 255);
	TEDeactivate(fullpath);
	if (doexportscrap)
	    writescrap();
    }
}

/*
** This routine is called repeatedly if the operator presses in the controls
** of scroll bars.  It is optional (installed via SetCtlAction).  Without it
** the scroll bars only accept a single click in any part of them.  Generally
** users will expect scrolling to be continuous, so you want one of these.
**
** This routine is _not_ called when dragging the thumb.
*/

void
fileaction(theControl, partCode)
ControlID theControl;
short partCode;
{
    short i, j;

    j = curdir->nitems - curdir->viewitems;
    if (partCode == inUpButton) {
	if (curdir->curview > curdir->firstvis)
	    scrollfileto(curdir->curview - 1);
    } else if (partCode == inDownButton) {
	if (curdir->curview < j)
	    scrollfileto(curdir->curview + 1);
    } else if (partCode == inPageUp) {
        i = (curdir->curview < curdir->viewitems + curdir->firstvis)
	    ? curdir->curview - curdir->firstvis : curdir->viewitems;
        scrollfileto(curdir->curview - i);
    } else if (partCode == inPageDown) {
        i = (curdir->curview < (j - curdir->viewitems))
	    ? curdir->viewitems : (j - curdir->curview);
        scrollfileto(curdir->curview + i);
    }
}

void
diraction(theControl, partCode)
ControlID theControl;
short partCode;
{
    short i, j;

    j = curpath.nitems - curpath.viewitems;
    if (partCode == inUpButton) {
	if (curpath.curview > curpath.firstvis)
	    scrolldirto(curpath.curview - 1);
    } else if (partCode == inDownButton) {
	if (curpath.curview < j)
	    scrolldirto(curpath.curview + 1);
    } else if (partCode == inPageUp) {
        i = (curpath.curview < curpath.viewitems + curpath.firstvis)
	    ? curpath.curview - curpath.firstvis : curpath.viewitems;
        scrolldirto(curpath.curview - i);
    } else if (partCode == inPageDown) {
        i = (curpath.curview < (j - curpath.viewitems))
	    ? curpath.viewitems : (j - curpath.curview);
        scrolldirto(curpath.curview + i);
    }
}

/*
** Scroll file list to item #.
*/

scrollfileto(item)
int item;
{
    short i, j, sign;
    struct adirent *dip;
    Rect scratch;

    if (item == curdir->curview || item < curdir->firstvis
	|| item > curdir->nitems)
        return;

    i = curdir->viewitems - 1;	/* Find the last item that can start the */
    j = curdir->lastvis;	/* list so it fills the display completely. */
    dip = &curdir->adir[j];
    while (j > curdir->firstvis && i) {
	if (curdir->unfinished && dip->d_ino == UNKNOWN)
	    finishoff(curdir, dip);
	if (!(dip->d_ino & (EMPTY | NAMEHIDE | PERMHIDE))) {
	    i--;
	}
	dip--;
	j--;
    }
    if (item > j)		/* Don't start displaying too far into list. */
	item = j;

    sign = (curdir->curview - item) < 0 ? 1 : -1; 
    i = 0;
    j = curdir->curview;
    dip = &curdir->adir[j];
    while (j != item) {		/* Find # of visible positions to move. */
	if (curdir->unfinished && dip->d_ino == UNKNOWN)
	    finishoff(curdir, dip);
	if (!(dip->d_ino & (EMPTY | NAMEHIDE | PERMHIDE))) {
	    i -= sign;
	}
	j += sign;
	dip += sign;
    }

    if (curdir->curview != item) {
	curdir->curview = item;
	removefilebar();
	scratch = Dialog[DIALOG_files].rect;
	InsetRect(&scratch, 1, 1);
	scratch.edge.right -= 16;
	ScrollRect(&scratch, 0, finfo.leading * i, 0);
	showfiles();
	putupfilebar();
	setfilebar();
    }
}

/*
** Scroll directory list to item #.
*/

scrolldirto(item)
int item;
{
    short i;
    Rect scratch;

    if (item == curpath.curview || item < 0
	|| item > curpath.nitems - curpath.viewitems)
        return;
    i = curpath.curview - item;
    curpath.curview = item;
    removedirbar();
    scratch = Dialog[DIALOG_directories].rect;
    InsetRect(&scratch, 1, 1);
    scratch.edge.right -= 16;
    ScrollRect(&scratch, 0, finfo.leading * i, 0);
    showdirs();
    putupdirbar();
}

/*
** Re-draw file list scrollbar.
*/

setfilebar()
{
    static barstate;		/* So we don't keep doing hilite of 0 which */
				/* messes up the arrows when mouse scrolling */
    if (curdir) {
	if (curdir->displayable <= curdir->viewitems) {
	    if (!barstate) {
		HiliteControl(filescroll, 255);
		barstate = 1;
	    }
	} else {
	    if (barstate) {
		HiliteControl(filescroll, 0);
		barstate = 0;
	    }
	    Set4CtlVals(filescroll, 0, curdir->displayable,
			ScaleRS(curdir->curview - curdir->firstvis,
				curdir->displayable,
				(curdir->displayable - curdir->viewitems)),
			curdir->viewitems);
	}
    }
}

/*
** Re-draw directory list scrollbar.
*/

setdirbar()
{
    static barstate;		/* So we don't keep doing hilite of 0 which */
				/* messes up the arrows when mouse scrolling */
    if (curpath.displayable <= curpath.viewitems) {
	if (!barstate) {
	    HiliteControl(dirscroll, 255);
	    barstate = 1;
	}
    } else {
	if (barstate) {
	    HiliteControl(dirscroll, 0);
	    barstate = 0;
	}
	Set4CtlVals(dirscroll, 0, curpath.displayable,
		    ScaleRS(curpath.curview - curpath.firstvis,
			    curpath.displayable,
			    (curpath.displayable - curpath.viewitems)),
		    curpath.viewitems);
    }
}

/*
** Prepare any storage needed by the dialog(s).
*/

initdialog()
{
    DMItem *itemp;
    int nitems, item;
    Rect scratch;
    ControlID cid;

    nitems = sizeof(Dialog)/sizeof(DMItem);
    for (item=0; item<nitems; item++) {
	itemp = &Dialog[item];
	switch (itemp->type) {
	  case DMPUSHBUT:
	    cancelpush = NewControl(window, &itemp->rect, itemp->text,
				    0, 0, 0, 1, pushButProc, 0);
	    HiliteControl(cancelpush, 255);
	    ShowControl(cancelpush);
	    break;

	  case DMDPUSHBUT:
	    okpush = NewControl(window, &itemp->rect,
				(mode == dosave) ? "Save" : itemp->text,
				0, 0, 0, 1, defButProc, 0);
	    HiliteControl(okpush, 255);
	    ShowControl(okpush);
	    break;

	  case DMTEXTEDIT:
	    scratch = itemp->rect;
	    InsetRect(&scratch, 2, 2);
	    fullpath = TENew(&scratch, &scratch);
	    TESetText(0, 1024, fullpath); /* Allow a 1K path. */
	    SetWordBreak(wordbrk, fullpath);
	    if (mode == dosave) {
		TEActivate(fullpath);
		teactive = 1;
	    }
	    break;

	  case DMPICKLIST:
	    scratch = itemp->rect;
	    scratch.edge.left = scratch.edge.right - 17;
	    cid = NewControl(window, &scratch, itemp->text,
			     0, 0, 0, 10, scrollBarProc, 0);
	    HiliteControl(cid, 255);
	    ShowControl(cid);
	    SetCtlScrollRate(cid, 5); /* Pace them. */
	    if (item == DIALOG_files) {
		filescroll = cid;
		SetCtlAction(cid, fileaction);
	    } else if (item == DIALOG_directories) {
		dirscroll = cid;
		SetCtlAction(cid, diraction);
	    }
	    break;
	}
    }
    nitems = sizeof(Dialog2)/sizeof(DMItem);
    for (item=0; item<nitems; item++) {
	itemp = &Dialog2[item];
	switch (itemp->type) {
	  case DMDPUSHBUT:
	    okpush2 = NewControl(window, &itemp->rect, itemp->text,
				0, 0, 0, 1, defButProc, 0);
	    break;
	}
    }
    nitems = sizeof(Dialog3)/sizeof(DMItem);
    for (item=0; item<nitems; item++) {
	itemp = &Dialog3[item];
	switch (itemp->type) {
	  case DMPUSHBUT:
	    cancelpush3 = NewControl(window, &itemp->rect, itemp->text,
				    0, 0, 0, 1, pushButProc, 0);
	    break;

	  case DMDPUSHBUT:
	    okpush3 = NewControl(window, &itemp->rect, itemp->text,
				0, 0, 0, 1, defButProc, 0);
	    break;
	}
    }
}

/*
** What they say!
*/

drawdialogs()
{
    drawdialog();
    if (!(permerror | overwrite))
	putupdirbar();
    drawdialog2();
    drawdialog3();
}

drawdialog()
{
    DMItem *itemp;
    int nitems, item;
    Rect scratch;
    static int firsttime = 1;

    if (!firsttime) {
	if (!(permerror | overwrite)) {
	    if (!IsCtlActive(cancelpush))
		ShowControl(cancelpush);
	    else
		DrawControl(cancelpush);
	    if (!IsCtlActive(okpush))
		ShowControl(okpush);
	    else
		DrawControl(okpush);
	}
	DrawControl(filescroll);
	DrawControl(dirscroll);
    }
    firsttime = 0;
    nitems = sizeof(Dialog)/sizeof(DMItem);
    for (item=0; item<nitems; item++) {
	itemp = &Dialog[item];
	switch (itemp->type) {
	  case DMTEXT:
	    TextFont(itemp->font);
	    TextSize(itemp->fontsize);
	    DrawJustText(itemp->text, StringLen(itemp->text), &itemp->rect,
			 centerJustify, centerJustify);
	    break;

	  case DMTEXTEDIT:
	    TEUpdate(&itemp->rect, fullpath);
	    FrameRect(&itemp->rect);
	    break;

	  case DMBOX:
	    if (item != DIALOG_Frame)
		FrameRect(&itemp->rect);
	    break;

	  case DMPICKLIST:
	    FrameRect(&itemp->rect);
	    if (item == DIALOG_files) {
		showfiles();
		if (!(permerror | overwrite))
		    putupfilebar();
		setfilebar();
	    }
	    if (item == DIALOG_directories) {
		showdirs();
	    }
	    break;
	}
    }
}

drawdialog2()
{
    DMItem *itemp;
    int nitems, item;
    Rect scratch;

    if (permerror) {
	HideControlNoInval(cancelpush);
	HideControlNoInval(okpush);
	HiliteControl(dirscroll, 255);
	HiliteControl(filescroll, 255);
	nitems = sizeof(Dialog2)/sizeof(DMItem);
	for (item=0; item<nitems; item++) {
	    itemp = &Dialog2[item];
	    switch (itemp->type) {
	      case DMDPUSHBUT:
		if (!IsCtlActive(okpush2))
		    ShowControl(okpush2);
		else
		    DrawControl(okpush2);
		break;

	      case DMTEXT:
		TextFont(itemp->font);
		TextSize(itemp->fontsize);
		DrawJustText(itemp->text, StringLen(itemp->text), &itemp->rect,
			     centerJustify, centerJustify);
		break;

	      case DMBOX:
		if (item != DIALOG2_Frame) {
		    scratch = itemp->rect;
		    FrameRect(&scratch);
		    if (item == DIALOG2_Frame2) {
			InsetRect(&scratch, 1, 1);
			EraseRect(&scratch);
		    }
		    if (itemp->attr & BDOUBLE) {
			InsetRect(&scratch, 1, 1);
			PenSize(2, 2);
			FrameRect(&scratch);
			PenNormal();
		    }
		}
		break;
	    }
	}
    }
}

drawdialog3()
{
    DMItem *itemp;
    int nitems, item;
    Rect scratch;

    if (overwrite) {
	HideControlNoInval(cancelpush);
	HideControlNoInval(okpush);
	HiliteControl(dirscroll, 255);
	HiliteControl(filescroll, 255);
	nitems = sizeof(Dialog3)/sizeof(DMItem);
	for (item=0; item<nitems; item++) {
	    itemp = &Dialog3[item];
	    switch (itemp->type) {
	      case DMPUSHBUT:
		if (!IsCtlActive(cancelpush3))
		    ShowControl(cancelpush3);
		else
		    DrawControl(cancelpush3);
		break;

	      case DMDPUSHBUT:
		if (!IsCtlActive(okpush3))
		    ShowControl(okpush3);
		else
		    DrawControl(okpush3);
		break;

	      case DMTEXT:
		TextFont(itemp->font);
		TextSize(itemp->fontsize);
		DrawJustText(itemp->text, StringLen(itemp->text), &itemp->rect,
			     centerJustify, centerJustify);
		break;

	      case DMBOX:
		if (item != DIALOG3_Frame) {
		    scratch = itemp->rect;
		    FrameRect(&scratch);
		    if (item == DIALOG3_Frame2) {
			InsetRect(&scratch, 1, 1);
			EraseRect(&scratch);
		    }
		    if (itemp->attr & BDOUBLE) {
			InsetRect(&scratch, 1, 1);
			PenSize(2, 2);
			FrameRect(&scratch);
			PenNormal();
		    }
		}
		break;
	    }
	}
    }
}
