/*
** Definitions for the Standard File Picker program.
*/

#include <sys/types.h>
#include <sys/stat.h>

char *getcwd(char *buf, int size);
char *twiddle(char *string, int uid);

struct adirent {		/* A directory entry. */
    short d_ino;		/* Flags, inode, or whatever. */
    short d_size;		/* Size of name area. */
    char *d_name;
};

struct dircache {
    struct dircache *next;	/* Pointer to next cached directory. */
    struct adirent *adir;	/* Pointer to in-memory directory array. */
    struct stat dstat;		/* Status of this dir, for cache invals. */
    char *name;			/* Full pathname of this directory. */
    short namelen;		/* String length of pathname. */
    short nitems;		/* Number of items (dir. entries) in list. */
    short displayable;		/* Number of displayable items in list. */
    short selectable;		/* Number of selectable items in list. */
    short unfinished;		/* Number of un-statted items left. */
    short curbar;		/* Current highlit (selected) item (index). */
    short curview;		/* Current starting view item (index). */
    short curviewend;		/* Current ending view item (index). */
    short firstvis;		/* First visible item (index). */
    short lastvis;		/* Last visible item (index). */
    short viewitems;		/* # of items viewable in associated picker. */
} *dirchain, *curdir, curpath;

struct adirent searchstring;	/* For keyboard-entered search. */
struct adirent basename;	/* For Save-As base name. */
int uid, gid, mode, options;
char *initialpath;
char *shmatchstr;
struct regexp *regmatch;

short fileviewitems, dirviewitems;

#define UNKNOWN  0x00		/* Flag bits for directory entry types. */
#define EMPTY	 0x01
#define FILE	 0x02
#define DIR 	 0x04
#define NAMEHIDE 0x08
#define PERMHIDE 0x10
#define NAMEGRAY 0x20
#define PERMGRAY 0x40

#define SF_DOTFILES 1		/* Show files whose names start with a '.' */
#define SF_GRAYBAD 2		/* Gray out unsuitable files/directories. */
#define SF_HIDEBAD 4		/* Hide unsuitable files/directories. */

enum {				/* Modes */
    doopen,			/* Open file. */
    dosave,			/* Save file. */
    dodiropen,			/* Select dir to read (all?) from. */
    dodirsave			/* Select dir to save in. */
};
