/*
**  Do shell-style pattern matching for ?, \, [], and * characters.
**  Might not be robust in face of malformed patterns; e.g., "foo[a-"
**  could cause a segmentation violation.  It is 8bit clean.
**
**  Written by Rich $alz, mirror!rs, Wed Nov 26 19:03:17 EST 1986.
**  Rich $alz is now <rsalz@bbn.com>.
**  Special thanks to Lars Mathiesen <thorinn@diku.dk> for the ABORT code.
**  This can greatly speed up failing wildcard patterns.  For example:
**	pattern: -*-*-*-*-*-*-12-*-*-*-m-*-*-*
**	text 1:	 -adobe-courier-bold-o-normal--12-120-75-75-m-70-iso8859-1
**	text 2:	 -adobe-courier-bold-o-normal--12-120-75-75-X-70-iso8859-1
**  Text 1 matches with 51 calls, while text 2 fails with 54 calls.  Without
**  the ABORT, then it takes 22310 calls to fail.  Ugh.
*/

#define TRUE 1
#define FALSE 0
#define ABORT (-1)

#define NEGATE_CLASS	'^'

/* Forward declaration. */
static int domatch();

/*
**  See if the text matches the p, which has an implied leading asterisk.
*/

static int
star(text, p)
register char *text;
register char *p;
{
    register int ret;

    do
	ret = domatch(text++, p);
    while (ret == FALSE);
    return ret;
}


/*
**  Match text and p, return TRUE, FALSE, or ABORT.
*/

static int
domatch(text, p)
register char *text;
register char *p;
{
    register int last;
    register int matched;
    register int reverse;

    for ( ; *p; text++, p++) {
	if (*text == '\0' && *p != '*')
	    return ABORT;
	switch (*p) {
	  case '\\':		/* Literal match with following character. */
	    p++;		/* and FALLTHROUGH */
	  default:
	    if (*text != *p)
		return FALSE;
	    continue;
	  case '?':		/* Match anything. */
	    continue;
	  case '*':		/* Trailing star matches everything. */
	    return *++p ? star(text, p) : TRUE;
	  case '[':
	    if (reverse = p[1] == NEGATE_CLASS)
		/* Inverted character class. */
		p++;
	    for (last = 0400, matched = FALSE; *++p && *p != ']'; last = *p)
		/* This next line requires a good C compiler. */
		if (*p == '-' ? *text <= *++p && *text >= last : *text == *p)
		    matched = TRUE;
	    if (matched == reverse)
		return FALSE;
	    continue;
	}
    }
    return *text == '\0';
}


/*
**  User-level routine.  Returns TRUE or FALSE.
*/

int
shellmatch(text, p)
char *text;
char *p;
{
    return domatch(text, p) == TRUE;
}
