/*
** Look for (and perform) "~", "~/..." and "~name/..." substitution,
** and $ENV environment variable substitution.
**
** Returns a pointer to a string, which may be the original string
** or a malloc'd buffer with the expanded string in it.
*/

#include <pwd.h>
#include <string.h>
#include <malloc.h>

char *
twiddle(string, uid)
char *string;
int uid;
{
    char *cp, *scopy, *subbed, *alloca(), *getenv();
    struct passwd *pwd, *getpwuid(), *getpwnam();

    if (string[0] == '~') {
	scopy = alloca(strlen(string) + 1);
	strcpy(scopy, string);
	if (cp = strchr(scopy, '/')) /* Strip any trailing garbage. */
	    *cp = 0;
	if (!scopy[1]) {	/* Do substitution. */
	    pwd = getpwuid(uid); /* They want their own home dir. */
	} else {		/* Someone else's home dir. */
	    if (!(pwd = getpwnam(&scopy[1]))) { /* Nope, they */
		Bell();
		pwd = getpwuid(uid); /* really want their own home dir! */
	    }
	}
	if (!pwd) {
	    return 0;
	}
	subbed = malloc(strlen(pwd->pw_dir) + 1 + (cp ? strlen(&cp[1])+1 : 0));
	strcpy(subbed, pwd->pw_dir);
	if (cp) {		/* Add trailing stuff if any existed. */
	    *cp = '/';
	    strcat(subbed, cp);
	}
	string = subbed;
    } else if (string[0] == '$') { /* Handle environment variable subs. */
	cp = string;
	if (!(string = getenv(&string[1]))) {
	    cp[0] = '~';	/* Failure, return home directory. */
	    cp[1] = 0;
	    Bell();
	    return twiddle(cp, uid);
	}
    }
    return string;
}


/*
** Canonicalize a path name.  Removes adjacent slashes and /./ strings,
** and strip out path elements eaten by ../ sequences.
*/

canon(s)
char *s;
{
    char *p1, *p2;
    int hold;

    p1 = p2 = s;
    while (p1 && *p1 && (p1 = strchr(p1, '/'))) {
	hold = 0;
	while (p1[1] == '/') {	/* Lop out adjacent slashes. */
	    strcpy(p1, &p1[1]);
	    hold = 1;
	}
	while (!strncmp(p1, "/./", 3) || !strcmp(p1, "/.")) {
	    strcpy(p1, &p1[2]);	/* Lop out './' path elements. */
	    hold = 1;
	}
	while (!strncmp(p1, "/../", 4) || !strcmp(p1, "/..")) {
	    strcpy(p2, &p1[3]);	/* Back up one in directory path. */
	    p1 = s;		/* Start over, so can handle multiple */
	    hold = 1;		/* instances of '../' patheaters. */
	}
	p2 = p1;
	if (!hold)		/* Don't move on 'till nothing's changed. */
	    p1++;
    }
}
