/***********************************************************************
*
*
*                Second layer of viewer, displays files
*
*  ----------------------------------------------------------------
*
*                     ISC Systems Corporation
*                   Confidential / Proprietary
*      Copyright (C) 1990 by ISC-Bunker Ramo / Unpublished Work
*       For Use by ISC-Bunker Ramo and Authorized Users Only
*                      All rights reserved.
*
*  Author:  Carl Paukstis       September, 1990
*
*  Notes:
*
**********************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <isc/keycodes.h>
#include <isc/window.h>
#include <isc/controls.h>
#include <dlgman.h>
#include "moview.h"

#define CHUNKSIZE 2048

static FView *FVroot;
static int FVfill(ViewerWindow *VW);

FView *
FVOpen( short rows,
			 short columns,
			 short x,
			 short y,
			 short font,
	         short fsize,
             short tabstop,
			 char *fname,
			 int grafFd )
{
	FView *FV;
	unsigned char *buf;
	int bufsize;
	int size;
	int inputFd;

	if ((inputFd = open(fname, O_RDONLY, 0)) < 0)
		return NULL;

	bufsize = CHUNKSIZE;
	if ((buf = (unsigned char *)malloc(bufsize)) == NULL)
		return NULL;
	size = read(buf,bufsize,inputFd);

	if ((size = read(inputFd,buf,bufsize)) < 0)
	{
		close(inputFd);
		free(buf);
		return NULL;
	}
	if ((FV = (FView *)malloc(sizeof(*FV))) == NULL)
	{
		close(inputFd);
		free(buf);
		return NULL;
	}
	FV->inputFd = inputFd;
	FV->chunksize = CHUNKSIZE;
	FV->FVMagic = FVMAGIC;			/* make valid struct */
	FV->buflen = CHUNKSIZE;
	FV->inputEOF = FALSE;

	if (size < bufsize)
	{
		bufsize = size;
		FV->inputEOF = TRUE;
	}
	if ((FV->VW=VWOpen(rows,columns,x,y,font,fsize,tabstop,fname,buf,bufsize,grafFd,FVfill)) == NULL)
	{
		free(FV);
		close(inputFd);
		return NULL;
	}
	FV->next = FVroot;		/* link new struct at head of list */
	FVroot = FV;
	return FV;
}

static int 
FVfill(ViewerWindow *VW)
{
	FView *FV;
	unsigned char *nextbyte;
	int size;

	if (VW->inputEOF)
	{
		return 0;
	}
	for (FV = FVroot; FV; FV = FV->next)	/* look for our FV struct by */
		if (FV->VW == VW)					/* checking the associated VW */
			break;

	if (!FV)
		return 0;							/* somebody's hosed-up */

	if (VW->textBuffCount + CHUNKSIZE > FV->buflen)
	{
		unsigned char *save = VW->textBuff;
		FV->buflen += FV->chunksize;			/* grow the text buffer */
		VW->textBuff = (unsigned char *)realloc(VW->textBuff, FV->buflen + 1);
		if (! VW->textBuff)
		{
			FV->buflen -= FV->chunksize;
			VW->textBuff = save;
			return FALSE;
		}
		FV->chunksize <<= 1;				/* logarithmic increase buf size */
	}
	nextbyte = VW->textBuff + VW->textBuffCount;
	*nextbyte = '\n';		/* make sure there's one more char for caller */

	size = read(FV->inputFd, nextbyte, CHUNKSIZE);
	if (size != CHUNKSIZE)
	{
		*(VW->textBuff + VW->textBuffCount + size) = '\n';	/* ensure newline */
		if (size <= 0)
		{
			FV->inputEOF = TRUE;
			close(FV->inputFd);
			size = 0;
		}
	}
	VW->textBuffCount += size;
	return size;
}


/*
**  "Close" a file-viewer window - kill the window, de-allocate the
**  structure and associated data, forget about it.
*/
BOOL 
FVClose (FView *FV)
{
	ViewerWindow *VW = FV->VW;
	FView *search, *prev;

	if (!FV || FV->FVMagic != FVMAGIC)
		return FALSE;
	if (VW->textBuff) free(VW->textBuff);
	if (FV->inputFd > 0 && ! FV->inputEOF) close(FV->inputFd);
	(void)VWClose(VW);

	prev = (FView *)&FVroot;
	for (search = FVroot; search; search = search->next)
	{
		if (search == FV)
			break;
		prev = search;
	}
	if (search)
	{
		prev->next = search->next;		/* unlink struct */
	}
	FV->FVMagic = 0;					/* flag invalid struct */
	free(FV);
	return TRUE;
}
