/***********************************************************************
*
*      Text file viewer program using its own windowman window.
*
*  ----------------------------------------------------------------
*
*                     ISC Systems Corporation
*                   Confidential / Proprietary
*      Copyright (C) 1990 by ISC-Bunker Ramo / Unpublished Work
*       For Use by ISC-Bunker Ramo and Authorized Users Only
*                      All rights reserved.
*
*  Author:  Carl Paukstis       June, 1990
*
*  Notes:
*
*
**********************************************************************/

#include <stdio.h>
#ifdef lint
#include <unistd.h>
#endif
#include <fcntl.h>
#include <errno.h>
#include <malloc.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <isc/keycodes.h>
#include <isc/window.h>
#include <isc/controls.h>
#include <isc/dlgman.h>
#include <isc/textedit.h>
#include <dnix/fcodes.h>
#include <moDshow.h>
#include <moview.h>
#include <mousecur.h>

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#ifndef MAXPATH
#define MAXPATH 64
#endif

#define FILEMENU	1			/* The "file" menu */
#define	OPENITEM	1			/* "Open..." */
#define POSITEM		2			/* "Show Position..." */
#define SAVEITEM	3			/* "Write..." */
#define	COPYITEM	5			/* "Copy" */
#define	QUITITEM	7			/* "Quit" */

#define	CHUNKSIZE	4096	/* bytes at a time to read from input */
#define MAXCHUNK	262144L	/* Maximum realloc increment. */

typedef struct {
	void (*func)(void *, int, int);
	int waitthing;
} NWCOMP;

/*
**  ugly global variables 
*/

BOOL			fromPipe = FALSE;/* are we displaying our stdin? */
BOOL			attributes = TRUE;/* 8th bit indicates attribute char? */
int				grafFd;			/* FD attached to windowman */
int				tqFd;			/* trap queue FD */
int				inputFd = -1;	/* our (current) input file */
char *			Myname;			/* name of this program */
char *			usertitle = 0;	/* user-supplied title */
ViewerWindow *	VW;				/* our one viewer window */
char			wtitle[80];		/* window title buffer */
static char		Pathnamebuf[256];
int				FPPid;			/* PID of file-picker process */
SFDialog 		OpenSFD;
static NWCOMP	nwcomp;			/* for nowait-wait completion junk */

/*
**  Macros to deal with the pathname ring and position info
*/
#define PATHSIZE 16
#define NEXTPATH() (paths[((curpath==pathcount-1)?curpath=0:curpath+=1)])
#define PREVPATH() (paths[((curpath<=0)?curpath=pathcount-1:curpath-=1)])
#define CURRPATH() (paths[curpath])
#define ADDPATH(name) if(pathcount>PATHSIZE-1)\
	{ pathcount=PATHSIZE; strcpy(NEXTPATH(),name);}\
	else {strcpy(paths[pathcount],name);curpath=pathcount++;}
#define SAVEPOS(where)  (oldpos[curpath]=(where))
#define RESTOREPOS() (oldpos[curpath])
static char paths[PATHSIZE][MAXPATH];
static int oldpos[PATHSIZE] = 0;
static int pathcount = 0;
static int curpath = -1;

/*
** Your friendly local functions...
*/
static int		getNewFile(void);
static void		getComp(void * comp, int rpar, int rn);
static char *	jimsitoa (val, radix, leadfill, ndigits);
static void		keyboardevent(EventRecord *ep);
static void		menuevent();
void			moactivate(EventRecord *ep);
void			scrapevent(EventRecord *ep);
static int		mofill(ViewerWindow *VW);
static void		putComp(void * comp, int rpar, int rn);
static void		showPosition(void);
static void		swapFiles(char pathname[]);
static int		saveFile(void);

static void		usage(void);

int
main(int argc, char * argv[])
{
    extern void PrintRelInfo(void);
    extern int optind;
    extern char * optarg, *getenv();
    char *title;
    char *searcharg = 0;
    char *argp;
    int flag;
    short font = 5;
    short size = 9;
    char *path = getenv("WINDOWS_ON");
    int forkme = TRUE;
    int restricted;
    short xpos = 32, ypos = 40, 
    windowRows = 40, windowColumns = 80, tabstop = 8;

	/* setup global Myname pointing to basename of what we were called */
    Myname = strrchr(argv[0],'/');
    if (Myname == NULL)
	Myname = argv[0];
    else
	Myname++;

    restricted = 0;
				/* collect the arguments */
    while ((flag = getopt(argc, argv, "AFf:c:p:r:RT:t:Vx:y:S:s:")) != EOF) {
	switch (flag) {
	  case 'A':
	    attributes = FALSE;
	    break;

	  case 'c':
	    windowColumns = atoi(optarg);
	    break;

	  case 'F':
	    forkme = FALSE;
	    break;

	  case 'f':
	    font = atoi(optarg);
	    break;

	  case 'p':
	    path = optarg;
	    break;

	  case 'r':
	    windowRows = atoi(optarg);
	    break;

	  case 'R':
	    restricted = 1;
	    break;

	  case 'S':
	    searcharg = optarg;
	    break;

	  case 's':
	    size = atoi(optarg);
	    break;

	  case 'T':
	    usertitle = optarg;
	    break;

	  case 't':
	    tabstop = atoi(optarg);
	    break;

	  case 'V':
	    PrintRelInfo();
	    return 0;

	  case 'x':
	    xpos = atoi(optarg);
	    break;

	  case 'y':
	    ypos = atoi(optarg);
	    break;

	  case 'U':
	  case '?':
	  default:
	    usage();
	    return 2;
	}
    }

    if (argc > optind) {	/* Get the filename (if any). */
	while (argc > optind) {
	    ADDPATH(argv[optind]);
	    ++optind;
	}
	(void)NEXTPATH();
    } else {
	errno = 0;
	if (isatty(0) != 1) {
	    fromPipe = TRUE;
	    inputFd = fileno(stdin);
	}
    }

	
    if (fromPipe) {		/* open some input file */
	if (usertitle)
	    title = "";
	else
	    title = "(stdin)";
    } else {
	if (CURRPATH()[0] == '\0') {
	    usage();
	    return 2;
	}
	if ((inputFd = open(CURRPATH(), O_RDONLY, 0)) < 0) {
	    char *msg = ": cannot open ";
	    write(2,Myname, strlen(Myname));
	    write(2,msg,strlen(msg));
	    write(2,CURRPATH(), strlen(CURRPATH()));
	    write(2,"\n",1);
	    return 1;
	}
	title = CURRPATH();
    }

    if (forkme) {		/* if we should separate from parent */
	int (*oldsig)();

	oldsig = signal(SIGHUP, SIG_IGN);
	if (fork())
	    return 0;
	setpgrp();
	signal(SIGHUP, oldsig);
    }

				/* initialize the graphics toolbox */
    if ((tqFd = InitGraf()) < 0  || (grafFd = ConnectGraf(path)) < 0) {
	char *msg = ": Sorry, cannot connect to your graphics terminal.\n";
	write(2,"\n",1);
	write(2,Myname, strlen(Myname));
	write(2,msg,strlen(msg));
	return 1;
    }
    if (forkme && ! fromPipe) {
	int fd=0;
	while (fd < inputFd)
	    close(fd++);	/* Get rid of nasty standard fd's. */
    }
    Event(updateEvt, VWupdate);
    Event(mouseEvt, VWmouse);
    Event(windowEvt, VWwindow);
    Event(menuEvt, menuevent);
    Event(keyEvt, keyboardevent);
    Event(activateEvt, moactivate);
    Event(scrapEvt, scrapevent);
    Event(noEvent, (void (*)())VWnoevent); /* read file while idle */
    wtitle[0] = '\0';
    if (usertitle)
	strcpy(wtitle, usertitle);
    else
	strcpy(wtitle, title);
    VW = VWOpen(windowRows, windowColumns, xpos, ypos, font, size, tabstop,
		wtitle, malloc(1), 0/*bufsize*/, grafFd, mofill);
    SearchVW = VW;
    NewMenu(FILEMENU, "File");
    AppendMenu(FILEMENU, "/OOpen\311;Show Position\311;/WWrite\311;(-;"
	       "(/CCopy;(-;/QQuit");
				/* now let 'em look at it. */
    if (restricted) {
	DisableItem(FILEMENU, OPENITEM);
	DisableItem(FILEMENU, SAVEITEM);
    }
    if (argp = path) {		/* Handle copying video to other WS. */
	char *cp, buf[80]; int fd, wid; long winds;
	Rect r;

	winds = 0;
	while (argp = strchr(argp, '\n')) {
	    argp++;
	    cp = strchr(argp, '\n');
	    if (!cp)
		cp = argp + strlen(argp);
	    if (cp - argp) {
		strncpy(buf, argp, cp - argp);
		buf[cp - argp] = 0;
		if ((fd = ConnectGraf(buf)) >= 0) {
		    SetRect(&r, VW->xpos, VW->ypos,
			    VW->xpos+VW->width, VW->ypos+VW->height);
		    strcat(strcpy(buf, "Copy of "), wtitle);
		    wid = NewWindow(&r, buf, showWindow,
				    documentProc, frontWindow,
				    hasSendToBack, fd);
		    SetPort(wid);
		    ClipRect(&VW->clipRect);
		    PenMode(srcXor);
		    TextMode(srcOr);
		    TextFont(VW->font);
		    TextFace(VW->face);
		    TextSize(VW->size);
		    winds |= 1 | 1 << wid;
		}
	    }
	}
	if (winds) {
	    SetPorts(winds);
	}
    }

    if (fromPipe)
	if (mofill(VW) == 0) {
	    write(2, "mo: empty pipe\n", 15);
	    VWClose(VW);
	    exit(1);
	}
    if (searcharg) {
	register int point;
	if (ValSearchArg(0, 0, 0, searcharg)
	    && (point=VWSearch(VW, VW->regexbuf, 1)) > 0) {
	    VW->hiLiteLine = point;
	    VW->screenNext = VW->screenFirst = point - (VW->rows >>1);
	}
    }
    while (1) {			/* Stay in event loop forever. */
	(void) GetEvent(anyWindow, noEvent);
	flag = EventError() >> 16;
	if (flag == grafFd)	/* Failure on main display connection. */
	    break;
	DisconnectGraf(EventError() >> 16); /* Must have been a copy node. */
    }
    return 1;
}

/*
**  Display the usage message
*/
static void
usage(void)
{
	char *usagetag = "Usage: ";
	char *msg = " [-V] | [-x xpos] [-y ypos] [-c columns] [-r rows] [-f font] \n"
		"        [-t tabstop] [-T title] [-A] [-F] [-R] [-s size] [-S searcharg]\n"
		"        [-p wmnode] file\n";
	write(2,usagetag,strlen(usagetag));
	write(2,Myname,strlen(Myname));
	write(2,msg,strlen(msg));
}

/*
**  Buffer-filling routine.  Called-back from VWnoevent; reads the 
**  input file into the buffer when the user is idle.
*/
static int 
mofill(ViewerWindow *myVW)
{
	unsigned char *nextbyte;
	unsigned char *bufend;
	long size, count;
	unsigned char readbuf[CHUNKSIZE];
	static long buflen = 0, chunksize = CHUNKSIZE;

	if (myVW->inputEOF)			/* we must've read it all last time */
	{
		return 0;
	}

	if (myVW->textBuffCount + CHUNKSIZE > buflen)
	{
		unsigned char *save = myVW->textBuff;
		buflen += chunksize;			/* grow the text buffer */
		myVW->textBuff = realloc(myVW->textBuff, buflen + 1);
		if (! myVW->textBuff)
		{
			buflen -= chunksize;
			myVW->textBuff = save;
			myVW->inputtruncated = TRUE;
			return FALSE;
		}
		if (chunksize < MAXCHUNK)
		    chunksize <<= 1; /* exponentially increase buf size */
	}
	bufend = myVW->textBuff + myVW->textBuffCount;

	do {
	    size = read(inputFd, readbuf, CHUNKSIZE);
	} while (size < 0 && errno == EINTR);
	if (size != CHUNKSIZE) {
	    if (size <= 0) {
		close(inputFd);
		size = 0;
		if (fromPipe) { /* Ensure pipe-writer child */
		    dnix(F_NOWAIT | F_WAIT, 0, 0, 0, 0, 0, tqFd, 0);
		}	/* is waited for. */
		if (myVW->textBuffCount+size && *(bufend + size - 1) != '\n') {
		    readbuf[size++] = '\n'; /* ensure trailing newline */
		}
	    }
	}
	count = size;
	nextbyte = readbuf;
	while (count--)
	{
		unsigned char *pappend(register unsigned char *, register int);
		bufend = pappend(bufend, *(nextbyte++));
	}
	myVW->textBuffCount = bufend - myVW->textBuff;
	return size;
}
/*
**  Activate event handler
*/
void
moactivate(ep)
EventRecord *ep;
{
    int whichport;

    if (VW && ep->who == VW->mainWindow && ep->message) {
	if (FPPid) {
		/* main window being activated, and picker window exists */
	    SendBehind(VW->mainWindow, secondWindow);
	    return;
	}
    }
    VWactivate(ep);
}

copy_on()
{
    EnableItem(FILEMENU, COPYITEM);
}

copy_off()
{
    DisableItem(FILEMENU, COPYITEM);
}

/*
**  Menu event handler
*/
static void
menuevent(ep)
EventRecord *ep;
{
	int epmsg = ep->message;		/* save it in case we do more events */
	unsigned int menuid = ep->message >>16;
	unsigned int menuitem = ep->message & 0xFFFF;

	switch (menuid)
	{
	case FILEMENU:
		switch (menuitem)
		{
		case OPENITEM:			/* Open */
			BeginMenu(epmsg);
			getNewFile();
			EndMenu(epmsg);
			break;
		case POSITEM:			/* Show Position */
			BeginMenu(epmsg);
			showPosition();
			EndMenu(epmsg);
			break;
		case SAVEITEM:			/* Write */
			BeginMenu(epmsg);
			saveFile();
			EndMenu(epmsg);
			break;
		case COPYITEM:			/* Copy */
			BeginMenu(epmsg);
			copy(VW);
			EndMenu(epmsg);
			break;
		case QUITITEM:			/* Quit */
			bailout();
			break;
		default:				/* Crazy - can't be here */
			exit(0);
			break;
		}
		break;
	default:
		VWmenu(ep);
	}
	return;
}

static void
keyboardevent(EventRecord *ep)
{
    extern short newcopy;
    int key;

    if (ep->who != VW->mainWindow)
	return;
    key = ep->message & ~F_KEYDOWN & ~F_KBDIII;
    switch (key) {
      case F_SHIFT | 'P':	/* Previous file */
      case 'p':
	if (pathcount > 1) {
	    SAVEPOS(VW->screenFirst);
	    swapFiles(PREVPATH());
	}
	break;

      case F_SHIFT | 'N':	/* Next file */
      case 'n':
	if (pathcount > 1) {
	    SAVEPOS(VW->screenFirst);
	    swapFiles(NEXTPATH());
	}
	break;

      case F_SHIFT | 'E':	/* Another file */
	getNewFile();
	break;

      case F_CTRL | F_SHIFT | 'Q': /* quit */
      case F_FUNC | K_RESET:
      case F_FUNC | K_ESC:
      case F_CTRL | 'q':
      case F_SHIFT | 'Q':
      case 'q':
	bailout();
	break;
#if 0
      case F_ALT | 'p': {
	  int i = curpath;
	  fprintf(stderr, "Current = %d\n", curpath);
	  do {
	      fprintf(stderr, "current %d: '%s'", curpath, CURRPATH());
	      fprintf(stderr, " Next: '%s'\n",NEXTPATH()); }
	  while (curpath != i);
	  break;
      }
#endif
      default:
	VWkeyboard(ep);
    }
}

bailout()
{
    if (newcopy) {
	writescrap();
	FlushTwice();
    }
    exit(0);
}

/*
**  Dialog to get new filename & open the new file
*/
static int 
valpath(char *path)
{
    int fd;
    if ((fd = open(path,O_RDONLY|O_NOCTTY,0)) >= 0) {
	close(fd);
	return DMok;
    }

    Bell();
    return DM_FAIL;
}

static int 
getNewFile(void)
{
	nwcomp.func = getComp;
	nwcomp.waitthing = 0;
	
	OpenSFD.options = SF_GETFILE;
	OpenSFD.initialname = CURRPATH();
	OpenSFD.title = "mo: Open File\311";
	FPPid = NWFilePicker(&OpenSFD, tqFd);
	dnix(F_NOWAIT|F_WAIT, 0, &nwcomp.waitthing, sizeof(nwcomp.waitthing),
				0,0, tqFd, &nwcomp);
}

static void
getComp(void * comp, int rpar, int rn)
{
	if (rpar != FPPid) {
		dnix(F_NOWAIT|F_WAIT, 0, &nwcomp.waitthing, sizeof(nwcomp.waitthing),
			 0,0, tqFd, &nwcomp);
		return;
	}

	FPPid = 0;					/* flag no file-picker active */
	if (OpenSFD.returnedname) {
		if (VW)						/* we're really initialized and running */
		{
			SelectWindow(dlgWindow);
			SendBehind(VW->mainWindow, frontWindow);
			SAVEPOS(VW->screenFirst);
			ADDPATH(OpenSFD.returnedname);
			swapFiles(OpenSFD.returnedname);
		}
		free(OpenSFD.returnedname);
		return;
	}
}

static int 
saveFile(void)
{
	nwcomp.func = putComp;
	nwcomp.waitthing = 0;
	
	if (! VW->inputEOF) {
		SetMouseCursor(MC_WATCH);
		while (! VW->inputEOF)
			VWnoevent(VW);
		SetMouseCursor(MC_ARROW);
	}
	OpenSFD.options = SF_PUTFILE;
	OpenSFD.initialname = CURRPATH();
	OpenSFD.title = "mo: Save File\311";
	FPPid = NWFilePicker(&OpenSFD, tqFd);
	dnix(F_NOWAIT|F_WAIT, 0, &nwcomp.waitthing, sizeof(nwcomp.waitthing),
				0,0, tqFd, &nwcomp);
}

static void
putComp(void * comp, int rpar, int rn)
{
	int fd;
	
	if (rpar != FPPid) {
		dnix(F_NOWAIT|F_WAIT, 0, &nwcomp.waitthing, sizeof(nwcomp.waitthing),
			 0,0, tqFd, &nwcomp);
		return;
	}

	FPPid = 0;					/* flag no file-picker active */

	if (!OpenSFD.returnedname)
		return;

	if ((fd = open(OpenSFD.returnedname,O_WRONLY|O_APPEND|O_CREAT,0666)) < 0) {
		exit(1);
	}
	write(fd,VW->textBuff, VW->textBuffCount);
	close(fd);
	free(OpenSFD.returnedname);
	return;
}

static void
swapFiles(char pathname[])
{
	Rect rect;
	int fd = open(pathname, O_RDONLY,0);
	if (fd < 0)
		exit(1);
	if (! VW->inputEOF)
		close(inputFd);
	wtitle[0] = '\0';
	if (usertitle)
		strcpy(wtitle, usertitle);
	else
		strcpy(wtitle, pathname);
	SetWTitle(VW->mainWindow, wtitle);
	SelectWindow(dlgWindow);
	inputFd = fd;
	VW->textBuffCount = VW->lineCount = VW->maxWidth = VW->firstColumn = 0;
	VW->linesDone = 0;
	VW->hiLiteLine = -1;
	VW->inputEOF = FALSE;
	Event(noEvent, (void (*)())VWnoevent);
	EraseRect(&VW->clipRect);
	SetRect(&rect, 0, 0, 30000, 30000);
	InvalRect(&rect);
	VW->screenFirst = RESTOREPOS();		/* back to where we were */
}

/*
**  Dialog to show current position in the file
*/
static void
showPosition(void)
{
    DMDcb *cbptr;
    extern DMItem moDshow[];
    Rect dlgwinrect;
    int curport;

    GetPort(&curport);
    SetRect(&dlgwinrect, SearchVW->xpos + 6, SearchVW->ypos + 6,
	    SearchVW->xpos + 6 + 215, SearchVW->ypos + 6 + 100);
    dlgWindow = NewWindow(&dlgwinrect, "mo: Show Position\311", showWindow, 
			  dBoxProc, frontWindow, noOptions, grafFd);

    cbptr = DMNew(moDshow, dlgWindow, (void *)0);
    dialogUp = cbptr;
    if (cbptr == NULL) {
	char msg[] = "Can't open dialog.\n";
	write(1,msg,strlen(msg));
	exit(1);
    }
    DMSetText(cbptr, MODSHOW_lineof, jimsitoa(VW->screenFirst+1,10,' ',0));
    DMSetText(cbptr, MODSHOW_byteof, jimsitoa(VW->lines[VW->screenFirst]+1,10,' ',0));
    DMSetText(cbptr, MODSHOW_linemax, jimsitoa(VW->lineCount,10,' ',0));
    DMSetText(cbptr, MODSHOW_bytemax, jimsitoa(VW->textBuffCount - 1, 10,' ',0));
    if (VW->inputEOF)
	DMSetText(cbptr, MODSHOW_warning, "Top line is");
    else
	DMSetText(cbptr, MODSHOW_warning, "(still reading)");
    (void)DMCollect(cbptr);

    dialogUp = NULL;
    DMDispose(cbptr);
    DisposeWindow(dlgWindow); 
    SetPort(curport);
}

static char *
jimsitoa (val, radix, leadfill, ndigits)
int val, radix, leadfill, ndigits;
{
    short i;
    char *p, *p2, out;
    static char buf[20];
    unsigned long test, remainder;

    p = buf;
    remainder = val;
    i = ndigits;			/* Make sure there're at least the number */
    while (i--)				/* of leading whatevers that we've asked for */
		*p++ = leadfill;	/* as the field width. */
    p2 = p;
    if (radix == 16) 		/* Hex output? */
	{
		test = 0x10000000L;
	}
    else 
	{
		test = 1000000000L;	/* Decimal output? */
		if (val < 0) 
		{
			remainder = -val;
			*p++ = '-';
		}
    }
    while (test > remainder)	/* Scale down to match number. */
		test /= radix;
    if (remainder)
	{
		while (test) 
		{
			out = remainder / test;
			*p++ = out + ((out >= 10) ? '7' : '0');
			remainder %= test;
			test /= radix;
		}
	}
    else
	{
		*p++ = '0';
	}
    *p = 0;
    if (ndigits) 		/* A field width set? */
	{
		if ((p2 + ndigits) >= p) /* Make sure we never truncate anything. */
			return p - ndigits;
		else
			return p2;
    }
    return buf;			/* No field width, return minimum number. */
}
