/***********************************************************************
*
*                   Standard Mouse Cursor shapes
*
*                     ISC Systems Corporation
*                   Confidential / Proprietary
*              Copyright (C) 1991,1992 by ISC-Bunker Ramo
*       For Use by ISC-Bunker Ramo and Authorized Users Only
*                      All rights reserved.
*
*  Author:       Carl Paukstis    June, 1992
*  Stolen-from:  Cathy Higel      June 1991 (who stole it elsewhere)
*
*  Notes:  Support routines for dialog box to load a file
*
**********************************************************************/

#include <isc/window.h>
#include <mousecur.h>

void
SetMouseCursor(int cursortype)
{
	static Cursor watch_cursor = {				/* watch cursor */
		{ 0x0000, 0x07C0, 0x07C0, 0x07C0, 0x0820, 0x1110, 0x2108, 0x2108,
		  0x2208, 0x2408, 0x1010, 0x0820, 0x07C0, 0x07C0, 0x07C0, 0x0000 },
		{ 0x0FE0, 0x0FE0, 0x0FE0, 0x1FE0, 0x1FF0, 0x3FF8, 0x7FFC, 0x7FFC,
		  0x7FFC, 0x7FFC, 0x3FF8, 0x1FF0, 0x1FE0, 0x0FE0, 0x0FE0, 0x0FE0 },
		{ 8,8 }};
	static Cursor system_cursor = {				/* pointer cursor */
		{ 0x0000, 0x4000, 0x6000, 0x7000, 0x7800, 0x7C00, 0x7E00, 0x7F00,
		  0x7F80, 0x7C00, 0x6C00, 0x4600, 0x0600, 0x0300, 0x0300, 0x0000 },
		{ 0xC000, 0xE000, 0xF000, 0xF800, 0xFC00, 0xFE00, 0xFF00, 0xFF80,
		  0xFFC0, 0xFFE0, 0xFE00, 0xEF00, 0xCF00, 0x8780, 0x0780, 0x0380 },
		{ 1,1 }};
	static Cursor hand_cursor = {				/* hand cursor */
		{ 0x0000, 0x0000, 0x0000, 0x0000, 0x0700, 0x1900, 0x2200, 0x4700,
		  0xc7fe, 0x8c01, 0x97fe, 0xe410, 0x87e0, 0x8420, 0xc7c0, 0xef80 },
		{ 0x0000, 0x0000, 0x0000, 0x0000, 0x0700, 0x1f00, 0x3e00, 0x7f00,
		  0xfffe, 0xffff, 0xfffe, 0xfff0, 0xffe0, 0xffe0, 0xffc0, 0x7f80 },
		{ 9,16 }};
	static Cursor palm_cursor = {				/* flat hand */
		{ 0x0000, 0x0000, 0x0540, 0x0AB0, 0x0AA8, 0x1AA8, 0x2AA8, 0x2808,
		  0x2008, 0x2008, 0x1FF0, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 },
		{ 0x0000, 0x0540, 0x1FF0, 0x1FF8, 0x1FFC, 0x3FFC, 0x7FFC, 0x7FFC,
		  0x7FFC, 0x7FFC, 0x3FF8, 0x1FF0, 0x0000, 0x0000, 0x0000, 0x0000 },
		{ 8,8 }};
	static Cursor text_cursor = {				/* text mode indicator	*/
		{ 0x0000, 0x0630, 0x0140, 0x0080, 0x0080, 0x0080, 0x0080, 0x0080,
		  0x0080, 0x0080, 0x0080, 0x0080, 0x0080, 0x0140, 0x0630, 0x0000 },
		{ 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
		  0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 },
		{ 8,8 }};
	Cursor	*new_cursor;

	switch (cursortype) {
	case MC_TEXT:
		new_cursor = &text_cursor;
		break;
	case MC_WATCH:
		new_cursor = &watch_cursor;
		break;
	case MC_HAND:
		new_cursor = &hand_cursor;
		break;
	case MC_PALM:
		new_cursor = &palm_cursor;
		break;
	case MC_ARROW:
	default:
		new_cursor = &system_cursor;
		break;
	}
	SetCursor(new_cursor);
	FlushWindow();
}
