/***********************************************************************
*
*
*                Viewer module for displaying text
*
*  ----------------------------------------------------------------
*
*                     ISC Systems Corporation
*                   Confidential / Proprietary
*      Copyright (C) 1990 by ISC-Bunker Ramo / Unpublished Work
*       For Use by ISC-Bunker Ramo and Authorized Users Only
*                      All rights reserved.
*
*  Author:  Carl Paukstis       August, 1990
*
*  Notes:
*
**********************************************************************/

#include <stdio.h>
#ifdef lint
#include <unistd.h>
#endif
#include <malloc.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <isc/keycodes.h>
#include <isc/window.h>
#include <isc/controls.h>
#include <isc/dlgman.h>
#include <isc/textedit.h>
#include <mousecur.h>
#include <moDfind.h>
#include "moview.h"

#define INIT register char *sp = instring;
#define GETC() (*sp++)
#define PEEKC() (*sp)
#define UNGETC(c) (--sp)
#define RETURN(c) return sp;
#define ERROR(c) return NULL
#include <regexp.h>

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif
#define	FINDMENU	3101		/* The "find" menu */
#define	SEARCHITEM	1			/* "Search for..." */
#define	NEXTITEM	2			/* "Find Next" */
#define	PREVITEM	3			/* "Find Previous */

#define	TABMENU		3102		/* The "tabs" menu (with no specific items) */

#define	LINEBUFSIZE	256			/* max length of one line, with trailing NUL */

ViewerWindow *SearchVW;
int dlgWindow = -1;				/* used for dialog box */
DMDcb *dialogUp;				/* if we have a dialog up, its cb is here */
#define ST_LITERAL	1
#define ST_CASEIN	2
#define ST_REGEX	3
static int searchType = ST_LITERAL;	/* literal/case-insensitive/regex */

/*
**  ugly global variables 
*/
extern BOOL			attributes;

/*
**  External declarations
*/
extern Icon viewerIcon;

/*
**  Local (static) declarations
*/
static ViewerWindow *VWroot = NULL;

static void ctlAction(ControlID theControl, short partCode);
static void doIconify(register ViewerWindow *VW);
static void drawWindow(register ViewerWindow *VW, BOOL erase);
static int getSearchArg(int *direction);
static void hiLite(register ViewerWindow *VW, register int line, BOOL lite);
static void scrollHoriz(register ViewerWindow *VW, register int howmuch);
static void scrollVert(register ViewerWindow *VW, register int howmany);
static void moveWithControl(ViewerWindow *VW, ControlID theControl);
static void putOneLine(register ViewerWindow *VW, unsigned char *in, 
					   register short v, register short h, 
					   register int offset, register int length);
static void searchAndDisplay(register ViewerWindow *VW, int direction, 
							 char *searcharg);
static void reValControls(register ViewerWindow *VW);
/*
**  Initialization routine - setup new ViewerWindow struct,
**  initialize assorted parameters, etc.
*/
ViewerWindow *
VWOpen	(short rows,
			 short columns,
			 short x,
			 short y,
			 short font,
		     short size,
             short tabstop,
		     char *title,
			 unsigned char *buffer,
		     int buffsize,
			 int grafFd,
			 int (* filler)(ViewerWindow *) )
{
	Rect myWinRect;
	register ViewerWindow *VW;
	Rect button;

	if ((VW = (ViewerWindow *)malloc(sizeof(*VW))) == NULL)
		return NULL;

	VW->rows = (rows == 0) ? 40 : rows;
	VW->columns = (columns == 0) ? 80 : columns;
	VW->xpos = (x == 0) ? 28 : x;
	VW->ypos = (y == 0) ? 32 : y;
	VW->iconified = VW->inTrackControl = FALSE;
	VW->lines = NULL;
	VW->textBuff = buffer;
	VW->hiLiteLine = -1;
	VW->tabStop = tabstop;
	VW->fillbuffer = filler;
	VW->grafFd = grafFd;
	if (filler)
		VW->inputEOF = FALSE;
	else
		VW->inputEOF = TRUE;
	VW->linesDone = FALSE;
	VW->textBuffCount = buffsize;
	*VW->regexbuf = 0x7F;
	VW->lineCount =
		VW->screenNext =
		VW->screenFirst =
		VW->firstColumn =
		VW->maxWidth =
		0;
	/* create window as hidden... */
    SetRect(&myWinRect, VW->xpos, VW->ypos, VW->xpos+VW->width, 
		VW->ypos+VW->height);
    VW->mainWindow = NewWindow(&myWinRect, title, hideWindow, documentProc, 
		frontWindow, hasGoAway|hasZoom|hasSendToBack|hasGrow|hasDrag, grafFd);
    SetPort(VW->mainWindow);
	/* set up the font stuff... */
	VW->font = font;
	VW->face = normalText;
	VW->size = size;
	TextFont(VW->font);
	TextFace(VW->face);
	TextSize(VW->size);
	GetFontInfo(&VW->fInfo);
	/* figure out height x width for text, plus scrollbars. */
	VW->fontHeight = VW->fInfo.ascent + VW->fInfo.descent + VW->fInfo.leading;
	VW->fontWidth = VW->fInfo.widMax;
	VW->height = VW->fontHeight * VW->rows + 17;
	VW->width = VW->fontWidth * VW->columns + 17;
	VW->moveFirst = VW->fInfo.ascent + 1;
	
	SizeWindow(VW->mainWindow, VW->width, VW->height, updateRequest);

    PenMode(srcXor);
	TextMode(srcOr);
    MouseMode(F_KEYDOWN | F_KEYUP | F_MOUSE1);

	/* set up the scroll bars */
    InitControls();
    button.edge.left = VW->width - 16;
    button.edge.right = VW->width + 1;
    button.edge.top = -1;
    button.edge.bottom = VW->height - 15;
    VW->vscrollbar = NewControl(VW->mainWindow, &button, "V", 1, 0, 0, 100,
			   scrollBarProc, 0);
    SetCtlAction(VW->vscrollbar, ctlAction);
    SetCtlScrollRate(VW->vscrollbar, 5); /* Pace this one. */

    button.edge.left = -1;
    button.edge.right = VW->width - 15;
    button.edge.top = VW->height - 16;
    button.edge.bottom = VW->height + 1;
    VW->hscrollbar = NewControl(VW->mainWindow, &button, "H", 1, 0, 0, 100,
			   scrollBarProc, 0);
	SetCtlAction(VW->hscrollbar, ctlAction);
    SetCtlScrollRate(VW->hscrollbar, 5); /* Pace this one. */
/* Prepare menu stuff */
	NewMenu(FINDMENU, "Find");
	AppendMenu(FINDMENU, "/sSearch For\311");
	AppendMenu(FINDMENU, "(/nSearch Next");
	AppendMenu(FINDMENU, "(/pSearch Previous");
	NewMenu(TABMENU, "Tabs");
	AppendMenu(TABMENU, "Show as ^I");
	AppendMenu(TABMENU, " 4 Columns");
	AppendMenu(TABMENU, " 8 Columns");
	AppendMenu(TABMENU, "12 Columns");
	CheckItem(TABMENU, (VW->tabStop / 4) + 1, 1);
/* Show them the window and return */
	ShowWindow(VW->mainWindow);
	SelectWindow(VW->mainWindow);
	SetRect(&VW->clipRect, 0, 0, VW->width - 17, VW->height - 17);
	ClipRect(&VW->clipRect);
	VW->next = VWroot;		/* link new struct at head of list */
	VWroot = VW;
	VW->VWMagic = VWMAGIC;
	return VW;
}


/*
**  Display a windowfull of text, starting with the line at VW->screenNext
*/
static void 
drawWindow(register ViewerWindow *VW, BOOL erase)
{
	register short vert = VW->moveFirst;
	register int newnext;
	int dif;
	int curport;

	GetPort(&curport);
	if (!(curport & 0x80000000L) && curport != VW->mainWindow)
		SetPort(VW->mainWindow);
	if (theFont != VW->font)
	{
		TextFont(VW->font);
		TextFace(VW->face);
		TextSize(VW->size);
	}

	if (erase)
		EraseRect(&VW->clipRect);
	else
		endspan(VW);
	/*
	**  A little trickery here to make sure that we don't prematurely
	**  truncate the bottom of the display if we haven't read in the
	**  whole file yet.
	*/
	while (VW->screenNext > (VW->lineCount - VW->rows))
	{
		if (VW->linesDone)
			VW->screenNext = VW->lineCount - VW->rows;
		else
			VWnoevent(VW);		/* this will update VW->lineCount */
	}

	if (VW->screenNext < 0)	VW->screenNext = 0;		/* don't start too soon */

	newnext = VW->screenNext + VW->rows;		/* how much will fit? */
	if (newnext > VW->lineCount) newnext = VW->lineCount;

	dif = VW->screenNext - VW->screenFirst;			/* how far do we move? */
	if (dif && dif > -VW->rows && dif < VW->rows && ! erase)		
		/* less than a screenful, but not a simple redraw */
	{
		VW->screenNext = VW->screenFirst + VW->rows;
		scrollVert(VW, dif);
	}
	else
	{
		if (VW->screenNext != VW->screenFirst)	/* we're not redrawing */
		{
			EraseRect(&VW->clipRect);
			VW->screenFirst = VW->screenNext;	/* save which line's on top */
		}
		for (; VW->screenNext < newnext; ++VW->screenNext)
		{
			putOneLine(VW, VW->textBuff + VW->lines[VW->screenNext], vert, 
					   VW->fInfo.leading, VW->firstColumn, VW->columns);
			vert += VW->fontHeight;
		}
	}
	hiLite(VW, VW->hiLiteLine, 1);
	reValControls(VW);						/* update scroll bars */
	if (VW->startrow >= 0) {	/* COPY selection up? */
		VW->lastrow = VW->startrow;
		VW->lastcol = VW->startcol;
		dospan(VW, &VW->endrow, &VW->endcol); /* Redo hilighting. */
	}
	if (!(curport & 0x80000000L) && curport != VW->mainWindow)
		SetPort(curport);
}

static void 
hiLite(register ViewerWindow *VW, register int line, BOOL lite)
{
	register int offset;
	register short vpos;
	Rect rect;

	offset = line - VW->screenFirst;
	if (offset < 0 || offset >= VW->rows)
		return;
	vpos = VW->moveFirst + offset * VW->fontHeight;
	SetRect(&rect, VW->clipRect.edge.left, vpos - VW->fInfo.ascent - 1,
			VW->clipRect.edge.right, vpos + VW->fInfo.descent + 1);
	EraseRect(&rect);
	putOneLine(VW, VW->textBuff + VW->lines[line], vpos,
			   VW->fInfo.leading, VW->firstColumn, VW->columns);
	if (lite)
		InvertRect(&rect);
	return;
}

/*
**  Draw one line of text (or part thereof for horizontal scrolling)
*/
static void 
putOneLine(register ViewerWindow *VW, unsigned char *in, register short vpos, 
	register short hpos, register int offset, register int length)
{
	char outbuff[LINEBUFSIZE];
	register char *out;			/* where to put translated characters */
	int face;					/* the face of the moment */
	int dlen;					/* length of string to draw */
	int howmany;				/* number of characters passed to DrawText */
	char *wherefrom;			/* start of characters passed to DrawText */
	int llen = 0;				/* length of line drawn so far */
	int stoplen = offset + length; /* length after which we can stop drawing */

	while (llen < stoplen && *in != '\n') {
		face = normalText;
		out = outbuff;
		while (*in != '\0' && *in != '\n' && out < (outbuff+sizeof(outbuff))) {
			if (*in == '\t' && VW->tabStop) {		/* translate tabs */
				int pad = VW->tabStop - ((out - outbuff + llen) % VW->tabStop);
				while (pad--)
					*out++ = ' ';
				++in;					/* skip the tab character */
				continue;
			}
			if (*in < ' ') {			/* control char */
				*out++ = '^';
				*out++ = *in++ | 0x40;	/* display controls in '^X' format */
				continue;
			}
			if (attributes && (*in > 0x7F))	{
				if (out != outbuff)	/* change of mode */
					break;
				if (*in >= NM_CHAR && *in <= ALL_CHAR)
					face = (*in & 7) << 1;
				in++;
			}
			else
			{
				*out++ = *in++;
			}
		}
		dlen = out - outbuff;
		if (llen <= offset) {	/* not displaying yet */
			if (llen+dlen <= offset) {	/* Time to start yet? */
				llen += dlen;
				continue;
			}
			howmany = llen+dlen-offset;
			wherefrom = out - howmany;
		} else {
			howmany = dlen;
			wherefrom = outbuff;
		}
		llen += dlen;
		if (theFace != face)
			TextFace(face);
		MoveTo(hpos, vpos);
		if (howmany > length)
			howmany = length;
		hpos += howmany * VW->fontWidth;
		if (hpos > VW->width) {	/* Will it run off the end of the window? */
			howmany -= (hpos - VW->width) / VW->fontWidth; /* Trim to fit. */
			stoplen = 0;		/* Force an exit of the loop (hack!). */
		}
		DrawText(wherefrom, 0, howmany);
	}
	if (llen > VW->maxWidth)
		VW->maxWidth = llen;
	if (theFace != normalText)
		TextFace(normalText);
}

int
VWnoevent(parm)
ViewerWindow *parm;
{
#define CHUNKSIZE 1000			/* number of lines in a chunk */
	int size;
	unsigned char *nextbyte;
	register ViewerWindow *VW;

	VW = parm;
	if (!VW)					/* called us from noevent, not directly */
		for (VW=VWroot; VW; VW = VW->next)	/* look for window needing work */
			if (!VW->linesDone)
				break;
	if (! VW)
		return FALSE;					/* not my window, no action */

	if (VW->linesDone)
		return FALSE;
	if (! VW->inputEOF && VW->fillbuffer)	/* more file and a way to get it */
	{
		if (VW->fillbuffer(VW) == 0)		/* go get it */
			VW->inputEOF = TRUE;
		if (VW->inputtruncated) {
			truncated(VW);
			goto nomo;
		}
	}

	if (! VW->lines)
	{
		VW->lines = (int *)malloc(sizeof(int) * CHUNKSIZE);
		if (! VW->lines)
		{
			char *msg = ": can't allocate memory\n";
/*			write(2, Myname, strlen(Myname)); */
			write(2, msg, strlen(msg));
			exit (1);
		}
		VW->linesSize = CHUNKSIZE;
		VW->lines[0] = 0;					/* first line starts at offset 0 */
	}
	nextbyte = VW->textBuff + VW->lines[VW->lineCount];
	size = VW->textBuffCount - VW->lines[VW->lineCount];
	while (size--)
		if (*nextbyte++ == '\n')		/* cumulatively count lines in input */
		{
			if (++VW->lineCount + 1 >= VW->linesSize) /* one extra */
			{
				int *save = VW->lines;
				VW->linesSize += CHUNKSIZE;		/* grow line offsets array */
				VW->lines = (int *)realloc(VW->lines, 
									VW->linesSize * sizeof(VW->lines[0]));
				if (VW->lines == NULL)
				{
					VW->linesSize -= CHUNKSIZE;
					VW->lines = save;
					truncated(VW);
					return FALSE;
				}
			}
			VW->lines[VW->lineCount] = nextbyte - VW->textBuff;	/* save offset */
		}
  nomo:
	reValControls(VW);
	VW->lines[VW->lineCount + 1] = VW->textBuffCount; /* for search */
	if (VW->inputEOF) {
		VW->linesDone = TRUE;
	}
	return TRUE;
}

/*
** Input was truncated due to lack of memory.  Try to put some sort of
** notice to this effect into the viewed buffer.
*/

truncated(VW)
ViewerWindow *VW;
{
	char *msg;
	int len, size;
	unsigned char *nextbyte;

	if (!VW->linesSize)
		return;
	msg = "\205...Truncated due to lack of memory!\200\n";
	len = StringLen(msg);
	VW->inputtruncated = TRUE;	/* Just in case. */
	while (((size = VW->textBuffCount - VW->lines[VW->lineCount]) < len)
		   || (VW->lineCount >= VW->linesSize))
		--VW->lineCount;		/* Make room for warning message. */
	nextbyte = VW->textBuff + VW->lines[VW->lineCount];
	strcpy(nextbyte, msg);
	nextbyte += len;
	VW->lines[++VW->lineCount] = nextbyte - VW->textBuff;
	VW->lines[VW->lineCount + 1] = VW->textBuffCount; /* for search */
}

/*
**  Keyboard event handler 
*/
void 
VWkeyboard(EventRecord *ep)
{
    unsigned int key;
#define WAITCMD 0				/* state is waiting for command */
#define WAITMARK 1				/* state is waiting for mark letter */
#define WAITJUMP 2				/* state is waiting for jump letter */
	register ViewerWindow *VW;
	int direction;

	for (VW=VWroot; VW; VW = VW->next)	/* look for one of my windows */
		if (VW->mainWindow == ep->who)
			break;
	if (! VW)
		return;							/* not my window, no action */

    if (ep->message & F_KEYUP)
		return;
	key = ep->message & ~F_KEYDOWN & ~F_KBDIII;
	if (key == (F_CTRL | F_SHIFT | 'I') || key == (F_CTRL | 'i'))
	{
		doIconify(VW);
		return;
	}
	if (VW->iconified)
	{
		Bell();
		return;
	}

	ObscureCursor();
	if (VW->kbstate == WAITMARK)		/* which mark do they want to set */
	{
		if (key >= 'A' && key <= 'Z')
			key += 0x20;				/* convert to lowercase */
		if (key < 'a' || key > 'z')
			Bell();
		else
			VW->marks[key - 'a'] = VW->screenFirst;
		VW->kbstate = WAITCMD;
		return;
	}
	if (VW->kbstate == WAITJUMP)		/* where do they want to jump? */
	{
		if (key >= 'A' && key <= 'Z')
			key += 0x20;				/* convert to lowercase */
		if (key < 'a' || key > 'z')
			Bell();
		else
		{
			VW->screenNext = VW->marks[key - 'a'];
			drawWindow(VW, FALSE);
		}
		VW->kbstate = WAITCMD;
		return;
	}

	switch (key)
	{
	case F_CTRL | F_SHIFT | 'C':
	case F_CTRL | 'c':
		copy(VW);
		break;

	case F_FUNC | K_PGUP:						/* plain page up */
	case F_FUNC | K_NPAD9:
	case F_CTRL | F_SHIFT | 'B':
	case F_CTRL | 'b':
	case 'b':
	case F_CTRL | F_SHIFT | 'Z':
	case F_CTRL | 'z':
	case 'z':

		if (VW->screenNext > VW->rows)
		{ 
			VW->screenNext = VW->screenFirst - VW->rows; 	/* back one page */
			drawWindow(VW, FALSE); 
		}
		else
			Bell();
		break;
	case F_FUNC | K_PGDN:						/* plain page down */
	case F_FUNC | K_NPAD3:
	case F_CTRL | F_SHIFT | 'F':
	case F_CTRL | 'f':
	case 'f':
	case F_CTRL | F_SHIFT | 'V':
	case F_CTRL | 'v':
	case 'v':
	case ' ':
		if (VW->screenNext < VW->lineCount)
			drawWindow(VW, FALSE);
		else
			Bell();
		break;
	case F_FUNC | K_UP:							/*  plain up arrow */
	case F_FUNC | K_NPAD8:
	case F_CTRL | F_SHIFT | 'K':
	case F_CTRL | 'k':
	case 'k':
	case 'y':
		if (VW->screenNext > VW->rows)
		{
			VW->screenNext = VW->screenFirst - 1;
			drawWindow(VW, FALSE);
		}
		else
			Bell();
		break;
	case F_FUNC | K_DOWN:						/* plain down arrow */
	case F_FUNC | K_NPAD2:
	case F_FUNC | K_NPENTER:
	case F_FUNC | K_NPLUS:
	case F_FUNC | K_ENTER:
	case F_CTRL | F_SHIFT | 'J':
	case F_CTRL | 'j':
	case 'j':
	case 'e':
		if (VW->screenNext < VW->lineCount)		/* if not already at end */
		{
			VW->screenNext = VW->screenFirst + 1;
			drawWindow(VW, FALSE);
		}
		else
			Bell();
		break;
	case F_FUNC | K_LEFT:						/* plain left arrow */
	case F_FUNC | K_NPAD4:
	case 'h':
		if (VW->firstColumn > VW->tabStop)
		{
			scrollHoriz(VW, - VW->tabStop);
		}
		else
		{
			scrollHoriz(VW, -VW->firstColumn);
		}
		break;
	case F_CTRL | F_FUNC | K_LEFT:				/* CTRL left arrow */
	case F_CTRL | F_FUNC | K_NPAD4:
	case F_CTRL | F_SHIFT | F_FUNC | K_LEFT:
	case F_CTRL | F_SHIFT | F_FUNC | K_NPAD4:
		if (VW->firstColumn > 0)
		{
			scrollHoriz(VW,-1);
		}
		else
		{
			Bell();
		}
		break;	
	case F_FUNC | K_RIGHT:						/* plain right arrow */
	case F_FUNC | K_NPAD6:
	case 'l':
		if (VW->firstColumn + VW->tabStop < VW->maxWidth - VW->columns)
		{
			scrollHoriz(VW,VW->tabStop);
		}
		else
		{
			scrollHoriz(VW,VW->maxWidth - VW->columns - VW->firstColumn);
		}
		break;	
	case F_CTRL | F_FUNC | K_RIGHT:				/* CTRL right arrow */
	case F_CTRL | F_FUNC | K_NPAD6:
	case F_CTRL | F_SHIFT | F_FUNC | K_RIGHT:
	case F_CTRL | F_SHIFT | F_FUNC | K_NPAD6:
		if (VW->firstColumn < VW->maxWidth - VW->columns)
		{
			scrollHoriz(VW,1);
		}
		else
		{
			Bell();
		}
		break;	
	case F_ALT | '0':
	case F_ALT | '4':
	case F_ALT | '8':
		VW->tabStop = key & 0x0F;
		VW->screenNext = VW->screenFirst;
		drawWindow(VW, TRUE);
		break;
	case F_ALT | F_FUNC | K_LEFT:				/* ALT left arrow */
	case F_ALT | F_FUNC | K_NPAD4:
		VW->xpos -= 10;
		MoveWindow(ep->who, VW->xpos, VW->ypos, 0);
		break;
	case F_ALT | F_FUNC | K_RIGHT:				/* ALT right arrow */
	case F_ALT | F_FUNC | K_NPAD6:
		VW->xpos += 10;
		MoveWindow(ep->who, VW->xpos, VW->ypos, 0);
		break;
	case F_ALT | F_FUNC | K_UP:					/* ALT up arrow */
	case F_ALT | F_FUNC | K_NPAD8:
		VW->ypos -= 10;
		MoveWindow(ep->who, VW->xpos, VW->ypos, 0);
		break;
	case F_ALT | F_FUNC | K_DOWN:				/* ALT down arrow */
	case F_ALT | F_FUNC | K_NPAD2:
		VW->ypos += 10;
		MoveWindow(ep->who, VW->xpos, VW->ypos, 0);
		break;
	case F_FUNC | K_NPAD5:						/* redraw */
	case F_CTRL | F_SHIFT | 'L':
	case F_CTRL | 'l':
	case F_CTRL | F_SHIFT | 'R':
	case F_CTRL | 'r':
	case F_SHIFT | 'R':
	case 'r':
		VW->screenNext = VW->screenFirst;
		drawWindow(VW, TRUE);
		break;
	case F_FUNC | K_END:						/* End */
	case F_FUNC | K_NPAD1:
	case F_FUNC | K_PF4:
	case F_SHIFT | 'G':
		if (! VW->inputEOF) {
			SetMouseCursor(MC_WATCH);
			while (! VW->inputEOF)
				VWnoevent(VW);
			SetMouseCursor(MC_ARROW);
		}
		if (VW->screenNext != VW->lineCount)
		{
			VW->screenNext = VW->lineCount;
			drawWindow(VW, FALSE);
		}
		break;
	case F_FUNC | K_HOME:						/* Home */
	case F_FUNC | K_NPAD7:
	case F_FUNC | K_PF3:
	case 'g':
		if (VW->screenFirst != 0)
		{
			VW->screenNext = 0;
			drawWindow(VW, FALSE);
		}
		break;
	case F_CTRL | F_SHIFT | 'D':				/* half-screen down */
	case F_CTRL | 'd':
	case F_SHIFT | 'D':
	case 'd':
	case F_CTRL | F_FUNC | K_DOWN:
	case F_CTRL | F_FUNC | K_NPAD2:
	case F_SHIFT | F_FUNC | K_DOWN:
	case F_SHIFT | F_FUNC | K_NPAD2:
	case F_CTRL | F_SHIFT | F_FUNC | K_DOWN:
	case F_CTRL | F_SHIFT | F_FUNC | K_NPAD2:
		VW->screenNext = VW->screenFirst + VW->rows / 2;
		drawWindow(VW, FALSE);
		break;
	case F_CTRL | F_SHIFT | 'U':				/* half-screen up */
	case F_CTRL | 'u':
	case F_SHIFT | 'U':
	case 'u':
	case F_CTRL | F_FUNC | K_UP:
	case F_CTRL | F_FUNC | K_NPAD8:
	case F_SHIFT | F_FUNC | K_UP:
	case F_SHIFT | F_FUNC | K_NPAD8:
	case F_CTRL | F_SHIFT | F_FUNC | K_UP:
	case F_CTRL | F_SHIFT | F_FUNC | K_NPAD8:
		VW->screenNext = VW->screenFirst - VW->rows / 2;
		drawWindow(VW, FALSE);
		break;
	case F_CTRL | F_SHIFT | 'A':				/* beginning of line */
	case F_CTRL | 'a':
	case F_FUNC | K_PF1:
	case F_SHIFT | F_FUNC | K_LEFT:
	case F_SHIFT | F_FUNC | K_NPAD4:
	case '0':
		scrollHoriz(VW, - VW->firstColumn);
		break;
	case F_CTRL | F_SHIFT | 'E':				/* end of line */
	case F_CTRL | 'e':
	case F_FUNC | K_PF2:
	case F_SHIFT | F_FUNC | K_RIGHT:
	case F_SHIFT | F_FUNC | K_NPAD6:
	case F_SHIFT | '$':
		if (VW->maxWidth > VW->columns)
			scrollHoriz(VW, VW->maxWidth - VW->columns - VW->firstColumn);
		break;
	case F_SHIFT | 'M':						/* set mark */
	case 'm':
		VW->kbstate = WAITMARK;
		break;
	case '\'':								/* goto mark */
		VW->kbstate = WAITJUMP;
		break;
	case '/':								/* search */
		if (getSearchArg(&direction))
		{
			hiLite(VW, VW->hiLiteLine, 0); /* un-hilite previous, if any */
			VW->hiLiteLine = -1;				/* new search */
			searchAndDisplay(VW, direction, VW->regexbuf);
		}
		break;
	case F_CTRL | F_SHIFT | 'P':			/* search previous */
	case F_CTRL | 'p':
		searchAndDisplay(VW, -1, VW->regexbuf);
		break;
	case F_CTRL | F_SHIFT | 'N':			/* search next */
	case F_CTRL | 'n':
		searchAndDisplay(VW, 1, VW->regexbuf);
		break;
	default:
		Bell();
	}
}
/*
**
**  Move the display horizontally by some number of columns.
**  Negative is left, positive is right.
**  Update VW->firstColumn to reflect the move, redisplay as necessary.
**
*/
static void 
scrollHoriz(register ViewerWindow *VW, register int howmuch)
{
	int prevfirst = VW->firstColumn;
	Rect rect = VW->clipRect;
	register int newmax = ((VW->screenNext > VW->lineCount) ? VW->lineCount : VW->screenNext);
	register short vert = VW->moveFirst;
	int horiz;
	int start;
	int curport;

	GetPort(&curport);
	if (!(curport & 0x80000000L) && curport != VW->mainWindow)
		SetPort(VW->mainWindow);

	VW->firstColumn += howmuch;
	if (VW->firstColumn < 0) VW->firstColumn = 0;		/* leftmost column */
	if (! (howmuch = VW->firstColumn - prevfirst)) return;	/* didn't move */
	ScrollRect(&VW->clipRect, -howmuch * VW->fontWidth, 0, updateRequest);
	if (howmuch > 0)
	{
		start = VW->firstColumn + VW->columns - howmuch;
		horiz = (VW->columns - howmuch) * VW->fontWidth + VW->fInfo.leading;
		rect.edge.left = rect.edge.right - howmuch * VW->fontWidth;
	}
	else
	{
		start = VW->firstColumn;
		horiz = VW->fInfo.leading;
		rect.edge.right = rect.edge.left - howmuch * VW->fontWidth;
		howmuch = -howmuch;			/* absolute value, for putOneLine */
	}
	ValidRect(&rect);

	VW->screenNext = VW->screenFirst;	/* back to current page */
	if (VW->screenNext < 0) VW->screenNext = 0;
	for (; VW->screenNext < newmax; ++VW->screenNext)
	{
		putOneLine(VW, VW->textBuff + VW->lines[VW->screenNext], vert, horiz, start, howmuch);
		vert += VW->fontHeight;
	}
	hiLite(VW, VW->hiLiteLine, 1);
	reValControls(VW);
	if (!(curport & 0x80000000L) && curport != VW->mainWindow)
		SetPort(curport);
}

/*
**  Scroll the text vertically, negative is toward beginning of file.
*/
static void 
scrollVert(register ViewerWindow *VW, register int howmany)
{
	Rect openrect;
	register int howfar;
	register int drawwhere;
	register int drawwhich;

	if (!howmany) return;

/*
**  Prevent scrolling off the front or back end of the text.
*/
	if (-howmany > VW->screenFirst)
		howmany = -VW->screenFirst;
	else if (howmany > VW->lineCount - VW->screenNext + 1)
		howmany = VW->lineCount - VW->screenNext;

/*
**  Calculate the rectangle "opened up" by the scrolling.
**  Then do the scroll, and ValidRect the opened area.
*/
	howfar =  howmany * VW->fontHeight;
	openrect = VW->clipRect;
	if (howmany > 0)
	{
		openrect.edge.top = openrect.edge.bottom - howfar;
		drawwhich = VW->screenNext;
		VW->screenNext += howmany;
		VW->screenFirst += howmany;
	}
	else
	{
		openrect.edge.bottom = openrect.edge.top - howfar;
		drawwhich = VW->screenFirst + howmany;
		VW->screenNext += howmany;
		VW->screenFirst += howmany;
		howmany = -howmany;
	}
	ScrollRect(&VW->clipRect, 0, -howfar, updateRequest);
	ValidRect(&openrect);
/*
** Draw the lines which belong in the opened area.
*/
	drawwhere = (drawwhich - VW->screenFirst) * VW->fontHeight + VW->moveFirst;
	while (howmany--)
	{
		putOneLine(VW, VW->textBuff + VW->lines[drawwhich], 
				   drawwhere, VW->fInfo.leading, 
				   VW->firstColumn, VW->columns);
		drawwhere += VW->fontHeight;
		++drawwhich;
	}
	reValControls(VW);
}

/*
** Mouse-driven COPY helpers.
*/

void
startspan(VW, row, col)
ViewerWindow *VW;
short row, col;
{
    if (VW->startrow >= 0) {
		removespan(VW);			/* Remove highlighting. */
    }
    VW->lastrow = VW->startrow = VW->endrow = row;
    VW->lastcol = VW->startcol = VW->endcol = col;
}

dospan(VW, currow, curcol)
ViewerWindow *VW;
short *currow, *curcol;
{
    Rect hiliterect;
    short row, col, scolonrow, newcol, srow, newrow;

    if (VW->startrow < 0)
		return;
    if (*currow < 0)			/* Constrain span variables to valid range. */
		*currow = 0;
    if (*currow >= VW->rows)
		*currow = VW->rows - 1;
    if (*curcol < 0)
		*curcol = 0;
    if (*curcol >= VW->columns)
		*curcol = VW->columns - 1;

    newcol = col = *curcol;		/* Normalize the span. */
    scolonrow = VW->lastcol;
    newrow = row = *currow;
    srow = VW->lastrow;
    if (row < srow) {
		short i;

		row = srow;
		srow = newrow;
		i = VW->lastcol;
		VW->lastcol = col;
		col = i;
    } else if (row == srow && col < scolonrow) {
		col = scolonrow;
		scolonrow = newcol;
    }

    if (row != srow) {			/* Row changed? */
		hiliterect.edge.top = srow * VW->fontHeight + 1;
		hiliterect.edge.bottom = hiliterect.edge.top + VW->fontHeight;
		hiliterect.edge.left = VW->lastcol*VW->fontWidth + VW->fInfo.leading;
		hiliterect.edge.right = VW->columns*VW->fontWidth + VW->fInfo.leading;
		InvertRect(&hiliterect); /* Do first (partial) row. */
		if ((srow + 1) < row) {	/* Spanned more than one line? */
			hiliterect.edge.top = hiliterect.edge.bottom;
			hiliterect.edge.bottom = row * VW->fontHeight + 1;
			hiliterect.edge.left = VW->fInfo.leading;
			InvertRect(&hiliterect); /* Do middle row(s).  (Whole ones.) */
		}
		hiliterect.edge.top = hiliterect.edge.bottom;
		hiliterect.edge.bottom += VW->fontHeight;
		hiliterect.edge.left = VW->fInfo.leading;
		hiliterect.edge.right = col * VW->fontWidth + VW->fInfo.leading;
		InvertRect(&hiliterect); /* Do last (partial) row. */
		VW->lastrow = newrow;
		VW->lastcol = newcol;
    } else if (col != scolonrow) { /* Column changed? */
		hiliterect.edge.top = row * VW->fontHeight + 1;
		hiliterect.edge.bottom = hiliterect.edge.top + VW->fontHeight;
		hiliterect.edge.left = scolonrow * VW->fontWidth + VW->fInfo.leading;
		hiliterect.edge.right = col * VW->fontWidth + VW->fInfo.leading;
		InvertRect(&hiliterect);
		VW->lastcol = newcol;
    }
}

removespan(VW)
ViewerWindow *VW;
{
    if (VW->startrow >= 0) {
		VW->lastrow = VW->startrow;
		VW->lastcol = VW->startcol;
		dospan(VW, &VW->endrow, &VW->endcol); /* Remove highlighting. */
		VW->startrow = -1;		/* Mark highlighting as gone. */
		copy_off();				/* Disallow a COPY. */
    }
}

endspan(VW)
ViewerWindow *VW;
{
    removespan(VW);
}

short newcopy;					/* New outgoing scrap. */
char *scraphandle;
long scraphandlesize;

void
scrapevent(ep)
EventRecord *ep;
{
    if (ep->message) {
		if (newcopy)
			writescrap();
    }
}

void
killscrap()
{
    if (scraphandle) {
		free(scraphandle);
		scraphandle = 0;
    }
}

writescrap()
{
	ZeroScrap();
	if (scraphandle) {
	    PutScrap(scraphandlesize, 'TEXT', scraphandle);
	}
	PublishScrap(0);
	newcopy = 0;
}

/*
** Put the selected data into the private scrap, and make it
** available for export.
*/

copy(VW)
ViewerWindow *VW;
{
    short row, col, srow, scol, erow, ecol;
    short mrow, mcol, flag, i, offset;
    unsigned char *p, *q, *limit, *malloc();

    if (VW->startrow < 0)
		return;
    killscrap();
    srow = VW->startrow;
    scol = VW->startcol;
    erow = VW->endrow;
    ecol = VW->endcol;
    mrow = VW->rows;
    mcol = VW->maxWidth;
    if (erow < srow) {		/* Normalize the input range. */
		i = erow;
		erow = srow;
		srow = i;
      swapcol:
		i = ecol;
		ecol = scol;
		scol = i;
    } else if (erow == srow && ecol < scol) {
		goto swapcol;
    }
	p = VW->textBuff + VW->lines[i = srow + VW->screenFirst]
		+ scol + VW->firstColumn;
	limit = VW->textBuff + VW->lines[i + 1];
	if (p > limit)
		p = limit;
	q = VW->textBuff + VW->lines[i = erow + VW->screenFirst]
		+ ecol + VW->firstColumn;
	limit = VW->textBuff + VW->lines[i + 1];
	if (q > limit)
		q = limit;
    i = scraphandlesize = q - p;
    newcopy = 1;				/* Set export-on-deactivate flag. */
    if (i && (scraphandle = malloc(i))) {
		memcpy(scraphandle, p, i); /* Copy the data to our local scrap. */
		copy_on();
    }
    endspan(VW);
}

/*
**  Mouse event handler.
**
** Right here we only handle the buttons.  Scroll bars are taken care of
** elsewhere (because we want continuous scrolling on a press).  This is,
** however, where thumb (elevator box) drags are reported.
*/

void 
VWmouse(EventRecord *ep)
{
    ControlID p;
    short i, row, col;
    Rect rect;
    register ViewerWindow *VW;
    int curport;

    for (VW=VWroot; VW; VW = VW->next) /* look for one of my windows */
	if (VW->mainWindow == ep->who)
	    break;
    if (!VW)
	return;			/* not my window, no action */

	if (FindControl(&ep->where.corner.topLeft, ep->who, &p)) {	  
		if (ep->message & F_KEYDOWN)  /* Mouse-down event? */
		{
			GetPort(&curport);
			if (!(curport & 0x80000000L) && curport != VW->mainWindow)
				SetPort(VW->mainWindow);
			SetRect(&rect,0,0,30000,30000);
			ClipRect(&rect);
			VW->inTrackControl = TRUE;	/* because noEvent may be called */
		    i = TrackControl(p, &ep->where.corner.topLeft, ctlAction); 
			VW->inTrackControl = FALSE;
			ClipRect(&VW->clipRect);
			if (i == inThumb)
			{					/* Account for thumb drags. */
				moveWithControl(VW, p);
			}
			if (!(curport & 0x80000000L) && curport != VW->mainWindow)
				SetPort(curport);
		}
		return;
	} else {					/* Mouse COPY selection. */
		row = ep->where.edge.top - 1; /* Compute the row, column. */
		row /= VW->fontHeight;
		col = ep->where.edge.left - VW->fInfo.leading;
		col /= VW->fontWidth;
		if (ep->message & F_KEYDOWN) {
			if ((ep->message & F_KEYMASK) == 1) {
				MouseMode(F_KEYDOWN | F_KEYUP | F_MOUSE1); /* Allow drags. */
				startspan(VW, row, col);
			}
		} else if (ep->message & F_KEYUP) {
			if ((ep->message & F_KEYMASK) == 1) {
				MouseMode(F_KEYDOWN | F_KEYUP);	/* Normal mousies. */
				dospan(VW, &row, &col);
				VW->endrow = row;
				VW->endcol = col; /* Spanned range? */
				if (row != VW->startrow || col != VW->startcol) {
					copy_on();	/* Allow a COPY. */
				} else {
					endspan(VW);
				}
			}
		} else if (ep->message & F_MOUSE1) { /* Was a drag event. */
			dospan(VW, &row, &col);
		}
	}
}

/*
**  Move the display to track the (new) position of the control.
*/
static void 
moveWithControl(ViewerWindow *VW, ControlID theControl)
{
	int v = GetCtlValue(theControl);
	if (theControl == VW->vscrollbar)
	{
		if (v == 100)			/* special case for zipping to end of file */
			while(! VW->inputEOF)
				VWnoevent(VW);
		VW->screenNext = (v * (VW->lineCount - VW->rows)) / 100;
		drawWindow(VW, FALSE);		/* show 'em  the page */
	}
	if (theControl == VW->hscrollbar)
	{
		VW->firstColumn = (v * (VW->maxWidth - VW->columns)) / 100;
		VW->screenNext = VW->screenFirst;
		drawWindow(VW, TRUE);
	}
}

/*
** This routine is called repeatedly if the operator presses in the controls
** of scroll bars.  It is optional (installed via SetCtlAction).  Without it
** the scroll bars only accept a single click in any part of them.  Generally
** users will expect scrolling to be continuous, so you want one of these.
**
** This routine is _not_ called when dragging the thumb.
*/

static void 
ctlAction(ControlID theControl, short partCode)
{
	Rect rect;
	register ViewerWindow *VW;

	for (VW=VWroot; VW; VW = VW->next)	/* look for one of my windows */
		if (theControl == VW->vscrollbar || theControl == VW->hscrollbar)
			break;
	if (! VW)
		return;							/* not my window, no action */

	ClipRect(&VW->clipRect); 
	if (theControl == VW->vscrollbar) 
	{
		if (partCode == inUpButton)
		{
			VW->screenNext = VW->screenFirst - 1;
			drawWindow(VW, FALSE);
		}
		else if (partCode == inDownButton)
		{
			VW->screenNext = VW->screenFirst + 1;
			drawWindow(VW, FALSE);
		}
		else if (partCode == inPageUp)
		{ 
			VW->screenNext = VW->screenFirst - VW->rows; /* prev page */
			drawWindow(VW, FALSE); 
		}
		else if (partCode == inPageDown)
			drawWindow(VW, FALSE);
	}
	if (theControl == VW->hscrollbar)
	{
		if (partCode == inUpButton)
			scrollHoriz(VW,-1);
		else if (partCode == inDownButton)
			scrollHoriz(VW,1);
		else if (partCode == inPageUp)
			scrollHoriz(VW,-5);
		else if (partCode == inPageDown)
			scrollHoriz(VW,5);
	}
	SetRect(&rect, 0, 0, 30000, 30000);
	ClipRect(&rect);
}

/*
**  window event handler
*/	
void 
VWwindow(EventRecord *ep)
{
    Rect rect;
    register ViewerWindow *VW;
    int whichport;

    for (VW=VWroot; VW; VW = VW->next) /* look for one of my windows */
	if (VW->mainWindow == ep->who)
	    break;
    if (!VW)
	return;			/* not my window, no action */

    GetPort(&whichport);
    if (!(whichport & 0x80000000L) && whichport != ep->who)
	SetPort(ep->who);
    switch (ep->message & F_KEYMASK)  {
      case windowGoaway:
	VWClose(VW);
	ReleaseEvent();
	break;

      case windowMove:
	VW->xpos += ep->where.edge.left;
	VW->ypos += ep->where.edge.top;
	MoveWindow(ep->who, VW->xpos, VW->ypos, 0);
	break;

      case windowResize:
	if (!VW->iconified)  {
	    VW->screenNext = VW->screenFirst;	/* start of redisplay  */
	    SetRect(&rect, VW->width-17, VW->height-17, VW->width, VW->height);
	    VW->width += ep->where.edge.left;
	    VW->height += ep->where.edge.top;
	    VW->width = ((VW->width - 17)/ VW->fontWidth) * VW->fontWidth + 17;
	    VW->height = ((VW->height - 17) / VW->fontHeight) * VW->fontHeight + 17;
	    SetRect(&VW->clipRect, 0, 0, 30000, 30000);
	    ClipRect(&VW->clipRect);
	    EraseRect(&rect);	/* Erase old grow box. */
	    rect.edge.left = VW->width - 16;
	    rect.edge.top = -1;
	    HideControl(VW->vscrollbar);
	    MoveControl(VW->vscrollbar, rect.edge.left, rect.edge.top);
	    SizeControl(VW->vscrollbar, 17, VW->height - 14);
	    rect.edge.left = -1;
	    rect.edge.top = VW->height - 16;
	    HideControl(VW->hscrollbar);
	    MoveControl(VW->hscrollbar, rect.edge.left, rect.edge.top);
	    SizeControl(VW->hscrollbar, VW->width - 14, 17);
	    SizeWindow(ep->who, VW->width, VW->height, noupdateRequest);
	    ShowControl(VW->vscrollbar);
	    ShowControl(VW->hscrollbar);
	    DrawGrowIcon(ep->who); /* Draw new grow box. */
	    SetRect(&VW->clipRect, 0, 0, VW->width - 17, VW->height - 17);
	    ClipRect(&VW->clipRect);
	    VW->rows = VW->clipRect.edge.bottom / VW->fontHeight;
	    VW->columns = VW->clipRect.edge.right / VW->fontWidth;
	    drawWindow(VW, FALSE); 
	}
	break;

      case windowZoom:
	doIconify(VW);
	break;
      default:
	break;
    }
    if (!(whichport & 0x80000000L) && whichport != ep->who)
	SetPort(whichport);	/* restore viewport */
}

static void 
doIconify(register ViewerWindow *VW)
{
	if (VW->iconified) 
	{
		EnableMenu(FINDMENU);
		EnableMenu(TABMENU);
		UnIconifyWindow(VW->mainWindow);
	} 
	else 
	{
		DisableMenu(FINDMENU);
		DisableMenu(TABMENU);
		IconifyWindow(VW->mainWindow);
	}
	VW->iconified ^= 1;
	return;
}
/*
**  Set the control values for the scrollbars to indicate our
**  current position in the input file.
*/
static void 
reValControls(register ViewerWindow *VW)
{
	Rect rect;
	register int size;
	register int pct;

	if (dialogUp)				/* don't screw up dialog box */
		return;

	TextFont(VW->font);		/* get around control manager "feature" */
	TextFace(VW->face);
	TextSize(VW->size);

	SetRect(&rect, 0, 0, 30000, 30000);	/* no clip rect */
	ClipRect(&rect);
	size = (100 * VW->columns) / (VW->maxWidth ? VW->maxWidth : 1);
	if (VW->maxWidth > VW->columns)
		pct = (100 * VW->firstColumn) / (VW->maxWidth - VW->columns);
	else
		pct = 100;
	if (size >= 100 || !VW->active)
		HiliteControl(VW->hscrollbar, 255);		/* inactive */
	else
		HiliteControl(VW->hscrollbar, 1);
	SetCtlValues(VW->hscrollbar, pct, size);
	size = (100 * VW->rows) / (VW->lineCount ? VW->lineCount : 1);
	if (VW->lineCount > VW->rows)
		pct = (100 * (VW->screenNext - VW->rows)) / (VW->lineCount - VW->rows);
	else
		pct = 100;
	if (size >= 100 || !VW->active)
		HiliteControl(VW->vscrollbar, 255);		/* inactive */
	else
		HiliteControl(VW->vscrollbar, 1);
	SetCtlValues(VW->vscrollbar, pct, size);
	ClipRect(&VW->clipRect);
}

/*
**  Window update event handler
*/
void 
VWupdate(EventRecord *ep)
{
    Rect rect;
    int whichport, save;
    register ViewerWindow *VW;

    GetPort(&whichport);
    if (ep->who != dlgWindow) {		/* if it's not for my dialog window... */
	for (VW=VWroot; VW; VW = VW->next)	/* look for one of my windows */
	    if (VW->mainWindow == ep->who
		|| ((whichport & 0x80000000L) && (whichport & 1<<ep->who)
		    && (whichport & 1<<VW->mainWindow)))
		break;
	if (! VW)
	    return;							/* not my window, no action */
    }

    SetPort(ep->who);
    BeginUpdate(ep->who);
    if (ep->who == dlgWindow)
	DMUpdate(dialogUp);
    else {
	if (VW->iconified) {
	    SetRect(&rect, 0, 0, 32, 32);
	    PlotIcon(&rect, &viewerIcon);
	} else {
	    TextFont(VW->font);	/* 'cuz dlgman leaves it wrong... */
	    TextFace(VW->face);
	    TextSize(VW->size);
	    SetRect(&rect, 0, 0, 30000, 30000);
	    ClipRect(&rect);
	    DrawControls(ep->who);
	    DrawGrowIcon(ep->who);
	    ClipRect(&VW->clipRect);
	    VW->screenNext = VW->screenFirst;
	    save = VW->mainWindow; /* Spoof in case of copy video. */
	    VW->mainWindow = ep->who;
	    drawWindow(VW, TRUE);
	    VW->mainWindow = save;
	}
    }

    EndUpdate(ep->who);
    SetPort(whichport);
}

/*
**  Activate event handler
*/
void
VWactivate(EventRecord *ep)
{
    static int lastevent;
    register ViewerWindow *VW;
    Rect rect;
    int whichport;
    unsigned short hilite_v, hilite_h;

    if (ep->who == dlgWindow) {
	lastevent = ep->message;
	return;
    }

    for (VW=VWroot; VW; VW = VW->next) /* look for one of my windows */
	if (VW->mainWindow == ep->who)
	    break;
    if (!VW)
	return;			/* not my window, no action */

    GetPort(&whichport);
    SetPort(ep->who);
    SetRect(&rect, 0, 0, 30000, 30000);	/* no clip rect */
    ClipRect(&rect);
    if (ep->message) {
	if (dialogUp) {		/* main activated, and dialog window exists */
	    SetPort(VW->mainWindow);
	    if (lastevent)	/* last wind event activated dialog */
		SendBehind(VW->mainWindow, dlgWindow);
	    else {
		SelectWindow(dlgWindow);
	    }
	    SetPort(dlgWindow);
	    return;
	} else {		/* VW->mainWindow becoming the front window */
	    VW->active = 1;
	    if (VW->lineCount <= VW->rows) /* whole file in one screen */
		hilite_v = 255;	/* scrollbar inactive */
	    else
		hilite_v = 0;	/* scrollbar active */
	    if (VW->maxWidth <= VW->columns)
		hilite_h = 255;	/* scrollbar inactive */
	    else
		hilite_h = 0;	/* scrollbar active */
	}
    } else {			/* de-activate event */
	VW->active = 0;
	hilite_v = hilite_h = 255; /* scrollbar inactive */
	if (newcopy)
	    writescrap();
    }
    HiliteControl(VW->hscrollbar, hilite_h);
    HiliteControl(VW->vscrollbar, hilite_v);
    ClipRect(&VW->clipRect);
    SetPort(whichport);
}

/*
**  menu event handler
*/

void
VWmenu(EventRecord *ep)
{
	int epmsg = ep->message;		/* save it in case we do more events */
	unsigned int menuid = ep->message >>16;
	unsigned int menuitem = ep->message & 0xFFFF;
	register ViewerWindow *VW;

	for (VW=VWroot; VW; VW = VW->next)	/* look for one of my windows */
		if (VW->mainWindow == ep->who)
			break;
	if (! VW)
		return;							/* not my window, no action */

	SearchVW = VW;
    BeginMenu(epmsg);
	switch (menuid)
	{
	case FINDMENU:
		switch (menuitem)
		{
		int direction;
		case SEARCHITEM:									/* search... */
			if (getSearchArg(&direction))
			{
				hiLite(VW, VW->hiLiteLine, 0); /* un-hilite previous, if any */
				VW->hiLiteLine = -1;				/* new search */
				searchAndDisplay(VW, direction, VW->regexbuf);
			}
			break;
		case NEXTITEM:
			searchAndDisplay(VW, 1, VW->regexbuf);
			break;
		case PREVITEM:
			searchAndDisplay(VW, -1, VW->regexbuf);
			break;
		}
		break;
	case TABMENU:
		CheckItem(TABMENU, (VW->tabStop / 4) + 1, 0);
		VW->tabStop = 4 * (menuitem - 1);
		CheckItem(TABMENU, (VW->tabStop / 4) + 1, 1);
		VW->screenNext = VW->screenFirst; 
		drawWindow(VW, TRUE);
		break;
	default:
		break;
	}
    EndMenu(epmsg);
}

static int
getSearchArg(int *direction)
{
	static char searchargbuf[256] = "";
	DMDcb *cbptr;
	extern DMItem moDfind[];
	int group, curport;
	Rect dlgwinrect;
	int retval = FALSE;

	SetRect(&dlgwinrect, SearchVW->xpos + 6, SearchVW->ypos + 6,
			SearchVW->xpos + 6 + 480, SearchVW->ypos + 6 + 80);
	GetPort(&curport);
	dlgWindow = NewWindow(&dlgwinrect, "mo: Search\311", showWindow, 
		dBoxProc, frontWindow, noOptions, SearchVW->grafFd);
	cbptr = DMNew(moDfind, dlgWindow, (void *)0);
	dialogUp = cbptr;
	if (cbptr == NULL)
	{
		char msg[] = "Can't open dialog.\n";
		write(1,msg,strlen(msg));
		exit(1);
	}
	DMSetText(cbptr, MODFIND_searcharg, searchargbuf);
	TESetSelect(0, 31000, DMGetId(cbptr, MODFIND_searcharg));
	for(;;)
	{
		group = DMCollect(cbptr);
		if (group == DMcancel)
			break;
		DMGet(cbptr, MODFIND_searcharg, searchargbuf);
		if (group == 253)		/* backward */
			*direction = -1;
		else
			*direction = 1;
		retval = TRUE;
		break;
	}
	dialogUp = NULL;
	DMDispose(cbptr);
	DisposeWindow(dlgWindow);
	dlgWindow = -1;
	
	SetPort(curport);
	SelectWindow(SearchVW->mainWindow);
	return retval;
}

int 
ValSearchArg(DMDcb *cbp, DMItem *dip, unsigned char when, DMTEVal *arg)
{
	if (searchType == ST_REGEX) {
		if (compile(arg->text, SearchVW->regexbuf, 
					SearchVW->regexbuf + sizeof(SearchVW->regexbuf), '\0'))
			return DM_PASS;
	} else {
		strcpy(SearchVW->regexbuf, arg->text);
		return DM_PASS;
	}
	return DM_FAIL;
}

int
fastSearch(ViewerWindow *VW, char *arg, int arglen, int point)
{
	int len, left, mid, right, offset;
	char *found = 0, *from = (char *)VW->textBuff + VW->lines[point];
	
	while (!found) {
		len = (char *)VW->textBuff + VW->lines[VW->lineCount + 1] - from;
		if (searchType == ST_LITERAL)
			found = FastFind(from,arg,len,arglen);
		else
			found = FastCIFind(from,arg,len,arglen);
		if (VW->inputEOF)
			break;
		VWnoevent(VW);		/* more file to look in */
		from += len;
	}
	if (!found)
		return -1;
	/* found points to character at start of match */
	offset = found - (char *)VW->textBuff;
	left = point;
	right = VW->lineCount;
	for (;;) {					/* search to find line number from offset */
		mid = (right + left) >>1;
		if (offset < VW->lines[mid])
			right = mid;
		else
			if (offset < VW->lines[mid + 1])
				break;
			else
				left = mid + 1;
	}
	return mid;
}
		
int
VWSearch (register ViewerWindow *VW, register char *arg, 
		  register int direction)
{
	register int point;	
	int rv, arglen=strlen(arg);
	char myline[256];
	char *in, *out, *end = myline+sizeof(myline);

	/* see if last found is now on the display */
	if (VW->hiLiteLine >= VW->screenFirst  && VW->hiLiteLine < VW->screenNext)
		point = VW->hiLiteLine + direction;		/* start from last found */
	else
		/* start from top:bottom of current display */
		point = VW->screenFirst + ((direction > 0) ? 0 : VW->rows);

	SetMouseCursor(MC_WATCH);
	if (direction > 0 && (searchType == ST_LITERAL || searchType == ST_CASEIN)){
		rv = point = fastSearch(VW, arg, arglen, point);
	} else {
		while (point <= VW->lineCount && point >= 0) {
			if (! VW->inputEOF)
				VWnoevent(VW);
			in = (char *)VW->textBuff + VW->lines[point];
			out=myline;
			while (*in != 0 && *in != '\n' && out < end)
				*out++ = *in++;
			*out = '\0';
			switch (searchType) {
			case ST_REGEX:
				rv = step(myline, arg);
				break;
			case ST_LITERAL:
				rv = (int)FastFind(myline, arg,
						   StringLen(myline), arglen);
				break;
			case ST_CASEIN:
				rv = (int)FastCIFind(myline, arg,
						    StringLen(myline), arglen);
				break;
			}
			if (rv) {
				SetMouseCursor(MC_ARROW);
				break;
			} else {
			    rv = -1;
			}
			point += direction;
		}
	}
	SetMouseCursor(MC_ARROW);
	if (rv >= 0)
		return point;
	return -1;
}
	
static void
searchAndDisplay(register ViewerWindow *VW, int direction, char *searcharg)
{
	int point;
	char mycopy[256];
	char *in, *out = mycopy, *max = mycopy+sizeof(mycopy);

	if (*searcharg == 0x7F)
	{
		Bell();
		return;
	}
	
	if ((point = VWSearch(VW, searcharg, direction)) >= 0)
	{
		hiLite(VW, VW->hiLiteLine, 0);
		EnableItem(FINDMENU, NEXTITEM);
		EnableItem(FINDMENU, PREVITEM);
		VW->hiLiteLine = point;

		if (point < VW->screenFirst || point >= VW->screenNext) {
			VW->screenNext = point - (VW->rows / 2);
			drawWindow(VW, FALSE);
		} else {
			hiLite(VW, VW->hiLiteLine, 1);
		}
	}
	else
	{
		Bell();
	}
}

/*
**  "Close" a viewer window - kill the window, de-allocate the
**  structure and associated data, forget about it.
*/
BOOL 
VWClose (ViewerWindow *VW)
{
	ViewerWindow *search, *prev;

	if (dialogUp || !VW || VW->VWMagic != VWMAGIC)
		return FALSE;
	DisposeWindow(VW->mainWindow);
	if (VW->lines) free((char *)VW->lines);

	prev = (ViewerWindow *)VWroot;
	for (search = VWroot; search; search = search->next)
	{
		if (search == VW)
			break;
		prev = search;
	}
	if (search)
	{
		prev->next = search->next;		/* unlink struct */
	}
	VW->VWMagic = 0;					/* flag as invalid struct */
	free((char *)VW);
	return TRUE;
}

/*
**  "Validate" radiobutton for search/find type
*/
int
Vbut(DMDcb *cbp, DMItem *dip, unsigned char when, short value, short partcode)
{
	searchType = dip->usercode;
	return DM_PASS;
}

/*
**  Prefill radiobuttons for search/find type
*/
int
Pbut(DMDcb *cbp, DMItem *dip, int *value)
{
	if (dip->usercode == searchType) {
		*value = 1;
		return 1;
	} else {
		*value = 0;
		return 0;
	}
}
