/**********************************************************************
*
*        Structure and function definitions for viewer library
*  ----------------------------------------------------------------
*
*                     ISC Systems Corporation
*                   Confidential / Proprietary
*      Copyright (C) 1990 by ISC-Bunker Ramo / Unpublished Work
*       For Use by ISC-Bunker Ramo and Authorized Users Only
*                      All rights reserved.
*
*  Author:  Carl Paukstis       August, 1990
*
*  Notes:
*
**********************************************************************/

#ifndef _MOVIEW_H			/* multiple-inclusion interlock */
#define _MOVIEW_H

#ifndef FALSE
#define TRUE 1
#define FALSE 0
#endif

typedef int BOOL;

#define VWMAGIC 0xDEADBEEF
#define FVMAGIC 0xBEADFEED

typedef struct VW {
	struct	VW *next;		/* link to next struct */
	int		grafFd;
	int		mainWindow;
	Rect	clipRect;		/* clipping rectangle */
	int		xpos;			/* x coordinate of window top left */
	int		ypos;			/* y coordinate of window top left */
	short	width;			/* current window width (pixels) */
	short	height;			/* current window height (pixels) */
	short	rows;			/* number of lines in window in current font */
	short	columns;		/* max number of columns in window, current font */
	short	font;			/* current font number */
	short	face;			/* current face */
	short	size;			/* current size */
	FontInfo fInfo;			/* info about the current font */
	short	fontHeight;		/* total height of current font (pixels) */
	short	fontWidth;		/* max width of current font (pixels) */
	BOOL	iconified;		/* is our window currently iconified? */
	BOOL	active;			/* window activated (in front)? */
	short	moveFirst;		/* where to MoveTo to draw first line */
	int		screenNext;		/* line number of next line to display */
	int		screenFirst;	/* line number of first line on screen */
	int		firstColumn;	/* first text column to display */
	int		tabStop;		/* width of a tab stop (chars) */
	int		hiLiteLine;		/* line number to be hilit when displayed */
	ControlID hscrollbar;	/* horizontal scroll bar */
	ControlID vscrollbar;	/* vertical scroll bar */
	char	regexbuf[256];	/* compiled search expression */
	int		kbstate;		/* state for keyboard event handler */
	int		inTrackControl;	/* are we currently doing a track? */
	unsigned char *	textBuff;		/* buffer used to hold input text */
	int		textBuffCount;	/* number of bytes of text in buffer + 1 */
	BOOL	inputEOF;		/* have we filled the buffer as far as possible? */
	BOOL	inputtruncated;	/* Input truncated (lack of memory)? */
	int		(*fillbuffer)(struct VW *);	/* function to fill buffer, or NULL */
	int *	lines;			/* array of offsets to text lines */
	int		linesSize;		/* size of lines array, incl. unused space */
	int		lineCount;		/* number of lines of text in buffer */
	BOOL	linesDone;		/* have we found/recorded all lines in buffer? */
	short	maxWidth;		/* length (bytes) of longest line */
	int		marks[26];		/* position markers set by user */
	short startrow;			/* Used to mark range for COPY. */
	short startcol;
	short endrow;
	short endcol;
	short lastrow, lastcol;	/* For remembering highlighting stages. */
	int		VWMagic;		/* magic number to verify valid struct */
} ViewerWindow;


typedef struct fview {		/* file-viewer definition struct */
	struct fview *next;			/* link to next struct */
	int			inputFd;		/* FD for open input file */
	unsigned 	chunksize;		/* size increment for data buffer */
	int 		buflen;			/* current allocated size of buffer */
	BOOL		inputEOF;		/* have we read the whole file? */
	ViewerWindow *VW;			/* associated viewer window */
	int			FVMagic;		/* magic number to verify valid struct */
	} FView;

FView *FVOpen( 				/* open file-viewer window */
			short rows,			/* number of text rows */
			short columns,		/* and columns */
			short x,			/* x coordinate of top left corner (pixels) */
			short y,			/* ... and y coordinate */
			short font,			/* font number */
			short size,			/* font size */
            short tabstop,		/* initial tab value */
			char *fname,		/* file pathname */
			int grafFd );		/* graphics connection */

BOOL FVClose(FView *FV);	/* close file-viewer window */

ViewerWindow *VWOpen (		/* open text-buffer-viewer window */
			short rows,			/* number of text rows */
			short columns,		/* number of text columns */
			short x,			/* x coordinate of upper left corner */
			short y,			/* ... and y coordinate */
			short font,			/* display font number */
		    short size,			/* font size */
            short tabstop,		/* initial tab stop setting */
			char *title,		/* window title */
			unsigned char *buffer,		/* text buffer to display */
			int buffsize,		/* number of bytes of text in the buffer */
			int grafFd,			/* graphics connection */
			int (* filler)(ViewerWindow *) );	/* RESERVED, pass a NULL */

void VWactivate(EventRecord *ep);	/* activate event handler */
void VWkeyboard(EventRecord *ep);	/* keyboard event handler */
void VWmenu(EventRecord *ep);		/* menu event handler */
void VWmouse(EventRecord *ep);		/* mouse event handler */
int  VWnoevent();					/* non-event handler */
void VWupdate(EventRecord *ep);		/* update event handler */
void VWwindow(EventRecord *ep);		/* window event handler */
int VWSearch (register ViewerWindow *VW, register char *arg, 
			  register int direction);

BOOL VWClose(ViewerWindow *VW);		/* close text-buffer-viewer window */

extern ViewerWindow *SearchVW;
extern int dlgWindow;			/* used for dialog box */
extern DMDcb *dialogUp;			/* if we have a dialog up, its cb is here */

#define NM_CHAR 0x80			/* "normal" mode start indicator */
#define BO_CHAR 0x81			/* "boldface" mode */
#define UL_CHAR 0x82			/* "underline" mode */
#define IT_CHAR 0x84			/* "italic" mode */
#define ALL_CHAR 0x87			/* All styles together. */

#endif /* _MOVIEW_H */
