/***********************************************************************
*
*                File Selector Dialog support routines
*
*                     ISC Systems Corporation
*                   Confidential / Proprietary
*              Copyright (C) 1991,1992 by ISC-Bunker Ramo
*       For Use by ISC-Bunker Ramo and Authorized Users Only
*                      All rights reserved.
*
*  Author:       Carl Paukstis    June, 1992
*  Stolen-from:  Cathy Higel      June 1991 (who stole it elsewhere)
*
*  Notes:  Support routines for dialog box to load a file
*
**********************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/times.h>
#include <dnix/fcodes.h>
#include <malloc.h>
#include <isc/window.h>
#include <isc/controls.h>
#include <isc/dlgman.h>
#include <isc/keycodes.h>
#include <isc/cd_clunkdraw.h>
#include <isc/textedit.h>
#include <mousecur.h>

#define MAX_FILENAME (256)
#define MAX_DIR (32)
#define MAX_FILE (512)

#define	DIALNEW		240

static int namecomp(a, b);
static void FreeLists(void);
static char pathname[MAX_FILE + 1];
static int dir_count;
static char Filename[MAX_FILENAME];
static char directory[MAX_FILENAME];
static char *dir[MAX_DIR];
static char *file[MAX_FILE];
static struct stat filestat;
static int (*Validfile)(char *);

int
SelectFile (int window, int hoff, int voff, char *deflt, char *result, DMDcb **dialog, int val())
{
	DMDcb *save = *dialog;
	DMDcb *cbptr;
	extern DMItem Dselfile[];
	int retval = 1;
	int group;

	strcpy(Filename,deflt);
	Validfile = val;
	cbptr = DMNewOffset(Dselfile, window, (void *)0, voff, hoff);
	DMTEActivate(cbptr, "fullname");
	*dialog = cbptr;
	if (cbptr == NULL)	{
		char msg[] = "Can't open dialog.\n";
		write(1,msg,strlen(msg));
		exit(1);
	}
	DMSetText(cbptr, "fullname", Filename);
	TESetSelect(0, 31000, DMGetId(cbptr, "fullname"));
	group = DMCollect(cbptr);
	if (group == DMcancel) {
		retval = 0;
	} else {
		DMGet(cbptr, "fullname", Filename);
		strcpy(result, Filename);
		retval = DMok;
	}

	*dialog = save;
	DMDispose(cbptr);
	FlushWindow();
	FreeLists();
	return retval;
}
/*
** Version of "access()" which works with the _effective_, rather
** than _real_ userid.
*/
int
myaccess(char *path, int mode)
{
	int euid, egid;
	int abits;
	struct stat sbuf;

	if (stat(path, &sbuf))
		return -1;
	euid=geteuid();
	egid=getegid();
	abits = sbuf.st_mode;		/* file mode; perms in low-order 9 bits */
	if (euid == sbuf.st_uid)
		abits >>= 6;			/* we own it */
	else if (egid == sbuf.st_gid)
		abits >>= 3;			/* we're in the group */
	abits &= 07;				/* low-order 3 bits */
	if (abits & mode)
		return 0;
	errno = EACCES;
	return -1;
}

/***********************************************************************
**							Dialog PreFills
***********************************************************************/

PFdirectories(DMDcb *dialog, DMItem *item, DMPick *pick)
{
	char *cp;
	int  namelength = 0;

	strcpy(directory, Filename);	/* look for a directory slash */
	cp = strrchr(directory, '/');
	if (cp) {
		*(++cp) = 0;
	} else {
		if (pathname[0] != '\0') {	/* restore previous path used */
			namelength = strlen(pathname);
			while (namelength > 0) {
				if (pathname[namelength] == '/')
					break;
				namelength--;
			}
			strncpy(directory, pathname, namelength);
		} else {
			getcwd(directory, MAX_FILENAME); /* get current directory */
			strcat(directory, "/");
		}
	}

	/*
	** gen the list of directories to get to our current location
	*/
	dir_count = Build_Dir_List();
	pick->length = dir_count;
	pick->arrayhead = &dir[0];
	pick->selitem = dir_count - 1;
	pick->topitem = (pick->selitem > 6) ? pick->selitem - 6 : 1;
}

PFfiles(DMDcb *dialog, DMItem *item, DMPick *pick)
{
	char *cp;
	int findex;

	SetMouseCursor(MC_WATCH);

	/*
	** gen the list of files in the current directory
	*/
	pick->length = Build_File_List();
	pick->arrayhead = &file[0];

	/*
	** Look for our current file, so we can highlight it
	*/
	cp = strrchr(Filename, '/');
	if (cp)
		++cp;
	else
		cp = Filename;

	for (findex = 0; findex < pick->length; ++findex)
		if (strcmp(file[findex], cp) == 0) {
			pick->selitem = findex;
			pick->topitem = (pick->selitem > 2) ? pick->selitem - 3 : 1;
			break;
		}
	SetMouseCursor(MC_ARROW);
}


PFfullname(DMDcb *dialog, DMItem *item, char *cp)
{
	strcpy(cp, pathname);	/* restore previous path used */
}

/***********************************************************************
**					Dialog Validation routines
***********************************************************************/

VALdirectories(DMDcb *dialog, DMItem *item, int when, short indx, short partcode)
{
	DMPick list;
	short i;
	char tempname[MAX_FILENAME];

	if (partcode == inList) {

		/* build the new directory head	*/
		tempname[0] = 0;
		for (i = 0; i <= indx; i++)
			strcat(tempname, dir[i]);

		/* if it's different then the current head */
		if (strcmp(directory, tempname)) {
			SetMouseCursor(MC_WATCH);
			strcpy(directory, tempname);
			dir_count = Build_Dir_List();
			list.length = dir_count;
			list.arrayhead = &dir[0];
			DMChangePick(dialog, "directories", &list);
			DMSetPickItem(dialog, "directories", dir_count-1);
			list.length = Build_File_List();
			list.arrayhead = &file[0];
			DMChangePick(dialog, "files", &list);
			DMSetText(dialog, "fullname",tempname);
			SetMouseCursor(MC_ARROW);
		}
	}
	return DM_PASS;
}

VALfiles(DMDcb *dialog, DMItem *item, int when, short indx, short partcode)
{
	DMPick list;
	int status = DM_PASS;
	char tempname[MAX_FILENAME];

	if (partcode == inList || partcode == inListDbl) {

		if (strchr(file[indx], '/')) {	/* did we select a directory */
			strcpy(tempname, directory);
			strcat(tempname, file[indx]);
			if (myaccess(tempname, 05)) {
				status = DM_FAIL;
			} else {
				SetMouseCursor(MC_WATCH);
				strcat(directory, file[indx]);
				dir_count = Build_Dir_List();
				list.length = dir_count;
				list.arrayhead = &dir[0];
				DMChangePick(dialog, "directories", &list);
				DMSetPickItem(dialog, "directories", dir_count-1);
				list.length = Build_File_List();
				list.arrayhead = &file[0];
				DMChangePick(dialog, "files", &list);
				DMSetText(dialog, "fullname",tempname);
				SetMouseCursor(MC_ARROW);
			}
		} else {	/* we selected a file */
			strcpy(tempname, directory);
			strcat(tempname, file[indx]);
			DMSetText(dialog, "fullname",tempname);
			FlushWindow();
			status = DM_PASS;
			if (partcode == inListDbl)
				return DMok;
		}
	}		
	return status;
}

VALload(DMDcb *dialog, DMItem *item, int when, int value)
{
	ControlID cid;
	DMPick list;
	int i;
	struct stat statbuf;
	char buffer[MAX_FILENAME], *cp;

	DMGet(dialog, "fullname", buffer);	/* get anything in the fullname */

	if (!strlen(buffer)) 
		return DMok;


	/* is it a file */
	if (stat(buffer, &statbuf)) 
		return (*Validfile)(buffer);
	if (S_ISREG(statbuf.st_mode)) 
		return (*Validfile)(buffer);

	/*
	** if its a directory, build a new picklist
	*/
	if (!S_ISDIR(statbuf.st_mode)) return DM_FAIL;
	if (myaccess(buffer, 05)) return DM_FAIL;
	for (cp = buffer; *cp; cp++) ;
	if ( *(cp-1) != '/' ) {
		*(cp++) = '/';
		*cp = 0;
	}

	/*
	** switch to the new directory
	*/
	SetMouseCursor(MC_WATCH);
	strcpy(directory, buffer);
	dir_count = Build_Dir_List();
	list.length = dir_count;
	list.arrayhead = &dir[0];
	DMChangePick(dialog, "directories", &list);
	DMSetPickItem(dialog, "directories", dir_count-1);
	list.length = Build_File_List();
	list.arrayhead = &file[0];
	DMChangePick(dialog, "files", &list);
	SetMouseCursor(MC_ARROW);
	return DM_PASS;
}

/***********************************************************************
** build a list of files,
** returning the count of files found.
***********************************************************************/
Build_File_List()
{
	register DIR *dp;
	register struct dirent *dep;
	register int count = 0;
	struct stat statbuf;
	char tempname[1024];

	/*
	** open the directory
	*/
	dp = opendir(directory);
	if ( !dp ) return 0;

	/*
	** build the picklist
	*/
	while ((dep = readdir(dp))) {
		/*
		** skip some obvious junk
		*/
		if (dep->d_name[0] == '.') continue;

		/*
		** try to stat the item
		*/
		strcpy(tempname, directory);
		strcat(tempname, dep->d_name);
		if (stat(tempname, &statbuf)) continue;

		/*
		** it must be a file or a directory
		*/
		if (!S_ISREG(statbuf.st_mode) && !S_ISDIR(statbuf.st_mode)) continue;

		/*
		** try to add it to the list
		*/
		if (count < MAX_FILE) {
			if (!file[count]) {
				file[count] = malloc(16);
				if (!file[count]) break;
			}
			strcpy(file[count], dep->d_name);
			if (S_ISDIR(statbuf.st_mode)) strcat(file[count], "/");
			count++;
		}
	}

	/*
	** cleanup
	*/
	closedir(dp);
	if ( count ) shellsort(&file[0], count, sizeof(char *), namecomp);
	return count;
}

Build_Dir_List()
{
	register int count = 0;
	register char *src = directory, *dst;

	/*
	** pick the directory apart
	*/
	while (*src && count < MAX_DIR)	{
		/*
		** get us some storage if needed
		*/
		if (!dir[count]) {
			dir[count] = malloc(16);
			if (!dir[count]) break;
		}

		/*
		** copy out the next directory
		*/
		dst = dir[count];
		do {
			*dst++ = *src;
		} while (*src && *(src++) != '/');
		*dst = 0;
		count++;
	}
	return count;
}

/***********************************************************************
** Shell sort with the calling sequence of qsort.
***********************************************************************/
#define A(v, i)  ((v) + ((i) * size))

shellsort(v, n, size, compar)
char *v;
int n, size;
int (*compar)();
{
	register int j;
	int gap, i;

	for (gap = n/2; gap > 0; gap /= 2)
		for (i = gap; i < n; i++)
			for (j = i-gap; j >= 0; j -= gap) {
				if ((*compar)(A(v,j), A(v,j+gap)) <= 0) break;
				qexch(A(v,j), A(v,j+gap), size);
			}
}

static qexch(char *p1, char *p2, int n)
{
	register char save;

	while ( n-- ) {
		save = *p1;
		*p1++ = *p2;
		*p2++ = save;
	}
}

/***********************************************************************
** compare two names
***********************************************************************/
static int
namecomp(a, b)
char **a, **b;
{
	return(strcmp(*a, *b));
}

static void
FreeLists(void)
{
	register int i;

	for (i = 0; i < MAX_DIR; i++) {
		if ( dir[i] ) free(dir[i]);
		dir[i] = NULL;
	}
	for (i = 0; i < MAX_FILE; i++) {
		if ( file[i] ) free(file[i]);
		file[i] = NULL;
	}
}

