/*
**  Test program for viewer library
*/

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <isc/window.h>
#include <isc/controls.h>
#include <isc/keycodes.h>
#include "viewer.h"

static ViewerWindow *VW;
static int mywindow;
static Rect myrect;
static char *fname;
static int inputFd;
static char *buf;
static struct stat st;
static int	grafFd;
static char **Argv;

static int mykeyboard(EventRecord *);
static int myupdate(EventRecord *);
int
main (int argc, char *argv[])
{
	char *Myname;
	static int trapqfd;
	Argv = argv;
	
	Myname = strrchr(argv[0],'/');
	if (Myname == NULL)
		Myname = argv[0];
	else
		Myname++;
	
 	/* initialize the graphics toolbox */
    if ((trapqfd = InitGraf()) < 0) 
	{
		char *msg = ": Error from InitGraf\n";
		write(2,Myname, strlen(Myname));
		write(2,msg,strlen(msg));
		return 1;
    }
    if ((grafFd = ConnectGraf((char *)0)) < 0) 
	{
		char *msg = ": Error from ConnectGraf\n";
		write(2,Myname, strlen(Myname));
		write(2,msg,strlen(msg));
		return 1;
    }
    SetRect(&myrect, 50, 50, 500, 300);
    mywindow = NewWindow(&myrect, "", showWindow, plainDBox, 
		frontWindow, 0, grafFd);

	/* open some input file */
	fname = argv[1];
	inputFd = open(fname, 0, 0);
	if (inputFd < 0)			/* they gave us none, or a bad one */
	{
		fprintf(stderr,"can't open '%s'\n",fname);
		exit(1);
	}
	if (fstat(inputFd,&st) < 0)
	{
		fprintf(stderr, "can't fstat\n");
		exit(1);
	}
	if ((buf = malloc(st.st_size)) == NULL)
	{
		fprintf(stderr, "can't malloc\n");
		exit(1);
	}
	if (read(inputFd, buf, st.st_size) != st.st_size)
	{
		fprintf(stderr, "Can't read\n");
		exit(1);
	}
	close(inputFd);
	if ((VW=VWOpen(40,80,32,48,5,fname,buf,st.st_size,grafFd,NULL)) == NULL)
	{
		fprintf(stderr, "can't VWOpen\n");
		exit(1);
	}
    Event(updateEvt, myupdate);
    Event(mouseEvt, VWmouse);
    Event(windowEvt, VWwindow);
    Event(menuEvt, VWmenu);
    Event(keyEvt, mykeyboard);
	Event(activateEvt, VWactivate);
	Event(noEvent, VWnoevent);
    (void) GetEvent(anyWindow, noEvent);
}

static int myupdate(EventRecord *ep)
{
	if (ep->who == mywindow)
	{
		BeginUpdate(ep->who);
		SetPort(mywindow);
		EraseRect(&myrect);
		TextFont(9);
		TextFace(normalText);
		TextSize(36);
		MoveTo(0,100);
		DrawText("Foo Bar Bozo",0,12);
		EndUpdate(ep->who);
	}
	else
		VWupdate(ep);
}

static int mykeyboard(EventRecord *ep)
{
	unsigned int key;

	if (ep->message & F_KEYUP)
		return 0;
	key = ep->message & ~F_KEYDOWN & ~F_KBDIII;

	if (key == 'a')
	{
		if (!VWClose(VW))
		{
			fprintf(stderr, "Can't VWClose\n");
			exit(1);
		}
		/* reopen the input file */
		inputFd = open(fname, 0, 0);
		if (inputFd < 0)			/* they gave us none, or a bad one */
		{
			fprintf(stderr,"can't open '%s'\n",fname);
			exit(1);
		}
		if (fstat(inputFd,&st) < 0)
		{
			fprintf(stderr, "can't fstat\n");
			exit(1);
		}
		if ((buf = malloc(st.st_size)) == NULL)
		{
			fprintf(stderr, "can't malloc\n");
			exit(1);
		}
		if (read(inputFd, buf, st.st_size) != st.st_size)
		{
			fprintf(stderr, "Can't read\n");
			exit(1);
		}
		close(inputFd);
		if ((VW=VWOpen(40,80,32,48,5,fname,buf,st.st_size,grafFd,NULL)) == NULL)
		{
			fprintf(stderr, "can't VWOpen\n");
			exit(1);
		}
		return;
	}
	if (key == 'n' && Argv[2])
	{
		/* open the second input file */
		inputFd = open(Argv[2], 0, 0);
		if (inputFd < 0)			/* they gave us none, or a bad one */
		{
			fprintf(stderr,"can't open '%s'\n",Argv[2]);
			exit(1);
		}
		if (fstat(inputFd,&st) < 0)
		{
			fprintf(stderr, "can't fstat\n");
			exit(1);
		}
		if ((buf = malloc(st.st_size)) == NULL)
		{
			fprintf(stderr, "can't malloc\n");
			exit(1);
		}
		if (read(inputFd, buf, st.st_size) != st.st_size)
		{
			fprintf(stderr, "Can't read\n");
			exit(1);
		}
		close(inputFd);
		if ((VW=VWOpen(40,80,32,48,5,Argv[2],buf,st.st_size,grafFd,NULL)) == NULL)
		{
			fprintf(stderr, "can't VWOpen\n");
			exit(1);
		}
		return;
	}
	if (key == (F_SHIFT | 'Q'))
		exit(0);
	VWkeyboard(ep);
}
