/*
** Program to spew junk onto the network, to eat up bandwidth.
**
** Usage:  blower [-b] [-n nwrites] [-N totalwrites] [-s packsize] [-V]
**
** Version 1.0		12/14/92	Jim Cathey
**	Original version.
*/

#include <malloc.h>
#include <signal.h>
#include <dnix/errno.h>
#include <dnix/fcodes.h>
#include <dnix/itrq.h>
#include <dnix/cheapdrv.h>

#define EBUFSIZE 2062		/* One-at-a-time write buffer size. */
#define NWRITES 4		/* Number of these writes up at once. */

struct ebuf {
    union {
	unsigned short packetsize;
	void (*iodone)();	/* Indirect completion routine. */
    } u;
    int reqnum;
    unsigned char data[EBUFSIZE];
};

struct ebuf *ebufs;
int etherfd, trapqfd;
struct itrq trapnode;
void writedone();
char *mysprintf();
char *progname;
int nextbuf;
int nwrites;			/* Number of simultaneous writes up. */
int total;			/* Number of total writes to do. */
int packnum;			/* Packet number written. */
int writesup;			/* Number of writes currently posted. */
int packsize;			/* Size of packets to write. */
int broads;			/* Broadcast noise. */

main(argc, argv)
int argc;
char *argv[];
{
    int i;
    long bufsize;
    char scratch[256];
    char *argp;
    register void (**procptr)(void *struc, int retparm, int reqnum);
    void *calloc();

    nwrites = NWRITES;
    total = -1;			/* Forever. */
    packsize = 1024;
    progname = *argv;
    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'b':		/* Broadcast packets. */
		broads = 1;
		break;

	      case 'n':		/* Number of concurrent writes. */
		if (!*argp)
		    --argc, argp = *++argv;
		i = myatoi(argp);
		if (i >= 1 && i <= 10)
		    nwrites = i;
		break;

	      case 'N':		/* Number of total writes. */
		if (!*argp)
		    --argc, argp = *++argv;
		total = myatoi(argp);
		break;

	      case 's':		/* Size of writes. */
		if (!*argp)
		    --argc, argp = *++argv;
		i = myatoi(argp);
		if (i >= sizeof(struct etherheader) && i <= EBUFSIZE)
		    packsize = i;
		break;

	      case 'V':		/* Version printout. */
		strcpy(scratch, progname);
		strcat(scratch, " Version 1.0\n");
		write(1, scratch, strlen(scratch));
		exit(0);
		break;
	    }
	} else {
	    ;
	}
    }
    ebufs = (struct ebuf *) calloc(sizeof(struct ebuf), nwrites);
    if (!ebufs) {
      nomemory:
	mysprintf(scratch, "%s: Unable to allocate buffers.\n", progname);
	(void) write(2, scratch, strlen(scratch));
	exit(2);
    }
    i = dnix(F_LOCKMEM, ebufs, nwrites * sizeof(struct ebuf));
    if (i < 0) {
	mysprintf(scratch, "%s: Unable to lock down I/O buffer memory (%d).\n",
		  progname, i);
	(void) write(2, scratch, strlen(scratch));
	exit(3);
    }
    etherfd = open("/dev/sniffer", 2);
    if (etherfd < 0) {
	mysprintf(scratch, "%s: Unable to open /dev/sniffer.\n", progname);
	(void) write(2, scratch, strlen(scratch));
	exit(4);
    }
				/* We can only do I/O from locked bufs now.  */
    if ((i = dnix(F_READ, etherfd,
		  ((struct etherheader *)ebufs->data)->dstaddr,
		  sizeof(struct etheraddr), CHEAPF_OWNADDRESS, 0)) <= 0) {
	exit(5);
    }
    if (broads) {		/* Broadcast address? */
	for (i=0; i < sizeof(struct etheraddr); i++)
	    ((struct etherheader *)ebufs->data)->dstaddr[i] = ~0;
    }
    if ((i = dnix(F_READ, etherfd,
		  ((struct etherheader *)ebufs->data)->srcaddr,
		  sizeof(struct etheraddr), CHEAPF_OWNADDRESS, 0)) <= 0) {
	exit(5);
    }
    if ((i = dnix(F_READ, etherfd,
		  &((struct etherheader *)ebufs->data)->protocol,
		  sizeof(short), CHEAPF_PROTOCOL, 0)) <= 0) {
	exit(5);
    }
    for (i=1; i<nwrites; i++)
	memcpy(ebufs[i].data, ebufs[0].data, sizeof(struct etherheader));
    trapqfd = dnix(F_OTQ);
    if (trapqfd < 0) {
	mysprintf(scratch, "%s: Unable to open trap queue.\n", progname);
	(void) write(2, scratch, strlen(scratch));
	exit(6);
    }

    i = nwrites;		/* How many no-wait writes to post? */
    while (i--) {		/* Post them. */
	postwrite();
    }
    while (1) {			/* Wait for event completions. */
	if (read(trapqfd, &trapnode, sizeof(trapnode))
	    == sizeof(trapnode)) {
	    procptr = (void (**) ()) trapnode.it_oid;
	    (**procptr)(procptr, trapnode.it_rpar, trapnode.it_rn);
	}
    }
}

postwrite()
{
    int reqnum;

    ebufs[nextbuf].u.iodone = writedone;
    *(long *) &ebufs[nextbuf].data[16] = packnum++;
    reqnum = ebufs[nextbuf].reqnum = 
	(total) ? 
	    dnix(F_WRITE | F_NOWAIT, etherfd, ebufs[nextbuf].data, packsize,
		 0, 0, trapqfd, &ebufs[nextbuf]) : -1;
    if (total > 0)
	total--;
    if (reqnum >= 0)
	writesup++;
    nextbuf++;
    if (nextbuf >= nwrites)
	nextbuf = 0;
}

/*
** Called when packet(s) are written.
*/

void
writedone(peb, it_rpar, it_rn)
struct ebuf *peb;
int it_rpar, it_rn;
{
    char scratch[256];

    writesup--;
    if (it_rpar >= 0) {
	peb->reqnum = 0;
	if (total)
	    postwrite();
    } else if (it_rpar != ECANCL) {
	mysprintf(scratch, "%s: Error %d.\n", progname, it_rpar);
	(void) write(2, scratch, strlen(scratch));
	exit(7);
    }
    if (!writesup)
	exit(0);
}

/*
** Sprintf replacement.
**
** Handles format strings containing:
**
**   %[1..9]s, %[0][1..9]d, %[0][1..9]x, %[0][1..9]X
**
** It's much smaller than the standard one.  Returns pointer to the NULL.
*/

char *
mysprintf(buffer, fmt, arglist)
char *buffer, *fmt;
char *arglist;
{
    register char *p, *q, *p2;
    short ndigits;
    int leadfill;
    char *itoa();
    union arg {
	long lng;
	char *ptr;
    } *argptr;

    argptr = (union arg *) &arglist;
    p = fmt;
    q = buffer;
    while (*p) {            /* While there's some format string. */
	if (*p == '%') {    /* Is it a format control? */
	    p++;            /* Yes, which one (skip over % char)? */
	    if (*p == '0')
		leadfill = *p++;
	    else
		leadfill = ' ';
	    if (*p >= '1' && *p <= '9')	/* Format output length? */
		ndigits = *p++ - '0';
	    else
		ndigits = 0;
	    if (*p == 's') {
		p2 = argptr++->ptr; /* String. */
	    } else if (*p == 'd') {
		p2 = itoa(argptr++->lng, 10, leadfill, ndigits);
	    } else if (*p == 'x' || *p == 'X') {
		p2 = itoa(argptr++->lng, 16, leadfill, ndigits);
	    }
	    p++;		/* Skip over format control. */
	    if (ndigits)
		while (*p2 && ndigits--) /* Copy the inclusion. */
		    *q++ = *p2++;
	    else
		while (*p2)	/* Copy the inclusion. */
		    *q++ = *p2++;
	} else			/* No, it's a literal -- copy it. */
	    *q++ = *p++;
    }
    *q = 0;			/* Null-terminate it. */
    return q;
}

char *
itoa(val, radix, leadfill, ndigits)
int val, radix, leadfill, ndigits;
{
    short i;
    char *p, *p2, out;
    static char buf[20];
    unsigned long test, remainder;

    p = buf;
    remainder = val;
    i = ndigits;		/* Make sure there're at least the number */
    while (i--)			/* of leading whatevers that we've asked for */
	*p++ = leadfill;	/* as the field width. */
    p2 = p;
    if (radix == 16) {		/* Hex output? */
	test = 0x10000000L;
    } else {
	test = 1000000000L;	/* Decimal output? */
	if (val < 0) {
	    remainder = -val;
	    *p++ = '-';
	}
    }
    while (test > remainder)	/* Scale down to match number. */
	test /= radix;
    if (remainder)
	while (test) {
	    out = remainder / test;
	    *p++ = out + ((out >= 10) ? '7' : '0');
	    remainder %= test;
	    test /= radix;
	}
    else
	*p++ = '0';
    *p = 0;
    if (ndigits) {		/* A field width set? */
	if ((p2 + ndigits) >= p) /* Make sure we never truncate. */
	    return p - ndigits;
	else
	    return p2;
    }
    return buf;			/* No field width, return minimum number. */
}

/*
** Provide our own atoi() function because the standard one uses ctype
** and is overly large (for some uses).
*/

static
myatoi(p)			/* Parse out a number. */
register char *p;
{
    register int val;
    register char chr;
    
    val = 0;
    while (1) {
	chr = *p++ - '0';
	if (chr < 0 || chr > 9)
	    break;
	val = (val * 10) + chr;
    }
    return val;
}
