/*
** Structure definitions for the dmap packets.
*/

struct wsinfo {
    char wsname[8];		/* system name */
    long wsserial;		/* system serial number */
    char wslevel;
    char wsonline;
    long wsconnect;		/* up time */
    char wsrel[8];		/* dnix release level */
    char wsver[8];		/* dnix version number */
};

struct dmap {
    unsigned char dnpname[8];	/* system name */
    long dnpserial;		/* system serial number */
    long sendtime;		/* time packet was sent */
    long recvtime;		/* time packet was received */
    long boottime;		/* time system was booted */
    long loadavg[3];		/* load averages */
    char dnprel[9];		/* dnix release level */
    char dnpver[9];		/* dnix version number */
    struct wsinfo ws[1436 / sizeof(struct wsinfo)];
};

struct duser {
    char line[8];		/* users tty line */
    char user[8];		/* users name */
    long time;			/* users login time */
    long idle;			/* users idle time */
};

struct dwho {
    char protocol;		/* protocol version # */
    char packet;		/* packet type, see below */
    char pad[2];
    long sendtime;		/* time packet was sent */
    long recvtime;		/* time packet was received */
    char sysname[32];		/* system name */
    long loadave[3];		/* system load average */
    long boottime;		/* time the system was booted */
    struct duser we[1024 / sizeof(struct duser)];
};

struct dmapmsg {
    struct etheraddr dstaddr;
    struct etheraddr srcaddr;
    short protocol;
    short rectype;
    short framelen;
    union {
        struct dmap dm;
        struct dwho wh;
    } ms;
};

struct dmapmsg2 {		/* New dmap packet format. */
    struct etheraddr dstaddr;
    struct etheraddr srcaddr;
    short protocol;
    short rectype;
    unsigned char dnpname[8];
    long uptime;		/* Time system has been running (seconds). */
    long boottime;		/* Time system was booted. */
    long sendtime;		/* Time message was sent. */
    long dnpserial;		/* System serial number. */
    char dnprel[9];		/* DNIX release level. */
    char dnpver[9];		/* DNIX version number. */
    short filler;		/* To make minimal-sized ethernet packet. */
};

#define DMAPREC (1)		/* Values of rectype */
#define DWHOREC (2)
#define DMAPREC2 (3)		/* New DMAP (small) format packet. */
