/*
** Sniffer packet to Network General converter.
**
** Usage:
**	dsniff2ng [-V] [file]
**
** Where
**	-V	prints version number and exits,
**	file	Packets in this file, otherwise read from standard in.
*/

#include <stdio.h>
#include <errno.h>
#include <dnix/cheapdrv.h>
#include <dnix/fcodes.h>
#include "ngsniff.h"

struct packetform {
    struct sniffheader snh;
    struct etherheader eth;
    unsigned char data[8192];	/* Way bigger than typically needed. */
};

void die();
unsigned long dotime();
char *progname;

main(argc, argv)
int argc;
char *argv[];
{
    FILE *fp, *outfp;
    unsigned short packetsize, psize;
    struct packetform packet;
    char *filename, *argp;
    int packnum;
    unsigned long timestamp;
    Header_Record nghdr;
    Version_Record ngver;
    Frame_Record ngfrm;

    progname = *argv;
    filename = 0;
    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'V':		/* Version printout. */
		printf("%s: Version 1.0\n", progname);
		exit(0);
		break;

	      default:
		fprintf(stderr, "Usage: %s [-V] [packetfile]\n", progname);
		exit(1);
		break;
	    }
	} else {
	    filename = argp;
	}
    }
    if (filename)
	fp = fopen(filename, "r");
    else
	fp = fdopen(0, "r");
    if (!fp) {
	fprintf(stderr, "%s: Unable to read input file %s\n", progname,
		filename ? filename : "");
	exit(2);
    }
				/* Emit NG header crap */
    if (fwrite("TRSNIFF data    \032", 17, 1, stdout) != 1)
	die();
    nghdr.type = LIBI(REC_VERS);
    nghdr.length = LIBI(sizeof(ngver));
    nghdr.rsvd = LIBI(0);
    if (fwrite(&nghdr, sizeof(nghdr), 1, stdout) != 1)
	die();

    ngver.maj_vers = LIBI(1);
    ngver.min_vers = LIBI(0);
    ngver.date.value[0] = LIBI(0);
    ngver.date.value[1] = LIBI(0);
    ngver.type = REC_FRAME;
    ngver.network = N_ETHERNET;
    ngver.format = 1;
    ngver.timeunit = T_MICOM;
    ngver.rsvd[0] = LIBI(0);
    ngver.rsvd[1] = LIBI(0);
    ngver.rsvd[2] = LIBI(0);
    if (fwrite(&ngver, sizeof(ngver), 1, stdout) != 1)
	die();

    nghdr.type = LIBI(REC_FRAME);
    ngfrm.time_low = LIBI(0);
    ngfrm.time_mid = LIBI(0);
    ngfrm.time_high = 0;
    ngfrm.time_day = 0;
    ngfrm.frame_status = 0;
    ngfrm.flags = 0;
    ngfrm.true_size = LIBI(0);
    ngfrm.rsvd = LIBI(0);

    setvbuf(fp, 0, _IOFBF, 4096);
    packnum = -1;
    while (fread(&packetsize, sizeof packetsize, 1, fp) == 1) {
	packnum++;
	packet.snh.statusbits = 0;
	if (packetsize > sizeof(packet)) {
	    fprintf(stderr, "ERROR -- %d byte packet is too big!\n",
		    packetsize);
	    fseek(fp, packetsize, 1); /* Skip over it. */
	} else if (packetsize <
		   (sizeof(struct etherheader) + sizeof(struct sniffheader))) {
	    if (!packetsize)	/* It's a runt packet. */
		packetsize = sizeof packetsize;
	    if (fread(&packet, packetsize, 1, fp) != 1)
		break;
	    fprintf(stderr, "#%-4d  RUNT   %4d bytes: ", packnum,
		   packet.snh.length - sizeof(struct sniffheader));
	    {
		int i, j, k;
		unsigned char *bufp1;

		i = packet.snh.length - sizeof(struct sniffheader);
		bufp1 = (unsigned char *) &packet.eth;
		k = 0;
		for (j = 0; j < i; j++) { /* Display hex. */
		    if (!(j % 2) && (j != 0)) {
			putc(' ', stderr);
			k++;
		    }
		    fprintf(stderr, "%02X", *bufp1++);
		    k += 2;
		}
		fprintf(stderr, "%*s\n", 34 - k, "");
	    }
	} else {
	    if (fread(&packet, packetsize, 1, fp) != 1)
		break;
	    if (packet.snh.statusbits == 0xFF) { /* It's a format-override, */
		packnum--;	/* so skip it. */
		continue;
	    }
	    if (!packnum)	/* First packet? */
		dotime(&packet); /* Prime the timer logic. */
	    if (!(packet.snh.statusbits & PACKET_GOOD)) { /* Errored? */
		continue;	/* Skip it. */
	    }
	    timestamp = dotime(&packet);
	    ngfrm.time_low = LIBI(timestamp & 0xFFFF);
	    ngfrm.time_mid = LIBI(timestamp >> 16);
	    psize = packetsize - sizeof(struct sniffheader);
	    ngfrm.size = LIBI(psize);
	    nghdr.length = LIBI(sizeof(ngfrm) + psize);
	    if (fwrite(&nghdr, sizeof(nghdr), 1, stdout) != 1)
		die();
	    if (fwrite(&ngfrm, sizeof(ngfrm), 1, stdout) != 1)
		die();
	    if (fwrite(&packet.eth, psize, 1, stdout) != 1)
		die();
	}
    }
    nghdr.type = LIBI(REC_EOF);
    nghdr.length = LIBI(0);
    if (fwrite(&nghdr, sizeof(nghdr), 1, stdout) != 1)
	die();
    exit(0);
}

/*
** Convert timestamp to a Network General 1/2 microsecond timestamp.
*/

#define SCALE 100000000.0

unsigned long
dotime(packet)
struct packetform *packet;
{
    static unsigned long basetick;
    static double basetime;
    double abstime, hi, lo;
    unsigned short num, denom;

    num = packet->snh.fracticknum;
    denom = packet->snh.fractickdenom;
    if (!basetick)		/* Start ticker at 0 for dump. */
	basetick = packet->snh.tickstamp;
    abstime = SCALE * (packet->snh.tickstamp - basetick)
	/ packet->snh.tickrate;
    abstime += (SCALE * num / denom) / packet->snh.tickrate;
    if (!basetime)
	basetime = abstime;

    lo = modf(abstime/50.0, &hi);
    return (unsigned long) hi;
}

void
die()
{
    fprintf(stderr, "%s: Write Error %d on output\n", progname, errno);
    exit(3);
}
