/*
** Function to dump out an ARP protocol frame.
*/

#include <stdio.h>
#include <dnix/cheapdrv.h>
#include <sys/types.h>
#include "protos.h"

/*
 * Ethernet Address Resolution Protocol.
 *
 * See RFC 826 for protocol description.  Structure below is adapted
 * to resolving internet addresses.  Field names used correspond to 
 * RFC 826.
 */
struct	ether_arp {
	u_short	arp_hrd;	/* format of hardware address */
#define ARPHRD_ETHER 	1	/* ethernet hardware address */
	u_short	arp_pro;	/* format of proto. address (ETHERPUP_IPTYPE) */
	u_char	arp_hln;	/* length of hardware address (6) */
	u_char	arp_pln;	/* length of protocol address (4) */
	u_short	arp_op;
#define	ARPOP_REQUEST	1	/* request to resolve address */
#define	ARPOP_REPLY	2	/* response to previous request */
	u_char	arp_xsha[6];	/* sender hardware address */
	u_char	arp_xspa[4];	/* sender protocol address */
	u_char	arp_xtha[6];	/* target hardware address */
	u_char	arp_xtpa[4];	/* target protocol address */
};

char *arps[] = {0, "Request", "Reply",
		    "Reverse Request", "Reverse Reply",
};

dumparp(buf, len, rawdata)
void *buf;
int len;
int rawdata;			/* TRUE to dump data part of frame. */
{
    return dumparprarp(buf, len, rawdata, "ARP");
}

/*
** Dump a reverse-ARP protocol frame.
*/

dumprarp(buf, len, rawdata)
void *buf;
int len;
int rawdata;			/* TRUE to dump data part of frame. */
{
    return dumparprarp(buf, len, rawdata, "RARP");
}

static
dumparprarp(buf, len, rawdata, str)
void *buf;
int len;
int rawdata;			/* TRUE to dump data part of frame. */
char *str;
{
    char from[14], to[14];
    unsigned char *cp;
    struct ether_arp *arp;

    arp = (struct ether_arp *) buf;
    printf("\t%s\tOpcode: %s (%d)\n", str,
	   (arp->arp_op > sizeof(arps)/sizeof(arps[0]))
	   ? "??" : arps[arp->arp_op], arp->arp_op);	   
    makename(arp->arp_xsha, from);
    makename(arp->arp_xtha, to);
    printf("\t\tTo   %-13.13s (%d.%d.%d.%d)\n", to, arp->arp_xtpa[0],
	   arp->arp_xtpa[1], arp->arp_xtpa[2], arp->arp_xtpa[3]);
    printf("\t\tFrom %-13.13s (%d.%d.%d.%d)\n", from, arp->arp_xspa[0],
	   arp->arp_xspa[1], arp->arp_xspa[2], arp->arp_xspa[3]);
    cp = (unsigned char *) &arp[1];
    return cp - (unsigned char *) buf;
}
