/*
** Function to dump out an Appletalk protocol frame.
*/

#include <stdio.h>
#include <dnix/cheapdrv.h>
#include "atalk.h"
#include "protos.h"

/*
** CAP/UAB seems to use a LLAP header on the front of an ELAP packet.
** This is strange.
*/

dumpatalk(buf, len, rawdata)
void *buf;
int len;
int rawdata;			/* TRUE to dump data part of frame. */
{
    unsigned char *cp;
    struct etalkmsg *f;
    int ddplen;

    f = (struct etalkmsg *) buf;
    printf("\tELAP\tTo node $%02X from node $%02X, type $%02X ",
	   f->dstnode, f->srcnode, f->type);
    if (f->type >= 0x80) {	/* LAP control packet. */
	cp = ((unsigned char *) &f->type) + 1;
	switch (f->type) {
	  case 0x81:		/* LAP enquiry packet. */
	    printf("(lapENQ)\n");
	    break;
	  case 0x82:		/* LAP ack of an enquiry packet. */
	    printf("(lapACK)\n");
	    break;
	  case 0x84:		/* LAP rts. */
	    printf("(lapRTS)\n");
	    break;
	  case 0x85:		/* LAP cts. */
	    printf("(lapCTS)\n");
	    break;
	  default:		/* Illegal. */
	    printf("(UNKNOWN)\n");
	    break;
	}
    } else {			/* LAP data packet format. */
	switch (f->type) {
	  case 0x01:		/* LAP data (DDP short header). */
	    printf("(DDP Short)\n");
	    ddplen = ((f->ddp.s.len1 & 3) << 8) | f->ddp.s.len2;
	    printf("\tDDP\tLength %d, To socket $%02X from socket $%02X, "
		   "type $%02X ", ddplen, f->ddp.s.dsock,
		   f->ddp.s.ssock, f->ddp.s.type);
	    cp = dumpddp(&f->ddp.s.data, f->ddp.s.type,
			 ddplen - (&f->ddp.s.data - &f->ddp.s.len1),
			 rawdata);
	    break;

	  case 0x02:		/* LAP data (DDP extended header). */
	    printf("(DDP Extended)\n");
	    ddplen = ((f->ddp.l.len1 & 3) << 8) | f->ddp.l.len2;
	    printf("\tDDP\tLength %d, Hop count %d, Checksum $%04X\n",
		   ddplen, (f->ddp.l.len1 & 0x3C) >> 2,
		   f->ddp.l.chk1 << 8 | f->ddp.l.chk2);
	    printf("\t\tDest net %d, Src net %d, Dest Node $%02X, "
		   "Src Node $%02X\n",
		   f->ddp.l.dnet1 << 8 | f->ddp.l.dnet2,
		   f->ddp.l.snet1 << 8 | f->ddp.l.snet2,
		   f->ddp.l.dnode, f->ddp.l.snode);
	    printf("\t\tTo socket $%02X from socket $%02X, type $%02X ",
		   f->ddp.l.dsock, f->ddp.l.ssock, f->ddp.l.type);
	    cp = dumpddp(&f->ddp.l.data, f->ddp.l.type,
			 ddplen - (&f->ddp.l.data - &f->ddp.l.len1),
			 rawdata);
	    break;
	  default:		/* Illegal. */
	    printf("(UNKNOWN)\n");
	    cp = (unsigned char *) buf + 1;
	    break;
	}
    }
    return cp - (unsigned char *) buf;
}

/*
** Apple's Phase 2 Ethertalk doesn't seem to have an LLAP header, but
** just drops straight into DDP after the SNAP header.
*/

dumpatalk2(buf, len, rawdata)
void *buf;
int len;
int rawdata;			/* TRUE to dump data part of frame. */
{
    unsigned char *cp;
    struct ddplong *f;
    int ddplen;

    f = (struct ddplong *) buf;
    printf("\tELAP 2\tTo net %d node $%02X, From net %d node $%02X (DDP Extended)\n",
		   f->dnet1 << 8 | f->dnet2, f->dnode, 
		   f->snet1 << 8 | f->snet2, f->snode);
    ddplen = ((f->len1 & 3) << 8) | f->len2;
    printf("\tDDP\tLength %d, Hop count %d, Checksum $%04X\n",
	   ddplen, (f->len1 & 0x3C) >> 2, f->chk1 << 8 | f->chk2);
    printf("\t\tTo socket $%02X from socket $%02X, type $%02X ", f->dsock,
	   f->ssock, f->type);
    cp = dumpddp(&f->data, f->type, ddplen - (&f->data - &f->len1),
		 rawdata);
    return cp - (unsigned char *) buf;
}

unsigned char *
dumpddp(d, ddptype, len, rawdata)
struct ddpdata *d;
int ddptype, len, rawdata;
{
    union {
	struct stuple *st;
	struct etuple *et;
	char *cp;
    } t;
    unsigned char *cp;
    int count, net, i;

    switch (ddptype) {
      case 0x01:		/* RTMP Data packet. */
	printf("(RTMP Data)\n");
	printf("\tRTMP\tRouter's net #%d, Router's node $%02X\n",
	       (d->u.rtmpd.rnet1 << 8) | d->u.rtmpd.rnet2, d->u.rtmpd.rnode);
	len -= 4;
	if (!d->u.rtmpd.u2.rtmps.zero1 && !d->u.rtmpd.u2.rtmps.zero2
	    && d->u.rtmpd.u2.rtmps.hex82 == 0x82) {
	    t.st = &d->u.rtmpd.u2.rtmps.tuple;
	    while (len > 0) {
		printf("\t\tNet %d, Distance %d\n",
		       (t.st->net1 << 8) | t.st->net2,
		       t.st->distance & 0x1F);
		len -= sizeof(struct stuple);
		t.st++;
	    }
	    return (unsigned char *) t.st;
	} else {		/* Extended network. */
	    t.et = &d->u.rtmpd.u2.rtmpl.tuple;
	    while (len > 0) {
		if (t.et->distance & 0x80) {
		    printf("\t\tExtended Net %d-%d, Distance %d\n",
			   (t.et->rstart1 << 8) | t.et->rstart2,
			   (t.et->rend1 << 8) | t.et->rend2,
			   t.et->distance & 0x1F);
		    len -= sizeof(struct etuple);
		    t.cp = (char *) t.et + sizeof(struct etuple); /* CC BUG! */
		} else {
		    printf("\t\tNet %d, Distance %d\n",
			   (t.st->net1 << 8) | t.st->net2,
			   t.st->distance & 0x1F);
		    len -= sizeof(struct stuple);
		    t.cp = (char *) t.st + sizeof(struct stuple); /* CC BUG! */
		}
	    }
	    return (unsigned char *) t.et;
	}
	break;

      case 0x05:		/* RTMP Request. */
	switch (d->u.rtmpr.function) {
	  case 0x01:
	    printf("(RTMP Request)\n");
	    printf("\tRTMP\tRequest %d\n", d->u.rtmpr.function);
	    break;

	  case 0x02:
	    printf("(RTMP RDR)\n");
	    printf("\tRTMP\tRoute Data Request, Split Horizon\n");
	    break;

	  case 0x03:
	    printf("(RTMP RDR)\n");
	    printf("\tRTMP\tRoute Data Request, non-Split Horizon\n");
	    break;

	  default:
	    printf("(RTMP UNKNOWN)\n");
	    printf("\tRTMP\tUnknown Request %d\n", d->u.rtmpr.function);
	    break;
	}
	return ((unsigned char *) d->u.rtmpr.function) + 1;
	break;

      case 0x04:		/* Appletalk Echo Protocol. */
	printf("(AEP)\n");
	switch (d->u.aer.function) {
	  case 0x01:
	    printf("\tAEP\tEcho Request\n");
	    break;
	  case 0x02:
	    printf("\tAEP\tEcho Response\n");
	    break;
	  default:
	    printf("\tAEP\tUnknown Echo Function %d\n", d->u.aer.function);
	    break;
	}
	if (rawdata)
	    hd(&d->u.aer.data, len - 1, 0);
	return (unsigned char *) &d->u.aer.data;
	break;

      case 0x02:		/* NBP */
	printf("(NBP)\n");
	count = d->u.nbp.function & 0x0F;
	switch (d->u.nbp.function >> 4) {
	  case 0x01:		/* NBP BrRq */
	    printf("\tNBP\tBrRq ID #%d\n", d->u.nbp.nbpid);
	    return dumpnbptupl(&d->u.nbp.tuple, count);
	    break;
	  case 0x02:		/* NBP LkUp */
	    printf("\tNBP\tLkUp ID #%d\n", d->u.nbp.nbpid);
	    return dumpnbptupl(&d->u.nbp.tuple, count);
	    break;
	  case 0x03:		/* NBP LkUp-Reply */
	    printf("\tNBP\tLkUp-Reply ID #%d\n", d->u.nbp.nbpid);
	    return dumpnbptupl(&d->u.nbp.tuple, count);
	    break;
	  case 0x04:		/* NBP FwdReq */
	    printf("\tNBP\tFwdReq\n", d->u.nbp.nbpid);
	    return dumpnbptupl(&d->u.nbp.tuple, count);
	    break;
	  default:
	    printf("\tNBP\tUnknown Function %d\n", d->u.nbp.function >> 4);
	    return (unsigned char *) &d->u.nbp.nbpid;
	    break;
	}
	break;

      case 0x06:		/* Zone Info Protocol. */
	printf("(ZIP)\n");
	count = d->u.zip.netcount;
	switch (d->u.zip.function) {
	  case 0x01:		/* ZIP Query. */
	    printf("\tZIP\tQuery\n");
	    cp = &d->u.zip.net1;
	    while (count--) {
		printf("\t\tNet #%d\n", (cp[0] << 8) | cp[1]);
		cp += 2;
	    }
	    return cp;
	    break;
	  case 0x02:		/* ZIP Reply. */
	    printf("\tZIP\tReply\n");
	    cp = &d->u.zip.net1;
	    while (count--) {
	      zipagain:
		net = (cp[0] << 8) | cp[1];
		printf("\t\tNet #%d a.k.a. '", net);
		cp += 2;
		i = *cp++;
		while (i--)
		    putchar(*cp++);
		printf("'\n");
		if (net == ((cp[0] << 8) | cp[1]))
		    goto zipagain;
	    }
	    return cp;
	    break;
	  case 0x05:		/* ZIP GetNetInfo Request. */
	    printf("\tZIP\tGetNetInfo Request for zone '");
	    cp = &d->u.zip.net1 + 4;
	    i = *cp++;
	    while (i--)
		putchar(*cp++);
	    printf("'\n");
	    return cp;
	    break;
	  case 0x06:		/* ZIP GetNetInfo Reply. */
	    printf("\tZIP\tGetNetInfo Reply\n");
	    net = d->u.zip.netcount;
	    printf("\t\t");
	    if (net & 0x80) {
		printf("Zone Invalid");
		if (net & 0x60)
		    printf(", ");
	    }
	    if (net & 0x40) {
		printf("Use Broadcast");
		if (net & 0x20)
		    printf(", ");
	    }
	    if (net & 0x20) {
		printf("Only One Zone");
	    }
	    cp = &d->u.zip.net1;
	    printf("\n\t\tNet %d-%d, Zone '", (cp[0] << 8) | cp[1],
		   (cp[2] << 8) | cp[3]);
	    cp += 4;
	    i = *cp++;
	    while (i--)
		putchar(*cp++);
	    putchar('\'');
	    if (i = *cp++) {
		printf(", Zone Multicast '");
		while (i--)
		    putchar(*cp++);
		putchar('\'');
	    }
	    if (net & 0x80) {
		if (i = *cp++) {
		    printf(", Default Zone '");
		    while (i--)
			putchar(*cp++);
		    putchar('\'');
		}
	    }
	    putchar('\n');
	    return cp;
	    break;
	  case 0x07:		/* ZIP Notify. */
	    printf("\tZIP\tNotify\n");
	    net = d->u.zip.netcount;
	    printf("\t\t");
	    if (net & 0x80) {
		printf("Zone Invalid");
		if (net & 0x60)
		    printf(", ");
	    }
	    if (net & 0x40) {
		printf("Use Broadcast");
		if (net & 0x20)
		    printf(", ");
	    }
	    if (net & 0x20) {
		printf("Only One Zone");
	    }
	    cp = &d->u.zip.net1;
	    printf("\n\t\tOld Zone '");
	    cp += 4;
	    i = *cp++;
	    while (i--)
		putchar(*cp++);
	    putchar('\'');
	    if (i = *cp++) {
		printf(", New Zone Multicast '");
		while (i--)
		    putchar(*cp++);
		putchar('\'');
	    }
	    if (net & 0x80) {
		if (i = *cp++) {
		    printf(", New Zone '");
		    while (i--)
			putchar(*cp++);
		    putchar('\'');
		}
	    }
	    putchar('\n');
	    return cp;
	    break;
	  case 0x08:		/* ZIP Extended Reply. */
	    printf("\tZIP\tExtended Reply\n");
	    printf("\t\t<<Not Parsed>>\n");
	    return (unsigned char *) &d->u.zip.netcount;
	    break;
	};
	break;

      case 0x03:		/* Appletalk Transaction Protocol. */
	printf("(ATP)\n");
	switch ((i = d->u.atp.control) >> 6) {
	  case 0:
	    printf("\tATP\tUNKNOWN $%02X\n", i);
	    break;
	  case 1:
	    printf("\tATP\tTReq, Bmap $%02X, ", i);
	    if (i & 0x20) {
		printf("XO");
		if (i & 0x18)
		    putchar(' ');
	    }
	    if (i & 0x10) {
		printf("EOM");
		if (i & 0x08)
		    putchar(' ');
	    }
	    if (i & 0x08) {
		printf("STS");
	    }
	    if (!(i & 0x20))
		printf(", ");
	    else {
		switch (i & 0x07) {
		  case 0:
		    printf(" 30 sec ");
		    break;
		  case 1:
		    printf(" 1 min ");
		    break;
		  case 2:
		    printf(" 2 min ");
		    break;
		  case 3:
		    printf(" 4 min ");
		    break;
		  case 4:
		    printf(" 8 min ");
		    break;
		  default:
		    printf(" ?? (%d) ", i & 0x07);
		    break;
		}
		printf("tmout, ");
	    }
	    printf("Trn ID #%d, UData $%08X\n",
		   (d->u.atp.tid1 << 8) | d->u.atp.tid2,
		   (d->u.atp.user[0] << 24) | (d->u.atp.user[1] << 16) |
		   (d->u.atp.user[2] << 8) | d->u.atp.user[3]);
	    break;
	  case 2:
	    printf("\tATP\tTResp, Seq #%d, ", d->u.atp.bitseq);
	    if (i & 0x20) {
		printf("XO");
		if (i & 0x18)
		    putchar(' ');
	    }
	    if (i & 0x10) {
		printf("EOM");
		if (i & 0x08)
		    putchar(' ');
	    }
	    if (i & 0x08)
		printf("STS");
	    if (i & 0x38)
		printf(", ");
	    printf("Trans ID #%d, UsrData $%08X\n",
		   (d->u.atp.tid1 << 8) | d->u.atp.tid2,
		   (d->u.atp.user[0] << 24) | (d->u.atp.user[1] << 16) |
		   (d->u.atp.user[2] << 8) | d->u.atp.user[3]);
	    break;
	  case 3:
	    printf("\tATP\tTRel, Seq #%d, ", d->u.atp.bitseq);
	    if (i & 0x20)
		printf("XO ");
	    if (i & 0x10)
		printf("EOM ");
	    if (i & 0x08)
		printf("STS ");
	    printf("\n");
	    break;
	}
	if (rawdata)
	    hd(&d->u.atp.data, len - 8, 0);
	return (unsigned char *) &d->u.atp.data;
	break;

      case 0x07:		/* Appletalk Data Stream Protocol. */
	printf("(ADSP)\n");
	printf("\tADSP\t");
	if (d->u.adsp.descriptor & 0x80) { /* ADSP control. */
	    printf("Control ");
	    switch (d->u.adsp.descriptor & 0x0F) {
	      case 0:
		if (d->u.adsp.descriptor & 0x40)
		    printf("Probe ");
		else
		    printf("Ack ");
		break;
	      case 1:
		printf("OpenReq ");
		break;
	      case 2:
		printf("OpenAck ");
		break;
	      case 3:
		printf("OpenReq&Ack ");
		break;
	      case 4:
		printf("OpenDeny ");
		break;
	      case 5:
		printf("CloseAdvice ");
		break;
	      case 6:
		printf("FwdReset ");
		break;
	      case 7:
		printf("FwdResetAck ");
		break;
	      case 8:
		printf("RetransAdvice ");
		break;
	      default:
		printf("UNKNOWN (%d) ", d->u.adsp.descriptor & 0x0F);
		break;
	    }
	} else {
	    printf("Data ");
	}
	if (d->u.adsp.descriptor & 0x40)
	    printf("ACKR ");
	if (d->u.adsp.descriptor & 0x20)
	    printf("EOM ");
	if (d->u.adsp.descriptor & 0x10)
	    printf("ATN ");
	printf("\n\t\tSrcConnID %d, FirstSeq %d, NextSeq %d, RcvWdw %d\n",
	       (d->u.adsp.sconnid[0] << 8) | d->u.adsp.sconnid[1],
	       (d->u.adsp.pfbseq[0] << 24) | (d->u.adsp.pfbseq[1] << 16) |
	       (d->u.adsp.pfbseq[2] << 8) | d->u.adsp.pfbseq[3],
	       (d->u.adsp.pnrseq[0] << 24) | (d->u.adsp.pnrseq[1] << 16) |
	       (d->u.adsp.pnrseq[2] << 8) | d->u.adsp.pnrseq[3],
	       (d->u.adsp.prwdw[0] << 8) | d->u.adsp.prwdw[1]);
	cp = &d->u.adsp.data;
	if (d->u.adsp.descriptor & 0x10) {
	    printf("\t\tAttnCode %d\n", (cp[0] << 8) | cp[1]);
	    cp += 2;
	    len -= 2;
	}
	if (d->u.adsp.descriptor >= 0x81 && d->u.adsp.descriptor <= 0x84) {
	    printf("\t\tOpenDialog ProtVrsn $%02X, DstConnID %d, AttnSeq %d\n",
		   (cp[0] << 8) | cp[1], (cp[2] << 8) | cp[3],
		   (cp[4] << 24) | (cp[5] << 16) | (cp[6] << 8) | cp[7]);
	    cp += 8;
	    len -= 8;
	}
	if (rawdata)
	    hd(cp, len - 13, 0);
	return cp;
	break;

      default:
	printf("(UNKNOWN -- $%03X)\n", ddptype);
	return (unsigned char *) d;
	break;
    }
}

unsigned char *
dumpnbptupl(n, count)
struct nbptuple *n;
int count;
{
    unsigned char *cp;
    int i;

    while (count--) {
	printf("\t\t    Net #%d, Node ID $%02X, Socket $%02X, Enum %d\n",
	       (n->net1 << 8) | n->net2, n->node, n->sock, n->enumer);
	cp = &n->obj;
	printf("\t\t\tObject: '");
	i = *cp++;
	while (i--)
	    putchar(*cp++);
	printf("', Type: '");
	i = *cp++;
	while (i--)
	    putchar(*cp++);
	printf("', Zone: '");
	i = *cp++;
	if (!i)
	    putchar('*');
	else
	    while (i--)
		putchar(*cp++);
	printf("'\n");
	n = (struct nbptuple *) cp;
    }
    return cp;
}
