/*
** Dump out a BPDU (Bridge Protocol Data Unit, Spanning Tree) frame.
** From ANSI/IEEE Std 802.1D, 1993 Edition.
*/

#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include "protos.h"

				/* This structure is on an odd boundary */
				/* (due to the SAP header) and is an odd */
				/* length.  The following (adjoining) */
				/* structure thus is on an even boundary, */
				/* and is safe to access as short/long even */
				/* on a 68010. */
struct bpdu0 {
    unsigned char protid[2];	/* Protocol ID. */
    unsigned char protverid;	/* Protocol Version ID. */
    unsigned char bpdutype;	/* BPDU Type. */
    unsigned char flags;
};

struct bpdu {
    unsigned short rootpri;	/* Root Identifier priority. */
    unsigned char rootid[6];	/* Root Identifier (MAC address). */
    unsigned long rootcost;	/* Root Path Cost. */
    unsigned short bridgepri;	/* Bridge Identifier priority. */
    unsigned char bridgeid[6];	/* Bridge Identifier (MAC address). */
    unsigned short portid;	/* Port Identifier. */
    unsigned short msgage;	/* Message Age. */
    unsigned short maxage;	/* Max Age. */
    unsigned short hellotime;	/* Hello Time. */
    unsigned short fwddelay;	/* Forward Delay. */
};

dumpbpdu(buf, len, rawdata)
unsigned char *buf;
int len, rawdata;
{
    unsigned char *buffer;
    struct bpdu0 *bpdu0;
    struct bpdu *bpdu;
    int protocol, ii, jj, kk, ll;
    char root[14], bridge[14];

	/*
	** decode BPDU header
	*/
    bpdu0 = (struct bpdu0 *) buf;
    bpdu =  (struct bpdu *) &bpdu0[1];
    buffer = (unsigned char *) bpdu0;
    protocol = (bpdu0->protid[0] << 8) | bpdu0->protid[1];
    printf("\tBPDU %d (V%d) ", protocol, bpdu0->protverid);
    if (protocol != 0) {
	printf("Unknown Protocol!\n");
	protocol = &bpdu0->protverid - buf;
	len -= protocol;
	buffer += protocol;
    } else if (bpdu0->protverid != 0) {
	printf("Unknown Protocol Version!\n");
	protocol = &bpdu0->bpdutype - buf;
	len -= protocol;
	buffer += protocol;
    } else switch (bpdu0->bpdutype) {
      case 0:
	printf("Spanning Tree Configuration");
	if (bpdu0->flags & 0x80)
	    printf(" TOPO_CHG_ACK");
	if (bpdu0->flags & 0x01)
	    printf(" TOPO_CHG");
	makename(bpdu->rootid, root);
	makename(bpdu->bridgeid, bridge);
	printf("\n\t\tRoot Pri   %d, Root ID   %s, Root Cost %ld\n",
	       bpdu->rootpri, root, bpdu->rootcost);
	printf("\t\tBridge Pri %d, Bridge ID %s, Port ID 0x%X\n",
	       bpdu->bridgepri, bridge, bpdu->portid);
	ii = (bpdu->msgage * 1000) / 256;
	jj = (bpdu->maxage * 1000) / 256;
	kk = (bpdu->hellotime * 1000) / 256;
	ll = (bpdu->fwddelay * 1000) / 256;
	printf("\t\tMsg Age %d.%03d, Max Age %d.%03d, Hello Time %d.%03d, "
	       "Fwd Delay %d.%03d\n", ii / 1000, ii % 1000,
	       jj / 1000, jj % 1000, kk / 1000, kk % 1000,
	       ll / 1000, ll % 1000);
	buffer += sizeof(struct bpdu0) + sizeof(struct bpdu);
	len -= sizeof(struct bpdu0) + sizeof(struct bpdu);
	break;

      case 0x80:
	printf("Spanning Tree Topology Change Notification\n");
	buffer += sizeof(struct bpdu0) - sizeof(bpdu0->flags);
	len -= sizeof(struct bpdu0) - sizeof(bpdu0->flags);
	break;

      default:
	printf("Spanning Tree UNKNOWN Type\n");
	buffer += sizeof(struct bpdu0) - sizeof(bpdu0->flags);
	len -= sizeof(struct bpdu0) - sizeof(bpdu0->flags);
	break;
    }
    if (len <= 0)
	return buffer - buf;

    if (rawdata)		/* Now dump whatever is left. */
	hd(buffer, len, 0);
    return buffer - buf;
}
