/*
** Function to dump out a dmap protocol frame.
*/

#include <stdio.h>
#include <dnix/cheapdrv.h>
#include "dmap.h"
#include "protos.h"

/*
** Dump DMAP/DWHO protocol request.
*/

dumpdmap(buffer, byte_count)
void *buffer;
int byte_count;
{
    int i, temp;
    long deltat;
    struct dmapmsg *msgp;
    struct dmapmsg2 *msgp2;
    struct wsinfo *wsp;
    struct duser *up;
    char *ctime();

    msgp = (struct dmapmsg *) buffer;
    switch (msgp->rectype) {
      case DMAPREC:		/* Old DMAP record. */
	temp = ((msgp->ms.dm.dnpserial >> 27) & 0x1F) + '@';
	printf("\tDMAP\tpacket from %s, Serial No. %c%08ld\n",
	       nametostr(msgp->ms.dm.dnpname, 0),
	       temp, msgp->ms.dm.dnpserial & 0x7FFFFFFL);
	printf("\t\tSent %.24s\n\t\tRunning since %.24s\n",
	       ctime(&msgp->ms.dm.sendtime), ctime(&msgp->ms.dm.boottime));
	printf("\t\tLoad averages:  %d.%02d (1 min), %d.%02d (10 min), "
	       "%d.%02d (15 min)\n",
	       msgp->ms.dm.loadavg[0] / 100, msgp->ms.dm.loadavg[0] % 100,
	       msgp->ms.dm.loadavg[1] / 100, msgp->ms.dm.loadavg[1] % 100,
	       msgp->ms.dm.loadavg[2] / 100, msgp->ms.dm.loadavg[2] % 100);
	printf("\t\tOS Release %.9s, Version %.9s\n", msgp->ms.dm.dnprel,
	       msgp->ms.dm.dnpver);
	wsp = msgp->ms.dm.ws;
	i = (struct wsinfo *) ((long) msgp + byte_count) - wsp;
	while (i--) {
	    temp = ((wsp->wsserial >> 27) & 0x1F) + '@';
	    if (wsp->wsonline) {
		printf("\t\t\tWorkstation %8.8s, Serial No. %c%08ld  UP\n",
		   wsp->wsname, temp, wsp->wsserial & 0x7FFFFFFL);
		printf("\t\t\t  Level %d, since %.24s\n", 
		       wsp->wslevel, ctime(&wsp->wsconnect));
		printf("\t\t\t  running OS Release %.8s, Version %.8s\n",
		       wsp->wsrel, wsp->wsver);
	    } else {
		printf("\t\t\tWorkstation %8.8s, Serial No. *********  DOWN\n",
		   wsp->wsname);
	    }
	    wsp++;
	}
	return (char *) wsp - (char *) buffer;
	break;

      case DWHOREC:		/* Old DWHO record. */
	printf("\tDWHO\tpacket (Version %d) from %s\n", msgp->ms.wh.protocol,
	       msgp->ms.wh.sysname);
	printf("\t\tSent %.24s\n", ctime(&msgp->ms.wh.sendtime));
	printf("\t\tRunning since %.24s\n", ctime(&msgp->ms.wh.boottime));
	printf("\t\tLoad averages:  %d.%02d (1 min), %d.%02d (10 min), "
	       "%d.%02d (15 min)\n",
	       msgp->ms.wh.loadave[0] / 100, msgp->ms.wh.loadave[0] % 100,
	       msgp->ms.wh.loadave[1] / 100, msgp->ms.wh.loadave[1] % 100,
	       msgp->ms.wh.loadave[2] / 100, msgp->ms.wh.loadave[2] % 100);
	up = msgp->ms.wh.we;
	i = (struct duser *) ((long) msgp + byte_count) - up;
	while (i--) {
	    printf("\t\t  User %8.8s on line %8.8s since %.24s\n",
		   up->user, up->line, ctime(&up->time));
	    deltat = msgp->ms.wh.sendtime - up->idle;
	    printf("\t\t\t\t\t    Idle since %.24s\n", ctime(&deltat));
	    up++;  
	}
	return (char *) up - (char *) buffer;
	break;

      case DMAPREC2:		/* New DMAP record. */
	msgp2 = (struct dmapmsg2 *) buffer;
	temp = ((msgp2->dnpserial >> 27) & 0x1F) + '@';
	if (msgp2->boottime) {
	    printf("\tDMAP 2\tpacket from %s, Serial No. %c%08ld\n",
		   nametostr(msgp2->dnpname, 0),
		   temp, msgp2->dnpserial & 0x7FFFFFFL);
	    printf("\t\tSent %.24s\n", ctime(&msgp2->sendtime));
	    printf("\t\tRunning since %.24s\n", ctime(&msgp2->boottime));
	    printf("\t\tOS Release %.9s, Version %.9s\n", msgp2->dnprel,
		   msgp2->dnpver);
	} else {		/* Detected death notification. */
	    printf("\tDMAP 2\t%s isn't 'alive' anymore\n",
		   nametostr(msgp2->dnpname, 0));
	    printf("\t\tSent %.24s\n", ctime(&msgp2->sendtime));
	}
	return sizeof(struct dmapmsg2);
	break;
    }
}
