/*
** Function to dump out an IP protocol frame.
*/

#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netdb.h>
#include "protos.h"

#define TRUE (1)
#define FALSE (0)

char *decode_protocol(int);
char *decode_port(int, int);
char *decode_ip_offset();
char *decode_tcp_flags();
char *decode_icmp_type(int);
char *decode_icmp_subcode(int, int);
char *inet_ntoa();
int nbport(int, int, int);
void *alloca();

struct pseudohdr {	       /* Pseudo-header oft-included in IP checksum. */
    struct in_addr ip_src, ip_dst;
    unsigned short protocol;
    unsigned short len;
};

dumpip(buf, len, rawdata)
unsigned char *buf;
int len, rawdata;
{
    unsigned char *buffer;
    struct ip *ip;
    struct icmp *icmp;
    struct udphdr *udp;
    struct pseudohdr *phdr;
    struct tcphdr *tcp;
    struct hostent *host;
    int count, sport, dport;
    int ip_len, ip_opt_len, tcp_opt_len, tcp_data_len;
    char *name;

	/*
	** decode IP header
	*/
    ip = (struct ip *) buf;
    buffer = (unsigned char *) ip;
    ip_len = ip->ip_hl << 2;
    ip_opt_len = ip_len - sizeof(struct ip);
    count = 0;
    printf("\tIP (V%d)\t", ip->ip_v);
    if (ip->ip_v != IPVERSION) {
	printf("Unknown Version!\n");
	return &ip->ip_tos - buf;
    }
    printf("To %s from %s, type %d (%s)\n", inet_ntoa(&ip->ip_dst, 0),
	   inet_ntoa(&ip->ip_src, 1), ip->ip_p, decode_protocol(ip->ip_p));
    printf("\t\t%d byte header%s, %d bytes of data, Time to live: %d hops\n",
	   ip_len, (ip_opt_len ? " (incl. opts)" : ""), ip->ip_len - ip_len,
	   ip->ip_ttl);
    printf("\t\tPrecedence %d%s%s%s, Hdr Cksum: $%04X %s\n", ip->ip_tos >> 5,
	   (ip->ip_tos & 0x10) ? " Low Delay" : "",
	   (ip->ip_tos & 0x80) ? " Hi Thruput" : "",
	   (ip->ip_tos & 0x04) ? " Hi Rel" : "", ip->ip_sum,
	   ipcksum(ip, sizeof(*ip)) ? "(BAD)" : "(OK)");
    printf("\t\tPacket ID: %d%s%s, Offset: %d\n", ip->ip_id,
	   (ip->ip_off & IP_DF) ? " Unfragmentable" : "",
	   (ip->ip_off & IP_MF) ? " More fragments" : "", ip->ip_off & 0x1FFF);

    buffer += sizeof(struct ip);
    len -= sizeof(struct ip);
    if (len <= 0)
	return buffer - buf;

    if (ip_opt_len) {		/* Any Options? */
	unsigned char *p;
	int i, j;

	i = ip_opt_len;
	p = buffer;
	while (i++) {
	    printf("\t\tIP Option $%02X: %sClass %d, Number %d -- ", *p,
		   (*p & 0x80) ? "Copy, " : "", (*p >> 5) & 3, *p & 0x1F);
	    switch (*p & 0x7F) {
	      case 0: printf("End of Options\n"); break;

	      case 1: printf("NOP\n"); break;

	      case 2:
		printf("Security Restrictions\n");
		hd(&p[1], 10, 0);
		i += 10;
		p += 10;
		break;

	      case 3:
		printf("Loose Src Route\n");
	      route:
		j = 4;
		while (j < p[2]) {
		    printf("\t\t\t%s\n", inet_ntoa(&p[j], 0));
		    j += 4;
		}
		break;

	      case 7: printf("Record Route\n");
		goto route;

	      case 8: printf("Stream ID $%02X%02X%02X\n", p[1], p[2], p[3]);
		i += 3;
		p += 3;
		break;

	      case 9: printf("Strict Src Route\n");
		goto route;

	      case 0x44: printf("Timestamp\n");
		hd(&p[4], p[1] - 4, p[2] - 4);
		i += p[1] - 1;
		p += p[1] - 1;
		break;

	      default: printf("Unknown\n"); break;
	    }
	    p++;
	}
	buffer += ip_opt_len;
	len -= ip_opt_len;
    }

    sport = dport = 0;
    switch (ip->ip_p) {		/* Decode UDP/ICMP/TCP header. */
      case IPPROTO_UDP:
	udp = (struct udphdr *) buffer;
	phdr = alloca(sizeof(*phdr) + len); /* Our checksummer works on */
	memcpy(&phdr[1], buffer, len); /* blocks of memory, and UDP has a */
	phdr->ip_src = ip->ip_src; /* pseudo-header included in this sum. */
	phdr->ip_dst = ip->ip_dst; /* Make a contiguous block of memory to */
	phdr->protocol = ip->ip_p; /* checksum. */
	phdr->len = udp->uh_ulen;
	printf("\tUDP\tSrc Port: %s (%u), ",
	       decode_port(sport = udp->uh_sport, ip->ip_p), udp->uh_sport);
	printf("Dest Port: %s (%u)\n\t\t%d byte header, %d byte%s of data, "
	       "Cksum: $%04X %s\n", decode_port(dport=udp->uh_dport, ip->ip_p),
	       udp->uh_dport, sizeof(*udp), udp->uh_ulen - sizeof(*udp),
	       ((udp->uh_ulen - sizeof(*udp)) == 1) ? "" : "s", udp->uh_sum,
	       ipcksum(phdr, phdr->len + sizeof(*phdr)) ? "(BAD)" : "(OK)");
	buffer += sizeof(struct udphdr);
	len -= sizeof(struct udphdr);
	break;

      case IPPROTO_TCP:
	tcp = (struct tcphdr *) buffer;
	phdr = alloca(sizeof(*phdr) + len); /* Our checksummer works on */
	memcpy(&phdr[1], buffer, len); /* blocks of memory, and TCP has a */
	phdr->ip_src = ip->ip_src; /* pseudo-header included in this sum. */
	phdr->ip_dst = ip->ip_dst; /* Make a contiguous block of memory to */
	phdr->protocol = ip->ip_p; /* checksum. */
	phdr->len = ip->ip_len - ip_len;
	tcp_opt_len = (tcp->th_off << 2) - sizeof(struct tcphdr);
	tcp_data_len = ip->ip_len - ip_len - sizeof(*tcp) - tcp_opt_len;
	printf("\tTCP\tSrc Port: %u (%s), ", tcp->th_sport,
	       decode_port(sport = tcp->th_sport, ip->ip_p));
	printf("Dest Port: %u (%s)\n", tcp->th_dport,
	       decode_port(dport = tcp->th_dport, ip->ip_p));
	printf("\t\tSeq #: %u, Ack #: %u\n", tcp->th_seq, tcp->th_ack);
	printf("\t\tData Offset: %d%s, Flags: $%02X - %s\n", tcp->th_off,
	       (tcp_opt_len > 0 ? " (Options Present)" : ""),
	       tcp->th_flags, decode_tcp_flags(tcp->th_flags));
	printf("\t\t%d byte header%s, %d byte%s of data\n",
	       sizeof(*tcp) + tcp_opt_len,
	       (tcp_opt_len ? " (incl. opts)" : ""), tcp_data_len,
	       (tcp_data_len == 1) ? "" : "s");
	printf("\t\tWindow: %u, Cksum $%04X %s, Urgent Ptr: %u\n",
	       tcp->th_win, tcp->th_sum,
	       ipcksum(phdr, phdr->len + sizeof(*phdr)) ? "(BAD)" : "(OK)",
	       tcp->th_urp);
	buffer += sizeof(struct tcphdr);
	len -= sizeof(struct tcphdr);
	if (tcp_opt_len) {
	    int optlen;
	    u_short maxseg;
	    u_char *cp;

	    if (tcp_opt_len > 0)
		printf("\t\tTCP Options\n");
	    cp = (u_char *)buffer;
	    buffer += tcp_opt_len;
	    len -= tcp_opt_len;
	    for (; tcp_opt_len > 0; tcp_opt_len -= optlen, cp += optlen) {
		switch (cp[0]) {
		  case TCPOPT_EOL:
		    printf("\t\t\tTCPOPT_EOL:     %02X\n", cp[0]);
		    optlen = tcp_opt_len;
		    break;

		  case TCPOPT_NOP:
		    printf("\t\t\tTCPOPT_NOP:     %02X\n", cp[0]);
		    optlen = 1;
		    break;

		  case TCPOPT_MAXSEG:
		    optlen = cp[1];
		    if (optlen <= 0)
			break;
		    if (optlen != 4)
			continue;
		    maxseg = *(u_short *) (cp + 2);
		    printf("\t\t\tTCPOPT_MAXSEG:  %02X (%d)\n", cp[0], maxseg);
		    break;

		  default:
		    optlen = cp[1];
		    if (optlen <= 0)
			optlen = tcp_opt_len;
		    printf("\t\t\tUnknown Option: %02X\n", cp[0]);
		    break;
		}
	    }
	}
	break;

      case IPPROTO_ICMP:
	icmp = (struct icmp *) buffer;
	printf("\tICMP\tType: %s (%d), Subcode: %s (%d), Cksum $%04X %s\n", 
	       decode_icmp_type(icmp->icmp_type), icmp->icmp_type,
	       decode_icmp_subcode(icmp->icmp_type, icmp->icmp_code),
	       icmp->icmp_code, icmp->icmp_cksum,
	       ipcksum(icmp, ip->ip_len - ip_len) ? "(BAD)" : "(OK)");
	switch (icmp->icmp_type) {
	  case ICMP_ECHOREPLY:
	  case ICMP_ECHO:
	  case ICMP_IREQ:
	  case ICMP_IREQREPLY:
	    printf("\t\tID: %d, Sequence #%d\n", icmp->icmp_id,
		   icmp->icmp_seq);
	    buffer += 8;
	    len -= 8;
	    break;

	  case ICMP_UNREACH:
	  case ICMP_SOURCEQUENCH:
	    buffer += 8;
	    len -= 8;
	    break;

	  case ICMP_REDIRECT:
	    printf("\t\tGateway %s\n", inet_ntoa(&buffer[4], 0));
	    buffer += 8;
	    len -= 8;
	    break;

	  case ICMP_TIMXCEED:
	    buffer += 8;
	    len -= 8;
	    break;

	  case 9:		/* Router Advertisement */
	    buffer += 8 + icmp->icmp_pptr * 8;
	    len -= 8 + icmp->icmp_pptr * 8;
	    break;

	  case 10:		/* Router Solicited */
	    buffer += 8;
	    len -= 8;
	    break;

	  case ICMP_PARAMPROB:	  
	    printf("\t\tByte %d is bad\n", icmp->icmp_pptr);
	    buffer += 8;
	    len -= 8;
	    break;

	  case ICMP_TSTAMP:
	  case ICMP_TSTAMPREPLY:
	    printf("\t\tID: %d, Sequence #%d\n", icmp->icmp_id,
		   icmp->icmp_seq);
	    printf("\t\tTimestamps: Originate %d, Rx %d, Tx %d\n",
		   icmp->icmp_otime, icmp->icmp_rtime, icmp->icmp_ttime);
	    buffer += 20;
	    len -= 20;
	    break;

	  case ICMP_MASKREQ:
	  case ICMP_MASKREPLY:
	    printf("\t\tID: %d, Sequence #%d\n", icmp->icmp_id,
		   icmp->icmp_seq);
	    printf("\t\tSubnet Mask: $%08X\n", icmp->icmp_mask);
	    buffer += 12;
	    len -= 12;
	    break;

	  default:
	    buffer += sizeof(struct icmp);
	    len -= sizeof(struct icmp);
	    break;
	}
	break;
    }
    if (dport = nbport(sport, dport, ip->ip_p)) {
	switch (dport) {
	  case 1: count = dumpnbname(buffer, len, rawdata); break;
	  case 2: count = dumpnbdgram(buffer, len, rawdata); break;
	  case 3: count = dumpnbsess(buffer, tcp_data_len, rawdata); break;
	  default: count = 0; break;
	}
	buffer += count;
	len -= count;
    }
    if (len <= 0)
	return buffer - buf;

    if (rawdata && len)		/* Now dump whatever is left. */
	hd(buffer, len, 0);
    return buffer - buf;
}

char *
decode_protocol(prot)
int prot;
{
    static char cprot[6];

    switch (prot) {
      case IPPROTO_IP: return("IP");
      case IPPROTO_ICMP: return("ICMP");
      case IPPROTO_IGMP: return("IGMP");
      case IPPROTO_GGP: return("GGP");
      case IPPROTO_TCP: return("TCP");
      case IPPROTO_EGP: return("EGP");
      case IPPROTO_PUP: return("PUP");
      case IPPROTO_UDP: return("UDP");
      case IPPROTO_HELLO: return("HELLO");
      case IPPROTO_ND: return("ND");
      case IPPROTO_RAW: return("RAW");
      default: sprintf(cprot, "%d", prot);
	return cprot;
    }
}

char *
decode_port(port, protocol)
int port, protocol;
{
    switch (port) {
      case 5: return("Remote Job Entry");
      case 7: return("Echo");
      case 9: return("Discard");
      case 11: return("Systat");
      case 13: return("Daytime");
      case 15:
	if (protocol == IPPROTO_TCP)
	    return("Netstat");
	break;
      case 17: return("Quote of the Day");
      case 19: return("Character Generator");
      case 20:
	if (protocol == IPPROTO_TCP)
	    return("ftp-data");
	break;
      case 21:
	if (protocol == IPPROTO_TCP)
	    return("ftp");
	break;
      case 23:
	if (protocol == IPPROTO_TCP)
	    return("telnet");
	break;
      case 25:
	if (protocol == IPPROTO_TCP)
	    return("SMTP");
	break;
      case 37: return("Time");
      case 39:
	if (protocol == IPPROTO_UDP)
	  return("Resource Location Protocol");
	break;
      case 42: return("Host Name Server");
      case 43: 
	if (protocol == IPPROTO_TCP)
	    return("Who is");
	break;
      case 53: return("Domain Name Server");
      case 57:
	if (protocol == IPPROTO_TCP)
	    return("mtp");
      case 67:
	if (protocol == IPPROTO_UDP)
	    return("Bootstrap Protocol Server");
	break;
      case 69:
	if (protocol == IPPROTO_UDP)
	    return("Trivial File Transfer");
	break;
      case 70: return("gopher");
      case 77:
	if (protocol == IPPROTO_TCP)
	    return("rje/netrjs");
	break;
      case 79:
	if (protocol == IPPROTO_TCP)
	    return("Finger");
	break;
      case 87:
	if (protocol == IPPROTO_TCP)
	    return("link/ttylink");
	break;
      case 95:
	if (protocol == IPPROTO_TCP)
	    return("supdup");
	break;
      case 101:
	if (protocol == IPPROTO_TCP)
	    return("hostnames");
	break;
      case 102:
	if (protocol == IPPROTO_TCP)
	    return("iso-tsap");
	break;
      case 103:
	if (protocol == IPPROTO_TCP)
	    return("dictionary/webster/x400");
	break;
      case 104:
	if (protocol == IPPROTO_TCP)
	    return("x400-snd");
	break;
      case 105:
	if (protocol == IPPROTO_TCP)
	    return("csnet-ns");
	break;
      case 109:
	if (protocol == IPPROTO_TCP)
	    return("pop/pop-2");
	break;
      case 110:
	if (protocol == IPPROTO_TCP)
	    return("pop-3");
	break;
      case 111: return("portmap/sunrpc");
      case 113:
	if (protocol == IPPROTO_TCP)
	    return("auth");
	break;
      case 115:
	if (protocol == IPPROTO_TCP)
	    return("sftp");
	break;
      case 117:
	if (protocol == IPPROTO_TCP)
	    return("path/uucp-path");
	break;
      case 119:
	if (protocol == IPPROTO_TCP)
	    return("nntp");
	break;
      case 123:
	if (protocol == IPPROTO_UDP)
	    return("Network Time Protocol");
	break;
      case 137:
	if (protocol == IPPROTO_UDP)
	    return("nbname");
	break;
      case 138:
	if (protocol == IPPROTO_UDP)
	    return("nbdatagram");
	break;
      case 139:
	if (protocol == IPPROTO_TCP)
	    return("nbsession");
	break;
      case 144:
	if (protocol == IPPROTO_TCP)
	    return("NeWS");
	break;
      case 153:
	if (protocol == IPPROTO_UDP)
	    return("sgmp");
	break;
      case 158:
	if (protocol == IPPROTO_TCP)
	    return("tcprepo");
	break;
      case 161:
	if (protocol == IPPROTO_UDP)
	    return("snmp");
	break;
      case 162:
	if (protocol == IPPROTO_UDP)
	    return("snmp-trap");
	break;
      case 170:
	if (protocol == IPPROTO_TCP)
	    return("print-srv");
	break;
      case 175:
	if (protocol == IPPROTO_TCP)
	    return("vmnet");
	break;
      case 315:
	if (protocol == IPPROTO_UDP)
	    return("load");
	break;
      case 400:
	if (protocol == IPPROTO_TCP)
	    return("vmnet0");
	break;
      case 500:
	if (protocol == IPPROTO_UDP)
	    return("sytek");
	break;
      case 512:
	if (protocol == IPPROTO_TCP)
	    return("exec");
	else
	    return("biff");
      case 513:
	if (protocol == IPPROTO_TCP)
	    return("login");
	else
	    return("rwhod");
      case 514:
	if (protocol == IPPROTO_TCP)
	    return("shell");
	else
	    return("syslog");
      case 515:
	if (protocol == IPPROTO_TCP)
	    return("printer");
	break;
      case 517:
	if (protocol == IPPROTO_UDP)
	    return("talk");
	break;
      case 518:
	if (protocol == IPPROTO_UDP)
	    return("ntalk");
	break;
      case 520:
	if (protocol == IPPROTO_TCP)
	    return("efs");
	else
	    return("routed");
      case 525:
	if (protocol == IPPROTO_UDP)
	    return("timed");
      case 526:
	if (protocol == IPPROTO_TCP)
	    return("tempo");
	break;
      case 530:
	if (protocol == IPPROTO_TCP)
	    return("courier");
	break;
      case 531:
	if (protocol == IPPROTO_TCP)
	    return("conference");
	else
	    return("rvd-control");
	break;
      case 532:
	if (protocol == IPPROTO_TCP)
	    return("netnews");
	break;
      case 533:
	if (protocol == IPPROTO_UDP)
	    return("netwall");
	break;
      case 540:
	if (protocol == IPPROTO_TCP)
	    return("uucp");
	break;
      case 543:
	if (protocol == IPPROTO_TCP)
	    return("klogin");
	break;
      case 544:
	if (protocol == IPPROTO_TCP)
	    return("kshell");
	break;
      case 550:
	if (protocol == IPPROTO_UDP)
	    return("new-rwho");
	break;
      case 556:
	if (protocol == IPPROTO_TCP)
	    return("remotefs");
	break;
      case 560:
	if (protocol == IPPROTO_UDP)
	    return("rmonitor");
	break;
      case 561:
	if (protocol == IPPROTO_UDP)
	    return("monitor");
	break;
      case 600:
	if (protocol == IPPROTO_TCP)
	    return("garcon");
	break;
      case 601:
	if (protocol == IPPROTO_TCP)
	    return("maitrd");
	break;
      case 602:
	if (protocol == IPPROTO_TCP)
	    return("busboy");
	break;
      case 700:
	if (protocol == IPPROTO_UDP)
	    return("acctmaster");
	break;
      case 701:
	if (protocol == IPPROTO_UDP)
	    return("acctslave");
	break;
      case 702:
	if (protocol == IPPROTO_UDP)
	    return("acct");
	break;
      case 703:
	if (protocol == IPPROTO_UDP)
	    return("acctlogin");
	break;
      case 704:
	if (protocol == IPPROTO_UDP)
	    return("acctprinter/elcsd");
	break;
      case 705:
	if (protocol == IPPROTO_UDP)
	    return("acctinfo");
	break;
      case 706:
	if (protocol == IPPROTO_UDP)
	    return("acctslave2");
	break;
      case 707:
	if (protocol == IPPROTO_UDP)
	    return("acctdisk");
	break;
      case 750: return("kerberos");
      case 751: return("kerberos_master");
      case 752:
	if (protocol == IPPROTO_UDP)
	    return("passwd_server");
	break;
      case 753:
	if (protocol == IPPROTO_UDP)
	    return("userreg_server");
	break;
      case 754:
	if (protocol == IPPROTO_TCP)
	    return("krb_prop");
	break;
      case 888:
	if (protocol == IPPROTO_TCP)
	    return("erlogin");
	break;
      case 1109:
	if (protocol == IPPROTO_TCP)
	    return("kpop");
	break;
      case 1167:
	if (protocol == IPPROTO_UDP)
	    return("phone");
	break;
      case 1524:
	if (protocol == IPPROTO_TCP)
	    return("ingreslock");
	break;
      case 1666:
	if (protocol == IPPROTO_UDP)
	    return("maze");
	break;
      case 2049:
	if (protocol == IPPROTO_UDP)
	    return("nfs");
	break;
      case 2053:
	if (protocol == IPPROTO_TCP)
	    return("knetd");
	break;
      case 2105:
	if (protocol == IPPROTO_TCP)
	    return("eklogin");
	break;
      case 5555:
	if (protocol == IPPROTO_TCP)
	    return("rmt");
	break;
      case 5556:
	if (protocol == IPPROTO_TCP)
	    return("mtb");
	break;
      case 9535:
	if (protocol == IPPROTO_TCP)
	    return("man");
	break;
      case 9536:
	if (protocol == IPPROTO_TCP)
	    return("w");
	break;
      case 9537:
	if (protocol == IPPROTO_TCP)
	    return("mantst");
	break;
      case 10000:
	if (protocol == IPPROTO_TCP)
	    return("bnews");
	else
	    return("rscs0");
      case 10001:
	if (protocol == IPPROTO_TCP)
	    return("queue");
	else
	    return("rscs1");
      case 10002:
	if (protocol == IPPROTO_TCP)
	    return("poker");
	else
	    return("rscs2");
      case 10003:
	if (protocol == IPPROTO_TCP)
	    return("gateway");
	else
	    return("rscs3");
      case 10004:
	if (protocol == IPPROTO_TCP)
	    return("remp");
	else
	    return("rscs4");
      case 10005:
	if (protocol == IPPROTO_UDP)
	    return("rscs5");
	break;
      case 10006:
	if (protocol == IPPROTO_UDP)
	    return("rscs6");
	break;
      case 10007:
	if (protocol == IPPROTO_UDP)
	    return("rscs7");
	break;
      case 10008:
	if (protocol == IPPROTO_UDP)
	    return("rscs8");
	break;
      case 10009:
	if (protocol == IPPROTO_UDP)
	    return("rscs9");
	break;
      case 10010:
	if (protocol == IPPROTO_UDP)
	    return("rscsa");
	break;
      case 10011:
	if (protocol == IPPROTO_UDP)
	    return("rscsb");
	break;
      case 10012: return("qmaster");
    }
    return("UNKNOWN");
}

char *
decode_tcp_flags(flags)
u_char flags;
{
    static char buffer[40];

    buffer[0] = 0;
    if (flags & TH_URG)
	strcat(buffer, "URG ");
    if (flags & TH_ACK)
	strcat(buffer, "ACK ");
    if (flags & TH_PUSH)
	strcat(buffer, "PUSH ");
    if (flags & TH_RST)
	strcat(buffer, "RST ");
    if (flags & TH_SYN)
	strcat(buffer, "SYN ");
    if (flags & TH_FIN)
	strcat(buffer, "FIN");
    return(buffer);
}

char *
decode_icmp_type(type)
int type;
{
    switch (type) {
      case ICMP_ECHOREPLY:
	return("Echo Reply");

      case ICMP_UNREACH:
	return("Dest Unreachable");

      case ICMP_SOURCEQUENCH:
	return("Source Quench");

      case ICMP_REDIRECT:
	return("Redirect");

      case ICMP_ECHO:
	return("Echo Request");

      case 9:
	return("Router Advertisement");

      case 10:
	return("Router Solicitation");

      case ICMP_TIMXCEED:
	return("Time Exceeded");

      case ICMP_PARAMPROB:
	return("Parameter Problem");

      case ICMP_TSTAMP:
	return("Timestamp Request");

      case ICMP_TSTAMPREPLY:
	return("Timestamp Reply");

      case ICMP_IREQ:
	return("Information Request");

      case ICMP_IREQREPLY:
	return("Information Reply");

      case ICMP_MASKREQ:
	return("Address Mask Request");

      case ICMP_MASKREPLY:
	return("Address Mask Reply");

      default:
	return("Unknown");
    }
}

char *
decode_icmp_subcode(type, subcode)
int type, subcode;
{
    switch (type) {
      case ICMP_UNREACH:
	switch (subcode) {
	  case ICMP_UNREACH_NET:
	    return("Network Unreachable");

	  case ICMP_UNREACH_HOST:
	    return("Host Unreachable");

	  case ICMP_UNREACH_PROTOCOL:
	    return("Protocol Unreachable");

	  case ICMP_UNREACH_PORT:
	    return("Port Unreachable");

	  case ICMP_UNREACH_NEEDFRAG:
	    return("Frag. Needed but Disallowed");

	  case ICMP_UNREACH_SRCFAIL:
	    return("Source Route Failed");

	  default:
	    return("Unknown");
	}
	break;

      case ICMP_REDIRECT:
	switch (subcode) {
	  case ICMP_REDIRECT_NET:
	    return("Redirect for Net");

	  case ICMP_REDIRECT_HOST:
	    return("Redirect for Host");

	  case ICMP_REDIRECT_TOSNET:
	    return("Redirect for Type of Service & Net");

	  case ICMP_REDIRECT_TOSHOST:
	    return("Redirect for Type of Service & Host");

	  default:
	    return("Unknown");
	}
	break;

      case ICMP_TIMXCEED:
	if (subcode == ICMP_TIMXCEED_INTRANS)
	    return("Time to Live Exceeded");
	else if (subcode == ICMP_TIMXCEED_REASS)
	    return("Frag Reassy Time Exceeded");
	else
	    return("Unknown");
	break;

      default:
	return("");
    }
}

char *
inet_ntoa(sp, sel)
unsigned char *sp;
int sel;
{
    static char buf[2][16];

    sprintf(buf[sel], "%d.%d.%d.%d", sp[0], sp[1], sp[2], sp[3]);
    return buf[sel];
}


/* Decide if further NetBIOS disassembler may be applied. */

int
nbport(sport, dport, protocol)
int sport, dport, protocol;
{
    if ((sport == 137 || dport == 137) && protocol == IPPROTO_UDP)
	return 1;		/* nbname */

    if ((sport == 138 || dport == 138) && protocol == IPPROTO_UDP)
	return 2;		/* nbdatagram */

    if ((sport == 139 || dport == 139) && protocol == IPPROTO_TCP)
	return 3;		/* nbsession */
    return 0;
}
