/*
** Function to dump out a DEC LAT protocol frame.
**
** Very rudimentary at this time
*/

#include <stdio.h>
#include <dnix/cheapdrv.h>
#include <sys/types.h>
#include "protos.h"

/*
** definition of a multicast message
**
** the text area is broken down as follows:
**
** XX                   group number
**
** XX string    node name
** XX string    node description
**
** XX           number of services available
**
** XX           rating for primary service
** XX string    name of primary service
** XX string    description of primary service
**
** XX           rating for alternate service (if any)
** XX string    name of alternate service (if any)
** XX string    description of alternate service (if any)
**
** XX           nclass
** XX           service 0
*/
struct multi_cast {
    u_char type;
    u_char circuit_timer;
    u_char highprot;
    u_char lowprot;
    u_char curprot;
    u_char eco_level;
    u_char multi_id;
    u_char change_flags;
    u_short frame_size;
    u_char mc_timer;
    u_char host_status;
    u_char nodegr_len;
    u_char text[256];
};

/*
** my groups and services
*/
#define NUM_GROUPS (1);
#define GROUP1 (0x01)
#define NUM_SERVICES (1)
#define SERVICE1 (1)                    /* Interactive Terminal */

/*
** definitions for header flags
*/
#define RESP_REQ (0x01)
#define TO_HOST (0x02)

/*
** definitions for header type
*/
#define MSG_RUN (0x00)
#define MSG_START (0x04)
#define MSG_STOP (0x08)
#define MSG_SERVICE (0x28)
#define MSG_COMMAND (0x30)
#define MSG_SOLICIT (0x38)
#define MSG_RESPONSE (0x3C)

/*
** header for command messages
*/
struct lat_command {
    u_char type;		/* message type and flags */
    u_char prtcl_format;	/* protocol format */
    u_char high_prtcl_ver;
    u_char low_ptrcl_ver;
    u_char cur_prtcl_ver;
    u_char cur_prtcl_eco;
    u_short frame_size;
    u_short request_identifier;
    u_short entry_identifier;
    u_char command_type;
#define CMD_SOLICIT_NONQUEUED (1)
#define CMD_SOLICIT_QUEUED (2)
#define CMD_CANCEL (3)
#define CMD_STATUS_ENTRY (4)
#define CMD_STATUS_QUEUE (5)
#define CMD_STATUS_MINE (6)
    u_char command_modifier;
#define MOD_STATUS_TIMER (0x01)
#define MOD_STATUS_CHANGE (0x02)
};

/*
** header for solicit messages
*/
struct lat_solicit {
    u_char type;		/* message type and flags */
    u_char prtcl_format;	/* protocol format */
    u_char high_prtcl_ver;
    u_char low_ptrcl_ver;
    u_char cur_prtcl_ver;
    u_char cur_prtcl_eco;
    u_short frame_size;
    u_short request_identifier;
    u_char command_type;
    u_char command_modifier;
};

/*
** header for solicit responses
*/
struct lat_response {
    u_char type;		/* message type and flags */
    u_char prtcl_format;	/* protocol format */
    u_char high_prtcl_ver;
    u_char low_ptrcl_ver;
    u_char cur_prtcl_ver;
    u_char cur_prtcl_eco;
    u_short frame_size;
    u_short request_identifier;
    u_short request_status;
    u_short node_status;
    struct etheraddr node_addr;
    u_short multi_timer;
};

/*
** basic header common to all messages
*/
struct lat_header {
    u_char type;		/* message type and flags */
#define MSG_TYPE_MASK (0xFC)
#define MSG_FLAGS_MASK (0x03)
    u_char slots;		/* number of slots present */
    u_short dest_id;		/* destination ID */
    u_short sorc_id;		/* source ID */
    u_char seq_num;		/* sequence number */
    u_char ack_num;		/* acknowledge number */
};

dumplat(buf, len, rawdata)
void *buf;
int len;
int rawdata;			/* TRUE to dump data part of frame. */
{
    struct lat_header *lat;

    lat = (struct lat_header *) buf;
    printf("\tLAT\tOpcode: %s (%d) %s %s\n",
	   lattype(lat->type & MSG_TYPE_MASK),
	   lat->type & MSG_TYPE_MASK, (lat->type & RESP_REQ) ? "RESPREQ" : "",
	   (lat->type & TO_HOST) ? "TOHOST" : "");
    switch (lat->type & MSG_TYPE_MASK) {
      case MSG_RUN: return 1;
      case MSG_START: return 1;
      case MSG_STOP: return 1;
      case MSG_SERVICE: return new_host(lat);
      case MSG_COMMAND: return 1;
      case MSG_SOLICIT: return 1;
      case MSG_RESPONSE: return 1;
      default: return 1;
    }
}

static char *
lattype(cmd)
unsigned char cmd;
{
    switch (cmd) {
      case MSG_RUN: return "RUN";
      case MSG_START: return "START";
      case MSG_STOP: return "STOP";
      case MSG_SERVICE: return "SERVICE";
      case MSG_COMMAND: return "COMMAND";
      case MSG_SOLICIT: return "SOLICIT";
      case MSG_RESPONSE: return "RESPONSE";
    }
    return "???";
}

/*
** process a new host message
*/
new_host(multi)
struct multi_cast *multi;
{
    int i, services, len;
    u_char *cp, ch, rating;

    cp = (u_char *)&multi->nodegr_len;
    cp += (*cp + 1);		/* step over the groups. */
    cp += (*cp + 1);		/* step over the node name. */
    cp += (*cp + 1);		/* step over the node desc. */
    services = *(cp++);		/* number of services avail. */

    while (services--) {	/* For each service offered... */
        /*
        ** cp now points at the primary service entry
        */
        rating = *(cp++);	/* get the rating */
        len = *(cp++);		/* length of primary name */
	printf("\t\tName: %.*s, Rating: %d, ", len, cp, rating);
	cp += len;

        len = *(cp++);		/* length of primary desc */
	printf("Desc: %.*s\n", len, cp);
        cp += len;
    }
    cp += 2;			/* Skip over # of svcs & class (SERVICE1) */
    return cp - (u_char *) multi;
}
