/*
** Functions to dump out NetBIOS in IP frames.
*/

#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include "netbios.h"
#include "protos.h"

#define TRUE (1)
#define FALSE (0)

char *decode_name();
void *alloca();

dumpnbname(buf, len, rawdata)
unsigned char *buf;
int len, rawdata;
{
    unsigned char *buffer;
    unsigned short us, rrtype;
    unsigned long ul;
    struct nmheader *nm;
    int count;

    nm = (struct nmheader *) buf;
    buffer = (unsigned char *) nm;
    printf("\tNBNAME\ttrn_id: %d, ", nm->trn_id);
    switch (nm->opcode_flags_rcode & SUBOP) {
	case OP_QUERY:   printf("Query"); break;
	case OP_REGIS:   printf("Registration"); break;
	case OP_RELEASE: printf("Release"); break;
	case OP_WACK:    printf("WACK"); break;
	case OP_REFRESH: printf("Refresh"); break;
	default:         printf("%02X (Unknown)",
				(nm->opcode_flags_rcode & SUBOP)>>11); break;
    }
    printf(" %s, flags: %02X -",
	   (nm->opcode_flags_rcode & OPCODE & RBIT) ? "RSP" : "REQ",
	   (nm->opcode_flags_rcode & FLAGS) >> 4);
    if (nm->opcode_flags_rcode & FLAGS & FL_BROAD)
	printf(" BROADCAST");
    if (nm->opcode_flags_rcode & FLAGS & FL_RA)
	printf(" RECURSAVAIL");
    if (nm->opcode_flags_rcode & FLAGS & FL_RD)
	printf(" RECURSDESIRED");
    if (nm->opcode_flags_rcode & FLAGS & FL_TC)
	printf(" TRUNCATED");
    if (nm->opcode_flags_rcode & FLAGS & FL_AA)
	printf(" AUTH_ANS");
    printf("\n\t\trcode: ");
    switch (nm->opcode_flags_rcode & RCODE) {
      case 0: printf("OK"); break;
      case FMT_ERR: printf("FMT_ERR"); break;
      case SRV_ERR: printf("SRV_ERR"); break;
      case NAM_ERR: printf("NAM_ERR"); break;
      case IMP_ERR: printf("IMP_ERR"); break;
      case RFS_ERR: printf("RFS_ERR"); break;
      case ACT_ERR: printf("ACT_ERR"); break;
      case CFT_ERR: printf("CFT_ERR"); break;
      default: printf("%X (Unknown)", nm->opcode_flags_rcode & RCODE); break;
    }
    printf(", (qdcount: %d, ancount: %d, nscount: %d, arcount: %d)\n",
	   nm->qdcount, nm->ancount, nm->nscount, nm->arcount);
    buffer += sizeof(*nm);
    if (count = nm->qdcount) while (count--) {
	printf("\t\tQuestion Name: ");
	buffer += dumpname(buf, buffer);
	printf("\t\tQuestion Type: ");
	switch (us = *(unsigned short *) buffer) {
	  case 0x0020: printf("NB"); break;
	  case 0x0021: printf("NBSTAT"); break;
	  default:  printf("%04X (Unknown)", us); break;
	}
	buffer += sizeof(us);
	printf(", Question Class: ");
	switch (us = *(unsigned short *) buffer) {
	  case 0x0001: printf("INET\n"); break;
	  default:  printf("%04X (Unknown)\n", us); break;
	}
	buffer += sizeof(us);
    }
    if (count = nm->ancount) while (count--) {
	printf("\t\tAnswer Name: ");
	buffer += dumpname(buf, buffer);
	printf("\t\tAnswer Type: ");
	switch (rrtype = *(unsigned short *) buffer) {
	  case 0x0001: printf("A"); break;
	  case 0x0002: printf("NS"); break;
	  case 0x000A: printf("NULL"); break;
	  case 0x0020: printf("NB"); break;
	  case 0x0021: printf("NBSTAT"); break;
	  default:  printf("%04X (Unknown)", rrtype); break;
	}
	buffer += sizeof(rrtype);
	printf(", Answer Class: ");
	switch (us = *(unsigned short *) buffer) {
	  case 0x0001: printf("INET\n"); break;
	  default:  printf("%04X (Unknown)\n", us); break;
	}
	buffer += sizeof(us);
	ul = *(unsigned long *) buffer;
	buffer += sizeof(ul);
	printf("\t\tTTL: %d, ", ul);
	us = *(unsigned short *) buffer;
	buffer += sizeof(us);
	printf("rdlength: %d\n", us);
	if (rrtype == 0x20) { /* NB */
	    while (us >= 6) {
		rrtype = *(unsigned short *) buffer;
		buffer += sizeof(rrtype);
		us -=  sizeof(rrtype);
		printf("\t\t%s Name,", (rrtype & 0x8000) ? "Group" : "Unique");
		printf(" Owner type: ");
		switch (rrtype>>13 & 3) {
		  case 0: printf("BNODE"); break;
		  case 1: printf("PNODE"); break;
		  case 2: printf("MNODE"); break;
		  default: printf("%X (Unknown)", rrtype>>13 & 3); break;
		}
		printf(", Owner Address: %s\n", inet_ntoa(buffer, 0));
		buffer += 4;
		us -= 4;
	    }
	} else if (rawdata) {	/* Now dump resource data. */
	    hd(buffer, us, 0);
	    buffer += us;
	}
    }
    if (count = nm->nscount) while (count--) {
	printf("\t\tAuthority Name: ");
	buffer += dumpname(buf, buffer);
	printf("\t\tAuthority Type: ");
	switch (rrtype = *(unsigned short *) buffer) {
	  case 0x0001: printf("A"); break;
	  case 0x0002: printf("NS"); break;
	  case 0x000A: printf("NULL"); break;
	  case 0x0020: printf("NB"); break;
	  case 0x0021: printf("NBSTAT"); break;
	  default:  printf("%04X (Unknown)", rrtype); break;
	}
	buffer += sizeof(rrtype);
	printf(", Authority Class: ");
	switch (us = *(unsigned short *) buffer) {
	  case 0x0001: printf("INET\n"); break;
	  default:  printf("%04X (Unknown)\n", us); break;
	}
	buffer += sizeof(us);
	ul = *(unsigned long *) buffer;
	buffer += sizeof(ul);
	printf("\t\tTTL: %d, ", ul);
	us = *(unsigned short *) buffer;
	buffer += sizeof(us);
	printf("rdlength: %d\n", us);
	if (rrtype == 0x02) {	/* NS */
	    printf("\t\tRedirect Name: ");
	    buffer += dumpname(buf, buffer);
	} else if (rawdata) {		/* Now dump resource data. */
	    hd(buffer, us, 0);
	    buffer += us;
	}
    }
    if (count = nm->arcount) while (count--) {
	printf("\t\tAdditional Name: ");
	buffer += dumpname(buf, buffer);
	printf("\t\tAdditional Type: ");
	switch (rrtype = *(unsigned short *) buffer) {
	  case 0x0001: printf("A"); break;
	  case 0x0002: printf("NS"); break;
	  case 0x000A: printf("NULL"); break;
	  case 0x0020: printf("NB"); break;
	  case 0x0021: printf("NBSTAT"); break;
	  default:  printf("%04X (Unknown)", us); break;
	}
	buffer += sizeof(rrtype);
	printf(", Additional Class: ");
	switch (us = *(unsigned short *) buffer) {
	  case 0x0001: printf("INET\n"); break;
	  default:  printf("%04X (Unknown)\n", us); break;
	}
	buffer += sizeof(us);
	ul = *(unsigned long *) buffer;
	buffer += sizeof(ul);
	printf("\t\tTTL: %d, ", ul);
	us = *(unsigned short *) buffer;
	buffer += sizeof(us);
	printf("rdlength: %d\n", us);
	if (rrtype == 0x01) {	/* A */
	    printf("\t\tRedirect Address: %s\n", inet_ntoa(buffer, 0));
	} else if (rrtype == 0x20) { /* NB */
	    rrtype = *(unsigned short *) buffer;
	    printf("\t\t%s Name,", (rrtype & 0x8000) ? "Group" : "Unique");
	    printf(" Owner type: ");
	    switch (rrtype>>13 & 3) {
	      case 0: printf("BNODE"); break;
	      case 1: printf("PNODE"); break;
	      case 2: printf("MNODE"); break;
	      default: printf("%X (Unknown)", rrtype>>13 & 3); break;
	    }
	    printf(", Owner Address: %s\n", inet_ntoa(&buffer[2], 0));
	} else if (rawdata)		/* Now dump resource data. */
	    hd(buffer, us, 0);
	buffer += us;
    }
    len -= buffer - buf;
    if (len <= 0)
	return buffer - buf;

    if (rawdata) {		/* Now dump whatever is left. */
	hd(buffer, len, 0);
	buffer += len;
    }
    return buffer - buf;
}

dumpnbdgram(buf, len, rawdata)
unsigned char *buf;
int len, rawdata;
{
    unsigned char *buffer;
    struct ndheader *nd;
    struct nderror *nde;
    int code;

    nd = (struct ndheader *) buf;
    nde = (struct nderror *) buf;
    buffer = (unsigned char *) nd;
    printf("\tNBDGRAM\t");
    code = 0;
    switch (nd->msg_type) {
      case DIRECT_UNIQUE_DGRAM: printf("Direct Unique"); break;
      case DIRECT_GROUP_DGRAM: printf("Direct Group"); break;
      case BROADCAST_DGRAM: printf("Broadcast"); break;
      case DATAGRAM_ERROR: printf("Error"); code = 1; break;
      case DATAGRAM_QUERY_REQUEST: printf("Query Request"); code = 2; break;
      case DATAGRAM_ACK_QUERY_RSP: printf("Query ACK"); code = 2; break;
      case DATAGRAM_NAK_QUERY_RSP: printf("Query NAK"); code = 2; break;
      default: printf("Unknown Type: %02X", nd->msg_type); break;
    }
    if (nd->flags & MORE)
	printf(" MORE");
    if (nd->flags & FIRST)
	printf(" FIRST");
    printf(" SNT=");
    switch (nd->flags & SNT) {
      case BNODE: printf("BNODE"); break;
      case PNODE: printf("PNODE"); break;
      case MNODE: printf("MNODE"); break;
      case NBDD: printf("NBDD"); break;
    }
    printf(" dgm_id: %d\n", nd->dgm_id);
    printf("\t\tSource %s port %d\n",
	   inet_ntoa(&nd->source_ip), nd->source_port);
    switch (code) {
      case 0: printf("\t\tlength: %d offset: %d\n",
		     nd->dgm_length, nd->packet_offset);
	buffer += sizeof(*nd);
	if (nd->flags & FIRST) {
	    printf("\t\tSource Name: ");
	    buffer += dumpname(buf, buffer);
	    printf("\t\tDestination Name: ");
	    buffer += dumpname(buf, buffer);
	}
	break;

      case 1: printf("\t\tError %02X", nde->error_code);
	switch (nde->error_code) {
	  case DNAME_MISSING: printf(" (Destination Name not Present)");
	    break;
	  case INVALID_SNAME: printf(" (Invalid Source Name Format)");
	    break;
	  case INVALID_DNAME: printf(" (Invalid Destination Name Format)");
	    break;
	  default: printf(" (Unknown)"); break;
	}
	putchar('\n');
	buffer += sizeof(*nde);
	break;

      case 2:
	buffer += sizeof(*nd);
	if (nd->flags & FIRST) {
	    printf("\t\tDestination Name: ");
	    buffer += dumpname(buf, buffer);
	}
	break;
    }
    len -= buffer - buf;
    if (len <= 0)
	return buffer - buf;

    if (rawdata) {		/* Now dump whatever is left. */
	hd(buffer, len, 0);
	buffer += len;
    }
    return buffer - buf;
}

/*
** This function is buggy, in that if a message sent through a session
** won't fit in a single frame, its fragmentation is unaccounted for.  Later
** parts of the message will probably show up as a mangled session frame,
** as we don't track sessions ourself.  In addition, any subsequent
** disassembly of the message will fail, as it's incomplete in each
** frame.
*/
dumpnbsess(buf, len, rawdata)
unsigned char *buf;
int len, rawdata;
{
    unsigned char *buffer;
    struct nsheader *ns;
    unsigned long length;
    unsigned short us;

    ns = (struct nsheader *) buf;
    buffer = (unsigned char *) ns;
    printf("\tNBSESS\t");
    if (len >= sizeof(*ns)) {
	switch (ns->type) {
	  case MESSAGE: printf("Message"); break;
	  case REQUEST: printf("Request"); break;
	  case POSITIVE: printf("Positive Response"); break;
	  case NEGATIVE: printf("Negative Response"); break;
	  case RETARGET: printf("Retarget Response"); break;
	  case KEEPALIVE: printf("Keep Alive"); break;
	  default: printf("Unknown Type: %02X", ns->type); break;
	}
	length = (ns->flags & EXTENSION) << 16 | ns->length;
	printf(", length: %ld\n", length);
	buffer += sizeof(*ns);
	switch (ns->type) {
	  case REQUEST: 
	    printf("\t\tTo: ");
	    buffer += dumpname(buf, buffer);
	    printf("\t\tFrom: ");
	    buffer += dumpname(buf, buffer);
	    break;

	  case NEGATIVE: 
	    printf("\t\tError: ");
	    buffer += dumpname(buf, buffer);
	    switch (*buffer) {
	      case 0x80: printf("Not listening on called name"); break;
	      case 0x81: printf("Not listening for calling name"); break;
	      case 0x82: printf("Called name not present"); break;
	      case 0x83: printf("Insufficient resources"); break;
	      case 0x8F: printf("Unspecified error"); break;
	      default: printf("%02X (Unknown)", *buffer); break;
	    }
	    buffer++;
	    putchar('\n');
	    break;

	  case RETARGET:
	    printf("\t\tRetarget to %s, ", inet_ntoa(buffer));
	    buffer += 4;
	    us = *(unsigned short *) buffer;
	    buffer += sizeof(us);
	    printf("port %d\n", us);
	    break;
	}
	len -= buffer - buf;
	if (len <= 0)
	    return buffer - buf;
    } else {
	printf("<empty>\n");
    }
    if (rawdata && len) {	/* Now dump whatever is left. */
	hd(buffer, len, 0);
	buffer += len;
    }
    return buffer - buf;
}

/* Dump out a name, possibly just a pointer to another name within frame. */

static int
dumpname(base, np)
unsigned char *base, *np;
{
    int offset;

    if ((*np & 0xC0) == 0xC0) {	/* Pointer? */
	offset = (*np & 0x3F) << 8 | np[1];
	dumpname(base, &base[offset]);
	return 2;
    } else
	return dumpcrap(np);
}


static void
dumpchar(chr)
unsigned char chr;
{
    if (chr >= ' ' && chr <= '~')
	putchar(chr);
    else
	printf("<%02X>");
}

/* Dump out one of those doubly-mangled NetBIOS names. */
/* These are the stupidest things I've ever seen. */

static int
dumpcrap(np)
unsigned char *np;
{
    int ii;
    unsigned char chr, *buf;

    buf = np;
    while (ii = *buf++) {
	if (ii == 0x20) {	/* 'Compression' tag? */
	    while (ii--) {
		if (ii & 1) {
		    chr = (*buf++ - 'A') << 4;
		} else {
		    chr = chr | (*buf++ - 'A');
		    dumpchar(chr);
		}
	    }
	} else {
	    dumpchar(ii);
	}
    }
    putchar('\n');
    return buf - np;
}
