/*
** Function to dump out a safelink (fileserver) protocol frame.
*/

#include <stdio.h>
#include <malloc.h>
#include <search.h>
#include <string.h>
#include <sys/types.h>
#include <dnix/cheapdrv.h>
#include <dnix/dutsname.h>
#include <dnix/fcodes.h>
#include <dnix/errno.h>
#include <net/x25.h>
#include "protos.h"
#include "ws.h"

extern int packnum;
static char Destination[32];	/* Name holders. */
static char Source[32];
static char *func;

dumpsafe(buf, len, rawdata)
void *buf;
int len;
int rawdata;			/* TRUE to dump data part of frame. */
{
    int len2;
    unsigned char *cp;
    char *ptype;
    unsigned short byte;
    short reject;
    short n_r, n_s, v_r, v_s, pf;
    struct wsbuf *tws;
    char tempstr[256];

    cp = (unsigned char *) buf;
    makename(cp, Destination);
    makename(&cp[6], Source);
    reject = 0;
    len2 = (cp[14] << 8) | cp[15];
    len = len2 - 4;		/* Length after L2 header taken. */
    fputs("\tSAFE ", stdout);
    switch (cp[16]) {		/* Level 2 address type */
      case 1:
	fputs("B\t", stdout);
	break;
      case 3:
	fputs("A\t", stdout);
	break;
      default:
	fputs("?\t", stdout);
	break;
    }
    fputs("L2 ", stdout);
 /* Level 2 control byte */
 /* assume we want sequence numbers, and poll/final bit */
    n_s = (cp[17] >> 1) & 7;
    n_r = (cp[17] >> 5) & 7;
    pf = (cp[17] >> 4) & 1;
    if ((cp[17] & 0x01) == 0) {
    /* information */
	printf("INFO %s         N(S): %d  N(R): %d  ",
	       pf ? "P " : "- ", n_s, n_r);
	tws = (struct wsbuf *) & cp[14];	/* 1st 'int' isn't in the
						 * frame */
	dump_wsbuf(tws);
	if (rawdata && tws->w_bufsz) {
	    putchar('\n');
	    hd((unsigned char *) &tws->w_buf, tws->w_bufsz, 0);
	} else
	    putchar('\n');
    } else {
	if (!(cp[17] & 0x02)) {	/* Supervisory. */
	    byte = cp[17] & 0x0F;
	    if (byte == X25_RR)
		func = "RR  ";
	    else if (byte == X25_RNR)
		func = "RNR ";
	    else if (byte == X25_REJ)
		func = "REJ ";
	    else {
		sprintf(tempstr, "-%02X-", byte);
		func = tempstr;
	    }
	    printf("%s %s                  N(R): %d\n", func,
		   pf ? "PF" : "- ", n_r);
	} else {		/* Unnumbered. */
	    ptype = "PF";
	    byte = cp[17] & ~X25_POLL;
	    switch (byte) {
	      case X25_MUI:
		func = "MUI ";
		break;
	      case X25_SABM:
		func = "SABM";
		ptype = "P";
		break;
	      case X25_DM:
		func = "DM  ";
		ptype = "F";
		break;
	      case X25_DISC:
		func = "DISC";
		ptype = "P";
		break;
	      case X25_UA:
		func = "UA  ";
		ptype = "F";
		break;
	      case X25_FRMR:
		func = "FRMR";
		reject = 1;
		ptype = "F";
		break;
	      case X25_XID:
		func = "XID ";
		break;
	      default:
		sprintf(tempstr, "-%02X-", cp[17]);
		func = tempstr;
		break;
	    }
	    printf("%s %-2s\n", func, pf ? ptype : "-");
	    if (byte == X25_MUI) {
		len--;
		switch (cp[18]) {
		  case '1':
		    printf("\t\t\tSearch for primary service '%.*s'\n",
			   len, &cp[19]);
		    break;
		  case '2':
		    printf("\t\t\tSearch for secondary service '%.*s'\n",
			   len, &cp[19]);
		    break;
		  case '3':
		    printf("\t\t\tSearch for tertiary service '%.*s'\n",
			   len, &cp[19]);
		    break;
		  case '!':
		    printf("\t\t\tCan supply service '%.*s'\n", len, &cp[19]);
		    break;
		  default:
		    printf("\t\t\t<<Unknown -- $%02X>>\n", cp[18]);
		    break;
		}
	    }
	}
    }
    if (reject) {		/* Unnumbered Frame Rejected, show why... */
	short bits;

	bits = cp[20] & 0xF;
	switch (bits) {
	  case 1:
	    func = "Invalid Control field";
	    break;
	  case 3:
	    func = "Invalid I-field";
	    break;
	  case 4:
	    func = "I-field too large";
	    break;
	  case 8:
	    func = "Invalid N(R)";
	    break;
	  default:
	    func = "<Unknown reason>";
	    break;
	}
	printf("\t\tFRMR info: Rejected CONTROL field: %02X, "
	       "V(S) & V(R) = %02X, %s (%02X)\n",
	       cp[18], cp[19], func, cp[20]);
    }
    rawdata = &cp[14 + len2] - (unsigned char *) buf;
    len = (char *) &tws->w_buf - (char *) buf;
    return (rawdata > len) ? len : rawdata;
}

static char *fh_fn[] = {"FH_ALLOC", "FH_DEALLOC", "FH_LOAD",
			    "FH_STAT", "FH_CLOSE", "FH_CLOSERQ",
			    "FH_CTTY", "FH_READ", "FH_WRITE",
			    "FH_IPMNT", "FH_IPUMNT", "FH_SWAPIN",
			    "FH_SWAPOUT", "FH_START", "FH_STOP",
			    "FH_KILL", "FH_CLOCK", "FH_ACCESS",
			    "FH_QSTAT"
};

static char *f_fn[] = {"F_OPEN", "F_MVDIR", "F_CREATE",
			   "F_READ", "F_WRITE", "F_IOCR",
			   "F_IOCW", "F_UNLINK", "F_LINK",
			   "F_LOCATE", "F_CPROT", "F_CUID",
			   "F_CGID", "F_MOUNT", "F_MKNOD",
			   "F_FPROT", "F_EXCLUS", "F_PIPE",
			   "F_MKDIR", "F_RMDIR", "F_ALLOC",
			   "F_LOCKF", "F_UWRITE", "F_UREAD",
			   "F_WAIT", "F_PAUSE", "F_TRACE",
			   "F_DNIXSEM", "F_NAP", "F_TERMIN",
			   "F_TRPLU", "F_CANCEL", "F_CLOSE",
			   "F_SEEK", "F_GTIME", "F_STIME",
			   "F_SPRI", "F_SUID", "F_SGID",
			   "F_OTQ", "F_FORK", "F_EXEC",
			   "F_SPAWN", "F_AMOUNT", "F_PASSRQ",
			   "F_INPUT", "F_OUTPUT", "F_MINP",
			   "F_MOUT", "F_AUTO", "F_GPID",
			   "F_GUID", "F_GEUID", "F_GGID",
			   "F_GEGID", "F_SCMASK", "F_DLU",
			   "F_SDLU", "F_SLUF", "F_ORQ",
			   "F_EXIT", "F_SIGNAL", "F_ALARM",
			   "F_KILL", "F_SPLIM", "F_BREAK",
			   "F_DSPR", "F_SPG", "F_LOGC",
			   "F_TMKIO", "F_PSTAT", "F_OLURQ",
			   "F_unassigned", "F_WHREQ", "F_T2URD",
			   "F_T2UWR", "F_SEGP", "F_FCNTL",
			   "F_GPGRP", "F_SPGRP", "F_GPPID",
			   "F_UNAME", "F_RDREQ", "F_T1REQ",
			   "F_GROI", "F_MEMRES", "F_GIVEUPCPU"
};

static char *f2_fn[] = {"F_BSDSTUFF", "F_LOGSTRUCT",
			    "F_TESTTQ", "F_LOCKMEM", "F_WSET",
			    "F_SYSINFO", "F_FSSCTL",
			    "F_FREEMEM", "F_VDATA"
};

char *
func_name(int func)
{
    if (func >= 1 && func <= 87)
	return (f_fn[func - 1]);
    if (func >= 100 && func <= 118)
	return (fh_fn[func - 100]);
    if (func >= 119 && func <= 127)
	return (f2_fn[func - 119]);
    return ("???");
}


void
i2s(buf, num)
char *buf;
int num;
{
    if (num <= 0)
	sprintf(buf, "%d,", num);
    else
	sprintf(buf, "%02X,", num);
}

static char *tracker(struct wsbuf *tws);

dump_wsbuf(tws)
struct wsbuf *tws;
{
    char t1[32], t2[32];

    i2s(t1, tws->w_id);
    i2s(t2, tws->w_rpar);
    printf("\n\t  w_fc   = %08X %s %s\n", tws->w_fc, func_name(tws->w_fc),
	   tracker(tws));
    printf("\t  w_id   = %08X,  w_rpar  = %08X,  w_bufsz  = %08X (%d)\n",
	   tws->w_id, tws->w_rpar, tws->w_bufsz, tws->w_bufsz);
    printf("\t  w_pri  = %08X,  w_flags = %08X,  w_oflag  = %08X\n",
	   tws->w_pri, tws->w_flags, tws->w_oflag);
    printf("\t  w_bcnt = %08X,  w_dev   = %08X,  w_offlag = %08X\n",
	   tws->w_bcnt, tws->w_dev, tws->w_offlag);
    printf("\t  w_prad = %08X,  w_uid   = %08X,  w_rdfid  = %08X\n",
	   tws->w_prad, tws->w_uid, tws->w_rdfid);
    printf("\t  w_fid  = %08X,  w_gid   = %08X,  w_ext1   = %08X\n",
	   tws->w_fid, tws->w_gid, tws->w_ext1);
    printf("\t  w_par  = %08X,  w_nph   = %08X,  w_ext2   = %08X\n",
	   tws->w_par, tws->w_nph, tws->w_ext2);
    printf("\t  w_kpar = %08X,  w_npl   = %08X,  w_tid    = %08X",
	   tws->w_kpar, tws->w_npl, tws->w_tid);
}

/*
** Attempt to track filenames versus file-id's for easier human
** reading.  This only works if you see the open go by.
*/

static char *
tracker(tws)
struct wsbuf *tws;
{
    int doname;
    char *name;
    static char buf[256], buf2[4] = {"..."};

    if (name = findoverride(packnum)) {
	newname(tws->w_npl, tws->w_fid, name, strlen(name), 0, 0);
    }
    buf[0] = 0;
    doname = 1;
    switch (tws->w_fc) {
      case FH_QSTAT:		/* One-shot named commands, no name history. */
      case F_UNLINK:		/* These commands use filenames, and don't */
      case F_MOUNT:		/* create FID's from them.  Because we don't */
      case F_CPROT:		/* remember the names, we can't show it */
      case F_CUID:		/* on the acknowledgement packet. */
      case F_CGID:
      case F_MKDIR:
      case F_RMDIR:
      case F_MKNOD:
	if (!tws->w_rpar)
	    newname(tws->w_npl, tws->w_nph | tws->w_fc, tws->w_buf,
		    tws->w_bufsz, buf, sizeof(buf));
	else {
	    delname(tws->w_npl, tws->w_nph | tws->w_fc, buf, sizeof(buf));
	    if (tws->w_rpar == ESWAP)
		strncat(buf, " -- Handler", sizeof(buf));
	    else if (tws->w_rpar < 0) {
		strncat(buf, " ", sizeof(buf));
		strncat(buf, errnum(tws->w_rpar), sizeof(buf));
	    }
	}
	pidtoname(tws->w_npl, tws->w_nph >> 16, buf, sizeof(buf), 1);
	return buf;
	break;

      case F_OPEN:		/* Persistent name history addition. */
      case F_LOCATE:		/* These commands add a FID to the table, */
      case F_CREATE:		/* based on an ASCII filename. */
	if (!tws->w_rpar) {
	    newname(tws->w_npl, tws->w_nph | tws->w_fc, tws->w_buf,
		    tws->w_bufsz, buf, sizeof(buf));
	} else if (tws->w_rpar < 0) {
	    delname(tws->w_npl, tws->w_nph | tws->w_fc, buf, sizeof(buf));
	    if (tws->w_rpar == ESWAP) {
		strncat(buf, " -- Handler", sizeof(buf));
	    } else {
		strncat(buf, " ", sizeof(buf));
		strncat(buf, errnum(tws->w_rpar), sizeof(buf));
	    }
	} else {
	    newfid(tws->w_npl, tws->w_nph | tws->w_fc, tws->w_rpar,
		   buf, sizeof(buf));
	}
	pidtoname(tws->w_npl, tws->w_nph >> 16, buf, sizeof(buf), 1);
	return buf;
	break;

      case FH_CLOSE:		/* Remove a name from the name history. */
	delname(tws->w_npl, tws->w_fid, buf, sizeof(buf));
	pidtoname(tws->w_npl, tws->w_nph >> 16, buf, sizeof(buf), 1);
	return buf;
	break;

      case FH_LOAD:		/* Commands to use the name history. */
	addpid(tws->w_npl, tws->w_nph >> 16, tws->w_fid);
	doname = 0;		/* Fall through! */
      case F_READ:		/* These commands have only a FID. */
      case FH_READ:
      case F_WRITE:
      case FH_WRITE:
      case F_IOCR:
      case F_IOCW:
      case FH_STAT:
      case FH_ACCESS:
      case F_ALLOC:
	makefname(tws->w_npl, tws->w_fid, buf, sizeof(buf));
	if (tws->w_rpar < 0) {
	    strncat(buf, " ", sizeof(buf));
	    strncat(buf, errnum(tws->w_rpar), sizeof(buf));
	}
	pidtoname(tws->w_npl, tws->w_nph >> 16, buf, sizeof(buf), doname);
	return buf;
	break;

				/* Link uses the name history _and_ has */
      case F_LINK:		/* another filename buried in the packet. */
	if (!tws->w_rpar) {	/* Initial request. */
	    makefname(tws->w_npl, tws->w_par, buf, sizeof(buf));
	    strncat(buf, " to ", sizeof(buf));
	    newname(tws->w_npl, tws->w_nph | tws->w_fc, tws->w_buf,
		    tws->w_bufsz, buf, sizeof(buf));
	} else {
	    strcpy(buf, "to ");
	    delname(tws->w_npl, tws->w_nph | tws->w_fc, &buf[3],sizeof(buf)-3);
	    if (tws->w_rpar < 0) { /* An error. */
		if (tws->w_rpar == ESWAP) {
		    strncat(buf, " -- Handler", sizeof(buf));
		} else {
		    strncat(buf, " ", sizeof(buf));
		    strncat(buf, errnum(tws->w_rpar), sizeof(buf));
		}
	    }
	}
	pidtoname(tws->w_npl, tws->w_nph >> 16, buf, sizeof(buf), 1);
	return buf;
	break;

				/* Everything else just show the */
      default:			/* name of the process doing it. */
	pidtoname(tws->w_npl, tws->w_nph >> 16, buf, sizeof(buf), 1);
	return buf;
    }
    return "";
}

/*
** Handle a lookup tree for resolving file ID numbers and names.
**
** We use the SVID's tsearch package.  Internally this uses nodes of
** the form:  struct node { void *key; struct node *left; struct node *right;};
** where the key field is a pointer to one of our node structures.  The
** tsearch package takes care of creating/deleting its internal nodes.
** NB: the root pointers actually point to the internal node structures,
** not the external ones, though I've declared them as the latter.  Since
** the root pointers are only looked at by the tsearch package this
** 'error' is inconsequential.  The tsearch package is poorly documented,
** and part of it even lies!
*/

static struct node {
    long nodeid;
    long pidfileid;
    char *name;
} *fidroot, *pidroot;

/*
** Node comparitor.  Compares (numerically) the two key fields of our node.
*/

static int
ncompare(n1, n2)
struct node *n1, *n2;
{
    unsigned char *cp1, *cp2;
    int i;

    cp1 = (unsigned char *) &n1->nodeid;
    cp2 = (unsigned char *) &n2->nodeid;
    i = sizeof(n1->nodeid) + sizeof(n1->pidfileid);
    while (i--) {
	if (*cp1 != *cp2) {
	    if (*cp1 > *cp2) {
		return 1;
	    } else {
		return -1;
	    }
	}
	cp1++;
	cp2++;
    }
    return 0;
}

/*
** Convert a node/file ID to a name using the search tree.
*/

static void
makefname(nodeid, fileid, to, size)
long nodeid;
short fileid;
char *to;
{
    struct node n, **npp, **tfind();

    n.nodeid = nodeid;
    n.pidfileid = fileid;
    if (npp = tfind(&n, &fidroot, ncompare)) {
	strncat(to, (*npp)->name, size);
    }
}

/*
** Add a new filename to the tree.  The fileID has the requesting PID combined
** with the function code in it at first, and will be filled in later with the
** real FID (when we see that packet).
*/

static void
newname(nodeid, fileid, name, nsize, to, size)
long nodeid, fileid;
char *name, *to;
int nsize, size;
{
    char *store;
    struct node *nstore;

    if (store = malloc(nsize + 1)) { /* Store the node. */
	strncpy(store, name, nsize);
	store[nsize] = 0;
	if (nstore = (struct node *) malloc(sizeof(struct node))) {
	    nstore->nodeid = nodeid;
	    nstore->pidfileid = fileid;
	    nstore->name = store;
	    tsearch(nstore, &fidroot, ncompare);
	    if (to)
		strncat(to, store, size);
	} else {
	    free(store);
	}
    }
}

/*
** Add a pid reference to the table.
*/

static void
addpid(nodeid, pid, fileid)
long nodeid;
short pid, fileid;
{
    struct node n, *nstore, **npp, **tfind();
    char *cp;

    n.nodeid = nodeid;
    n.pidfileid = pid;
    if (!(npp = tfind(&n, &pidroot, ncompare))) { /* No existing PID name. */
	n.pidfileid = fileid;
	if (npp = tfind(&n, &fidroot, ncompare)) {
	    n.name = cp = (*npp)->name;
	    if (cp = strrchr(cp, '/'))
		n.name = cp + 1;
	    if (nstore = (struct node *) malloc(sizeof(struct node))) {
		if (cp = malloc(strlen(n.name) + 1)) {
		    strcpy(cp, n.name);
		    n.name = cp;
		    n.pidfileid = pid;
		    *nstore = n;
		    tsearch(nstore, &pidroot, ncompare);
		} else {
		    free(nstore);
		}
	    }
	}
    } else {			/* PID name exists. */
	nstore = *npp;
	n.pidfileid = fileid;	/* Is it the same? */
	if (npp = tfind(&n, &fidroot, ncompare)) { /* Have we a filename? */
	    n.name = cp = (*npp)->name;
	    if (cp = strrchr(cp, '/'))
		n.name = cp + 1;
	    if (strcmp(nstore->name, n.name)) {	/* Name's different. */
		free(nstore->name); /* Replace it. */
		if (nstore->name = malloc(strlen(n.name) + 1)) {
		    strcpy(nstore->name, n.name);
		}
	    }
	}
    }
}

/*
** Convert a PID to a name using the search tree.
*/

static void
pidtoname(nodeid, pid, to, size, doname)
long nodeid;
short pid;
char *to;
int size, doname;
{
    struct node n, **npp, **tfind();
    char buf[16];

    n.nodeid = nodeid;
    n.pidfileid = pid;
    if (npp = tfind(&n, &pidroot, ncompare)) {
	sprintf(buf, ", PID %d ", pid);
	strncat(to, buf, size);
	if (doname)
	    strncat(to, (*npp)->name, size);
    }
}

/*
** Look up the nodeid/pid, and replace the pid with the fileid.
** (Two-phase open, we get the name and the fid in separate packets,
**  so we use the pid in place of the fid temporarily.)
*/

static void
newfid(nodeid, pid, fileid, to, size)
long nodeid, pid;
short fileid;
char *to;
int size;
{
    struct node n, *nstore, **nfound, **tfind();

    n.nodeid = nodeid;
    n.pidfileid = pid;
    if (nfound = tfind(&n, &fidroot, ncompare)) {
	nstore = *nfound;
	n.name = nstore->name;
	tdelete(&n, &fidroot, ncompare);
	free(nstore);
	n.pidfileid = fileid;
	if (nfound = tfind(&n, &fidroot, ncompare)) { /* In case of */
	    nstore = *nfound;	/* duplicates due to errors in the */
	    free(nstore->name); /* fileserver!  (ONP did it.) */
	    tdelete(nstore, &fidroot, ncompare);
	    free(nstore);
	}
	if (fileid >= 0
	    && (nstore = (struct node *) malloc(sizeof(struct node)))) {
	    nstore->nodeid = nodeid;
	    nstore->pidfileid = fileid;
	    nstore->name = n.name;
	    tsearch(nstore, &fidroot, ncompare);
	    strncat(to, n.name, size);
	} else {
	    free(n.name);
	}
    }
}

/*
** Look up the nodeid/fileid, and delete it.  Before deleting it return
** the filename (if the nodeid/fileid record existed).
*/

static void
delname(nodeid, fileid, buf, size)
long nodeid, fileid;
char *buf;
int size;
{
    struct node n, *nstore, **nfound, **tfind();

    buf[0] = 0;
    n.nodeid = nodeid;
    n.pidfileid = fileid;
    if (nfound = tfind(&n, &fidroot, ncompare)) {
	nstore = *nfound;
	strncpy(buf, nstore->name, size);
	free(nstore->name);
	tdelete(&n, &fidroot, ncompare);
	free(nstore);
    }
}

/*
** Subroutines to print out the filename tree (for debugging).
*/

#if 0
printnode(node, order, level)
struct node **node;
int order, level;
{
    if (order == preorder || order == leaf) {
	printf("%s is $%X $%X\n", (*node)->name, (*node)->nodeid,
	       (*node)->pidfileid);
    }
}

debuglist(root)
struct node *root;
{
    printf("Tree contains:\n");
    twalk(root, printnode);
}
#endif

char *errs[] = {
    "",
    "EPERM",			/* -1 Not owner */
    "ENOENT",			/* -2 No such file or directory */
    "ESRCH",			/* -3 No such process */
    "EINTR",			/* -4 Interrupted system call */
    "EIO",			/* -5 I/O error */
    "ENXIO",			/* -6 No such device or address */
    "E2BIG",			/* -7 Arg list too long */
    "ENOEXEC",			/* -8 Exec format error */
    "EBADF",			/* -9 Bad file number */

    "ECHILD",			/* -10 No children */
    "EAGAIN",			/* -11 No more processes */
    "ENOMEM",			/* -12 Not enough core */
    "EACCES",			/* -13 Permission denied */
    "EFAULT",			/* -14 Bad address */
    "ENOTBLK",			/* -15 Block device required */
    "EBUSY",			/* -16 Mount device busy */
    "EEXIST",			/* -17 File exists */
    "EXDEV",			/* -18 Cross-device link */
    "ENODEV",			/* -19 No such device */

    "ENOTDIR",			/* -20 Not a directory */
    "EISDIR",			/* -21 Is a directory */
    "EINVAL",			/* -22 Invalid argument */
    "ENFILE",			/* -23 File table overflow */
    "EMFILE",			/* -24 Too many open files */
    "ENOTTY",			/* -25 Not a typewriter */
    "ETXTBSY",			/* -26 Text file busy */
    "EFBIG",			/* -27 File too large */
    "ENOSPC",			/* -28 No space left on device */
    "ESPIPE",			/* -29 Illegal seek */

    "EROFS",			/* -30 Read-only file system */
    "EMLINK",			/* -31 Too many links */
    "EPIPE",			/* -32 Broken pipe */
    "EDOM",			/* -33 Argument too large */
    "ERANGE",			/* -34 Result too large */
    "ENOMSG",			/* -35 No message of desired type */
    "EIDRM",			/* -36 Identifier removed */
    "ECHRNG",			/* -37 Channel number out of range */
    "EL2NSYNC",			/* -38 Level 2 not syncronized */
    "EL3HLT",			/* -39 Level 3 halted */

    "EL3RST",			/* -40 Level 3 reset */
    "ELNRNG",			/* -41 Link number out of range */
    "EUNATCH",			/* -42 Protocol driver not attached */
    "ENOCSI",			/* -43 No CSI structure availible */
    "EL2HLT",			/* -44 Level 2 halted */
    "EDEADLK",			/* -45 Record locking deadlock */
    "ENOLCK",			/* -46 No lock structures available */

    "", "", "", "", "", "", "", "", "", "", "", "", "",
/*
 *	5.3 additions
 */
    "ENOSTR",			/* -60 Device not a stream */
    "ENODATA",			/* -61 no data (for no delay io) */
    "ETIME",			/* -62 timer expired */
    "ENOSR",			/* -63 out of streams resources */

    "ENONET",			/* -64 Machine is not on the network */
    "ENOPKG",			/* -65 Package not installed */
    "EREMOTE",			/* -66 The object is remote */
    "ENOLINK",			/* -67 the link has been severed */
    "EADV",			/* -68 advertise error */
    "ESRMNT",			/* -69 srmount error */

    "ECOMM",			/* -70 Communication error on send */
    "EPROTO",			/* -71 Protocol error */
    "", "",
    "EMULTIHOP",		/* -74 multihop attempted */
    "ELBIN",			/* -75 Inode is remote (not really error) */
    "EDOTDOT",			/* -76 Cross mount point (not really error) */
    "EBADMSG",			/* -77 trying to read unreadable message */
    "ENAMETOOLONG",		/* -78 File name too long */
    "EOVERFLOW",		/* -79 Value too large for data type */

    "ENOTUNIQ",			/* -80 given log. name not unique */
    "EBADFD",			/* -81 f.d. invalid for this operation */
    "EREMCHG",			/* -82 Remote address changed */

/* shared library problems */
    "ELIBACC",			/* -83 Can't access a needed shared lib. */
    "ELIBBAD",			/* -84 Accessing a corrupted shared lib. */
    "ELIBSCN",			/* -85 .lib section in a.out corrupted. */
    "ELIBMAX",			/* -86 Attempting to link in too many libs. */
    "ELIBEXEC",			/* -87 Attempting to exec a shared library. */
    "EILSEQ",			/* -88 Illegal byte sequence */
    "ENOSYS",			/* -89 Function not implemented */

    "ELOOP",			/* -90 Too many levels of symbolic links */
    "ESTART",			/* -91 Restart interrupted system call */
    "ESTRIPE",			/* -92 Streams pipe error */
    "ENOTEMPTY",		/* -93 Directory not empty */
    "EUSERS",			/* -94 Too many users */
    "ENOTSOCK",			/* -95 Socket operation on non-socket */
    "EDESTADDREQ",		/* -96 Destination address required */
    "EMSGSIZE",			/* -97 Message too long */
    "EPROTOTYPE",		/* -98 Protocol wrong type for socket */
    "ENOPROTOOPT",		/* -99 Protocol not available */

    "", "", "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "", "",

/*
 *	5.2 DNIX Specials (-119 - -128) These differ from SYS 5.4
 */
    "ENONODES",			/* -119 No more nodes */
    "EREQAB",			/* -120 Request aborted */
    "EBADFC",			/* -121 Illegal function code */
    "ESWAP",			/* -122 Refered process is swapped */
    "ECANCL",			/* -123 Refered request is canceled */
    "EDRVNF",			/* -124 Device driver not found */
    "ETIMOUT",			/* -125 Timeout on driver */
    "ELOCKOK",			/* -126 Locking is ok - no open node needed */
    "ENOLOCK",			/* -127 No locks for this process any more */
    "EWAIT",			/* -128 Request waiting for termination */
    "",
    "EPROTONOSUPPORT",		/* -130 Protocol not supported */
    "ESOCKTNOSUPPORT",		/* -131 Socket type not supported */
    "EOPNOTSUPP",		/* -132 Op not supported on transport endpoint */
    "EPFNOSUPPORT",		/* -133 Protocol family not supported */
    "EAFNOSUPPORT",		/* -134 Address family not supp by proto family */
    "EADDRINUSE",		/* -135 Address already in use */
    "EADDRNOTAVAIL",		/* -136 Can't assign requested address */
    "ENETDOWN",			/* -137 Network is down */
    "ENETUNREACH",		/* -138 Network is unreachable */
    "ENETRESET",		/* -139 Network dropped connection on reset */

    "ECONNABORTED",		/* -140 Software caused connection abort */
    "ECONNRESET",		/* -141 Connection reset by peer */
    "ENOBUFS",			/* -142 No buffer space available */
    "EISCONN",			/* -143 Transport endpoint already connected */
    "ENOTCONN",			/* -144 Transport endpoint not connected */

    "", "", "", "", "", "", "", "",
    "ESHUTDOWN",		/* -153 Can't send after transport endpoint shutdown */
    "ETOOMANYREFS",		/* -154 Too many references: cannot splice */
    "ETIMEDOUT",		/* -155 Connection timed out */
    "ECONNREFUSED",		/* -156 Connection refused */
    "EHOSTDOWN",		/* -157 Host is down */
    "EHOSTUNREACH",		/* -158 No route to host */
    "EALREADY",			/* -159 Operation already in progress */

    "EINPROGRESS",		/* -160 Operation now in progress */
    "ESTALE",			/* -161 Stale NFS handle */
};

/*
** Convert error number to a string.
*/

static char *
errnum(err)
int err;
{
    static char buf[20];

    if (err < 0 && -err < sizeof(errs)/sizeof(errs[0])) {
	sprintf(buf, "%s (%d)", errs[-err], err);
	return buf;
    } else
	return "";
}
