/*
** Dump out 802.3 SAP packets.
*/

#include <stdio.h>
#include <string.h>
#include <dnix/cheapdrv.h>
#include <net/x25.h>
#include "protos.h"
#include "sapsnap.h"
#include "protocols.h"

dumpsap(p, size, rawdata)
void *p;
int size;
int rawdata;
{
    struct saphdr *sap;
    char *cp, buf2[32], line[132], *ptype;
    int len;
    int rseq;			/* Receive sequence #, N(R) from X.25 doc. */
    int sseq;			/* Send sequence #, N(S) from X.25 doc. */
    int pf;			/* Poll/Final bit from X.25 doc. */
    int ssap, dsap;
    unsigned short l2cmd;

    sap = (struct saphdr *) p;
    cp = (char *) &sap[1];
    switch (dsap = (sap->dsap & ~1)) {
      case IP_SAP:
	ptype = "IP";
	break;

      case ARP_SAP:
	ptype = "ARP";
	break;

      case SPAN_SAP:
	ptype = "Spanning Tree";
	break;

      case NETBIOS_SAP:
	ptype = "NETBIOS";
	break;

      case NETWARE_SAP:
	ptype = "NETWARE";
	break;

      default:
	ptype = "Unknown";
	break;
    }
    printf("\tSAP %s\t802.3 Length: %d, DSAP: $%02X%s, "
	   "SSAP: $%02X (%s)\n", (sap->ssap & 1) ? "CMD" : "RSP",
	   len = sap->eth.protocol, sap->dsap,
	   (sap->dsap & 1) ? " (group)" : "", sap->ssap, ptype);

    ssap = sap->ssap & ~1;
    len -= 4;			/* Normal SAP header. */
    line[0] = '\0';		/* Clear any existing line. */
    strcat(line, "\t\tL2 ");
    if (sap->ssap & 1)
	ptype = "P";
    else
	ptype = "F";
    sseq = rseq = -1;
    l2cmd = sap->control;
    l2cmd = (l2cmd >> 8) | (l2cmd << 8);
    pf = (l2cmd & 0x100);	/* Split out poll/final bit. */
    if (!(l2cmd & 1)) {		/* Info frame? */
	strcat(line, "INFO  ");
	sseq = (l2cmd >> 1) & 0x7F;
	rseq = l2cmd >> 9;
    } else {			/* A Supervisory or Unnumbered frame. */
	if ((l2cmd & 0x03) == 0x01) { /* Supervisory frame. */
	    rseq = l2cmd >> 9;
	    switch (l2cmd & 0x0F) {
	      case X25_RR:	/* Receiver Ready */
		strcat(line, "RR    ");
		break;
	      case X25_RNR:	/* Receiver Not Ready */
		strcat(line, "RNR   ");
		break;
	      case X25_REJ:	/* Reject */
		strcat(line, "REJ   ");
		break;
	      default:
		sprintf(buf2, "$%04X ", l2cmd);
		strcat(line, buf2);
		break;
	    }
	} else {		/* Unnumbered (non-sequenced) frame. */
	    l2cmd &= 0xFF;	/* Only an 8-bit command. */
	    cp--;
	    len++;
	    pf = (l2cmd & X25_POLL); /* Split out poll/final bit. */
	    switch (l2cmd & ~X25_POLL) {
	      case 0x03:	/* UI command */
		strcat(line, "UI    ");
		break;

	      case X25_XID:	/* Exchange ID command. */
		strcat(line, "XID   ");
		break;

	      case 0xE3:	/* Test command. */
		strcat(line, "TEST  ");
		break;

	      case 0x6F:	/* SABME command */
		strcat(line, "SABME ");
		break;

	      case X25_DISC:	/* Disconnect Command. */
		strcat(line, "DISC  ");
		break;

	      case X25_UA:	/* Unnumbered Acknowledgement. */
		strcat(line, "UA    ");
		break;

	      case X25_DM:	/* Disconnected Mode. */
		strcat(line, "DM    ");
		break;

	      case X25_FRMR:	/* Frame Reject. */
		strcat(line, "FRMR  ");
		break;

	      default:
		sprintf(buf2, "-%02X-  ", l2cmd);
		strcat(line, buf2);
		break;
	    }
	}
    }
    sprintf(buf2, "%1s         ", pf ? ptype : "-");
    strcat(line, buf2);
    if (sseq < 0)
	strcat(line, "          ");
    else {
	sprintf(buf2, "N(S): %2d  ", sseq);
	strcat(line, buf2);
    }
    if (rseq < 0)
	strcat(line, "          ");
    else {
	sprintf(buf2, "N(R): %2d  ", rseq);
	strcat(line, buf2);
    }
    puts(line);
    switch (dsap) {
      case IP_SAP:
	pf = dumpip(cp, len, rawdata);
	len -= pf;
	cp += pf;
	break;

      case ARP_SAP:
	pf = dumparp(cp, len, rawdata);
	len -= pf;
	cp += pf;
	break;

      case SPAN_SAP:
	pf = dumpbpdu(cp, len, rawdata);
	len -= pf;
	cp += pf;
	break;

      case NETBIOS_SAP:
	break;

      case NETWARE_SAP:
	pf = dumpxns(cp, len, rawdata);
	len -= pf;
	cp += pf;
	break;
    }
    if (len > 0 && rawdata)
	hd((unsigned char *) cp, len, 0);
    return cp - (char *) p;
}
