/*
** Function to dump out a Wincom protocol frame.
*/

#include <stdio.h>
#include <sys/types.h>
#include <dnix/cheapdrv.h>
#include <dnix/dutsname.h>
#include "protos.h"
#include "wincom.h"

static char *device(unsigned char ch);
static int dumpwin(void *p, int size, int rawdata, char *string);

dumpwincom(buf, len, rawdata)
void *buf;
int len;
int rawdata;			/* TRUE to dump data part of frame. */
{
    unsigned char *cp, cmd;
    int len2;

    cp = (unsigned char *) buf + 16; /* Move to Wincom data. */
    len2 = ((cp[-2] << 8) | cp[-1]) - 2;
    return dumpwin(cp, len2, rawdata, "\tWINCOM\t") + 16;
}

static
dumpwin(buf, len, rawdata, str)
void *buf;
int len;
int rawdata;			/* TRUE to dump data part of frame. */
char *str;			/* String to put at front of first line. */
{
    unsigned char *cp, *cp2, cmd;

    fputs(str, stdout);
    cp = (unsigned char *) buf;
    if ((*cp & CONTROL) != CONTROL) {
	printf("Invalid packet, byte = $%02X\n", *cp);
    } else {
	switch ((cmd = *cp++) & CODE) {
	  case STX:
	    printf("STX %s\n", device(cmd));
	    if (len > 3) {
		switch (cmd & DEVICE) {
		  case KBD_ID:
		    cp += dumpkbd(cp, len - 3, rawdata);
		    break;

		  case VAL_ID:
		    cp += dumpval(cp, len - 3, rawdata);
		    break;

		  case DOC_ID:
		    cp += dumpdoc(cp, len - 3, rawdata);
		    break;

		  case SRW_ID:
		    cp += dumpmag(cp, len - 3, rawdata);
		    break;

		  case PIN_ID:
		    cp += dumppin(cp, len - 3, rawdata);
		    break;

		  case SDO_ID:
		    cp += dumpsdo(cp, len - 3, rawdata);
		    break;
		}
		printf("\t\tLRC (%02X)\n", *cp++);
		cp += dumpwin(cp, 1, rawdata, "\t\t"); /* Do ETX. */
	    } else {
		fputs("\t\tPacket too short!\n", stdout);
	    }
	    break;

	  case ETX:
	    printf("ETX %s\n", device(cmd));
	    break;

	  case RESET:
	    printf("RESET %s\n", device(cmd));
	    break;

	  case RESYNC:
	    printf("RESYNC %s\n", device(cmd));
	    break;

	  case ACK:
	    printf("ACK %s\n", device(cmd));
	    break;

	  case PNM_REQ:
	    printf("PNM Request %s\n", device(cmd));
	    break;

	  default:
	    cp--;
	    printf("Unknown (%02X) %s\n", cmd, device(cmd));
	    break;
	}
    }
    return cp - (unsigned char *) buf;
}

static char *
device(ch)
unsigned char ch;
{
    static char buf[20];

    switch (ch & DEVICE) {
      case KBD_ID:
	return "Keyboard";
	break;

      case VAL_ID:
	return "Validation or Val/Journal Printer";
	break;

      case DOC_ID:
	return "Document Printer";
	break;

      case SRW_ID:
	return "Magcard Reader/Writer";
	break;

      case PIN_ID:
	return "Atalla PIN Pad";
	break;

      case SDO_ID:
	return "System Driver";
	break;

      default:
	sprintf(buf, "Device %d", (ch & DEVICE) >> 3);
	return buf;
	break;
    }
}

static
dumpkbd(cp, len, rawdata)
unsigned char *cp;
int len, rawdata;
{
    int i;

    i = len - 1;
    switch (*cp++) {
      case 0xBF:		/* Extended mode lead-in. */
	i--;
	switch (*cp++) {
	  case 0xB1:		/* Key status. */
	    printf("\t\tKey Status: ");
	    if (cp[0] & 0x1C) {	/* More than 1 keylock on? */
		if (cp[0] & 0x20)
		    printf("SUP1 ");
		if (cp[0] & 0x10)
		    printf("SUP2 ");
		if (cp[0] & 0x08)
		    printf("SUP3 ");
		if (cp[0] & 0x04)
		    printf("SUP4 ");
	    } else {		/* Normal keyboard, only 1 keylock. */
		if (cp[0] & 0x20)
		    printf("SUP ");
	    }
	    if (cp[1] & 0x20)
		printf("CAPS ");
	    if (cp[1] & 0x08)
		printf("ALT ");
	    if (cp[1] & 0x04)
		printf("CTRL ");
	    if (cp[1] & 0x02)
		printf("LSHIFT ");
	    if (cp[1] & 0x01)
		printf("RSHIFT ");
	    printf("\n");
	    break;

	  case 0xBA:		/* Shifted key. */
	    printf("\t\tSHIFT ");
	    goto keycode;
	    break;

	  case 0xBB:		/* CTRL-ed key. */
	    printf("\t\tCTRL ");
	    goto keycode;
	    break;

	  case 0xBC:		/* ALT-ed key. */
	    printf("\t\tALT ");
	    goto keycode;
	    break;

	  default:		/* Regular keystroke. */
	    printf("\t\t");
	    i += 1;		/* Went too far, back up. */
	    cp--;
	  keycode:
	    printf("Keycode%s: ", (i > 1) ? "s" : "");
	    while (i--) {
		printf("$%02X ", *cp);
		if (*cp >= ' ' && *cp <= '~')
		    printf("(%c) ", *cp);
		cp++;
	    }
	    printf("\n");
	}
	break;

      case 0x01:		/* Beep 1/10 second */
	printf("\t\tBeep .1 second\n");
	break;

      case 0x02:		/* Beep 2/10 second */
	printf("\t\tBeep .2 second\n");
	break;

      case 0x03:		/* Beep 3/10 second */
	printf("\t\tBeep .3 second\n");
	break;

      case 0x04:		/* Beep 4/10 second */
	printf("\t\tBeep .4 second\n");
	break;

      case 0x05:		/* Beep 5/10 second */
	printf("\t\tBeep .5 second\n");
	break;

      case 0x06:		/* Beep 6/10 second */
	printf("\t\tBeep .6 second\n");
	break;

      case 0x07:		/* Beep 7/10 second */
	printf("\t\tBeep .7 second\n");
	break;

      case 0x08:		/* Beep 8/10 second */
	printf("\t\tBeep .8 second\n");
	break;

      case 0x09:		/* Beep 9/10 second */
	printf("\t\tBeep .9 second\n");
	break;

      case 0x0A:		/* Beep 1 second */
	printf("\t\tBeep 1 second\n");
	break;

      case 0x10:		/* Default Track */
	printf("\t\tSelect MagCard Default Track\n");
	break;

      case 0x11:		/* Track 1 */
	printf("\t\tSelect MagCard Track 1\n");
	break;

      case 0x12:		/* Track 2 */
	printf("\t\tSelect MagCard Track 2\n");
	break;

      case 0x13:		/* Track 3 */
	printf("\t\tSelect MagCard Track 3\n");
	break;

      case 0x21:		/* Select standard mode. */
	printf("\t\Select Standard Mode\n");
	break;

      case 0x22:		/* Select extended mode. */
	printf("\t\tSelect Extended Mode\n");
	break;

      case 0x53:		/* Key Status Request. */
	printf("\t\tKey Status Request\n");
	break;

      case 0x5F:		/* Device Status Request. */
	printf("\t\tDevice Status Request\n");
	break;

      case 0xB2:		/* PIN message. */
	printf("\t\tPIN message\n");
	if (i > 0) {
	    if (rawdata)
		hd(cp, i, 0);
	    else
		printf("\t\t...\n");
	}
	break;

      case 0xB4:		/* Magcard message. */
	printf("\t\tMagcard message\n");
	if (i > 0) {
	    if (rawdata)
		hd(cp, i, 0);
	    else
		printf("\t\t...\n");
	}
	break;

      case 0xB5:		/* Magcard error. */
	printf("\t\tMagcard error %c\n", *cp);
	break;
    }
    return len;
}

static
dumpval(cp, len, rawdata)	/* Val & Val/Jrnl printer dumper. */
unsigned char *cp;
int len, rawdata;
{
    int len2, i;

    i = 0;
    len2 = len;
    fputs("\t\t", stdout);
    if (len2 == 1) {		/* Printer completion? */
	if (*cp == 0xB4) {
	    fputs("Perform Self Test", stdout);
	} else {
	    if (*cp & 1) {	/* Val & Val/Jrnl */
		if (i++)
		    fputs(", ", stdout);
		fputs("Print Complete", stdout);
	    }
	    if (*cp & 2) {	/* Val/Jrnl */
		if (i++)
		    fputs(", ", stdout);
		fputs("Invalid Message", stdout);
	    }
	    if (*cp & 4) {	/* Val & Val/Jrnl */
		if (i++)
		    fputs(", ", stdout);
		fputs("Paper Fault/Motor Stall", stdout);
	    }
	    if (*cp & 8) {	/* Val/Jrnl */
		if (i++)
		    fputs(", ", stdout);
		fputs("Interlock Open", stdout);
	    }
	    if (*cp & 0x10) {	/* Val & Val/Jrnl */
		if (i++)
		    fputs(", ", stdout);
		fputs("Receive Error/Data Bad", stdout);
	    }
	    if (*cp & 0x40) {	/* Val & Val/Jrnl */
		if (i++)
		    fputs(", ", stdout);
		fputs("Line Length Error/Paper Low", stdout);
	    }
	}
	fputs("\n", stdout);
    } else {			/* Print data. */
	len2--;
	switch (*cp++) {
	  case 0xB1:		/* Val/Jrnl Validation Print */
	    printf("Val/Jrnl Validation Print\n");
	    break;

	  case 0xB2:		/* Val/Jrnl Journal Print */
	    printf("Val/Jrnl Journal Print\n");
	    break;

	  case 0xB3:		/* Val/Jrnl Validation & Journal Print */
	    printf("Val/Jrnl Validation & Journal Print\n");
	    break;

	  case 0xB4:		/* Validation Print */
	    printf("Validation Print\n");
	    break;

	  default:
	    printf("Unknown Print\n");
	    cp--;
	    len2++;
	    break;
	}
	if (len2 > 0) {
	    if (rawdata)
		hd(cp, len2, 0);
	    else
		printf("...\n");
	}
    }
    return len;
}

static
dumpdoc(cp, len, rawdata)
unsigned char *cp;
int len, rawdata;
{
    int i, j;

    fputs("\t\t", stdout);
    switch (*cp) {
				/* Messages TO printer. */
      case PRT_HDR_LR:		/* Print header left/right */
	fputs("Print Header\n", stdout);
	dphdr(&cp[1]);
	break;

      case PRT_HDR_AX:		/* Print header aux */
	fputs("Aux Print Header\n", stdout);
	dphdr(&cp[1]);
	break;

      case PRT_DATA_LR:		/* Print data left/right */
	fputs("Print Data\n", stdout);
	dpdata(&cp[1], len - 1, rawdata);
	break;

      case PRT_DATA_AX:		/* Print data aux */
	fputs("Aux Print Data\n", stdout);
	dpdata(&cp[1], len - 1, rawdata);
	break;

      case AX_REQ_OK:		/* Aux request OK */
	fputs("Aux Request OK\n", stdout);
	break;

      case AX_REQ_NOK:		/* Aux request not OK */
	fputs("Aux Request NOT OK\n", stdout);
	break;

      case SIG_HDR_LR:		/* Signature header left/right */
	fputs("Signature Header\n", stdout);
	break;

      case SIG_HDR_AX:		/* Signature header aux */
	fputs("Aux Signature Header\n", stdout);
	break;

      case SIG_DATA_LR:		/* Signature data left/right */
	fputs("Signature Data\n", stdout);
	break;

      case SIG_DATA_AX:		/* Signature data aux */
	fputs("Aux Signature Data\n", stdout);
	break;

      case IDENTIFY_LR:		/* Identify left/right */
	fputs("Identify\n", stdout);
	break;

      case IDENTIFY_AX:		/* Identify aux */
	fputs("Aux Identify\n", stdout);
	break;

				/* Messages FROM printer. */
      case 0x01:		/* L/R Completion. */
	fputs("Print Completion\n", stdout);
	break;

      case 0x02:		/* Aux Completion. */
	fputs("Aux Print Completion\n", stdout);
	break;

      case 0x04:		/* Signature Not Completed, L/R. */
	fputs("Signature Not Completed\n", stdout);
	break;

      case 0x05:		/* Signature Not Completed, Aux. */
	fputs("Aux Signature Not Completed\n", stdout);
	break;

      case 0x06:		/* 5400/5401 L/R Identity */
	fputs("Type 5400/5401\n", stdout);
	break;

      case 0x07:		/* 5400/5401 Aux Identity */
	fputs("Aux Type 5400/5401\n", stdout);
	break;

      case 0x0A:		/* 5402 L/R Identity */
	fputs("Type 5402\n", stdout);
	break;

      case 0x0B:		/* 5402 Aux Identity */
	fputs("Aux Type 5402\n", stdout);
	break;

      case 0x0C:		/* 5402 Enhanced L/R Identity */
	fputs("Type 5402 Enhanced\n", stdout);
	break;

      case 0x0D:		/* 5402 Enhanced Aux Identity */
	fputs("Aux Type 5402 Enhanced\n", stdout);
	break;

      case 0x0E:		/* Unsolicited L/R Clear. */
	fputs("Unsolicited CLEAR\n", stdout);
	break;

      case 0x0F:		/* Unsolicited Aux Clear. */
	fputs("Unsolicited Aux CLEAR\n", stdout);
	break;

      default:
	i = *cp >> 4;
	j = *cp & 0x0F;
	if (i >= 1 && i <= 9	/* Aux print request? */
	    && j >= 0 && j <= 9) { /* (2-digit, [1-9][0-9]) */
	    if (i == 1 && j == 8)
		printf("Aux RESET\n", *cp);
	    else
		printf("Aux Print Request #%02X\n", *cp);
	} else if (i == 10	/* Aux print request? */
		 && j >= 0 && j <= 9) /* (2-digit, A[0-9]) -> 0[0-9] */
	    printf("Aux Print Request #%02X\n", *cp - 0xA0);
	else			/* Unrecognized. */
	    printf("??? $%02X\n", *cp);
	break;
    }
    return len;
}

static
dphdr(hp)
unsigned char *hp;
{
    int lpi;

    lpi = 60 / (hp[0] & 0x3F);
    printf("\t\t\t%d Pitch, %d Steps/Line (%d LPI), Line 1 offset: %d\n",
	    (hp[0] & 0x40) ? 10 : 12, hp[0] & 0x3F, lpi, hp[1]);
    printf("\t\t\tC-fold: At line %d, Skip: %d lines, ", hp[2], hp[3]);
    printf("Line # After: %d\n", hp[4]);
    printf("\t\t\tLines/page: %d, Print on Line %d\n", hp[5], hp[6]);
}

static
dpdata(dp, len, rawdata)
unsigned char *dp;
int len, rawdata;
{
    switch (*dp) {
      case 0:
	printf("\t\t\tAux Teller Reset\n");
	break;

      case DN_LINES:		/* Vertical tab down x lines */
	printf("\t\t\tVertical TAB Down %d lines\n", dp[1]);
	break;

      case UP_LINES:		/* Vertical tab up x lines */
	printf("\t\t\tVertical TAB Up %d lines\n", dp[1]);
	break;

      case DN_STEPS:		/* Vertical tab down x steps */
	printf("\t\t\tVertical TAB Down %d steps\n", dp[1]);
	break;

      case UP_STEPS:		/* Vertical tab up x steps */
	printf("\t\t\tVertical TAB Up %d steps\n", dp[1]);
	break;
    }
    len -= 2;
    if (len > 0) {
	if (rawdata)
	    hd(&dp[2], len, 0);
	else
	    printf("\t\t\t... (%d byte%s)\n", len, len == 1 ? "" : "s");
    }
}

static
dumpmag(cp, len, rawdata)	/* Magstripe Reader/Writer. */
unsigned char *cp;
int len, rawdata;
{
    int i;

    i = len - 1;
    printf("\t\t");
    switch (*cp++) {
      case 0x08:		/* Abort Operation */
	printf("Abort Operation\n");
	break;

      case 0x0B:		/* Completion. */
	if (i == 1)
	    printf("Write Completion\n");
	else {
	    printf("Read Completion\n");
	    if (i > 1) {
		if (rawdata)
		    hd(cp, i, 0);
		else
		    printf("\t\t...\n");
	    }
	}
	break;

      case 0x0C:		/* Error detected. */
	printf("Error Detected: %c\n", *cp);
	break;

      case 0x0E:		/* Invalid Command. */
	printf("Command Rejected\n");
	break;

      case 0x02:		/* Read Track 2 */
      case 0x12:
      case 0x22:
	printf("Read Track 2\n");
	break;

      case 0x04:		/* Read Track 3 */
      case 0x14:
      case 0x24:
	printf("Read Track 3\n");
	break;

      case 0x03:		/* Write Track 2 */
      case 0x13:
      case 0x23:
	printf("Write Track 2\n");
	goto prdata;
	break;

      case 0x05:		/* Write Track 3 */
      case 0x15:
      case 0x25:
	printf("Write Track 3\n");
      prdata:
	if (*cp != 0x0B)
	    cp++, i--;
	if (i > 1) {
	    if (rawdata)
		hd(&cp[1], i - 1, 0);
	    else
		printf("\t\t...\n");
	}
	break;

    }
    return len;
}

static
dumppin(cp, len, rawdata)	/* Atalla PIN pad. */
unsigned char *cp;
int len, rawdata;
{
    if (len) {
	if (rawdata)
	    hd(cp, len, 0);
	else
	    printf("\t\t...\n");
    }
    return len;
}

static
dumpsdo(cp, len, rawdata)	/* System Driver goodies. */
unsigned char *cp;
int len, rawdata;
{
    if (len == 1 && *cp == 0x16)
	printf("\t\tRT Sync\n");
    else {
	if (len) {
	    if (rawdata)
		hd(cp, len, 0);
	    else
		printf("\t\t...\n");
	}
    }
    return len;
}

