/*
** Function to dump out an XNS (Xerox Network Service) protocol frame.
**
** Among other things, this is used for Novell.  Some information on
** Novell may be found in the November '93 Dr. Dobb's Journal, starting
** on page 123.
*/

#include <stdio.h>
#include <dnix/cheapdrv.h>
#include <sys/types.h>
#include "protos.h"

struct xnshdr {
    u_short checksum;
    u_short length;
    u_char ctrl;
    u_char type;
    u_long destnet;
    u_char desthost[6];
    u_short destsock;
    u_long srcnet;
    u_char srchost[6];
    u_short srcsock;
};

struct spxhdr {
    u_char connctrl;
    u_char datatype;
    u_short srcconn;
    u_short destconn;
    u_short seqnum;
    u_short acknum;
    u_short allonum;
};

struct ncphdr {
    u_short type;		/* Packet type. */
    u_char seqnum;		/* Packet sequence number. */
    u_char connidlo;		/* Server connection #, low. */
    u_char taskno;		/* Client task #. */
    u_char connidhi;		/* Server connection #, high. */
    union {
	struct {
	    u_char fcode;	/* NCP Function code. */
	    u_char sfunlenhi;	/* Optional subfunction length, high. */
	    u_char sfunlenlow;	/* Optional subfunction length, low. */
	    u_char sfun;	/* Optional subfunction code. */
	} req;
	struct {
	    u_char compl;	/* Completion code. */
	    u_char connstat;	/* Connection status. */
	} reply;
    } u;
};

char *xnss[] = {
    "Unknown", "RIP", "Echo", "Error", "PEP", "SPP", /* 0-5 */
    "??", "??", "??", "??", "??", "??", "??", "??", "??", "??", /* 6-15 */
    "Experimental", "NCP", "Experimental", "Experimental", /* 16-19 */
    "Experimental", "Experimental", "Experimental", "Experimental", /* 20-23 */
    "Experimental", "Experimental", "Experimental", "Experimental", /* 24-27 */
    "Experimental", "Experimental", "Experimental", "Experimental" /* 28-31 */
};

static char *socktype(int socket);

dumpxns(buf, len, rawdata)
void *buf;
int len;
int rawdata;			/* TRUE to dump data part of frame. */
{
    char from[14], to[14], *cp, *alloca();
    struct xnshdr *xns;
    struct spxhdr *spx;
    struct ncphdr *ncp;
    int i;

    cp = alloca(len);		/* Make local copy, as buf might be */
    memcpy(cp, buf, len);	/* misaligned causing coredump on 68010. */
    buf = cp;
    xns = (struct xnshdr *) buf;
    i = xns->type & 0x1F;
    printf("\tIPX\tCksum: $%04X, Length: %d, Hop %d, Type %d (%s)\n",
	   xns->checksum, xns->length, xns->ctrl & 0x0F, i, xnss[i]);
    makename(xns->srchost, from);
    makename(xns->desthost, to);
    printf("\t\tTo   %-13.13s Net %d Socket %04X%s\n", to, xns->destnet,
	   xns->destsock, socktype(xns->destsock));
    printf("\t\tFrom %-13.13s Net %d Socket %04X%s\n", from, xns->srcnet,
	   xns->srcsock, socktype(xns->srcsock));
    cp = (char *) &xns[1];
    if (i == 5) {		/* SPX? */
	spx = (struct spxhdr *) &xns[1];
	printf("\tSPX\t");
	i = 0;
	if (spx->connctrl & 0x80)
	    i++,printf(" SYS");
	if (spx->connctrl & 0x40)
	    i++,printf(" ACK-REQD");
	if (spx->connctrl & 0x20)
	    i++,printf(" ATTN");
	if (spx->connctrl & 0x20)
	    i++,printf(" EOM");
	if (i)
	    printf(" ");
	printf("Type: %d", spx->datatype);
	if (spx->datatype == 0xFE)
	    printf(" (End of Connection)");
	if (spx->datatype == 0xFF)
	    printf(" (End of Connection w/ACK)");
	printf("\n");
	printf("\t\tTo %d From %d, Seq#: %d, Ack#: %d, Allo#: %d\n",
	       spx->destconn, spx->srcconn, spx->seqnum, spx->acknum,
	       spx->allonum);
	cp = (char *) &spx[1];
    } else if (i == 17) {	/* NCP? */
	ncp = (struct ncphdr *) &xns[1];
	printf("\tNCP\tType $%04X", ncp->type);
	switch (ncp->type) {
	  case 0x1111:		/* Create service connection. */
	    printf(" (Create Service Connection)");
	    break;
	  case 0x2222:		/* Request. */
	    printf(" (Request)");
	    break;
	  case 0x3333:		/* Reply. */
	    printf(" (Reply)");
	    break;
	  case 0x5555:		/* Destroy service connection. */
	    printf(" (Destroy Service Connection)");
	    break;
	  case 0x7777:		/* Burst Packet. */
	    printf(" (Burst Packet)");
	    break;
	  case 0x9999:		/* Request being processed response. */
	    printf(" (Request Being Processed)");
	    break;
	}
	printf(" Seq#: %d, Conn#: %d, Task#: %d\n", ncp->seqnum,
	       ncp->connidlo | (ncp->connidhi << 8), ncp->taskno);
	cp = (char *) &ncp[1];
    }
    len -= cp - (char *) buf;
    if (len > 0 && rawdata) {
	hd((unsigned char *) cp, len, 0);
	cp += len;
    }
    return cp - (char *) buf;
}

static char *
socktype(socket)
{
    switch (socket) {
      case 0x0001: return " (Routing Information Packet)";
      case 0x0002: return " (Echo Protocol Packet)";
      case 0x0003: return " (Error Handler Packet)";
      case 0x0451: return " (Novell File Svc Packet)";
      case 0x0452: return " (Novell Svc Advert Packet)";
      case 0x0453: return " (Novell Route Info Packet)";
      case 0x0455: return " (Novell NetBIOS Packet)";
      case 0x0456: return " (Novell Diags Packet)";
      default: return "";
    }
}
