/*
** Convert a bitmap file to a Picture File.  The picture file may
** then be embedded within a program by the program coffwrap (and the
** linker), for use within dialogs (or wherever).
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <malloc.h>
#include <isc/window.h>
#include <isc/bitobj.h>

char versstr[] = "Version 1.0";
char *progname;			/* Our program name. */
char *paintfile;		/* Input bitmap file. */
char *pictfile;			/* Output PICR file. */

short bits_per_pixel = 1;
short pal_entries, pix_bits;
BitMap source;
Rect dest;
PicHandle thepict;
int mode;
struct BITOBJPAL {
    struct BITOBJ header;
    short pal_entries;
};

main(argc, argv)
int argc;
char *argv[];
{
    int i;
    char *cp;

    progname = argv[0];

    /*
    ** Process command line switches.
    */

    if (argc < 3) {
	version();
	return 1;
    }
    paintfile = argv[1];
    pictfile = argv[2];
    for (i=1; i<argc; i++) {	/* Search out arguments. */
	cp = argv[i];
	if (*cp++ == '-') {	/* Command-line switch? */
	    switch (*cp++) {	/* Yes, which one? */
	      case 'm':
		if (!*cp)
		    i++, cp = argv[i];
		mode = atoi(cp);
		break;

	      default:		/* Version printer. */
		version();
		return 0;
		break;
	    }
	}
    }

    /*
    **	Initialize the graphics toolbox.
    */

    if (InitGraf() < 0) {
	return 1;
    }

    ConnectGraf("");		/* Get a NULL graphics connection. */
    if (!loadfile(paintfile))
	return 2;
    thepict = OpenPicture(&source.bounds);
    CopyBits(&source, 0, &source.bounds, &source.bounds, mode, 0);
    ClosePicture();
    if ((i = creat(pictfile, 0644)) < 0)
	return 3;
    write(i, *thepict, (*thepict)->picSize);
    close(i);
    return 0;
}

/*
** Convert a width to RowBytes
*/
short Width2RowBytes(width)
short width;
{
    return ( ((short) (width - 1) / 16 + 1) << 1 );
}

/*
** Adjust RowBytes for a pixel map (color)
*/
short 
CalcRowBytes(rowbytes)
short rowbytes;
{
 /*
 ** If odd rowBytes, then shift out pixel info.
 */
    if (rowbytes & 1)
	rowbytes = (rowbytes >> 3) / (1 << ((rowbytes & 7) >> 1));
    return rowbytes;
}

int
loadfile(filename)
char *filename;
{
    struct BITOBJ *BitObj;
    struct PALETTE pal;
    struct stat lstatbuf;
    long position;
    int fd, Read_Size = BUFSIZ, height;
    char header[BUFSIZ], *base, *scratch, *buffer;

    if (filename && !source.baseAddr) {
	/*
	** Try to stat the picture.
	*/
	if (stat(filename, &lstatbuf))
	    return 0;
	if (lstatbuf.st_size < Read_Size)
	    Read_Size = lstatbuf.st_size;

	/*
	** Try to open the picture.
	*/
	fd = open(filename, 0);
	if (fd < 0)
	    return 0;

	/*
	** Read in the Picture Header.
	*/
	if (read(fd, header, Read_Size) != Read_Size) {
	  closenquit:
	    close(fd);
	    return 0;
	}

	/*
	** Now figure out how to decode it.
	*/
	BitObj = (struct BITOBJ *) header;
	if (Is_Isc(BitObj)) {
	    /*
	    ** Create the source BitMap
	    */
	    source.rowBytes = Width2RowBytes(BitObj->width) * bits_per_pixel;
	    SetRect(&source.bounds, 0, 0, BitObj->width, BitObj->height);
	    height = source.bounds.edge.bottom - source.bounds.edge.top;
	    dest = source.bounds;

	    /*
	    ** Get us a buffer for the decoded BitMap.
	    */
	    source.baseAddr = malloc(source.rowBytes * height);
	    if (!source.baseAddr) {
		goto closenquit;
	    }
	    if (pal_entries)
		source.rowBytes = (source.rowBytes << 3) | pix_bits;

	    /*
	    ** Now read in and decode the image.
	    */
	    position = sizeof(bitobj);
	    if (pal_entries) {
		position += sizeof(short);
		lseek(fd, position, 0);
		while (pal_entries--) {
		    if (read(fd, &pal, sizeof(pal)) != sizeof(pal)) {
		      freenquit:
			free(source.baseAddr);
			source.baseAddr = 0;
			goto closenquit;
		    }
		    SetRGBColor(pal.ind, pal.red, pal.green, pal.blue);
		    position += sizeof(pal);
		}
	    }
	    lstatbuf.st_size -= position;
	    lseek(fd, position, 0);
	    switch (BitObj->compress) {
	      case BO_COMP_NONE:
		if (read(fd, source.baseAddr, lstatbuf.st_size)
		    != lstatbuf.st_size) {
		    goto freenquit;
		}
		break;

	      case BO_COMP_PACKBITS:
		buffer = malloc(lstatbuf.st_size);
		if (!buffer) {
		    goto freenquit;
		}
		if (read(fd, buffer, lstatbuf.st_size) != lstatbuf.st_size) {
		    free(buffer);
		    goto freenquit;
		}
		base = buffer;
		scratch = source.baseAddr;
		while (height--) /* Unpack the picture. */
		    UnpackBits(&base, &scratch, BitObj->rowBytes);
		free(buffer);
		break;
	    }
	} else {		/* Is a MacPaint file. */
	    /*
	    ** Create the source BitMap.
	    */
	    source.rowBytes = 72;
	    SetRect(&source.bounds, 0, 0, 576, 720);
	    height = source.bounds.edge.bottom - source.bounds.edge.top;
	    dest = source.bounds;

	    /*
	    ** Get us a buffer for the decoded BitMap.
	    */
	    source.baseAddr = malloc(source.rowBytes * height);
	    if (!source.baseAddr) {
		goto closenquit;
	    }

	    /*
	    ** Now read and decode the image.
	    */
	    lstatbuf.st_size -= 512; /* File size. */
	    lseek(fd, 512L, 0);	/* Skip the file's header. */
	    buffer = malloc(lstatbuf.st_size);
	    if (!buffer) {
		goto freenquit;
	    }
	    if (read(fd, buffer, lstatbuf.st_size) != lstatbuf.st_size) {
		free(buffer);
		goto freenquit;
	    }
	    base = buffer;
	    scratch = source.baseAddr;
	    while (height--)	/* Unpack the picture. */
		UnpackBits(&base, &scratch, 72);
	    free(buffer);
	}
	close(fd);
    }
    return 1;
}

Is_Isc(header)
struct BITOBJ *header;
{
    /*
    ** Must be a known compression type.
    */
    if (header->compress != BO_COMP_NONE
	&& header->compress != BO_COMP_PACKBITS)
	return 0;

    /*
    ** Currently everything is version zero.
    */
    if (header->version)
	return 0;

    /*
    ** The rowBytes must work with the width.
    */
    if (Width2RowBytes(header->width) != CalcRowBytes(header->rowBytes))
	return 0;

    /*
    ** Figure out the new bits per pixel (color).
    */
    if (header->rowBytes & 1) {
	bits_per_pixel = (1 << ((header->rowBytes & 7) >> 1));
	pix_bits = (header->rowBytes & 7); /* Save global pix_bits. */
    }
    /*
    ** Get the number of palette entries (if any).
    */
    if (header->flags & BO_FL_PALETTE) {
	struct BITOBJPAL *IscHeadPal = (struct BITOBJPAL *) header;

	pal_entries = IscHeadPal->pal_entries;
    }
    /*
    ** OK, assume its a valid BitObj.
    */
    return 1;
}

#define PRINT(x) p = (x); write(1, p, StringLen(p));

version()
{
    char *p;

    PRINT(versstr);
}

usage()
{
    char *p;

    version();
    PRINT("Usage:\n");
    PRINT("Usage: makepictfile paintfile pictfile\n");
}
