/*
** NetBIOS structures from RFC1002, for NB over IP.
*/

/* Name Service packet header */

struct nmheader {
    unsigned short trn_id;
    unsigned short opcode_flags_rcode;
    unsigned short qdcount;
    unsigned short ancount;
    unsigned short nscount;
    unsigned short arcount;
};

#define OPCODE			0xF800
# define RBIT			 0x8000
# define SUBOP			 0x7800
#  define OP_QUERY		  (0<<11)
#  define OP_REGIS		  (5<<11)
#  define OP_RELEASE		  (6<<11)
#  define OP_WACK		  (7<<11)
#  define OP_REFRESH		  (8<<11)
#define FLAGS			0x07F0
# define FL_BROAD		 0x0010	/* Broadcast */
# define FL_RA			 0x0080 /* Recursion Available */
# define FL_RD			 0x0100 /* Recursion Desired */
# define FL_TC			 0x0200 /* Truncation */
# define FL_AA			 0x0400 /* Authoritative Answer */
#define RCODE			0x000F
# define FMT_ERR		 0x1	/* Format Error */
# define SRV_ERR		 0x2	/* Server Failure */
# define NAM_ERR		 0x3	/* Name Error */
# define IMP_ERR		 0x4	/* Unsupported Request */
# define RFS_ERR		 0x5	/* Refused Error */
# define ACT_ERR		 0x6	/* Active Error */
# define CFT_ERR		 0x7	/* Conflict Error */

/***********************************************************************/

/* Datagram packet header */

struct ndheader {
    unsigned char msg_type;
    unsigned char flags;
    unsigned short dgm_id;
    unsigned long source_ip;
    unsigned short source_port;
    unsigned short dgm_length;
    unsigned short packet_offset; /* Used when fragmenting. */
};

struct nderror {
    unsigned char msg_type;
    unsigned char flags;
    unsigned short dgm_id;
    unsigned long source_ip;
    unsigned short source_port;
    unsigned char error_code;
};

/* msg_type */

#define DIRECT_UNIQUE_DGRAM	0x10
#define DIRECT_GROUP_DGRAM	0x11
#define BROADCAST_DGRAM		0x12
#define DATAGRAM_ERROR		0x13
#define DATAGRAM_QUERY_REQUEST	0x14
#define DATAGRAM_ACK_QUERY_RSP	0x15
#define DATAGRAM_NAK_QUERY_RSP	0x16

/* flags */
#define MORE			0x01	/* More fragments follow (len & off) */
#define FIRST			0x02	/* First (possibly only) fragment */
#define SNT			0x0C	/* Source End-Node Type */
# define BNODE			 0x00	/* Broadcast Node */
# define PNODE			 0x04	/* Point-to-point Node */
# define MNODE			 0x08	/* Mixed (B & P) Node */
# define NBDD			 0x0C	/* NB Datagram Distribution Node */

/* error_code */

#define DNAME_MISSING		0x82
#define INVALID_SNAME		0x83
#define INVALID_DNAME		0x84

/***********************************************************************/

/* Session packet header */

struct nsheader {
    unsigned char type;
    unsigned char flags;
    unsigned short length;	/* Total length of trailer fields. */
/* Trailer fields here */
};

/* type */

#define MESSAGE			0x00	/* Session Message */
#define REQUEST			0x81	/* Session Request */
#define POSITIVE		0x82	/* Positive Session Response */
#define NEGATIVE		0x83	/* Negative Session Response */
#define RETARGET		0x84	/* Retarget Session Response */
#define KEEPALIVE		0x85	/* Session Keep Alive */

/* flags */
#define EXTENSION		0x01	/* Length Extension */

/***********************************************************************/
/***********************************************************************/
/***********************************************************************/
