/*
** Network General packet to Sniffer converter.
**
** Usage:
**	ng2dsniff [-V] [file]
**
** Where
**	-V	prints version number and exits,
**	file	Packets in this file, otherwise read from standard in.
*/

#include <stdio.h>
#include <errno.h>
#include <dnix/cheapdrv.h>
#include <dnix/fcodes.h>
#include "ngsniff.h"

struct packetform {
    struct sniffheader snh;
    struct etherheader eth;
    unsigned char data[8192];	/* Way bigger than typically needed. */
};

void rdie(), fdie(), die();
unsigned long dotime();
char *progname;

main(argc, argv)
int argc;
char *argv[];
{
    FILE *fp, *outfp;
    unsigned short packetsize, psize;
    struct packetform packet;
    char *filename, *argp, buf[17];
    int packnum;
    Header_Record nghdr;
    Version_Record ngver;
    Frame_Record ngfrm;
    double timestamp, ipart;

    progname = *argv;
    filename = 0;
    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'V':		/* Version printout. */
		printf("%s: Version 1.0\n", progname);
		exit(0);
		break;

	      default:
		fprintf(stderr, "Usage: %s [-V] [packetfile]\n", progname);
		exit(1);
		break;
	    }
	} else {
	    filename = argp;
	}
    }
    if (filename)
	fp = fopen(filename, "r");
    else
	fp = fdopen(0, "r");
    if (!fp) {
	fprintf(stderr, "%s: Unable to read input file %s\n", progname,
		filename ? filename : "");
	exit(2);
    }
				/* Eat NG header crap */
    if (fread(buf, sizeof buf, 1, fp) != 1)
	rdie();
    if (buf[16] != '\032')
	fdie();
    if (fread(&nghdr, sizeof nghdr, 1, fp) != 1)
	rdie();
    if (nghdr.type != LIBI(REC_VERS) || nghdr.length != LIBI(sizeof(ngver)))
	fdie();
    if (fread(&ngver, sizeof(ngver), 1, fp) != 1)
	rdie();
    if (ngver.type != REC_FRAME || ngver.network != N_ETHERNET
	|| ngver.format != 1 || ngver.rsvd[0] != 0)
	fdie();
    packnum = -1;
    packet.snh.tickrate = 50;
    packet.snh.statusbits = PACKET_GOOD;
    packet.snh.fractickdenom = 0x7FFF; /* Maximum resolution. */
    while (fread(&nghdr, sizeof nghdr, 1, fp) == 1) {
	if (nghdr.type == LIBI(REC_FRAME)) {
	    if (fread(&ngfrm, sizeof(ngfrm), 1, fp) != 1)
		rdie();
	    timestamp = (double) ngfrm.time_high;
	    timestamp = timestamp * 65536.0 + (double) LIBI(ngfrm.time_mid);
	    timestamp = timestamp * 65536.0 + (double) LIBI(ngfrm.time_low);
	    timestamp = timestamp / netunits[ngver.timeunit]; /* In usec. */
	    timestamp /= 20000.0; /* Number of 50-Hz ticks, for timestamp */
	    modf(timestamp, &ipart);
	    packet.snh.tickstamp = (unsigned long) ipart;
	    timestamp -= ipart;	/* Now turn the fraction into our best-res */
	    modf(timestamp * 32767, &ipart); /* fractional representation. */
	    packet.snh.fracticknum = (short) ipart;

	    psize = LIBI(ngfrm.size);
	    if (fread(&packet.eth, psize, 1, fp) != 1)
		rdie();
	    packnum++;
	    packetsize = packet.snh.length = psize + sizeof(struct sniffheader);
	    psize = packetsize;
	    if (packetsize & 1)	/* Need pad? */
		packetsize++;
	    if (fwrite(&packetsize, sizeof(packetsize), 1, stdout) != 1)
		die();
	    if (fwrite(&packet, psize, 1, stdout) != 1)
		die();
	    if (psize & 1)	/* Add pad? */
		fputc(0, stdout);
	} else if (nghdr.type == LIBI(REC_EOF)) {
	    exit(0);
	} else {		/* Skip unknown record type. */
	    fseek(fp, LIBI(nghdr.length), 1);
	}
    }
    fdie();
}

void
rdie()
{
    fprintf(stderr, "%s: Read Error %d on input\n", progname, errno);
    exit(2);
}

void
fdie()
{
    fprintf(stderr, "%s: Format Error on input\n", progname);
    exit(2);
}

void
die()
{
    fprintf(stderr, "%s: Write Error %d on output\n", progname, errno);
    exit(3);
}
