#! /bin/sh
#
#  NAME
#	sniffer.ins7 - sniffer installation script
#
#	Copyright (c) 1991 ISC-BUNKER RAMO
#	  All Rights Reserved
#
#	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ISC-BR
#	The copyright notice above does not evidence any
#	actual or intended publication of such source code.
#	This Module contains Proprietary Information of
#	ISC-BR and should be treated as confidential.
#
#  DESCRIPTION
#       D-NIX Ethernet Sniffer installation script.
#
#  USER MODIFIABLE VARIABLES
#	None
#
#  EXIT CODES
#	 0: successful execution
#	 1: fatal error
#	 2: warning
#	 3: interruption
#

# ======================================================= #
#  User Modifiable Variables                              #
# ======================================================= #
NETPASS=${NETPASS:-iscisc}	# default netkey password
NODENAME=${NODENAME:-unknown}		# default system id

# ======================================================= #
#  WARNING: Users do not alter anything below this line.  #
# ======================================================= #

#
#  External variables imported from GENASYS
#
PKG=${PKG:-sniffer}		# package abbreivation
export PKG
ROOT=${ROOT:-`pwd`}
export ROOT
DISTDIR=${DISTDIR:-$ROOT/usr/adm/dist}
export DISTDIR
HISTORY=${HISTORY:-$DISTDIR/history}
export HISTORY
PKGDIR=${PKGDIR:-$DISTDIR/$PKG}	# package info directory
export PKGDIR
CLEAR=${CLEAR:-'echo "\n"'}
export CLEAR
NEWBOX=${NEWBOX:-FALSE}
export NEWBOX
BATCH=${BATCH:-FALSE}
export BATCH
VERBOSE=${VERBOSE:-TRUE}
export VERBOSE
PATH=${PATH:-":/bin:/usr/bin:/etc"}
export PATH
HARDNAME=${HARDNAME:-DNP-010}
export HARDNAME
SYSTYPE=${SYSTYPE:-DNP}
export SYSTYPE
CPUTYPE=${CPUTYPE:-M68010}
export CPUTYPE
BLDTYPE=${BLDTYPE:-ISC}
export BLDTYPE
RUNLEVEL=${RUNLEVEL:-0}
export RUNLEVEL
OSREL=${OSREL:-0}
export OSREL
OSVER=${OSVER:-0}
export OSVER
GENVER=${GENVER:-0}
export GENVER

#
#  Fixup target directory path (if needed)
#
if [ "$ROOT" = "/" ]
then
	ROOT=""
fi

#
#  Package variables and constants
#
PKGPROG=`basename $0`		# package install script
PKGVER='$Revision: 1.0 $'	# package RCS version
PKGNAME="Sniffer"		# package name
PKGTEMP=/tmp/$PKG.temp		# package tmp work file
PKGERRS=FALSE			# Fatal errors
PKGWARN=FALSE			# Warnings
PKGINTR=FALSE			# Interruptions
PKGBOOT=0			# 0  = Reboot not required

SYSIDFILE=/etc/systemid

#
#  Required files list
#
RQDFILES="\
	/bin/cat \
	/bin/copy \
	/bin/expr \
	/bin/sh \
	/bin/ls \
	/bin/chmod \
	/etc/mknod"

#
#  Display package version level
#
if [ $VERBOSE = TRUE ]
then	echo "D-NIX $PKGNAME Installation Script  $PKGVER\n"
fi

#
#  Check required files list - all here?
#
NOFILE=FALSE
for file in $RQDFILES
do
	if [ ! -x $file ]
	then
		echo "sniffer: Required file $file not found"
		NOFILE=TRUE
	fi
done
if [ $NOFILE = TRUE ]
then
	exit 1
fi

#
#  Initialize trap for various interrupt conditions
#
trap 'PKGINTR=TRUE; exit 3' 1 2 3 15

#
#  Exit with appropriate status set
#
trap '  if   [ $PKGERRS = TRUE ]	# Fatal error
	then
		PKGBOOT=1
	elif [ $PKGINTR = TRUE ]	# Interruption
	then
		PKGBOOT=3
	elif [ $PKGWARN = TRUE ]	# Warning
	then
		PKGBOOT=`/bin/expr $PKGBOOT + 2`
	fi
	exit $PKGBOOT
	
	' 0

# ================================================= #
#  Sniffer specific stuff starts here               #
# ================================================= #

#
#  Add the required devices as needed
#
if [ ! -c $ROOT/dev/sniffer -a -c $ROOT/dev/dnet ]
then
	set `/bin/ls -l $ROOT/dev/dnet`
	major=$5
	/etc/mknod $ROOT/dev/sniffer c ${major%,} 9
	/bin/chmod 600 $ROOT/dev/sniffer
fi

if [ ! -c $ROOT/dev/hdlcsniff -a -c $ROOT/dev/hdlc ]
then
	set `/bin/ls -l $ROOT/dev/hdlc`
	major=$5
	let minor="$6 + 128"
	/etc/mknod $ROOT/dev/hdlcsniff c ${major%,} $minor
	/bin/chmod 600 $ROOT/dev/hdlcsniff
fi

if [ ! -c $ROOT/dev/winsniff -a -c $ROOT/dev/windrv ]
then
	set `/bin/ls -l $ROOT/dev/windrv`
	major=$5
	let minor="$6 + 128"
	/etc/mknod $ROOT/dev/winsniff c ${major%,} $minor
	/bin/chmod 600 $ROOT/dev/winsniff
fi

#
#  Make an addresses file if needed
#
if [ ! -f usr/lib/sniffer/addresses ]
then
	/bin/copy -m usr/lib/sniffer/addresses+ usr/lib/sniffer/addresses
	/bin/chmod 600 usr/lib/sniffer/addresses
fi

# ================================================= #
#  Sniffer specific stuff ends here                 #
# ================================================= #

#
#  Update history for old genasys
#
if [ $GENVER = 0 -a -w $HISTORY -a -f $PKGDIR/version ]
then
	/bin/cat $PKGDIR/version >> $HISTORY
fi
exit 0
