/*
** Protocol types for some common interesting Ethernet protocols.  These
** are either in the original Ethertype field (bytes 12 & 13 of the frame),
** or are in the SNAP header of an 802.3 frame.
*/

#define SAFE 0xFF00		/* ISC Safelink (fileserver). */
#define BOOT 0xFF01		/* ISC Boot/Dump protocol. */
#define DNET 0xFF02		/* ISC D-net (netman) protocol. */
#define DMAP 0xFF03		/* ISC dmap protocol. */
#define ATLK 0x809B		/* Appletalk . */
#define AARP 0x80F3		/* Appletalk Address Resolution Protocol. */
#define LAT  0x6004		/* DEC LAT protocol. */
#define IP   0x0800		/* Internet IP protocol. */
#define ARP  0x0806		/* Internet ARP protocol. */
#define RARP 0x8035		/* Reverse ARP protocol. */
#define DECB 0x8038		/* DEC Bridge Management. */
#define NOVELL 0x8137		/* Novell XNS/IPX */

/*
** SAP Protocol types.  These are only 1 byte in size, not the two bytes
** of the traditional Ethertype.  (For now, we don't have any detailed SAP 
** disassemblers, so this fake code only enables SAP header display.)
*/

#define SAP 0x02

#define NULL_SAP    0x00	/* Null LSAP */
#define INDIVID_SAP 0x02	/* Individual LLC Sublayer Management Func. */
#define GROUP_SAP   0x03	/* Group LLC Sublayer Management Func. */
#define INDIVSNA_SAP 0x04	/* Individual IBM SNA Path Control. */
#define GROUPSNA_SAP 0x05	/* Group IBM SNA Path Control. */
#define IP_SAP      0x06	/* ARPANET Internet Protocol (IP). */
#define SNA1_SAP    0x08	/* SNA. */
#define SNA2_SAP    0x0C	/* SNA. */
#define PROWAYMI_SAP 0x0E	/* PROWAY (IEC955) Network Mgmt & Init. */
#define TI_SAP      0x18	/* Texas Instruments. */
#define SPAN_SAP    0x42	/* IEEE 802.1 Bridge Spanning Tree Protocol. */
#define RS511_SAP   0x4E	/* EIA RS-511 Manufacturing Msg Service. */
#define ISO8208_SAP 0x7E	/* ISO 8208 (X.25 over IEEE 802.2 Typ2 LLC). */
#define XNS_SAP     0x80	/* Xerox Network Systems (XNS). */
#define NESTAR_SAP  0x86	/* Nestar. */
#define PROWAYAS_SAP 0x8E	/* PROWAY (IEC955) Active Station List Maint */
#define ARP_SAP     0x98	/* ARPANET Addr Resolution Protocol (ARP). */
#define VINES_SAP   0xBC	/* Banyan VINES. */
#define SNAP_SAP    0xAA	/* Subnetwork Access Protocol (SNAP). */
#define NETBIOS_SAP 0xF0	/* NETBios SSAP/DSAP. */
#define NETWARE_SAP 0xE0	/* Novell NetWare SSAP/DSAP. */
#define INDIVLAN_SAP 0xF4	/* Individual IBM LAN Management. */
#define GROUPLAN_SAP 0xF5	/* Group IBM LAN Management. */
#define RPL_SAP     0xF8	/* IBM Remote Program Load (RPL). */
#define UB_SAP      0xFA	/* Ungermann-Bass. */
#define ISONLP_SAP  0xFE	/* ISO Network Layer Protocol. */
#define GLOBAL_SAP  0xFF	/* Global LSAP. */

/*
** Dummy protocol types for sniffing non-Ethernet connections.  These
** 'protocols' aren't actually ever seen on the Ethernet, but are used
** so that the same packet disassembler may be used for all.
*/

#define WINCOM 0x0001		/* Wincom protocol. */
