/*
** GUI front-end for the Ethernet Sniffer.
**
** Usage:  sniffer [-a] [-b] [-c] [-D Dstaddrfilter] [-fd filterdata]
**		   [-fi] [-fo filteroffset] [-fx] [-hd] [-hg histogram]
**		   [-hl hexlimit] [-n bufsize] [-nose] [-ov overrides]
**		   [-p] [-P Protocolfilter] [-S Srcaddrfilter] [-s]
**		   [-td trigdata] [-ti] [-to trigoffset] [-tp trigpos]
**		   [-tr] [-ts] [-V] [-va] [-vd] [-vr] [-vs] [-vt] [-wan] [-win]
**		   [settingsfile]
**
** where
**	-b, -c, -D, -n, -p, -P, and -S flags pre-fill fields that control
**	the same lettered options to the sniffnet program.  The -nose
**	option shuts off the display of the large bitmapped artwork, for
**	better performance over slow WAN connections.  -V prints the
**	version number and exits.  -a enables animation, -s enables sound.
**
** Version 1.0		12/15/92	Jim Cathey
**	Original version.
*/

#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <malloc.h>
#include <fcntl.h>
#include <unistd.h>
#include <dnix/fcodes.h>
#include <dnix/cheapdrv.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <isc/window.h>
#include <isc/keycodes.h>
#include <isc/controls.h>
#include <isc/textedit.h>
#include <isc/dlgman.h>
#include <isc/vterm.h>
#include <isc/bitobj.h>
#include <dnix/dutsname.h>
#include "sharedmem.h"
#include "Dmain.h"
#include "Dtrig.h"
#include "Dpick.h"
#include "Dlimit.h"
#include "Dlevel1.h"
#include "Dover.h"
#include "Dfilt.h"
#include "Dhelp.h"
#include "Dhist.h"

/*
** Layer control bits in the usercode of Dialog items:
**
** Bit 0 is on for Capture elements,
** Bit 1 is on for Show elements.
**
** Bits 2-4 select items shown on a per-level-1-interface basis:
** Bit 2 is on for LAN (capture) elements,
** Bit 3 is on for WAN (capture) elements,
** Bit 4 is on for WINCOM (capture) elements.
*/
#define DCAPS 1
#define DSHOWS 2
#define DLEVEL1MASK 0x1C
#define DLAN 4
#define DWAN 8
#define DWINCOM 16

int trapqfd, wmfd, windowid;
char *nodename, *progname;
Rect winrect;
int shmkey, shmid;
struct sniffstat *shmp;
DMDcb *dialog, *dialog2;
extern Picture nose, snort;
int sniffchild;			/* Child process of sniffer capture program. */
int sniffing;			/*   " is sniffing (vs writing file). */
int statchild;			/* Child process of sniffer stats program. */
int openchild;			/* Child process of Open File dialog. */
int savechild;			/* Child process of Save File dialog. */
int lsetchild;			/* Child process of Load Settings dialog. */
int ssetchild;			/* Child process of Save Settings dialog. */
int copychild;			/* Child process of file copier. */
int viewchild;			/* Child process of file viewer. */
int helpchild;			/* Child process of help viewer. */
int histchild;			/* Child process of histogram viewer. */
int editchild;			/* Child process of address editor. */
char *sniffer = "./sniffnet";
char *sniffhdlc = "./sniffhdlc";
char *sniffwin = "./sniffwin";
char *snifferdump = "./snifferdump";
char *statprog = "./packstats";
char *copier = "/bin/copy";
char *viewer = "/usr/local/mo";
char *picker = "/usr/lib/graphics/filepicker";
char *nroff = "/usr/bin/nroff";
char *cat = "/bin/cat";		/* In case nroff's not there, revert to cat. */
char *nulldev = "/dev/null";
char *addresses = "./addresses";
char *hm1 = "-H";
char *hm2 = "-R";
char *helps[] = {"sniffer.8", "blower.8", "sniffdump.8", "sniffhdlc.8", 
		     "sniffwin.8", "packstats.8", "sniffnet.8"};

SFDialog sfo, sfs, sflss;
int hasfile;			/* TRUE if a trace file is 'loaded', */
				/* -1 if it's not to be removed. */
char filename[80];		/* The trace filename. */
char fltrhexprot[24];		/* Storage for hex filter/display protocol */
char disphexprot[24];		/* values, since we've only 1 TE for both. */
char fltrsrcaddr[256];		/* Ditto for source and destination */
char dispsrcaddr[256];		/* filter addresses. */
char fltrdstaddr[256];
char dispdstaddr[256];
char trigoffset[5] = {"0"};
char trigdata[256];
char filtoffset[5] = {"0"};
char filtdata[256];
char trigpos[5] = {"0"};
char trigat[6];
char overrides[256];
char *picklabel;
DMPick pickarray;
int fltrsrc;			/* Ditto for the source/dest/conversational */
int dispsrc;			/* filter enabling. */
int fltrdst;
int dispdst;
int fltrconv;
int dispconv;
int trigger, triggerinv;
int datafilt, filterexc;
int nostop;
int normal = 1;
int tpos = DTRIG_trig0;
int iconified;
int nonose;			/* TRUE to suppress bitmapped artwork. */
int animated;			/* TRUE to show capture bargraphs. */
int sound;			/* TRUE to add sound. */
int bflag, cflag, pflag;
char *Dflag, *nflag, *Pflag, *Sflag;
short mywindowisinfront, doimportscrap = 1, doexportscrap;
char scrapbuf[256];		/* Cut/Paste buffer. */
int scrapsize;			/* Amount in buffer. */
int datalimit;
char limitsize[5];
int level1;			/* Which hardware interface to tap, */
#define L1LAN 0			/* 0 == LAN, 1 == WAN, 2 == wincom. */
#define L1WAN 1
#define L1WINCOM 2

int lasthelp = DHELP_sniffer;	/* Last help screen shown. */
#define HIST_SIZE 1
#define HIST_SADR 2
#define HIST_DADR 4
#define HIST_PROT 8
int histogram = HIST_SIZE|HIST_SADR|HIST_DADR|HIST_PROT;
int allheaders;
int datapacks;
int rawpacks;
int symbolics;
int deltatime;
struct dutsname duts;           /* For clock tick rate, et al. */
int lastticks, lastseen;	/* For animated bandwidth calculation. */

struct overpacket {		/* For packet-override dummy packets. */
    unsigned short packsize;
    struct sniffheader snh;
    struct etherheader eth;
};

struct nowaitblock {		/* No-wait control structure. */
    void (*completionroutine)();
    int requestnum;
    int param;			/* Length of nap or process exit status. */
    int arg;			/* Misc argument. */
};

void waitdone();		/* Forward declare completion routines. */
void readdone();
void napdone();
void waitforchild(void);
void postnowaitwait(void);
void killnowaitwait(void);
void writescrap();
void toggleicon();
int children;			/* Count of child processes. */
static struct nowaitblock nowaitwait = {waitdone, 0, 0};
static struct nowaitblock nowaitread = {readdone, 0, 0};
static struct nowaitblock nowaitnap = {napdone, 0, 500};

void updateevent();		/* Forward declare event handlers. */
void menuevent();
void activatevent();
void windowevent();
void menuevent();
void scrapevent();
void MyAction(ControlID theControl, short partCode);

/* Menus and menu items. */
#define FILEMENU 1
#define FILEOPEN 1
#define FILESAVE 3
#define FILELSET 5
#define FILESSET 6
#define FILEQUIT 8

#define EDITMENU 2
#define EDITCUT 1
#define EDITCOPY 2
#define EDITPASTE 3
#define EDITADDR 5

#define MISCMENU 3
#define MISCLIMIT 1
#define MISCHDWR 2
#define MISCOVERRIDE 3
#define MISCHISTOGRAM 4
#define ANIMATION 5
#define SOUND 6
#define HELP 7

/*
** Error messages that match the exit statuses of the sniffnet program.
*/
char landev[] = {"Unable to open /dev/sniffer!"};
char wandev[] = {"Unable to open /dev/hdlcsniff!"};
char wincomdev[] = {"Unable to open /dev/winsniff!"};

char *snifferrs[] = {
    "",
    "Your sniffer has expired!  Get a new one.",
    "Unable to allocate enough memory!  (Try a smaller capture buffer.)",
    "Unable to lock down I/O buffer memory!  (Try a smaller capture buffer.)",
    landev,			/*** CHANGED PROGRAMMATICALLY! ***/
    "Either you aren\325t superuser or this OS doesn\325t support sniffing!",
    "Unable to open a trap queue!",
    "Error detected when reading packets!"
};

/*
** Miscellaneous error messages.
*/
char *nomsg = "";
char *savefailed = "File save failed!";
char *openfailed = "File open failed!";
char *badcapture = "Bad capture!";
char *sniffdied = "Sniffnet died from signal ";

main(argc, argv)
int argc;
char *argv[];
{
    int len;
    char scratch[80], *cp, *getenv();

    dnix(F_UNAME, &duts, sizeof(struct dutsname));
    if ((trapqfd = InitGraf()) < 0) {
	exit(1);
    }

    scratch[0] = scratch[sizeof scratch] = 0;
    cp = getenv("WINDOWS_ON");
    if (cp && strchr(cp, '.')) {
	if ((wmfd = open("/etc/domainid", O_RDONLY|O_NOCTTY)) >= 0) {
	    if (read(wmfd, scratch, sizeof(scratch) - 1) > 0) {
		cp = strchr(scratch, '\n');
		if (cp)		/* Knock off newline. */
		    *cp = 0;
	    }
	    close(wmfd);
	}
    }
    nodename = malloc(StringLen(duts.nodename) + StringLen(scratch) + 1);
    strcpy(nodename, duts.nodename);
    strcat(nodename, scratch);

    if ((wmfd = ConnectGraf(0)) < 0) {
	char *p;
	p = "Unable to connect to a Window Manager\n";
	write(2, p, StringLen(p));
	exit(2);
    }

    setpgrp();
    processargs(argc, argv);
    winrect = Dmain[DMAIN_MainFrame].rect;
    OffsetRect(&winrect, 30, 50);
    strcpy(scratch, "Network Sniffer : ");
    strncat(scratch, nodename, sizeof scratch);
    windowid = NewWindow(&winrect, scratch,
			 showWindow, documentProc, frontWindow, 
			 hasDrag|hasGoAway|hasZoom|hasSendToBack, wmfd);
    SetPort(windowid);
    BeginUpdate(windowid);
    EndUpdate(windowid);
    InitControls();
    TEInit();
    DMInit();
    PLInit();

    Event(updateEvt, updateevent); /* Bind the standard event handlers. */
    Event(activateEvt, activatevent);
    Event(windowEvt, windowevent);
    Event(menuEvt, menuevent);
    Event(scrapEvt, scrapevent);

    NewMenu(FILEMENU, "File");
    AppendMenu(FILEMENU, "/OOpen\311;-;(/SSave As\311;-;Restore Settings\311;Save Settings\311;-;/QQuit");
    NewMenu(EDITMENU, "Edit");
    AppendMenu(EDITMENU, "/XCut;/CCopy;/VPaste;-;Addresses File\311");
				/* If we don't have our TTY on stdin */
				/* (i.e. we were started in the background) */
				/* then the editor probably won't have a */
				/* good idea about the actual size of its */
				/* window, so block the Edit Addresses func. */
				/* This is easier to do than trying to wade */
				/* through all the environment stuff to make */
				/* it work right (though such is possible). */
    if (!isatty(0))
	DisableItem(EDITMENU, EDITADDR);
    NewMenu(MISCMENU, "Misc");
    AppendMenu(MISCMENU, "Hex Limit\311;Hardware Intfc\311;"
	       "Format Overrides\311;Histograms\311;Animation;Sound;Help\311");
    CheckItem(MISCMENU, ANIMATION, animated);
    if (!WMIsLocal(wmfd)) {	/* No sound if displaying remotely. */
	DisableItem(MISCMENU, SOUND);
	sound = 0;
    } else {
	CheckItem(MISCMENU, SOUND, sound);
    }
    dialog = DMNew(Dmain, windowid, 0);
    applyargs();
    checkprogs();
#if 0
    DMCollect(dialog);
#else
/*
** Do everything DMCollect does, but don't let the ESC key be a synonym
** for the STOP button more than once.  This prevents (unfortunately common)
** double actuation, which loses all the sniffed data.
*/
    {
	Point click;
	EventRecord *ep;
	int group;

	MouseMode(F_KEYDOWN | F_KEYUP | F_MOUSE1);
	group = 0;
	while (group < DMout) {
	    do {
		ep = GetEvent(1 << windowid, 1 << mouseEvt | 1 << keyEvt);
		if (!ep || ep == (EventRecord *) -1) {
		    ep = 0;
		    break;		/* Bail out on 'errors'. */
		}
	    } while (ep->message & F_KEYUP || !(dialog->flags & DIenable));
	    if (!ep)
		break;
	    click = ep->where.corner.topLeft;
	    if ((ep->what == mouseEvt) && (ep->message & F_KEYDOWN)) {
		group = DMMouseEvent(dialog, ep);
	    } else if (ep->what == keyEvt) {
		if ((ep->message & (F_KEYMASK|F_CTRL)) == ('i' | F_CTRL))
		    toggleicon();
		else if ((ep->message & F_KEYMASK) != K_ESC || sniffing)
		    group = DMKeyEvent(dialog, ep);
	    } else if (ep->what == mouseEvt && (ep->message & F_KEYDOWN)) {
		if (!PtInRect(&click, &dialog->dialog->rect)) {
		    Bell();
		}
	    }
	}
    }
#endif
    if (sniffchild)
	stopsniff();
    if (hasfile > 0)
	unlink(filename);
    if (doexportscrap) {
	writescrap();
	FlushTwice();
    }
    if (sflss.returnedname) {	/* Restore settings re-execs the sniffer. */
	close(wmfd);
	close(trapqfd);
	execl("/bin/sh", "sh", "-c", sflss.returnedname, 0);
    }
    return 0;
}

processargs(argc, argv)
int argc;
char *argv[];
{
    int enabled;
    char *argp;
    char scratch[256];

    enabled = 0;
    progname = *argv;
    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'a':		/* Animated progress report. */
		animated = 1;
		break;

	      case 'b':		/* Recv bad packets too. */
		bflag = 1;
		break;

	      case 'c':		/* Receive both ends of a conversation. */
		cflag = 1;
		break;

	      case 'D':		/* Destination address filter. */
		if (!*argp)
		    --argc, argp = *++argv;
		Dflag = argp;
		break;

	      case 'f':		/* Filter options. */
		switch (*argp++) {
		  case 'd':
		    if (!*argp)
			--argc, argp = *++argv;
		    strncpy(filtdata, argp, sizeof filtdata);
		    break;

		  case 'i':
		    datafilt = 1;
		    break;

		  case 'o':
		    if (!*argp)
			--argc, argp = *++argv;
		    strncpy(filtoffset, argp, sizeof filtoffset);
		    break;

		  case 'x':
		    filterexc = 1;
		    break;
		}
		break;

	      case 'h':		/* Hex/Histogram options. */
		switch (*argp++) {
		  case 'd':
		    datalimit = 1;
		    break;

		  case 'g':
		    if (!*argp)
			--argc, argp = *++argv;
		    histogram = atoi(argp);
		    break;

		  case 'l':
		    if (!*argp)
			--argc, argp = *++argv;
		    strncpy(limitsize, argp, sizeof limitsize);
		    break;
		}
		break;

	      case 'n':		/* Buffer size. */
		if (*argp == 'o') {
		    nonose = 1;
		} else {
		    if (!*argp)
			--argc, argp = *++argv;
		    nflag = argp;
		}
		break;

	      case 'o':		/* Format Overrides. */
		if (*argp++ == 'v') {
		    if (!*argp)
			--argc, argp = *++argv;
		    strncpy(overrides, argp, sizeof overrides);
		}
		break;

	      case 'p':		/* Promiscuous mode. */
		pflag = 1;
		break;

	      case 'P':		/* Protocol filter. */
		if (!*argp)
		    --argc, argp = *++argv;
		Pflag = argp;
		break;

	      case 'S':		/* Source address filter. */
		if (!*argp)
		    --argc, argp = *++argv;
		Sflag = argp;
		break;

	      case 's':		/* Audio packet capture feedback. */
		sound = 1;
		break;

	      case 't':		/* Trigger options. */
		switch (*argp++) {
		  case 'd':
		    if (!*argp)
			--argc, argp = *++argv;
		    strncpy(trigdata, argp, sizeof trigdata);
		    break;

		  case 'i':
		    triggerinv = 1;
		    break;

		  case 'o':
		    if (!*argp)
			--argc, argp = *++argv;
		    strncpy(trigoffset, argp, sizeof trigoffset);
		    break;

		  case 'p':
		    if (!*argp)
			--argc, argp = *++argv;
		    tpos = atoi(argp);
		    break;

		  case 'r':
		    trigger = 1;
		    triggerinv = normal = nostop = 0;
		    break;

		  case 's':
		    nostop = 1;
		    normal = trigger = triggerinv = 0;
		    break;
		}
		break;

	      case 'V':		/* Version printout. */
		strcpy(scratch, progname);
		strcat(scratch, " Version 1.1\n");
		write(1, scratch, StringLen(scratch));
		exit(0);
		break;

	      case 'v':
		switch (*argp) {
		  case 'a':
		    allheaders = 1;
		    break;

		  case 'd':
		    datapacks = 1;
		    break;

		  case 'r':
		    rawpacks = 1;
		    break;

		  case 's':
		    symbolics = 1;
		    break;

		  case 't':
		    deltatime = 1;
		    break;
		}
		break;

	      case 'w':		/* WAN/Wincom Level 1 interface */
		switch (*argp) {
		  case 'a':
		    level1 = L1WAN;
		    break;

		  case 'i':
		    level1 = L1WINCOM;
		    break;
		}
		break;
	    }
	} else {		/* Assume it's a 'settings' file. */
	    close(wmfd);
	    close(trapqfd);
	    execl("/bin/sh", "sh", "-c", argp, 0);
	}
    }
}

/*
** Apply command-line arguments that have to be done _after_ the
** dialog is created.
*/
applyargs()
{
    if (bflag)			/* Recv bad packets too. */
	SetCtlValue(DMGetId(dialog, DMAIN_BadPacks), 1);

    if (cflag) {		/* Receive both ends of a conversation. */
	SetCtlValue(DMGetId(dialog, DMAIN_Conversation), 1);
	if (Dflag && Sflag)
	    DMItemEnable(dialog, DMAIN_Conversation);
    }
    if (Dflag) {		/* Destination address filter. */
	DMSetText(dialog, DMAIN_DstFilterAddr, Dflag);
	DMItemEnable(dialog, DMAIN_DstFilterAddr);
	SetCtlValue(DMGetId(dialog, DMAIN_Destination), 1);
    }

    if (nflag)			/* Buffer size. */
	DMSetText(dialog, DMAIN_bufsize, nflag);

    if (pflag)			/* Promiscuous mode. */
	DMExclusive(dialog, DMAIN_PromiscuousPacks, 3);

    if (Pflag) {		/* Protocol filter. */
	DMExclusive(dialog, DMAIN_HexProt, 2);
	DMSetText(dialog, DMAIN_HexProtTE, Pflag);
	strncpy(fltrhexprot, Pflag, sizeof fltrhexprot);
	DMItemEnable(dialog, DMAIN_HexProtTE);
    }

    if (Sflag) {		/* Source address filter. */
	DMSetText(dialog, DMAIN_SrcFilterAddr, Sflag);
	DMItemEnable(dialog, DMAIN_SrcFilterAddr);
	SetCtlValue(DMGetId(dialog, DMAIN_Source), 1);
    }

    if (level1) {		/* Handle initial WAN/Wincom Level 1 Intfc. */
	if (level1 > L1WAN) {
	    level1 = L1LAN;
	    setlevel1(1);
	} else {
	    setlevel1(0);
	}
    }
    if (allheaders)
	SetCtlValue(DMGetId(dialog, DMAIN_AllHeaders), 1);
    if (datapacks)
	SetCtlValue(DMGetId(dialog, DMAIN_DataAftrHdr), 1);
    if (rawpacks)
	SetCtlValue(DMGetId(dialog, DMAIN_DispRawPacks), 1);
    if (symbolics)
	SetCtlValue(DMGetId(dialog, DMAIN_DispSymbolic), 1);
    if (deltatime)
	SetCtlValue(DMGetId(dialog, DMAIN_DeltaTime), 1);
}

/*
** Make sure our support programs are present.
*/

checkprogs()
{
    checkprog(sniffer)
	|| checkprog(snifferdump)
	    || checkprog(statprog)
		|| checkprog(copier)
		    || checkprog(viewer)
			|| checkprog(picker)
			    || checkperm();
}

checkprog(prog)
char *prog;
{
    char scratch[256];

    if (access(prog, 01)) {
	strcpy(scratch, "Missing or nonexecutable ");
	strcat(scratch, prog);
	strcat(scratch, " program!");
	setstatus(scratch);
	return 1;
    }
    return 0;
}

/*
** Quick check to see if sufficient permission exists to sniff,
** for a warning message.
*/

checkperm()
{
    struct stat fs;

    if (geteuid()
	&& (stat(sniffer, &fs) || !(fs.st_mode & S_ISUID) || fs.st_uid)) {
	setstatus("You don\325t have permission to sniff.");
    }
}

void
killscrap()
{
    scrapsize = 0;		/* Delete private scrap. */
    doimportscrap = 0;
}

void
fetchscrap()
{
    long size;

    killscrap();		/* Create private scrap. */
    scrapsize = GetScrap(scrapbuf, 'TEXT', sizeof(scrapbuf));
    TESetScrapLen(scrapsize);
    if (scrapsize) {
	EnableItem(EDITMENU, EDITPASTE);
	BlockMove(scrapbuf, TEGetScrapAddr(), scrapsize);
    } else
	DisableItem(EDITMENU, EDITPASTE);
}

void
writescrap()
{
    if (doexportscrap) {
	ZeroScrap();		/* Export private scrap. */
	PutScrap(scrapsize, 'TEXT', scrapbuf);
	PublishScrap(0);
	doexportscrap = 0;
    }
}

void
activatevent(ep)
EventRecord *ep;
{
    short i;

    if (ep->message & 1) {	/* Activating? */
	mywindowisinfront = 1;
	if (doimportscrap)
	    fetchscrap();
	if (!iconified) {
	    if (savechild || openchild)
		SendBehind(windowid, backWindow);
	    else if (dialog2)
		DMDlgEnable(dialog2);
	    else
		DMDlgEnable(dialog);
	}
    } else {			/* Deactivating. */
	mywindowisinfront = 0;
	if (doexportscrap)
	    writescrap();
	if (!iconified) {
	    if (dialog2)
		DMDlgDisable(dialog2);
	    else
		DMDlgDisable(dialog);
	}
    }
}

void
scrapevent(ep)
EventRecord *ep;
{
    if (!ep->message) {         /* Someone published new scrap. */
        doimportscrap = 1;
        if (mywindowisinfront)
            fetchscrap();
    } else {                    /* Someone wants our (?) scrap. */
        if (doexportscrap && mywindowisinfront)
            writescrap();
    }
}

void
updateevent(ep)
EventRecord *ep;
{
    SetPort(ep->who);
    BeginUpdate(ep->who);
    if (iconified)
	drawicon();
    else {
	DMUpdate(dialog);
	if (dialog2)
	    DMUpdate(dialog2);
    }
    EndUpdate(ep->who);
}

/*
** Makes BIG assumptions about the icon buried in our object code.
*/
drawicon()
{
    extern bitobj icon[];
    Rect rect;

    SetRect(&rect, 0, 0, 32, 32);
    PlotIcon(&rect, (Icon *) &icon[1]);
}

/*
** 'Validation' routine for the Nose UserItem.  This routine actually
** does the drawing of the item.
*/

#define NOSE !nonose

drawnose(cbp, dip, when)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
{
    Rect rnose;
    PicPtr theNose;

    theNose = &nose;
    rnose = dip->rect;
    rnose.edge.right = rnose.edge.left + theNose->picFrame.edge.right;
    rnose.edge.bottom = rnose.edge.top + theNose->picFrame.edge.bottom;
    if (NOSE)
	DrawPicture(&theNose, &rnose);
}

drawsnort(cbp, dip, when)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
{
    Rect rsnort;
    PicPtr theSnort;

    theSnort = &snort;
    rsnort = dip->rect;
    rsnort.edge.right = rsnort.edge.left + theSnort->picFrame.edge.right;
    rsnort.edge.bottom = rsnort.edge.top + theSnort->picFrame.edge.bottom;
    if (sniffchild) {
	if (nonose)
	    setstatus("Sniffing\311");
	else {
	    if (!animated)
		DrawPicture(&theSnort, &rsnort);
	    else if (!dialog2)
		drawprogress(1);
	}
    } else {
	if (animated)
	    rsnort = Dmain[DMAIN_status].rect;
	if (animated || NOSE)
	    EraseRect(&rsnort);
    }
}

static void
uniconify()
{
    EnableMenu(FILEMENU);
    EnableMenu(MISCMENU);
    UnIconifyWindow(windowid);
    BeginUpdate(windowid);
    DMUpdate(dialog);
    if (dialog2)
	DMUpdate(dialog2);
    EndUpdate(windowid);
}

static void
iconify()
{
    DisableMenu(FILEMENU);
    DisableMenu(MISCMENU);
    IconifyWindow(windowid);
    BeginUpdate(windowid);
    drawicon();
    EndUpdate(windowid);
}

static void
toggleicon()
{
    iconified = !iconified;
    if (!iconified) {		/* Inverted test! */
	uniconify();
    } else {
	iconify();
    }
}

void
windowevent(ep)
EventRecord *ep;
{
    switch (ep->message & F_KEYMASK) {
      case windowGoaway:
	if (sflss.returnedname) {
	    free(sflss.returnedname);
	    sflss.returnedname = 0;
	}
	ReleaseEvents();
	break;

      case windowZoom:
	toggleicon();
	break;

      case windowMove:
	OffsetRect(&winrect, ep->where.edge.left, ep->where.edge.top);
	MoveWindow(ep->who, winrect.edge.left, winrect.edge.top, 0);
	break;
    }
}

static void
starteditor()
{
    int i, status;
    char *editor, *line, scratch[80];
    struct vterm_window wind;

    if (!(editor = getenv("VISUAL")))
	if (!(editor = getenv("EDITOR")))
	    editor = "/usr/bin/dmacs";
    if (!(line = getenv("LINE")))
	line = "/dev/tty";
    status = dnix(F_IOCR, 0, &wind, sizeof(wind), VTERM_WSTATS, 0);
    if (!(editchild = fork())) {
	if (shmp)
	    shmdt(shmp);
	for (i=0; i<_NFILE; i++) { /* Child. */
	    close(i);
	}
	dup(dup(i = open(line, O_RDWR))); /* Stdin/out/error for the editor. */
	if (i < 0)
	    exit(1);
	if (status > 0)
	    dnix(F_IOCW, 0, &wind, status, VTERM_WSTATS, 0); /* Dupe */
	ioctl(0, VTERM_NOLURK, 0); /* window characteristics. */
	strcpy(scratch, "Network Addresses : ");
	strncat(scratch, nodename, sizeof scratch);
	dnix(F_IOCW, 0, scratch, strlen(scratch), VTERM_RENAME, 0);
	execlp(editor, editor, addresses, 0);
	_exit(1);
    } else if (editchild > 0) { /* Parent? */
	waitforchild();
    } else {			/* Fork failed. */
	editchild = 0;
    }
}

setlevel1(wincom)
int wincom;
{
    int capmode;

    DMGet(dialog, DMAIN_CapMode, &capmode);
    if (level1) {
	snifferrs[4] = wandev;
	DMCodeHide(dialog, DLAN, DLEVEL1MASK);
	if (capmode) {		/* Capture mode settings? */
	    DMCodeHide(dialog, DWINCOM, DLEVEL1MASK);
	    DMCodeShow(dialog, DWAN, DLEVEL1MASK);
	}
	capmode = 0;
    } else {			/* Not WAN, */
	if (wincom) {		/* Wincom L1? */
	    level1 = L1WINCOM;
	    snifferrs[4] = wincomdev;
	    DMCodeHide(dialog, DLAN, DLEVEL1MASK);
	    if (capmode) {		/* Capture mode settings? */
		DMCodeHide(dialog, DWAN, DLEVEL1MASK);
		DMCodeShow(dialog, DWINCOM, DLEVEL1MASK);
	    }
	    capmode = 0;
	} else {		/* Must be LAN. */
	    snifferrs[4] = landev;
	}
    }
    return capmode;
}

void
menuevent(ep)
EventRecord *ep;
{
    int i, j;
    short menu, item;
    long message;
    DMDcb *dp;
    char *cp, scratch[5];

    BeginMenu(message = ep->message);
    menu = (message >> 16) & 0xFFFF;
    item = message & 0xFFFF;
    switch (menu) {
      case FILEMENU:
	switch (item) {
	  case FILEOPEN:
	    sfo.options = SF_GETFILE;
	    if (sfo.initialname)
		free(sfo.initialname);
	    sfo.initialname = sfo.returnedname;
	    i = NWFilePicker(&sfo, trapqfd);
	    if (i >= 0) {
		DMItemDisable(dialog, DMAIN_GoButton);
		openchild = i;
		waitforchild();
	    }
	    break;

	  case FILESAVE:
	    sfs.options = SF_PUTFILE;
	    if (sfs.initialname)
		free(sfs.initialname);
	    sfs.initialname = sfs.returnedname;
	    i = NWFilePicker(&sfs, trapqfd);
	    if (i >= 0) {
		DMItemDisable(dialog, DMAIN_GoButton);
		savechild = i;
		waitforchild();
	    }
	    break;

	  case FILELSET:
	    sflss.options = SF_GETFILE;
	    if (sflss.initialname)
		free(sflss.initialname);
	    sflss.initialname = sflss.returnedname;
	    i = NWFilePicker(&sflss, trapqfd);
	    if (i >= 0) {
		DMItemDisable(dialog, DMAIN_GoButton);
		lsetchild = i;
		waitforchild();
	    }
	    break;

	  case FILESSET:
	    sflss.options = SF_PUTFILE;
	    if (sflss.initialname)
		free(sflss.initialname);
	    sflss.initialname = sflss.returnedname;
	    i = NWFilePicker(&sflss, trapqfd);
	    if (i >= 0) {
		DMItemDisable(dialog, DMAIN_GoButton);
		ssetchild = i;
		waitforchild();
	    }
	    break;

	  case FILEQUIT:
	    if (sflss.returnedname) {
		free(sflss.returnedname);
		sflss.returnedname = 0;
	    }
	    ReleaseEvents();
	    break;
	}
	break;

      case EDITMENU:
	if ((dp = dialog2) || (dp = dialog)) { /* An active dialog? */
	    if (dp->curTid) {	/* An active TE within it? */
		switch (item) {
		  case EDITCUT:
		    TECut(dp->curTid);
		  getscrap:
		    killscrap();
		    doexportscrap = 1; /* Set export-on-deactivate flag. */
		    scrapsize = TEGetScrapLen();
		    if (scrapsize > sizeof scrapbuf)
			scrapsize = sizeof scrapbuf;
		    if (cp = TEGetScrapAddr())
			BlockMove(cp, scrapbuf, scrapsize);
		    break;

		  case EDITCOPY:
		    TECopy(dp->curTid);
		    goto getscrap;
		    break;

		  case EDITPASTE:
		    TEPaste(dp->curTid);
		    break;
		}
	    }
	}
	switch (item) {
	  case EDITADDR:
	    DisableItem(EDITMENU, EDITADDR);
	    starteditor();
	    break;
	}
	break;

      case MISCMENU:
	i = 0;
	if (item < ANIMATION || item == HELP) {
	    DisableMenu(FILEMENU);
	    DisableMenu(MISCMENU);
	    DMDlgDisable(dialog);
	}
	switch (item) {
	  case MISCLIMIT:
	    dialog2 = DMNew(Dlimit, windowid, 0);
	    if (DMCollect(dialog2) == DMok) {
		DMGet(dialog2, DLIMIT_limithd, &datalimit);
		DMGet(dialog2, DLIMIT_limit, limitsize);
	    }
	    break;

	  case MISCHDWR:
	    dialog2 = DMNew(Dlevel1, windowid, 0);
	    if (DMCollect(dialog2) == DMok) {
		DMGet(dialog2, DLEVEL1_wan, &level1);
		DMGet(dialog2, DLEVEL1_wincom, &j);
		i = setlevel1(j);
	    }
	    break;

	  case MISCOVERRIDE:
	    dialog2 = DMNew(Dover, windowid, 0);
	    if (DMCollect(dialog2) == DMok) {
		DMGet(dialog2, DOVER_overrides, overrides);
	    }
	    break;

	  case MISCHISTOGRAM:
	    dialog2 = DMNew(Dhist, windowid, 0);
	    i = histogram;
	    if (!hasfile)
		DMItemDisable(dialog2, DHIST_ok);
	    if (DMCollect(dialog2) == DMok) {
		runhistogram(histogram);
		InvalRect(&Dmain[DMAIN_StopButton].rect);
	    } else {		/* Dialog cancelled. */
		histogram = i;
	    }
	    i = 0;
	    break;

	  case ANIMATION:
	    animated = !animated;
	    CheckItem(MISCMENU, ANIMATION, animated);
	    break;

	  case SOUND:
	    sound = !sound;
	    CheckItem(MISCMENU, SOUND, sound);
	    break;

	  case HELP:
	    dialog2 = DMNew(Dhelp, windowid, 0);
	    i = lasthelp;
	    if (DMCollect(dialog2) == DMok) {
		switch (lasthelp) {
		  case DHELP_sniffer: i = 0; break;
		  case DHELP_sniffwin: i = 4; break;
		  case DHELP_snifferdump: i = 2; break;
		  case DHELP_packstats: i = 5; break;
		  case DHELP_sniffnet: i = 6; break;
		  case DHELP_blower: i = 1; break;
		  case DHELP_sniffhdlc: i = 3; break;
		}
		runhelp(helps[i]);
	    } else {		/* Dialog cancelled. */
		lasthelp = i;
	    }
	    i = 0;
	    break;
	}
	if (item < ANIMATION || item == HELP) {
	    DMDisposeNoInval(dialog2);
	    dialog2 = 0;
	    if (i) {		/* Turning back on LAN capture goodies? */
		DMCodeHide(dialog, DWAN, DLEVEL1MASK);
		DMCodeHide(dialog, DWINCOM, DLEVEL1MASK);
		DMCodeShow(dialog, DLAN, DLEVEL1MASK);
	    }
	    DMDlgEnableUpdate(dialog);
	    EnableMenu(FILEMENU);
	    EnableMenu(MISCMENU);
	}
	break;
    }
    EndMenu(message);
}

/*
** This routine actually runs the Sniff.
*/

ValGo(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    return startsniff();
}

/*
** Make up a name for the selected protocol.
*/
char *
protname()
{
    int j;

    DMGet(dialog, DMAIN_SafeProt, &j);
    if (j)
	return "safe";
    else if (DMGet(dialog, DMAIN_BootProt, &j), j)
	return "boot";
    else if (DMGet(dialog, DMAIN_DnetProt, &j), j)
	return "dnet";
    else if (DMGet(dialog, DMAIN_DmapProt, &j), j)
	return "dmap";
    else if (DMGet(dialog, DMAIN_IpProt, &j), j)
	return "ip";
    else if (DMGet(dialog, DMAIN_ArpProt, &j), j)
	return "arp";
    else if (DMGet(dialog, DMAIN_AtalkProt, &j), j)
	return "atalk";
    else if (DMGet(dialog, DMAIN_AarpProt, &j), j)
	return "aarp";
    else if (DMGet(dialog, DMAIN_HexProt, &j), j) {
	return fltrhexprot;
    }
}

/*
** Start a sniffing process for capturing into a file.
*/

startsniff()
{
    return startsniffnet(0);
}

/*
** Start a sniffing process, either for capturing into a file or
** for feeding a pipeline to the histogram process.
*/

startsniffnet(hist)
int hist;			/* FD when starting a histogram capture. */
{
    int i, j, capt;
    char *args[40];		/* execv()'s argument list. */
    char *itoa();
    char bufsiz[5];
    char saddr[20];
    char daddr[20];

    if (sniffchild) {
	Bell();
	return DM_FAIL;
    }
    capt = !hist;
    clearstatus();
    if (DMGet(dialog, DMAIN_CapMode, &i), i) {
	DMGet(dialog, DMAIN_HexProtTE, fltrhexprot);
	DMGet(dialog, DMAIN_SrcFilterAddr, fltrsrcaddr);
	DMGet(dialog, DMAIN_DstFilterAddr, fltrdstaddr);
	DMGet(dialog, DMAIN_Source, &fltrsrc);
	DMGet(dialog, DMAIN_Destination, &fltrdst);
	DMGet(dialog, DMAIN_Conversation, &fltrconv);
    }
    if (hasfile && capt) {
	if (hasfile > 0)
	    unlink(filename);
	hasfile = 0;
	DisableItem(FILEMENU, FILESAVE);
    }
    if (capt) {
	overrides[0] = 0;	/* Clear the packet format overrides. */
	strcpy(filename, "/tmp/sniffXXXXXX");
	if (!mktemp(filename))
	    return DM_FAIL;
    }
    i = j = 0;
    if (level1 == L1WAN) {
	args[i++] = sniffhdlc;
    } else if (level1 == L1WINCOM) {
	args[i++] = sniffwin;
    } else {
	args[i++] = sniffer;
	if (!animated) {
	    args[i++] = "-m";	/* Multiple-read mode. */
	} else {
	    struct tms timebuf;

	    lastticks = times(&timebuf);
	    lastseen = 0;
	}
    }
    args[i++] = "-l";		/* Lock process in memory. */
    DMGet(dialog, DMAIN_bufsize, bufsiz);
    args[i++] = "-n";		/* Size of sniff buffer. */
    args[i++] = capt ? bufsiz : "0";
    if (level1 == L1LAN) {
	DMGet(dialog, DMAIN_PromiscuousPacks, &j);
	if (j)
	    args[i++] = "-p";
	DMGet(dialog, DMAIN_BadPacks, &j);
	if (j)
	    args[i++] = "-b";
	if (fltrsrc) {
	    args[i++] = "-S";
	    strncpy(saddr, fltrsrcaddr, sizeof saddr);
	    if (procaddr(saddr)) {
		setstatus("Source capturing filter name not found!");
		return DM_FAIL;
	    } else
		args[i++] = saddr;
	}
	if (fltrdst) {
	    args[i++] = "-D";
	    strncpy(daddr, fltrdstaddr, sizeof daddr);
	    if (procaddr(daddr)) {
		setstatus("Destination capturing filter name not found!");
		return DM_FAIL;
	    } else
		args[i++] = daddr;
	}
	if (fltrconv)
	    args[i++] = "-c";
	DMGet(dialog, DMAIN_NoProt, &j);
	if (!j) {		/* Protocol filter? */
	    args[i++] = "-P";	/* Yes. */
	    args[i++] = protname();
	}
    }
    if (nostop || hist)
	args[i++] = "-s";
    if (trigger && capt) {
	args[i++] = "-to";
	args[i++] = trigoffset;
	args[i++] = "-td";
	args[i++] = trigdata;
	args[i++] = "-tp";
	args[i++] = trigpos;
	if (triggerinv)
	    args[i++] = "-ti";
    }
    if (datafilt | filterexc) {
	args[i++] = "-fo";
	args[i++] = filtoffset;
	args[i++] = "-fd";
	args[i++] = filtdata;
	if (filterexc)
	    args[i++] = "-fx";
    }
    if (sound)
	args[i++] = "-a";
    args[i] = 0;
    if (animated	/* Shared-memory stats for Ethernet/Wincom sniff? */
	&& (level1 == L1LAN || level1 == L1WINCOM)
	&& (shmkey = ftok(args[0], 0)) != (key_t) -1) {
	if ((shmid = shmget(shmkey, sizeof(*shmp), IPC_CREAT | 0666)) != -1) {
	    if ((shmp = (struct sniffstat *) shmat(shmid, 0, 0))
		== (struct sniffstat *) -1) {
		shmp = 0;
	    } else {
		memset(shmp, 0, sizeof(*shmp));
		args[i++] = "-k";
		args[i++] = itoa(shmkey, 10, 0, 0);
		args[i++] = 0;
	    }
	}
    }
    if (!(sniffchild = fork())) {
	for (i=0; i<_NFILE; i++) { /* Child. */
	    if (capt || i != hist)
		close(i);
	}
	open(nulldev, O_RDONLY); /* Stdin for the sniffer. */
	if (hist) {		/* Stdout for the sniffer. */
	    dup(hist); close(hist);
	} else {
	    if ((i = creat(filename, 0600)) < 0)
		_exit(1);
	}
	dup(0);			/* Stderr for the sniffer. */
	execv(args[0], args);
	_exit(1);
    } else if (sniffchild > 0) { /* Parent? */
	sniffing = 1;
	DMItemEnable(dialog, DMAIN_StopButton);
	DMItemDisable(dialog, DMAIN_ViewButton);
	DMItemDisable(dialog, DMAIN_GoButton);
	DisableItem(FILEMENU, FILEOPEN);
	waitforchild();
	nowaitwait.arg = hist;	/* Mark what kind it is. */
	if (animated) {
	    napdone(&nowaitnap, 0, 0); /* Start the poller. */
	    drawprogress(1);
	} else
	    drawsnort(dialog, &Dmain[DMAIN_Snort], DMbeforeDrawing);
    } else {			/* Fork failed. */
	sniffchild = 0;
	return DM_FAIL;
    }
    return DM_PASS;
}

PreSize(cbp, dip, cp)
DMDcb *cbp;
DMItem *dip;
char *cp;
{
    strcpy(cp, "128");
}

PFtrigoffset(cbp, dip, cp)
DMDcb *cbp;
DMItem *dip;
char *cp;
{
    strcpy(cp, trigoffset);
}

PFtrigdata(cbp, dip, cp)
DMDcb *cbp;
DMItem *dip;
char *cp;
{
    strcpy(cp, trigdata);
}

PFtrigmatch(cbp, dip, value)
DMDcb *cbp;
DMItem *dip;
int *value;
{
    *value = trigger ^ triggerinv;
}

PFtrigmismatch(cbp, dip, value)
DMDcb *cbp;
DMItem *dip;
int *value;
{
    *value = triggerinv;
}

PFtrigfull(cbp, dip, value)
DMDcb *cbp;
DMItem *dip;
int *value;
{
    *value = normal;
}

PFtrignone(cbp, dip, value)
DMDcb *cbp;
DMItem *dip;
int *value;
{
    *value = nostop;
}

PFtrigpos(cbp, dip, value)
DMDcb *cbp;
DMItem *dip;
int *value;
{
    *value = (dip->ident == tpos);
    if (!trigger)		/* Disable trigger position selectors? */
	DMPreItemDisable(cbp, dip);
}

PFfiltoffset(cbp, dip, cp)
DMDcb *cbp;
DMItem *dip;
char *cp;
{
    strcpy(cp, filtoffset);
}

PFfiltdata(cbp, dip, cp)
DMDcb *cbp;
DMItem *dip;
char *cp;
{
    strcpy(cp, filtdata);
}

PFdatafilt(cbp, dip, value)
DMDcb *cbp;
DMItem *dip;
int *value;
{
    *value = datafilt;
}

PFdataexcl(cbp, dip, value)
DMDcb *cbp;
DMItem *dip;
int *value;
{
    *value = filterexc;
}

PFpicklabel(cbp, dip, cp)
DMDcb *cbp;
DMItem *dip;
char *cp;
{
    strcpy(cp, picklabel);
}

PFlimit(cbp, dip, cp)
DMDcb *cbp;
DMItem *dip;
char *cp;
{
    strcpy(cp, limitsize);
}

PFlimithd(cbp, dip, value)
DMDcb *cbp;
DMItem *dip;
int *value;
{
    *value = datalimit;
}

PFlan(cbp, dip, value)
DMDcb *cbp;
DMItem *dip;
int *value;
{
    *value = !level1;
}

PFwan(cbp, dip, value)
DMDcb *cbp;
DMItem *dip;
int *value;
{
    *value = (level1 == L1WAN);
}

PFwincom(cbp, dip, value)
DMDcb *cbp;
DMItem *dip;
int *value;
{
    *value = (level1 == L1WINCOM);
}

PFoverrides(cbp, dip, cp)
DMDcb *cbp;
DMItem *dip;
char *cp;
{
    strcpy(cp, overrides);
}

sorter(p1, p2)
char **p1, **p2;
{
    return strcmp(*p1, *p2);
}

/*
** Read the addresses file and stuff the pick list with it.
*/

PFpicker(cbp, dip, psp)
DMDcb *cbp;
DMItem *dip;
DMPick *psp;
{
    int fd, i, len;
    struct stat fs;
    char *fbuf, *cp, *cp2, *cp3, *cp4, *alloca();

    if (!pickarray.arrayhead && !stat(addresses, &fs)) {
	if (fbuf = alloca(fs.st_size + 1)) {
	    if ((fd = open(addresses, O_RDONLY|O_NOCTTY)) >= 0) {
		if (read(fd, fbuf, fs.st_size) == fs.st_size) {
		    fbuf[fs.st_size] = 0;
		    i = 0;	/* Count valid entries. */
		    cp = fbuf;
		    while (cp && cp < (fbuf + fs.st_size)) {
			if (*cp != '#')
			    i++;
			if (cp = strchr(cp, '\n'))
			    cp++;
		    }
		    if (i) {	/* Got valid entries?  Load 'em up! */
			if (pickarray.arrayhead =
			    (char **) calloc(i, sizeof(char *))) {
			    i = 0;
			    cp = fbuf;
			    while (cp && cp < (fbuf + fs.st_size)) {
				if (*cp != '#') {
				    if ((cp2 = strchr(cp, ' ')) ||
					(cp2 = strchr(cp, '\t'))) {
					cp4 = cp2; /* End of address. */
					while (*cp2 == ' ' ||
					       *cp2 == '\t')
					    cp2++; /* Skip whitespace. */
					cp3 = cp2;
					while (*cp3 && *cp3 != '\n')
					    cp3++;
					len = cp3 - cp2;
					if (cp3 = calloc(len+2+cp4-cp, 1)) {
					    strncpy(cp3, cp2, len);
					    strncpy(&cp3[len+1], cp, cp4-cp);
					    pickarray.arrayhead[i++] = cp3;
					    pickarray.length = i;
					}
				    }
				}
				if (cp = strchr(cp, '\n'))
				    cp++;
			    }
			    if (pickarray.length) { /* Sort the list. */
				qsort(pickarray.arrayhead, pickarray.length,
				      sizeof(pickarray.arrayhead[0]), sorter);
			    }
			}
		    }
		}
		close(fd);
	    }
	}
    }
    if (psp) {
	*psp = pickarray;
    }
}

PFhelp(cbp, dip, value)
DMDcb *cbp;
DMItem *dip;
int *value;
{
    *value = (dip->ident == lasthelp);
}

ValHelp(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    lasthelp = dip->ident;
}

PFhist(cbp, dip, value)
DMDcb *cbp;
DMItem *dip;
int *value;
{
    int val;

    val = 0;
    switch (dip->ident) {
      case DHIST_sizes: if (histogram & HIST_SIZE) val = 1; break;
      case DHIST_srcaddr: if (histogram & HIST_SADR) val = 1; break;
      case DHIST_dstaddr: if (histogram & HIST_DADR) val = 1; break;
      case DHIST_protocols: if (histogram & HIST_PROT) val = 1; break;
    }
    *value = val;
}

ValHist(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    switch (dip->ident) {
      case DHIST_sizes:
	if (value)
	    histogram |= HIST_SIZE;
	else
	    histogram &= ~HIST_SIZE;
	break;

      case DHIST_srcaddr:
	if (value)
	    histogram |= HIST_SADR;
	else
	    histogram &= ~HIST_SADR;
	break;

      case DHIST_dstaddr:
	if (value)
	    histogram |= HIST_DADR;
	else
	    histogram &= ~HIST_DADR;
	break;

      case DHIST_protocols:
	if (value)
	    histogram |= HIST_PROT;
	else
	    histogram &= ~HIST_PROT;
	break;
    }
}

ValFresh(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    if (value)
	DMItemEnable(dialog2, DHIST_ok);
    else if (!hasfile)
	DMItemDisable(dialog2, DHIST_ok);
}

/*
** Run the triggering dialog.
*/

ValTrig(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    int savetpos;

    DisableMenu(FILEMENU);
    DisableMenu(MISCMENU);
    DMDlgDisable(dialog);
    savetpos = tpos;
    dialog2 = DMNew(Dtrig, windowid, 0);
    if (DMCollect(dialog2) == DMok) {
	DMGet(dialog2, DTRIG_trigmatch, &trigger);
	DMGet(dialog2, DTRIG_trigmismatch, &triggerinv);
	if (triggerinv)
	    trigger = 1;
	DMGet(dialog2, DTRIG_trignone, &nostop);
	DMGet(dialog2, DTRIG_trigfull, &normal);
	DMGet(dialog2, DTRIG_trigoffset, trigoffset);
	DMGet(dialog2, DTRIG_trigdata, trigdata);
    } else {
	tpos = savetpos;
    }
    switch (tpos) {
      case DTRIG_trig0:
	strcpy(trigpos, "0");
	break;

      case DTRIG_trig25:
	strcpy(trigpos, "25");
	break;

      case DTRIG_trig50:
	strcpy(trigpos, "50");
	break;

      case DTRIG_trig75:
	strcpy(trigpos, "75");
	break;

      case DTRIG_trig100:
	strcpy(trigpos, "100");
	break;
    }
    DMDisposeNoInval(dialog2);
    dialog2 = 0;
    EnableMenu(FILEMENU);
    EnableMenu(MISCMENU);
    DMDlgEnableUpdate(dialog);
    return DM_PASS;
}

Valtrigoffset(cbp, dip, when, val)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
DMTEVal *val;
{
    int ch;

    ch = val->keycode & 0xFFFF;
    if (!(val->keycode & F_FUNC)) {
	if (ch >= 'a' && ch <= 'f') {
	    val->keycode &= ~('a' - 'A');
	    ch &= ~('a' - 'A');
	}
    }
    if ((val->keycode & F_FUNC) ||
	(ch >= '0' && ch <= '9') || (ch >= 'A' && ch <= 'F')) {
	return DM_PASS;
    } else {
	return DM_FAIL;
    }
}

Valtrigdata(cbp, dip, when, val)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
DMTEVal *val;
{
    int ch;

    ch = val->keycode & F_KEYMASK;
    if (!(val->keycode & (F_FUNC|F_CTRL|F_ALT))) {
	if (ch >= 'a' && ch <= 'f') {
	    val->keycode &= ~('a' - 'A'); /* Make hex chars uppercase. */
	    ch &= ~('a' - 'A');
	}
	if (ch == 'X') {	/* Force X's to be lowercase. */
	    val->keycode |= ('a' - 'A');
	    ch |= ('a' - 'A');
	}
    }
    if ((val->keycode & (F_FUNC|F_CTRL|F_ALT)) ||
	(ch >= '0' && ch <= '9') || (ch >= 'A' && ch <= 'F') /* Hexers, */
	|| (ch == 'x') || (ch == '@') || (ch == '$') || (ch == '%') /* base, */
	|| (ch == ' ') || (ch == ',') || (ch == '-')	/* separators, */
	|| (ch == '<') || (ch == '>') /* repeat count delimiters, */
	|| (ch == '[') || (ch == ']')) { /* offset delimiters. */
	return DM_PASS;
    } else {
	return DM_FAIL;
    }
}

ValTrigPos(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    tpos = dip->ident;
    return DM_PASS;
}

/*
** Run the data filtering dialog.
*/

ValFilt(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    DisableMenu(FILEMENU);
    DisableMenu(MISCMENU);
    DMDlgDisable(dialog);
    dialog2 = DMNew(Dfilt, windowid, 0);
    if (DMCollect(dialog2) == DMok) {
	DMGet(dialog2, DFILT_datafilt, &datafilt);
	DMGet(dialog2, DFILT_dataexcl, &filterexc);
	DMGet(dialog2, DFILT_filtoffset, filtoffset);
	DMGet(dialog2, DFILT_filtdata, filtdata);
    }
    DMDisposeNoInval(dialog2);
    dialog2 = 0;
    EnableMenu(FILEMENU);
    EnableMenu(MISCMENU);
    DMDlgEnableUpdate(dialog);
    return DM_PASS;
}

Valdatafilt(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    DMRItem *rip;
    int i, j;

    if (DMFindByIdent(cbp, DFILT_dataexcl, &dip, &rip))
	SetCtlValue(rip->id, 0);
}

Valdataexcl(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    DMRItem *rip;

    if (DMFindByIdent(cbp, DFILT_datafilt, &dip, &rip))
	SetCtlValue(rip->id, 0);
}

ValStopwhen(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    if (dip->ident == DTRIG_trigmatch || dip->ident == DTRIG_trigmismatch) {
	DMItemGrpEnable(cbp, 1);
	DMItemGrpEnable(cbp, 3);
    } else {
	DMItemGrpDisable(cbp, 1);
	DMItemGrpDisable(cbp, 3);
    }
}

ValStop(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    return stopsniff();
}

stopsniff()
{
    if (sniffchild) {		/* Stop the child prematurely. */
	kill(sniffchild, 2);	/* SIGINT. */
	sniffing = 0;
	return DM_PASS;
    } else {			/* Stop that! */
	return DM_FAIL;
    }
}

ValView(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    return viewfile();
}

/*
** A change has occurred in the capture address filtering.
*/

ValFltrAddr(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    ControlID cid;
    int i, enabled;

    enabled = 0;
    if (DMGet(dialog, DMAIN_Source, &i), i) {
	enabled++;
	DMItemEnable(dialog, DMAIN_SrcFilterAddr);
	DMItemEnable(dialog, DMAIN_PickSrc);
	if (dip == &Dmain[DMAIN_Source])
	    DMTEActivate(dialog, DMAIN_SrcFilterAddr);
    } else {
	DMItemDisable(dialog, DMAIN_SrcFilterAddr);
	DMItemDisable(dialog, DMAIN_PickSrc);
    }
    if (DMGet(dialog, DMAIN_Destination, &i), i) {
	enabled++;
	DMItemEnable(dialog, DMAIN_DstFilterAddr);
	DMItemEnable(dialog, DMAIN_PickDst);
	if (dip == &Dmain[DMAIN_Destination])
	    DMTEActivate(dialog, DMAIN_DstFilterAddr);
    } else {
	DMItemDisable(dialog, DMAIN_DstFilterAddr);
	DMItemDisable(dialog, DMAIN_PickDst);
    }
    if (enabled == 2) {
	DMItemEnable(dialog, DMAIN_Conversation);
    } else {			/* Turning off conversational mode. */
	cid = DMGetId(dialog, DMAIN_Conversation);
	SetCtlValue(cid, 0);
	DMItemDisable(dialog, DMAIN_Conversation);
    }
    return DM_PASS;
}

/*
** A change has occurred in the capture protocol filter settings.
*/

ValFltrProt(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    ControlID cid;
    int i;

    if (DMGet(dialog, DMAIN_HexProt, &i), i) {
	DMItemEnable(dialog, DMAIN_HexProtTE);
	if (dip == &Dmain[DMAIN_HexProt])
	    DMTEActivate(dialog, DMAIN_HexProtTE);
    } else {
	DMItemDisable(dialog, DMAIN_HexProtTE);
    }
    return DM_PASS;
}

/*
** A change has occurred in the protocol display disassembler settings.
*/

ValDispProt(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    int i, j;

    if (dip == &Dmain[DMAIN_DAllProt]) {
	if (DMGet(dialog, DMAIN_DAllProt, &i), i) {
	    SetCtlValue((ControlID) DMGetId(dialog, DMAIN_DSafeProt), 0);
	    SetCtlValue((ControlID) DMGetId(dialog, DMAIN_DBootProt), 0);
	    SetCtlValue((ControlID) DMGetId(dialog, DMAIN_DDnetProt), 0);
	    SetCtlValue((ControlID) DMGetId(dialog, DMAIN_DDmapProt), 0);
	    SetCtlValue((ControlID) DMGetId(dialog, DMAIN_DIpProt), 0);
	    SetCtlValue((ControlID) DMGetId(dialog, DMAIN_DArpProt), 0);
	    SetCtlValue((ControlID) DMGetId(dialog, DMAIN_DAtalkProt), 0);
	    SetCtlValue((ControlID) DMGetId(dialog, DMAIN_DAarpProt), 0);
	    SetCtlValue((ControlID) DMGetId(dialog, DMAIN_DHexProt), 0);
	    DMItemDisable(dialog, DMAIN_HexProtTE);
	}
    } else {
	j = 0;
	DMGet(dialog, DMAIN_DSafeProt, &i); j |= i;
	DMGet(dialog, DMAIN_DBootProt, &i); j |= i;
	DMGet(dialog, DMAIN_DDnetProt, &i); j |= i;
	DMGet(dialog, DMAIN_DDmapProt, &i); j |= i;
	DMGet(dialog, DMAIN_DIpProt, &i); j |= i;
	DMGet(dialog, DMAIN_DArpProt, &i); j |= i;
	DMGet(dialog, DMAIN_DAtalkProt, &i); j |= i;
	DMGet(dialog, DMAIN_DAarpProt, &i); j |= i;
	DMGet(dialog, DMAIN_DHexProt, &i); j |= i;
	if (j)
	    SetCtlValue((ControlID) DMGetId(dialog, DMAIN_DAllProt), 0);
	if (DMGet(dialog, DMAIN_DHexProt, &i), i) {
	    DMItemEnable(dialog, DMAIN_HexProtTE);
	    if (dip == &Dmain[DMAIN_DHexProt])
		DMTEActivate(dialog, DMAIN_HexProtTE);
	} else {
	    DMItemDisable(dialog, DMAIN_HexProtTE);
	}
    }
    return DM_PASS;
}

/*
** Switch dialog from capture to view mode.
*/

ValMode(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    ControlID cid;
    int i;

    if (DMGet(dialog, DMAIN_CapMode, &i), i) { /* Capture settings. */
	EnableItem(MISCMENU, MISCHDWR);
	DMCodeHide(dialog, DSHOWS, DCAPS|DSHOWS);
	DMCodeShow(dialog, DCAPS, DLEVEL1MASK|DCAPS|DSHOWS);
	switch (level1) {
	  case L1LAN:
	    DMCodeShow(dialog, DLAN|DCAPS, DLEVEL1MASK|DCAPS|DSHOWS);
	    break;

	  case L1WAN:
	    DMCodeShow(dialog, DWAN|DCAPS, DLEVEL1MASK|DCAPS|DSHOWS);
	    break;

	  case L1WINCOM:
	    DMCodeShow(dialog, DWINCOM|DCAPS, DLEVEL1MASK|DCAPS|DSHOWS);
	    break;
	}
	DMGet(dialog, DMAIN_HexProtTE, disphexprot);
	DMSetText(dialog, DMAIN_HexProtTE, fltrhexprot);
	DMGet(dialog, DMAIN_SrcFilterAddr, dispsrcaddr);
	DMSetText(dialog, DMAIN_SrcFilterAddr, fltrsrcaddr);
	DMGet(dialog, DMAIN_DstFilterAddr, dispdstaddr);
	DMSetText(dialog, DMAIN_DstFilterAddr, fltrdstaddr);
	DMGet(dialog, DMAIN_Source, &dispsrc);
	SetCtlValue((ControlID) DMGetId(dialog, DMAIN_Source), fltrsrc);
	DMGet(dialog, DMAIN_Destination, &dispdst);
	SetCtlValue((ControlID) DMGetId(dialog, DMAIN_Destination), fltrdst);
	DMGet(dialog, DMAIN_Conversation, &dispconv);
	SetCtlValue((ControlID) DMGetId(dialog, DMAIN_Conversation), fltrconv);
	ValFltrProt(dialog, 0, 0, 0, 0);
    } else {			/* Display settings. */
	DisableItem(MISCMENU, MISCHDWR);
	DMCodeHide(dialog, DCAPS, DCAPS|DSHOWS);
	DMCodeShow(dialog, DSHOWS, DCAPS|DSHOWS);
	DMGet(dialog, DMAIN_HexProtTE, fltrhexprot);
	DMSetText(dialog, DMAIN_HexProtTE, disphexprot);
	DMGet(dialog, DMAIN_SrcFilterAddr, fltrsrcaddr);
	DMSetText(dialog, DMAIN_SrcFilterAddr, dispsrcaddr);
	DMGet(dialog, DMAIN_DstFilterAddr, fltrdstaddr);
	DMSetText(dialog, DMAIN_DstFilterAddr, dispdstaddr);
	DMGet(dialog, DMAIN_Source, &fltrsrc);
	SetCtlValue((ControlID) DMGetId(dialog, DMAIN_Source), dispsrc);
	DMGet(dialog, DMAIN_Destination, &fltrdst);
	SetCtlValue((ControlID) DMGetId(dialog, DMAIN_Destination), dispdst);
	DMGet(dialog, DMAIN_Conversation, &fltrconv);
	SetCtlValue((ControlID) DMGetId(dialog, DMAIN_Conversation), dispconv);
	if (DMGet(dialog, DMAIN_DHexProt, &i), i) {
	    DMItemEnable(dialog, DMAIN_HexProtTE);
	} else {
	    DMItemDisable(dialog, DMAIN_HexProtTE);
	}
    }
    ValFltrAddr(dialog, 0, 0, 0, 0);
    return DM_PASS;
}

/*
** Run the address picker dialog.
*/

ValPick(cbp, dip, when, value, partcode)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
short value;
short partcode;
{
    short item;

    DisableMenu(FILEMENU);
    DMDlgDisable(dialog);
    picklabel = (dip->ident == DMAIN_PickSrc)
	? "Source Address" : "Destination Address";
    dialog2 = DMNew(Dpick, windowid, 0);
    DMKeys2Pick(dialog2, DPICK_picker);
    if (DMCollect(dialog2) == DMok) {
	DMGet(dialog2, DPICK_picker, &item);
	DMSetText(dialog, (dip->ident == DMAIN_PickSrc)
		  ? DMAIN_SrcFilterAddr : DMAIN_DstFilterAddr,
		  pickarray.arrayhead[item]);
    }
    DMDisposeNoInval(dialog2);
    dialog2 = 0;
    EnableMenu(FILEMENU);
    DMDlgEnableUpdate(dialog);
    return DM_PASS;
}

ValDex(cbp, dip, when, val)
DMDcb *cbp;
DMItem *dip;
unsigned char when;
DMTEVal *val;
{
    return DMValNumeric(cbp, dip, when, val->text, 4, 0, 8192);
}

clearstatus()
{
    setstatus(nomsg);
}

setstatus(s)
char *s;
{
    DMSetText(dialog, DMAIN_status, s);
}

/*
** Set the Font/Face/Size & drawing mode for the progress indicator.
*/

static void
ffs(mode)
int mode;
{
    if (theFont != 3)
	TextFont(3);
    if (theFace != normalText)
	TextFace(normalText);
    if (theSize != 9)
	TextSize(9);
    if (theTextMode != mode)
	TextMode(mode);
}

/*
** Draw the animated capturing progress indicator.  Tries very hard to
** avoid drawing anything extra, for efficiency.
*/

drawprogress(labels)
int labels;			/* Draw labels too. */
{
    int ticks, seen, left, bot;
    static short rect1, rect2;
    static int int1, int2, int3, int4, int5;
    Rect r, r1;
    char buf[256], *mysprintf();

    if (shmp && !iconified) {
	r = Dmain[DMAIN_status].rect;
	r.edge.bottom--;
	left = r.edge.left;
	r.edge.left += 50;
	r.edge.top = r.edge.bottom - 10;
	if (labels) {
	    ffs(srcOr);
	    MoveTo(left, r.edge.bottom - 1);
	    DrawText("Captured", 0, 8);
	    FrameRect(&r);	/* Capture buffer percentage full. */
	    rect1 = -1;
	}
	r1 = r;
	InsetRect(&r1, 1, 1);
	if (shmp->capbufsize) {
	    r1.edge.right = r1.edge.left + ScaleRS(shmp->capbytes >> 9,
				    r1.edge.right - r1.edge.left,
				    shmp->capbufsize >> 9);
	    if (r1.edge.right < r1.edge.left)
		r1.edge.right = r1.edge.left;
	    if (rect1 != r1.edge.right) {
		rect1 = r1.edge.right;
		FillRect(&r1, blackPattern);
		r1.edge.left = r1.edge.right;
		r1.edge.right = r.edge.right - 1;
		EraseRect(&r1);
	    }
	}
	OffsetRect(&r, 0, -15);
	if (labels) {
	    ffs(srcOr);
	    MoveTo(left, r.edge.bottom - 1);
	    DrawText("Net B/W", 0, 7);
	    FrameRect(&r);	/* Bandwidth Utilization. */
	    rect2 = -1;
	}
	r1 = r;
	InsetRect(&r1, 1, 1);
	seen = shmp->seenbytes;
	ticks = shmp->curtime;
	if (ticks != lastticks) {
	    r1.edge.right = r1.edge.left
		+ ScaleRS((short) (MulS((seen - lastseen) >> 9,
					duts.clocktics)) /
			  (short) (ticks - lastticks),
			  r1.edge.right - r1.edge.left, 2048);
	    if (r1.edge.right < r1.edge.left)
		r1.edge.right = r1.edge.left;
	    if (rect2 != r1.edge.right) {
		rect2 = r1.edge.right;
		FillRect(&r1, blackPattern);
		r1.edge.left = r1.edge.right;
		r1.edge.right = r.edge.right - 1;
		EraseRect(&r1);
	    }
	}
	lastticks = ticks;
	lastseen = seen;

#define OFFS1 70		/* Distance from "Packs" to digits */
#define OFFS2 28		/* Distance from "Cap" to digits */
#define OFFS3 110		/* Distance from "Packets" to "Cap" */
#define OFFS4 170		/* Distance from "Packets" to "Bad" */
#define OFFS5 70		/* Distance from "Bytes" to digits */
#define OFFS6 130		/* Distance from "Bytes" to "Cap" */

	OffsetRect(&r, 0, -40);
	if (labels) {
	    ffs(srcOr);
	    MoveTo(left, bot = r.edge.bottom - 1);
	    DrawText("Sniffing...", 0, 11);
	}
	OffsetRect(&r, 0, 12);
	bot = r.edge.bottom - 1;
	if (labels) {
	    ffs(srcOr);
	    MoveTo(left, bot);
	    DrawText("Packs Seen:", 0, 11);
	    int1 = -1;
	    MoveTo(left + OFFS3, bot);
	    DrawText("Cap:", 0, 4);
	    int2 = -1;
	    MoveTo(left + OFFS4, bot);
	    DrawText("Bad:", 0, 4);
	    int3 = -1;
	}
	if (int1 != shmp->seenpackets) {
	    int1 = shmp->seenpackets;
	    ffs(srcCopy);
	    mysprintf(buf, "%d", shmp->seenpackets);
	    MoveTo(left + OFFS1, bot);
	    DrawString(buf);
	}
	if (int2 != shmp->cappackets) {
	    int2 = shmp->cappackets;
	    ffs(srcCopy);
	    MoveTo(left + OFFS3 + OFFS2, bot);
	    mysprintf(buf, "%d", shmp->cappackets);
	    DrawString(buf);
	}
	if (int3 != shmp->badpackets) {
	    int3 = shmp->badpackets;
	    ffs(srcCopy);
	    MoveTo(left + OFFS4 + OFFS2, bot);
	    mysprintf(buf, "%d", shmp->badpackets);
	    DrawString(buf);
	}
	OffsetRect(&r, 0, 10);
	bot = r.edge.bottom - 1;
	if (labels) {
	    ffs(srcOr);
	    MoveTo(left, bot);
	    DrawText("Bytes Seen:", 0, 11);
	    int4 = -1;
	    MoveTo(left + OFFS6, bot);
	    DrawText("Cap:", 0, 4);
	    int5 = -1;
	}
	if (int4 != shmp->seenbytes) {
	    int4 = shmp->seenbytes;
	    ffs(srcCopy);
	    mysprintf(buf, "%d", shmp->seenbytes);
	    MoveTo(left + OFFS5, bot);
	    DrawString(buf);
	}
	if (int5 != shmp->capbytes) {
	    int5 = shmp->capbytes;
	    ffs(srcCopy);
	    MoveTo(left + OFFS6 + OFFS2, bot);
	    mysprintf(buf, "%d", shmp->capbytes);
	    DrawString(buf);
	}
	ffs(srcOr);
    }
}

/*
** The sniffer capture process has exited.
*/

void
sniffdone(nwcb)
struct nowaitblock *nwcb;
{
    char scratch[80], *p;
    int i;

    sniffchild = 0;
    sniffing = 0;
    if (shmp) {			/* Destroy shared memory. */
	shmdt(shmp);
	shmctl(shmid, IPC_RMID, 0);
	shmp = 0;
    }
    DMItemDisable(dialog, DMAIN_StopButton);
    EnableItem(FILEMENU, FILEOPEN);
    drawsnort(dialog, &Dmain[DMAIN_Snort], DMbeforeDrawing);
    if (nwcb->param) {		/* Child exited with error? */
	if (!nwcb->arg) {
	    unlink(filename);	/* Trace file is no good. */
	    hasfile = 0;
	    DisableItem(FILEMENU, FILESAVE);
	    DMItemDisable(dialog, DMAIN_ViewButton);
	}
	DMItemEnable(dialog, DMAIN_GoButton);
	if (!(i = (nwcb->param & 0xFF))) {
	    setstatus(snifferrs[nwcb->param >> 8]);
	} else {
	    strcpy(scratch, sniffdied);
	    p = scratch + StringLen(sniffdied);
	    i &= 0x7F;
	    if (i > 9)
		*p++ = i/10 + '0';
	    i %= 10;
	    *p++ = i + '0';
	    *p++ = '!';
	    *p = 0;
	    setstatus(scratch);
	}
	Bell();
    } else if (!nwcb->arg || hasfile) {
	hasfile = 1;
	EnableItem(FILEMENU, FILESAVE);
	DMItemEnable(dialog, DMAIN_ViewButton);
	if (!statfile())
	    DMItemEnable(dialog, DMAIN_GoButton);
    } else {
	DMItemEnable(dialog, DMAIN_GoButton);
    }
    if (dialog2)		/* A dialog's over our buttons, refresh */
	DMUpdate(dialog2);	/* it since we just scribbled on it. */
}

/*
** Append the packet format overrides to the packet file, if any.
** These are placed into a dummy packet at the end of the file, which
** has a special status so that the rest of the program suite will
** ignore it.  Multiple override packets may be appended (if the file
** is saved many times), but only the last one will be used.  The packet
** is formatted in such a way that it can be recognized merely by seeking
** to the end and reading it.
*/

void
appendoverrides()
{
    int fd, i;
    struct overpacket *pp;

    if (!overrides[0])		/* No overrides?  Forget it. */
	return;
    i = sizeof(*pp) + sizeof overrides;
    pp = alloca(i);
    memset(pp, 0, i);
    pp->packsize = pp->snh.length = i - sizeof(pp->packsize);
    pp->snh.statusbits = ~0;	/* Special mark. */
    BlockMove(overrides, &pp[1], sizeof overrides);
    if ((fd = open(filename, O_RDWR | O_APPEND)) >= 0) {
	write(fd, pp, i);
	close(fd);
    }
}

/*
** Slurp the packet format overrides (if any) out of the packet file.
** These are placed into a dummy packet at the end of the file, which
** has a special status so that the rest of the program suite will
** ignore it.  Multiple override packets may have been appended (if the file
** was saved many times), but only the last one will be used.  The packet
** is formatted in such a way that it can be recognized merely by seeking
** to the end and reading it.
*/

void
openoverrides()
{
    int fd, i;
    struct overpacket *pp;

    i = sizeof(*pp) + sizeof overrides;
    pp = alloca(i);
    if ((fd = open(filename, O_RDONLY)) >= 0) {
	if (lseek(fd, -i, 2)) {
	    if (read(fd, pp, i) == i) {
		i -= sizeof(pp->packsize);
		if (pp->packsize == i && pp->snh.length == i
		    && pp->snh.statusbits == 0xFF)
		    BlockMove(&pp[1], overrides, sizeof overrides);
	    }
	}
	close(fd);
    }
}

/*
** The Save file picker process has exited.  Copy the captured file to
** the given name, and add any format overrides to the end of it.
*/

void
savedone(nwcb)
struct nowaitblock *nwcb;
{
    savechild = 0;
    if (!iconified)
	SelectWindow(windowid);
    DMItemEnable(dialog, DMAIN_GoButton);
    if (nwcb->param) {		/* Child exited with error? */
	setstatus(savefailed);
	Bell();
    } else if (sfs.returnedname && hasfile) {
	appendoverrides();	/* Add overrides to file first, if any. */
	unlink(sfs.returnedname);
	if (link(filename, sfs.returnedname)) {
	    if (errno == EXDEV) { /* Cross-device, must do a copy instead. */
		filecopy(filename, sfs.returnedname);
	    } else {		/* Save failed. */
		setstatus(savefailed);
		Bell();
	    }
	}
    }
}

/*
** The Save Settings picker process has exited.  Make a settings file.
*/

void
ssetdone(nwcb)
struct nowaitblock *nwcb;
{
    char *cp, buf[1024], buf2[256];
    int fd, j;

    ssetchild = 0;
    if (!iconified)
	SelectWindow(windowid);
    DMItemEnable(dialog, DMAIN_GoButton);
    if (nwcb->param) {		/* Child exited with error? */
	setstatus(savefailed);
	Bell();
    } else if (sflss.returnedname) {
	if ((fd = open(sflss.returnedname, O_RDWR|O_NOCTTY|O_TRUNC|O_CREAT,
		       0700)) >= 0) {
	    cp = mysprintf(buf, ":\nexec ./%s", progname);
	    if (animated)
		cp = mysprintf(cp, " -a");
	    DMGet(dialog, DMAIN_BadPacks, &j);
	    if (j)
		cp = mysprintf(cp, " -b");
	    DMGet(dialog, DMAIN_Conversation, &j);
	    if (j)
		cp = mysprintf(cp, " -c");
	    DMGet(dialog, DMAIN_DstFilterAddr, buf2);
	    DMGet(dialog, DMAIN_Destination, &j);
	    if (j)
		cp = mysprintf(cp, " -D %s", buf2);
	    if (nonose)
		cp = mysprintf(cp, " -nose");
	    DMGet(dialog, DMAIN_bufsize, buf2);
	    cp = mysprintf(cp, " -n %s", buf2);
	    DMGet(dialog, DMAIN_PromiscuousPacks, &j);
	    if (j)
		cp = mysprintf(cp, " -p");
	    DMGet(dialog, DMAIN_HexProtTE, buf2);
	    DMGet(dialog, DMAIN_NoProt, &j);
	    if (!j)
		cp = mysprintf(cp, " -P %s", protname());
	    DMGet(dialog, DMAIN_SrcFilterAddr, buf2);
	    DMGet(dialog, DMAIN_Source, &j);
	    if (j)
		cp = mysprintf(cp, " -S %s", buf2);
	    if (sound)
		cp = mysprintf(cp, " -s");
	    write(fd, buf, cp - buf);
	    cp = buf;
	    if (nostop)
		cp = mysprintf(cp, " -ts");
	    if (trigdata[0])
		cp = mysprintf(cp, " -td '%s'", trigdata);
	    if (trigoffset[0])
		cp = mysprintf(cp, " -to %s", trigoffset);
	    if (trigger)
		cp = mysprintf(cp, " -tr");
	    if (triggerinv)
		cp = mysprintf(cp, " -ti");
	    cp = mysprintf(cp, " -tp %d", tpos);
	    if (filtdata[0])
		cp = mysprintf(cp, " -fd '%s'", filtdata);
	    if (filtoffset[0])
		cp = mysprintf(cp, " -fo %s", filtoffset);
	    if (filterexc)
		cp = mysprintf(cp, " -fx");
	    if (datafilt)
		cp = mysprintf(cp, " -fi");
	    if (overrides[0])
		cp = mysprintf(cp, " -ov '%s'", overrides);
	    if (datalimit)
		cp = mysprintf(cp, " -hd");
	    if (limitsize[0])
		cp = mysprintf(cp, " -hl '%s'", limitsize);
	    cp = mysprintf(cp, " -hg %d", histogram);
	    DMGet(dialog, DMAIN_DispSymbolic, &j);
	    if (j)
		cp = mysprintf(cp, " -vs");
	    DMGet(dialog, DMAIN_DeltaTime, &j);
	    if (j)
		cp = mysprintf(cp, " -vt");
	    DMGet(dialog, DMAIN_DataAftrHdr, &j);
	    if (j)
		cp = mysprintf(cp, " -vd");
	    DMGet(dialog, DMAIN_DispRawPacks, &j);
	    if (j)
		cp = mysprintf(cp, " -vr");
	    DMGet(dialog, DMAIN_AllHeaders, &j);
	    if (j)
		cp = mysprintf(cp, " -va");
	    if (level1)
		cp = mysprintf(cp, level1 > L1WAN ? " -win" : " -wan");
	    cp = mysprintf(cp, "\n");
	    write(fd, buf, cp - buf);
	    close(fd);
	} else {		/* Save failed. */
	    setstatus(savefailed);
	    Bell();
	}
    }
}

/*
** The Load Settings picker process has exited.
*/

void
lsetdone(nwcb)
struct nowaitblock *nwcb;
{
    lsetchild = 0;
    if (!iconified)
	SelectWindow(windowid);
    DMItemEnable(dialog, DMAIN_GoButton);
    if (nwcb->param) {		/* Child exited with error? */
	setstatus(openfailed);
	Bell();
    } else if (sflss.returnedname) {
	ReleaseEvents();
    }
}

void
opendone(nwcb)
struct nowaitblock *nwcb;
{
    openchild = 0;
    if (!iconified)
	SelectWindow(windowid);
    DMItemEnable(dialog, DMAIN_GoButton);
    if (nwcb->param) {		/* Child exited with error? */
	setstatus(openfailed);
	Bell();
    } else if (sfo.returnedname) {
	overrides[0] = 0;	/* Clear the packet format overrides. */
	if (hasfile > 0)
	    unlink(filename);
	hasfile = -1;
	EnableItem(FILEMENU, FILESAVE);
	DMItemEnable(dialog, DMAIN_ViewButton);
	strncpy(filename, sfo.returnedname, sizeof filename);
	openoverrides();
	statfile();
    }
}

/*
** Start a child process to copy a file.
*/

filecopy(from, to)
char *from, *to;
{
    int i;

    if (!(copychild = fork())) {
	if (shmp)
	    shmdt(shmp);
	for (i=0; i<_NFILE; i++) /* Child. */
	    close(i);
	open(nulldev, O_RDONLY); /* Stdin for the copier. */
	dup(0);			/* Stdout for the copier. */
	dup(0);			/* Stderr for the copier. */
	execl(copier, copier, "-m", from, to, 0);
	_exit(1);
    } else if (copychild > 0) { /* Parent? */
	DisableItem(FILEMENU, FILESAVE);
	waitforchild();
    } else {			/* Fork failed. */
	copychild = 0;
    }
}

/*
** The file copying process has exited.
*/

void
copydone(nwcb)
struct nowaitblock *nwcb;
{
    copychild = 0;
    if (nwcb->param) {		/* Child exited with error? */
	setstatus(savefailed);
	Bell();
    }
    if (hasfile)
	EnableItem(FILEMENU, FILESAVE);
}

/*
** Start a child process to view a file.  Actually starts two processes,
** a viewer, and a dumper that is piped to it.
*/

viewfile()
{
    char *args[40];		/* execv()'s argument list. */
    char *dm;			/* Data dump mode. */
    char *saddr;		/* Source viewing filter. */
    char *daddr;		/* Destination viewing filter. */
    char scratch[80];
    int i, j, datamode, pipes[2];

    if (!hasfile)
	return DM_FAIL;
    if (DMGet(dialog, DMAIN_ViewMode, &i), i) {
	DMGet(dialog, DMAIN_HexProtTE, disphexprot);
	DMGet(dialog, DMAIN_SrcFilterAddr, dispsrcaddr);
	DMGet(dialog, DMAIN_DstFilterAddr, dispdstaddr);
	DMGet(dialog, DMAIN_Source, &dispsrc);
	DMGet(dialog, DMAIN_Destination, &dispdst);
	DMGet(dialog, DMAIN_Conversation, &dispconv);
    }
    i = 0;
    args[i++] = snifferdump;
    args[i++] = filename;
    args[i++] = "-m";
    if (level1)			/* Don't whine about short packets from */
	args[i++] = "-s";	/* non-Ethernet interfaces. */
    if (datalimit) {
	args[i++] = "-dl";
	args[i++] = limitsize;
    }
    if (DMGet(dialog, DMAIN_DispSymbolic, &j), j) {
	args[i++] = "-f";
	args[i++] = addresses;
    }
    if (DMGet(dialog, DMAIN_DeltaTime, &j), j)
	args[i++] = "-d";
    if (DMGet(dialog, DMAIN_DispRawPacks, &j), j)
	args[i++] = "-r";
    if (DMGet(dialog, DMAIN_AllHeaders, &j), j)
	args[i++] = "-h";
    DMGet(dialog, DMAIN_DataAftrHdr, &datamode);
    dm = datamode ? hm2 : hm1;
    if (DMGet(dialog, DMAIN_DAllProt, &j), j) {
	if (datamode)
	    args[i++] = "-A";
	else
	    args[i++] = "-a";
    } else {
	if (DMGet(dialog, DMAIN_DSafeProt, &j), j) {
	    args[i++] = dm;
	    args[i++] = "safe";
	}
	if (DMGet(dialog, DMAIN_DBootProt, &j), j) {
	    args[i++] = dm;
	    args[i++] = "boot";
	}
	if (DMGet(dialog, DMAIN_DDnetProt, &j), j) {
	    args[i++] = dm;
	    args[i++] = "dnet";
	}
	if (DMGet(dialog, DMAIN_DDmapProt, &j), j) {
	    args[i++] = dm;
	    args[i++] = "dmap";
	}
	if (DMGet(dialog, DMAIN_DIpProt, &j), j) {
	    args[i++] = dm;
	    args[i++] = "ip";
	}
	if (DMGet(dialog, DMAIN_DArpProt, &j), j) {
	    args[i++] = dm;
	    args[i++] = "arp";
	}
	if (DMGet(dialog, DMAIN_DAtalkProt, &j), j) {
	    args[i++] = dm;
	    args[i++] = "atalk";
	}
	if (DMGet(dialog, DMAIN_DAarpProt, &j), j) {
	    args[i++] = dm;
	    args[i++] = "aarp";
	}
	if (DMGet(dialog, DMAIN_DHexProt, &j), j) {
	    args[i++] = dm;
	    args[i++] = disphexprot;
	}
    }
    if (dispconv)
	args[i++] = "-c";
    if (trigat[0]) {
	args[i++] = "-t";
	args[i++] = trigat;
    }
				/* Must be last!  (Due to variable length.) */
    if (dispsrc) {
	if (saddr = alloca(StringLen(dispsrcaddr)+1)) {
	    strcpy(saddr, dispsrcaddr);
	    while ((saddr = strtok(saddr, "\n\r\t "))
		   && i < sizeof(args)/sizeof(args[0])-3) {
		if (daddr = alloca(24)) {
		    strncpy(daddr, saddr, 24);
		    if (procaddr(daddr)) {
			setstatus("Source viewing filter name not found!");
			return DM_FAIL;
		    }
		}
		args[i++] = "-S";
		args[i++] = daddr;
		saddr = 0;
	    }
	}
    }
    if (dispdst) {
	if (daddr = alloca(StringLen(dispdstaddr)+1)) {
	    strcpy(daddr, dispdstaddr);
	    while ((daddr = strtok(daddr, "\n\r\t "))
		   && i < sizeof(args)/sizeof(args[0])-3) {
		if (saddr = alloca(24)) {
		    strncpy(saddr, daddr, 24);
		    if (procaddr(saddr)) {
			setstatus("Destination viewing filter name not found!");
			return DM_FAIL;
		    }
		}
		args[i++] = "-D";
		args[i++] = saddr;
		daddr = 0;
	    }
	}
    }
    if (overrides[0] && (dm = alloca(StringLen(overrides)+1))) {
	strcpy(dm, overrides);
	while ((dm = strtok(dm, "\n\r\t "))
	       && i < sizeof(args)/sizeof(args[0])-3) {
	    args[i++] = "-C";
	    args[i++] = dm;
	    dm = 0;
	}
    }
    args[i] = 0;		/* End of argument list. */

    if (pipe(pipes))
	return DM_FAIL;
    if (!(viewchild = fork())) {
	if (shmp)
	    shmdt(shmp);
	for (i=0; i<_NFILE; i++) /* Child (the viewer). */
	    if (i != pipes[0])
		close(i);
	if (pipes[0]) {
	    dup(pipes[0]);
	    close(pipes[0]);
	}
	strcpy(scratch, "Network Trace : ");
	strncat(scratch, nodename, sizeof scratch);
	execl(viewer, viewer, "-F", "-T", scratch, "-c", "85", 0);
	_exit(3);
    } else if (viewchild > 0) {
	waitforchild();		/* Parent (the sniffer). */
	if (!(i = fork())) {
	    for (i=0; i<_NFILE; i++) /* Child (the dumper). */
		if (i != pipes[1])
		    close(i);
	    if (pipes[1] != 1) {
		dup(dup(pipes[1]));
		close(0);
		close(pipes[1]);
	    }
	    execv(snifferdump, args);
	    _exit(2);
	} else if (i > 0) {
	    waitforchild();	/* Parent (the sniffer). */
	    close(pipes[0]);
	    close(pipes[1]);
	    return DM_PASS;
	}
    } else {			/* Fork failed. */
	viewchild = 0;
    }
    close(pipes[0]);
    close(pipes[1]);
    return DM_FAIL;
}

/*
** The file viewing process has exited.
*/

void
viewdone(nwcb)
struct nowaitblock *nwcb;
{
    viewchild = 0;
    if (nwcb->param) {		/* Child exited with error? */
	Bell();
    }
    if (!iconified)
	SelectWindow(windowid);
}

/*
** Start a child process to calculate stats about a capture file.
*/

int sinpipes[2], soutpipes[2];
char stats[128];

statfile()
{
    if (statchild)
	return 0;
    if (pipe(sinpipes))
	return 0;
    if (pipe(soutpipes)) {
	close(sinpipes[0]);
	close(sinpipes[1]);
	return 0;
    }
    if (!(statchild = fork())) {
	if (shmp)
	    shmdt(shmp);
	close(sinpipes[1]);	/* (Child) */
	close(soutpipes[0]);
	if (sinpipes[0] != 0) {
	    close(0);
	    dup(sinpipes[0]);	/* Child's stdin. */
	    close(sinpipes[0]);
	}
	if (soutpipes[1] != 1) {
	    close(1);
	    dup(soutpipes[1]);	/* Child's stdout. */
	    close(soutpipes[1]);
	}
	execl(statprog, statprog, "-b", filename, level1 ? "-s" : 0, 0);
	_exit(1);
    } else if (statchild > 0) {	/* (Parent) */
	close(sinpipes[0]);
	close(soutpipes[1]);
	nowaitread.requestnum = rdnix(F_NOWAIT | F_READ, soutpipes[0], stats,
				     sizeof stats, 0, 0, trapqfd, &nowaitread);
	waitforchild();
	setstatus("Calculating statistics...");
	return 1;
    } else {			/* Fork failed, no child. */
	close(sinpipes[0]);
	close(sinpipes[1]);
	close(soutpipes[0]);
	close(soutpipes[1]);
	statchild = 0;
	return 0;
    }
}

/*
** The file statistics process has exited.  Enable another capture.
*/

void
statdone(nwcb)
struct nowaitblock *nwcb;
{
    statchild = 0;
    if (nwcb->param) {
	setstatus(badcapture);
    }
    DMItemEnable(dialog, DMAIN_GoButton);
    if (dialog2)		/* A dialog's over our buttons, refresh */
	DMUpdate(dialog2);	/* it since we just scribbled on it. */
}

void
readdone(nwcb, rpar, reqnum)
struct nowaitblock *nwcb;
int rpar, reqnum;
{
    char *cp, *cp2;

    nwcb->requestnum = 0;
    trigat[0] = 0;
    if (rpar > 0) {            /* Successful completion of read. */
	stats[rpar] = 0;
	setstatus(stats);
	if (cp = FastFind(stats, " packets before ", rpar, 16)) {
	    cp2 = cp;
	    while (cp2 > stats && *--cp2 != ' ')
		;
	    if (cp2 > stats) {
		cp2++;
		strncpy(trigat, cp2, cp - cp2);
		trigat[cp - cp2] = 0;
	    }
	}
    }
    close(sinpipes[1]);
    close(soutpipes[0]);
}

/*
** Start a child process to view help.  Actually starts two processes,
** a viewer, and an nroff that is piped to it.
*/

runhelp(file)
char *file;
{
    char *args[4];		/* execv()'s argument list. */
    char scratch[80];
    int i, pipes[2];

    i = 0;
    if (access(nroff, X_OK)) {	/* Nroff missing? */
	args[i++] = cat;	/* Use cat instead. */
    } else {
	args[i++] = nroff;
	args[i++] = "-man";
    }
    args[i++] = file;
    args[i] = 0;		/* End of argument list. */

    if (pipe(pipes))
	return DM_FAIL;
    if (!(helpchild = fork())) {
	if (shmp)
	    shmdt(shmp);
	for (i=0; i<_NFILE; i++) /* Child (the viewer). */
	    if (i != pipes[0])
		close(i);
	if (pipes[0]) {
	    dup(pipes[0]);
	    close(pipes[0]);
	}
	strcpy(scratch, "Sniffer Help : ");
	strncat(scratch, nodename, sizeof scratch);
	execl(viewer, viewer, "-F", "-T", scratch, 0);
	_exit(3);
    } else if (helpchild > 0) {
	waitforchild();		/* Parent (the sniffer). */
	if (!(i = fork())) {
	    for (i=0; i<_NFILE; i++) /* Child (nroff). */
		if (i != pipes[1])
		    close(i);
	    if (pipes[1] != 1) {
		dup(dup(pipes[1]));
		close(0);
		close(pipes[1]);
	    }
	    nice(4);		/* Nroff is a pig.  Be nice. */
	    execv(args[0], args);
	    _exit(2);
	} else if (i > 0) {
	    waitforchild();	/* Parent (the sniffer). */
	    close(pipes[0]);
	    close(pipes[1]);
	    return DM_PASS;
	}
    } else {			/* Fork failed. */
	helpchild = 0;
    }
    close(pipes[0]);
    close(pipes[1]);
    return DM_FAIL;
}

/*
** The help viewing process has exited.
*/

void
helpdone(nwcb)
struct nowaitblock *nwcb;
{
    helpchild = 0;
    
    if (nwcb->param) {		/* Child exited with error? */
	Bell();
    }
    if (!iconified)
	SelectWindow(windowid);
}

/*
** Start a child process to view a histogram.  Actually starts two processes,
** a viewer, and a packstats that is piped to it.  (Three if a fresh capture
** is called for.)
*/

runhistogram(types)
int types;
{
    char *args[9];		/* execv()'s argument list. */
    char scratch[80];
    int i, j, fresh, pipes[2], pipes2[2];

    i = 0;
    DMGet(dialog2, DHIST_fresh, &fresh);
    if (fresh && sniffchild)
	return DM_FAIL;
    args[i++] = statprog;
    if (types & HIST_SIZE)
	args[i++] = "-BS";
    if (types & HIST_SADR)
	args[i++] = "-BA";
    if (types & HIST_DADR)
	args[i++] = "-BD";
    if (types & HIST_PROT)
	args[i++] = "-BP";
    if (DMGet(dialog, DMAIN_DispSymbolic, &j), j) {
	args[i++] = "-f";
	args[i++] = addresses;
    }
    args[i++] = filename;	/* Must be last! */
    args[i] = 0;		/* End of argument list. */

    if (pipe(pipes))
	return DM_FAIL;
    if (!(histchild = fork())) {
	if (shmp)
	    shmdt(shmp);
	for (i=0; i<_NFILE; i++) /* Child (the viewer). */
	    if (i != pipes[0])
		close(i);
	if (pipes[0]) {
	    dup(pipes[0]);
	    close(pipes[0]);
	}
	strcpy(scratch, "Packet Histogram");
	if (histogram & (histogram - 1)) /* More than one histogram? */
	    strcat(scratch, "s");
	strcat(scratch, " : ");
	strncat(scratch, nodename, sizeof scratch);
	execl(viewer, viewer, "-F", "-T", scratch, 0);
	_exit(3);
    } else if (histchild > 0) {
	waitforchild();		/* Parent (the sniffer). */
	if (fresh) {
	    if (pipe(pipes2)) {
		close(pipes[0]);
		close(pipes[1]);
		return DM_FAIL;
	    }
	    startsniffnet(pipes2[1]); /* Start a sniffnet child, to a pipe. */
	    close(pipes2[1]);
	}
	if (!(i = fork())) {
	    if (shmp)
		shmdt(shmp);
	    for (i=0; i<_NFILE; i++) { /* Child (packstats). */
		if (i == pipes[1])
		    continue;
		if (fresh && i == pipes2[0])
		    continue;
		close(i);
	    }
	    if (fresh) {
		if (pipes2[0] != 0) {
		    dup(pipes2[0]); close(pipes2[0]);
		}
		if (pipes[1] != 1) {
		    dup(pipes[1]);
		    close(pipes[1]);
		}
		for (i=0; i<sizeof(args) / sizeof(args[0]); i++) {
		    if (!args[i]) {
			args[i-1] = 0; /* Cancel filename. */
			break;
		    }
		}
	    } else {
		if (pipes[1] != 1) {
		    dup(dup(pipes[1]));
		    close(0);
		    close(pipes[1]);
		}
		nice(4);	/* Be nice. */
	    }
	    execv(args[0], args);
	    _exit(2);
	} else if (i > 0) {
	    waitforchild();	/* Parent (the sniffer). */
	    close(pipes[0]);
	    close(pipes[1]);
	    if (fresh)
		close(pipes2[0]);
	    return DM_PASS;
	}
    } else {			/* Fork failed. */
	histchild = 0;
    }
    close(pipes[0]);
    close(pipes[1]);
    return DM_FAIL;
}

/*
** The histogram viewing process has exited.
*/

void
histdone(nwcb)
struct nowaitblock *nwcb;
{
    histchild = 0;
    
    if (nwcb->param) {		/* Child exited with error? */
	Bell();
    }
    if (!iconified)
	SelectWindow(windowid);
}

int ishex(register unsigned char ch);

ishex(ch)
unsigned char ch;
{
    if ((ch >= '0' && ch <= '9') || (ch >= 'a' && ch <= 'f')
	|| (ch >= 'A' && ch <= 'F') || (ch == '-'))
	return 1;
    return 0;
}

/*
** Convert a node name to a hex address, using the addresses file.
** If it fails, try it raw as a hex address;
*/

procaddr(name)
char *name;
{
    int i, len;

    if (!pickarray.arrayhead)	/* Make sure we have address list. */
	PFpicker(0, 0, 0);
    for (i=0; i<pickarray.length; i++) {
	len = StringLen(pickarray.arrayhead[i]);
	if (!strcmp(pickarray.arrayhead[i], name)) {
	    strcpy(name, &pickarray.arrayhead[i][len+1]);
	    return 0;
	}
    }
    len = StringLen(name);	/* No name match, is it all hex? */
    for (i=0; i<len; i++)
	if (!ishex(name[i]))
	    return 1;		/* No, is garbage then. */
    return 0;
}

/*
** Child process control subroutines.
*/

void
waitdone(nwcb, rpar, reqnum)
struct nowaitblock *nwcb;
int rpar, reqnum;
{
    nwcb->requestnum = 0;
    if (rpar >= 0) {            /* Successful completion of wait. */
	if (rpar == sniffchild) { /* What child is this... */
	    sniffdone(nwcb);
	} else if (rpar == statchild) {
	    statdone(nwcb);
	} else if (rpar == openchild) {
	    opendone(nwcb);
	} else if (rpar == savechild) {
	    savedone(nwcb);
	} else if (rpar == lsetchild) {
	    lsetdone(nwcb);
	} else if (rpar == ssetchild) {
	    ssetdone(nwcb);
	} else if (rpar == copychild) {
	    copydone(nwcb);
	} else if (rpar == viewchild) {
	    viewdone(nwcb);
	} else if (rpar == helpchild) {
	    helpdone(nwcb);
	} else if (rpar == histchild) {
	    histdone(nwcb);
	} else if (rpar == editchild) {
	    editchild = 0;
	    EnableItem(EDITMENU, EDITADDR);
	    if (pickarray.arrayhead) {
		free(pickarray.arrayhead);
		pickarray.arrayhead = 0;
	    }
	}
	if (--children > 0)
	    postnowaitwait();
    } else {
				/* Do nothing, responding to the CANCEL. */
    }
}

/*
** Add a child process to the no-wait wait queue.
*/

void
waitforchild()
{
    if (++children == 1)
	postnowaitwait();
}

void
postnowaitwait()
{
    if (children && !nowaitwait.requestnum)
	nowaitwait.requestnum = dnix(F_NOWAIT | F_WAIT, 0, &nowaitwait.param,
				     sizeof(nowaitwait.param), 0, 0,
				     trapqfd, &nowaitwait);
}

void
killnowaitwait()
{
    int i;

    if (nowaitwait.requestnum) {
	i = dnix(F_CANCEL, nowaitwait.requestnum);
	nowaitwait.requestnum = 0;
    }
}

void
napdone(nwcb, rpar, reqnum)
struct nowaitblock *nwcb;
int rpar, reqnum;
{
    nwcb->requestnum = 0;
    if (rpar >= 0) {            /* Successful completion of nap. */
	if (animated) {
	    if (!dialog2) {	/* No secondary dialog up? */
		drawprogress(0);
	    }
	    nwcb->requestnum = dnix(F_NOWAIT | F_NAP, nwcb->param, 0, 0, 0, 0,
				    trapqfd, nwcb);
	}
    } else {
				/* Do nothing, responding to the CANCEL. */
    }
}

/*
** Sprintf replacement.
**
** Handles format strings containing:
**
**   %[1..9]s, %[0][1..9]d, %[0][1..9]x, %[0][1..9]X
**
** It's much smaller than the standard one.  Returns pointer to the NULL.
*/

char *
mysprintf(buffer, fmt, arglist)
char *buffer, *fmt;
char *arglist;
{
    register char *p, *q, *p2;
    short ndigits;
    int leadfill;
    char *itoa();
    union arg {
	long lng;
	char *ptr;
    } *argptr;

    argptr = (union arg *) &arglist;
    p = fmt;
    q = buffer;
    while (*p) {            /* While there's some format string. */
	if (*p == '%') {    /* Is it a format control? */
	    p++;            /* Yes, which one (skip over % char)? */
	    if (*p == '0')
		leadfill = *p++;
	    else
		leadfill = ' ';
	    if (*p >= '1' && *p <= '9')	/* Format output length? */
		ndigits = *p++ - '0';
	    else
		ndigits = 0;
	    if (*p == 's') {
		p2 = argptr++->ptr; /* String. */
	    } else if (*p == 'd') {
		p2 = itoa(argptr++->lng, 10, leadfill, ndigits);
	    } else if (*p == 'x' || *p == 'X') {
		p2 = itoa(argptr++->lng, 16, leadfill, ndigits);
	    }
	    p++;		/* Skip over format control. */
	    if (ndigits)
		while (*p2 && ndigits--) /* Copy the inclusion. */
		    *q++ = *p2++;
	    else
		while (*p2)	/* Copy the inclusion. */
		    *q++ = *p2++;
	} else			/* No, it's a literal -- copy it. */
	    *q++ = *p++;
    }
    *q = 0;			/* Null-terminate it. */
    return q;
}

char *
itoa(val, radix, leadfill, ndigits)
int val, radix, leadfill, ndigits;
{
    short i;
    char *p, *p2, out;
    static char buf[20];
    unsigned long test, remainder;

    p = buf;
    remainder = val;
    i = ndigits;		/* Make sure there're at least the number */
    while (i--)			/* of leading whatevers that we've asked for */
	*p++ = leadfill;	/* as the field width. */
    p2 = p;
    if (radix == 16) {		/* Hex output? */
	test = 0x10000000L;
    } else {
	test = 1000000000L;	/* Decimal output? */
	if (val < 0) {
	    remainder = -val;
	    *p++ = '-';
	}
    }
    while (test > remainder)	/* Scale down to match number. */
	test /= radix;
    if (remainder)
	while (test) {
	    out = remainder / test;
	    *p++ = out + ((out >= 10) ? '7' : '0');
	    remainder %= test;
	    test /= radix;
	}
    else
	*p++ = '0';
    *p = 0;
    if (ndigits) {		/* A field width set? */
	if ((p2 + ndigits) >= p) /* Make sure we never truncate. */
	    return p - ndigits;
	else
	    return p2;
    }
    return buf;			/* No field width, return minimum number. */
}
