/*
** Sniffer packet dumper, converts to a form suitable for feeding
** to a Verilog simulation (via $readmemh system task).
**
** Usage:
**	tovlog [-V] [file]
**
** Where
**	-V	prints version number and exits,
**	file	Packets in this file, otherwise read from standard in.
*/

#include <stdio.h>
#include <dnix/cheapdrv.h>
#include <dnix/fcodes.h>

#define FORPC			/* DEFINEd for Carriage Returns too. */

struct packetform {
    struct sniffheader snh;
    struct etherheader eth;
    unsigned char data[8192];	/* Way bigger than typically needed. */
};

main(argc, argv)
int argc;
char *argv[];
{
    FILE *fp, *outfp;
    unsigned short packetsize;
    struct packetform packet;
    char *filename, *progname, *argp;
    int ii, packnum;
    char namebuf[100];

    progname = *argv;
    filename = 0;
    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'V':		/* Version printout. */
		printf("%s: Version 1.0\n", progname);
		exit(0);
		break;

	      default:
		fprintf(stderr, "Usage: %s [-V] [packetfile]\n", progname);
		exit(1);
		break;
	    }
	} else {
	    filename = argp;
	}
    }
    if (filename)
	fp = fopen(filename, "r");
    else
	fp = fdopen(0, "r");
    if (!fp) {
	fprintf(stderr, "%s: Unable to read input file %s\n", progname,
		filename ? filename : "");
	exit(2);
    }

    setvbuf(fp, 0, _IOFBF, 4096);
    packnum = -1;
    while (fread(&packetsize, sizeof packetsize, 1, fp) == 1) {
	packnum++;
	packet.snh.statusbits = 0;
	if (packetsize > sizeof(packet)) {
	    printf("ERROR -- %d byte packet is too big!\n", packetsize);
	    fseek(fp, packetsize, 1); /* Skip over it. */
	} else if (packetsize <
		   (sizeof(struct etherheader) + sizeof(struct sniffheader))) {
	    if (!packetsize)	/* It's a runt packet. */
		packetsize = sizeof packetsize;
	    if (fread(&packet, packetsize, 1, fp) != 1)
		break;
	    printf("#%-4d  RUNT   %4d bytes: ", packnum,
		   packet.snh.length - sizeof(struct sniffheader));
	    {
		int i, j, k;
		unsigned char *bufp1;

		i = packet.snh.length - sizeof(struct sniffheader);
		bufp1 = (unsigned char *) &packet.eth;
		k = 0;
		for (j = 0; j < i; j++) { /* Display hex. */
		    if (!(j % 2) && (j != 0)) {
			putc(' ', stdout);
			k++;
		    }
		    printf("%02X", *bufp1++);
		    k += 2;
		}
		printf("%*s\n", 34 - k, "");
	    }
	} else {
	    if (fread(&packet, packetsize, 1, fp) != 1)
		break;
	    if (packet.snh.statusbits == 0xFF) { /* It's a format-override, */
		packnum--;	/* so skip it. */
		continue;
	    }
	    if (!(packet.snh.statusbits & PACKET_GOOD)) { /* Errored? */
		continue;	/* Skip it. */
	    }
	    sprintf(namebuf, "arlpkt%03d.in", packnum);
	    if (!(outfp = fopen(namebuf, "w"))) {
		fprintf(stderr, "%s: Unable to open output file %s\n",
			progname, namebuf);
		exit(3);
		
	    }
#ifdef FORPC
	    fprintf(outfp, "/* Length */ %08X\r\n",
		    packetsize - sizeof(struct sniffheader));
#else
	    fprintf(outfp, "/* Length */ %08X\n",
		    packetsize - sizeof(struct sniffheader));
#endif
	    hd(&packet.eth, packet.snh.length - sizeof(struct sniffheader),
	       outfp);
	    fclose(outfp);
	}
    }
    exit(0);
}

/*
** Dump data in raw hex format.
*/

hd(buffer, byte_count, outfp)
unsigned char *buffer;
int byte_count;
FILE *outfp;
{
    unsigned char *bufp1, *bufp2, ch, ch2;
    char linebuf[80], *p;
    int to_prt;
    int i, j, k;

    bufp1 = bufp2 = buffer;
    for (i = 0; i < byte_count; i += 16) {
	to_prt = (i + 16 > byte_count) ? byte_count - i : 16 ;
	j = bufp1 - buffer;
	p = linebuf;		/* Don't use printf, for speed. */
	*p++ = '/';
	*p++ = '*';
	*p++ = ' ';
	ch = j >> 8;
	ch2 = '0' + (ch & 0x0F);
	if (ch2 > '9')
	    ch2 += 'A' - '9' - 1;
	*p++ = ch2;
	ch = j & 0xFF;
	ch2 = '0' + (ch >> 4);
	if (ch2 > '9')
	    ch2 += 'A' - '9' - 1;
	*p++ = ch2;
	ch2 = '0' + (ch & 0x0F);
	if (ch2 > '9')
	    ch2 += 'A' - '9' - 1;
	*p++ = ch2;
	*p++ = ':';
	*p++ = ' ';
	*p++ = '*';
	*p++ = '/';
	*p++ = ' ';
	*p++ = ' ';
	for (j = 0; j < to_prt; j++) { /* Display hex. */
	    if (!(j % 4) && (j != 0)) {
		*p++ = ' ';
	    }
	    ch = *bufp1++;
	    ch2 = '0' + (ch >> 4);
	    if (ch2 > '9')
		ch2 += 'A' - '9' - 1;
	    *p++ = ch2;
	    ch2 = '0' + (ch & 0x0F);
	    if (ch2 > '9')
		ch2 += 'A' - '9' - 1;
	    *p++ = ch2;
	}
	for (; j < 16; j++) {	/* Fill out rest of line with zeros. */
	    if (!(j % 4) && (j != 0)) {
		*p++ = ' ';
	    }
	    *p++ = '0';
	    *p++ = '0';
	}
	*p++ = ' ';
	*p++ = ' ';
	*p++ = '/';
	*p++ = '*';
	*p++ = ' ';
	for (j = 0; j < to_prt; j++) { /* Display ASCII */
	    if (*bufp2 >= ' ' && *bufp2 <= 0x7E)
		*p++ = *bufp2;
	    else
		*p++ = '.';
	    bufp2++; 
	}
	*p++ = ' ';
	*p++ = '*';
	*p++ = '/';
#ifdef FORPC
	*p++ = '\r';
#endif
	*p++ = '\n';
	*p++ = 0;
	fputs(linebuf, outfp);
    }
}
