/*
 *	Structures used for diskless workstation implementation
 */
#ifndef _DNIX_WS_H
#define _DNIX_WS_H

#define FH_ALLOC	100		/* fc = file handler routine */
#define FH_DEALLOC	101
#define FH_LOAD		102
#define FH_STAT		103
#define FH_CLOSE	104
#define FH_CLOSERQ	105
#define FH_CTTY		106
#define FH_READ		107
#define FH_WRITE	108
#define FH_IPMNT	109
#define FH_IPUMNT	110
#define FH_SWAPIN	111
#define FH_SWAPOUT	112
#define FH_START	113
#define FH_STOP		114
#define FH_KILL		115
#define FH_CLOCK	116
#define FH_ACCESS	117
#define FH_QSTAT	118

#define WSBUFSZ	1024

struct wsbuf
{
	int		w_slid;		/* SAFE LINK id */
	struct wsbuf *	w_next;		/* Link to next buffer */
	int		w_id;		/* Machine id provided by FSS */
	int		w_bufsz;	/* Size of buffer to transfer */
	int		w_pri;		/* Priority of request */
	int		w_rpar;		/* Ruturn parameter */
	int		w_fc;		/* Function code */
	int		w_bcnt;		/* Buffer size */
	int		w_prad;		/* Random address */
	int		w_fid;		/* inode pointer */
	int		w_par;		/* Parameter */
	int		w_kpar;		/* Second parameter */
	int		w_flags;	/* Request flags */
	int		w_dev;		/* Device */
	int		w_uid;		/* User id */
	int		w_gid;		/* Group id */
	int		w_nph;		/* Node and process id */
	int		w_npl;		/* More node and process id */
	int		w_oflag;	/* Open node flags */
	int		w_offlag;	/* More open node flags */
	int		w_rdfid;	/* Root directory inode pointer */
	int		w_ext1;		/* For future extensions */
	int		w_ext2;		/* For future extensions */
	int		w_ext3;		/* For future extensions */
	int		w_tid;		/* Transaction ID. */
	char		w_buf[WSBUFSZ];	/* Buffer for data */
};

struct touch {		/* Structure used to pass inode touch information */
	int	t_fid;
	short	t_flags;
};

#define WSHEADERSZ	((int)sizeof(struct wsbuf) - WSBUFSZ)
/* DNP20 should be changed to reference the disk device when ready */
#if WP4 
#	define WSDEV		makedev(4,0x40)
#else
#	if WP5 | DNP20 | WP6
#		if FH_STUBBS
#			define WSDEV		makedev(4,0x40)
#		else
#			define WSDEV		makedev(17,0x40)
#		endif
#	endif
#	if DS60M10 | DS90M10 | DSM20
#		define WSDEV		makedev(17,0x40)
#	endif
#endif
	
/*
 *	Following structure holds statistics for station
 */
struct wsstat {
	struct dutsname	st_sysinfo;	/* Workstations sysinfo structure */
	int		st_level;	/* File server level */
	int		st_online;	/* Number of connections */
	int		st_offline;	/* Number of broken connections */
	time_t		st_time;	/* Last connect/disconnect time */
	time_t		st_ctime;	/* Acc total connect time */
	time_t		st_dtime;	/* Acc total disconnect time */
	int		st_req;		/* Number of requests */
	int		st_rxdata;	/* Amount of data received */
	int		st_txdata;	/* Amount of data transmitted */
	int		st_read;	/* Number of read requests */
	int		st_szread;	/* Number of bytes read */
	int		st_write;	/* Number of write requests */
	int		st_szwrite;	/* Number of bytes written */
	int		st_ofile;	/* Number of open files */
	int		st_maxofile;	/* Max number of files opened */
	int		st_swapblk;	/* Number of swap blocks allocated */
	int		st_maxblk;	/* Max number of swap blocks alloc'd */
	int		st_load;	/* Amount of data loaded (exec) */
	int		st_swapout;	/* Amount of data swapped out */
	int		st_swapin;	/* Amount of data swapped in */
	int		st_loadread;	/* Amount of load data read from disk to page cache */
	int		st_swapread;	/* Amount of swap data read from disk to page cache */
	short		st_flags	/* Status flags copied from s_flags */
};

#ifdef FSS
/*
 *	Structures local to file system server
 */

/*
 *	Structure defining special FSS link
 */
struct fssl {
	short	l_dev;
	short	l_d1;
	short	l_f1;
	short	l_f2;
};

/*
 *	Structure defining an open file in FSS opened by a remote
 *	work station.
 */
struct fdesc {
	int	f_ip;		/* Inode pointer */
	short	f_flag;		/* Open Node flags */
	short	f_fflag;	/* Fcntl Flags */
};

/*
 *	Structure defining a mounted handler in FSS opened by a remote
 *	work station.
 */
struct hdesc {
	int	h_ip;		/* Inode pointer */
	int	h_prad;		/* address of open node in WS */
};

/*
 *	Structure defining a swap disk block owned by a
 *	remote work station.
 */
union bdesc {
	struct {
		unsigned short		f_tbl;	/* Top bit is inuse flag */
						/* Bottom 3 bits are swap block table number */
		unsigned short		f_nxt;	/* Index to next free block in table */
	} b_free;
	daddr_t b_dadr;				/* Disk address */
};

#define	BDFREE(bp)	(bp->b_free.f_tbl & 0x8000)

/*
 * Structure of swap table control block .
 */

struct	stdesc	{
	struct	bdesc	*sw_adr;	/* base table pointer */
	struct	bdesc	sw_ffree;	/* first free block in chain */
	unsigned short	sw_freecnt;	/* number of free blocks in chain */
};

/*
 *	Structure defining a connection of a remote work station
 */

#define WSNAMESZ	9		/* Length of workstation name */
#define FSS_MAXLINKS	64		/* Max number of softlinks per workstation */
#define	FSS_MAXFILES	144		/* Max number of open files per workstation */
#define	FSS_MAXHANDLERS	12		/* Max number of open handlers per worstation */
#define	FSS_MAXSWAPTBLS	8		/* Max number of Swap Block Tables per workstation */
#define	FSS_MAXWSTN	128		/* Max number of workstations supported */
#define	FSS_SBPST	512		/* Number of swap blocks per swap table */

struct workstn {
	int		s_slid;			/* SAFE LINK id */
	short		s_datasize;		/* Max databuffer size in requests and responces */
	short		s_pagesize;		/* Workstation page size */
	short		s_flags;		/* Some flags */
	char		s_name[WSNAMESZ];	/* Name of workstation */
	struct wsstat	s_stat;			/* Statistics on station */
	struct fssl    *s_endlinks;		/* End of defined softlinks */
	struct fssl	s_links[FSS_MAXLINKS];	/* Table of Soft links definitions */
	struct fdesc	s_files[FSS_MAXFILES];	/* Table of open fd's */
	struct hdesc	s_hndl[FSS_MAXHANDLERS];/* Table of Mounted handlers */
	struct stdesc   s_blocks[FSS_MAXSWAPTBLS];	/* Table of open swap block Tables */
	struct bdesc   *s_ffree;		/* first free swap block */
	short		s_nfiles;		/* number of open files */
	short		s_nhndl;		/* number of mounted handlers */
	short		s_sseqnum;		/* sequence number for sent messages */
	short		s_rseqnum;		/* sequence number for receive messages */
	short		s_count;		/* number of active requests  */
};
#endif

/*
 *  Workstation status flags
 */
#define SF_ACTIVE	0x0001	/* Workstation is activated */
#define SF_CLOSING	0x0002	/* Workstation is closing down */
#define	SF_OPEN		0x0004	/* Workstation is open for normal traffic */
#define SF_CLOSPND	0x0008	/* Workstation has close pending */

/*
 *	Sub functions for user request to control fileserver
 */

#define FSS_ADDLINK	0		/* Add a soft link */
#define FSS_STAT	1		/* Read status */
#define FSS_STATC	2		/* Read and clear status */
#define FSS_ADDWS	3		/* Add a workstation */
#define FSS_RMWS	4		/* Remove a workstation */
#define FSS_KILLWS	5		/* Kill a workstation */
#define FSS_ACT		6		/* Activate workstation */
#define FSS_DEACT	7		/* Deactivate workstation */

#endif	/* _DNIX_WS_H */

