/*
**
** Test the Control Manager stuff.
**
** Also serves as an example application.
*/

#include <isc/window.h>
#include <isc/keycodes.h>
#include <isc/controls.h>

#define XWINL 100		/* For creating our window. */
#define XWINR 400
#define YWINT 100
#define YWINB 200

void updateevent();		/* Forward declare event handlers. */
void menuevent();
void mouseevent();
void keyboardevent();
void activatevent();
void MyAction(ControlID theControl, short partCode);

ControlID controls[10];		/* The controls we'll be using. */
int mywindow;			/* The window we'll be using. */

main(argc, argv)
int argc;
char *argv[];
{
    short i;
    int trapqfd, wmfd;
    Rect winrect, button;
    char *sysid;

    if (argc == 2)		/* Allow an override of where to send */
	sysid = argv[1];	/* our window. */

    /*
    **	Initialize the graphics toolbox.
    */

    if ((trapqfd = InitGraf()) < 0) {
	exit(1);
    }

    /*
    **	Establish a connection to a window manager.
    */

    if ((wmfd = ConnectGraf(sysid)) < 0) {
	exit(2);
    }

    SetRect(&winrect, XWINL, YWINT, XWINR, YWINB);
    mywindow = NewWindow(&winrect, "Control Manager Test",
			 showWindow, dBoxProc, frontWindow, 
			 noOptions, wmfd);
    SetPort(mywindow);
    if (mywindow < 0) {
	exit(3);
    }

    InitControls();		/* We have to do this before using controls. */
    Event(updateEvt, updateevent); /* Bind the standard event handlers. */
    Event(mouseEvt, mouseevent);
    Event(keyEvt, keyboardevent);
    Event(activateEvt, activatevent);

    SetRect(&button, 10, 10, 80, 30); /* Create the controls. */
    i = 0;
    controls[i++] = NewControl(mywindow, &button, "Button", 1,
			       0, 0, 1, pushButProc, i);
    SetRect(&button, 110, 10, 180, 22);
    controls[i++] = NewControl(mywindow, &button, "Check", 1,
			       2, 0, 2, checkBoxProc, i);
    OffsetRect(&button, 0, 40);
    controls[i++] = NewControl(mywindow, &button, "Radio", 1,
			       2, 0, 2, radioButProc, i);
    /* Scroll bars should be 17 pixels in the skinny direction. */
    SetRect(&button, 200, 10, 217, 90);
    /* We initialize a scroll bar with a 0..100 range (arbitrary). */
    controls[i] = NewControl(mywindow, &button, "V Scroll Bar", 1,
			     0, 0, 100, scrollBarProc, i);
    /* Set this scroll bar at 10 (out of 100), with a height of 80 (of 100). */
    SetCtlValues(controls[i], 10, 80);
    SetCtlAction(controls[i++], MyAction); /* For continuous scrolling while */
				/* the buttons of scroll bar are pressed. */

    SetRect(&button, 20, 73, 184, 90);
    controls[i] = NewControl(mywindow, &button, "H Scroll Bar", 1,
			     0, 0, 100, scrollBarProc, i);
    /* Set this scroll bar at 0 (out of 100), with a height of 50 (of 100). */
    SetCtlValues(controls[i], 0, 50);
    SetCtlScrollRate(controls[i], 5); /* Pace this one. */
    SetCtlAction(controls[i++], MyAction); /* For continuous scrolling while */
				/* the buttons of scroll bar are pressed. */
    controls[i] = 0;		/* End marker. */

    (void) GetEvent(anyWindow, noEvent); /* Get events until time to quit. */
/*
** Unlike in IM, we must call KillControls manually upon exit.  There is no
** tie to the window system here.
*/

    KillControls(mywindow);	/* Clean up (unnecessary, actually). */
    DisposeWindow(mywindow);
    DisconnectGraf(wmfd);
    EndGraf();
    exit(0);
}

/*
** Handle keyboard input from the operator.  Here we use it to test various
** features of the controls.
*/

void
keyboardevent(ep)
EventRecord *ep;
{
    short key, i;
    Rect rect;

    if (ep->message & F_KEYUP)
	return;
    key = ep->message & F_KEYMASK;
    if (key == 'Q' || key == 'q' || key == K_ESC) { /* Quit? */
	ReleaseEvent();		/* Makes the GetEvent (above) return. */
	return;
    }
    if (key == 'D' || key == 'd') { /* Deactivate controls? */
	i = 0;
	while (controls[i])
	    HiliteControl(controls[i++], 255); /* 255 marks it inactive. */
    }
    if (key == 'A' || key == 'a') { /* Re-activate controls? */
	i = 0;
	while (controls[i])
	    HiliteControl(controls[i++], 0); /* 0 mark all of it active. */
    }
    if (key == 'H' || key == 'h') { /* Hide controls? */
	i = 0;
	while (controls[i])
	    HideControl(controls[i++]);
    }
    if (key == 'S' || key == 's') { /* Show controls? */
	i = 0;
	while (controls[i])
	    ShowControl(controls[i++]);
    }
}

/*
** On an activate event we always appropriately activate/deactivate our
** Controls.  This is primarily done to hide scroll bars on windows that
** can't take events (so it's obvious you can't use them then).
*/

void
activatevent(ep)
EventRecord *ep;
{
    short i;

    if (ep->message & 1) {	/* Activating? */
	i = 0;
	while (controls[i])
	    HiliteControl(controls[i++], 0); /* Mark all of them active. */
    } else {			/* Deactivating. */
	i = 0;
	while (controls[i])
	    HiliteControl(controls[i++], 255); /* Mark all of them inactive. */
    }
}

/*
** On an update event we always call DrawControls.  It will try to draw all
** of the visible controls associated with this window.
**
** We have to draw _everything_ that's supposed to be visible in the window at
** this time.  Here that is only the controls, which DrawControls does for us.
*/

void
updateevent(ep)
EventRecord *ep;
{
    SetPort(ep->who);
    BeginUpdate(ep->who);
    DrawControls(ep->who);	/* Re-draw all the controls. */
    EndUpdate(ep->who);
}

/*
** If we have controls up we search for a hit on one.  If there is we can
** (optionally) call TrackControl which does all highlighting for us, and 
** returns the part code of the control the mouse is _released_ in.  
** It is customary to reject a mouse control action unless both the hit
** and release are on the same control.  This behavior is not enforced by
** the Control Manager, so you can have it how you want it.  You'll have to
** manually highlight buttons to get a flash if you don't call TrackControl.
**
** Right here we only handle the buttons.  Scroll bars are taken care of
** elsewhere (because we want continuous scrolling on a press).  This is,
** however, where thumb (elevator box) drags are reported.
*/

void
mouseevent(ep)
EventRecord *ep;
{
    ControlID p;
    short i;

    if (ep->message & F_KEYDOWN) { /* Mouse-down event? */
	if (FindControl(&ep->where.corner.topLeft, ep->who, &p)) {
	    i = TrackControl(p, &ep->where.corner.topLeft, (void (*)()) -1);
	    if (i == inCheckBox)
		SetCtlValue(p, !GetCtlValue(p)); /* Toggle state buttons. */
	    if (i == inThumb)
		Bell();		/* Acknowledge thumb drags. */
	}
    }
}

/*
** This routine is called repeatedly if the operator presses in the controls
** of scroll bars.  It is optional (installed via SetCtlAction).  Without it
** the scroll bars only accept a single click in any part of them.  Generally
** users will expect scrolling to be continuous, so you want one of these.
**
** This routine is _not_ called when dragging the thumb.
*/

void
MyAction(theControl, partCode)
ControlID theControl;
short partCode;
{
    if (partCode == inUpButton)
	SetCtlValue(theControl, GetCtlValue(theControl)-5);
    else if (partCode == inDownButton)
	SetCtlValue(theControl, GetCtlValue(theControl)+5);
    else if (partCode == inPageUp)
	SetCtlValue(theControl, GetCtlValue(theControl)-20);
    else if (partCode == inPageDown)
	SetCtlValue(theControl, GetCtlValue(theControl)+20);
}
