/*
** Test copy and video spy functions, mouse window crossing functions,
** and GetEvent error handling functions.
**
** Spy via an `od /netphys/TEST_sysid-to-spy-on/spy-on/sysid-to-send-to'
**  e.g. "od /netphys/TEST_eagle-01/spy-on/eagle-02"
**
** Spies may be shut off with "od /netphys/TEST_eagle-01/spy-off/eagle-02"
** or the like.
**
** Video copy sysid's are command line args.
**  e.g. "test11 eagle-02 eagle-03"
*/

#include <stdio.h>
#include <dnix/fcodes.h>
#include <isc/keycodes.h>
#include <isc/window.h>

#define WIDTH (400)		/* Window definitions. */
#define HIGTH (200)

/*#define DEBUG			/* DEFINED for debug printing. */

static int xpos = 132;
static int ypos = 148;
static char title[] = "Mouse testing.";
static Rect rect1, rect2;

static char *windowsystem;

int trapqfd;
int windowid, windowid2;
long portbits = 0;

struct windowctl {
	short active;
	short fd;
	char name[10];
} windowtab[32];

static Cursor mycursor = {
    {0x0100, 0x0100, 0x0100, 0x0100, 0x0100, 0x0100, 0x0100, 0xFFFE,
     0x0100, 0x0100, 0x0100, 0x0100, 0x0100, 0x0100, 0x0100, 0x0000},
    {0x0380, 0x0380, 0x0380, 0x0380, 0x0380, 0x0380, 0xFFFE, 0xFFFE,
     0xFFFE, 0x0380, 0x0380, 0x0380, 0x0380, 0x0380, 0x0380, 0x0000},
    {8, 8}
};

void updateevent();
void mouseevent();
void keyevent();
void windowevent();
void secondwindevent();
extern char *getenv();

main(argc, argv)
int argc;
char *argv[];
{
    int i, j;
    int fd;
    Rect winrect;
    char *argp;
    EventRecord *ep;

    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
#ifdef DEBUG
	fprintf(stderr, "Error %d in InitGraf\n", -trapqfd);
#endif
	exit(1);
    }
    if (!(windowsystem = getenv("WINDOWS_ON"))) {
#ifdef DEBUG
	fprintf(stderr, "Must be started from a window system.\n");
#endif
	exit(2);
    }
    
    Event(updateEvt, updateevent);
    Event(mouseEvt, mouseevent);
    Event(secondWindEvt, secondwindevent);
    Event(keyEvt, keyevent);
    Event(windowEvt, windowevent);

    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'x':
		xpos = atoi(argp);
		break;
	      case 'y':
		ypos = atoi(argp);
		break;
	    }
	} else {
	    copyvideo(argp, "Video copy");
	}
    }
    if ((fd = ConnectGraf(NULL)) < 0) {	/* Initialize graphics toolbox. */
#ifdef DEBUG
	fprintf(stderr, "Error %d in ConnectGraf\n", -fd);
#endif
	exit(1);
    }
    SetRect(&winrect, xpos, ypos, xpos+WIDTH, ypos+HIGTH);
    OffsetRect(&winrect, -100, -100);
    windowid2 = NewWindow(&winrect, "Drag-onto Window", showWindow, 
			 documentProc, frontWindow, isShy, fd);
    SetPort(windowid2);
    MouseMode(F_KEYDOWN | F_KEYUP | F_MOUSE1);
    OffsetRect(&winrect, 100, 100);
    windowid = NewWindow(&winrect, title, showWindow, documentProc,
			 frontWindow, hasDrag, fd);
    SetPort(windowid);
    MouseMode(F_KEYDOWN | F_KEYUP | F_MOUSE1 | mouseCrossWinds);
    portbits |= (1 << windowid);
    SetPorts(portbits);
    PenMode(srcXor);
    SetCursor(&mycursor);
    MouseMode(F_KEYDOWN | F_KEYUP | F_MOUSE1 | mouseCrossWinds);

    startserver();

    do {
	ep = GetEvent(anyWindow, noEvent);
	if (ep == (EventRecord *) -1) {
	    i = EventError();
#ifdef DEBUG
	    printf("Error %d on a connection.\n", (short) (i & 0xFFFF));
#endif
	    copyconnoff(i >> 16);
	    if (fd == (i >> 16))
		break;		/* Time to quit, main window is down. */
	}
    } while (ep);
}

int serverdone();
int serverfd;
int (*servernode)() = serverdone;
char serverbuffer[256];

/* Report to network as server of name 'TEST_`windowsystem`' */

startserver()
{
    char pathbuffer[256];

    strcpy(pathbuffer,"/netphys/listen/TEST_");
    strcat(pathbuffer, windowsystem);
#ifdef DEBUG
    printf("Listen path = %s\n", pathbuffer);
    fflush(stdout);
#endif
    if ((serverfd = open(pathbuffer, AM_REWR)) >= 0) {
	serverlisten();
    }
}

/* Setup a nowait read on server fd to wait for connection request */

serverlisten()
{
    dnix(F_READ | F_NOWAIT, serverfd, serverbuffer, sizeof(serverbuffer), 0, 0,
	 trapqfd, &servernode) ;
}

/* Connection request receive. Buffer contains path, servicename and open parameters. */

serverdone(np, rpar, rn)
void *np;
int rpar, rn;
{
    register int i, fd;
    register char *cp;
    char pathbuffer[256];

#ifdef DEBUG
    printf("Listen read terminated, rpar = %d\n", rpar);
    fflush(stdout);
#endif
    if (rpar > 1) {
	cp = serverbuffer+1;	/* Bypass netman MQ flag byte */
	rpar -= 1;
	cp[rpar] = 0;
#ifdef DEBUG
	printf("Connect name = %s\n", cp);
	fflush(stdout);
#endif
	strcpy(pathbuffer,"/netphys/");	/* Create full path to open */
	strcat(pathbuffer,cp);	/* requested connection. */
	i = StringLen(cp)+1;	/* Bypass connect path name */
	cp += i;
	rpar -= i;
	i = StringLen(cp)+1;	/* Bypass service name */
	cp += i;
	rpar -= i;
#ifdef DEBUG
	printf("Connect path = %s\n", pathbuffer);
	fflush(stdout);
#endif
/*
** We do this open so's to notify Netman that we've accepted the connection,
** which will take the other side's open off the server queue and put it on
** our queue (all this is internal to Netman).
** 
** This (kludgy) action is what completes the other side's open.  We don't
** need the connection here, but this is one way (the easiest) to notify
** Netman of our actions since we don't have anything but a name at this
** point in time.  (We could have made some write to the Netman fd, but this
** is how it is.  The assumption was made that there would be data exchanged.)
*/
	if ((fd = open(pathbuffer, AM_REWR)) >= 0) { /* Open fd to */
	    if (rpar > 0) {	/* requested connection. */
#ifdef DEBUG
		printf("Call parameters: %s\n", cp);
		fflush(stdout);
#endif
		for (i = 0; i < rpar; i++) {
		    if (cp[i] == '/') break;
		}
		cp[i] = 0;
		if (!strcmp(cp, "spy-on")) {
		    if (rpar > i+1) {
			copyvideo(&cp[i+1], "Video spy");
		    }
		    close(fd);
		} else if (!strcmp(cp, "spy-off")) {
		    copyvideooff(&cp[i+1]);
		    close(fd);
		} else if (!strcmp(cp, "p+message")) {
/*			hnd_p+msg(fd);	*/
		} else if (!strcmp(cp, "p8message")) {
/*			hnd_p8msg(fd);	*/
		} else {
		    close(fd);
		}
	    }
	}
	serverlisten();		/* Wait for next connection request */
    }
}

void
keyevent(ep)
EventRecord *ep;
{
    if (ep->who == windowid) {	/* Main window. */
	if ((ep->message & F_KEYMASK) == K_ESC) {
	    ReleaseEvent();	/* Time to exit. */
	} else {
	    ObscureCursor();
	}
    } else if (ep->who != windowid2) { /* Spy window. */
	if ((ep->message & F_KEYMASK) == K_ESC) {
	    copyoff(ep->who);
	}
    }
}

/*
** Start a spy to the named workstation.
*/

copyvideo(system, title)
char *system, *title;
{
    register int fd, id;
    Rect winrect;
    char windowtitle[100];

    if ((fd = ConnectGraf(system)) < 0) { /* Make connection to */
	return -1;		/* handler/server. */
    }
    strcpy(windowtitle, title);
    strcat(windowtitle, " from system: '");
    strcat(windowtitle, windowsystem);
    strcat(windowtitle, "'");
    SetRect(&winrect, xpos, ypos, xpos+WIDTH, ypos+HIGTH);
    if ((id = NewWindow(&winrect, windowtitle, showWindow, documentProc,
			frontWindow, hasDrag, fd)) < 0) {
	DisconnectGraf(fd);
	return(-1);
    }
    SetPort(id);
    MouseMode(F_KEYDOWN | F_KEYUP | F_MOUSE1 | mouseCrossWinds);
    PenMode(srcXor);
    windowtab[id].active = 1;
    windowtab[id].fd = fd;
    strncpy(windowtab[id].name, system, sizeof(windowtab[id].name));
    portbits |= (1 << id);
    SetPorts(portbits);
    return id;
}

/*
** Shut off a spy copy to the named workstation.
*/

copyvideooff(which)
char *which;
{
    int id;

    for (id = 0; id < (sizeof(windowtab) / sizeof(windowtab[0])); id++) {
	if (windowtab[id].active && !strncmp(which, windowtab[id].name)) {
	    copyoff(id);
	}
    }
}

/*
** Shut off a spy copy to the connection specified by the file descriptor.
**
** For handling errors from GetEvent().
*/

copyconnoff(fd)
int fd;
{
    int id;

    for (id = 0; id < (sizeof(windowtab) / sizeof(windowtab[0])); id++) {
	if (windowtab[id].active && windowtab[id].fd == fd) {
	    copyoff(id);
	}
    }
}

/*
** Shut off a spy copy.
*/

copyoff(id)
register int id;
{
    if (windowtab[id].active) {
	windowtab[id].active = 0;
	SetPort(id);
	DisposeWindow(id);
	DisconnectGraf(windowtab[id].fd);
	portbits &= ~(1 << id);
	SetPorts(portbits);
    }
}

void
updateevent(ep)
EventRecord *ep;
{
    long saveport;

    GetPort(&saveport);
    SetPort(ep->who);
    BeginUpdate(ep->who);
    if (ep->who == windowid2)	/* Drag-onto window. */
	FrameRect(&rect2);
    else
	FrameRect(&rect1);	/* Normal case. */
    EndUpdate(ep->who);
    SetPort(saveport);
}

void
mouseevent(ep)
EventRecord *ep;
{
    Rect *scratch;
    long saveport;
    static long curport;
    static Point hit, drag;

    if (ep->who != windowid && ep->who != windowid2)
	return;			/* No mousies from spy windows. */
    if (ep->who == windowid) {
	scratch = &rect1;
	if (curport != ep->who) {
	    SetPorts(portbits);
	}
    } else {
	scratch = &rect2;
	if (curport != ep->who) {
	    SetPort(ep->who);
	}
    }
    curport = ep->who;
    PenMode(srcXor);
    FrameRect(scratch);
    if (ep->message & F_KEYDOWN) {
	hit = ep->where.corner.topLeft;
    }
    if (ep->message & F_MOUSE1) {
	drag = ep->where.corner.topLeft;
	Pt2Rect(&hit, &drag, scratch);
    }
    FrameRect(scratch);
}

void
secondwindevent(ep)
EventRecord *ep;
{
    Rect scratch;
    long saveport;

    if (ep->who != windowid) {	/* From a window crossing? */
	FrameRect(&rect2);	/* Remove (if any) drawing rectangle. */
	SetRect(&rect2, 0, 0, 0, 0);
	GetPort(&saveport);
	SetPort(ep->who);
	if ((ep->message & F_KEYDOWN) && !(ep->message & F_MOUSE1)) {
	    SetRect(&scratch, 0, 0, 1000, 1000); /* Window exit. */
	    EraseRect(&scratch);
	} else if (ep->message & F_KEYDOWN) { /* Window entrance. */
	    SetRect(&scratch, 0, 0, 1000, 1000);
	    EraseRect(&scratch);
	}
	if (ep->message & F_KEYUP) { /* Mouse release in window. */
	    SetRect(&scratch, 0, 0, 1000, 1000);
	    InvertRect(&scratch);
	}
	if ((ep->message & (F_KEYDOWN | F_KEYUP)) == 0) { /* Normal drag. */
	    scratch.corner.topLeft = ep->where.corner.topLeft;
	    scratch.edge.bottom = scratch.edge.top + 6;
	    scratch.edge.right = scratch.edge.left + 6;
	    FillRect(&scratch, blackPattern);
	}
	SetPort(saveport);
    }
}

void
windowevent(ep)
EventRecord *ep;
{
    switch (ep->message & F_KEYMASK) {
      case windowMove:
	xpos += ep->where.edge.left;
	ypos += ep->where.edge.top;
	MoveWindow(ep->who, xpos, ypos, 0);
	break;
    }
}

