/*
** Test the use of both WM and Clunkdraw at once.
**
** Demonstrates both off-screen drawing and the use of the local Clunkdraw
** connection (if any) to do CopyBits drawing, tests WM's polygon connection.
*/

#include <stdio.h>
#include <malloc.h>

char *errs[] = {"",		/* No error for 0. */
		    "Unable to attach video memory (must setuid root?)",/* 1 */
		    "shared.code file missing.", /* 2 */
		    "shared.code file unreadable.", /* 3 */
		    "Bad format on shared.code file.", /* 4 */
		    "Unable to make semaphore for shared.code file.", /* 5 */
		    "Unable to spawn grafserver.", /* 6 */
		    "Unable to find/create shared memory identifier.", /* 7 */
		    "Unable to attach shared memory segment.", /* 8 */
		    "/dev/graphics is missing/unusable.", /* 9 */
		    "grafserver failed: "}; /* 10 */

#include <errno.h>

#include <isc/window.h>
#include <isc/cd_clunkdraw.h>

#define XWINL 40		/* Window definitions. */
#define XWINR 560
#define YWINT 40
#define YWINB 140

main(argc, argv)
int argc;
char *argv[];
{
    int i, window;
    int trapqfd, fd;
    Rect winrect, myrect;
    GrafPort port;
    BitMap offscreen;
    WindowPtr wptr;
    PolyHandle mypoly;
    RgnHandle myregion;
    void (*initgrafarg)();

    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
	fprintf(stderr, "%s: Error %d in InitGraf().\n", argv[0], -trapqfd);
	exit(1);
    }
    if ((fd = ConnectGraf(0)) < 0) {	/* Initialize graphics toolbox. */
	fprintf(stderr, "%s: Unable to connect to Window Manager (Error #%d in ConnectGraf).\n", 
		argv[0], -fd);
	exit(1);
    }
		/* If we're local, ask for video drawing priveleges. */
    		/* Otherwise, we can only do off-screen work. */
    initgrafarg = WMIsLocal(fd) ? 0 : CD_OpenPort;
    if (i = CD_InitGraf(initgrafarg)) { /* Initialize graphics toolbox. */
	if (i > 10 || i < 0)
	    i=1;
	fprintf(stderr,
		"%s: Can't initialize Clunkdraw (%d, errno = %d), %s",
		argv[0], i, errno, errs[i]);
	if (i == 10 && errno > 200)
	    fprintf(stderr, "%s\n", errs[errno - 200]);
	else
	    fprintf(stderr, "\n");
	exit(i);
    }

    SetRect(&winrect, XWINL, YWINT, XWINR, YWINB); /* Make our WM window. */
    window = NewWindow(&winrect, "Offscreen Bitmap Test", showWindow,
		       noGrowDocProc, frontWindow, noOptions, fd);
    SetPort(window);
    wptr = GetWPtr(window);	/* Is window available for Clunkdraw too? */

    CD_OpenPort(&port);		/* Create the offscreen port and bitmap. */
    CD_OffsetRect(&winrect, -winrect.edge.left, -winrect.edge.top);

    offscreen.bounds = winrect;
    offscreen.rowBytes = (((winrect.edge.right - winrect.edge.left - 1)
			   >> 4) + 1) << 1;
    offscreen.baseAddr = malloc((winrect.edge.bottom - winrect.edge.top)
				* offscreen.rowBytes);
    port.portRect = winrect;
    CD_RectRgn(port.visRgn, &winrect);
    CD_RectRgn(port.clipRgn, &winrect);
    CD_SetPortBits(&offscreen);
    CD_EraseRect(&winrect);	/* Finally offscreen bitmap's built & clean. */

    InsetRect(&winrect, 20, 20); /* Now draw into the offscreen bitmap. */
    CD_FillRoundRect(&winrect, 8, 8, &ltGray);
    CD_FrameRoundRect(&winrect, 8, 8);
    InsetRect(&winrect, 100, -10);
    CD_FillRect(&winrect, &gray);
    CD_PenSize(2, 2);
    CD_FrameRect(&winrect);
    InsetRect(&winrect, -120, -10);

    myregion = CD_NewRgn();	/* Make region to test WM region delivery. */
    SetRect(&myrect, 20, 20, 30, 50);
    CD_OpenRgn();
    CD_FrameOval(&myrect);
    OffsetRect(&myrect, 60, 0);
    CD_FrameOval(&myrect);
    CD_MoveTo(25, 30);
    CD_LineTo(85, 30);
    CD_LineTo(85, 40);
    CD_LineTo(25, 40);
    CD_LineTo(25, 30);
    CD_CloseRgn(myregion);
    CD_OffsetRgn(myregion, 20, 20);

    mypoly = CD_OpenPoly();	/* Make polygon to test WM poly delivery. */
    drawstar();
    CD_ClosePoly();
/*
** Don't make the poly first and then make the region because making a saved
** region can destroy the saved polygon.  You can, of course, copy the saved
** polygon somewhere else after making it -- in which case it doesn't matter
** which order you do things in.
*/

    if (wptr)		/* Make sure WM has gotten and done everything */
	FlushTwice();	/* if we're going to draw via Clunkdraw. */
    CD_SetPort(wptr);
    for (i=notSrcBic; i>=srcCopy; i--) { /* Draw result. */
	if (wptr) {		/* If local, use Clunkdraw. */
	    CD_EraseRect(&wptr->portBits.bounds);
	    CD_CopyBits(&offscreen, &wptr->portBits, &winrect, &wptr->portRect,
			 i, 0L);
	} else {		/* If remote use WM. */
	    EraseRect(&winrect);
	    CopyBits(&offscreen, 0, &winrect, &winrect, i, 0);
	}
    }
    if (wptr) {
	CD_FramePoly(mypoly);	/* Image to Clunkdraw if we can, */
	CD_FillRgn(myregion, &black);
    } else {
	FramePoly(mypoly);	/* else image to WM. */
	FillRgn(myregion, &black);
	FlushWindow();		/* If we're using WM, flush it before sleep. */
    }
    CD_KillPoly(mypoly);	/* Dispose of the saved polygon */
    CD_DisposeRgn(myregion);	/* and the saved region. */
    sleep(5);

    if (wptr && argc > 1) {	/* Do we want to test auto-cleanup of Clunk- */
	SetRect(&winrect, XWINL, YWINT, XWINR, YWINB); /* draw resources? */
	wptr = CD_NewWindow(&winrect, "Cleanup Test", showWindow, /* Yes. */
		       noGrowDocProc, (WindowPtr) -1, noOptions, 0);
	sleep(2);
    }
    CD_EndGraf();		/* Clean up and close our Clunkdraw conn. */
    DisconnectGraf(fd);		/* Disconnect from the WM. */
    EndGraf();			/* Shut down graphics hookup. */
}

drawstar()			/* For making a polygon. */
{
    CD_MoveTo(12, 100);
    CD_LineTo(50, 0);
    CD_LineTo(88, 100);
    CD_LineTo(0, 30);
    CD_LineTo(100, 30);
    CD_LineTo(12, 100);
}
