/*
**
** Test Window Manager locking functions.
**
*/

#include <stdio.h>
#include <errno.h>
#include <isc/window.h>
#include <isc/keycodes.h>

#define XWINL 100
#define XWINR 404
#define YWINT 100
#define YWINB 200

int mywindow;
Rect winrect;

int doupdate();			/* Event handling routines. */
int dokey();
int doactivate();
int windowevent();

main(argc, argv)
int argc;
char *argv[];
{
    short i;
    int trapqfd, wmfd;
    char *sysid;

    sysid = 0;
    if (argc == 2)
	sysid = argv[1];

    /*
    **	Initialize the graphics toolbox.
    */

    if ((trapqfd = InitGraf()) < 0) {
	fprintf(stderr, "%s: Error %d in InitGraf().\n", 
		argv[0], -trapqfd);
	exit(1);
    }

    /*
    **	Establish a connection to a window manager.
    */

    if ((wmfd = ConnectGraf(sysid)) < 0) { /* Initialize graphics toolbox. */
	fprintf(stderr, "%s: Unable to connect to Window Manager (Error #%d in ConnectGraf).\n", 
		argv[0], -wmfd);
	exit(1);
    }

    SetRect(&winrect, XWINL, YWINT, XWINR, YWINB);
    mywindow = NewWindow(&winrect, "Locking Test",
			       showWindow, documentProc, frontWindow, 
			       hasDrag|hasGoAway|hasSendToBack, wmfd);

    SetPort(mywindow);
    if (mywindow < 0) {
	fprintf(stderr, "Window Create failed, errno=%d.\n", errno);
	exit(3);
    }
    Event(updateEvt, doupdate);
    Event(keyEvt, dokey);
    Event(activateEvt, doactivate);
    Event(windowEvt, windowevent);

    (void) GetEvent(anyWindow, noEvent);
}

dokey(ep)
EventRecord *ep;
{
    short key;
    Rect rect;

    if (ep->message & F_KEYUP)
	return;
    SetPort(ep->who);
    key = ep->message & F_KEYMASK;
    if (ep->message & F_FUNC) {
	switch (key) {
	  case K_ESC:
	  case K_RESET:
	    exit(0);
	    break;

	  case K_ENTER:
	  case K_SEND:
	  case K_NPENTER:
	    LockWM();
	    break;

	  case K_BACKSP:
	  case K_BCKSP2:
	    UnlockWM();
	    break;
	}
    }
}

doactivate(ep)
EventRecord *ep;
{
}

doupdate(ep)
EventRecord *ep;
{
    SetPort(ep->who);
    BeginUpdateRect(ep->who, &ep->where);
    EraseRect(&ep->where);
    TextFont(0);
    TextFace(normalText);
    MoveTo(10, 40);
    DrawString("Return ");
    TextFace(italicText);
    DrawString("Locks ");
    TextFace(normalText);
    DrawString("the Window Manager.");
    MoveTo(10, 52);
    DrawString("Backspace ");
    TextFace(italicText);
    DrawString("Unlocks ");
    TextFace(normalText);
    DrawString("the Window Manager.");
    EndUpdate(ep->who);
}

windowevent(ep)
EventRecord *ep;
{
    SetPort(ep->who);
    switch (ep->message & F_KEYMASK) {
      case windowGoaway:
	exit(0);
	break;

      case windowMove:
	OffsetRect(&winrect, ep->where.edge.left, ep->where.edge.top);
	MoveWindow(ep->who, winrect.edge.left, winrect.edge.top, 0);
	break;
    }
}
