/*
** Test the use of both WM and Clunkdraw at once.
**
** Demonstrates both off-screen drawing and the use of the local Clunkdraw
** connection (if any) to do CopyBits drawing.
*/

#include <stdio.h>
#include <malloc.h>

char *errs[] = {"",		/* No error for 0. */
		    "Unable to attach video memory (must setuid root?)",/* 1 */
		    "shared.code file missing.", /* 2 */
		    "shared.code file unreadable.", /* 3 */
		    "Bad format on shared.code file.", /* 4 */
		    "Unable to make semaphore for shared.code file.", /* 5 */
		    "Unable to spawn grafserver.", /* 6 */
		    "Unable to find/create shared memory identifier.", /* 7 */
		    "Unable to attach shared memory segment.", /* 8 */
		    "/dev/graphics is missing/unusable.", /* 9 */
		    "grafserver failed: "}; /* 10 */

#include <errno.h>

#include <isc/window.h>
#include <isc/cd_clunkdraw.h>

#define XWINL 40		/* Window definitions. */
#define XWINR 560
#define YWINT 40
#define YWINB 340

main(argc, argv)
int argc;
char *argv[];
{
    int i, window;
    int trapqfd, fd;
    Rect winrect, rect;
    GrafPort port;
    BitMap offscreen;
    WindowPtr wptr;
    void (*initgrafarg)();

    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
	fprintf(stderr, "%s: Error %d in InitGraf().\n", argv[0], -trapqfd);
	exit(1);
    }
    if ((fd = ConnectGraf(0)) < 0) {	/* Initialize graphics toolbox. */
	fprintf(stderr, "%s: Unable to connect to Window Manager (Error #%d in ConnectGraf).\n", 
		argv[0], -fd);
	exit(1);
    }
		/* If we're local, ask for video drawing priveledges. */
    		/* Otherwise, we can only do off-screen work. */
    initgrafarg = WMIsLocal(fd) ? 0 : CD_OpenPort;
    if (i = CD_InitGraf(initgrafarg)) { /* Initialize graphics toolbox. */
	if (i > 10 || i < 0)
	    i=1;
	fprintf(stderr,
		"%s: Can't initialize Clunkdraw (%d, errno = %d), %s",
		argv[0], i, errno, errs[i]);
	if (i == 10 && errno > 200)
	    fprintf(stderr, "%s\n", errs[errno - 200]);
	else
	    fprintf(stderr, "\n");
	exit(i);
    }

    SetRect(&winrect, XWINL, YWINT, XWINR, YWINB);
    window = NewWindow(&winrect, "Offscreen Bitmap Test", showWindow,
		       noGrowDocProc, frontWindow, noOptions, fd);
    SetRect(&winrect, 40, 40, 560, 140);
    SetPort(window);
    wptr = GetWPtr(window);

    CD_OpenPort(&port);		/* Create the offscreen port and bitmap. */
    CD_OffsetRect(&winrect, -winrect.edge.left, -winrect.edge.top);

    offscreen.bounds = winrect;
    offscreen.rowBytes = (((winrect.edge.right - winrect.edge.left - 1)
			   >> 4) + 1) << 1;
    offscreen.baseAddr = malloc((winrect.edge.bottom - winrect.edge.top)
				* offscreen.rowBytes);
    port.portRect = winrect;
    CD_RectRgn(port.visRgn, &winrect);
    CD_RectRgn(port.clipRgn, &winrect);
    CD_SetPortBits(&offscreen);
    CD_EraseRect(&winrect);

    InsetRect(&winrect, 20, 20); /* Now draw into the offscreen bitmap. */
    CD_FillRoundRect(&winrect, 8, 8, &ltGray);
    CD_FrameRoundRect(&winrect, 8, 8);
    InsetRect(&winrect, 100, -10);
    CD_FillRect(&winrect, &gray);
    CD_PenSize(2, 2);
    CD_FrameRect(&winrect);
    InsetRect(&winrect, -120, -10);

    if (wptr)		/* Make sure WM has gotten and done everything. */
	FlushTwice();
    CD_SetPort(wptr);
    rect = winrect;
    rect.edge.right *= 2;
    rect.edge.bottom *= 2;
    for (i=notSrcBic; i>=srcCopy; i--) { /* Draw result. */
	if (wptr) {		/* If local, use Clunkdraw. */
	    CD_EraseRect(&wptr->portBits.bounds);
	    CD_CopyBits(&offscreen, &wptr->portBits, &winrect, &winrect,i, 0L);
	} else {		/* If remote use WM. */
	    EraseRect(&rect);
	    if (argc > 1)
		CopyBits(&offscreen, 0, &winrect, &rect, i, 0);
	    else
		CopyBits(&offscreen, 0, &winrect, &winrect, i, 0);
	}
    }
    if (!wptr)			/* If we're using WM, flush it before sleep. */
	FlushWindow();
    sleep(5);

    CD_EndGraf();		/* Clean up and close. */
    DisconnectGraf(fd);
    EndGraf();
}
