/*
** Test CopyBits call.
**
** Use Arrow keys, Insert, Delete, PgUp, PgDn, Home, and End keys to vary
** the uses of CopyBits.
**
** Uses direct Clunkdraw calls if/when it can.  This's much faster and 
** eliminates some variables from the test.
**
*/

#include <stdio.h>
#include <errno.h>
#include <isc/window.h>
#include <isc/keycodes.h>
#include <isc/cd_clunkdraw.h>

Rect globalrect, windowrect, view, srcrect;
BitMap source;
int mywindow;

struct BITOBJ {			/* Header for included Data Dictionary */
    short version;		/* bitmap object. */
    short rowBytes;
    short width;
    short height;
    short hres;
    short vres;
    short flags;
    short compress;
};
extern struct BITOBJ holly;
extern char hollybits[];

/* Clunkdraw initialization error messages. */
char *errs[] = {"",		/* No error for 0. */
		    "Unable to attach video memory (must setuid root?)",/* 1 */
		    "shared.code file missing.", /* 2 */
		    "shared.code file unreadable.", /* 3 */
		    "Bad format on shared.code file.", /* 4 */
		    "Unable to make semaphore for shared.code file.", /* 5 */
		    "Unable to spawn grafserver.", /* 6 */
		    "Unable to find/create shared memory identifier.", /* 7 */
		    "Unable to attach shared memory segment.", /* 8 */
		    "/dev/graphics is missing/unusable.", /* 9 */
		    "grafserver failed: "}; /* 10 */

WindowPtr wptr;

main(argc, argv)
int argc;
char *argv[];
{
    int i;
    int trapqfd, wmfd, doupdate(), dokey(), dowindow();
    char *itoa(), namebuf[160];

    /*
     **	Initialize the graphics toolbox.
     */
    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
	strcpy(namebuf, argv[0]);
	strcat(namebuf, ": Error ");
	strcat(namebuf, itoa(-trapqfd));
	strcat(namebuf, " in InitGraf().\n");
	write(2, namebuf, StringLen(namebuf));
	exit(1);
    }
    if ((wmfd = ConnectGraf(0)) < 0) {	/* Initialize graphics toolbox. */
	strcpy(namebuf, argv[0]);
	strcat(namebuf, ": Unable to connect to Window Manager (Error #");
	strcat(namebuf, itoa(-wmfd));
	strcat(namebuf, " in ConnectGraf).\n");
	write(2, namebuf, StringLen(namebuf));
	exit(1);
    }
    /*
     **	Create a window.
     */
    SetRect(&globalrect, 32, 40, 262, 240);
    mywindow = NewWindow(&globalrect, "CopyBits Test", showWindow,documentProc,
			 frontWindow, hasDrag|hasSendToBack|hasGoAway, wmfd);
    SetPort(mywindow);
    windowrect = globalrect;
    OffsetRect(&windowrect, -windowrect.edge.left, -windowrect.edge.top);
    if (argc == 1)		/* Argument disables Clunkdraw access. */
	wptr = GetWPtr(mywindow); /* Is window available for Clunkdraw too? */
    if (wptr) {		/* If we're local, ask for video drawing priveleges. */
	if (i = CD_InitGraf(0)) { /* Initialize graphics toolbox. */
	    strcpy(namebuf, argv[0]);
	    strcat(namebuf, ": Clunkdraw initialization failed (Err #");
	    strcat(namebuf, itoa(i));
	    if (i > 10 || i < 0)
		i=1;
	    strcat(namebuf, ", errno=");
	    strcat(namebuf, itoa(errno));
	    strcat(namebuf, "), ");
	    strcat(namebuf, errs[i]);
	    if (i == 10 && errno > 200)
		strcat(namebuf, errs[errno - 200]);
	    strcat(namebuf, "\n");
	    write(2, namebuf, StringLen(namebuf));
	    exit(i);
	}
    }

    source.rowBytes = holly.rowBytes;
    SetRect(&source.bounds, 0, 0, holly.width, holly.height);
    srcrect = source.bounds;
    source.baseAddr = hollybits;
    view = source.bounds;

    Event(updateEvt, doupdate);
    Event(keyEvt, dokey);
    Event(windowEvt, dowindow);
    GetEvent(anyWindow, noEvent); /* Do the program. */

    DisposeWindow(mywindow);	/* Delete the window. */
    DisconnectGraf(wmfd);	/* Disconnect from Window Manager. */
    EndGraf();			/* Un-init graphics toolbox. */
    if (wptr)
	CD_EndGraf();
    exit(0);
}

doupdate(evtptr)
EventRecord *evtptr;
{
    Rect scratch;

    SetPort(evtptr->who);
    BeginUpdateRect(evtptr->who, &evtptr->where);
    if (SectRect(&evtptr->where, &view, &scratch)) {
	if (wptr) {
	    FlushTwice();
	    CD_SetPort(wptr);
	    CD_CopyBits(&source, &wptr->portBits, &srcrect, &view, srcCopy, 0);
	} else {
	    CopyBits(&source, 0, &srcrect, &view, srcCopy, 0);
	}
    }
    EndUpdate(evtptr->who);
}

moveme(dh, dv)			/* Move the window. */
int dh, dv;
{
    OffsetRect(&globalrect, dh, dv);
    MoveWindow(mywindow, globalrect.edge.left, globalrect.edge.top, 0);
}

shiftme(dh, dv)			/* Scroll the window. */
int dh, dv;
{
    OffsetRect(&view, dh, dv);
    ScrollRect(&windowrect, dh, dv, -1);
}

dokey(evtptr)
EventRecord *evtptr;
{
    Rect scratch;
    int altflag;

    if (evtptr->message & F_KEYUP) /* Skip key-up's. */
	return;
    if (!(evtptr->message & F_FUNC)) { /* Skip non-function keys. */
	Bell();
	return;
    }
    SetPort(evtptr->who);
    altflag = evtptr->message & F_ALT;
    switch (evtptr->message & F_KEYMASK) {
      case K_UP:
	if (altflag)
	    moveme(0, -18);
	else
	    shiftme(0, -11);
	break;

      case K_DOWN:
	if (altflag)
	    moveme(0, 18);
	else
	    shiftme(0, 11);
	break;

      case K_RIGHT:
	if (altflag)
	    moveme(18, 0);
	else
	    shiftme(11, 0);
	break;

      case K_LEFT:
	if (altflag)
	    moveme(-18, 0);
	else
	    shiftme(-11, 0);
	break;

      case K_PGUP:		/* Zoom in. */
	if (altflag) {
	    InsetRect(&srcrect, 10, 10);
	    InsetRect(&view, 10, 10);
	    EraseRect(&windowrect);
	    InvalRect(&windowrect);
	} else {
	    view.edge.right += 10;
	    view.edge.bottom += 10;
	    InvalRect(&view);
	}
	break;

      case K_PGDN:		/* Zoom out. */
	if (altflag) {
	    InsetRect(&srcrect, -10, -10);
	    InsetRect(&view, -10, -10);
	    EraseRect(&windowrect);
	    InvalRect(&windowrect);
	} else {
	    scratch = view;
	    scratch.edge.left = scratch.edge.right - 10;
	    EraseRect(&scratch);
	    scratch.edge.left = view.edge.left;
	    scratch.edge.top = scratch.edge.bottom - 10;
	    EraseRect(&scratch);
	    view.edge.right -= 10;
	    view.edge.bottom -= 10;
	    InvalRect(&view);
	}
	break;

      case K_HOME:		/* Reset the view rectangle & repaint. */
	if (!EqualRect(&view, &source.bounds)) {
	    view = source.bounds;
	    EraseRect(&windowrect);
	    InvalRect(&windowrect);
	}
	if (!EqualRect(&srcrect, &source.bounds)) {
	    srcrect = source.bounds;
	    EraseRect(&windowrect);
	    InvalRect(&windowrect);
	}
	break;

      case K_END:		/* Repaint the window. */
	EraseRect(&windowrect);
	InvalRect(&windowrect);
	break;

      case K_INSERT:		/* Add a ClipRect to the party. */
	scratch = windowrect;
	scratch.edge.right = (scratch.edge.right + scratch.edge.left) / 2;
	ClipRect(&scratch);
	EraseRect(&scratch);
	InvalRect(&scratch);
	break;

      case K_DELETE:		/* Remove the ClipRect. */
	ClipRect(&windowrect);
	EraseRect(&windowrect);
	InvalRect(&windowrect);
	break;

      case K_RESET:		/* Quit the program. */
      case K_ESC:
	ReleaseEvent();
	break;

      default:			/* Holler if we don't understand. */
	Bell();
	break;
    }
}

dowindow(evtptr)
EventRecord *evtptr;
{
    switch (evtptr->message & F_KEYMASK) {
      case windowGoaway:
	ReleaseEvent();
	break;

      case windowMove:
	OffsetRect(&globalrect, evtptr->where.edge.left,
		   evtptr->where.edge.top);
	MoveWindow(evtptr->who, globalrect.edge.left, globalrect.edge.top, 0);
	break;
    }
}

char *
itoa(val)
int val;
{
    static char buf[20];
    char *p;
    int test;

    p = buf;
    test = 1000000;
    if (val < 0) {
	val = -val;
	*p++ = '-';
    }
    while (test > val)
	test /= 10;
    if (val)
	while (test) {
	    *p++ = val / test + '0';
	    val %= test;
	    test /= 10;
	}
    else
	*p++ = '0';
    *p = 0;
    return buf;
}
