/*
** Test Bold text measurements.
*/

#include <stdio.h>
#include <errno.h>
#include <isc/window.h>
#include <isc/keycodes.h>
#include <isc/cd_clunkdraw.h>

char foo[] = {"This is bold text"};
Rect globalrect, windowrect, view, srcrect;
int mywindow;

main(argc, argv)
int argc;
char *argv[];
{
    int font, size;
    int trapqfd, wmfd;
    Point penpos;
    char *itoa(), namebuf[160];

    /*
     **	Initialize the graphics toolbox.
     */
    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
	strcpy(namebuf, argv[0]);
	strcat(namebuf, ": Error ");
	strcat(namebuf, itoa(-trapqfd));
	strcat(namebuf, " in InitGraf().\n");
	write(2, namebuf, StringLen(namebuf));
	exit(1);
    }
    if ((wmfd = ConnectGraf(0)) < 0) {	/* Initialize graphics toolbox. */
	strcpy(namebuf, argv[0]);
	strcat(namebuf, ": Unable to connect to Window Manager (Error #");
	strcat(namebuf, itoa(-wmfd));
	strcat(namebuf, " in ConnectGraf).\n");
	write(2, namebuf, StringLen(namebuf));
	exit(1);
    }
    /*
     **	Create a window.
     */
    SetRect(&globalrect, 32, 40, 362, 240);
    OffsetRect(&globalrect, 200, 0);
    mywindow = NewWindow(&globalrect, "Text Measurement Test", showWindow,
			 documentProc, frontWindow, noOptions, wmfd);
    SetPort(mywindow);

    MoveTo(100, 100);
    TextFace(boldText);
    size = 6;
    if (argc > 1)
	font = atoi(argv[1]);
    if (argc > 2)
	size = atoi(argv[2]);
    TextFont(font);
    TextSize(size);
    DrawString(foo);
    GetPen(&penpos);
    printf("Font: %d, size: %d points.\n", font, size);
    printf("Measured width: %d pixels.\n", penpos.h - 100);
    printf("Calculated width: %d pixels.\n", StringWidth(foo));
    printf("Locally Calculated width: %d pixels.\n",
	   LocalStringWidth(font, boldText, size, foo));
    FlushWindow();
    SetRect(&globalrect, 100, 100, penpos.h, 105);
    InvertRect(&globalrect);
    FlushWindow();
    globalrect.edge.top -= 10;
    EraseRect(&globalrect);
    FlushWindow();
testtables(font, size);
    DisposeWindow(mywindow);	/* Delete the window. */
    DisconnectGraf(wmfd);	/* Disconnect from Window Manager. */
    EndGraf();			/* Un-init graphics toolbox. */
    exit(0);
}

char *
itoa(val)
int val;
{
    static char buf[20];
    char *p;
    int test;

    p = buf;
    test = 1000000;
    if (val < 0) {
	val = -val;
	*p++ = '-';
    }
    while (test > val)
	test /= 10;
    if (val)
	while (test) {
	    *p++ = val / test + '0';
	    val %= test;
	    test /= 10;
	}
    else
	*p++ = '0';
    *p = 0;
    return buf;
}

testtables(font, size)
int font, size;
{
    register FontRec *frp;
    unsigned char *_wmgetfonttable();
    int face;
    char *s;
    register char *owtab;
    int i, j;

    face = 0;
    frp = (FontRec *)_wmgetfonttable(font, face, size);

    switch (frp->fontType) {
      case 1:
	s = "ISC format Roman";
	break;

      case 2:
	s = "ISC format Asian";
	break;

      case 0x9000:
      case 0x9001:
      case 0xD000:
      case 0xD001:
	s = "Macintosh format Roman proportional";
	break;

      case 0xB000:
	s = "Macintosh format Roman fixed-pitch";
	break;

      default:
	printf("Font table not a valid font format ($%04X).\n",
	       frp->fontType);
	exit(5);
	break;
    }
    printf("Dump of font table, font %d size %d:\n\tFormat: %s font.\n", 
	   font, size, s);
    printf("\tfontType\t$%04X\n", frp->fontType);
    printf("\tfirstChar\t$%02X (%d)\n", frp->firstChar, frp->firstChar);
    printf("\tlastChar\t$%02X (%d)", frp->lastChar, frp->lastChar);
    printf(", %d character span\n", 1 + frp->lastChar - frp->firstChar);
    printf("\twidMax\t\t%d pixels\n", frp->widMax);
    printf("\tkernMax\t\t%d pixels\n", frp->kernMax);
    printf("\tnDescent\t%d pixels", frp->nDescent);
    if (frp->nDescent != -frp->descent)
	puts(" (Error?)");
    else
	puts("");
    printf("\tfRectWidth\t%d pixels\n", frp->fRectWidth);
    printf("\tfRectHeight\t%d pixels\n", frp->fRectHeight);
    printf("\tascent\t\t%d pixels\n", frp->ascent);
    printf("\tdescent\t\t%d pixels\n", frp->descent);
    printf("\tleading\t\t%d pixels (Total vertical pitch: %d pixels)\n",
	   frp->leading, frp->ascent + frp->descent + frp->leading);
    printf("\trowWords\t%d words\n\n", frp->rowWords);

    owtab = (char *) &frp[1]; /* Proportional font format, get pointer */
    j = 0;
    if (frp->owTLoc) {
	for (i=frp->firstChar; i<=(frp->lastChar + 2); i++) {
	    switch (frp->fontType) {
	      case 1:
		printf("\tChar %02X (%d):\tWidth: %3d", i, i, 
		       frp->owTLoc ? owtab[j] : frp->widMax);
		break;

	      case 0x9000:
	      case 0x9001:
	      case 0xB000:
	      case 0xD000:
	      case 0xD001:
		printf("\tChar %02X (%d):\t Width: %3d", i, i,
		       frp->owTLoc ? owtab[j] : frp->widMax);
		break;
	    }
	    if (frp->owTLoc && owtab[j] == -1)
		printf(" (Missing)");
	    if (i >= ' ' && i <= 126)
		printf("  '%c'", i);
	    else if (i == (frp->lastChar + 1))
		printf("  Missing Symbol");
	    puts("");
	    j++;
	}
    }
    if (frp->fontType == 2)
	printf("\tAsian font image file: '%s'\n", &frp[1]);
    
}
