/*
** Test the Color Picker/File Picker service calls.
*/

#include <stdio.h>
#include <isc/window.h>
#include <dnix/fcodes.h>

				/* Color palette I/O structure. */
struct iostuff {
    long size;			/* Size of this structure (in bytes). */
    short ncolors;		/* Number of entries following. */
    struct iopalette {
	unsigned short index, red, green, blue;
    } colors[8];
} instruct;

SFDialog fp;

void waitdone(), waitdone2();

struct nwblock {		/* No-wait wait completion block. */
    void (*completion)();
    int exitstat;
} waitchild = {waitdone, 0};

struct nwblock waitchild2 = {waitdone2, 0};

main(argc, argv)
int argc;
char *argv[];
{
    int i, reqnum;
    struct iopalette *p;
    int trapqfd, child;

    trapqfd = InitGraf();
    if (argc > 1) {		/* Test File Picker. */
	fp.options = SF_GETFILE;
	fp.initialname = "test24.c";
	if (FilePicker(&fp) >= 0) { /* Test the blocking version. */
	    dumpit2();		/* OK, show us what we picked. */
	    fp.options = SF_PUTFILE; /* Test the no-wait version. */
	    fp.initialname = 0;
	    if ((child = NWFilePicker(&fp, trapqfd)) >= 0) {
		reqnum = rdnix(F_NOWAIT | F_WAIT, 0, &waitchild2.exitstat, 
			       sizeof waitchild2.exitstat, 0, 0, trapqfd,
			       &waitchild2);
		GetEvent(noWindow, noEvent);
	    } else {
		puts("NW File Pick failed.");
		exit(2);
	    }
	    exit(0);
	} else {
	    puts("File Pick failed.");
	}
    } else {			/* Test Color Picker. */
	instruct.size = sizeof instruct; /* Preload our color pick struct. */
	instruct.ncolors = 8;	/* There are 8 colors to set. */
	p = &instruct.colors[0];
	for (i=0; i < 8; i++) {
	    p->index = i + 8;	/* We want colors 8 to 15 to be set. */
	    p->red = p->green = p->blue = 0xFFFF; /* Initially all white. */
	    p++;
	}

	setvbuf(stdout, 0L, _IOFBF, 0); /* Fully buffered stdout. */
	if (ColorPicker(&instruct, 9) >= 0) /* Test the blocking version. */
	    dumpit();		/* OK, show us what changed. */

	if ((child = NWColorPicker(&instruct, 10, trapqfd)) >= 0) {
	    rdnix(F_NOWAIT | F_WAIT, 0, &waitchild.exitstat, 
		  sizeof waitchild.exitstat, 0, 0, trapqfd, &waitchild);
	    GetEvent(noWindow, noEvent); /* Test the no-wait version. */
	}
    }
    EndGraf();
    exit(1);
}

void
waitdone(p, rpar, reqnum)	/* No-wait wait completion routine. */
struct nwblock *p;
int rpar, reqnum;
{
    if (!p->exitstat)		/* Color pick successful? */
	dumpit();		/* Show us your goodies! */
    EndGraf();
    exit(0);
}

dumpit()			/* Print out the (modified) pick structure. */
{
    int i;
    struct iopalette *p;

    p = &instruct.colors[0];
    for (i=0; i < instruct.ncolors; i++) {
	printf("Color %2d = [$%04X, $%04X, $%04X]\n",
	       p->index, p->red, p->green, p->blue);
	p++;
    }
    puts("");
    fflush(stdout);
}

void
waitdone2(p, rpar, reqnum)	/* No-wait wait completion routine. */
struct nwblock *p;
int rpar, reqnum;
{
    if (!p->exitstat)		/* Color pick successful? */
	dumpit2();		/* Show us your goodies! */
    EndGraf();
    exit(0);
}

dumpit2()			/* Print out the file picked. */
{
    if (fp.returnedname) {
	puts(fp.returnedname);
    }
}
