/*
** Hack program to test the Touch Screen driver's transformation matrix
** code and the calibration stuff.
*/

#include <stdio.h>
#include <isc/window.h>
#include <isc/keycodes.h>

void updateevent();
void mouseevent();
void keyevent();

int window, wmfd;
HdwrType theHdwrRecord;
struct ts_transform tform;

main(argc, argv)
int argc;
char *argv[];
{
    int i, j;
    int trapqfd;
    Rect winrect, button;
    char *argp;

    tform.tformA[0][0] = tform.tformA[1][1] = Int2Fix(1); /* Default. */
    i = open("/dev/mouse", 0);	/* Pick up OS driver transform matrix. */
    ioctl(i, GETTRANSFORM, &tform);
    close(i);

    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
	exit(1);
    }
    if ((wmfd = ConnectGraf(0)) < 0) { /* Initialize graphics toolbox. */
	exit(2);
    }
    Event(updateEvt, updateevent);
    Event(mouseEvt, mouseevent);
    Event(keyEvt, keyevent);

    GetHdwrType(&theHdwrRecord);
    SetRect(&winrect, 0, 0, theHdwrRecord.hdots, theHdwrRecord.vdots);
    window = NewWindow(&winrect, "Touch Screen Calibration Test", showWindow,
		       documentProc, frontWindow, noOptions, wmfd);
    SetPort(window);
    TextMode(srcCopy);
    MouseMode(F_KEYDOWN | F_KEYUP | F_MOUSE1);
    if (GetEvent(anyWindow, noEvent)) /* Process user interaction. */
	return 1;
    return 0;
}

void
mouseevent(ep)
EventRecord *ep;
{
    Point src, dst;
    char buffer[128];

    src = ep->where.corner.botRight;
    transform(&dst, &src);
    mysprintf(buffer, "Raw [%d, %d] -> [%d, %d]          ", src.h, src.v,
	      dst.h, dst.v);
    MoveTo(10, 110);
    DrawString(buffer);
    mysprintf(buffer, "Driver cooked: [%d, %d]           ",
	      ep->where.edge.left, ep->where.edge.top);
    MoveTo(10, 130);
    DrawString(buffer);
}

void
keyevent(ep)
EventRecord *ep;
{
    ReleaseEvent();
}

void
updateevent(ep)
EventRecord *ep;
{
    SetPort(ep->who);
    BeginUpdate(ep->who);
    MoveTo(10, 50);
    DrawString("Test the Touch Screen mouse driver calibration.");
    MoveTo(10, 70);
    DrawString("Press any key to quit.");
    EndUpdate(ep->who);
}

/*
** Kjell's matrix transformation code.
**
** The transform() routine is actually in the OS touch-screen driver,
** and is used to convert raw device coordinates into the scaled/rotated
** screen-sized range.  It is included here for informational purposes.
**
** The setuptransform() routine makes the necessary transformation matrix
** to convert one set of coordinates to another.
**
** These routines were lifted from the P+ document alignment program.
*/

transform(dest, source)
short *dest, *source;
{
    int i;

    for (i = 0; i < 2; i++) {
	dest[i] = Fix2Int(FixMul(Int2Fix(source[0]), tform.tformA[i][0])
	    + FixMul(Int2Fix(source[1]), tform.tformA[i][1]) + tform.tformB[i]);
    }
}

/*
** Create a transformation matrix out of three points P, Q, and R,
**
** +-----------------------+
** |                       |
** |   * P            * Q  |
** |                       |
** |                       |
** |                       |
** |                       |
** |                       |
** |                       |
** |                       |
** |                       |
** |                  * R  |
** |                       |
** +-----------------------+
**
** with both input (raw mouse) and output (pixel) space coordinates
** given for all three points.  The matrix will then map input space
** to output space by applying the above transform() function.
**
**                        +=     =+
** The matrix is of form: | A B 0 |   so that: X' = AX + BY + E
**                        | C D 0 |            Y' = CX + DY + F
**                        | E F 1 |
**                        +=     =+
**
** This is a standard two-dimensional coordinate transformation matrix that
** can do transformation (shifting), scaling, and rotation.  Coefficients
** A-F are stored in a 2x2 matrix for A-D (matrix A), and a separate 1x2
** matrix for E-F (matrix B) for space efficiency.  Coefficients A-F are
** in fixed-point (16.16) format for speed/space reasons (inside the
** touch-screen driver).
*/

/*
** Sprintf replacement.
**
** Handles format strings containing:
**
**	 %s, %[0][1..9]d, %[0][1..9]x, %[0][1..9]X
**
** It's much smaller than the standard one.
*/
mysprintf(buffer, fmt, arglist)
char *buffer, *fmt;
char *arglist;
{
    register char *p, *q, *p2;
    short ndigits;
    int leadfill;
    char *itoa();
    union arg {
		long lng;
		char *ptr;
    } *argptr;
	
    argptr = (union arg *) &arglist;
    p = fmt;
    q = buffer;
    while (*p) {		/* While there's some format string. */
	if (*p == '%') {	/* Is it a format control? */
	    p++;		/* Yes, which one (skip over % char)? */
	    if (*p == '0')
		leadfill = *p++;
	    else
		leadfill = ' ';
	    if (*p >= '1' && *p <= '9')	/* Format output length? */
		ndigits = *p++ - '0';
	    else
		ndigits = 0;
	    if (*p == 's') {
		p2 = argptr++->ptr; /* String. */
	    } else if (*p == 'd') {
		p2 = itoa(argptr++->lng, 10, leadfill, ndigits);
	    } else if (*p == 'x' || *p == 'X') {
		p2 = itoa(argptr++->lng, 16, leadfill, ndigits);
	    }
	    p++;		/* Skip over format control. */
	    while (*p2)		/* Copy the inclusion. */
		*q++ = *p2++;
	} else			/* No, it's a literal -- copy it. */
	    *q++ = *p++;
    }
    *q = 0;			/* Null-terminate it. */
    return q - buffer;		/* Return number of characters processed. */
}

char *
itoa(val, radix, leadfill, ndigits)
int val, radix, leadfill, ndigits;
{
    short i;
    char *p, *p2, out;
    static char buf[20];
    unsigned long test, remainder;
	
    p = buf;
    remainder = val;
    i = ndigits;		/* Make sure there're at least the number */
    while (i--)			/* of leading whatevers that we've asked for */
	*p++ = leadfill;	/* as the field width. */
    p2 = p;
    if (radix == 16) {		/* Hex output? */
	test = 0x10000000L;
    } else {
	test = 1000000000L;	/* Decimal output? */
	if (val < 0) {
	    remainder = -val;
	    *p++ = '-';
	}
    }
    while (test > remainder)	/* Scale down to match number. */
	test /= radix;
    if (remainder)
	while (test) {
	    out = remainder / test;
	    *p++ = out + ((out >= 10) ? '7' : '0');
	    remainder %= test;
	    test /= radix;
	}
    else
	*p++ = '0';
    *p = 0;
    if (ndigits) {		/* A field width set? */
	if ((p2 + ndigits) >= p) /* Make sure we never truncate anything. */
	    return p - ndigits;
	else
	    return p2;
    }
    return buf;			/* No field width, return minimum number. */
}
