/*
**
** Test the standard patterns.
**
*/

#include <isc/window.h>
#include <isc/keycodes.h>

#define WIDTH (600)		/* Window definitions. */
#define HEIGHT (400)

static int xpos = 16;
static int ypos = 48;
static char title[] = "Standard patterns for numeric reference.";
static char *path;

void updateevent();
void keyevent();

main(argc, argv)
int argc;
char *argv[];
{
    int i, j;
    int trapqfd, fd;
    Rect winrect;
    int window;
    char *argp, buf[160];

    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'x':
		xpos = atoi(argp);
		break;
	      case 'y':
		ypos = atoi(argp);
		break;
	    }
	} else {
	    path = argp;
	}
    }
    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
	mysprintf(buf, "%s: Error %d in InitGraf().\n", argv[0], -trapqfd);
	write(2, buf, StringLen(buf));
	return 1;
    }
    if ((fd = ConnectGraf(path)) < 0) {	/* Initialize graphics toolbox. */
	mysprintf(buf, "%s: Unable to connect to Window Manager (Error #%d in ConnectGraf).\n", 
		  argv[0], -fd);
	write(2, buf, StringLen(buf));
	return 2;
    }
    SetRect(&winrect, xpos, ypos, xpos+WIDTH, ypos+HEIGHT);
    window = NewWindow(&winrect, title, showWindow, documentProc,
		       frontWindow, noOptions, fd);
    SetPort(window);

    Event(updateEvt, updateevent);
    Event(keyEvt, keyevent);
    GetEvent(anyWindow, noEvent);
    return 0;
}

void
updateevent(ep)
EventRecord *ep;
{
    BeginUpdate(ep->who);
    staticinfo();
    EndUpdate(ep->who);
}

void
keyevent(ep)
EventRecord *ep;
{
    long msg;

    msg = ep->message;
    if (msg & F_KEYUP)
	return;
    if (msg & F_FUNC) {
	switch (msg & F_KEYMASK) {
	  case K_RESET:
	  case K_ESC:
	    ReleaseEvent();
	    break;

	  default:
	    Bell();
	}
    } else {
	Bell();
    }
}

staticinfo()
{
    int i, x, y;
    Rect rect;
    char buffer[80];
    for (i = 0; i < 64; i++) {
	x = (i % 8) * 64 + 44;
	y = (i / 8) * 48 + 20;
	MoveTo((i<10) ? x+22 : x+16, y-6);
	mysprintf(buffer, "%d", i);
	DrawString(buffer);
	SetRect(&rect, x-1, y-1, x+49, y+25);
	FrameRect(&rect);
	InsetRect(&rect, 1, 1);
	FillRect(&rect, (Pattern *) i);
    }
}

/*
** Sprintf replacement.
**
** Handles format strings containing:
**
**	 %s, %[0][1..9]d, %[0][1..9]x, %[0][1..9]X
**
** It's much smaller than the standard one.
*/
mysprintf(buffer, fmt, arglist)
char *buffer, *fmt;
char *arglist;
{
    register char *p, *q, *p2;
    short ndigits;
    int leadfill;
    char *itoa();
    union arg {
	long lng;
	char *ptr;
    } *argptr;
	
    argptr = (union arg *) &arglist;
    p = fmt;
    q = buffer;
    while (*p) {		/* While there's some format string. */
	if (*p == '%') {	/* Is it a format control? */
	    p++;		/* Yes, which one (skip over % char)? */
	    if (*p == '0')
		leadfill = *p++;
	    else
		leadfill = ' ';
	    if (*p >= '1' && *p <= '9')	/* Format output length? */
		ndigits = *p++ - '0';
	    else
		ndigits = 0;
	    if (*p == 's') {
		p2 = argptr++->ptr; /* String. */
	    } else if (*p == 'd') {
		p2 = itoa(argptr++->lng, 10, leadfill, ndigits);
	    } else if (*p == 'x' || *p == 'X') {
		p2 = itoa(argptr++->lng, 16, leadfill, ndigits);
	    }
	    p++;		/* Skip over format control. */
	    while (*p2)		/* Copy the inclusion. */
		*q++ = *p2++;
	} else			/* No, it's a literal -- copy it. */
	    *q++ = *p++;
    }
    *q = 0;			/* Null-terminate it. */
    return q - buffer;		/* Return number of characters processed. */
}

char *
itoa(val, radix, leadfill, ndigits)
int val, radix, leadfill, ndigits;
{
    short i;
    char *p, *p2, out;
    static char buf[20];
    unsigned long test, remainder;
	
    p = buf;
    remainder = val;
    i = ndigits;		/* Make sure there're at least the number */
    while (i--)			/* of leading whatevers that we've asked for */
	*p++ = leadfill;	/* as the field width. */
    p2 = p;
    if (radix == 16) {		/* Hex output? */
	test = 0x10000000L;
    } else {
	test = 1000000000L;	/* Decimal output? */
	if (val < 0) {
	    remainder = -val;
	    *p++ = '-';
	}
    }
    while (test > remainder)	/* Scale down to match number. */
	test /= radix;
    if (remainder)
	while (test) {
	    out = remainder / test;
	    *p++ = out + ((out >= 10) ? '7' : '0');
	    remainder %= test;
	    test /= radix;
	}
    else
	*p++ = '0';
    *p = 0;
    if (ndigits) {		/* A field width set? */
	if ((p2 + ndigits) >= p) /* Make sure we never truncate anything. */
	    return p - ndigits;
	else
	    return p2;
    }
    return buf;			/* No field width, return minimum number. */
}
