/*
**
** Test the Picture recording/playback calls.
**
*/

#include <isc/window.h>
#include <isc/keycodes.h>

#define WIDTH (600)		/* Window definitions. */
#define HEIGHT (400)

void updateevent(), keyevent(), menuevent();
PicHandle thepict;

struct BITOBJ {			/* Header for included Data Dictionary */
    short version;		/* bitmap object. */
    short rowBytes;
    short width;
    short height;
    short hres;
    short vres;
    short flags;
    short compress;
};
extern struct BITOBJ holly;
extern char hollybits[];

main(argc, argv)
int argc;
char *argv[];
{
    int trapqfd, fd;
    Rect winrect;
    int window;

    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
	return 1;
    }
    if ((fd = ConnectGraf(0)) < 0) { /* Initialize graphics toolbox. */
	return 2;
    }
    SetRect(&winrect, 20, 50, 20 + WIDTH, 50 + HEIGHT);
    window = NewWindow(&winrect, "Test Pictures", showWindow, documentProc, 
		       frontWindow, noOptions, fd);
    SetPort(window);
    SetRect(&winrect, 0, 0, WIDTH, HEIGHT);

    Event(updateEvt, updateevent);
    Event(keyEvt, keyevent);
    Event(menuEvt, menuevent);
    NewMenu(1, "Edit");
    AppendMenu(1, "/CCopy");

    thepict = OpenPicture(&winrect);
    staticinfo();
    ClosePicture();

    GetEvent(anyWindow, noEvent);
    KillPicture(thepict);
    return 0;
}

void
updateevent(ep)
EventRecord *ep;
{
    Rect teensy;

    BeginUpdate(ep->who);
    DrawPicture(thepict, &(*thepict)->picFrame);
    SetRect(&teensy, 0, 0, WIDTH/10, HEIGHT/10);
    OffsetRect(&teensy, 6, HEIGHT/2);
    DrawPicture(thepict, &teensy);
    OffsetRect(&teensy, 0, HEIGHT/5);
    teensy.edge.bottom += 20;
    teensy.edge.right += 40;
    DrawPicture(thepict, &teensy);
    EndUpdate(ep->who);
}

void
keyevent()
{
    ReleaseEvent();
}

void
menuevent(ep)
EventRecord *ep;
{
    BeginMenu(ep->message);
    ZeroScrap();
    PutScrap((*thepict)->picSize, 'PICR', *thepict);
    PublishScrap(0);
    EndMenu(ep->message);
}

staticinfo()
{
    int i, x, y;
    int xbias;
    int ybias;
    BitMap source;
    Rect view;
    PolyHandle mypoly;
    char *cp;

    xbias = WIDTH/2;
    ybias = HEIGHT-20;
    TextFont(0);
    TextSize(12);
    MoveTo(1, ybias);
    LineTo(WIDTH-2, ybias);
    MoveTo(4, ybias - 4);
    DrawString(cp = "X Axis");
    PicComment(24, StringLen(cp) + 1, cp);
    MoveTo(xbias, 1);
    LineTo(xbias, HEIGHT - 2);
    MoveTo(xbias + 4, 12);
    DrawString(cp = "Y Axis");
    PicComment(42, StringLen(cp) + 1, cp);
    for (i = -xbias; i < WIDTH - xbias; i++) {
	x = xbias + i;
	y = ybias - (i*i)/100;
	if (y >= 0) {
	    SetPixel(x, y);
	}
    }

    source.rowBytes = holly.rowBytes;
    SetRect(&source.bounds, 0, 0, holly.width, holly.height);
    source.baseAddr = hollybits;
    view = source.bounds;
    OffsetRect(&view, xbias + 20, 40);
    CopyBits(&source, 0, &source.bounds, &view, srcOr, 0);

    PenSize(20, 20);
    view = source.bounds;
    OffsetRect(&view, 20, 20);
    FrameRoundRect(&view, 10, 4);
    PenNormal();
    MoveTo(view.edge.left, view.edge.top);
    Line(view.edge.right - view.edge.left, view.edge.bottom - view.edge.top);
    MoveTo(view.edge.right, view.edge.top);
    LineTo(view.edge.left, view.edge.bottom);
    mypoly = OpenPoly();
    drawstar(WIDTH/2 + 40, HEIGHT-200);
    ClosePoly();
    mypoly = DupePoly(mypoly);
    FramePoly(mypoly);
    OffsetPoly(mypoly, 80, 80);
    FillPoly(mypoly, greyPattern);
    KillPoly(mypoly);
}

drawstar(h, v)			/* For making a polygon. */
int h, v;
{
    MoveTo(h+12, v+100);
    LineTo(h+50, v+0);
    LineTo(h+88, v+100);
    LineTo(h+0, v+30);
    LineTo(h+100, v+30);
    LineTo(h+12, v+100);
}

#if 0
void
StdComment(kind, dataSize, data)
short kind;
short dataSize;
void *data;
{
    printf("Comment: %d (%s)\n", kind, dataSize ? data : "");
}
#endif
