/*
**
** Test the Picture parsing call.  Gets the test picture from the
** clipboard, you can run test30 to put one there.
**
*/

#include <isc/window.h>

PicHandle thepict;

main(argc, argv)
int argc;
char *argv[];
{
    int trapqfd, fd;
    long size;
    void *malloc();

    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
	return 1;
    }
    if ((fd = ConnectGraf(0)) < 0) { /* Initialize graphics toolbox. */
	return 2;
    }
    size = GetScrap(0, 'PICR', 0);
    if (size) {
	thepict = malloc(sizeof(PicPtr));
	*thepict = malloc(size);
	GetScrap(*thepict, 'PICR', size);
	printf("Picture (PICR) from the ClipBoard:\n");
	ParsePicture(thepict);
    } else {
	printf("No picture (PICR) on the ClipBoard.\n");
    }
    return 0;
}

char *
verber(verb)
int verb;
{
    switch (verb) {
      case 0: return "frame"; break;
      case 1: return "paint"; break;
      case 2: return "erase"; break;
      case 3: return "invert"; break;
      case 4: return "fill"; break;

      case 5: return "set"; break;
      case 6: return "sect"; break;
      case 7: return "union"; break;
      case 8: return "diff"; break;
      case 9: return "xor"; break;
    }
}

void
StdClipRect(int verb, Rect *rp)
{
    printf(" ClipRect %s [(%d, %d) - (%d, %d)]\n", verber(verb),
	   rp->edge.left, rp->edge.top, rp->edge.right, rp->edge.bottom);
}

void
StdBits(sbits, dummy, srect, drect, mode, dummy2)
BitMap *sbits;
int dummy;
Rect *srect;
Rect *drect;
int mode;
int dummy2;
{
    printf(" BitMap [(%d, %d) - (%d, %d)] mode %d\n",
	   drect->edge.left, drect->edge.top, drect->edge.right,
	   drect->edge.bottom, mode);
}

void
StdLine(from, to, gs)
Point *from;
Point *to;
GrafState *gs;
{
    printf(" Line from (%d, %d) to (%d, %d) Mode %d, Size (%d, %d)\n",
	   from->h, from->v, to->h, to->v, gs->pnMode,
	   gs->pnSize.h, gs->pnSize.v);
}

void
StdText(text, count, gs)
char *text;
int count;
GrafState *gs;
{
    printf(" Text \"%*s\" at (%d, %d), FFSM: %d, %d, %d, %d\n", count, text,
	   gs->pnLoc.h, gs->pnLoc.v, gs->txFont, gs->txFace,
	   gs->txSize, gs->txMode);
}

void
StdRGBColor(int index, int red, int green, int blue)
{
    printf(" RGB Color set %d: $%04X, $%04X, $%04X, $%04X\n",
	   index, red & 0xFFFF, green & 0xFFFF, blue & 0xFFFF);
}

void
StdRect(verb, rp, gs)
int verb;
Rect *rp;
GrafState *gs;
{
    printf(" Rectangle %s: [(%d, %d) - (%d, %d)]\n", verber(verb),
	   rp->edge.left, rp->edge.top, rp->edge.right, rp->edge.bottom);
}

void
StdOval(verb, rp, gs)
int verb;
Rect *rp;
GrafState *gs;
{
    printf(" Oval %s: [(%d, %d) - (%d, %d)]\n", verber(verb),
	   rp->edge.left, rp->edge.top, rp->edge.right, rp->edge.bottom);
}

void
StdRRect(verb, rp, hrad, vrad, gs)
int verb;
int hrad;
int vrad;
Rect *rp;
GrafState *gs;
{
    printf(" Round Rect %s: [(%d, %d) - (%d, %d)]  Rad: (%d, %d)\n",
	   verber(verb),
	   rp->edge.left, rp->edge.top, rp->edge.right, rp->edge.bottom,
	   hrad, vrad);
}

void
StdArc(verb, rp, startAngle, arcAngle, gs)
int verb;
Rect *rp;
int startAngle;
int arcAngle;
GrafState *gs;
{
    printf(" Arc %s [(%d, %d) - (%d, %d)]  From: %d for %d\n",
	   verber(verb),
	   rp->edge.left, rp->edge.top, rp->edge.right, rp->edge.bottom,
	   startAngle, arcAngle);
}

void
StdWedge(verb, rp, startAngle, arcAngle, gs)
int verb;
Rect *rp;
int startAngle;
int arcAngle;
GrafState *gs;
{
    printf(" Wedge %s [(%d, %d) - (%d, %d)]  From: %d for %d\n",
	   verber(verb),
	   rp->edge.left, rp->edge.top, rp->edge.right, rp->edge.bottom,
	   startAngle, arcAngle);
}

void
StdPoly(int verb, PolyHandle ph, GrafState *gs)
{
    printf(" Polygon %s\n", verber(verb));
}

void
StdRgn(int verb, RgnHandle rh, GrafState *gs)
{
    printf(" Region %s\n", verber(verb));
}

void
StdSetClip(int verb, RgnHandle rh)
{
    printf(" Clip Region %s\n", verber(verb));
}

void
StdText16(short *text, int count, GrafState *gs)
{
    printf(" 16-bit Text, %d characters\n", count);
}

void
StdJustText(txt, count, rect, hflag, vflag, gs)
char *txt;
int count;
Rect *rect;
int hflag;
int vflag;
GrafState *gs;
{
    printf(" Justified Text \"%*s\"\n", count, txt);
}
