/*
**
** Test the Event System and Menu stuff.
**
*/

#include <stdio.h>
#include <isc/window.h>
#include <isc/keycodes.h>

#define WIDTH (600)		/* Window sizes. */
#define HEIGHT (220)

#define MENU1	1
#define MENU2	2
#define MENU5	5
#define MENU7	7
#define MENU8	8

static int xpos = 32;
static int ypos = 48;
static char title[] = "Menu testing.";
static char *path;
static char message[100];

void updateevent();
void menuevent();
void mouseevent();
void keyboardevent();
void windowevent();
void scrapevent();
int nullevent();

main(argc, argv)
int argc;
char *argv[];
{
    int i, j;
    int trapqfd, fd;
    Rect winrect;
    int window;
    char *argp;

    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'x':
		xpos = atoi(argp);
		break;
	      case 'y':
		ypos = atoi(argp);
		break;
	    }
	}
	else {
	    path = argp;
	}
    }
    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
	return 1;
    }
    if ((fd = ConnectGraf(path)) < 0) {	/* Connect to WM. */
	return 2;
    }
    SetRect(&winrect, xpos, ypos, xpos+WIDTH, ypos+HEIGHT);
    window = NewWindow(&winrect, title, showWindow, documentProc, frontWindow,
		       hasDrag|hasSendToBack|hasZoom|hasGoAway, fd);
    SetPort(window);
    SetKbdLEDs(1);
    MouseMode(mouseAll);

    Event(updateEvt, updateevent);
    Event(menuEvt, menuevent);
    Event(mouseEvt, mouseevent);
    Event(keyEvt, keyboardevent);
    Event(windowEvt, windowevent);
    Event(scrapEvt, scrapevent);
    Event(noEvent, nullevent);

				/* First build the menus themselves. */
    NewMenu(MENU1, "File");
    AppendMenu(MENU1, "/NNew;-;/OOpen;-;Quit;Other");
    SetItemStyle(MENU1, 5, outlineText);
    SetItemKey(MENU1, 1, F_FUNC | K_F1);
    SetItemKey(MENU1, 5, 'Q' | F_CTRL | F_SHIFT); /* F_SHIFT makes menu key */
    SetItemKey(MENU1, 6, 'q' | F_CTRL | F_SHIFT); /* case-sensitive. */
    NewMenu(MENU2, "Edit");
    AppendMenu(MENU2, "/ZUndo;-;!\036/XCut;/CCopy;(/VPaste");
    NewMenu(MENU5, "Disabled");
    AppendMenu(MENU5, "Item1;-;!\036Item3;Item4;(Item5");
    DisableMenu(MENU5);
    NewMenu(MENU7, "Settings");
    AppendMenu(MENU7, "(Local Echo;-;Diamond;Coal");
    SetItemKey(MENU7, 3, K_F10+F_FUNC+F_CTRL+F_ALT);
    SetItemKey(MENU7, 4, F_ALT + 'A');
    NewMenu(MENU8, "Styles");
    AppendMenu(MENU8, "/PPlain Text;/B<BBold;/I<IItalic;/U<UUnderline;/X<XStrikeout;/O<OOutline;A Long one Just For Fun");
    (void) GetEvent(anyWindow, noEvent);
    return 0;
}

void
keyboardevent(ep)
EventRecord *ep;
{
    short key;
    Rect rect;

    key = ep->message & F_KEYMASK;
    if (!(ep->message & (F_CTRL | F_ALT | F_FUNC)))
	if (key == 'Q' || key == 'q')
	    exit(0);
    sprintf(message, "Keyevent  $%08X, (%d, %d), (%d, %d) %08X", ep->message,
	    ep->where.edge.left, ep->where.edge.top, ep->where.edge.right,
	    ep->where.edge.bottom, ep->when);
    SetRect(&rect, 0, 120, WIDTH, 150);
    EraseRect(&rect);
    MoveTo(10, 140);
    DrawString(message);
}

void
updateevent(ep)
EventRecord *ep;
{
    BeginUpdate(ep->who);
    TextFace(boldText);
    MoveTo(10, 50);
    DrawString("Test of Pulldown menus (quit with 'q').");
    TextFace(normalText);
    MoveTo(10, 70);
    DrawString("Hit ALT F1 - F12 to bring up menu.");
    MoveTo(10, 90);
    DrawString("Then select item by pressing F1 - F12, or arrow keys.");
    MoveTo(10, 110);
    DrawString("Select with ENTER, Cancel with ESC.");
    TextFace(underlineText);
    MoveTo(10, 140);
    DrawString(message);
    EndUpdate(ep->who);
}

void
menuevent(ep)
EventRecord *ep;
{
    int menuid, menuitem;
    Rect rect;

    BeginMenu(ep->message);
    menuid = ep->message >> 16;
    menuitem = ep->message & 0xFFFF;
    sprintf(message, "Menu event %d, %d", menuid, menuitem);
    SetRect(&rect, 0, 160, WIDTH, 180);
    EraseRect(&rect);
    MoveTo(10, 170);
    DrawString(message);
    EndMenu(ep->message);
}

void
mouseevent(ep)
EventRecord *ep;
{
    Rect rect;
    static int count;

    count++;
    sprintf(message, "Mouse event $%08X, (%d, %d), (%d, %d) %08X %d",
	    ep->message, ep->where.edge.left, ep->where.edge.top,
	    ep->where.edge.right, ep->where.edge.bottom, ep->when, count);
    SetRect(&rect, 0, 120, WIDTH, 150);
    EraseRect(&rect);
    MoveTo(10, 140);
    DrawString(message);
}

void
windowevent(ep)
EventRecord *ep;
{
    Rect rect;

    sprintf(message, "Window event $%08X, (%d, %d), (%d, %d) %08X", 
	    ep->message, ep->where.edge.left, ep->where.edge.top,
	    ep->where.edge.right, ep->where.edge.bottom, ep->when);
    SetRect(&rect, 0, 160, WIDTH, 180);
    EraseRect(&rect);
    MoveTo(10, 170);
    DrawString(message);
}

void
scrapevent(ep)
EventRecord *ep;
{
    Rect rect;

    sprintf(message, "Scrap event $%08X, (%d, %d), (%d, %d) %08X", 
	    ep->message, ep->where.edge.left, ep->where.edge.top,
	    ep->where.edge.right, ep->where.edge.bottom, ep->when);
    SetRect(&rect, 0, 160, WIDTH, 180);
    EraseRect(&rect);
    MoveTo(10, 170);
    DrawString(message);
}

int
nullevent()
{
    Rect rect;
    static int pass = 0;
    char message[100];

    pass++;
    sprintf(message, "Null event #%d", pass);
    SetRect(&rect, 0, 200, WIDTH, 210);
    EraseRect(&rect);
    MoveTo(10, 210);
    DrawString(message);
    return (pass % 10);		/* Stop (temporarily) after 10 null events. */
}
