/*
**
** Test mouse.
**
*/

#include <stdio.h>
#include <isc/keycodes.h>
#include <isc/window.h>
#include <isc/controls.h>
ControlID hscrollbar, vscrollbar;
Rect button;

#define WIDTH (400)				/* Window definitions. */
#define HEIGHT (200)
#define SMALLWIDTH (200)
#define SMALLHEIGHT (20)

static int xpos = 32;
static int ypos = 48;
static int width = WIDTH;
static int height = HEIGHT;

static char title[] = "Mouse testing";
static char *path;
static Rect rect;
static Point hit, drag;

static Cursor mycursor = {
	{ 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0xfffe,
	  0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0x100, 0 },
	{ 0x380, 0x380, 0x380, 0x380, 0x380, 0x380, 0xfffe, 0xfffe,
	  0xfffe, 0x380, 0x380, 0x380, 0x380, 0x380, 0x380, 0 },
	{ 8, 8}
};

int updateevent();
int mouseevent();
int windowevent();
int menuevent();

main(argc, argv)
int argc;
char *argv[];
{
    int i, j;
    int trapqfd, fd;
    Rect winrect;
    int window;
    char *argp;

    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'x':
		xpos = atoi(argp);
		break;
	      case 'y':
		ypos = atoi(argp);
		break;
	    }
	} else {
	    path = argp;
	}
    }
    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
	fprintf(stderr,"Error %d in GrafInit\n",-trapqfd);
	exit(1);
    }
    if ((fd = ConnectGraf(path)) < 0) {	/* Initialize graphics toolbox. */
	fprintf(stderr,"Error %d in ConnectGraf\n",-fd);
	exit(1);
    }
    SetRect(&winrect, xpos, ypos, xpos+width, ypos+height);
    window = NewWindow(&winrect, title, showWindow, documentProc, frontWindow, 
		       hasGoAway|hasZoom|hasSendToBack|hasGrow|hasDrag, fd);
    SetPort(window);
    PenMode(srcXor);
    SetCursor(&mycursor);
    MouseMode(F_KEYDOWN | F_KEYUP | F_MOUSE1);

    NewMenu(1, "Quit");
    AppendMenu(1, "/QReally");
    Event(updateEvt, updateevent);
    Event(mouseEvt, mouseevent);
    Event(windowEvt, windowevent);
    Event(menuEvt, menuevent);

    InitControls();
    button.edge.left = width - 16;
    button.edge.right = width + 1;
    button.edge.top = -1;
    button.edge.bottom = height - 15;
    vscrollbar = NewControl(window, &button, "V", 1, 0, 0, 10,
			   scrollBarProc, 0);
    button.edge.left = -1;
    button.edge.right = width - 15;
    button.edge.top = height - 16;
    button.edge.bottom = height + 1;
    hscrollbar = NewControl(window, &button, "H", 1, 0, 0, 10,
			   scrollBarProc, 0);

    while ((j = GetKey(window)) != EOF
	   && (j &= F_KEYMASK) != 'q' && j != 'Q') {
	ObscureCursor();
    }
}

windowevent(ep)
EventRecord *ep;
{
    static int zoomeddown;
    Rect rect;

    switch (ep->message & F_KEYMASK) {
      case windowGoaway:
	DisposeWindow(ep->who);
	ReleaseEvent();
	break;

      case windowMove:
	xpos += ep->where.edge.left;
	ypos += ep->where.edge.top;
	MoveWindow(ep->who, xpos, ypos, 0);
	break;

      case windowResize:
	if (!zoomeddown) {
	    SetRect(&rect, width-17, height-17, width, height);
	    EraseRect(&rect);	/* Erase old grow box. */
	    width += ep->where.edge.left;
	    height += ep->where.edge.top;
	    SizeWindow(ep->who, width, height, updateRequest);
	    rect.edge.left = width - 16;
	    rect.edge.top = -1;
	    HideControl(vscrollbar);
	    MoveControl(vscrollbar, rect.edge.left, rect.edge.top);
	    SizeControl(vscrollbar, 17, height - 14);
	    ShowControl(vscrollbar);
	    rect.edge.left = -1;
	    rect.edge.top = height - 16;
	    HideControl(hscrollbar);
	    MoveControl(hscrollbar, rect.edge.left, rect.edge.top);
	    SizeControl(hscrollbar, width - 14, 17);
	    ShowControl(hscrollbar);
	    DrawGrowIcon(ep->who); /* Draw new grow box. */
	}
	break;

      case windowZoom:
	if (zoomeddown) {
	    SizeWindow(ep->who, width, height, updateRequest);
	} else {
	    SizeWindow(ep->who, SMALLWIDTH, SMALLHEIGHT, updateRequest);
	}
	zoomeddown ^= 1;
	break;
    }
}

updateevent(ep)
EventRecord *ep;
{
    BeginUpdate(ep->who);
    DrawControls(ep->who);
    DrawGrowIcon(ep->who);
    FrameRect(&rect);
    EndUpdate(ep->who);
}

mouseevent(ep)
EventRecord *ep;
{
    Rect scratch;

    PenMode(srcXor);
    FrameRect(&rect);
/*	printf("Mouse event 0x%x, (%d, %d)\n", ep->message,
		   ep->where.edge.left, ep->where.edge.top);
*/
    if (ep->message & F_KEYDOWN) {
	hit = ep->where.corner.topLeft;
    }
    if (ep->message & F_MOUSE1) {
	drag = ep->where.corner.topLeft;
	Pt2Rect(&hit, &drag, &rect);
    }
    FrameRect(&rect);
}

menuevent(ep)
EventRecord *ep;
{
    BeginMenu(ep->message);
    EndMenu(ep->message);
    exit(0);
}
