/*
**
** Test Summagraphics tablet transformation testing.
**
*/

#include <stdio.h>
#include <isc/keycodes.h>
#include <isc/window.h>

#define WIDTH (300)				/* Window definitions. */
#define HIGTH (400)

static int xpos = 32;
static int ypos = 48;
static char title[] = "Summa transformation testing.";
static char *path;
static Rect rect;
static Rect docrect;

#define TOP 	50
#define BOTTOM	350
#define LEFT	50
#define RIGHT	250

Point dest[3];
Point source[3];

int updateevent();
int mouseevent();
int keyevent();

main(argc, argv)
int argc;
char *argv[];
{
    int i, j;
    int trapqfd, fd;
    Rect winrect;
    int window;
    char *argp;

	while (--argc > 0) {
		argp = *++argv;
		if (*argp == '-') {
			argp++;
			switch (*argp++) {
				case 'x':
					xpos = atoi(argp);
					break;
				case 'y':
					ypos = atoi(argp);
					break;
			}
		}
		else {
			path = argp;
		}
	}
	if ((trapqfd = InitGraf()) < 0) {		/* Initialize graphics toolbox. */
		fprintf(stderr,"Error %d in GrafInit\n",-trapqfd);
		exit(1);
	}
	if ((fd = ConnectGraf(path)) < 0) {		/* Initialize graphics toolbox. */
		fprintf(stderr,"Error %d in ConnectGraf\n",-fd);
		exit(1);
	}
	SetRect(&winrect, xpos, ypos, xpos+WIDTH, ypos+HIGTH);
	window = NewWindow(&winrect, title, showWindow, documentProc, frontWindow, noOptions, fd);
	SetPort(window);
	PenMode(srcCopy);
	MouseMode(F_KEYDOWN);
	HideCursor();

	Event(updateEvt,updateevent);
	Event(mouseEvt,mouseevent);
	Event(keyEvt, keyevent);

	getklick(window, &source[0], 20, 40, "Upper left corner");
	SetPt(&dest[0], LEFT, TOP);
	getklick(window, &source[1], 20, 40, "Upper right corner");
	SetPt(&dest[1], RIGHT, TOP);
	getklick(window, &source[2], 20, 40, "Lower left corner");
	SetPt(&dest[2], LEFT, BOTTOM);

	setuptransform(dest,source);

	SetRect(&winrect, 0, 0, WIDTH, HIGTH);
	EraseRect(&winrect);

	SetRect(&docrect, LEFT, TOP, RIGHT, BOTTOM);
	FrameRect(&docrect);

	PenMode(srcXor);
	MouseMode(mouseAll);

	while ((j = GetKey(window)) != EOF && (j &= F_KEYMASK) != 'q' && j != 'Q');
}

getklick(wid, pointp ,xpos ,ypos ,text)
int wid;
Point *pointp;
int xpos, ypos;
char *text;
{
	EventRecord *ep;
	Rect rect;
	MoveTo(xpos,ypos);
	DrawString(text);
	do {
		ep = GetEvent(1 << wid, 1 << mouseEvt);
	} while ((ep->message & F_KEYDOWN) == 0);
	SetRect(&rect, xpos, ypos-20, xpos+WIDTH, ypos+10);
	EraseRect(&rect);
	*pointp = ep->where.corner.botRight;
}

updateevent(ep)
EventRecord *ep;
{
	BeginUpdate(ep->who);
	FrameRect(&rect);
	EndUpdate(ep->who);
}

mouseevent(ep)
EventRecord *ep;
{
	FrameRect(&rect);
	transform(&ep->where.edge.top, &ep->where.edge.bottom);
	if (ep->message & F_KEYDOWN)
		rect.corner.topLeft = ep->where.corner.topLeft;
	if (ep->message & F_MOUSE1) {
		if (rect.edge.top <= ep->where.edge.top) {
			rect.edge.bottom = ep->where.edge.top;
		}
		if (rect.edge.left <= ep->where.edge.left) {
			rect.edge.right = ep->where.edge.left;
		}
	}
	FrameRect(&rect);
}

keyevent(ep)
EventRecord *ep;
{
    if (ep->message & F_KEYDOWN)
	if ((ep->message & F_KEYMASK) == 'q' ||
	    (ep->message & F_KEYMASK) == 'Q')
	    exit(0);
}

static long a[2][2], b[2];

setuptransform(dest,source)
short dest[][2], source[][2];
{
	int i;
	float a0, a1, b0, d1, tmp;
	d1 = (source[0][0] - source[1][0]);
	d1 /= (source[0][1] - source[1][1]);
	tmp = (source[0][0] - source[2][0]);
	tmp /= (source[0][1] - source[2][1]);
	d1 -= tmp;
	for (i = 0; i < 2; i++) {
		a1 = (dest[0][i] - dest[1][i]);
		a1 /= (source[0][1] - source[1][1]);
		tmp = (dest[0][i] - dest[2][i]);
		tmp /= (source[0][1] - source[2][1]);
		a1 -= tmp;
		a1 /= d1;
		a0 = (dest[0][i] - dest[1][i]) - a1 * (source[0][0] - source[1][0]);
		a0 /= (source[0][1] - source[1][1]);
		b0 = dest[0][i] - a0 * source[0][1] - a1 * source[0][0];
		a[i][0] = a1 * 65536;
		a[i][1] = a0 * 65536;
		b[i] = b0 + .5;
	}
}

transform(dest,source)
short *dest, *source;
{
	int i;
	for (i = 0; i < 2; i++) {
		dest[i] = FixMul(source[0], a[i][0]) + FixMul(source[1], a[i][1]) + b[i];
	}
}
