/*
**
** Test the Hardware Status call.
**
*/

#include <stdio.h>
#include <errno.h>
#include <isc/window.h>

#define XWINL 40		/* Window definitions. */
#define XWINR 490
#define YWINT 50
#define YWINB 270

HdwrType HdwrRecord;

main(argc, argv)
int argc;
char *argv[];
{
    short i;
    int trapqfd, wmfd;
    void doupdate();
    Rect winrect;
    int mywindow;
    char *sysid;

    if (argc == 2)
	sysid = argv[1];

    /*
    **	Initialize the graphics toolbox.
    */

    if ((trapqfd = InitGraf()) < 0) {
	fprintf(stderr, "%s: Error %d in InitGraf().\n", 
		argv[0], -trapqfd);
	exit(1);
    }

    /*
    **	Establish a connection to a window manager.
    */

    if ((wmfd = ConnectGraf(sysid)) < 0) {
	fprintf(stderr, "%s: Error %d in ConnectGraf().\n", 
		argv[0], -wmfd);
	exit(1);
    }

    SetRect(&winrect, XWINL, YWINT, XWINR, YWINB);
    mywindow = NewWindow(&winrect, "Workstation Hardware Characteristics",
			       showWindow, documentProc, frontWindow, 
			       noOptions, wmfd);
    SetPort(mywindow);
    if (mywindow < 0) {
	fprintf(stderr, "Window Create failed, errno=%d.\n", errno);
	exit(3);
    }

    Event(updateEvt, doupdate);
    GetHdwrType(&HdwrRecord);
    (void) GetEvent(anyWindow, 1<<keyEvt);

    DisposeWindow(mywindow);
    EndGraf();
}

void
doupdate(ep)
EventRecord *ep;
{
    SetPort(ep->who);
    BeginUpdate(ep->who);
    stats();
    EndUpdate(ep->who);
}

stats()
{
    int i, j;
    char buffer[80], buffer2[80];

    i = 20;
    j = 26;
    TextFont(2);
    TextSize(18);
    TextFace(underlineText);
    sprintf(buffer, "Hardware Information:");
    MoveTo(i, j);
    i += 20;
    j += 26;
    DrawString(buffer);
    TextFont(0);
    TextSize(12);
    TextFace(normalText);
    sprintf(buffer, "Screen Size:  %d (H) by %d (V).", HdwrRecord.hdots,
	    HdwrRecord.vdots);
    MoveTo(i, j);
    j += 20;
    DrawString(buffer);
    sprintf(buffer, "Screen Resolution:  %d (H) dots/inch by %d (V) dots/inch.",
	    HdwrRecord.hres, HdwrRecord.vres);
    MoveTo(i, j);
    j += 20;
    DrawString(buffer);
    sprintf(buffer, "Keyboard:  ");
    switch (HdwrRecord.kbdtype) {
      case 0:			/* None. */
	sprintf(buffer2, "None.");
	break;

      case 1:			/* PC AT keyboard. */
	sprintf(buffer2, "PC AT.");
	break;

      case 2:			/* ISC Wincom keyboard. */
	sprintf(buffer2, "ISC Wincom.");
	break;

      case 3:			/* Both. */
	sprintf(buffer2, "PC AT and ISC Wincom both.");
	break;
    }
    MoveTo(i, j);
    j += 20;
    DrawString(buffer);
    DrawString(buffer2);
    sprintf(buffer, "Monitor:  ");
    switch (HdwrRecord.montype) {
      case 0:			/* 9" Black & White. */
	sprintf(buffer2, "9\" Black & White.");
	break;

      case 1:			/* 14" Black & White. */
	sprintf(buffer2, "14\" Black & White.");
	break;

      case 2:			/* 14" Color. */
	sprintf(buffer2, "14\" Color.");
	break;
    }
    MoveTo(i, j);
    j += 20;
    DrawString(buffer);
    DrawString(buffer2);
    sprintf(buffer, "Processor:  Motorola ");
    switch (HdwrRecord.proctype) {
      case 0:			/* 68000. */
	sprintf(buffer2, "68000.");
	break;

      case 1:			/* 68010. */
	sprintf(buffer2, "68010.");
	break;

      case 2:			/* 68020. */
	sprintf(buffer2, "68020.");
	break;

      case 3:			/* 68030. */
	sprintf(buffer2, "68030.");
	break;

      case 4:			/* 68040. */
	sprintf(buffer2, "68040.");
	break;
    }
    MoveTo(i, j);
    j += 20;
    DrawString(buffer);
    DrawString(buffer2);

    sprintf(buffer, "Logic Box:  ");
    switch (HdwrRecord.nplanes) {
      case 1:			/* Monochrome. */
	sprintf(buffer2, "Monochrome.");
	break;

      default:			/* Color. */
	sprintf(buffer2, "Color (%d planes of %d available).",
		HdwrRecord.planesinuse, HdwrRecord.nplanes);
	break;
    }
    MoveTo(i, j);
    j += 20;
    DrawString(buffer);
    DrawString(buffer2);
    sprintf(buffer, "Total Memory:  %ld bytes.", HdwrRecord.memsize);
    MoveTo(i, j);
    j += 20;
    DrawString(buffer);

    sprintf(buffer, "Color Model:  ");
    switch (HdwrRecord.colormodel) {
      case 0:			/* Monochrome. */
	sprintf(buffer2, "Monochrome.");
	break;

      case 1:			/* Color. */
	sprintf(buffer2, "Old-style WP5 Color.");
	break;

      case 2:			/* WEB Color. */
	sprintf(buffer2, "Old-style WEB Color.");
	break;

      case 3:			/* WP6 Color. */
	sprintf(buffer2, "New-style WP6/WEB Color.");
	break;
    }
    MoveTo(i, j);
    j += 20;
    DrawString(buffer);
    DrawString(buffer2);
}
