/* Virtual Terminal Connection control block structure definition */

#include <dnix/itrq.h>

typedef	unsigned	char	BYTE;
typedef	unsigned	short	WORD;

/* define terminal Characteristics */
#define	MAXWINDOWS	10
#define	MAXROWS	24
#define	MAXCOLS	80
#define	MAXKEYS	256
#define	MAXNAMELEN	64
#define	MAXSPACES	4
#define MAXFONTNUMBER 32

/* define font Characteristics ( Must be fixed pitch font )*/
#define	FONTNUMBER	19
#define	CHARHIGH	13
#define	CHARWIDE	8

/* define window Characteristics */
#define	TWINTOP		40
#define	TWINLEFT	0

/* define offset of display within window ( this defines margin between Box and data ) */
#define	TOFFSET		2
#define LOFFSET		0
#define	BOFFSET		3

/* define offset between multiple terminal windows */
#define	VOFFSET		20
#define	HOFFSET		0


#define	FALSE	0
#define	TRUE	1
#define	UP		0
#define	DOWN	1
#define	BUCKETS	5
#define	NOCODE	0xFF

#ifdef	DEBUG
extern	int	debugsw;
#define	LOG(s,i)	if ( debugsw ) { printf(s,(i)),fflush(stdout);}
#else
#define	LOG(s,i)
#endif

#define	Setup_Menu 1
#define	AutoWrap	1
#define	Popup_Scroll	2
#define	Cursor_Blink	3
#define	Mouse_Input	4
#define Intl_Keyboard	5
#define NiceFlag	6
#define Other		7
#define History		8
#define	Clear		10
#define	Reset		11
#define Hide		12
#define Stty_Sane	14

#define	Edit_Menu 2
#define	CopyItem	1
#define PasteItem	2

struct readrq {					/* Structure of a read request. */
	struct readrq *next;
	WORD readcheck;				/* Flag, is readrn a NW read chk? */
	WORD readrn;				/* Request # that's waiting on empty queue. */
	int	 readbc;				/* Byte count for read buffer. */
	WORD pid;					/* Only 1 read request per pid. */
};

struct connect {		/* Structure of an extra connection (copy). */
    struct connect *next;
    char *nodename;
    char connectid;
    char refcount;
};

struct vidcopy {		/* Structure of a video [hot] copy. */
    struct vidcopy *next;
    struct connect *connect;
    Rect wrect;			/* Global location. */
    char windowid;
    char marker;
};

struct vtcb {
	int		(*nap_routine)();		/* pointer to timeout handler ( MUST BE FIRST ) */
	WORD	naprn;					/* request number of nap for cancel */
	WORD	blinkrn;				/* request number of nap for cursor blink */
	WORD	connect;				/* open vtcb table index */
	WORD	flag;					/* open / active flags */
	WORD	lurk;					/* leave window open until close event */
	WORD	iconified;				/* TRUE when window's iconified. */
	short	windowid;				/* window ID in ccb */
	struct readrq *readrq;		/* Read request queue. */
	struct vidcopy *vidcopy;
	long ports;
	WORD	pid;					/* process id of connected task */
	WORD	writern;				/* Request # of pending write. */
	int		writebc;				/* byte count for pending write */
	short	row;					/* line number on screen */
	short	col;					/* column number */
	short	mrow;					/* Maximum line number on screen */
	short	mcol;					/* Maximum column number */
	WORD	att;					/* current write attribute */
	BYTE	CharMask;				/* BYTE to xor for SO / SI simulation */
	BYTE	SOMask;					/* BYTE to xor for SO  */
	BYTE	SIMask;					/* BYTE to xor for SI  */
	BYTE	filler;
	WORD	font;					/* current selected font */
	WORD	fontsize;				/* current selected font size */
	short refcount;				/* For multiple-open references. */
	FontInfo fontinfo;			/* Metrics of selected font */
	short	charwide;
	short	charhigh;
	short	charoffset;
	short	histlines;			/* Number of saved scroll history lines. */
	Rect	wrect;
	struct	vtmodebits	mode;		/* terminal mode flags ( autowrap etc. ) */
	struct	vtmodebits	save_mode;	/* terminal mode flags ( autowrap etc. ) */
	short	save_row;
	short	save_col;
	WORD	save_att;
	WORD	charwidth;				/* single / double BYTE width */
	void	(*state)();				/* routine for handling next Character */
	short	trow;
	short	brow;					/* scrolling region */
	char	tname[MAXNAMELEN];		/* name of window on screen */
	struct	termio	tty;			/* ioctl structure for virtual term */
	int		buckets[BUCKETS];		/* parameter buckets for ANSI commands */
	int 	*bucket;				/* current bucket */
	int		bcount;					/* count of number of filled buckets  */
	int		keycount;				/* number of key strokes in buffer */
	int		keyinp;					/* index of first free space in buffer */
	int		keyoutp;				/* output pointer for buffer */
	BYTE	*keybuffer;				/* buffer for keystrokes */
	int		maxkeys;				/* Size of key buffer. */
	BYTE	*display;				/* pointer to display plane */
	BYTE	*atts;					/* pointer to attribute plane */
	BYTE	*history;			/* Pointer to scrolled-off-top stuff. */
	short startrow;				/* Used to mark range for COPY. */
	short startcol;
	short endrow;
	short endcol;
	short lastrow, lastcol;		/* For remembering highlighting stages. */
};

struct nwblock {				/* Indirect pointer to completion routine */
	void (*iodone)();
	struct itrq trapnode;
	struct ireq request;		/* Handler's function request. */
};

#define	VT_INIT	1
#define	VT_ACTIVE 2

#define	Posn(x,y)	MoveTo((x)*Term->charwide+Term->charoffset, (y)*Term->charhigh+TOFFSET-Term->fontinfo.descent );
