/*
** Handle requests from the outside world (the fake tty connection).
*/

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dnix/errno.h>
#include <dnix/fcodes.h>
#include <dnix/ireq.h>
#include <termio.h>
#include <sys/ttold.h>
#ifndef _SYS5R3
#define _SYS5R3
#endif
#include <sys/signal.h>
#include <string.h>
#include <isc/window.h>
#include <isc/vterm.h>
#include "vtcb.h"
#include "protos.h"

#ifndef F_TESTTQ
#define F_TESTTQ 121		/* Trap-Q check opcode. */
#endif
#ifndef NWREADCHK
#define NWREADCHK (('p'<<8)|100) /* Should be in a header file. */
#endif
#ifndef TIOCGWINSZ
#define TIOCGWINSZ (TIOC|104)
#define TIOCSWINSZ (TIOC|103)
struct winsize {
	unsigned short ws_row;		/* rows, in characters */
	unsigned short ws_col;		/* columns, in characters */
	unsigned short ws_xpixel;	/* horizontal size, pixels */
	unsigned short ws_ypixel;	/* vertical size, pixels */
};
#endif

extern int Popup;
extern int Autowrap;
extern int Not_vt_ansi;
extern int CursorBlink;
extern int MouseInput;
extern int KeyFilter;
extern int trapqfd;
extern char basename[];
extern char *Device, *path;
extern int	Rows;
extern int	Columns;
extern int	Font;
extern int	Size;
#ifdef DEBUG
int errno;
#endif
static short hoffset = HOFFSET;	/* Initial window-to-window offsets. */
static short voffset = VOFFSET;
static short twintop = TWINTOP;
static short twinleft = TWINLEFT;

struct vtcb *connections[MAXWINDOWS]; /* Pointers to connection control blocks. */
struct connect *chead;
struct stat mdstat;				/* Mount directory status. */
int mountfd;					/* Mount device file number. */
int mounttq;					/* Mount device trap queue. */
struct nwblock handlerfunc = {tqdone};

#define BUFFERSIZE 2048
BYTE buffer[BUFFERSIZE];
static short bufpos;
static short deferred;
static char copystr[] = "Copy of ";

void
handlerinit()
{
    int status;
    char *p;

	if ((mounttq = rdnix(F_OTQ)) < 0) {
		LOG("Unable to open trap queue (error %d).\n", errno);
		exit(3);
	}
    if (stat(Device, &mdstat) < 0) { /* Stat the mount directory. */
	LOG("Stat error on mount directory errno=%d\n", errno);
	exit(4);
    }
    if (!(mdstat.st_mode & S_IFDIR)) {
	LOG("Error : mount directory in use\n", 0);
	exit(5);
    }

    status = rdnix(F_MOUNT, LU_ROOT, Device, StringLen(Device), 0, 0);
    if (status < 0) {		/* Now mount us on the mount device. */
	LOG("Mount error %d\n", status);
	exit(6);
    }

    mountfd = status;		/*  Make us look like a char special device. */
    mdstat.st_mode = S_IFCHR | 0666;
    mdstat.st_nlink = 0;
    mdstat.st_size = 0;
    time(&mdstat.st_ctime);	/* set creation / change time to mount time */
    mdstat.st_atime = mdstat.st_ctime;

    p = Device;			/* Find the end of the device string */
    while (*p++)		/* and attempt to atoi() the digits */
	;			/* (if any) at the end. */
    p--;			/* (Point to NULL.) */
    while (*--p >= '0' && *p <= '9')
	;			/* We'll use this for a minor device number. */
    if (*++p)			/* Any numerics at the end? */
	status = myatoi(p) & 0xFF; /* Use them for an 8-bit minor number. */
    else			/* No digits, use 128 as minor number. */
	status = 128;
    mdstat.st_rdev = (129 << 8) | status; /* 129 is major number. */
    
    handlerread();		/* Now begin handling requests. */
	tqread();
}

/*
** Read the mount point's trap queue.  We stack this trap queue on
** the normal one so that we can effectively poll the mount point
** for further requests while we're processing one already.  This
** allows us to combine the data from a series of writes, for better
** overall performance.  (We can't poll a mount point, but we _can_
** poll a trap queue.)
*/

void
tqread()
{
	int i;

	i = rdnix(F_READ | F_NOWAIT, mounttq, &handlerfunc.trapnode,
			  sizeof(struct itrq), 0, 0, trapqfd, &handlerfunc);
	LOG("Posted read of 2nd trap queue, reqnum %d.\n", i);
}

/*
** Read of mount point's trap queue has completed.  Call (in turn) the
** mount point request handler.  Arcane handler nasty:  Never, never
** post another read to the mount point before you've terminated the
** last one, as the OS can mix the two requests together (with less
** than useful results) if a second request of the same type from the
** same place comes in before the first is terminated).
*/

void
tqdone(hndl, it_rpar, it_rn)
struct nwblock *hndl;
int it_rpar, it_rn;
{
	int rpar, rn;
	struct nwblock acopy;

	LOG("2nd Trap queue read #%d ", it_rn);
	LOG("completed (%d).\n", it_rpar);
	acopy = *hndl;				/* So the next read doesn't clobber! */
	rpar = acopy.trapnode.it_rpar;
	rn = acopy.trapnode.it_rn;
	if (it_rpar == sizeof(struct itrq)) {
		if (handlerdone(&acopy, rpar, rn))
			handlerread();		/* Post another read. */
	} else {
		LOG("Bad trap queue read (%d).\n", it_rpar);
	}
	LOG("2nd Trap queue read #%d processing completed.\n", it_rn);
}

/*
** Check to see if we've got another WRITE waiting for the buffer we're
** processing, and that it will fit in the buffer.  If so, append it to
** the buffer before continuing processing.  Before checking for more
** input, we give up the CPU to other processes at this level (presumably
** our client is one of these).
*/

void
poll(term)
struct vtcb *term;
{
	struct vtcb *vtcb;
	int i;

	if (!deferred &&	/* Any more requests from the client? */
	    (rdnix(F_GIVEUPCPU), rdnix(F_TESTTQ, mounttq))) {
		LOG("Poll found something.\n", 0);
		vtcb = (struct vtcb *) handlerfunc.request.ir_opid;
		if ((handlerfunc.request.ir_fc != F_WRITE) || invalid_vtcb(vtcb)
			|| (vtcb != term)
			|| ((handlerfunc.request.ir_bc + bufpos) > sizeof(buffer))) {
			LOG("Deferring request found by the poll (%s).\n", 
				(handlerfunc.request.ir_fc != F_WRITE) ? "Not a write" :
				(invalid_vtcb(vtcb) ? "Invalid vtcb" :
				((vtcb != term) ? "For wrong vtcb" : "Too much data")));
			deferred = 1;
		} else {
			LOG("Doing poll-ahead of %d bytes.\n", handlerfunc.request.ir_bc);
			i = rdnix(F_READ, mounttq, &handlerfunc.trapnode,
					  sizeof(struct itrq), 0, 0, 0, 0);
			LOG("(Explicit read of 2nd trap queue, rpar %d, ",
				handlerfunc.trapnode.it_rpar);
			LOG("status %d.)\n", handlerfunc.trapnode.it_stat);
			tqdone(&handlerfunc, i, 0);
			LOG("Finished poll-ahead.\n", 0);
		}
	}
}

/*
** Read the mount point's request.
*/

void
handlerread()
{
	int i;

	i = rdnix(F_READ | F_NOWAIT, mountfd, &handlerfunc.request, 
			  sizeof(struct ireq), 0, 0, mounttq, &handlerfunc);
	LOG("Posted read of handler mount point, reqnum %d.\n", i);
}

/*
** Check to see if a vtcb pointer is for a valid vtcb.
*/

int
invalid_vtcb(vtcb)
struct vtcb *vtcb;
{
	int i;

	for (i = 0; i < MAXWINDOWS; i++) {
		if (vtcb && connections[i] == vtcb)
			return 0;
	}
	return 1;
}

int
checkprot(hndl)
struct nwblock *hndl;
{
	return !hndl->request.ir_uid || hndl->request.ir_uid == mdstat.st_uid;
}

/*
** Terminate handler request.
*/

void
termhreq(reqnum, status, mode)
int reqnum;
long status;
int mode;
{
	(void) rdnix(F_TERMIN, reqnum, status, mode);
}

/*
** A new read has been requested.  Only one read request per PID is allowed
** to prevent errors from blowing vterm's storage out of the water.  (This
** preserves old behavior --- vterm's always been snotty this way.)  If the
** new request is accepted, put it in the front of the chain.  This lets
** 'recent' activity grab the terminal.
*/

newrd(vtcb, hndl, check)
struct vtcb *vtcb;
struct nwblock *hndl;
int check;
{
	int pid;
	struct readrq *rq;

	pid = (hndl->request.ir_nph >> 16) & 0xFFFF;
	rq = vtcb->readrq;
	while (rq) {				/* Only allow one read/readcheck per PID. */
		if (rq->pid == pid) {
			LOG("New read rq failed:  PID %d was greedy!\n", pid);
			return 0;
		}
		rq = rq->next;
	}
	if (!(rq = (struct readrq *) malloc(sizeof(*rq)))) {
		LOG("New read rq failed:  Memory allocation failure!\n", 0);
		return 0;
	}
	rq->readrn = hndl->request.ir_rn; /* Make a new request. */ 
	rq->readbc = hndl->request.ir_bc;
	rq->pid = pid;
	rq->readcheck = check;		/* Read versus readcheck. */
	rq->next = vtcb->readrq;
	vtcb->readrq = rq;			/* Stick it in the front of the queue. */
	LOG("New read rq, PID %d ", pid);
	LOG("reqnum %d\n", rq->readrn);
	return 1;
}

/*
** Terminate a read request.
*/

void
termrdreq(vtcb, reqnum, rpar)
struct vtcb *vtcb;
int reqnum, rpar;
{
	struct readrq *rq, *rq2;

	rq = vtcb->readrq;
	rq2 = 0;
	while (rq) {				/* Find it in the queue. */
		if (rq->readrn == reqnum) {
			LOG("Read RN %d\n", reqnum);
			if (!rq2)			/* Unlink it from the chain. */
				vtcb->readrq = rq->next;
			else
				rq2->next = rq->next;
			free(rq);			/* Throw it away. */
		  termanyway:
			termhreq(reqnum, rpar, 0); /* Terminate it. */
			return;
		}
		rq2 = rq;
		rq = rq->next;
	}
	goto termanyway;			/* Not found, terminate it anyway. */
}

/*
** Read the request done.  This routine is re-entrant at times (during
** polling for extra output to append to a request).
*/

int
handlerdone(hndl, it_rpar, it_rn)
struct nwblock *hndl;
int it_rpar, it_rn;
{
	register struct vtcb *vtcb;
	int i, status, namelen, pid, reqnum, retval;
	char *src, *dst;
	struct vidcopy *vcp;
	static short entrycount;
	extern short newcopy;
	extern int vt_cnt;

	/*
	** Process the request.
	*/
	retval = 1;
	pid = (hndl->request.ir_nph >> 16) & 0xFFFF;
	LOG("Mount point read %d completed", it_rn);
	LOG(" (%d).\n", it_rpar);
	LOG("Handler request %d ", hndl->request.ir_rn);
	LOG("func: %d ", hndl->request.ir_fc);
	LOG("opid: $%08X ", hndl->request.ir_opid);
	LOG("upar: $%08X ", hndl->request.ir_upar);
	LOG("bc: %d, ", hndl->request.ir_bc);
	LOG("pid: %d\n", pid);

	entrycount++;
	LOG("Entry level: %d.\n", entrycount);
	vtcb = (struct vtcb *) hndl->request.ir_opid;
	if ((i = hndl->request.ir_fc) == F_READ || i == F_WRITE
		|| i == F_CANCEL || (i == F_IOCW && hndl->request.ir_upar != SF_SFTM)
		|| (i == F_IOCR && hndl->request.ir_upar != SF_STAT)) {
		if (invalid_vtcb(vtcb))	{ /* Valid vtcb? */
			LOG("Invalid vtcb ($%X) ", vtcb);
			LOG("for function %d.\n", i);
			hndl->request.ir_fc = -i;	/* Force EINVAL response later. */
		}
	}
			
	reqnum = hndl->request.ir_rn;
	switch (hndl->request.ir_fc) {
	case F_LOCATE:
		LOG("Locate Request %d terminated.\n", reqnum);
		termhreq(reqnum, NULL, 0);
		break;					/* Terminate LOCATE request. */

	case F_CREATE:
		time(&mdstat.st_atime);
	case F_OPEN:
		if (mdstat.st_nlink < MAXWINDOWS) {
			status = 0;			/* No permissions granted at first. */
			i = hndl->request.ir_upar & 3; /* Only look at R/W bits. */
			if (i == AM_READ || i == AM_REWR) { /* Check read permissions. */
				if (checkprot(hndl)) {
					if (!hndl->request.ir_uid || (mdstat.st_mode & 0400))
						status |= OP_READ;
				} else if (hndl->request.ir_gid == mdstat.st_gid) {
					if (mdstat.st_mode & 0040)
						status |= OP_READ;
				} else if (mdstat.st_mode & 0004)
					status |= OP_READ;
			}
			if (i == AM_WRITE || i == AM_REWR) { /* Check write perms. */
				if (checkprot(hndl)) {
					if (!hndl->request.ir_uid || (mdstat.st_mode & 0200))
						status |= OP_WRIT;
				} else if (hndl->request.ir_gid == mdstat.st_gid) {
					if (mdstat.st_mode & 0020)
						status |= OP_WRIT;
				} else if (mdstat.st_mode & 0002)
					status |= OP_WRIT;
			}
			if (!status) {		/* Permission denied. */
				LOG("Open Request %d terminated (denied).\n", reqnum);
				termhreq(reqnum, EPERM, 0);
				break;
			}
			vtcb = (struct vtcb *) malloc(sizeof(struct vtcb));
			memset(vtcb, 0, sizeof(struct vtcb));
			vtcb->refcount = 1;
			vtcb->keybuffer = (BYTE *) malloc(vtcb->maxkeys = MAXKEYS);
			LOG("Open request, vtcb = 0x%X", vtcb);
			vtcb->pid = pid;	/* Record the PID of the requester. */
			LOG(" pid = %d\n", vtcb->pid);
			vtcb->lurk = 1;
			for (i = 0; i < MAXWINDOWS; i++) {
				if (connections[i] == NULL) {
					connections[i] = vtcb;
					vtcb->connect = i;
					break;
				}
			}
			mdstat.st_nlink++;
			Reset_termio(vtcb);
			vtcb->font = Font;	/* To be replaced by parameterised stuff. */
			vtcb->mrow = Rows;
			vtcb->mcol = Columns;
			vtcb->fontsize = Size;
			vtcb->wrect.edge.left = twinleft + (vtcb->connect * hoffset);
			vtcb->wrect.edge.top = twintop + (vtcb->connect * voffset);
			namelen = StringLen(basename);
			(void) strcpy(vtcb->tname, basename);
			(void) strcat(vtcb->tname, "vtN");
			vtcb->tname[StringLen(vtcb->tname)-1] = '0' + (char) vtcb->connect;
			vtcb->mode.autowrap = Autowrap?TRUE:FALSE;
			vtcb->mode.cursorblink = CursorBlink?TRUE:FALSE;
			vtcb->mode.pop_scroll = Popup?TRUE:FALSE;
			vtcb->mode.not_vt_ansi = Not_vt_ansi?TRUE:FALSE;
			vtcb->mode.mouse_input = MouseInput?TRUE:FALSE;
			vtcb->mode.intl_kbd_filter = KeyFilter?TRUE:FALSE;
			vtcb->startrow = -1;	/* Mark highlighting as gone. */
			if (i = hndl->request.ir_bc) {
				if (i >= sizeof buffer) /* Get user parameters and override */
					i = sizeof buffer - 1; /* the defaults. */
				i = rdnix(F_UREAD, reqnum, buffer, i, 0, 0, 0);
				buffer[i] = 0;
				LOG("Open parameter string = %s\n", buffer);
				src = (char *) buffer;
				while (*src && src < (char *) &buffer[sizeof buffer]) {
					while (*src && *src == '/')	/* Skip over slashes. */
						src++;
					i = namelen;
					switch (*src++) {
					  case 0x0E: /* Control-N, slash delimited window name. */
						dst = &vtcb->tname[i];
						i = MAXNAMELEN-1-i;
						while (*src && *src != '/' && i--)
							*dst++ = *src++;
						*dst = 0;
						break;

					  case 'N':
						i = 0;	/* Fall into 'n' case. */
					  case 'n':
						(void) strncpy(&vtcb->tname[i], src, MAXNAMELEN-1-i);
						*src-- = 0; /* Force termination of arguments. */
						*src = '/';
						break;

					  case 'x':
						if ((i = myatoi(src)))
							vtcb->wrect.edge.left = i;
						else if (*src == '*') {
							hoffset = myatoi(&src[1]);
						  redox:
							vtcb->wrect.edge.left = twinleft +
								(vtcb->connect * hoffset);
						} else if (*src == '+')
							vtcb->wrect.edge.left += myatoi(&src[1]);
						else if (*src == '!') {
							twinleft = myatoi(&src[1]);
							goto redox;
						}
						break;

					  case 'y':
						if ((i = myatoi(src)))
							vtcb->wrect.edge.top = i;
						else if (*src == '*') {
							voffset = myatoi(&src[1]);
						  redoy:
							vtcb->wrect.edge.top = twintop +
								(vtcb->connect * voffset);
						} else if (*src == '+')
							vtcb->wrect.edge.top += myatoi(&src[1]);
						else if (*src == '!') {
							twintop = myatoi(&src[1]);
							goto redoy;
						}
						break;

					  case 'r':
						if ((i = myatoi(src))) vtcb->mrow = i;
						break;

					  case 'c':
						if ((i = myatoi(src))) vtcb->mcol = i;
						break;

					  case 'f':
						if ((i = myatoi(src))) vtcb->font = i;
						break;

					  case 's':
						if ((i = myatoi(src))) vtcb->fontsize = i;
						break;

					  case 'l':
						vtcb->lurk = 0;
						break;


					  case 'P':
						if (*src != '0')
							vtcb->mode.pop_scroll = TRUE;
						else
							vtcb->mode.pop_scroll = FALSE;
						break;

					  case 'A':
						if (*src != '0')
							vtcb->mode.autowrap = TRUE;
						else
							vtcb->mode.autowrap = FALSE;
						break;

					  case 'B':
						if (*src != '0')
							vtcb->mode.cursorblink = TRUE;
						else
							vtcb->mode.cursorblink = FALSE;
						break;

					  case 'X':
						if (*src != '0')
							vtcb->mode.not_vt_ansi = TRUE;
						else
							vtcb->mode.not_vt_ansi = FALSE;
						break;

					  case 'M':
						if (*src != '0')
							vtcb->mode.mouse_input = TRUE;
						else
							vtcb->mode.mouse_input = FALSE;
						break;

					  case 'K':
						if (*src != '0')
							vtcb->mode.intl_kbd_filter = TRUE;
						else
							vtcb->mode.intl_kbd_filter = FALSE;
						break;

					  case 0:
						src--;	/* Run off the end, force termination. */
						break;
					}
					while (*src && *src != '/')	/* Find next argument. */
						src++;
				}
			}
			strncpy((char *) buffer, path, sizeof buffer);
			if (src = strchr(buffer, '\n')) {
			    hotcopies(vtcb, src +1);
			}
		} else {		/* Too many connections. */
			vtcb = (struct vtcb *) EBUSY;
		}
								/* Terminate OPEN/CREATE request. */
		LOG("Open Request %d terminated.\n", reqnum);
		termhreq(reqnum, (long) vtcb,
			 status | OP_NRAD | OP_STRM | OP_APPD);
		break;

	case F_READ:		/* Keyboard request. */
		LOG("Read request vtcb = 0x%X", vtcb);
		LOG(" Request #%d\n", reqnum);
		if (!(vtcb->flag & VT_INIT))
			init_vt220(vtcb);
		if (newrd(vtcb, hndl, 0)) {
			if (vt220_rdchk(vtcb) == FALSE) {
				if (hndl->request.ir_upar & (O_NDELAY | O_NONBLOCK)) {
					LOG("Terminating O_NDELAY/O_NONBLOCK read request #%d with 0/EAGAIN\n",
						reqnum);
					termrdreq(vtcb, reqnum,
							  hndl->request.ir_upar & O_NONBLOCK ? EAGAIN : 0);
				} else if (!(vtcb->tty.c_lflag & ICANON)) {	/* Raw mode? */
					if (!vtcb->tty.c_cc[VMIN]) {
						if (!vtcb->tty.c_cc[VTIME]) { /* No timer, immediate */
							termrdreq(vtcb, reqnum, 0); /* return. */
							LOG("No data available, terminating RN #%d.\n", 
								reqnum);
						} else { /* We want a timer started now, not later. */
							vtcb->naprn = rdnix(F_NAP|F_NOWAIT,
												vtcb->tty.c_cc[VTIME] * 100,
												0, 0, 0, 0, trapqfd, vtcb);
							LOG("Timer started vtcb = %X", vtcb);
							LOG(" rn = %d\n", vtcb->naprn);
						}
					}
				}
			}
		} else {
			LOG("Terminating %d, read channel already busy.\n", reqnum);
			termhreq(reqnum, EBUSY, 0);
		}
		break;

	case F_WRITE:		/* Process terminal's character sequences. */
		LOG("Write request vtcb = 0x%X", vtcb);
		LOG(" Request #%d\n", reqnum);
		if (!(vtcb->flag & VT_INIT))
			init_vt220(vtcb);
		time(&mdstat.st_mtime);
		if (vtcb->mode.stopped) {
			if (vtcb->writern == 0) {
				vtcb->writern = reqnum;
				vtcb->writebc = hndl->request.ir_bc;
				LOG("Request %d Held pending XON\n", vtcb->writern);
				break;
			} else {
				status = EBUSY;
			}
		} else {				/* Process the data, */
			status = hndl->request.ir_bc;
			namelen = 0;		/* (user buffer offset) */
			while (i = status) { /* in pieces if necessary. */
				if (i > (sizeof(buffer) - bufpos))
					i = (sizeof(buffer) - bufpos);
				rdnix(F_UREAD, reqnum, &buffer[bufpos], i, namelen, 0, 0);
#if 0
				{
				    static int logdata = -1; char buf[20];
				    if (logdata < 0) {
					sprintf(buf, "/tmp/vti%d", getpid());
					logdata = creat(buf, 0666);
				    }
				    if (logdata >= 0)
					write(logdata, &buffer[bufpos], i);
				}
#endif
				status -= i;	/* (status == 0 when done.) */
				namelen += i;	/* (namelen == ir_bc when done.) */
				bufpos += i;
				if (entrycount == 1) {
					if (!status) { /* All read?  Release client early. */
						termhreq(reqnum, namelen, 0);
						handlerread(); /* Post another read. */
						LOG("Write Request #%d terminated (early)", reqnum);
						LOG(" with status %d.\n", namelen);
						reqnum = 0;	/* Don't terminate twice. */
						retval = 0;	/* Don't post yet another read. */
					} else {
						LOG("Processing incomplete client buffer.\n", 0);
						deferred = 1; /* Can't accept more until this */
					}			/* request is fully consumed & terminated. */
					vt220(vtcb, &buffer[bufpos-i], i); /* Do the work. */
					bufpos = 0;
				} else {
					LOG("Adding %d bytes to buffer, ", i);
					vt_cnt += i; /* Re-entered, add it to the list. */
					LOG("now has %d bytes ", bufpos);
					LOG("(%d unconsumed).\n", vt_cnt);
				}
			}
		}
		if (reqnum) {
			termhreq(reqnum, status ? status : namelen, 0);
			LOG("Write Request #%d terminated", reqnum);
			LOG(" with status %d.\n", status ? status : namelen);
		}
		if (vtcb->iconified == 1) {
			vtcb->iconified = 3;
			if (vtcb->vidcopy)
			    SetPorts(vtcb->ports);
			else
			    SetPort(vtcb->windowid);
			draw_icon(vtcb);
		}
		break;

	case F_CLOSE:		/* A graphics process is ending. */
		LOG("Close request vtcb = 0x%X\n", vtcb);
		if (vtcb && !invalid_vtcb(vtcb) && !--vtcb->refcount) {
			if (vtcb->flag & VT_INIT && vtcb->lurk) {
			    if (vtcb->pid) {
			/*
			** Orphan the window and change the title to let the user know.
			*/
					LOG("Close will orphan window %d\n", vtcb->windowid);
					vtcb->pid = 0;
					namelen = StringLen(vtcb->tname);
					(void) strncat(vtcb->tname,
								   " \321 (Orphaned)", MAXNAMELEN-1-namelen);
					SetPort(vtcb->windowid);
					SetWTitle(vtcb->windowid, vtcb->tname);
					if (vtcb->flag & VT_ACTIVE)
						VCursor(vtcb, FALSE);
					vtcb->mode.cursoron = FALSE; /* leave it off!!! */
			    }
			} else {		/* Get rid of the window entirely. */
				LOG("Close will dispose of window %d\n", vtcb->windowid);
				i = vtcb->windowid;
				dispose_vt220(vtcb);
				if (newcopy) {
					SetPort(i);
					writescrap();
				}
			}
		}
		break;					/* Closes are not to be terminated. */

	case F_CANCEL:
		LOG("Cancel request vtcb = 0x%X ", vtcb);
		if (vtcb) {
			termrdreq(vtcb, reqnum, EINTR);
			if (vtcb->writern == reqnum) {
				LOG("Write RN %d\n", reqnum);
				termhreq(reqnum, EINTR, 0);
				vtcb->writern = 0;
			}
		}
		break;					/* Cancels are not to be terminated. */

	case F_CUID:
		status = 0;
		if (checkprot(hndl)) {
			mdstat.st_uid = hndl->request.ir_upar;
			time(&mdstat.st_ctime);
		} else
			status = EPERM;
		termhreq(reqnum, status, 0);
		LOG("F_UID Request %d terminated\n", reqnum);
		break;

	case F_CGID:
		status = 0;
		if (checkprot(hndl)) {
			mdstat.st_gid = hndl->request.ir_upar;
			time(&mdstat.st_ctime);
		} else
			status = EPERM;
		termhreq(reqnum, status, 0);
		LOG("F_GID Request %d terminated\n", reqnum);
		break;

	case F_CPROT:
		status = 0;
		if (checkprot(hndl)) {
			mdstat.st_mode &= S_IFMT;
			mdstat.st_mode |= (hndl->request.ir_upar & ~S_IFMT);
			time(&mdstat.st_ctime);
		} else
			status = EPERM;
		termhreq(reqnum, status, 0);
		LOG("F_CPROT Request %d terminated\n", reqnum);
		break;

	case F_IOCR:
	case F_IOCW:
		status = 0;
		switch (hndl->request.ir_upar) {

		case NWREADCHK:
			LOG("ioctl: NWREADCHK vtcb=%x\n", vtcb);
			if (!newrd(vtcb, hndl, 1)) /* PID already busy? */
				status = EBUSY;	/* Fall into regular read check code. */
		case SF_RDCHK:
			if (!(vtcb->flag & VT_INIT))
				init_vt220(vtcb);
			LOG("ioctl: SF_RDCHK vtcb=0x%X\n", vtcb);
			if (!status) {
				status = count_ready(vtcb, 0);
				if (status) {
					if (!(vtcb->tty.c_lflag & ICANON)) { /* Raw -- enough? */
						if (status < vtcb->tty.c_cc[VMIN])
							status = 0;	/* No, bag it. */
					}
				}
			}
			if (hndl->request.ir_upar == NWREADCHK) {
				if (status)		/* Something there, terminate it now. */
					termrdreq(vtcb, reqnum, status);
				goto done;		/* Otherwise leave pending readcheck. */
			}
			break;

		case SF_SYNC:
			break;

		case SF_STAT:
			rdnix(F_UWRITE, reqnum, &mdstat, sizeof(mdstat), 0, 0, 0, 0);
			status = sizeof(mdstat);
			break;

		case SF_CTER:			/* OS asking whether we can be contrl term.  */
			vtcb->pid = pid;	/* (New process group leader.) */
			LOG("ioctl: SF_CTER vtcb = 0x%X\n", vtcb);
			break;				/* Yes. */

		case SF_SFTM:			/* Change times.  */
			if (checkprot(hndl)) {
				rdnix(F_UREAD, reqnum, &mdstat.st_atime, 2 * sizeof(time_t),
					  0, 0, 0, 0);
				time(&mdstat.st_ctime);
				status = 2 * sizeof(time_t);
			} else {
				status = EPERM;
			}
			LOG("ioctl: SF_SFTM vtcb = 0x%X\n", vtcb);
			break;

		case TCGETA:
			LOG("ioctl: TCGETA vtcb = 0x%X\n", vtcb);
			rdnix(F_UWRITE, reqnum, &vtcb->tty, 
				  hndl->request.ir_bc, 0, 0, 0, 0);
			status = sizeof(struct termio);
			break;

		case TCSETAF:			/* TCSETA plus flush of input queue. */
			vtcb->keycount = vtcb->keyinp = vtcb->keyoutp = 0;
		case TCSETA:
		case TCSETAW:
			LOG("ioctl: TCSETA vtcb = 0x%X\n", vtcb);
			i = hndl->request.ir_bc;
			if (i > sizeof vtcb->tty)
				i = sizeof vtcb->tty;
			rdnix(F_UREAD, reqnum, &vtcb->tty, i, 0, 0, 0);
			if (vtcb->tty.c_oflag & OPOST && vtcb->tty.c_oflag & ONLCR)
				vtcb->mode.autocr = TRUE;
			else
				vtcb->mode.autocr = FALSE;
			if (!(vtcb->tty.c_cflag & CBAUD)) {
				(void) kill(-vtcb->pid, SIGHUP); /* Hang up magic cookie, */
				(void) kill(vtcb->pid, SIGHUP);	/* signal process & group. */
			}
			break;

		case TIOCGETP:
#if 0
			LOG("ioctl: TIOCGETP vtcb = 0x%X\n", vtcb);
			simsgttyb(vtcb, reqnum);
			status = sizeof(struct sgttyb);
#endif
			break;

		case TIOCSETP:
		case TCSBRK:
			break;

		case TCFLSH:
			(void) rdnix(F_UREAD, reqnum, &status, sizeof(status),
						 0, 0, 0);
			if (!status || status == 2)	/* Flush input queue? */
				vtcb->keycount = vtcb->keyinp = vtcb->keyoutp = 0;
			status = 0;
			break;

		case TCXONC:
			(void) rdnix(F_UREAD, reqnum, &status, sizeof(status), 0, 0, 0);
			i = vtcb->mode.stopped;
			if (vtcb->flag & VT_INIT) {
				SetKbdLEDs(vtcb->mode.stopped);
				if (!(vtcb->mode.stopped = status ? FALSE : TRUE))
					removespan(vtcb);
				if (i && status && vtcb->writern) { /* Release held request? */
					restartflow(vtcb);
				}
			}
			status = 0;
			break;

		case VTERM_NOLURK:		/* Make window go away upon death. */
			vtcb->lurk = 0;
			break;

		case VTERM_LURK:		/* Make window stay upon death. */
			vtcb->lurk = 1;
			break;

		case VTERM_HEREBOY:		/* Make window come to front. */
			if (vtcb->flag & VT_INIT)
				SelectWindow(vtcb->windowid);
			break;

		case VTERM_HIDE:		/* Make window go to back. */
			if (vtcb->flag & VT_INIT)
				SendBehind(vtcb->windowid, 0);
			break;

		case VTERM_WSTATS:		/* Return window stats. */
			{
				struct vterm_window owind, nwind;

				i = hndl->request.ir_bc;
				if (i > sizeof(nwind))
					i = sizeof(nwind);
				owind.rows = vtcb->mrow;
				owind.cols = vtcb->mcol;
				owind.vtmode = vtcb->mode;
				owind.font = vtcb->font;
				owind.fontsize = vtcb->fontsize;
				owind.iconified = vtcb->iconified;
				owind.wpos = vtcb->wrect;
				nwind = owind;
				rdnix(F_UWRITE + hndl->request.ir_fc - F_IOCR, reqnum, &nwind,
					  i, 0, 0, 0, 0); /* Read/write the structure. */
				status = sizeof(nwind); /* Changed? */
				i = nwind.iconified;
				nwind.iconified = owind.iconified;
				if (memcmp(&owind, &nwind, sizeof(owind))) {
					vtcb->mrow = nwind.rows;
					vtcb->mcol = nwind.cols;
					vtcb->mode = nwind.vtmode;
					vtcb->font = nwind.font;
					vtcb->fontsize = nwind.fontsize;
					vtcb->wrect.edge.left = nwind.wpos.edge.left;
					vtcb->wrect.edge.top = nwind.wpos.edge.top;
					if (vtcb->flag & VT_INIT) {
						close_vt220(vtcb);
						init_vt220(vtcb);
						new_history(vtcb);
#ifdef SIGWINCH
						kill(-vtcb->pid, SIGWINCH);
						kill(vtcb->pid, SIGWINCH);
#endif
					}
				} else if (owind.iconified != i) {
				    if (vtcb->flag & VT_INIT) {
						if (i)
							doiconify(vtcb);
						else
							douniconify(vtcb);
				    }
				}
			}
			break;

		case VTERM_RENAME:		/* Rename the window. */
			strcpy((char *)buffer, copystr);
			src = (char *) buffer + sizeof copystr - 1;
			i = hndl->request.ir_bc;
			if (i > sizeof buffer - sizeof copystr)
				i = sizeof buffer - sizeof copystr;
			(void) rdnix(F_UREAD, reqnum, src, i, 0, 0, 0);
			strncpy((char *)vtcb->tname, src, sizeof(vtcb->tname));
			if (vtcb->flag & VT_INIT) {
			    SetWTitle(vtcb->windowid, src);
			    vcp = vtcb->vidcopy;
			    while (vcp) {
					SetPort(vcp->windowid);
					SetWTitle(vcp->windowid, (char *) buffer);
					vcp = vcp->next;
			    }
			}
			break;

		case VTERM_TEMPGOAWAY:		/* Make window go away. */
			if (vtcb->flag & VT_INIT) {
			    close_vt220(vtcb);
			}
			break;

		case VTERM_SETHISTORY:		/* Set/get window's history buf size. */
			{
			    unsigned short histlines;

			    histlines = vtcb->histlines; /* Cheap trick here! */
			    (void) rdnix(F_UWRITE + hndl->request.ir_fc - F_IOCR, reqnum,
							 &histlines, sizeof histlines, 0, 0, 0);
			    if (vtcb->flag & VT_INIT) {
					if (vtcb->histlines != histlines) {
						vtcb->histlines = histlines;
						new_history(vtcb);
					}
			    }
			    status = sizeof histlines;
			}
			break;

		case VTERM_GETHISTORY:		/* Get window's history buf. */
								/* The history buffer is split off from the */
								/* regular display buffer, but it probably */
								/* shouldn't be --- it makes the code harder */
								/* (but it minimized impact on old code.) */
			{
			    long histoffset, histsize;

				if ((status = hndl->request.ir_bc) > vtcb->mcol)
					status = vtcb->mcol; /* Read at most one line. */
				src = (char *) vtcb->history;
				(void) rdnix(F_UREAD, reqnum, &histoffset,
							 sizeof histoffset, 0, 0, 0);
				if (histoffset < 0)
					status = EINVAL;
				else {
					histsize = vtcb->histlines * vtcb->mcol;
					if (!src || histoffset >= histsize) { /* Read reg dsply. */
						if (src)
							histoffset -= histsize;
						histsize = vtcb->mrow * vtcb->mcol;
						src = (char *) vtcb->display;
						if (!src || histoffset >= histsize) {
							src = 0; /* Off end of regular display, stop. */
							status = EINVAL;
						}
					}
					if (src) {
						src += histoffset;
						(void) rdnix(F_UWRITE, reqnum, src, status, 0, 0, 0);
					}
				}
			}
			break;

		case VTERM_COPIESTO:	/* Copy (hot) the window. */
			dst = (char *) buffer;
			*dst = 0;			/* Make current video copy list available. */
			if (vcp = vtcb->vidcopy) {
				while (vcp && dst < (char *) &buffer[sizeof buffer / 2]) {
					strcpy(dst, vcp->connect->nodename);
					dst += StringLen(dst);
					if (vcp = vcp->next)
						*dst++ = '\n';
				}
			}
			i = hndl->request.ir_bc;
			if (i > sizeof(buffer) / 2)
				i = sizeof(buffer) / 2;
			BlockMove(buffer, &buffer[sizeof buffer / 2], sizeof buffer / 2);
			(void) rdnix(F_UWRITE + hndl->request.ir_fc - F_IOCR, reqnum,
						 buffer, i, 0, 0, 0); /* Read/write list. */
			status = StringLen((char *) buffer);
			if (memcmp(buffer, &buffer[sizeof buffer / 2], sizeof buffer / 2)){
				hotcopies(vtcb, (char *) buffer);
			}
			break;

		case TIOCGWINSZ:
			if (hndl->request.ir_fc == F_IOCR) {
				struct winsize wind;

				i = hndl->request.ir_bc;
				if (i > sizeof(wind))
					i = sizeof(wind);
				wind.ws_row = vtcb->mrow;
				wind.ws_col = vtcb->mcol;
				wind.ws_xpixel = 0; /* For now. */
				wind.ws_ypixel = 0;
				rdnix(F_UWRITE, reqnum, &wind, i, 0, 0, 0, 0);
				status = sizeof(wind);
			}
			break;

		case TIOCSWINSZ:
			if (hndl->request.ir_fc == F_IOCW) {
				struct winsize owind, nwind;

				i = hndl->request.ir_bc;
				if (i > sizeof(nwind))
					i = sizeof(nwind);
				owind.ws_row = vtcb->mrow;
				owind.ws_col = vtcb->mcol;
				owind.ws_xpixel = 0; /* For now. */
				owind.ws_ypixel = 0;
				rdnix(F_UREAD, reqnum, &nwind, i, 0, 0, 0, 0);
				status = sizeof(nwind);
				nwind.ws_xpixel = 0; /* For now. */
				nwind.ws_ypixel = 0;
				if (memcmp(&owind, &nwind, sizeof(owind))) { /* Changed? */
					vtcb->mrow = nwind.ws_row;
					vtcb->mcol = nwind.ws_col;
					if (vtcb->flag & VT_INIT) {
						close_vt220(vtcb);
						init_vt220(vtcb);
						new_history(vtcb);
#ifdef SIGWINCH
						kill(-vtcb->pid, SIGWINCH);
						kill(vtcb->pid, SIGWINCH);
#endif
					}
				}
			}
			break;

		default:
			LOG("Unimplemented ioctl sub function vtcb = 0x%X ", vtcb);
			LOG(" , code = %X\n", hndl->request.ir_upar);
			status = EINVAL;
			break;
		}

		LOG("Terminating request %d ", reqnum);
		LOG("with %d.\n", status);
		termhreq(reqnum, status, 0);
		break;

	default:
		LOG("Unimplemented request vtcb = 0x%X ", vtcb);
		LOG(" , code = %X\n", hndl->request.ir_fc);
		status = EINVAL;
		LOG("Terminating request %d with EINVAL.\n", reqnum);
		termhreq(reqnum, status, 0);
		break;
	}
  done:
	if (!--entrycount) {
		deferred = 0;
		tqread();
	}
	return retval;
}

/*
** Make hot copies of the given window.
*/
static void
hotcopies(vtcb, names)
struct vtcb *vtcb;		/* Window to copy. */
char *names;			/* List of all names to get hot copies. */
{
    int windowid;
    char *src, namebuf[256];
    struct vidcopy *vcp, *vcp1, *vcp2;
    struct connect *ccp;
    struct EventRecord ev;

    vcp = vtcb->vidcopy;	/* Propose all for deletion. */
    while (vcp) {
	vcp->marker = 0;
	vcp = vcp->next;
    }
    src = names;
    while (src = strtok(src, "\n")) {
	vcp = vtcb->vidcopy;	/* Already there? */
	while (vcp) {
	    if (vcp->connect->nodename
		&& !strcmp(vcp->connect->nodename, src)) {
		vcp->marker = 1;
		goto next;
	    }
	    vcp = vcp->next;
	}
	if (ccp = copyconn(src)) {
	    if (!(vcp = malloc(sizeof *vcp))) {
		delcopyconn(ccp);
		goto next;
	    }
	    vcp->connect = ccp;
	    if (vtcb->flag & VT_INIT) {
		strcpy(namebuf, copystr);
		strncat(namebuf, vtcb->tname, sizeof namebuf);;
		vcp->wrect = vtcb->wrect;
		if ((vcp->windowid = NewWindow(&vcp->wrect, namebuf,
					       showWindow, documentProc,
					       frontWindow,
					       hasDrag|hasGoAway|hasSendToBack,
					       ccp->connectid)) < 0) {
		    delcopyconn(ccp);
		    free(vcp);
		    continue;
		}
		vtcb->ports |= 1 << vcp->windowid;
		SetPort(vcp->windowid);
		TextFont(vtcb->font);
		TextSize(vtcb->fontsize);
	    } else {
		vcp->windowid = -1;
	    }
	    vcp->marker = 1;
	    vcp->next = vtcb->vidcopy;
	    vtcb->vidcopy = vcp;
	    LOG("Hot copy to %s\n", src);
	}
next:
	src = 0;
    }
    deletehotcopies(vtcb);	/* Take out unwanted copies. */
}

/*
** Find (or make) a connection to the named node.
*/
struct connect *
copyconn(name)
char *name;
{
    int wmfd;
    struct connect *ccp;

    ccp = chead;
    while (ccp) {
	if (ccp->nodename && !strcmp(ccp->nodename, name)) {
	    ccp->refcount++;
	    return ccp;		/* Found it, return pointer to it. */
	}
	ccp = ccp->next;
    }
    if ((wmfd = ConnectGraf(name)) >= 0) { /* Try to make it. */
	if (!(ccp = malloc(sizeof *ccp))) {
	    DisconnectGraf(wmfd);
	} else {
	    ccp->refcount = 1;
	    ccp->connectid = wmfd;
	    if (ccp->nodename = malloc(StringLen(name) + 1))
		strcpy(ccp->nodename, name);
	    ccp->next = chead;
	    chead = ccp;
	}
    }
    return ccp;
}

/*
** Delete reference to a copy connection.  After the
** last one's gone delete the connection altogether.
*/
void
delcopyconn(ccp)
struct connect *ccp;
{
    struct connect *ccp1, *ccp2;

    ccp2 = 0;
    ccp1 = chead;
    while (ccp1) {
	if (ccp == ccp1) {
	    if (!--ccp->refcount) {
		DisconnectGraf(ccp->connectid);
		free(ccp->nodename);
		ccp1 = ccp->next;
		if (ccp2)
		    ccp2->next = ccp1;
		else
		    chead = ccp1;
		free(ccp);
	    }
	    return;
	}
	ccp1 = ccp1->next;
    }
}

void
deletehotcopies(vtcb)
struct vtcb *vtcb;
{
    struct vidcopy *vcp, *vcp1, *vcp2;

    vcp2 = 0;
    vcp = vtcb->vidcopy;
    while (vcp) {		/* Delete unwanted copies. */
	if (!vcp->marker) {
	    if (vcp->windowid >= 0) {
		DisposeWindow(vcp->windowid);
		vtcb->ports &= ~(1 << vcp->windowid);
	    }
	    delcopyconn(vcp->connect);
	    vcp1 = vcp->next;
	    if (vcp2)
		vcp2->next = vcp1;
	    else
		vtcb->vidcopy = vcp1;
	    free(vcp);
	    vcp = vcp1;
	} else {
	    vcp2 = vcp;
	    vcp = vcp->next;
	}
    }
}

void
new_history(vtcb)
struct vtcb *vtcb;
{
	int i;

	if (vtcb->history) {
		free(vtcb->history);
		vtcb->history = 0;
	}
	if (i = vtcb->histlines * vtcb->mcol) {
		if (vtcb->history = malloc(i))
			memset(vtcb->history, ' ', i);
		else
			vtcb->histlines = 0;
	}
}

/*
** Restart a held output request.
*/

void
restartflow(vtcb)
struct vtcb *vtcb;
{
	int count, i, ubuf;

	count = vtcb->writebc;		/* Process the data */
	ubuf = 0;					/* (user buffer offset) */
	while (i = count) {			/* in pieces if necessary. */
		if (i > (sizeof(buffer) - bufpos))
			i = (sizeof(buffer) - bufpos);
		rdnix(F_UREAD, vtcb->writern, &buffer[bufpos], i, ubuf, 0, 0);
		count -= i;				/* (count == 0 when done.) */
		ubuf += i;				/* (ubuf == ir_bc when done.) */
		bufpos += i;
		if (!count) {			/* All read?  Release client early. */
			termhreq(vtcb->writern, vtcb->writebc, 0);
			LOG("Request %X terminated\n", vtcb->writern);
			vtcb->writern = 0;
		} else {
			LOG("Processing incomplete client buffer.\n", 0);
			deferred = 1; /* Can't accept more until this */
		}			/* request is fully consumed & terminated. */
		vt220(vtcb, &buffer[bufpos-i], i); /* Do the work. */
		bufpos = 0;
	}
}

void
close_all()
{
	int i, flag;

	(void) close(mountfd);		/* So no new connections are accepted. */
								/* This way there won't be any zombie */
	flag = FALSE;				/* processes left around when we exit. */
	utime(Device, &mdstat.st_atime); /* Propagate times to the mount point */
								/* so /bin/who's accounting works right.  */
	for (i = 0; i < MAXWINDOWS; i++) {
		if (connections[i]) {
			LOG(" i = %d ", i);
			LOG(" connections[i] = %X\n", connections[i]);
			if (connections[i]->pid) {
				flag = TRUE;
				LOG(" Sending SIGHUP to process group id = %d\n",
					connections[i]->pid);
				(void) kill(-connections[i]->pid, SIGHUP);
				(void) kill(connections[i]->pid, SIGHUP); /* If not pgrp. */
			}
			while (connections[i]->readrq)
				termrdreq(connections[i], connections[i]->readrq->readrn, 0);
			if (connections[i]->writern)
				termhreq(connections[i]->writern, ECANCL, 0);
		}
	}
	if (flag)
		nap(2000L);			/* Short pause allows other processes to run. */
	flag = FALSE;
	LOG("Testing for survivors...\n",0);
	for (i=0; i < MAXWINDOWS; i++) {
		if (connections[i] != NULL && connections[i]->pid)
			if (!(kill(connections[i]->pid, 0))) { /* Check for survivors. */
				LOG(" PID %d still alive\n", connections[i]->pid);
				flag = TRUE;
			}
	}
	if (flag) {
		nap(5000L);				/* Longer pause in case of long clean up. */
		for (i = 0; i < MAXWINDOWS; i++) {
			if (connections[i] != NULL && connections[i]->pid) {
				LOG("Sending SIGTERM to process group id = %d\n",
					connections[i]->pid); /* Start bayonetting the wounded. */
				(void) kill(-connections[i]->pid, SIGTERM);
				(void) kill(connections[i]->pid, SIGTERM);
			}
		}
	}
	nap(200L);				/* Short pause allows other processes to run. */
	flag = FALSE;
	for (i=0; i < MAXWINDOWS; i++) {
		if (connections[i] != NULL && connections[i]->pid)
			if (!(kill(connections[i]->pid, 0))) { /* Check for survivors. */
				LOG(" PID %d still alive\n", connections[i]->pid);
				flag = TRUE;
			}
	}
	if (flag) {
		nap(5000L);				/* Longer pause in case of long clean up. */
		for (i = 0; i < MAXWINDOWS; i++) {
			if (connections[i] != NULL && connections[i]->pid) {
				LOG("Sending SIGKILL to process group id = %d\n",
					connections[i]->pid);	/* Terminate with extreme */
				(void) kill(-connections[i]->pid, SIGKILL);	/* prejudice. */
				(void) kill(connections[i]->pid, SIGKILL);
			}
		}
	}
	LOG("Genocide complete.\n",0);
}

/*
**  dispose_vt220
**
**  dispose of the open channel
*/
void
dispose_vt220(vtcb)
struct vtcb *vtcb;
{
    struct vidcopy *vcp;

    if (vtcb->flag & VT_INIT)
	close_vt220(vtcb);

    vcp = vtcb->vidcopy;
    while (vcp) {
	vcp->marker = 0;
	vcp = vcp->next;
    }
    deletehotcopies(vtcb);
    connections[vtcb->connect] = NULL;
    free(vtcb->keybuffer);
    free(vtcb);
    mdstat.st_nlink--;
    if (!mdstat.st_nlink)	/* If last window is gone, dump the menus */
	Menu_kill();		/* as well (de-fragments WM's menu store). */
}
