/*$
** vterm -- External VT type terminal emulation using the window manager
**
**	Synopsis:
**		Performs daemon startup to detatch from all ctty / files / pgrps
**		Mounts on /dev/vterm or specified mount directory.
**		Opens connection to WM using /dev/graphics or specified path.
**		Opens window when it receives an open to the mount directory.
**
**	Startup switches:	-V		Print version number & exit.
**						-dname	Mount directory name.
**						-pname	Workstation ID of window handler.
**						-F		Do not fork / daemonize
**						-xn		Debug level enable ( vtermtest only )
**						-h		Send SIGHUP to all processes open toward
**								handler on exit
**						-A0|1	Set Default Autowrap to OFF | ON
**						-P0|1	Set Default Popup Scroll to OFF | ON
**						-X0|1	Set vt / non_vt_ansi default
**						-B0|1	Set cursor Blink default 
**						-M0|1	Set Mouse Input default 
**						-K0|1	Set International (Euro) Kbd filter default.
**						-fnn	Set Default Font Number
**						-snn	Set Default font size 
**						-cnn	Set Default number of columns
**						-rnn	Set Default number of rows 
**
** No switches set is equivalent to:
** 	vterm -d/dev/vterm -p/dev/graphics -x0 -P0 -A0 -X0 -B0 -M0 -K0 \
**			-f19 -s12 -r24 -c80
**
**	Notes:
**		Default -p is read from environment WINDOWS_ON variable
**				-F switch is assumed if Parent PID is 1
**
**
**  (c) Copyrighted by ISC System Corp., 1988          All rights reserved.
**
*/

#define VERSION "Vterm version 8.3\n"

#include <stdio.h>
#ifndef _SYS5R3
#define _SYS5R3
#endif
#include <signal.h>
#include <sys/types.h>			/* System types. */
#include <dnix/dutsname.h>
#include <dnix/fcodes.h>
#include <dnix/ireq.h>
#include <dnix/errno.h>
#include <sys/stat.h>
#include <termio.h>

#include <isc/window.h>
#include <isc/keycodes.h>
#include <isc/vterm.h>

#include "vtcb.h"
#include "vt220.h"
#include "protos.h"

#define M_UP (0x20)
#define M_DOWN (0x10)
#define M_ALT (0x08)
#define M_CAPS (0x04)
#define M_SHIFT (0x02)
#define M_CTRL (0x01)

extern struct stat mdstat;		/* Mount directory status. */
extern struct vtcb *connections[MAXWINDOWS];
extern char *out_table[];
extern BYTE ctrl_xlate[];

int trapqfd;			/* trap Q file ID */
int	wm_cfd;			/* window manager connection ID */
int debugsw = 0;
int forksw = 1;
int	hupflag = 0;
int	Popup = 0;
int	Autowrap = 0;
int	Not_vt_ansi = 0;
int	CursorBlink = 0;
int	MouseInput = 0;
int	KeyFilter = 0;
int	Rows = MAXROWS;
int	Columns = MAXCOLS;
int	Font = FONTNUMBER;
int	Size = 0;
short obscurecurs;

char *Device = "/dev/vterm";
char *version = VERSION;
char *path;
char basename[MAXNAMELEN];

static Icon myicon = {		/* Plain icon for iconified window. */
	0x00000000,	0x00000000,	0x0FFFFFF0,	0x10000008,
	0x13FFFFC8,	0x14000028,	0x14000028,	0x14000028,
	0x14000028,	0x14000028,	0x14000028,	0x14000028,
	0x14000028,	0x14000028,	0x14000028,	0x14000028,
	0x14000028,	0x14000028,	0x13FFFFC8,	0x10000008,
	0x10000008,	0x1FFFFFF0,	0x0007E000,	0x3FFFFFFC,
	0x40000002,	0x40000002,	0x40000002,	0x40000002,
	0x40000002,	0x40000002,	0x3FFFFFFC,	0x00000000
};

static Icon myicon2 = {		/* Lit up icon to indicate unseen output. */
	0x00000000,	0x00000000,	0x0FFFFFF0,	0x1FFFFFF8,
	0x1FFFFFF8,	0x1C000038,	0x1C000038,	0x1C000038,
	0x1C000038,	0x1C000038,	0x1C000038,	0x1C000038,
	0x1C000038,	0x1C000038,	0x1C000038,	0x1C000038,
	0x1C000038,	0x1C000038,	0x1fFFFFF8,	0x1FFFFFF8,
	0x1FFFFFF8,	0x1FFFFFF0,	0x0007E000,	0x3FFFFFFC,
	0x40000002,	0x40000002,	0x40000002,	0x40000002,
	0x40000002,	0x40000002,	0x3FFFFFFC,	0x00000000
};

unsigned char font_sizes[MAXFONTNUMBER] = {
	12, 12, 18,  9,
	12,  9, 12, 12,
	14, 18, 12, 18,
	12, 26, 12, 12,
	16, 16, 18, 12,
	 7,  7,  8, 12,
	12, 12, 18, 12,
	12, 12,  0
};

/*eject*/

int
main(argc, argv)
int argc;
char *argv[];
{
    int i;
    char *arg;
    struct dutsname duts;	/* For our system name et. al. */

    if (!(path = getenv("WINDOWS_ON")))
	path = "/dev/graphics";
    while(--argc) {
	arg = *++argv;

	if (*arg++ == '-') {
	    switch (*arg++) {
	      case 'V':
		(void) write(1, version, StringLen(version));
		exit(0);
		break;

	      case 'F':
		forksw = 0;
		break;
#ifdef DEBUG
	      case 'x':		/* the higher the 'arg' (1-9) the more trace */
				/* information gets displayed on the console */
		if ((debugsw = myatoi(arg))==0) {
		    debugsw = 1;
		}
		break;
#endif
	      case 'p':
		path = arg;	/* Window manager network connection. */
		break;

	      case 'd':
		Device = arg;	/* Alternate mount point. */
		break;

	      case 'h':
		hupflag = 1;
		break;

	      case 'A':
		if (*arg != '0')
		    Autowrap = 1;
		break;

	      case 'B':
		if (*arg != '0')
		    CursorBlink = 1;
		break;

	      case 'P':
		if (*arg != '0')
		    Popup = 1;
		break;

	      case 'X':
		if (*arg != '0')
		    Not_vt_ansi  = 1;
		break;

	      case 'M':
		if (*arg != '0')
		    MouseInput  = 1;
		break;
	
	      case 'K':
		if (*arg != '0')
		    KeyFilter  = 1;
		break;
	
	      case 'f':
		Font = myatoi(arg);
		break;

	      case 's':
		Size = myatoi(arg);
		break;

	      case 'r':
		Rows = myatoi(arg);
		break;

	      case 'c':
		Columns = myatoi(arg);
		break;

	    }
	}
    }

    (void) rdnix(F_UNAME, &duts, sizeof(struct dutsname));
    (void) strcpy(basename, duts.nodename);
    (void) strcat(basename, " : ");

    if (getppid() != 1 && forksw) { /* Detach ourself. */
	if (fork() != 0) {
	    exit(0);
	}
    }

    if (debugsw == 0) {
	for (i = 0; i < _NFILE; i++) {
	    (void) close(i);	/* Do the rest of the 'daemonization'. */
	}
	(void) dup(dup(open("/dev/null", 2)));
	(void) signal(SIGINT, SIG_IGN);
	if (hupflag == 0)
	    setpgrp();		/* Create our own process group. */
    }

    if (hupflag)
	if (signal(SIGHUP, doabort) == SIG_IGN)
	    signal(SIGHUP, SIG_IGN);

    (void) signal(SIGUSR1, dodebug);
    (void) signal(SIGTERM, doabort); /* So we can be killed by root. */
    (void) signal(SIGCLD, SIG_IGN); /* Don't wait for any children. */
    chdir("/dev");

	/* Initilize the window management system interface */

    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
#ifdef	DEBUG
	fprintf(stderr, "%s: Error %d in InitGraf().\n", argv[0], -trapqfd);
#endif
	exit(1);
    }

    i = 10;
    while (((wm_cfd = ConnectGraf(path)) < 0) && --i) {
#ifdef	DEBUG
	fprintf(stderr, "%s: Error %d in ConnectGraf().\n", argv[0], -wm_cfd);
#endif
	nap(1000);
    }
    if (wm_cfd < 0)
	exit(2);

    handlerinit();		/* Initialize handler */
    niceme();

	/*
	** Now begin handling requests.
	*/

    (void) Event(keyEvt, keyevent);
    (void) Event(updateEvt, updateevent);
    (void) Event(activateEvt, activate_vt220);
    (void) Event(menuEvt, menuevent);
    (void) Event(windowEvt, windowevent);
    (void) Event(mouseEvt, mouseevent);
    (void) Event(scrapEvt, scrapevent);

    while (1) {			/* Stay in event loop forever. */
	(void) GetEvent(noWindow, noEvent);
	i = EventError() >> 16;
	if (i == wm_cfd)	/* Failure on main display connection. */
	    break;
	takeoutfd(i);		/* Must have been a copy node. */
    }
    doabort();
}

struct vtcb *
findvtcb(windowid, vcpp)
int windowid;
struct vidcopy **vcpp;
{
    struct vtcb *vtcb;
    struct vidcopy *vcp;
    int i;

    for (i = 0 ; i < MAXWINDOWS; i++) {
	if ((vtcb = connections[i]) && (vtcb->flag & VT_INIT)) {
	    if (vtcb->windowid == windowid)
		return vtcb;
	    vcp = vtcb->vidcopy;
	    while (vcp) {
		if (vcp->windowid == windowid) {
		    if (vcpp)
			*vcpp = vcp;
		    return vtcb;
		}
		vcp = vcp->next;
	    }
	}
    }
    return 0;
}

/*
** We had an error on a connection.  Do the right thing.
*/
void
takeoutfd(fd)
int fd;
{
    struct vtcb *vtcb;
    struct vidcopy *vcp;
    int i;

    for (i = 0 ; i < MAXWINDOWS; i++) {
	if ((vtcb = connections[i]) && (vtcb->flag & VT_INIT)) {
	    vcp = vtcb->vidcopy;
	    while (vcp) {
		if (vcp->connect->connectid == fd)
		    vcp->marker = 0;
		vcp = vcp->next;
	    }
	    deletehotcopies(vtcb);
	}
    }
}

void
updateevent(ep)
EventRecord *ep;
{
    struct vtcb *vtcb;
    struct vidcopy *vcp;

    LOG("Update for window %X\n", ep->who);
    vcp = 0;
    if (!(vtcb = findvtcb(ep->who, &vcp))) {
	LOG("Update window not found\n", 0);
	return;
    }	
    LOG("vt220 update vtcb = 0x%X", vtcb);
    LOG("  Rect: top = %X", ep->where.edge.top);
    LOG(" left = %X", ep->where.edge.left);
    LOG(" bottom = %X", ep->where.edge.bottom);
    LOG(" right = %X\n", ep->where.edge.right);
    SetPort(ep->who);
    BeginUpdateRect(ep->who, &ep->where);
    if (vtcb->iconified) {
	draw_icon(vtcb);
    } else {
	repaint(vtcb, ep);	/* (This always removes the cursor.) */
	if (vtcb->mode.cursordisp) /* Force cursor (re)draw? */
	    cursortoggle(vtcb);
	if (vtcb->startrow >= 0) { /* COPY selection up? */
	    vtcb->lastrow = vtcb->startrow;
	    vtcb->lastcol = vtcb->startcol;
	    dospan(vtcb, &vtcb->endrow, &vtcb->endcol); /* Redo hilighting. */
	} else {
	    if (vtcb->flag & VT_ACTIVE && !vcp) {
		VCursor(vtcb, TRUE);
		if (vtcb->mode.cursorblink)
		    vtcb->blinkrn = rdnix(F_NAP+F_NOWAIT, 250, 0, 0, 0, 0,
					  trapqfd, vtcb);
	    }
	}
    }
    EndUpdate(ep->who);
    LOG("vt220 update finished\n", 0);
}

/*
** Trim the string in some way.
*/
static
trim(str)
char *str;
{
    char ch, *cp, *cp2;
    int trimmed;

    trimmed = 0;
    cp = cp2 = str;
    while (ch = *cp++) {
	if (ch == ' ') {
	    trimmed++;
	    continue;
	}
	if (ch == 'v' && *cp == 't' && cp[1] >= '0' && cp[1] <= '9' && !cp[2]){
	    trimmed++;
	    ch = cp[1];
	    cp += 2;
	}
	*cp2++ = ch;
    }
    *cp2 = 0;
    if (!trimmed) {
	cp = cp2 = str;
	while (ch = *cp++) {
	    if (ch == '-' || ch == ':') {
		trimmed++;
		continue;
	    }
	    *cp2++ = ch;
	}
	*cp2 = 0;
	if (!trimmed) {
	    cp2 = ((cp2 - str) / 2) + str;
	    *cp2++ = '.';
	    cp = cp2 + 1;
	    while (*cp2++ = *cp++)
		;
	    cp2--;
	}
    }
    return cp2 - str;
}


void
draw_icon(vtcb)			/* Draw our icon. */
struct vtcb *vtcb;
{
    Icon icon;
    char buf[80];
    Rect scratch, kbd;
    int row, col;
    BYTE *cp, *cp2;
    unsigned long *ip;

    SetRect(&scratch, 0, 0, 32, 32);
    if (vtcb->iconified == 1)
	icon = myicon;
    else
	icon = myicon2;

    cp = vtcb->display;		/* Greek the display. */
    cp2 = vtcb->atts;
    for (row = 0; row < vtcb->mrow; row++) {
	for (col = 0; col < vtcb->mcol; col++) {
	    if ((*cp++ != ' ') | (*cp2++ & (notSrcCopy<<4))) {
		ip = ((unsigned long *) &icon) + 6;
		ip += ScaleRU(row, 10, vtcb->mrow);
		*ip |= 0x01000000 >> ScaleRU(col, 17, vtcb->mcol);
	    }
	}
    }
    PlotIcon(&scratch, &icon);
    TextFont(21);		/* Use the teensy font to label the icon. */
    TextSize(7);		/* 7 point size */
    TextFace(normalText);
    TextMode(srcOr);
    SetRect(&kbd, 3, 24, 30, 28);
    strncpy(buf, vtcb->tname, sizeof buf);
    col = StringLen(buf);
    while (LocalTextWidth(21, 0, 7, buf, col) > kbd.edge.right-kbd.edge.left) {
	col = trim(buf);	/* Make the label fit by trimming. */
    }
    DrawJustText(buf, col, &kbd, centerJustify, centerJustify);
    TextFont(vtcb->font);
    TextSize(vtcb->fontsize);
    TextMode(vtcb->att >> 4);
    ValidRect(&scratch);	/* Avoid update event after iconify() call. */
}

/*eject*/
struct vtcb *infront;

void
activate_vt220(ep)
EventRecord *ep;
{
    extern short newscrap, newcopy;
    struct vidcopy *vcp;
    struct vtcb *vtcb;

    LOG("Activate for window %X ", ep->who);
    LOG(" message = %X\n", ep->message);
    vcp = 0;
    if (!(vtcb = findvtcb(ep->who, &vcp))) {
	LOG("Activate window not found\n", 0);
	return;
    }	
    if (vcp)			/* Nothing to do for copy windows. */
	return;
    LOG("Activate vtcb = %X\n", vtcb);
    SetPort(ep->who);
    
    if (ep->message) {		/* Activating window. */
	infront = vtcb;
	if (newscrap)
	    fetchscrap();
	CheckItem(Setup_Menu, AutoWrap, vtcb->mode.autowrap);
	CheckItem(Setup_Menu, Popup_Scroll, vtcb->mode.pop_scroll);
	CheckItem(Setup_Menu, Cursor_Blink, vtcb->mode.cursorblink);
	CheckItem(Setup_Menu, Mouse_Input, vtcb->mode.mouse_input);
	CheckItem(Setup_Menu, Intl_Keyboard, vtcb->mode.intl_kbd_filter);
	if (!vtcb->iconified) {
	    if (vtcb->vidcopy)
		SetPorts(vtcb->ports);
	    VCursor(vtcb, TRUE);
	    if (vtcb->mode.cursorblink)
		vtcb->blinkrn = rdnix(F_NAP|F_NOWAIT,250,0,0,0,0,trapqfd,vtcb);
	} else {		/* Iconified & to front, update greeking. */
	    draw_icon(vtcb);
	}
	vtcb->flag |= VT_ACTIVE;
    } else {			/* De-activating window. */
	infront = 0;
	if (newcopy)
	    writescrap();
	if (!vtcb->iconified) {
	    if (vtcb->blinkrn)
		(void) rdnix(F_CANCEL, vtcb->blinkrn );
	    vtcb->blinkrn = 0;
	    if (vtcb->vidcopy)
		SetPorts(vtcb->ports);
	    VCursor(vtcb, FALSE);
	} else {		/* Iconified & to back, update greeking. */
	    draw_icon(vtcb);
	}
	vtcb->flag &= ~VT_ACTIVE;
    }
}
/*eject*/
static BYTE erase[] = {'\010', ' ', '\010'};
static BYTE echok[] = {'\n'};

void
keyevent(ep)
EventRecord *ep;
{
    struct vtcb *vtcb;
    struct vidcopy *vcp;
    unsigned int ch;
    int i, sl_key;
    char *temp;
    Rect *rp;

				/* Discard all key-up messages. */
    if (ep->message & F_KEYUP) {
	LOG("keyevent %X discarded\n", ep->message);
	return;
    }
				/* Find associated control block(s). */
    vcp = 0;
    if (!(vtcb = findvtcb(ep->who, &vcp))) {
	LOG("Keyboard Event window not found %d\n", ep->who);
	return;
    }
    rp = vcp ? &vcp->wrect : &vtcb->wrect;
    if (vtcb->startrow >= 0)
	return;			/* Shut off keyboard when copying. */
    /*
    ** ISC's keyboard needs some additional work.  Already done by WM when
    ** the International input filter is on.
    */
    if (!(ep->message & F_KBDIII) && !vtcb->mode.intl_kbd_filter)
	fixisckeys(&ep->message);
    GetPort(&i);		/* Make sure correct port is set. */
    if (i & 0x80000000L || vcp) {
	if ((i & 0x7FFFFFFFL) != vtcb->ports)
	    SetPorts(vtcb->ports);
    } else {
	if (i != ep->who)
	    SetPort(ep->who);
    }
				/* Handle ALT-arrow keys for moving window. */
    if (ep->message & F_FUNC && ep->message & F_ALT) {
	switch (ep->message & F_KEYMASK) {
	  case K_UP:
	    OffsetRect(rp, 0, -16);
	    MoveWindow(ep->who, rp->edge.left, rp->edge.top, 0);
	    return;

	  case K_DOWN:
	    OffsetRect(rp, 0, 16);
	    MoveWindow(ep->who, rp->edge.left, rp->edge.top, 0);
	    return;

	  case K_LEFT:
	    OffsetRect(rp, -16, 0);
	    MoveWindow(ep->who, rp->edge.left, rp->edge.top, 0);
	    return;

	  case K_RIGHT:
	    OffsetRect(rp, 16, 0);
	    MoveWindow(ep->who, rp->edge.left, rp->edge.top, 0);
	    return;
	}
    }
    if (vcp && (ep->message & F_FUNC)) {
	ch = ep->message & F_KEYMASK;
	if (ch == K_ESC || ch == K_RESET) {
	    vcp->marker = 0;
	    deletehotcopies(vtcb);
	    return;
	}
    }
				/* If iconified, discard most input */
				/* but complain about it.  Also discard */
				/* input from copy windows. */
    if (vtcb->iconified || vcp) {
	SetPort(ep->who);	/* Only ding the guy who hit the key. */
	Bell();
	return;
    }
			/* If window's an orphan, toss most of the input. */
    if (vtcb->pid == 0) {
	if ((ep->message & F_FUNC) && !vtcb->refcount) {
	    ch = ep->message & F_KEYMASK;
	    if (ch == K_ESC || ch == K_RESET)
		dispose_vt220(vtcb);
	}
	return;
    }

    obscurecurs = 1;		/* (Obscure mouse cursor if typing.) */
				/* Mark cursor for later obscuration. */
				/* This cuts the number of packets in half */
				/* by avoiding the command now which would */
				/* be automatically flushed in many cases. */
    LOG("Keyboard window = %d", ep->who);
    LOG(" Key = %X\n", ep->message);

	if (vtcb->tty.c_iflag & IXON) {	/* Is flow control enabled? */
	    ch = 0xFFFF;	/* Look for 'S', 'Q' or the scroll_lock key. */
	    if (ep->message & F_CTRL) ch = (ep->message & F_KEYMASK) & ~0x20;
	    sl_key = FALSE;
	    if ((ep->message & F_FUNC) && (ep->message & F_KEYMASK) == K_SCROLL)
		sl_key = TRUE;
	    if (vtcb->mode.stopped) {
		if ((ch == 'Q') || (sl_key == TRUE) || 
		    ((ch != 'S') && (vtcb->tty.c_iflag & IXANY))) {
		    removespan(vtcb); /* Turn output on. */
		    vtcb->mode.stopped = FALSE;
		    SetKbdLEDs(vtcb->mode.stopped);
		    if (vtcb->writern)
			restartflow(vtcb); /* Unplug the held data. */
		}
	    } else if ((ch == 'S') || (sl_key == TRUE)) {
		vtcb->mode.stopped = TRUE; /* Shut output off. */
		SetKbdLEDs(vtcb->mode.stopped);
	    }
	    if ((ch == 'Q') || (ch == 'S') || (sl_key == TRUE))
		return;
	}

    if (!(ep->message & F_FUNC))  {
	ch = ep->message & F_KEYMASK;
      keypatch:
	if (ep->message & F_CTRL)
	    if ((ch = ctrl_xlate[ch]) == NOCODE)
		return;
	if (ep->message & F_ALT)
	    ch |= 0x80;
	if ((vtcb->tty.c_lflag & ISIG) &&
	    ((ch==vtcb->tty.c_cc[VINTR]) || (ch==vtcb->tty.c_cc[VQUIT]))) {
	    if ((vtcb->tty.c_lflag & NOFLSH) == 0)
		while (rmbuffer(vtcb) == TRUE)
		    ;
	    (void) kill(-vtcb->pid, (ch == vtcb->tty.c_cc[VINTR])
			? SIGINT : SIGQUIT);
	    LOG("Sending Signal %d to process group",
		(ch == vtcb->tty.c_cc[VINTR]) ? SIGINT : SIGQUIT);
	    LOG("%d\n", vtcb->pid);
	    return;
	} else if ((vtcb->tty.c_lflag & ICANON) && 
		   ((ch == vtcb->tty.c_cc[VERASE])
		    || (ch == vtcb->tty.c_cc[VKILL]) || 
		    ((ch == vtcb->tty.c_cc[VEOF])
		     && (vtcb->keybuffer[(vtcb->keyinp + vtcb->maxkeys - 1)
					 % vtcb->maxkeys] == '\\')))) {
	    if ((vtcb->keybuffer[(vtcb->keyinp + vtcb->maxkeys - 1)
				 % vtcb->maxkeys] != '\\')) {
		if (ch == vtcb->tty.c_cc[VERASE]) {
		    if (rmbuffer(vtcb) == TRUE
			&& vtcb->tty.c_lflag & ECHOE) {
			if (vtcb->tty.c_lflag & ECHO)
			    vt220(vtcb, erase, sizeof(erase));
			else
			    vt220(vtcb, &erase[1], sizeof(erase) - 1);
		    }
		} else if (ch != vtcb->tty.c_cc[VEOF]) {
		    i = 0;
		    while (rmbuffer(vtcb) == TRUE)
			i++;
		    if ((vtcb->tty.c_lflag & ECHO)) {
			if (vtcb->tty.c_lflag & ECHOK)
			    vt220(vtcb, echok, sizeof(echok));
			else if (vtcb->tty.c_lflag & ECHOE)
			    while (i--)
				vt220(vtcb, erase, sizeof(erase));
		    }
		}
	    } else {
		vtcb->keybuffer[(vtcb->keyinp + vtcb->maxkeys - 1)
				% vtcb->maxkeys] = ch;
	    }
	    return;
	}
	if (ch == CR && vtcb->tty.c_iflag & ICRNL)
	    ch = NL;
	if (vtcb->keycount == vtcb->maxkeys) {
	    if (rmbuffer(vtcb) == TRUE ) {
		if (vtcb->tty.c_lflag & ECHO  && vtcb->tty.c_lflag & ECHOE)
		    vt220(vtcb, erase, sizeof(erase));
	    }
	}
	if (vtcb->keycount < vtcb->maxkeys)
	    addbuffer(vtcb, ch);
    } else {
	ch = ep->message & 0x7F;
	if (ch == K_ENTER || ch == K_BACKSP || ch == K_BCKSP2 || ch == K_ESC) {
	    ch = *out_table[ch];
	    goto keypatch;
	}
	if (ep->message & F_SHIFT) {
	    if ((ch == K_NPLUS || ch == K_NMINUS))
		ch = ',';	/* Fake the missing npad key (K_NCOMMA). */
	    else if (ch == K_TAB)
		ch = 002;	/* Backtab character. */
	}
	if ((temp = out_table[ch])
	    && (vtcb->keycount + StringLen(temp)) < vtcb->maxkeys) {
	    int i, rawch;
	    i = 0; rawch = ch;
	    while ((ch = *temp++)) {
		if (rawch >= 0x61 && rawch <= 0x6C && (ep->message & F_CTRL)
		    && (i == 2)) { /* Ctrl-function keys? */
		    ch += 4;	/* Add some more then (HACK!). */
		}
		addbuffer(vtcb, ch);
		i++;
	    }
	}
    }

    if (((vtcb->tty.c_lflag & ECHO))
	|| ((ch == NL) && (vtcb->tty.c_lflag & ECHONL))) {
	if (ch != vtcb->tty.c_cc[VEOF])
	    vt220(vtcb, &vtcb->keybuffer[(vtcb->keyinp + vtcb->maxkeys - 1)
					 % vtcb->maxkeys], 1);
	else	
	    LOG("EOF %d: Not echoed\n",ch);
    }


	/* Only check for termination if no more keys. */
    if (!CheckNextEvent(vtcb->windowid, keyEvt)) {
	if (vtcb->readrq)	/* If outstanding read, */
	    (void) vt220_rdchk(vtcb); /* check for complete. */
    }
}

void
startspan(vtcb, row, col)
struct vtcb *vtcb;
short row, col;
{
    if (vtcb->startrow >= 0) {
	removespan(vtcb);	/* Remove highlighting. */
    }
    vtcb->lastrow = vtcb->startrow = vtcb->endrow = row;
    vtcb->lastcol = vtcb->startcol = vtcb->endcol = col;
}

void
dospan(vtcb, currow, curcol)
struct vtcb *vtcb;
short *currow, *curcol;
{
    Rect hiliterect;
    short row, col, scolonrow, newcol, srow, newrow;

    if (vtcb->startrow < 0)
	return;
    if (*currow < 0)		/* Constrain span variables to valid range. */
	*currow = 0;
    if (*currow >= vtcb->mrow)
	*currow = vtcb->mrow - 1;
    if (*curcol < 0)
	*curcol = 0;
    if (*curcol >= vtcb->mcol)
	*curcol = vtcb->mcol - 1;

    newcol = col = *curcol;	/* Normalize the span. */
    scolonrow = vtcb->lastcol;
    newrow = row = *currow;
    srow = vtcb->lastrow;
    if (row < srow) {
	short i;

	row = srow;
	srow = newrow;
	i = vtcb->lastcol;
	vtcb->lastcol = col;
	col = i;
    } else if (row == srow && col < scolonrow) {
	col = scolonrow;
	scolonrow = newcol;
    }

    if (row != srow) {		/* Row changed? */
	hiliterect.edge.top = srow * vtcb->charhigh + TOFFSET;
	hiliterect.edge.bottom = hiliterect.edge.top + vtcb->charhigh;
	hiliterect.edge.left = vtcb->lastcol*vtcb->charwide + vtcb->charoffset;
	hiliterect.edge.right = vtcb->mcol * vtcb->charwide + vtcb->charoffset;
	InvertRect(&hiliterect); /* Do first (partial) row. */
	if ((srow + 1) < row) {	/* Spanned more than one line? */
	    hiliterect.edge.top = hiliterect.edge.bottom;
	    hiliterect.edge.bottom = row * vtcb->charhigh + TOFFSET;
	    hiliterect.edge.left = vtcb->charoffset;
	    InvertRect(&hiliterect); /* Do middle row(s).  (Whole ones.) */
	}
	hiliterect.edge.top = hiliterect.edge.bottom;
	hiliterect.edge.bottom += vtcb->charhigh;
	hiliterect.edge.left = vtcb->charoffset;
	hiliterect.edge.right = col * vtcb->charwide + vtcb->charoffset;
	InvertRect(&hiliterect); /* Do last (partial) row. */
	vtcb->lastrow = newrow;
	vtcb->lastcol = newcol;
    } else if (col != scolonrow) { /* Column changed? */
	hiliterect.edge.top = row * vtcb->charhigh + TOFFSET;
	hiliterect.edge.bottom = hiliterect.edge.top + vtcb->charhigh;
	hiliterect.edge.left = scolonrow * vtcb->charwide + vtcb->charoffset;
	hiliterect.edge.right = col * vtcb->charwide + vtcb->charoffset;
	InvertRect(&hiliterect);
	vtcb->lastcol = newcol;
    }
}

void
removespan(vtcb)
struct vtcb *vtcb;
{
    extern BYTE *scraphandle;

    if (vtcb->startrow >= 0) {
	vtcb->lastrow = vtcb->startrow;
	vtcb->lastcol = vtcb->startcol;
	dospan(vtcb, &vtcb->endrow, &vtcb->endcol); /* Remove highlighting. */
	vtcb->startrow = -1;	/* Mark highlighting as gone. */
	vtcb->mode.stopped = FALSE; /* Restart channel. */
	SetKbdLEDs(vtcb->mode.stopped);
	DisableItem(Edit_Menu, 1); /* Disallow a COPY. */
	if (scraphandle)
	    EnableItem(Edit_Menu, 2);
    }
}

void
endspan(vtcb)
struct vtcb *vtcb;
{
    removespan(vtcb);
    VCursor(vtcb, TRUE);
    if (vtcb->writern)
	restartflow(vtcb);	/* Unplug the held data (if any). */
}

void
mouseevent(ep)
EventRecord *ep;
{
    struct vidcopy *vcp;
    struct vtcb *vtcb;
    long i;
    short row, col, flags;

    vcp = 0;
    if (!(vtcb = findvtcb(ep->who, &vcp))) {
	LOG("Event for window not found\n", 0);
	return;
    }
    if (vcp)			/* Nothing to do for copy windows. */
	return;
    LOG("Mouse event vtcb = 0x%X\n", vtcb);

    if (vtcb->iconified)
	return;

    GetPort(&i);		/* Make sure correct port is set. */
    if (i & 0x80000000L || vcp) {
	if ((i & 0x7FFFFFFFL) != vtcb->ports)
	    SetPorts(vtcb->ports);
    } else {
	if (i != ep->who)
	    SetPort(ep->who);
    }
    /*
    ** compute the row, col
    */
    row = ep->where.edge.top - TOFFSET;
    row /= vtcb->charhigh;
    col = ep->where.edge.left - vtcb->charoffset;
    col /= vtcb->charwide;

    if (!vtcb->mode.mouse_input) { /* Local COPY selection mouse operations? */
	if (ep->message & F_KEYDOWN) {
	    if ((ep->message & F_KEYMASK) == 1) {
		vtcb->mode.stopped = TRUE; /* Stop channel. */
		MouseMode(F_KEYDOWN | F_KEYUP | F_MOUSE1); /* Allow drags. */
		SetKbdLEDs(vtcb->mode.stopped);
		VCursor(vtcb, FALSE);
		startspan(vtcb, row, col);
	    }
	} else if (ep->message & F_KEYUP) {
	    if ((ep->message & F_KEYMASK) == 1) {
		MouseMode(F_KEYDOWN | F_KEYUP);	/* Normal mousies. */
		dospan(vtcb, &row, &col);
		vtcb->endrow = row;
		vtcb->endcol = col; /* Spanned range? */
		if (row != vtcb->startrow || col != vtcb->startcol) {
		    EnableItem(Edit_Menu, 1); /* Allow a COPY. */
		    DisableItem(Edit_Menu, 2); /* Disallow a PASTE. */
		} else {
		    endspan(vtcb);
		}
	    }
	} else if (ep->message & F_MOUSE1) { /* Was a drag event. */
	    dospan(vtcb, &row, &col);
	}
	return;
    }
    /*
    ** Remote mouse op, figure out the flags.
    */
    flags = 0;
    if (ep->message & F_MOUSE1)
	flags |= M_DOWN;
    else
	flags |= M_UP;
    if (ep->message & F_ALT) flags |= M_ALT;
    if (ep->message & F_SHIFT) flags |= M_SHIFT;
    if (ep->message & F_CTRL) flags |= M_CTRL;

    if (vtcb->keycount < (vtcb->maxkeys - 5)) {	/* Make sure we have room. */
	addbuffer(vtcb, 0x1B);			/* ESC to start things out, */
	addbuffer(vtcb, 'm');			/* lower case 'm' next, */
	addbuffer(vtcb, (row+'!'));		/* row comes next, */
	addbuffer(vtcb, (col+'!'));		/* col comes next, */
	addbuffer(vtcb, (flags+' '));		/* key flags end the list. */
	if (vtcb->readrq)
	    (void) vt220_rdchk(vtcb);
    }
    LOG("Mouse event finished\n", 0);
}

/*eject*/
void
windowevent(ep)
EventRecord *ep;
{
    struct vidcopy *vcp;
    struct vtcb *vtcb;
    Rect *rp;
    int i, dx, dy;

    LOG("Event for window %X\n", ep->who);
    vcp = 0;
    if (!(vtcb = findvtcb(ep->who, &vcp))) {
	LOG("Event for window not found\n", 0);
	return;
    }
    rp = vcp ? &vcp->wrect : &vtcb->wrect;
    LOG("Event vtcb = 0x%X", vtcb);
    LOG("  Message (from some flounder): %X", ep->message);
    LOG("  Rect: top = %X", ep->where.edge.top);
    LOG(" left = %X", ep->where.edge.left);
    SetPort(ep->who);
    switch (ep->message & F_KEYMASK) {
      case windowMove:
	OffsetRect(rp, ep->where.edge.left, ep->where.edge.top);
	MoveWindow(ep->who, rp->edge.left, rp->edge.top, 0);
	break;

      case windowZoom:
	if (vtcb->iconified) {
	    douniconify(vtcb);
	} else {
	    doiconify(vtcb);
	}
	break;

      case windowGoaway:
	LOG("GoAway request vtcb = 0x%X\n", vtcb);
	if (vtcb) {
	    if (vcp) {
		vcp->marker = 0;
		deletehotcopies(vtcb);
	    } else {
		if (vtcb->flag & VT_INIT) { /* A real window, is it an orphan? */
		    if (vtcb->pid == 0) {
			if (!vtcb->refcount)
			    dispose_vt220(vtcb); /* Yes, just kill it. */
		    } else {	/* Not an orphan, hit process(es) on head. */
			i = SIGHUP;
			if (ep->message & F_CTRL) /* CTRL-close: Terminate. */
			    i = SIGTERM;
			if (ep->message & F_ALT) /* ALT-close: Kill. */
			    i = SIGKILL;
			if (!(ep->message & F_SHIFT)) /* Shift: don't signal */
			    (void) kill(-vtcb->pid, i); /* the process grp. */
			(void) kill(vtcb->pid, i);
		    }
		}
	    }
	}
	break;

#ifdef SIGWINCH
      case windowResize:
	if (!vtcb->iconified) {
	    dx = ep->where.edge.left / vtcb->charwide;
	    dy = ep->where.edge.top / vtcb->charhigh;
	    if (dx || dy) {
		vtcb->mcol += dx;
		vtcb->mrow += dy;
		close_vt220(vtcb);
		init_vt220(vtcb);
		new_history(vtcb);
		kill(-vtcb->pid, SIGWINCH);
		kill(vtcb->pid, SIGWINCH);
	    }
	}
	break;
#endif
    }
    LOG("Window Event finished\n", 0);
}

/*eject*/
void
fixisckeys(keycode)		/* Fix brain-dead ISC keyboard's mess */
long *keycode;			/* it makes of multiple shifting keys. */
{				/* Also provides missing ASCII's. */
	unsigned short key;
	long flags;

	key = *keycode & F_KEYMASK;
	flags = *keycode & F_FLAGMASK;
	if (key == ESC) {		/* Missing twiddle & backquote. */
		if (flags & F_SHIFT) {
			flags &= ~F_FUNC;
			key = '~';
		}
		if (flags & F_CTRL) {
			flags &= ~(F_CTRL | F_FUNC);
			key = '`';
		}
	} else if (flags & F_FUNC)	/* Skip all other function keys. */
		return;

	if (flags & F_SHIFT && flags & F_ALT) { /* Keyboard refuses to do shift */
		if (key >= 'a' && key <= 'z') /* processing if any other modifier */
			key &= ~('a' - 'A'); /* is active.  Do uppercasing. */
		else
			switch (key) {
			case '1':
				key = '!';
				break;
			case '2':
				key = '@';
				break;
			case '3':
				key = '#';
				break;
			case '4':
				key = '$';
				break;
			case '5':
				key = '%';
				break;
			case '6':
				key = '^';
				break;
			case '7':
				key = '&';
				break;
			case '8':
				key = '*';
				break;
			case '9':
				key = '(';
				break;
			case '0':
				key = ')';
				break;
			case '-':
				key = '_';
				break;
			case '=':
				key = '+';
				break;
			case '\\':
				key = '|';
				break;
			case '[':
				key = '{';
				break;
			case ']':
				key = '}';
				break;
			case ';':
				key = ':';
				break;
			case '\'':
				key = '"';
				break;
			case ',':
				key = '<';
				break;
			case '.':
				key = '>';
				break;
			case '/':
				key = '?';
				break;
			}
	}
	*keycode = flags | key;
}

/*
** Count how many characters are ready to be supplied to a read.
** Take care of cooked/raw processing.  Returns -1 if the read
** should be terminated with 0 bytes read.
*/

int
count_ready(vtcb, kybuffer)
struct vtcb *vtcb;
BYTE *kybuffer;					/* Optional output buffer. */
{
	int wrcount, rdcount;
	BYTE ch;

	wrcount = 0;
	if ((vtcb->keycount >= vtcb->tty.c_cc[VMIN])
		|| (vtcb->tty.c_lflag & ICANON || vtcb->tty.c_cc[VTIME])) {
		if (vtcb->tty.c_lflag & ICANON) {
			for (rdcount=0, wrcount=0; rdcount < vtcb->keycount ;) {
				ch = vtcb->keybuffer[(vtcb->keyoutp + rdcount++) % vtcb->maxkeys];
				if (ch == vtcb->tty.c_cc[VEOF])
					break;
				if (kybuffer)
					kybuffer[wrcount] = ch;
				wrcount++;
				if ((ch == NL) || (ch == vtcb->tty.c_cc[VEOL])
					|| (ch == vtcb->tty.c_cc[VEOL2]))
					break;
			}
			if (!rdcount || !((ch == NL) || (ch == vtcb->tty.c_cc[VEOF])
							  || (ch == vtcb->tty.c_cc[VEOL])
							  || (ch == vtcb->tty.c_cc[VEOL2])))
				return 0;
			if (!wrcount && ch == vtcb->tty.c_cc[VEOF])	/* No data except */
				return -1;		/* VEOF, so signal immediate completion. */
		} else {				/* RAW mode. */
			for (rdcount = 0 ; rdcount < vtcb->keycount ; rdcount++) {
				if (kybuffer)
					kybuffer[rdcount] = vtcb->keybuffer[(vtcb->keyoutp
														 + rdcount) % vtcb->maxkeys];
			}
			wrcount = rdcount;
		}
	}
	return wrcount;
}

int
vt220_rdchk(vtcb)
struct vtcb *vtcb;
{
	BYTE *kybuffer, *alloca();
	int wrcount;

	LOG("rdchk keycount = %d", vtcb->keycount);
	LOG(" VMIN = %d", vtcb->tty.c_cc[VMIN]);
	LOG(" VTIME = %d", vtcb->tty.c_cc[VTIME]);
	LOG(" icanon = %d\n", vtcb->tty.c_lflag & ICANON);

	if (vtcb->naprn) {
		LOG("Cancelling nap request %d\n", vtcb->naprn);
		(void) rdnix(F_CANCEL, vtcb->naprn);
		vtcb->naprn = 0;
	}
	kybuffer = alloca(vtcb->maxkeys);
	wrcount = count_ready(vtcb, kybuffer);
	if (!(vtcb->tty.c_lflag & ICANON)) { /* RAW mode? */
		if (wrcount < vtcb->tty.c_cc[VMIN]) { /* Not gotten minimum yet? */
								/* We only need to handle VMIN > 0 */
								/* and VTIME > 0 case here.  The others */
								/* are handled at the time the read is */
								/* initially posted, or when the timer */
								/* expires. */
			if (vtcb->tty.c_cc[VTIME] /* Inter-character timer requested? */
				&& (wrcount || !vtcb->tty.c_cc[VMIN])) {
				vtcb->naprn = rdnix(F_NAP|F_NOWAIT,	/* Start the timer. */
									vtcb->tty.c_cc[VTIME] * 100,
									0, 0, 0, 0, trapqfd, vtcb);
				LOG("Timer started vtcb = %X", vtcb);
				LOG(" rn = %d\n", vtcb->naprn);
			}
			return 0;
		}
	}
	if (wrcount) {
		wrcount = terminate(vtcb, kybuffer, wrcount);
		return wrcount;
	}
	return 0;
}

int
terminate(vtcb, kybuffer, wrcount) /* Terminate read[check] request. */
struct vtcb *vtcb;
BYTE *kybuffer;
int wrcount;
{
    int tcount;
    struct readrq *rq;

    if (!(rq = vtcb->readrq))
	return 0;
    tcount = wrcount;
    if (wrcount < 0) {		/* Cooker says quit now with */
	tcount = -wrcount;	/* no actual read data. */
	wrcount = 0;
    }
    if (rq->readcheck) {	/* Terminating a no-wait readcheck? */
	termrdreq(vtcb, rq->readrn, wrcount);
    } else {
	LOG("Terminating Read Request for 0x%X ", vtcb);
	LOG(" rn = %d ", rq->readrn);
	LOG(" bc = %d ", rq->readbc);
	LOG(" cnt = %d\n", wrcount);
	if (wrcount > rq->readbc) /* Don't send too much out. */
	    wrcount = tcount = rq->readbc;
	if (wrcount)
	    (void) rdnix(F_UWRITE, rq->readrn, kybuffer, wrcount, 0, 0, 0, 0);
	termrdreq(vtcb, rq->readrn, wrcount);
	vtcb->keycount -= tcount;
	vtcb->keyoutp = (vtcb->keyoutp + tcount) % vtcb->maxkeys;
	if (!vtcb->keycount && vtcb->maxkeys > MAXKEYS) { /* Buffer emptied */
	    vtcb->keybuffer = (BYTE *) realloc(vtcb->keybuffer,
					       vtcb->maxkeys = MAXKEYS);
	    vtcb->keyinp = vtcb->keyoutp = 0; /* and is overlarge?  Shrink */
	}			/* it to normal.  */
    }
    time(&mdstat.st_atime);
    mdstat.st_mtime = mdstat.st_atime;
    return wrcount;
}

/*eject*/
int
vt220_rdtimer(vtcb, rpar, rn)
struct vtcb *vtcb;
int	rpar, rn;
{
	BYTE *kybuffer, *alloca();
	int wrcount;

	if (vtcb->naprn == rn) {	/* Read timeout timer? */
		vtcb->naprn = 0;		/* clear pending t/o flag */

		if (rpar == ECANCL)
			return 0;
		if (!vtcb->readrq)
			return 0;

		LOG("Timer expired on read check ", 0);
		kybuffer = alloca(vtcb->maxkeys);
		wrcount = count_ready(vtcb, kybuffer);
		wrcount = terminate(vtcb, kybuffer, wrcount);
	} else {					/* One of the blink timers. */
		if (vtcb->blinkrn != rn)
			return 0;			/* not valid anymore , ignore */	
		if (rpar == ECANCL) {	
			vtcb->blinkrn = 0;
			return 0;
		}
		if ((vtcb->flag & VT_ACTIVE)) {  /* are we the active window */
			SetPort(vtcb->windowid);
			if (vtcb->mode.cursordisp) {
				vtcb->blinkrn = rdnix(F_NAP+F_NOWAIT,150,0,0,0,0,trapqfd,vtcb);
				VCursor(vtcb, FALSE);
			} else {
				vtcb->blinkrn = rdnix(F_NAP+F_NOWAIT,350,0,0,0,0,trapqfd,vtcb);
				VCursor(vtcb,TRUE);
			}
		}
	}
	return 0;
}



/*eject*/
void
doabort()
{
	extern short newcopy;
	if (newcopy) {
	    if (infront)
		SetPort(infront->windowid);
	    writescrap();
	}
	close_all();
#if 0
	FlushWindow();
	DisconnectGraf(wm_cfd);
	EndGraf();
	sleep(2);
#endif
	exit(0);
}

void
dodebug()
{
    static int debug;

    signal(SIGUSR1, dodebug);
    debug = !debug;
    FlushDebug(debug);
}

/*
** Smaller than the standard atoi() function.
*/

int
myatoi(p)
register char *p;
{
	register int val, flag;
	register char chr;

	val = flag = 0;
	if (*p == '-') {
		flag = 1;
		p++;
	}
	while (1) {
		chr = *p++ - '0';
		if (chr < 0 || chr > 9)
			break;
		val = (val * 10) + chr;
	}
	if (flag)
		val = -val;
	return val;
}

void
douniconify(vtcb)
struct vtcb *vtcb;
{
    EventRecord scratch;
    struct vidcopy *vcp;

    UnIconifyWindow(vtcb->windowid);
    vtcb->iconified = 0;
    SetRect(&scratch.where, 0, 0, 32000, 32000);
    EraseRect(&scratch.where);
    scratch.who = vtcb->windowid;
    LOG("False ", 0);		/* Do an update here to save time and reduce */
    updateevent(&scratch);	/* timing windows for display errors. */
    vcp = vtcb->vidcopy;
    while (vcp) {
	SetPort(vcp->windowid);
	UnIconifyWindow(vcp->windowid);
	vcp = vcp->next;
    }
}

void
doiconify(vtcb)
struct vtcb *vtcb;
{
    struct vidcopy *vcp;

    IconifyWindow(vtcb->windowid);
    vtcb->iconified = 1;
    draw_icon(vtcb);		/* This saves us getting an update event. */
    vcp = vtcb->vidcopy;
    while (vcp) {
	SetPort(vcp->windowid);
	IconifyWindow(vcp->windowid);
	vcp = vcp->next;
    }
}
