/*$
**  vtmenu -  Menu driven fuctions for terminal Emulation
**
**	Synopsis:
**
**	External Function List:
**	Menu_init()
**	menu_vt220(ep)
**
**
**	Notes:
**
**
**  (c) Copyrighted by ISC System Corp., 1988          All rights reserved.
**
*/

#include <stdio.h>
#include <isc/window.h>
#include <sys/types.h>		/* System types. */
#include <dnix/fcodes.h>
#include <dnix/ireq.h>
#include <termio.h>
#include <isc/vterm.h>
#include "vtcb.h"
#include "protos.h"

extern	struct	vtcb	*Term;
extern int trapqfd;		/* Trap Q file ID. */
extern int mountfd;		/* Request Q file ID. */
short niceflag;
char *settings = "/usr/lib/graphics/vtsettings";
char *history = "/usr/bin/vthistory";
char *mo = "/usr/local/mo";
extern char *path, *version;

/*%
**  Menu_init -  Initialize menus for current window
**
**	Synopsis:
**
**	Usage:
**
**	Parameters:
**
**	Returns:
**
**	Errors:
**
**	Notes:
*/


void
Menu_init(vtcb)
struct	vtcb *vtcb;
{
    extern BYTE *scraphandle;

    NewMenu(Setup_Menu, "Setup"); /* First build the menus themselves. */
    AppendMenu(Setup_Menu, "Autowrap;Popup Scroll;Cursor Blink;Mouse Input;"
	       "Keyboard Int\325l;Nice;Misc Settings\311;Show History\311;-;"
	       "Clear;Reset;Hide;-;Stty SANE");
    NewMenu(Edit_Menu, "Edit");
    AppendMenu(Edit_Menu, "(Copy;(Paste");
    CheckItem(Setup_Menu, AutoWrap, vtcb->mode.autowrap); /* Now patch 'em */
    CheckItem(Setup_Menu, Popup_Scroll, vtcb->mode.pop_scroll);	/* up. */
    CheckItem(Setup_Menu, Cursor_Blink, vtcb->mode.cursorblink);
    CheckItem(Setup_Menu, Mouse_Input, vtcb->mode.mouse_input);
    CheckItem(Setup_Menu, Intl_Keyboard, vtcb->mode.intl_kbd_filter);
    CheckItem(Setup_Menu, NiceFlag, niceflag);
    if (scraphandle)
	EnableItem(Edit_Menu, 2);
}

/*
** Dispose of the menus if there're no windows left.  This helps WM unfragment
** its menu structures for the VTERM that's always running in a workstation.
*/

void
Menu_kill()
{
    DisposeMenu(Setup_Menu);
    DisposeMenu(Edit_Menu);
}

/*eject*/
static void itoa(register char *buf, register int val);

static void
itoa(buf, val)
char *buf;
int val;
{
    short i;
    char out;
    unsigned long test, remainder;
	
    remainder = val;
    test = 1000000000L;	/* Decimal output. */
    if (val < 0) {
	remainder = -val;
	*buf++ = '-';
    }
    while (test > remainder)	/* Scale down to match number. */
	test /= 10;
    if (remainder)
	while (test) {
	    out = remainder / test;
	    *buf++ = out + '0';
	    remainder %= test;
	    test /= 10;
	}
    else
	*buf++ = '0';
    *buf = 0;
}

/*%
**  Menu_vt220 - Handle Menu events
**
**	Synopsis:	Scans connection table to find vtcb pointer
**			
**
**	Usage:
**
**	Parameters:
**
**	Returns:
**
**	Errors:
**
**	Notes:
*/

void
menuevent(ep)
EventRecord *ep;
{
    struct vtcb *vtcb;
    int menuid, menuitem;
    int i, fd;

    SetPort(ep->who);
    BeginMenu(ep->message);
    menuid = ep->message >> 16;
    menuitem = ep->message & 0xFFFF;
    LOG("Menu event %d", menuid);
    LOG(" %d\n", menuitem);
    if (!(vtcb = findvtcb(ep->who, 0))) {
	LOG("Menu window not found\n",0);
	return;
    }
	
    switch (menuid) {
      case Setup_Menu:
	switch (menuitem) {
	  case AutoWrap:
	    vtcb->mode.autowrap = !vtcb->mode.autowrap;
	    CheckItem(Setup_Menu, AutoWrap, vtcb->mode.autowrap); 
	    break;

	  case Popup_Scroll:
	    vtcb->mode.pop_scroll = !vtcb->mode.pop_scroll;
	    CheckItem(Setup_Menu, Popup_Scroll, vtcb->mode.pop_scroll);
	    break;

	  case Cursor_Blink:
	    vtcb->mode.cursorblink = !vtcb->mode.cursorblink;
	    CheckItem(Setup_Menu, Cursor_Blink, vtcb->mode.cursorblink);
	    if (vtcb->mode.cursorblink)
		vtcb->blinkrn = rdnix(F_NAP+F_NOWAIT, 250, 0, 0, 0, 0,
				      trapqfd, vtcb);
	    else {
		VCursor(vtcb, TRUE);
		vtcb->blinkrn = 0;
	    }
	    break;

	  case Mouse_Input:
	    vtcb->mode.mouse_input = !vtcb->mode.mouse_input;
	    CheckItem(Setup_Menu, Mouse_Input, vtcb->mode.mouse_input);
	    break;

	  case Intl_Keyboard:
	    vtcb->mode.intl_kbd_filter = !vtcb->mode.intl_kbd_filter;
	    SetKbdXlate(vtcb->mode.intl_kbd_filter);
	    CheckItem(Setup_Menu, Intl_Keyboard, vtcb->mode.intl_kbd_filter);
	    break;

	  case NiceFlag:
	    niceme();
	    CheckItem(Setup_Menu, NiceFlag, niceflag);
	    break;

	  case Other:
	    fd = dnix(F_OLURQ, mountfd, vtcb, OP_READ | OP_WRIT | OP_NRAD);
	    if (fd >= 0) {
		vtcb->refcount++;
		if (!(i = fork())) { /* Child */
		    for (i=0; i<72; i++) {
			if (i != fd)
			    close(i);
		    }
		    if (fd != 0) {
			dup(fd); /* Put connection to parent on stdin. */
			close(fd);
		    }
		    dup(dup(0)); /* Pick up stdout/stderr. */
		    execl(settings, settings, path, vtcb->tname, version, 0);
		    _exit(1);
		}
		close(fd);
	    }
	    break;

				/* Do a "vthistory | mo" equivalent. */
				/* Make sure mo runs in restricted mode */
				/* because it's running as root! */
	  case History:
	    fd = dnix(F_OLURQ, mountfd, vtcb, OP_READ | OP_NRAD);
	    if (fd >= 0) {
		int pipes[2];
		char rows[10], cols[10], xpos[10], ypos[10];

		vtcb->refcount++;
		if (!(i = fork())) { /* Child (mo) */
		    pipe(pipes);
		    if (!(i = fork())) { /* Grandchild (vthistory) */
			for (i=0; i<72; i++) {
			    if (i != fd && i != pipes[1])
				close(i);
			}
			if (fd != 0) {
			    dup(fd); /* Put vterm on stdin. */
			    close(fd);
			}
			if (pipes[1] != 1) {
			    dup(pipes[1]);
			    close(pipes[1]);
			}
			dup(1); /* Pick up stderr. */
			execl(history, history, 0);
			_exit(1);
		    }
		    for (i=0; i<72; i++) {
			if (i != pipes[0])
			    close(i);
		    }
		    if (pipes[0] != 0) {
			dup(pipes[0]); /* Put vthistory on stdin. */
			close(pipes[0]);
		    }
		    dup(dup(0)); /* Pick up stdout/stderr. */
		    i = StringLen(vtcb->tname);
		    (void) strncat(vtcb->tname,
				   " History", MAXNAMELEN-1-i);
		    itoa(rows, vtcb->mrow);
		    itoa(cols, vtcb->mcol);
		    itoa(xpos, vtcb->wrect.edge.left);
		    itoa(ypos, vtcb->wrect.edge.top);
		    execl(mo, mo, "-F", "-R", "-p", path, "-T", vtcb->tname,
			  "-r", rows, "-c", cols, "-x", xpos, "-y", ypos, 0);
		    _exit(1);
		}
		close(fd);
	    }
	    break;

	  case Reset:
	    Reset_vt220(vtcb);
	    CheckItem(Setup_Menu, AutoWrap, vtcb->mode.autowrap); 
	    CheckItem(Setup_Menu, Popup_Scroll, vtcb->mode.pop_scroll);
	    CheckItem(Setup_Menu, Cursor_Blink, vtcb->mode.cursorblink);
	    CheckItem(Setup_Menu, Mouse_Input, vtcb->mode.mouse_input);
	    CheckItem(Setup_Menu, Intl_Keyboard, vtcb->mode.intl_kbd_filter);
	    if ( vtcb->mode.cursorblink )
		vtcb->blinkrn = rdnix(F_NAP+F_NOWAIT,250,0,0,0,0,trapqfd,vtcb);
	    else
		vtcb->blinkrn = 0;

				/* Fall through into clear routine */
	  case Clear:
	    if (vtcb->vidcopy)
		SetPorts(vtcb->ports);
	    removespan(vtcb);
	    if (vtcb->flag & VT_ACTIVE)
		VCursor(vtcb, FALSE);
	    home();
	    blockclr(0, vtcb->mrow);
	    if (vtcb->flag & VT_ACTIVE)
		VCursor(vtcb, TRUE);
	    if (vtcb->vidcopy)
		SetPort(ep->who);
	    break;

	  case Hide:
	    if (!vtcb->readrq && (vtcb->flag & VT_INIT)) {
		EndMenu(ep->message);
		close_vt220(vtcb);
	    }
	    break;

	  case Stty_Sane:
	    Reset_termio(vtcb);
	    break;
	}
	break;

      case Edit_Menu:
	switch (menuitem) {
	  case CopyItem:
	    copy();
	    break;

	  case PasteItem:
	    paste();
	    break;
	}
	break;
    }
    SetPort(ep->who);
    EndMenu(ep->message);
}

void
niceme()
{
    nice(niceflag ? -4 : 4);
    niceflag = !niceflag;
}
