/*
** GUI interface to vterm settings.  Normally run as a child of vterm.
**
** Will diddle any vterm window that's hooked to stdin.  Usage:
**
**	vtsettings [windows_on [windowname [versionlabel]]]
**
**
** Revision History:
**
**	1.1	02/07/95	Jim Cathey
**		Added "copies to" logic, cut & paste.
**
**	1.0	05/26/93	Jim Cathey
**		Original Release.
*/

#include <strings.h>
#include <isc/window.h>
#include <isc/keycodes.h>
#include <isc/controls.h>
#include <isc/textedit.h>
#include <isc/dlgman.h>
#include "Dsettings.h"
#include <isc/vterm.h>
#include <dnix/fcodes.h>

int windowid;
Rect winrect;
DMDcb *dialog;
char *progname, *winname, *vtversion, *copies, *mysprintf();
struct vterm_window wind;
unsigned short history;
short mywindowisinfront;
short doimportscrap = 1;
short doexportscrap;
char scrapbuf[256];		/* Cut/Paste buffer. */
int scrapsize;			/* Amount in buffer. */

/* Menus and menu items. */
#define EDITMENU 1
#define EDITCUT 1
#define EDITCOPY 2
#define EDITPASTE 3

void updateevent(), windowevent(), scrapevent(), activateevent(), menuevent();
void fetchscrap(), writescrap();

main(argc, argv)
int argc;
char *argv[];
{
    int group, trapqfd, wmfd;
    int rows, cols, xpos, ypos, font, fsize;
    int autowrap, keymode, mousemode, bugmode, popup, blink, newname, ncopies;
    char a_rows[4], a_cols[4], a_font[4], a_fsize[4], a_xpos[5], a_ypos[5];
    char a_history[5], a_name[80], a_copies[256], *path, *alloca();

    path = 0;
    progname = argv[0];
    if (argc > 1)
	path = argv[1];
    if (argc > 2)
	winname = argv[2];
    if (argc > 3) {
	if (vtversion = strchr(argv[3], '\n'))
	    *vtversion = 0;
	vtversion = argv[3];
    }
    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
        exit(1);
    }
    if ((wmfd = ConnectGraf(path)) < 0) { /* Connect to a window manager. */
        exit(2);
    }
    DMInit();			/* Initialize DM, controls, TE's. */
    InitControls();
    TEInit();
    Event(updateEvt, updateevent); /* Bind events. */
    Event(windowEvt, windowevent);
    Event(scrapEvt, scrapevent);
    Event(activateEvt, activateevent);
    Event(menuEvt, menuevent);

    SetRect(&winrect, 0, 0, Dsettings[0].rect.edge.right,
	    Dsettings[0].rect.edge.bottom);
    OffsetRect(&winrect, 60, 60); /* Create window. */
    windowid = NewWindow(&winrect, "VT Settings", showWindow, documentProc,
                         frontWindow, hasDrag|hasGoAway|hasSendToBack, wmfd); 
    SetPort(windowid);
    NewMenu(EDITMENU, "Edit");
    AppendMenu(EDITMENU, "/XCut;/CCopy;/VPaste");

    dnix(F_IOCR, 0, &wind, sizeof(wind), VTERM_WSTATS, 0);
    dnix(F_IOCR, 0, &history, sizeof(history), VTERM_SETHISTORY, 0);
    a_copies[0] = 0;
    dnix(F_IOCR, 0, a_copies, sizeof(a_copies), VTERM_COPIESTO, 0);
    copies = alloca(strlen(a_copies) + 1);
    strcpy(copies, a_copies);
    dialog = DMNew(Dsettings, windowid, 0);
    group = DMCollect(dialog);
    if (group == DMok) {	/* Collect Data and Dispose of Dialog. */
        DMGet(dialog, DSETTINGS_autowrap, &autowrap);
        DMGet(dialog, DSETTINGS_keymode, &keymode);
        DMGet(dialog, DSETTINGS_mousemode, &mousemode);
        DMGet(dialog, DSETTINGS_bugmode, &bugmode);
        DMGet(dialog, DSETTINGS_popup, &popup);
        DMGet(dialog, DSETTINGS_blink, &blink);
        DMGet(dialog, DSETTINGS_rows, a_rows); rows=atoi(a_rows);
        DMGet(dialog, DSETTINGS_cols, a_cols); cols=atoi(a_cols);
        DMGet(dialog, DSETTINGS_font, a_font); font=atoi(a_font);
        DMGet(dialog, DSETTINGS_fontsize, a_fsize); fsize=atoi(a_fsize);
        DMGet(dialog, DSETTINGS_xpos, a_xpos); xpos=atoi(a_xpos);
        DMGet(dialog, DSETTINGS_ypos, a_ypos); ypos=atoi(a_ypos);
	DMGet(dialog, DSETTINGS_name, a_name);
	DMGet(dialog, DSETTINGS_copies, a_copies);
        newname = strlen(a_name);
	if (!winname || !strcmp(a_name, winname))
	    newname = 0;	/* No change. */
	ncopies = strlen(a_copies);
	if (!copies || !strcmp(a_copies, copies))
	    copies = 0;		/* No change. */
        DMGet(dialog, DSETTINGS_history, a_history); history=atoi(a_history);
	DMDisposeNoInval(dialog);
	DisconnectGraf(wmfd);
	dnix(F_IOCR, 0, &wind, sizeof(wind), VTERM_WSTATS, 0);
	wind.vtmode.autowrap = autowrap;
	wind.vtmode.cursorblink = blink;
	wind.cols = cols;
	wind.font = font;
	wind.vtmode.intl_kbd_filter = keymode;
	wind.vtmode.mouse_input = mousemode;
	wind.vtmode.pop_scroll = popup;
	wind.rows = rows;
	wind.fontsize = fsize;
	wind.vtmode.not_vt_ansi = bugmode;
	wind.wpos.edge.left = xpos;
	wind.wpos.edge.top = ypos;
	dnix(F_IOCW, 0, &wind, sizeof(wind), VTERM_WSTATS, 0);
	dnix(F_IOCW, 0, &history, sizeof(history), VTERM_SETHISTORY, 0);
	if (newname)
	    dnix(F_IOCW, 0, a_name, newname+1, VTERM_RENAME, 0);
	if (copies)
	    dnix(F_IOCW, 0, a_copies, ncopies+1, VTERM_COPIESTO, 0);
    } else {
	DMDisposeNoInval(dialog);
    }
    if (doexportscrap) {
	writescrap();
	FlushTwice();
    }
    return 0;
}

void 
updateevent(ep)			/* Update event handler. */
EventRecord *ep;
{
    BeginUpdate(ep->who);
    DMUpdate(dialog);
    EndUpdate(ep->who);
}

void
windowevent(ep)			/* Window event handler. */
EventRecord *ep;
{
    switch (ep->message & F_KEYMASK) {
      case windowGoaway:
	exit(0);
	break;

      case windowMove:
	OffsetRect(&winrect, ep->where.edge.left, ep->where.edge.top);
	MoveWindow(ep->who, winrect.edge.left, winrect.edge.top, 0);
	break;
    }
}

void
activateevent(ep)
EventRecord *ep;
{
    if (ep->message) {		/* Activating window. */
	mywindowisinfront = 1;
	if (doimportscrap)
	    fetchscrap();
    } else {			/* De-activating window. */
	mywindowisinfront = 0;
	if (doexportscrap)
	    writescrap();
    }
}

void
killscrap()
{
    doimportscrap = 0;
}

void
scrapevent(ep)
EventRecord *ep;
{
    if (!ep->message) {		/* Someone published new scrap. */
	doimportscrap = 1;
	if (mywindowisinfront)
	    fetchscrap();
    } else {			/* Someone wants (?) our scrap.. */
	if (doexportscrap && mywindowisinfront)
	    writescrap();
    }
}

void
fetchscrap()
{
    long size;

    killscrap();
    scrapsize = GetScrap(scrapbuf, 'TEXT', sizeof(scrapbuf));
    TESetScrapLen(scrapsize);
    if (scrapsize) {
        EnableItem(EDITMENU, EDITPASTE);
        BlockMove(scrapbuf, TEGetScrapAddr(), scrapsize);
    } else
        DisableItem(EDITMENU, EDITPASTE);
}

void
writescrap()
{
    if (doexportscrap) {
	ZeroScrap();
	PutScrap(scrapsize, 'TEXT', scrapbuf);
	PublishScrap(0);
	doexportscrap = 0;
    }
}

void
menuevent(ep)
EventRecord *ep;
{
    int i, j;
    short menu, item;
    long message;
    DMDcb *dp;
    char *cp, scratch[5];

    BeginMenu(message = ep->message);
    menu = (message >> 16) & 0xFFFF;
    item = message & 0xFFFF;
    switch (menu) {
      case EDITMENU:
	if (dp = dialog) {	/* An active dialog? */
	    if (dp->curTid) {	/* An active TE within it? */
		switch (item) {
		  case EDITCUT:
		    TECut(dp->curTid);
		  getscrap:
		    killscrap();
		    doexportscrap = 1; /* Set export-on-deactivate flag. */
		    scrapsize = TEGetScrapLen();
		    if (scrapsize > sizeof scrapbuf)
			scrapsize = sizeof scrapbuf;
		    if (cp = TEGetScrapAddr())
			BlockMove(cp, scrapbuf, scrapsize);
		    break;

		  case EDITCOPY:
		    TECopy(dp->curTid);
		    goto getscrap;
		    break;

		  case EDITPASTE:
		    TEPaste(dp->curTid);
		    break;
		}
	    }
	}
	break;
    }
    EndMenu(message);
}

/*
** Validation/prefill Routines.
*/

PFcb(dialog, dip, value)	/* Prefill for checkboxes. */
DMDcb *dialog; DMItem *dip; int *value;
{
    switch (dip->ident) {
      case DSETTINGS_autowrap: *value = wind.vtmode.autowrap; break;
      case DSETTINGS_popup: *value = wind.vtmode.pop_scroll; break;
      case DSETTINGS_blink: *value = wind.vtmode.cursorblink; break;
      case DSETTINGS_mousemode: *value = wind.vtmode.mouse_input; break;
      case DSETTINGS_keymode: *value = wind.vtmode.intl_kbd_filter; break;
      case DSETTINGS_bugmode: *value = wind.vtmode.not_vt_ansi; break;
    }
}

PFtext(cbp, dip, cp)		/* Text. */
DMDcb *cbp; DMItem *dip; char *cp;
{
    char *p;

    p = "%d";
    switch(dip->ident) {
      case DSETTINGS_name: if (winname) strcpy(cp, winname); break;
      case DSETTINGS_rows: mysprintf(cp, p, wind.rows); break;
      case DSETTINGS_cols: mysprintf(cp, p, wind.cols); break;
      case DSETTINGS_font: mysprintf(cp, p, wind.font); break;
      case DSETTINGS_fontsize: mysprintf(cp, p, wind.fontsize); break;
      case DSETTINGS_xpos: mysprintf(cp, p, wind.wpos.edge.left); break;
      case DSETTINGS_ypos: mysprintf(cp, p, wind.wpos.edge.top); break;
      case DSETTINGS_history: mysprintf(cp, p, history); break;
      case DSETTINGS_vtversion: if (vtversion) strcpy(cp, vtversion); break;
      case DSETTINGS_copies: if (copies) strcpy(cp, copies); break;
    }
}

/*
** Sprintf replacement.
**
** Handles format strings containing:
**
**	 %s, %[0][1..9]d, %[0][1..9]x, %[0][1..9]X
**
** It's much smaller than the standard one.
*/

char *
mysprintf(buffer, fmt, arglist)
char *buffer, *fmt;
char *arglist;
{
    register char *p, *q, *p2;
    short ndigits;
    int leadfill;
    char *itoa();
    union arg {
	long lng;
	char *ptr;
    } *argptr;
	
    argptr = (union arg *) &arglist;
    p = fmt;
    q = buffer;
    while (*p) {		/* While there's some format string. */
	if (*p == '%') {	/* Is it a format control? */
	    p++;		/* Yes, which one (skip over % char)? */
	    if (*p == '0')
		leadfill = *p++;
	    else
		leadfill = ' ';
	    if (*p >= '1' && *p <= '9')	/* Format output length? */
		ndigits = *p++ - '0';
	    else
		ndigits = 0;
	    if (*p == 's') {
		p2 = argptr++->ptr; /* String. */
	    } else if (*p == 'd') {
		p2 = itoa(argptr++->lng, 10, leadfill, ndigits);
	    } else if (*p == 'x' || *p == 'X') {
		p2 = itoa(argptr++->lng, 16, leadfill, ndigits);
	    }
	    p++;		/* Skip over format control. */
	    while (*p2)		/* Copy the inclusion. */
		*q++ = *p2++;
	} else			/* No, it's a literal -- copy it. */
	    *q++ = *p++;
    }
    *q = 0;			/* Null-terminate it. */
    return q;			/* Return pointer to NULL. */
}

char *
itoa(val, radix, leadfill, ndigits)
int val, radix, leadfill, ndigits;
{
    short i;
    char *p, *p2, out;
    static char buf[20];
    unsigned long test, remainder;
	
    p = buf;
    remainder = val;
    i = ndigits;		/* Make sure there're at least the number */
    while (i--)			/* of leading whatevers that we've asked for */
	*p++ = leadfill;	/* as the field width. */
    p2 = p;
    if (radix == 16) {		/* Hex output? */
	test = 0x10000000L;
    } else {
	test = 1000000000L;	/* Decimal output? */
	if (val < 0) {
	    remainder = -val;
	    *p++ = '-';
	}
    }
    while (test > remainder)	/* Scale down to match number. */
	test /= radix;
    if (remainder)
	while (test) {
	    out = remainder / test;
	    *p++ = out + ((out >= 10) ? '7' : '0');
	    remainder %= test;
	    test /= radix;
	}
    else
	*p++ = '0';
    *p = 0;
    if (ndigits) {		/* A field width set? */
	if ((p2 + ndigits) >= p) /* Make sure we never truncate anything. */
	    return p - ndigits;
	else
	    return p2;
    }
    return buf;			/* No field width, return minimum number. */
}
