/*$
**  vtsubs - basic terminal emulation support routines 
**
**	Synopsis:
**
**	External Function List:
**		VCursor(vtcb, mode)			Enable / Disable cursor display
**		identify()					Respond to terminal identification request
**		addbuffer(vtcb,ch)			Add key character to read buffer
**		rmbuffer(vtcb)				Remove key character from buffer
**		repaint(vtcb,ep)			Redraw screen on update event
**		home()						Home the cursor
**		cul(nlines,scrollflag)		Move cursor up n lines
**		cdl(nlines,scrollfalg)		Move cursor down n lines
**		cbc(nchars)					Move cursor back n chars
**		scroll(dir,dead,open,lines)	Scroll screen area
**		clear(row,col,nchars)		Clear n chars
**		blockclr(row,nrows)			Clear n lines
**		blockfill(row,nrows)		Clear n lines in internal screen buffer
**		insert(nchars)				Insert n chars
**		delete(nchars)				Delete n chars
**
**
**	Notes:
**
**
**  (c) Copyrighted by ISC System Corp., 1988          All rights reserved.
**
*/

#include <stdio.h>
#include <sys/types.h>
#include <dnix/fcodes.h>
#include <dnix/ireq.h>
#include <sys/termio.h>
#include <sys/ttold.h>
#include <isc/window.h>
#include <isc/vterm.h>
#include "vtcb.h"
#include "vt220.h"
#include "protos.h"

#ifndef lint
	static void flushstring(int col, int row);
#endif

extern	struct	vtcb	*Term;
extern	BYTE	*stringptr;
extern	int		stringlen;
extern	BYTE	*vt_data;
extern	int	vt_cnt;

/*eject*/
/*%
**  - 
**
**	Synopsis:
**
**	Usage:
**
**	Parameters:
**
**	Returns:
**
**	Errors:
**
**	Notes:
*/

void
VCursor(vtcb, mode)
struct vtcb *vtcb;
int mode;
{
	LOG("Cursor change: newstate = %d,", mode);
	LOG(" enabled = %d,", vtcb->mode.cursoron);
	LOG(" curstate = %d\n", vtcb->mode.cursordisp);
	if (vtcb->mode.cursoron == FALSE) /* Cursor off? */
		return;			/* Don't bother. */
	if (vtcb->mode.cursordisp == mode) /* Already in requested state? */
		return;			/* Need to do nothing. */

	if (!vtcb->iconified)
	    cursortoggle(vtcb);
	vtcb->mode.cursordisp = mode;
}

void
cursortoggle(vtcb)
struct vtcb *vtcb;
{
	Rect cursor;
	short i, j;

	i = vtcb->col * vtcb->charwide + vtcb->charoffset;
	j = vtcb->row * vtcb->charhigh + TOFFSET;
	SetRect(&cursor, i, j, i + vtcb->charwide, j + vtcb->charhigh);
	InvertRect(&cursor);
}

/*eject*/
/*%
**  - 
**
**	Synopsis:
**
**	Usage:
**
**	Parameters:
**
**	Returns:
**
**	Errors:
**
**	Notes:
*/


static char ident[] = {"\033[?62;8;9;35c"};

void
identify()
{
	char	*temp,ch;
	struct vtcb *term;

	term=Term;
	if (term->keycount + sizeof(ident) < term->maxkeys) {
		temp = ident;
		while (ch = *temp++)
			addbuffer(term, ch);
	}
	if (term->readrq)
		(void) vt220_rdchk(term);
}

/*eject*/
/*%
**  - 
**
**	Synopsis:
**
**	Usage:
**
**	Parameters:
**
**	Returns:
**
**	Errors:
**
**	Notes:
*/

void
addbuffer(vtcb,ch)
struct	vtcb	*vtcb;
char	ch;
{
	if (++vtcb->keycount >= vtcb->maxkeys)
	    vtcb->keycount--;
	else {
	    vtcb->keybuffer[vtcb->keyinp++] = ch;
	    if (vtcb->keyinp >= vtcb->maxkeys)
		vtcb->keyinp = 0;
	}
}

/*eject*/
/*%
**  - 
**
**	Synopsis:
**
**	Usage:
**
**	Parameters:
**
**	Returns:
**
**	Errors:
**
**	Notes:
*/

rmbuffer(vtcb)
struct vtcb *vtcb;
{
	int temp;

	if (vtcb->keycount) {
		temp = (vtcb->keyinp + vtcb->maxkeys - 1) % vtcb->maxkeys;
		if (vtcb->keybuffer[temp] == NL)
			return FALSE;
		vtcb->keyinp = temp;
		vtcb->keycount--;
		return TRUE;
	} else
		return FALSE;
}

/*eject*/
/*%
**  repaint - redraw the current window contents
**
**	Synopsis:
**		scans the area of the screen defined by the update rectangle
**		and builds the required strings to redraw.
**
**	Usage:
**		repaint(vtcb, ep);
**
**	Parameters:
**		vtcb		pointer to virtual terminal control block
**		ep			pointer to event record of update
**
**	Returns:
**		void
**
**	Errors:
**		none
**
**	Notes:
**		Since this is a redisplay, we don't need to ever use srcCopy
**		mode (the dirty area is guaranteed to be white).  We'll switch
**		to srcOr mode for this case as it's about twice as fast, but
**		only if the text face is normalText.
*/

void
repaint(vtcb, ep)
struct vtcb *vtcb;
EventRecord *ep;
{
	short row, col, spacenum, offset;
	short trow, brow, lcol, rcol;
	short fcol, flag;
	BYTE *dispptr, *attptr;
	BYTE ch, att, catt;
	Rect cursor;
	struct vtcb *term;

	Term = term = vtcb;
	/* Figure out minimum to redraw. */
	trow = ((short) (ep->where.edge.top - TOFFSET)) / term->charhigh;
	if (trow < 0)
		trow = 0;
	brow = (short) (ep->where.edge.bottom - TOFFSET) / term->charhigh + 1;
	if (brow > term->mrow)
		brow = term->mrow;
	lcol = ((short) (ep->where.edge.left - vtcb->charoffset))
	    / term->charwide;
	if (lcol < 0)
		lcol = 0;
	rcol = ((short) (ep->where.edge.right - vtcb->charoffset * 2))
	    / term->charwide + 1;
	if (rcol > term->mcol)
		rcol = term->mcol;
	LOG("Update rectangle: trow = %d", trow);
	LOG(" brow = %d", brow);
	LOG("  lcol = %d", lcol);
	LOG("  rcol = %d\n", rcol);
	catt = -1;
	for (row = trow ; row < brow; row++) {
		offset = row * term->mcol + lcol;
		dispptr = &vtcb->display[offset];
		attptr = &vtcb->atts[offset];
		spacenum = 0;
		stringlen = 0;
		stringptr = dispptr;
		flag = FALSE;
		for (col = lcol; col < rcol; col++) {
			att = *attptr++;
			ch = *dispptr++;
			if ((ch > ' ') ||	/* If displayable character (non-blank). */
				((ch == ' ')  && (att & ((notSrcCopy<<4) | underlineText)))) {
				if (catt == att) { /* if attributes are same */
					if (stringlen == 0) { /* If first char of string */
						fcol = col;
						stringptr = dispptr - 1;
					}
					stringlen++;
				} else {	/* Attributes are not the same. */
					if (stringlen) { /* flush current string if present */
						if (flag == TRUE) { /* discard appended spaces */
							stringlen -= spacenum;
						}
						flushstring(fcol, row);
					}
					catt = att; /* Start new string. */
					stringptr = dispptr - 1;
					stringlen = 1;
					fcol = col;
					if (catt == (srcCopy << 4)) /* Set text mode to */
						TextMode(srcOr); /* current attribute. */
					else
						TextMode(catt>>4);
					TextFace(catt&0x0F); /* Set face too. */
				}
				flag = FALSE;	/* We're not collecting trailing spaces. */
				spacenum = 0;
			} else {		/* Not a displayable character. */
				if (flag == TRUE) { /* if currently appending spaces */
					if (spacenum < MAXSPACES) { /* Not yet at max, so */
						stringlen++; /* accumulate it. */
						spacenum++;
					} else {	/* Reached max for append, so discard */
						stringlen -= spacenum; /* and flush. */
						flag = FALSE;
						flushstring(fcol, row);
					}
				} else {	/* Not currently appending spaces. */
					if (stringlen) { /* We have a string already so start */
						if (catt == att) { /* appending spaces to it. */
							flag = TRUE;
							spacenum = 1;
							stringlen++;
						} else { /* Attributes differ, flush it out. */
							flushstring(fcol, row);
							stringptr = dispptr;
						}
					} else {	/* No current string, skip this character. */
						stringptr = dispptr;
						fcol = col;
					}
				}
			}
		}
		if (stringlen) { /* Reached end of row in region with current string */
			if (flag) {
				stringlen -= spacenum;
			}
			flushstring(fcol, row);
		}
	}
	/* Now wipe out the cursor. */
	col = vtcb->col * vtcb->charwide + vtcb->charoffset;
	row = vtcb->row * vtcb->charhigh + TOFFSET;
	SetRect(&cursor, col, row, col + vtcb->charwide, row + vtcb->charhigh);
	EraseRect(&cursor);
	offset = vtcb->row * vtcb->mcol + vtcb->col;
	dispptr = &vtcb->display[offset];
	if (*dispptr > ' ') {
		attptr = &vtcb->atts[offset];
		TextMode(*attptr >> 4);
		TextFace(*attptr & 0x0F);
		Posn(vtcb->col, vtcb->row + 1);
		DrawChar(*dispptr);
	}
	TextMode(term->att >> 4);
	TextFace(term->att & 0x0F);
}

static void
flushstring(col, row)
short col, row;
{
	Posn(col, row + 1);		/* Flush string. */
	DrawText((char*) stringptr, 0, stringlen);
	stringlen = 0;
}

short newscrap = 1;				/* New incoming scrap. */
short newcopy;					/* New outgoing scrap. */
BYTE *scraphandle;
long scraphandlesize;

void
scrapevent(ep)
EventRecord *ep;
{
    extern struct vtcb *infront;
    struct vtcb *vtcb;
    struct vidcopy *vcp;

    LOG("Scrap event for window %X\n", ep->who);
    LOG("Scrap event %d\n", ep->message);
    vcp = 0;
    if (!(vtcb = findvtcb(ep->who, &vcp))) {
	return;
    }	
    if (vcp)			/* Nothing to do for copy windows. */
	return;
    if (!ep->message) {
	newscrap = 1;
	if (infront) {
	    SetPort(ep->who);
	    fetchscrap();
	}
    } else {
	if (newcopy) {
	    SetPort(ep->who);
	    writescrap();
	}
    }
}

void
killscrap()
{
    if (scraphandle) {
	free(scraphandle);
	scraphandle = 0;
    }
    DisableItem(Edit_Menu, 2);
    newscrap = 0;
}

static WmanVrsns wmversn;

void
fetchscrap()
{
    long size;

    killscrap();
    if (!wmversn.clunk)
	GetWManVrsns(&wmversn);
    if (wmversn.clunk >= 0x0304 && (size = GetScrap(0, 'TEXT', 0))) {
	if (scraphandle = (BYTE *) malloc(size)) {
	    GetScrap(scraphandle, 'TEXT', scraphandlesize = size);
	    EnableItem(Edit_Menu, 2);
	}
    }
}

void
writescrap()
{
    if (wmversn.clunk >= 0x0304 && newcopy) {
	ZeroScrap();
	if (scraphandle) {
	    PutScrap(scraphandlesize, 'TEXT', scraphandle);
	}
	PublishScrap(0);
	newcopy = 0;
    }
}

/*
** Convert the selected part of the screen buffer into a bunch of
** newline terminated lines with trailing blanks removed.  Put it
** into the private scrap, and make it available for export.
*/

void
copy()
{
    extern struct vtcb *infront;
    short row, col, srow, scol, erow, ecol;
    short mrow, mcol, flag, i, offset;
    BYTE *dispptr, *p, *q, *outptr, *bufptr, *alloca(), *malloc();
    struct vtcb *term;

    if (!(term = infront) || term->startrow < 0)
	return;
    killscrap();
    srow = term->startrow;
    scol = term->startcol;
    erow = term->endrow;
    ecol = term->endcol;
    mrow = term->mrow;
    mcol = term->mcol;
    bufptr = outptr = alloca(mrow * (short) (mcol + 1)); /* Room for max. */
    if (erow < srow) {		/* Normalize the input range. */
	i = erow;
	erow = srow;
	srow = i;
      swapcol:
	i = ecol;
	ecol = scol;
	scol = i;
    } else if (erow == srow && ecol < scol) {
	goto swapcol;
    }
    for (row = srow; row <= erow; row++) {
	offset = row * mcol + scol;
	dispptr = p = &term->display[offset];
	i = (row == erow) ? ecol : mcol;
	q = &term->display[row * mcol + i];
	i -= scol;
	scol = 0;
	flag = 0;
	while (i--) {		/* Is line blank? */
	    if (*p++ > ' ') {
		flag = 1;
		break;
	    }
	}
	if (flag) {		/* Non-blank line? */
	    while (*--q <= ' ')	/* Yes, strip off trailing spaces. */
		;
	    i = (q - dispptr) + 1;
	    while (i--)		/* Copy out the string. */
		*outptr++ = *dispptr++;
	}
	if (ecol == mcol || row != erow)
	    *outptr++ = '\n';
    }
    i = scraphandlesize = outptr - bufptr;
    newcopy = 1;		/* Set export-on-deactivate flag. */
    if (i && (scraphandle = p = malloc(i))) {
	while (i--)		/* Copy the record to our local scrap. */
	    *p++ = *bufptr++;
	EnableItem(Edit_Menu, 2);
    }
    if (term->vidcopy)
	SetPorts(term->ports);
    endspan(term);
}

void
paste()
{
    char *temp, ch;
    struct vtcb *term;
    int count;
    BYTE *realloc();
    extern struct vtcb *infront;

    if (term = infront) {
	if ((count = (term->keycount + scraphandlesize)) > term->maxkeys) {
	    LOG("Re-allocating to %d bytes\n", count);
	    term->keybuffer = realloc(term->keybuffer, term->maxkeys = count);
	    if (!term->keybuffer)
		term->keybuffer = malloc(term->maxkeys = MAXKEYS);
	}
	LOG("Size %d bytes\n", count);
	LOG("Scrap size %d bytes\n", scraphandlesize);
	if (count = scraphandlesize) {
	    if (term->tty.c_lflag & ECHO) {
		vt220(term, scraphandle, count);
	    }
	    temp = (char *) scraphandle;
	    while (count--) {
		ch = *temp++;
		if (ch == '\n' && !(term->tty.c_iflag & ICRNL))
		    ch = '\r';
		addbuffer(term, ch);
	    }
	}
	if (term->readrq)
	    (void) vt220_rdchk(term);
    }
}

/*eject*/
/*%
**  - 
**
**	Synopsis:
**
**	Usage:
**
**	Parameters:
**
**	Returns:
**
**	Errors:
**
**	Notes:
*/

void
home()			/* Home the cursor. */
{
	struct vtcb *term;

	term=Term;
	term->col  = 0;
	if ( term->mode.origin )
		term->row = term->trow;
	else
		term->row = 0;
	term->mode.pendforwrd = FALSE;
}
/*eject*/
/*%
**  - 
**
**	Synopsis:
**
**	Usage:
**
**	Parameters:
**
**	Returns:
**
**	Errors:
**
**	Notes:
*/

void
cul(nlines, scrollflag)
int nlines;
int scrollflag;
{
    register BYTE *temp, tchar;
    int lcnt, cnt;
    struct vtcb *term;

    term=Term;
    if (term->row < term->trow) { /* Outside scrolling region. */
	term->row -= nlines;
	if (term->row < 0)
	    term->row = 0;
	nlines = 0;
    } else {
	lcnt = term->row  - term->trow;
	if (nlines < lcnt) {
	    term->row -= nlines;
	    nlines = 0;
	} else if (term->row != term->trow) {
	    term->row -= lcnt;
	    nlines -= lcnt;
	}
    }
    if (term->mode.pop_scroll)	/* Check for more data from client, and */
	poll(term);		/* add it in if there is. */
    if (scrollflag && nlines) {	/* Have to scroll something? */
	lcnt = 0;
	if (term->mode.pop_scroll) {
	    LOG(" pop_scrolld check vt_cnt = %d \n", vt_cnt);
	    cnt = vt_cnt;
	    temp = vt_data;
	    while ((cnt > 0) && (term->trow <= (term->row - lcnt))
		   && (lcnt < 6 ) && (lcnt < (term->brow - term->trow))) {
		while (cnt--) {
		    if (((tchar = *temp++) & 0x7F) < 0x20) {
			if (tchar == ESC && *temp == 'M') {
			    lcnt++;
			    break;
			}
			if ((tchar != CR) && (tchar != TAB) && tchar != BS)
			    goto done;
		    }
		}
	    }
	  done:
	    LOG("cul : lcnt = %d, ", lcnt);
	    nlines += lcnt;
	    LOG("lines = %d\n", nlines);
	}
	scroll(DOWN, term->brow - nlines + 1, term->trow, nlines);
	term->row += lcnt;
    }
    term->mode.pendforwrd = FALSE;
}

/*eject*/
/*%
**  - 
**
**	Synopsis:
**
**	Usage:
**
**	Parameters:
**
**	Returns:
**
**	Errors:
**
**	Notes:
*/

void
cdl(nlines, scrollflag)
int nlines;
int scrollflag;
{
    register BYTE *temp, tchar;
    short lcnt,cnt;
    struct vtcb *term;

    term=Term;
    if (term->row > term->brow) { /* Outside scrolling region. */
	term->row += nlines;
	if ( term->row >= term->mrow )
	    term->row = term->mrow-1;
	nlines = 0;
    } else {
	cnt = term->brow - term->row;
	if (nlines <= cnt) {
	    term->row += nlines;
	    nlines = 0;
	} else if (term->row < term->brow) {
	    term->row = term->brow;
	    nlines -= cnt;
	}
    }
    if (term->mode.pop_scroll)	/* Check for more data from client, and */
	poll(term);		/* add it in if there is. */
    if (scrollflag && nlines) {
	lcnt = 0;
	if (term->mode.pop_scroll) {
	    LOG(" pop_scroll check vt_cnt = %d \n", vt_cnt);
	    cnt = vt_cnt;
	    temp = vt_data;
	    while ((cnt > 0) && (term->trow <= (term->row - lcnt))
		   && (lcnt < 3 ) && (lcnt < (term->brow - term->trow))) {
		while (cnt--) {
		    if (((tchar = *temp++) & 0x7F) < 0x20) {
			if (tchar == NL) {
			    lcnt++;
			    break;
			}
			if ((tchar != CR) && (tchar != TAB) && tchar != BS)
			    goto done;
		    }
		}
	    }
	  done:
	    LOG("cdl : lcnt = %d, ", lcnt);
	    nlines += lcnt;
	    LOG("lines = %d\n", nlines);
	}
	scroll(UP, term->trow, term->row-nlines+1, nlines);
	term->row -= lcnt;
    }
    term->mode.pendforwrd = FALSE;
}

/*eject*/
/*%
**  - 
**
**	Synopsis:
**
**	Usage:
**
**	Parameters:
**
**	Returns:
**
**	Errors:
**
**	Notes:
*/

void
cbc(nchars)
int	nchars;
{
	struct vtcb *term;

	term=Term;
	term->col = max(term->col - nchars, 0);
	term->mode.pendforwrd = FALSE;
}

/*eject*/
/*%
**  - 
**
**	Synopsis:
**
**	Usage:
**
**	Parameters:
**
**	Returns:
**
**	Errors:
**
**	Notes:
*/

void
scroll(direction, dead_line, open_line, nlines)
int	direction;	/* Direction to scroll the screen. */
int	dead_line;	/* The row(s) that get clobbered during scroll. */
int	open_line;	/* The row(s) that get 'opened' after scroll. */
int	nlines;		/* Number of lines to scroll. */
{
	long bmcnt, size, offset;
	BYTE	*dest, *src;	/* Address of character. */
	Rect	t_rect;
	struct vtcb *term;

	term=Term;
	LOG("Scroll %d ",direction);
	LOG(" dead = %d",dead_line);
	LOG(" open = %d",open_line);
	LOG(" nlines = %d\n",nlines);

	if (open_line != dead_line) {	/* Don't scroll if only one line. */
		if (direction == UP)
		{
			size = (short) nlines * (short) term->mcol;
			if (term->history && dead_line == 0) { /* Scroll into history? */
				dest = term->history;
				src = dest + size;
				bmcnt = (short) (term->histlines - nlines) * term->mcol;
				BlockMove(src, dest, bmcnt); /* Scroll the history. */
				dest += bmcnt;
				src = term->display;
				BlockMove(src, dest, size); /* Copy into the history. */
			}
			offset = (short) dead_line * term->mcol;
			dest = &term->display[offset];
			src = dest + size;
			bmcnt = (short) (open_line - dead_line) * term->mcol;
			BlockMove(src, dest, bmcnt); /* Move characters. */
			dest = &term->atts[offset];
			src = dest + size;
			BlockMove(src, dest, bmcnt); /* Move attributes. */
			SetRect(&t_rect, term->charoffset,
					TOFFSET + (short) (dead_line) * term->charhigh,
					term->charoffset + /* setup rectangle for scrolling */
					term->charwide * term->mcol,
					TOFFSET + (short) (open_line+nlines) * term->charhigh);
			if (!term->iconified)
				ScrollRect(&t_rect, 0, -(term->charhigh * (short) nlines), -1);
		} else {				/* Direction == DOWN. */
			dest = &term->display[(short) (dead_line + nlines) * term->mcol];
			src = dest - (short) nlines * term->mcol;
			bmcnt = (short) (dead_line - open_line) * term->mcol;
			while (bmcnt--)
				*--dest = *--src;	/* move byte at a time */
			dest = &term->atts[(short) (dead_line + nlines) * term->mcol];
			src = dest - (short) nlines * term->mcol;
			bmcnt = (short) (dead_line - open_line) * term->mcol;
			while (bmcnt--)
				*--dest = *--src;	/* move byte at a time */
								/* Setup rectangle for scrolling. */
			SetRect(&t_rect, term->charoffset,
					TOFFSET + ((short) open_line * term->charhigh),
					term->charoffset + term->charwide * term->mcol,
					TOFFSET + (short) (dead_line + nlines) * term->charhigh);
			if (!term->iconified)
				ScrollRect(&t_rect, 0, (term->charhigh * (short) nlines),-1);
		}
	}
	blockfill(open_line, nlines);
	SetRect(&t_rect, term->charoffset, /* Setup rectangle for erasing. */
			TOFFSET + (short) (open_line) * term->charhigh,
			term->charoffset + term->charwide * term->mcol,
			TOFFSET + (short) (open_line + nlines) * term->charhigh);
	if (!term->iconified)
		ValidRect(&t_rect);		/* Suppress update for known area. */
}
/*eject*/
/*%
**  - 
**
**	Synopsis:
**
**	Usage:
**
**	Parameters:
**
**	Returns:
**
**	Errors:
**
**	Notes:
*/


void
clear(row, col, nchars)
int row, col, nchars;
{
	Rect t_rect;
	BYTE *temp;
	register WORD bcnt;
	short offset;
	struct vtcb *term;

	term=Term;
	if (!term->iconified) {
		SetRect(&t_rect, term->charoffset + col * term->charwide,
		    TOFFSET + (row) * term->charhigh,
		    term->charoffset + term->charwide * (col + nchars),
		    TOFFSET + (row + 1) * term->charhigh); /* Rect for erasing. */
		EraseRect(&t_rect);
	}
	offset = (row * term->mcol) + col;
	temp = &term->display[offset];
	bcnt = nchars;
	while (bcnt--)
		*temp++ = ' ';
	temp = &term->atts[offset];
	bcnt = nchars;
	while (bcnt--)
		*temp++ = srcCopy;
}

/*eject*/
/*%
**  - 
**
**	Synopsis:
**
**	Usage:
**
**	Parameters:
**
**	Returns:
**
**	Errors:
**
**	Notes:
*/

void
blockclr(row, nrows)
int row, nrows;
{
    Rect t_rect;
    struct vtcb *term;

    term = Term;
    if (!term->iconified) { /* Set up rectangle for erasing. */
	SetRect(&t_rect, 0, TOFFSET + (row) * term->charhigh,
		term->mcol * term->fontinfo.widMax + term->charoffset * 2,
		TOFFSET + (row + nrows + 1) * term->charhigh);
	EraseRect(&t_rect);
    }
    blockfill(row, nrows);
}

/*&
**  - 
**
**	Parameters:
**
**	Returns:
**
**	Errors:
*/


void
blockfill(row, nrows)
int row, nrows;
{
	register BYTE *temp;
	register WORD bcnt, bcnt2;
	register short offset;
	struct vtcb *term;

	term=Term;
	offset = row * term->mcol;
	temp = &term->display[offset];
	bcnt = nrows * term->mcol;
	bcnt2 = bcnt;
	while (bcnt2--)
		*temp++ = ' ';
	temp = &term->atts[offset];
	while (bcnt--)
		*temp++ = srcCopy;
}

/*eject*/
/*%
**  - 
**
**	Synopsis:
**
**	Usage:
**
**	Parameters:
**
**	Returns:
**
**	Errors:
**
**	Notes:
*/

void
insert(nchars)
short	nchars;
{
	BYTE	*src,*dest;
	Rect	t_rect;
	short	bmcnt;
	struct vtcb *term;

	term=Term;
	if (!term->iconified) {
		SetRect(&t_rect, term->charoffset+term->charwide*(term->col),
			TOFFSET+(term->row*term->charhigh),	/* setup rect for scrolling */
		    term->charoffset+term->charwide*term->mcol,
		    TOFFSET+(short)(term->row+1)*term->charhigh);

		LOG("Scroll rect l = %d",t_rect.edge.left);
		LOG(" t = %d",t_rect.edge.top);
		LOG(" r = %d",t_rect.edge.right);
		LOG(" b = %d\n",t_rect.edge.bottom);

		ScrollRect(&t_rect, term->charwide*nchars, 0, -1);
		SetRect(&t_rect,term->charoffset+term->charwide*(term->col),
			TOFFSET+(term->row*term->charhigh),	/* Rect for validating known */
		    term->charoffset + term->charwide * (short) (term->col + nchars),
		    TOFFSET+(short)(term->row+1)*term->charhigh); /* clean area. */

		LOG("Validate rect l = %d",t_rect.edge.left);
		LOG(" t = %d",t_rect.edge.top);
		LOG(" r = %d",t_rect.edge.right);
		LOG(" b = %d\n",t_rect.edge.bottom);

		ValidRect(&t_rect);
	}

	dest = &term->display[(short) (term->row + 1) * term->mcol - 1];
	src = dest - nchars;
	bmcnt = term->mcol-((term->col)+nchars);
	if (bmcnt > 0)
		while (bmcnt--)
			*dest-- = *src--;	/* move byte at a time */
	bmcnt = nchars;
	while (bmcnt--)
		*dest-- = ' ';
	dest = &term->atts[(short) (term->row + 1) * term->mcol - 1];
	src = dest - nchars;
	bmcnt = term->mcol-((term->col)+nchars);
	if (bmcnt > 0)
		while (bmcnt--)
			*dest = *src;	/* move byte at a time */
	bmcnt = nchars;
	while (bmcnt--)
		*dest-- = srcCopy;
}


/*eject*/
/*%
**  - 
**
**	Synopsis:
**
**	Usage:
**
**	Parameters:
**
**	Returns:
**
**	Errors:
**
**	Notes:
*/

void
delete(nchars)
short	nchars;
{
	BYTE	*src,*dest;
	Rect	t_rect;
	short	bmcnt;
	struct vtcb *term;

	term=Term;
	if (!term->iconified) {
		SetRect(&t_rect,term->charoffset+term->charwide*term->col,TOFFSET+(term->row*term->charhigh),
		    term->charoffset+term->charwide*term->mcol,TOFFSET+(term->row+1)*term->charhigh);		/* setup rectangle for scrolling */
		ScrollRect(&t_rect, -term->charwide*nchars, 0,-1);	/*	*/
		SetRect(&t_rect,term->charoffset+term->charwide*(term->mcol-nchars),TOFFSET+(term->row*term->charhigh),
		    term->charoffset+term->charwide*term->mcol,TOFFSET+(term->row+1)*term->charhigh);		/* setup rectangle for scrolling */
		ValidRect(&t_rect);
	}
	dest = &term->display[term->row*term->mcol+term->col];
	src = dest + nchars;
	bmcnt = term->mcol-(term->col+nchars);
	while ( bmcnt-- )
		*dest++ = *src++;	/* move byte at a time */
	bmcnt = nchars;
	while ( bmcnt-- )
		*dest = ' ';
	src = &term->atts[term->row*term->mcol+term->col];
	dest = src + nchars;
	bmcnt = term->mcol-(term->col+nchars);
	while ( bmcnt-- )
		*dest++ = *src++;	/* move byte at a time */
	bmcnt = nchars;
	while ( bmcnt-- )
		*dest = srcCopy;
}
/*eject*/
void
Reset_termio(vtcb)
struct vtcb *vtcb;
{
	vtcb->tty.c_iflag = (ICRNL|IXON);
	vtcb->tty.c_oflag = (OPOST|ONLCR);
	vtcb->tty.c_cflag = (CS8|CREAD|B9600|HUPCL);
	vtcb->tty.c_lflag = (ICANON|ISIG|ECHO|ECHOE);
	vtcb->tty.c_line = 0;
	vtcb->tty.c_cc[VINTR] = 0x03;
	vtcb->tty.c_cc[VQUIT] = 0x1C;
	vtcb->tty.c_cc[VERASE] = 0x7F;
	vtcb->tty.c_cc[VKILL] = 0x18;
	vtcb->tty.c_cc[VEOF] = 0x04;
	vtcb->tty.c_cc[VEOL] = 0;
	vtcb->tty.c_cc[VEOL2] = 0;
	vtcb->tty.c_cc[VSWTCH] = CSWTCH;
	vtcb->mode.autocr = TRUE;
}

#if 0
void
simsgttyb(vtcb, rn)
struct vtcb *vtcb;
int rn;
{
	struct sgttyb dummy;

	/*
	** build the sgttyb structure
	*/
	dummy.sg_ispeed = dummy.sg_ospeed = vtcb->tty.c_cflag&CBAUD;
	dummy.sg_erase = vtcb->tty.c_cc[VERASE];
	dummy.sg_kill = vtcb->tty.c_cc[VKILL];
	dummy.sg_flags = 0;
	if (vtcb->tty.c_lflag&ICANON)
	{
		if (vtcb->tty.c_oflag&OPOST)
			dummy.sg_flags |= O_CRMOD;
		else
			dummy.sg_flags |= O_RAW;
	}
	if (vtcb->tty.c_cflag&HUPCL) dummy.sg_flags |= O_HUPCL;
	if (vtcb->tty.c_lflag&XCASE) dummy.sg_flags |= O_LCASE;
	if (vtcb->tty.c_lflag&ECHO) dummy.sg_flags |= O_ECHO;
	if (vtcb->tty.c_cflag&PARODD)
		dummy.sg_flags |= O_ODDP;
	else if (vtcb->tty.c_iflag&INPCK)
		dummy.sg_flags |= O_EVENP;
	else
		dummy.sg_flags |= (O_ODDP|O_EVENP);
	if (vtcb->tty.c_oflag&ONLCR)
	{
		dummy.sg_flags |= O_CRMOD;
		if (vtcb->tty.c_oflag&CR1) dummy.sg_flags |= O_CR1;
		if (vtcb->tty.c_oflag&CR2) dummy.sg_flags |= O_CR2;
	}
	else
	{
		if (vtcb->tty.c_oflag&CR1) dummy.sg_flags |= O_NL1;
		if (vtcb->tty.c_oflag&CR2) dummy.sg_flags |= O_NL2;
	}
	if ((vtcb->tty.c_oflag&TAB3) == TAB3) dummy.sg_flags |= O_XTABS;
	if (vtcb->tty.c_oflag&FFDLY) dummy.sg_flags |= O_VTDELAY;
	if (vtcb->tty.c_oflag&BSDLY) dummy.sg_flags |= O_BSDELAY;

	(void) rdnix(F_UWRITE, rn, &dummy, sizeof(struct sgttyb), 0, 0, 0, 0);
}
#endif

/*eject*/
void
Reset_vt220(vtcb)
struct vtcb *vtcb;
{
	extern void NormDisp();

	vtcb->row = 0;
	vtcb->col = 0;
	vtcb->att = (srcCopy<<4)|normalText;
	vtcb->mode.cursoron = TRUE;
	vtcb->mode.cursordisp = FALSE;				/* not currently displayed on screen */
	vtcb->state = NormDisp;
	vtcb->trow = 0;
	vtcb->brow = vtcb->mrow-1;
	vtcb->charwidth = 1;
	vtcb->SOMask = 0x80;
	vtcb->SIMask = 0x00;
	vtcb->CharMask = 0x00;
	if ( vtcb->tty.c_oflag & OPOST && vtcb->tty.c_oflag & ONLCR )
		vtcb->mode.autocr = TRUE;
	else
		vtcb->mode.autocr = FALSE;
	TextMode(vtcb->att>>4);
	TextFace(vtcb->att&0x0F);
}
