/*
**   WCHAT  --  A WindowMan equivalent of UNIX 'chat'
**
** Version 2.1		01/17/95	Jim Cathey
**	Extended to multiple talkers per session.  Many other small changes
**	and cleanups.
**
** Version 2.0		01/12/95	Jim Cathey
**	Extended to multiple talkers per session.  Many other small changes
**	and cleanups.
**
** Version 1.0		11/xx/90	H. Marc Lewis
**	Original version, only allows two talkers.
*/

#define VERSION " V2.1"

#include <fcntl.h>
#include <malloc.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <isc/window.h>
#include <isc/keycodes.h>

#define MAXUSERS 5		/* Number of panes possible. */
#define LINESIZE 128		/* Number of chars (max) per line. */
#define CH 15			/* Line spacing in pixels for font 0. */

#define SESSION 1
#define ADDUITEM 1
#define QUITITEM 3
#define EDIT 2
#define COPYITEM 1
#define PASTEITEM 2

int CHATwid;
int wmfd;
int Trapqfd;
int portbits;
char *get_user_at(), *mysprintf(), *parseuser();

static Rect winrect;
static Rect slice;
static short left = 20;
static short right = 620;
static short top = 60;
static short bottom = 460;
static char titlbuf[32], *Title = titlbuf;

struct pane {
    char *name;
    Rect slice;			/* The pane's bounding rectangle. */
    Rect scrollslice;		/* The pane's scroll rectangle. */
    Rect lineslice;		/* The last lines's rectangle. */
    Rect titlebar;		/* The title bar. */
    short maxlines;		/* Number of lines in this pane. */
    short x;			/* x,y of bottom line. */
    short y;
    short pixels;		/* Current pixelwidth of last line. */
    short wnid;			/* Associated window/node ID. */
    short cursor;		/* Cursor on/off. */
    short active;		/* TRUE when connected & (maybe) talking. */
    short fd;			/* ConnectGraf's fd. */
    char outband;		/* Last line out-of-band msg? (weird font) */
    char adduser;		/* TRUE when adding user. */
    char *nextch;		/* Points to null-byte at end of last line. */
    char *line[12];
} chat[MAXUSERS];

static struct anode {
    Rect winrect;
    short notified;
    short infront;
    short pane;			/* Associated pane structure. */
    short newscrap;		/* New incoming scrap. */
    short newcopy;		/* New outgoing scrap. */
    char *scraphandle;
    long scraphandlesize;
} node[MAXUSERS];		/* Indexed by Window ID. */

static short users;
static short hasdomain;
static char sysbuffer[128];
char *progname;
char *getenv();

static char anon[] = {"Anonymous"};
static char addusr[] = {"Add user: "};
void myrefresh(), keyevent(), windowevent(), scrapevent(), menuevent();
void activateevent(), fetchscrap(), writescrap(), addmenus(), copy(), paste();

main(argc, argv)
int argc;
char *argv[];
{
    int i, fd;
    short got_user;
    EventRecord *ep;
    char *argp, *cp, *cp2;

    mysprintf(titlbuf, "wchat Session #%d", getpid());
    got_user = 0;
    users = 1;
    node[0].notified = 0;
    node[0].newscrap = 1;
    chat[0].name = get_user_at();
    progname = *argv;
    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    ++argp;
	    switch (*argp++) {
	      case 't':
		if (!*argp)
		    --argc, argp = *++argv;
		Title = argp;	/* Get node title for window. */
		break;

	      case 'V':
	      bailout:
		cp = mysprintf(sysbuffer, "%s: %s\n", progname, VERSION);
		write(2, sysbuffer, cp - sysbuffer);
		exit(0);
		break;

	      case 'x':
		FlushDebug(1);
		break;

	      default:
		usage(progname);
		goto bailout;
	    }
	} else {		/* We want to talk to him. */
	    if (users >= sizeof(chat)/sizeof(chat[0])) {
		cp = mysprintf(sysbuffer, "%s: too many (>4) users!\n",
			       progname);
		write(2, sysbuffer, cp - sysbuffer);
		exit(1);
	    }
	    cp = *argv;
	    got_user = 1;
	    if (!(cp = parseuser(cp)))
		exit(0);
	    chat[users++].name = cp;
	}
    }
    if (!got_user) {
	mysprintf(sysbuffer, "%s: with whom?\n", progname);
	write(2, sysbuffer, StringLen(sysbuffer));
	exit(4);
    }
				/* Now contact all Window Managers. */
    if ((Trapqfd = InitGraf()) < 0) {
	exit(errno);
    }
    wmfd = chat[0].fd = ConnectGraf(0);	/* Start with ours. */
    if (wmfd < 0) {
	cp = mysprintf(sysbuffer, "Unable to contact your Window Manager\n");
	write(2, sysbuffer, cp - sysbuffer);
	exit(errno);
    }
    for (i=1; i<users; i++) {	/* Now do the rest of the conference. */
	if (cp = strchr(chat[i].name, '@')) /* Of form user@sys, */
	    cp++;		/* so take the system part verbatim. */
	else
	    cp = chat[i].name;
	if ((fd = ConnectGraf(cp)) >= 0) { /* Connect to handler/server. */
	    chat[i].fd = fd;
	    if (strchr(cp, '.'))
		hasdomain++;
	} else {
	    cp = mysprintf(sysbuffer, "%s: Unable to connect to machine "
			   "'%s'\n", progname, cp);
	    write(2, sysbuffer, cp - sysbuffer);
	    exit(5);
	}
    }
    domainize();		/* If domain's given domain-ize all names. */

    Event(updateEvt, myrefresh); /* Bind event handlers. */
    Event(windowEvt, windowevent);
    Event(keyEvt, keyevent);
    Event(scrapEvt, scrapevent);
    Event(menuEvt, menuevent);
    Event(activateEvt, activateevent);
    SetRect(&node[0].winrect, left, top, right, bottom);
    CHATwid = chat[0].wnid = NewWindow(&node[0].winrect, Title, showWindow,
				       documentProc, frontWindow,
				       hasGoAway|hasSendToBack|hasDrag, wmfd);
    portbits = 1 << CHATwid;
    SetPort(CHATwid);
    BeginUpdate(CHATwid);	/* Prevent initial update event. */
    EndUpdate(CHATwid);
    addmenus();
    SetKbdXlate(1);
    SetRect(&winrect, 0, 0, right-left, bottom-top);

    for (i=0; i<users; i++)
	setup_structures(i, users, 0);
    draw_panes(0);
    cursor_on(CHATwid);
    for (i=1; i<users; i++)
	copyvideo(i);

	/* This program is event-driven, which means that */
	/* as keystrokes (or other events) come in they drive the logic */
	/* of the program.  This loop simply waits for events to arrive. */

    while (1) {
	ep = GetEvent(anyWindow, noEvent); /* An error if we ever come back. */
	if ((int) ep < 0) {
	    fd = EventError() >> 16;
	    if (fd == wmfd)
		exit(1);
	    for (i=0; i<users; i++) {
		if (fd == chat[i].fd) {
		    killoff(chat[i].wnid, 1);
		}
	    }
	}
    }
}

/*
** Add domains to all node names.
*/
domainize()
{
    int fd, i;
    char *cp, *cp2;

    if (hasdomain) {
	fd = -1;
	for (i=0; i<users; i++) {
	    if (!strchr(cp = chat[i].name, '.')) {
		if (fd < 0) {
		    fd = open("/etc/domainid", O_RDONLY|O_NOCTTY);
		    if (fd < 0)
			break;
		    cp2 = sysbuffer + read(fd, sysbuffer, sizeof sysbuffer - 1);
		    *cp2 = 0;
		    if (cp2 = strchr(sysbuffer, '\n'))
			*cp2 = 0;
		    close(fd);
		}
		cp2 = malloc(StringLen(cp) + StringLen(sysbuffer) + 1);
		strcat(strcpy(cp2, cp), sysbuffer);
		free(cp);
		chat[i].name = cp2;
	    }
	}
    }
}

/*
** Convert to user@machine as best we can.  Make sure result is free-able.
*/
char *
parseuser(username)
char *username;
{
    int fd, i;
    char *nodename, *cp;

    cp = username;
    if (!strchr(username, '@')) { /* Only a name? */
		/* It's a username, try for a registered destination. */
	strcpy(sysbuffer, "/netphys/WHEREIS_");
	strncat(sysbuffer, username, sizeof sysbuffer);
	nodename = username;	/* In case of failure */
	username = anon;	/* try using it raw. */
	if ((fd = open(sysbuffer, O_RDONLY|O_NOCTTY)) >= 0) {
	    if ((i = read(fd, sysbuffer, sizeof(sysbuffer))) > 0) {
		if (i == 1) {	/* Chat was denied. */
		    cp = mysprintf(sysbuffer, "%s: %s is denying "
				   "connections\n", progname,
				   nodename);
		    write(1, sysbuffer, cp - sysbuffer);
		    close(fd);
		    return 0;
		}
		nodename = &sysbuffer[1]; /* It worked, */
		username = cp; /* skip MQ byte. */
		if ((cp = strchr(username, '.')) && !strchr(nodename, '.')) {
		    strncat(&sysbuffer[1], cp, sizeof sysbuffer - 2);
		}
	    }
	    close(fd);
	}
	cp = malloc(StringLen(nodename) + StringLen(username) + 2);
	strcat(strcat(strcpy(cp, username), "@"), nodename);
    } else if (*username == '@') { /* Only a machine name given? */
	cp = malloc(StringLen(anon) + StringLen(username) + 1);
	strcat(strcpy(cp, anon), username);
    } else {
	cp = malloc(StringLen(username) + 1);
	strcpy(cp, username);
    }
    return cp;
}

void
myrefresh(ep)
EventRecord *ep;
{
    SetPort(ep->who);
    BeginUpdate(ep->who);
    draw_panes(1);
    EndUpdate(ep->who);
    SetPorts(portbits);
}

void
keyevent(ep)
EventRecord *ep;
{
    long ch, oldport;
    short wnid;

    if (ep->message & F_KEYUP)
	return;
    if (node[wnid = ep->who].notified) { /* Turn off the notify if the */
	GetPort(&oldport);		/* guy notified is typing. */
	SetPort(wnid);
	Notify(wnid, noNotify);
	node[wnid].notified = 0;
	SetPort(oldport);
    }
    ObscureCursor();		/* Turn off mouse cursor. */
    ch = ep->message & (F_FUNC | F_CTRL | F_ALT | F_KEYMASK);
    if (ch & F_FUNC) {
	if (!(ch & F_ALT)) {
	    switch (ch & F_KEYMASK) {
	      case K_ENTER:
	      case K_NPENTER:
		ch = '\n';
		break;

	      case K_ESC:
	      case K_RESET:
		ch = K_ESC;
		break;

	      case K_BACKSP:
	      case K_BCKSP2:
		ch = K_BACKSP;
		break;

	      default:
bail:
		GetPort(&oldport);
		SetPort(wnid);
		Bell();
		SetPort(oldport);
		return;
	    }
	} else {		/* ALT key's down. */
	    GetPort(&oldport);
	    SetPort(wnid);
	    switch (ch & F_KEYMASK) {
	      case K_UP:
		OffsetRect(&node[wnid].winrect, 0, -16);
		MoveWindow(wnid, node[wnid].winrect.edge.left,
			   node[wnid].winrect.edge.top, 0);
		break;

	      case K_DOWN:
		OffsetRect(&node[wnid].winrect, 0, 16);
		MoveWindow(wnid, node[wnid].winrect.edge.left,
			   node[wnid].winrect.edge.top, 0);
		break;

	      case K_LEFT:
		OffsetRect(&node[wnid].winrect, -16, 0);
		MoveWindow(wnid, node[wnid].winrect.edge.left,
			   node[wnid].winrect.edge.top, 0);
		break;

	      case K_RIGHT:
		OffsetRect(&node[wnid].winrect, 16, 0);
		MoveWindow(wnid, node[wnid].winrect.edge.left,
			   node[wnid].winrect.edge.top, 0);
		break;
	    }
	    SetPort(oldport);
	    return;
	}
    } else if (ch & F_CTRL) {
	if ((oldport = ch & F_KEYMASK) == 'g' || oldport == 'G') {
	    Bell();
	    return;
	}
	if (oldport == 'x' || oldport == 'X') {
	    eraseline(node[wnid].pane);
	    cursor_on(wnid);
	    return;
	}
	if (oldport == 'h' || oldport == 'H') {
	    add_char(K_BACKSP, wnid);
	    return;
	}
	if (oldport == 'd' || oldport == 'D') {
	    goto quit;
	}
	goto bail;
    } else if (ch & F_ALT) {
	goto bail;
    } else {
	ch &= F_KEYMASK;
    }
    if (ch == K_ESC) {		/* Time to quit? */
quit:
	if (wnid == CHATwid)
	    exit(0);
	else
	    killoff(wnid, 0);
    } else {
	add_char(ch, wnid);	/* Normal path. */
    }
}

void
addmenus()
{
    NewMenu(SESSION, "Misc");
    AppendMenu(SESSION, "/NAdd New User\311;-;/QQuit");
    if (users >= sizeof(chat)/sizeof(chat[0]))
	DisableItem(SESSION, ADDUITEM);
    NewMenu(EDIT, "Edit");
    AppendMenu(EDIT, "(/C<XCopy;(/VPaste");
}

/*
** Copy the window to the given system.  The connection to WM has already
** been made (as part of the initial validation logic).
*/
copyvideo(pid)
short pid;			/* Pane ID. */
{
    Rect scratch;
    int wnid;
    char *np;

    SetRect(&scratch, left, top, right, bottom);
    wnid = NewWindow(&scratch, Title, showWindow, documentProc,
		     frontWindow, hasGoAway|hasSendToBack|hasDrag,
		     chat[pid].fd);
    chat[pid].wnid = wnid;
    SetPort(wnid);
    node[wnid].pane = pid;
    node[wnid].winrect = scratch;
    addmenus();
    SetKbdXlate(1);
    Bell();
    Bell();
    Notify(wnid, phoneNotify);	/* In case the terminal is locked. */
    node[wnid].notified = 1;
    BeginUpdate(wnid);		/* Go ahead and draw the window fully */
    draw_panes(0);		/* now, there's no point in waiting */
    cursor_on(wnid);		/* for the initial update event so */
    EndUpdate(wnid);
    chat[pid].active = 1;	/* this will save a little time. */
    node[wnid].newscrap = 1;
    portbits |= (1 << wnid);
    SetPorts(portbits);		/* Now connect to all windows in parallel. */
}


killoff(wnid, msg)
int wnid;			/* Window/Node ID. */
int msg;			/* Message index. */
{
    cursor_off(wnid);
    copyoff(wnid);
    TextFont(7);
    chat[node[wnid].pane].outband = 1;
    add_line(msg ? "Connection failed..." : "User exited...", wnid);
    TextFont(0);
}

copyoff(wnid)
int wnid;			/* Node ID. */
{
    if (chat[node[wnid].pane].active) {
	chat[node[wnid].pane].active = 0;
	SetPort(wnid);
	DisposeWindow(wnid);
	DisconnectGraf(chat[node[wnid].pane].fd);
	portbits &= ~(1 << wnid);
	SetPorts(portbits);
    }
}

void
windowevent(ep)
EventRecord *ep;
{
    long oldport;
    short wnid;

    switch (ep->message & F_KEYMASK) {
      case windowGoaway:
	if (ep->who == CHATwid)
	    exit(0);
	else
	    killoff(ep->who, 0);
	break;

      case windowMove:
	GetPort(&oldport);
	SetPort(wnid = ep->who);
	OffsetRect(&node[wnid].winrect, ep->where.edge.left,
		   ep->where.edge.top);
	MoveWindow(wnid, node[wnid].winrect.edge.left,
		   node[wnid].winrect.edge.top, 0);
	SetPort(oldport);
	break;
    }
}

draw_panes(cursor)
int cursor;			/* TRUE to do cursors too. */
{
    short user;

    FillRect(&winrect, (Pattern *) 45);
    TextFont(0);
    TextFace(normalText);
    TextSize(12);
    for (user=0; user < users; user++) {
	draw_pane(user, cursor);
    }
}

draw_pane(user, cursor)
int user;
int cursor;			/* TRUE to do cursors too. */
{
    Rect slice;
    short is, js, ks, first;
    char ch;
    char buff[80];
    char *sp, *tp;

    slice = chat[user].slice;
    FrameRect(&slice);
    InsetRect(&slice, 1, 1);
    EraseRect(&slice);
    js = chat[user].y;
    FillRect(&chat[user].titlebar, greyPattern);
    sp = chat[user].name;
    tp = buff;
    ks = 2;
    *tp++ = ' ';
    while (*sp) {
	if (*sp == '@') {
	    *tp++ = ' ';
	    *tp++ = 'a';
	    *tp++ = 't';
	    *tp++ = ' ';
	    sp++;
	    ks += 3;
	} else
	    *tp++ = *sp++;
	ks++;
    }
    *tp++ = ' ';
    *tp++ = '\0';
    TextMode(srcCopy);
    DrawJustText(buff, ks, &chat[user].titlebar, centerJustify, centerJustify);
    TextMode(srcOr);
    for (first = 1, is = chat[user].maxlines-1; is >= 0; --is) {
	if (chat[user].line[is]) {
	    MoveTo(chat[user].x, js);
	    if (first && chat[user].outband)
		TextFont(7);
	    DrawString(chat[user].line[is]);
	    if (first && chat[user].outband)
		TextFont(0);
	}
	js -= CH;
	first = 0;
    }
    if (cursor && chat[user].cursor)
	cursor_on(chat[user].wnid);
}

usage(progname)
char *progname;
{
    char buf[400];

    mysprintf(buf, "usage: %s [-V] [-t title] user [...]\n", progname);
    strcat(buf, "\twhere 'title' is the optional window's title,\n");
    strcat(buf, "\tand 'user' is the name of whom to chat with (up to 4).\n");
    strcat(buf, "\tUsername formats are:\n");
    strcat(buf, "\t   'user' (if he's using the 'registerme' program),\n");
    strcat(buf, "\t   'user@machine' to send it to a known machine (display),\n");
    strcat(buf, "\t & '@machine' to send it to anyone at that machine.\n");
    write(2, buf, StringLen(buf));
}

add_line(line, wnid)
char *line;
short wnid;
{
    int pid;

    pid = node[wnid].pane;
    if (LocalStringWidth(theFont, theFace, theSize, line) >=
	chat[pid].lineslice.edge.right - chat[pid].lineslice.edge.left) {
	while (*line)
	    add_char(*line++, wnid);
    } else {
	if (chat[pid].nextch != chat[pid].line[chat[pid].maxlines-1])
	    scroll_up(pid);
	MoveTo(chat[pid].x, chat[pid].y);
	DrawString(line);
	strncpy(chat[pid].line[chat[pid].maxlines-1], line, LINESIZE);
	chat[pid].line[chat[pid].maxlines-1][LINESIZE-1] = 0;
    }
}


add_char(ch, wnid)
char ch;
short wnid;			/* Window/Node ID. */
{
    long oldport;
    short pid, is;
    char *cp;

    pid = node[wnid].pane;
    cursor_off(wnid);
    if (ch == '\n') {		/* Newline. */
	if (chat[pid].adduser) { /* Adding new user? */
	    chat[pid].adduser = 0;
	    chat[pid].pixels = 0;
	    is = chat[pid].maxlines - 1;
	    cp = chat[pid].line[is];
	    if (users < MAXUSERS)
		connectadd(cp);
	    eraseline(pid);
	    TextFont(0);
	    if (users < MAXUSERS) {
		GetPort(&oldport);
		SetPort(wnid);
		EnableItem(SESSION, ADDUITEM);
		SetPort(oldport);
	    }
	    chat[pid].outband = 0;
	} else {
	    scroll_up(pid);
	}
	cursor_on(wnid);
	return;
    }
    if (chat[pid].outband)
	TextFont(7);
    if (ch == K_BACKSP) {	/* Backspace. */
	if (chat[pid].nextch != chat[pid].line[chat[pid].maxlines-1]) {
				/* Then there's a character to remove. */
	    ch = *--chat[pid].nextch;
	    *chat[pid].nextch = 0;
	    is = LocalCharWidth(theFont, theFace, theSize, ch);
	    chat[pid].pixels -= is;
	    erase_char(pid);
	} else {
	    GetPort(&oldport);
	    SetPort(wnid);
	    Bell();		/* Can't back up. */
	    SetPort(oldport);
	}
	cursor_on(wnid);
	if (chat[pid].outband)
	    TextFont(0);
	return;
    }
    is = LocalCharWidth(theFont, theFace, theSize, ch);
    if (isspace(ch) && (chat[pid].pixels + is > chat[pid].lineslice.edge.right
			- chat[pid].lineslice.edge.left - 50)) {
	scroll_up(pid);
	cursor_on(wnid);
	if (chat[pid].outband)
	    TextFont(0);
	return;			/* Don't need to add the space. */
    } else if (!isspace(ch) && (chat[pid].pixels + is > chat[pid].lineslice.edge.right - chat[pid].lineslice.edge.left - 20))
	scroll_up(pid);
    *chat[pid].nextch++ = ch;
    MoveTo(chat[pid].x, chat[pid].y);
    DrawString(chat[pid].line[chat[pid].maxlines-1]);
    chat[pid].pixels += is;
    cursor_on(wnid);
    if (chat[pid].outband)
	TextFont(0);
}

eraseline(pid)
int pid;
{
    int is;

    is = chat[pid].maxlines - 1;
    chat[pid].nextch = chat[pid].line[is];
    memset(chat[pid].line[is], 0, LINESIZE);
    EraseRect(&chat[pid].lineslice);
    chat[pid].pixels = 0;
}

/*
** Add the given user to the session.
*/
connectadd(username)
char *username;
{
    int fd, i, newuser, reuse;
    char *cp, *cp1, *cp2;
    RgnHandle saveclip;

    cp = addusr;
    while (*cp) {		/* Validate the add-user message. */
	if (*cp++ != *username++)
	    return;
    }
    if (!*username || !(cp = parseuser(username)))
	return;			/* In denial! */
    if (username = strchr(cp, '@')) /* It's of form user@sys, */
	username++;		/* so take the system part verbatim. */
    else
	username = cp;
    if ((fd = ConnectGraf(username)) >= 0) { /* Connect to handler/server. */
	if (strchr(username, '.'))
	    hasdomain++;
	domainize();
	reuse = 0;
	newuser = users;
	for (i=0; i<users; i++) { /* Re-connecting a dropout? */
	    if (!(cp1 = strchr(cp, '@'))) /* (Only match display node ID's.) */
		cp1 = cp;
	    if (!(cp2 = strchr(chat[i].name, '@')))
		cp2 = chat[i].name;
	    if (!chat[i].active && !strcmp(cp1, cp2)) {
		newuser = i;
		reuse = 1;
		break;
	    }
	}
	chat[newuser].fd = fd;
	if (!reuse) {
	    chat[newuser].name = cp;
	    for (i=0; i<users+1; i++)
		setup_structures(i, users+1, 1);
	    ++users;
	    draw_panes(1);
	} else {		/* Lose the old "exited" message. */
	    eraseline(newuser);
	    chat[newuser].outband = 0;
	    if (strcmp(cp, chat[newuser].name)) { /* User name changed? */
		free(chat[newuser].name);
		chat[newuser].name = cp;
		saveclip = LocalGetClip();
		ClipRectSect(&chat[newuser].slice);
		TextFont(0);
		TextFace(normalText);
		TextSize(12);
		draw_pane(newuser, 0);
		SetClip(saveclip);
	    }
	}
	copyvideo(newuser);
	cursor_on(chat[newuser].wnid);
	if (users >= MAXUSERS)
	    DisableItem(SESSION, ADDUITEM);
    }
}

scroll_up(pid)
short pid;			/* Pane ID to scroll. */
{
    short is;

    ScrollRect(&chat[pid].scrollslice, 0, -CH, -1);
    ValidRect(&chat[pid].lineslice);
    if (chat[pid].line[0])
	free(chat[pid].line[0]);
    for (is = 0; is < (chat[pid].maxlines - 1); is++) {
	chat[pid].line[is] = chat[pid].line[is+1];
    }
    chat[pid].line[is] = calloc(LINESIZE, 1);
    chat[pid].pixels = 0;
    chat[pid].nextch = chat[pid].line[is];
}

/*
** Calculate pane structure elements based on total number of panes
** and whether or not it was already allocated earlier.
*/
setup_structures(pid, num, preserve)
int pid, num, preserve;
{
    struct pane *pp;
    Rect r;
    int height, oldmax, delta;

    pp = &chat[pid];
    SetRect(&r, 10, 0, right - left - 10, bottom - top - 10);
    height = (r.edge.bottom - r.edge.top) / num;
    r.edge.bottom = r.edge.top + height;
    OffsetRect(&r, 0, pid * height);
    r.edge.top += 10;
    pp->slice = r;
    InsetRect(&r, 1, 1);
    r.edge.top += CH + 1;
    pp->scrollslice = r;
    if (preserve)
	oldmax = pp->maxlines;
    pp->maxlines = (r.edge.bottom - r.edge.top) / CH;
    r.edge.top = r.edge.bottom - CH;
    pp->lineslice = r;
    r.edge.top = pp->slice.edge.top + 1;
    r.edge.bottom = pp->scrollslice.edge.top - 2;
    InsetRect(&r, 2, 2);
    pp->titlebar = r;
    pp->x = pp->lineslice.edge.left + 3;
    pp->y = pp->lineslice.edge.bottom - 5;

    if (preserve && pp->line[oldmax-1]) {
	delta = oldmax - pp->maxlines; /* Preserve last parts of panes. */
	for (height = 0; height < pp->maxlines; height++) {
	    if (pp->line[height])
		free(pp->line[height]);
	    pp->line[height] = pp->line[height + delta];
	    pp->line[height + delta] = 0;
	}
    } else {
	pp->line[pp->maxlines-1] = calloc(LINESIZE, 1);
	pp->pixels = 0;
	pp->nextch = pp->line[pp->maxlines-1];
    }
}


cursor_on(wnid)
short wnid;
{
    do_cursor(wnid, 1);
}

cursor_off(wnid)
short wnid;
{
    do_cursor(wnid, 0);
}

do_cursor(wnid, on)
short wnid;
short on;
{
    Rect cursor;
    int pid;
    short cx, cy;

    pid = node[wnid].pane;
/*    SetPort(wnid);	*/
    cx = chat[pid].x + chat[pid].pixels + 1;
    cy = chat[pid].y;
    chat[pid].cursor = on;
    SetRect(&cursor, cx, cy-9, cx+7, cy+3);
    if (on)
	FillRect(&cursor, blackPattern);
    else
	EraseRect(&cursor);
/*    SetPorts(portbits);	*/
}

erase_char(pid)
short pid;			/* Pane ID. */
{
    Rect cursor;
    short cx, cy;

    cx = chat[pid].x + chat[pid].pixels;
    cy = chat[pid].y;
    SetRect(&cursor, cx, cy-10, cx+12, cy+4);
    EraseRect(&cursor);
}

char *
get_user_at()
{
    char *cp, *id, *cpu;

    if (!(id  = getenv("LOGNAME")))
	id = anon;
    if (!(cpu = getenv("WINDOWS_ON")))
	cpu = "Loose Ends";
    if (cp = strchr(cpu, '\n'))	/* No list of names!  Only primary. */
	*cp = 0;
    cp = malloc(StringLen(id) + StringLen(cpu) + 5);
    strcat(strcat(strcpy(cp, id), " at "), cpu);
    return cp;
}

void
menuevent(ep)
EventRecord *ep;
{
    int menuid, menuitem;
    int i, fd, docopy, dopaste, doadd;
    long oldport;

    dopaste = doadd = docopy = 0;
    GetPort(&oldport);
    SetPort(ep->who);
    BeginMenu(ep->message);
    menuid = ep->message >> 16;
    menuitem = ep->message & 0xFFFF;
    switch (menuid) {
      case SESSION:
	switch (menuitem) {
	  case ADDUITEM:
	    doadd++;
	    DisableItem(SESSION, ADDUITEM);
	    break;

	  case QUITITEM:
	    if (ep->who == CHATwid)
		exit(0);
	    else
		killoff(ep->who, 0);
	    break;
	}
	break;

      case EDIT:
	switch (menuitem) {
	  case COPYITEM:
	    docopy++;
	    break;

	  case PASTEITEM:
	    dopaste++;
	    break;
	}
	break;
    }
    EndMenu(ep->message);
    SetPort(oldport);

    if (docopy)
	copy(ep->who);
    if (dopaste)
	paste(ep->who);
    if (doadd)
	adduser(ep->who);
}

void
activateevent(ep)
EventRecord *ep;
{
    int wnid;

    wnid = ep->who;
    if (ep->message) {		/* Activating? */
	node[wnid].infront = 1;
	if (node[wnid].newscrap)
	    fetchscrap(wnid);
    } else {
	node[wnid].infront = 0;
    }
}

void
scrapevent(ep)
EventRecord *ep;
{
    int wnid;

    wnid = ep->who;
    if (!ep->message) {
	node[wnid].newscrap = 1;
	if (node[wnid].infront)
	    fetchscrap(wnid);
    } else {
	if (node[wnid].newcopy)
	    writescrap(wnid);
    }
}

void
killscrap(wnid)
int wnid;
{
    if (node[wnid].scraphandle) {
	free(node[wnid].scraphandle);
	node[wnid].scraphandle = 0;
    }
    DisableItem(EDIT, PASTEITEM);
    node[wnid].newscrap = 0;
}

void
fetchscrap(wnid)
int wnid;
{
    long size, oldport;

    GetPort(&oldport);
    SetPort(wnid);
    killscrap(wnid);
    if (size = GetScrap(0, 'TEXT', 0)) {
	if (size > 1024)	/* Pasting large amounts of stuff is */
	    size = 1024;	/* stupid here, don't let it happen. */
	if (node[wnid].scraphandle = (char *) malloc(size)) {
	    GetScrap(node[wnid].scraphandle, 'TEXT',
		     node[wnid].scraphandlesize = size);
	    EnableItem(EDIT, PASTEITEM);
	}
    }
    SetPort(oldport);
}

void
writescrap(wnid)
int wnid;
{
#if 0
    long oldport;

    if (node[wnid].newcopy) {
	GetPort(&oldport);
	SetPort(wnid);
	ZeroScrap();
	if (node[wnid].scraphandle) {
	    PutScrap(node[wnid].scraphandlesize, 'TEXT',
		     node[wnid].scraphandle);
	}
	PublishScrap(0);
	node[wnid].newcopy = 0;
	SetPort(oldport);
    }
#endif
}

/*
** Convert the selected part of the screen buffer into a bunch of
** newline terminated lines with trailing blanks removed.  Put it
** into the private scrap, and make it available for export.
*/

void
copy(wnid)
int wnid;
{
#if 0
    extern struct vtcb *infront;
    short row, col, srow, scol, erow, ecol;
    short mrow, mcol, flag, i, offset;
    char *dispptr, *p, *q, *outptr, *bufptr, *alloca();
    struct vtcb *term;

    if (!(term = infront) || term->startrow < 0)
	return;
    killscrap();
    srow = term->startrow;
    scol = term->startcol;
    erow = term->endrow;
    ecol = term->endcol;
    mrow = term->mrow;
    mcol = term->mcol;
    bufptr = outptr = alloca(mrow * (short) (mcol + 1)); /* Room for max. */
    if (erow < srow) {		/* Normalize the input range. */
	i = erow;
	erow = srow;
	srow = i;
      swapcol:
	i = ecol;
	ecol = scol;
	scol = i;
    } else if (erow == srow && ecol < scol) {
	goto swapcol;
    }
    for (row = srow; row <= erow; row++) {
	offset = row * mcol + scol;
	dispptr = p = &term->display[offset];
	i = (row == erow) ? ecol : mcol;
	q = &term->display[row * mcol + i];
	i -= scol;
	scol = 0;
	flag = 0;
	while (i--) {		/* Is line blank? */
	    if (*p++ > ' ') {
		flag = 1;
		break;
	    }
	}
	if (flag) {		/* Non-blank line? */
	    while (*--q <= ' ')	/* Yes, strip off trailing spaces. */
		;
	    i = (q - dispptr) + 1;
	    while (i--)		/* Copy out the string. */
		*outptr++ = *dispptr++;
	}
	if (ecol == mcol || row != erow)
	    *outptr++ = '\n';
    }
    i = scraphandlesize = outptr - bufptr;
    node[wnid].newcopy = 1;	/* Set export-on-deactivate flag. */
    if (i && (scraphandle = p = malloc(i))) {
	while (i--)		/* Copy the record to our local scrap. */
	    *p++ = *bufptr++;
	EnableItem(EDIT, PASTEITEM);
    }
    endspan(term);
#endif
}

void
paste(wnid)
int wnid;
{
    char *temp;
    short count;

    if (count = node[wnid].scraphandlesize) {
	temp = (char *) node[wnid].scraphandle;
	while (count--)
	    add_char(*temp++, wnid);
    }
}

adduser(wnid)
int wnid;
{
    int pid;
    char *cp;

    pid = node[wnid].pane;
    if (!chat[pid].outband && !chat[pid].adduser) {
	if (chat[pid].pixels)
	    add_char('\n', wnid);
	chat[pid].outband = chat[pid].adduser = 1;
	TextFont(7);
	cp = addusr;
	while (*cp)
	    add_char(*cp++, wnid);
	TextFont(0);
    }
}

/*
 ** Sprintf replacement.
 **
 ** Handles format strings containing:
 **
 **   %[.][0..9][0..9]s, %[0][1..9]d, %[0][1..9]x, %[0][1..9]X, %c
 **
 ** It's much smaller than the standard one.
 */

char *
mysprintf(buffer, fmt, arglist)
char *buffer, *fmt;
char *arglist;
{
    register char *p, *q, *p2;
    short ndigits;
    int leadfill;
    char *itoa();
    union arg {
	long lng;
	char *ptr;
    } *argptr;

    argptr = (union arg *) &arglist;
    p = fmt;
    q = buffer;
    while (*p) {		/* While there's some format string. */
	if (*p == '%') {	/* Is it a format control? */
	    p++;		/* Yes, which one (skip over % char)? */
	    if (*p == '.')
		p++;
	    if (*p == '0')
		leadfill = *p++;
	    else
		leadfill = ' ';
	    if (*p >= '1' && *p <= '9') { /* Format output length? */
		ndigits = *p++ - '0';
		if (*p >= '0' && *p <= '9')
		    ndigits = ndigits * 10 + (*p++ - '0');
	    } else
		ndigits = 0;
	    if (*p == 'c')
		*q++ = argptr++->lng;
	    else if (*p == 's') {
		p2 = argptr++->ptr; /* String. */
		if (ndigits) {	/* String width specified? */
		    while (*p2 && ndigits--) /* Yes, copy until we */
			*q++ = *p2++; /* have copied all, or run out. */
		    while (ndigits-- > 0) /* Any field width left? */
			*q++ = ' '; /* Yes, fill with spaces. */
		} else {
		    while (*p2)	/* Copy the inclusion. */
			*q++ = *p2++;
		}
	    } else {
		if (*p == 'd') {
		    p2 = itoa(argptr++->lng, 10, leadfill, ndigits);
		} else if (*p == 'x' || *p == 'X') {
		    p2 = itoa(argptr++->lng, 16, leadfill, ndigits);
		}
		while (*p2)	/* Copy the inclusion. */
		    *q++ = *p2++;
	    }
	    p++;		/* Skip over format control. */
	} else			/* No, it's a literal -- copy it. */
	    *q++ = *p++;
    }
    *q = 0;			/* Null-terminate it. */
    return q;
}

char *
itoa(val, radix, leadfill, ndigits)
int val, radix, leadfill, ndigits;
{
    short i;
    char *p, *p2, out;
    static char buf[20];
    unsigned long test, remainder;

    p = buf;
    remainder = val;
    i = ndigits;		/* Make sure there're at least the number */
    while (i--)			/* of leading whatevers that we've asked for */
	*p++ = leadfill;	/* as the field width. */
    p2 = p;
    if (radix == 16) {		/* Hex output? */
	test = 0x10000000L;
    } else {
	test = 1000000000L;	/* Decimal output? */
	if (val < 0) {
	    remainder = -val;
	    *p++ = '-';
	}
    }
    while (test > remainder)	/* Scale down to match number. */
	test /= radix;
    if (remainder)
	while (test) {
	    out = remainder / test;
	    *p++ = out + ((out >= 10) ? '7' : '0');
	    remainder %= test;
	    test /= radix;
	}
    else
	*p++ = '0';
    *p = 0;
    if (ndigits) {		/* A field width set? */
	if ((p2 + ndigits) >= p) /* Make sure we never truncate. */
	    return p - ndigits;
	else
	    return p2;
    }
    return buf;			/* No field width, return minimum number. */
}
