/*
**
** winidle [-V] [-A] [-x offset] [-y offset] [-s entry] [system-name]
**
** This program displays a bar graph of the systems
** user, sys, wait, and idle time.
**
** Revision History:
**
** Version 1.0  05/12/90 Initial Release
**
** Version 1.1  05/15/90 Fix length of rect for check box.
**                       Add command line option to set sample rate and
**                       activate the alarm function.
**
** Version 1.2  06/14/90 Add my initials to the Version number.
**
** Version 1.3  08/09/90 Fix the widMax stuff.
**
*/
#include <stdio.h>
#include <sys/types.h>
#include <a.out.h>
#include <string.h>
#include <dnix/sysinfo.h>
#undef F_EXEC
#include <dnix/fcodes.h>
#include <isc/window.h>
#include <isc/keycodes.h>
#include <isc/controls.h>

#define TRUE (1)
#define FALSE (0)
#define VOFFSET (3)		/* offset from top */
#define HOFFSET (10)	/* offset from left */
#define TICKSIZE (2)	/* size of tick marks */
#define TICKOFFSET (3)	/* offset added to left edge for tick marks */
#define VERSION "winidle Version 1.3 BWA"

/*
** menu stuff
*/
#define M_SAMPLE (1)

#define ENTRY1 (1)
#define ENTRY2 (2)
#define ENTRY3 (3)
#define ENTRY4 (4)
#define ENTRY5 (5)
#define ENTRY6 (6)
#define ENTRY7 (7)

extern int optind, opterr;
extern char *optarg;

int kbdevent(), menuevent(), mouseevent();
int refreshevent(), windowevent(), activatevent();

/*
** Icon to display when iconified
*/
Icon idlecon = {
	0x00000000,
	0x00000000,
	0x00000000,
	0x00000000,
	0x00000000,
	0x00000000,
	0x00000000,
	0x00000000,
	0x00000000,
	0x7FFFFFFE,
	0x40000012,
	0x400A801E,
	0x40000012,
	0x55FFFE12,
	0x49FF0012,
	0x55800012,
	0x49F00012,
	0x40000012,
	0x40000012,
	0x5CA00012,
	0x55505552,
	0x5CA02A9E,
	0x40000012,
	0x7FFFFFFE,
	0x00000000,
	0x00000000,
	0x00000000,
	0x00000000,
	0x00000000,
	0x00000000,
	0x00000000,
	0x00000000
};

Rect world, screen, check;
FontInfo fontrec;
ControlID checkbox;
struct Sysinfo t1, t2;
double user, system, wait, idle;
int sample_rate = ENTRY1, rate_array[8] = {0, 1, 2, 5, 10, 15, 30, 60};
short window, trapq, wmfd;
short width, height, warning = FALSE, weRicon;

main(argc, argv)
int argc;
char *argv[];
{
	double tdiff;
	int ch;
	short first = TRUE, total, horz, vert, x = 5, y = 45;
	char Title[50];

	/*
	** parse our arg list
	*/
	opterr = FALSE;
	while ( (ch = getopt(argc, argv, "VAx:y:s:")) != EOF )
	{
		switch (ch)
		{
			case 'V':
			puts(VERSION);
			exit(0);
			break;

			case 'A':
			warning = TRUE;
			break;

			case 'x':
			x = atoi(optarg);
			break;

			case 'y':
			y = atoi(optarg);
			break;

			case 's':
			sample_rate = atoi(optarg);
			if (sample_rate < ENTRY1)
				sample_rate = ENTRY1;
			else if (sample_rate > ENTRY7)
				sample_rate = ENTRY7;
			break;

			case '?':
			default:
			puts("usage: winidle [-V] [-A] [-x offset] [-y offset] [-s entry] [system-name]");
			exit(1);
			break;
		}
	}

	/*
	** setup screen
	*/
	trapq = InitGraf();
	if (trapq < 0) exit(1);
	if (optind < argc)
		wmfd = ConnectGraf(argv[optind]);
	else
		wmfd = ConnectGraf(NULL);
	if (wmfd < 0) exit(2);
	gethostname(Title, 50);
	strcat(Title, " : winidle");
	SetRect(&screen, 0, 0, 285, 80);
	world = screen;
	OffsetRect(&world, x, y);
	window = NewWindow(&world, Title, showWindow, documentProc, frontWindow, hasDrag|hasGoAway|hasZoom|hasSendToBack, wmfd);
	weRicon = FALSE;
	SetPort(window);
	PenMode(srcCopy);
	TextMode(srcCopy);
	TextFont(5);
	TextSize(9);
	GetFontInfo(&fontrec);
	width = fontrec.widMax;
	height = fontrec.ascent + fontrec.descent;
	horz = HOFFSET + (width * 6);
	vert = VOFFSET + (height * 6);
	SetRect(&check, horz, vert, (horz + 20 + 110), (vert + 12));
 	InitControls();
	checkbox = NewControl(window, &check, "Idle Time Alarm", showControl, warning, FALSE, TRUE, checkBoxProc, 0);

	/*
	** setup our menu
	*/
	NewMenu(M_SAMPLE, "Sample-Rate");
	AppendMenu(M_SAMPLE, "1;2;5;10;15;30;60");
	CheckItem(M_SAMPLE, sample_rate, TRUE);

	/*
	** setup our events
	*/
	Event(keyEvt, kbdevent);
	Event(menuEvt, menuevent);
	Event(mouseEvt, mouseevent);
	Event(updateEvt, refreshevent);
	Event(windowEvt, windowevent);
	Event(activateEvt, activatevent);
	GetEvent(anyWindow, 1<<activateEvt);
	setscreen();

	tdiff = 0;
	while (TRUE)
	{
		/*
		** get sys info
		*/
		dnix(F_SYSINFO, F_SYS_GETSAR, &t1, sizeof(t1));

		if (first)
			first = FALSE;
		else
		{
			/*
			** compute the new values
			*/
			tdiff = t1.cpu[0] - t2.cpu[0]
				+ t1.cpu[1] - t2.cpu[1]
				+ t1.cpu[2] - t2.cpu[2]
				+ t1.cpu[3] - t2.cpu[3];
			if (tdiff <= 0) continue;

			user = (double)(t1.cpu[1] - t2.cpu[1]) * 100.0 / tdiff;
			system = (double)(t1.cpu[2] - t2.cpu[2]) * 100.0 / tdiff;
			wait = (double)(t1.cpu[3] - t2.cpu[3]) * 100.0 / tdiff;
			idle = (double)(t1.cpu[0] - t2.cpu[0]) * 100.0 / tdiff;

			/*
			** show the new values
			*/
			if ( !weRicon )
			{
				show(1, user, blackPattern);
				show(2, system, blackPattern);
				show(3, wait, blackPattern);
				show(4, idle, greyPattern);
				if (warning)
				{
					if (idle < 0.1) Bell();
				}
				FlushWindow();
			}

		}

		t2 = t1;		/* Age the data */

		/*
		** sleep until the next sample period
		*/
		total = 0;
		while (total < rate_array[sample_rate])
		{
			/*
			** check for events
			*/
			if (CheckEvent(anyWindow, noEvent) == EOF) exit(1);

			/*
			** sleep
			*/
			sleep(1);
			total++;
		}
	}
}

setscreen()
{
	register short horz, vert;

	/*
	** draw the tick marks
	*/
	horz = HOFFSET + (width * 10);
	vert = VOFFSET + height;
	drawticks(horz+TICKOFFSET, vert);
	vert += (height * 4);
	drawticks(horz+TICKOFFSET, vert);

	/*
	** draw the legends
	*/
	horz -= (width * 4);
	vert = VOFFSET + (height * 2);
	MoveTo(horz, vert);
	DrawString("User");
	vert += height;
	MoveTo(horz, vert);
	DrawString("Sys ");
	vert += height;
	MoveTo(horz, vert);
	DrawString("Wait");
	vert += height;
	MoveTo(horz, vert);
	DrawString("Idle");
}

drawticks(horz, vert)
short horz, vert;
{
	/*
	** draw the base line
	*/
	MoveTo(horz, vert);
	LineTo(horz+200, vert);

	/*
	** draw the tick marks
	*/
	MoveTo(horz, vert-TICKSIZE);
	LineTo(horz, vert+TICKSIZE);
	MoveTo(horz+20, vert-TICKSIZE);
	LineTo(horz+20, vert+TICKSIZE);
	MoveTo(horz+40, vert-TICKSIZE);
	LineTo(horz+40, vert+TICKSIZE);
	MoveTo(horz+60, vert-TICKSIZE);
	LineTo(horz+60, vert+TICKSIZE);
	MoveTo(horz+80, vert-TICKSIZE);
	LineTo(horz+80, vert+TICKSIZE);
	MoveTo(horz+100, vert-TICKSIZE);
	LineTo(horz+100, vert+TICKSIZE);
	MoveTo(horz+120, vert-TICKSIZE);
	LineTo(horz+120, vert+TICKSIZE);
	MoveTo(horz+140, vert-TICKSIZE);
	LineTo(horz+140, vert+TICKSIZE);
	MoveTo(horz+160, vert-TICKSIZE);
	LineTo(horz+160, vert+TICKSIZE);
	MoveTo(horz+180, vert-TICKSIZE);
	LineTo(horz+180, vert+TICKSIZE);
	MoveTo(horz+200, vert-TICKSIZE);
	LineTo(horz+200, vert+TICKSIZE);
}

show(entry, amount, pattern)
short entry;
double amount;
Pattern *pattern;
{
	Rect on, off;
	short horz, vert;
	char buffer[10];

	/*
	** display the numeric value
	*/
	sprintf(buffer, "%5.1f", amount);
	horz = HOFFSET;
	vert = VOFFSET + height + (height * entry);
	MoveTo(horz, vert);
	DrawText(buffer, 0, 5);

	/*
	** display the bar graph
	*/
	amount *= 2;
	horz = HOFFSET + (width * 10) + TICKOFFSET;
	vert -= (height - 1);
	SetRect(&on, horz, vert, (horz + amount), (vert + (height - 1)));
	SetRect(&off, (horz + amount), vert, (horz + 200), (vert + (height - 1)));
	FillRect(&on, pattern);
	EraseRect(&off);
}

/*
** event handlers
*/
refreshevent(ep)
EventRecord *ep;
{
	BeginUpdate(ep->who);
	DrawControls(ep->who);
	if (weRicon)
	{
		Rect dummy;

		dummy = screen;
		dummy.edge.right = dummy.edge.left + 32;
		dummy.edge.bottom = dummy.edge.top + 32;
		PlotIcon(&dummy, &idlecon);
	}
	else
	{
		setscreen();
		show(1, user, blackPattern);
		show(2, system, blackPattern);
		show(3, wait, blackPattern);
		show(4, idle, greyPattern);
	}
	EndUpdate(ep->who);
}

menuevent(ep)
EventRecord *ep;
{
	register short mid, item;

	BeginMenu(ep->message);
	mid = ep->message >> 16;
	item = ep->message & 0xFFFF;
	switch(mid)
	{
		case M_SAMPLE:
		if (item != sample_rate)
		{
			CheckItem(M_SAMPLE, sample_rate, FALSE);
			sample_rate = item;
			CheckItem(M_SAMPLE, sample_rate, TRUE);
		}
		break;
	}
	EndMenu(ep->message);
}

kbdevent(ep)
EventRecord *ep;
{
	/*
	** discard all keyup events
	*/
	if (ep->message & F_KEYUP) return;

	/*
	** handle ALT-arrow keys for moving window
	*/
	if (ep->message & F_ALT)
	{
		switch(ep->message & F_KEYMASK)
		{
			case K_UP:
			OffsetRect(&world, 0, -16);
			MoveWindow(ep->who, world.edge.left, world.edge.top, 0);
			break;

			case K_DOWN:
			OffsetRect(&world, 0, 16);
			MoveWindow(ep->who, world.edge.left, world.edge.top, 0);
			break;

			case K_LEFT:
			OffsetRect(&world, -16, 0);
			MoveWindow(ep->who, world.edge.left, world.edge.top, 0);
			break;

			case K_RIGHT:
			OffsetRect(&world, 16, 0);
			MoveWindow(ep->who, world.edge.left, world.edge.top, 0);
			break;

			default:
			Bell();
			break;
		}
	}
	else if (ep->message & F_FUNC)
	{
		switch(ep->message & F_KEYMASK)
		{
			case K_ESC:
			case K_RESET:
			DisposeWindow(ep->who);
			exit(0);
			break;

			case K_F1:
			if (warning)
				warning = FALSE;
			else
				warning = TRUE;
			SetCtlValue(checkbox, warning);
			break;

			default:
			Bell();
			break;
		}
	}
}

mouseevent(ep)
EventRecord *ep;
{
	ControlID cid;

	if (ep->message & F_KEYDOWN)
	{
		/*
		** are we in a control box
		*/
		if (FindControl(&ep->where.corner.topLeft, ep->who, &cid))
		{
			/*
			** find out where we are in the control box
			*/
			switch ( TrackControl(cid, &ep->where.corner.topLeft, (void (*)()) -1) )
			{
				case inCheckBox:
				if (warning)
					warning = FALSE;
				else
					warning = TRUE;
				SetCtlValue(cid, warning);
				break;
			}
		}
	}
}

windowevent(ep)
EventRecord *ep;
{
	switch(ep->message & F_KEYMASK)
	{
		case windowMove:
		OffsetRect(&world, ep->where.edge.left, ep->where.edge.top);
		MoveWindow(ep->who, world.edge.left, world.edge.top, 0);
		break;

		case windowZoom:
		if (weRicon)
		{
			UnIconifyWindow(window);
			weRicon = FALSE;
		}
		else
		{
			IconifyWindow(window);
			weRicon = TRUE;
		}
		break;

		case windowGoaway:
		DisposeWindow(ep->who);
		exit(0);
		break;
	}
}

activatevent(ep)
EventRecord *ep;
{
    if (ep->message & F_KEYMASK)
	    HiliteControl(checkbox, 0);
	else
	    HiliteControl(checkbox, 255);
}

puts(string)
char *string;
{
	char buffer[BUFSIZ];

	strcpy(buffer, string);
	strcat(buffer, "\n");
	write(1, buffer, strlen(buffer));
}
