/*
** Window Manager Setup program.  This program manipulates WM's options.
**
**	V3.4	Jim Cathey	5/12/92
**		Added in Elographics touch screen support.
**
**	V3.3	Jim Cathey	4/16/92
**		Repaired math in touch-screen calibration code.
**
**	V3.2	Jim Cathey	4/7/92
**		Added in touch-screen calibration code.
**
**	V3.1	Jim Cathey	8/6/91
**		Changed WEB message for whiners.  Fixed bug in HSV conversion.
**
**	V3.0	Jim Cathey	1/31/91
**		Added WP6 color settings.
**
**	V2.1	Jim Cathey	1/16/91
**		Added new mouse devices.  De-printf'd for 11K size reduction.
**
**	V2.0	Jim Cathey	4/4/90
**		Fixed HSV color model input, relabeled as HSB (for Hue,
**		Saturation, Brightness).  Defaults to HSB model for color
**		boards with color monitors, RGB model otherwise.
**
**	V1.9	Jim Cathey	3/1/90
**		Try to show that there are reduced color choices when the
**		color planes are reduced (color WP5's).
**
**	V1.8	Jim Cathey	2/9/90
**		Nasty message for WEB.
**
**	V1.7	Jim Cathey	12/13/89
**		Added TextSize calls.  Added version printing.
**
**	V1.6	Jim Cathey	11/15/89
**		Sped up initial screen draw by removing ValidRect calls (which
**		weren't needed since V1.1's mods were made).  Added setup
**		record validation code.
**		
**	V1.5	Jim Cathey	10/18/89
**		Added mouse type changer.
**		
**	V1.4	Jim Cathey	6/19/89
**		(This time for sure!!!)  Save of current settings _still_
**		didn't work properly.  Sigh...
**		
**	V1.3	Jim Cathey	6/6/89
**		Save of current settings _still_ didn't work properly.  Also
**		fixed menu settings on default selection.
**
**	V1.2	Jim Cathey	5/22/89
**		Save of current settings didn't work properly.
**
**	V1.1	Jim Cathey	5/19/89
**		Added iconification stuff.  Also sped up initial screen paint
**		by skipping redundant controls junk on initial screen paint.
**
**	V1.0	Jim Cathey	5/9/89
**
** Color transformations (HSB <-> RGB) from _Color Considerations_ 
** by Lee Baldwin, September 1984 Byte, pp 227.
*/

#include <stdio.h>
#include <sys/utsname.h>
#include <isc/window.h>
#include <isc/keycodes.h>
#include <isc/controls.h>
#include "libwindow.h"
#include "gfcodes.h"
#include "setup.h"

#define VERSION ("V 3.4")

#define OLD_SETUP_VERSION 0x101	/* Old version of setup file. */

struct oldsetup {
    short version;		/* File format version. */
    short clock;		/* Clock format. */
    short menuflash;		/* Number of times to flash item selection. */
    short dblclicktime;		/* Count (in ticks) for mouse double-click. */
    short hibrite;
    short lobrite;
    short volume;
    short dimtime;
    short mousetype;
    struct palette entries[16];
    short recplay;
    char recordfile[16];
};

void RGBxHSV(double R, double G, double B, double *H, double *S, double *V);
void HSVxRGB(double h, double s, double v, double *R, double *G, double *B);

struct {
    Point s;			/* Screen coordinates of point. */
    Point t;			/* Tablet coordinates of point. */
    long sumx, sumy;		/* For averaging multiple hits. */
    short n;
    short up;			/* TRUE if marker is on-screen. */
} calmarks[3];

#define WIDTH (445)		/* Window definitions. */
#define HEIGHT (220)
#define SMALLWIDTH (250)
#define SMALLHEIGHT (20)

#define XCALPOS 20		/* Touch-screen calibration mark offsets. */
#define YCALPOS 40

static int xpos = 12;
static int ypos = 44;
static int width = WIDTH;
static int height = HEIGHT;
static int iconified;
static int calmode;		/* TRUE if calibrating touch screen. */
static int firstactive = 1;

#define SETUPMENU	1
#define CLOCKMENU	2
#define MENUMENU	3
#define CLIKMENU	4
#define MOUSEMENU	5
#define EVENTMENU	6

#define CALMENUITEM	12	/* Item # of Calibration screen. */

Icon myicon = {
	0x00000000, 0x00000000, 0x00000000, 0x01E79E00,
	0x01249266, 0x01249E66, 0x01249E00, 0x01249200,
	0x01249200, 0x01249200, 0x01249200, 0x01E49200,
	0x01E49200, 0x01249200, 0x01249200, 0x01249200,
	0x01279200, 0x01279200, 0x01249200, 0x01249200,
	0x01249200, 0x01249200, 0x01249200, 0x01E79E00,
	0x00000000, 0x3BBAB800, 0x2212A800, 0x3B12B800,
	0x0A12A000, 0x3B93A000, 0x00000000, 0x00000000
};

static char setupfile[40] = "/usr/lib/graphics/setupfiles/";
static char *path;

struct setup setupdata;

struct utsname mysystem;
struct mycontrols {
    ControlID cid;
    short hilval;
} controls[18];
Rect swatches[8];		/* Color swatches. */
Rect punchout;			/* For knocking out swatches that can't be */
				/* used because of reduced color planes. */
short colormodel;		/* 0 = HSV, 1 = RGB. */
short colornum;
int shiftval;			/* For wp5/wp6 numeric mangling. */
HdwrType theHdwrRecord;
WmanVrsns versionrecord;
char WMversions[40];
int window, calwindow;
int web;			/* Flag true when connected to a WEB. */
int wp6;			/* TRUE if WP6 color model. */
int maxval;			/* 15 for wp5, 63 for wp6. */
int hueval;			/* 36 for wp5, 72 for wp6. */
int satval;			/* 25 for wp5, 50 for wp6. */
int testmode;			/* TRUE if castrated. */
int wmfd;
char title[40];

void updateevent();
void windowevent();
void menuevent();
void mouseevent();
void keyevent();
void activatevent();
void MyAction(ControlID theControl, short partCode);

main(argc, argv)
int argc;
char *argv[];
{
    int i, j;
    int trapqfd;
    Rect winrect, button;
    char *argp;

    while (--argc > 0) {
	argp = *++argv;
	if (*argp == '-') {
	    argp++;
	    switch (*argp++) {
	      case 'x':
		xpos = atoi(argp);
		break;
	      case 'y':
		ypos = atoi(argp);
		break;
	      case 't':
		testmode++;
		break;
	    }
	} else {
	    path = argp;
	}
    }
    if ((trapqfd = InitGraf()) < 0) { /* Initialize graphics toolbox. */
	exit(1);
    }
    if ((wmfd = ConnectGraf(path)) < 0) { /* Initialize graphics toolbox. */
	exit(2);
    }
    Event(updateEvt, updateevent);
    Event(windowEvt, windowevent);
    Event(menuEvt, menuevent);
    Event(mouseEvt, mouseevent);
    Event(keyEvt, keyevent);
    Event(activateEvt, activatevent);

    uname(&mysystem);		/* Build settings file name. */
    mysystem.nodename[8] = 0;
    strcat(setupfile, mysystem.nodename);
    strcpy(title, "Window Manager Settings for ");
    strcat(title, mysystem.nodename);
    SetRect(&winrect, xpos, ypos, xpos+WIDTH, ypos+HEIGHT);
    window = NewWindow(&winrect, title, showWindow, documentProc, frontWindow,
		       hasGoAway|hasZoom|hasSendToBack|hasGrow|hasDrag, wmfd);
    SetPort(window);
    GetHdwrType(&theHdwrRecord);
    GetWManVrsns(&versionrecord);
    web = (theHdwrRecord.colormodel == 2
	   || theHdwrRecord.colormodel == 4); /* Are we a WEB? */
    wp6 = (theHdwrRecord.colormodel == 3); /* Are we a WP6? */
    if (wp6) {
	maxval = 63;
	hueval = 72;		/* 5 deg/step for hue. */
	satval = 50;		/* 2% per step for sat/bri. */
	shiftval = 2;
    } else {
	maxval = 15;
	hueval = 36;		/* 10 deg/step for hue. */
	satval = 25;		/* 4% per step for sat/bri. */
	shiftval = 4;
    }
    if (web) {
	mysprintf(WMversions, "WindowWEB Manager V%X.%02X  Windows Server V%X.%02X",
		  versionrecord.wman >> 8, versionrecord.wman & 0xFF,
		  versionrecord.clunk >> 8, versionrecord.clunk & 0xFF);
	firstmenu(1);
    } else {
	mysprintf(WMversions, "Window Manager V%X.%02X  ClunkDraw V%X.%02X",
		  versionrecord.wman >> 8, versionrecord.wman & 0xFF,
		  versionrecord.clunk >> 8, versionrecord.clunk & 0xFF);
	j = 0;			/* Properly handle monochrome units, and */
	if (theHdwrRecord.nplanes > 1) /* mono monitors on color units. */
	    j = 2 | (theHdwrRecord.montype > 1); /* (Montype > 1 for color.) */

	InitControls();
	colormodel = 0;		/* Default to HSB color model. */
	if (j < 3)
	    colormodel = 1;	/* Use RGB model for mono board or monitor. */
	i = 0;
	SetRect(&button, 10, 40, 60, 60);
	controls[i].cid = NewControl(window, &button, "HSB", showControl,
			     !colormodel, 0, 1, radioButProc, i); /* #0 */
	if (j < 3)
	    HiliteControl(controls[i].cid, controls[i].hilval = 255);
	i++;
	OffsetRect(&button, 0, 20);
	controls[i].cid = NewControl(window, &button, "RGB", showControl,
			     colormodel, 0, 1, radioButProc, i); /* #1 */
	if (j < 2)
	    HiliteControl(controls[i].cid, controls[i].hilval = 255);
	i++;
	/* Scroll bars should be 17 pixels in the skinny direction. */
	SetRect(&button, 82, 20, 99, 110);
	controls[i].cid = NewControl(window, &button, "H/R", showControl,
				     0, 0, maxval, scrollBarProc, i); /* #2 */
	SetCtlAction(controls[i].cid, MyAction);
	SetCtlScrollRate(controls[i].cid, 5);
	if (j < 3)
	    HiliteControl(controls[i].cid, controls[i].hilval = 255);
	i++;
	OffsetRect(&button, 30, 0);
	controls[i].cid = NewControl(window, &button, "S/G", showControl,
				     0, 0, maxval, scrollBarProc, i); /* #3 */
	SetCtlAction(controls[i].cid, MyAction);
	SetCtlScrollRate(controls[i].cid, 5);
	if (j < 2)
	    HiliteControl(controls[i].cid, controls[i].hilval = 255);
	i++;
	OffsetRect(&button, 30, 0);
	controls[i].cid = NewControl(window, &button, "B/B", showControl,
				     0, 0, maxval, scrollBarProc, i); /* #4 */
	SetCtlAction(controls[i].cid, MyAction);
	SetCtlScrollRate(controls[i].cid, 5);
	if (j < 3)
	    HiliteControl(controls[i].cid, controls[i].hilval = 255);
	i++;
	if (wp6) {
	    SetRect(&button, 210, 65, 240, 77);
	} else {
	    SetRect(&button, 210, 80, 240, 92);
	}
	controls[i].cid = NewControl(window, &button, "0", showControl,
				     1, 0, 1, radioButProc, i); /* #5 */
	if (j < 2)
	    HiliteControl(controls[i].cid, controls[i].hilval = 255);
	i++;
	OffsetRect(&button, 60, 0);
	controls[i].cid = NewControl(window, &button, "1", showControl,
				     0, 0, 1, radioButProc, i); /* #6 */
	if (j < 2)
	    HiliteControl(controls[i].cid, controls[i].hilval = 255);
	i++;
	OffsetRect(&button, 60, 0);
	controls[i].cid = NewControl(window, &button, "2", showControl,
				     0, 0, 1, radioButProc, i); /* #7 */
	if (j < 2)
	    HiliteControl(controls[i].cid, controls[i].hilval = 255);
	i++;
	OffsetRect(&button, 60, 0);
	controls[i].cid = NewControl(window, &button, "3", showControl,
				     0, 0, 1, radioButProc, i); /* #8 */
	if (j < 2)
	    HiliteControl(controls[i].cid, controls[i].hilval = 255);
	i++;
	SetRect(&button, 210, 120, 240, 132);
	controls[i].cid = NewControl(window, &button, wp6 ? "4" : "0",
				     showControl,
				     0, 0, 1, radioButProc, i); /* #9 */
	if (j < 2)
	    HiliteControl(controls[i].cid, controls[i].hilval = 255);
	i++;
	OffsetRect(&button, 60, 0);
	controls[i].cid = NewControl(window, &button, wp6 ? "5" : "1",
				     showControl,
				     0, 0, 1, radioButProc, i); /* #10 */
	if (j < 2)
	    HiliteControl(controls[i].cid, controls[i].hilval = 255);
	i++;
	OffsetRect(&button, 60, 0);
	controls[i].cid = NewControl(window, &button, wp6 ? "6" : "2",
				     showControl,
				     0, 0, 1, radioButProc, i); /* #11 */
	if (j < 2)
	    HiliteControl(controls[i].cid, controls[i].hilval = 255);
	i++;
	OffsetRect(&button, 60, 0);
	controls[i].cid = NewControl(window, &button, wp6 ? "7" : "3",
				     showControl,
				     0, 0, 1, radioButProc, i); /* #12 */
	if (j < 2)
	    HiliteControl(controls[i].cid, controls[i].hilval = 255);
	i++;

	/* Scroll bars should be 17 pixels in the skinny direction. */
	SetRect(&button, 100, 150, 214, 167);
	controls[i].cid = NewControl(window, &button,"Brightness", showControl,
				     maxval, 0, maxval, scrollBarProc, i);
	SetCtlAction(controls[i].cid, MyAction); /* #13 */
	SetCtlScrollRate(controls[i].cid, 5);
	i++;

	OffsetRect(&button, 0, 30);
	controls[i].cid = NewControl(window, &button, "Dimness", showControl,
				     0, 0, maxval, scrollBarProc, i); /* #14 */
	SetCtlAction(controls[i].cid, MyAction);
	SetCtlScrollRate(controls[i].cid, 5);
	i++;

	SetRect(&button, 316, 150, 430, 167);
	controls[i].cid = NewControl(window, &button,"Bell Volume",showControl,
				     0, 0, 255, scrollBarProc, i); /* #15 */
	SetCtlAction(controls[i].cid, MyAction);
	SetCtlScrollRate(controls[i].cid, 5);
	i++;

	OffsetRect(&button, 0, 30);
	controls[i].cid = NewControl(window, &button, "Dim Time", showControl,
				     0, 0, 255, scrollBarProc, i); /* #16 */
	SetCtlAction(controls[i].cid, MyAction);
	SetCtlScrollRate(controls[i].cid, 5);
	i++;

	controls[i].cid = 0;		/* End marker. */
	if (wp6) {
	    SetRect(&swatches[0], 200, 30, 240, 60);
	} else {
	    SetRect(&swatches[0], 200, 30, 240, 70);
	}
	swatches[4] = swatches[0];
	swatches[1] = swatches[0];
	OffsetRect(&swatches[4], 0, 55);
	OffsetRect(&swatches[1], 60, 0);
	swatches[5] = swatches[1];
	swatches[2] = swatches[1];
	OffsetRect(&swatches[5], 0, 55);
	OffsetRect(&swatches[2], 60, 0);
	swatches[6] = swatches[2];
	swatches[3] = swatches[2];
	OffsetRect(&swatches[6], 0, 55);
	OffsetRect(&swatches[3], 60, 0);
	swatches[7] = swatches[3];
	OffsetRect(&swatches[7], 0, 55);

	if (theHdwrRecord.nplanes == 4) { /* Color unit (WP5)? */
	    punchout.corner.botRight = swatches[3].corner.botRight;
	    switch (theHdwrRecord.planesinuse) { /* Handle reduced planes. */
	      case 4:
		SetRect(&punchout, 0, 0, 0, 0);
		break;
	      case 3:
		punchout.corner.topLeft = swatches[2].corner.topLeft;
		break;
	      case 2:
	      case 1:
		punchout.corner.topLeft = swatches[1].corner.topLeft;
		break;
	    }
	}
	defaultsetup();		/* In case readsetup fails. */
	readsetup();		/* Get the current settings. */
	setsliders();		/* Properly display these values. */
	setcolindex(5, 0, 3);
	dosetup();		/* Send them to WM. */
    }
    if (GetEvent(anyWindow, noEvent)) /* Process user interaction. */
	return 1;		/* Continues after a ReleaseEvent (rets 0). */
				/* GetEvent returns EOF on connection error. */

    FlushTwice();		/* We need to make sure that all of our */
    savecurrent();		/* data is actually delivered.  If we quit */
				/* too soon, Netman might not deliver our */
				/* last packet. */
    return 0;
}

/*
** On a color WP5 the palette model is (using Karnaugh map notation):
**
**       \ Plane 1
**        \
** Plane 3 \ Plane 0
**          \
**   Plane 2 \   00      01      11      10         Color 'Name'
**            +-------+-------+-------+-------+    ==============
**         00 | 0     | 1     | 3     | 2     |    |
**            |       |       |       |       |    | Normal
**            | Black | White | Black |   X   |    |
**            +-------+-------+-------+-------+    |-------------
**            | 4     | 5     | 7     | 6     |    |
**         01 |       |       |       |       |    | Highlight #1
**            |   X   | Hil 1 | Black |   X   |    |
**            +-------+-------+-------+-------+    |-------------
**            | 12    | 13    | 15    | 14    |    |
**         11 |       |       |       |       |    | Highlight #3
**            |   X   | Hil 1 | Hil 3 |   X   |    |
**            +-------+-------+-------+-------+    |-------------
**            | 8     | 9     | 11    | 10    |    |
**         10 |       |       |       |       |    | Highlight #2
**            |   X   | Hil 2 | Black |   X   |    |
**            +-------+-------+-------+-------+    +-------------
**                   /        |        \
**      Don't Care  /         |         \  Don't Care
**                 /          |          \
**                /           |           \
**               /            |            \
**               | Background | Foreground |
**               |   Color    |   Color    |
**
** Highlight #1 is defaulted to a beige, Highlight #2 is a pastel green.
** Highlight #3 is beige, but the foreground becomes a red.
**
** Plane #0 is filled with 1's at all times.  (Because the hardware requires
** color #0 to be black, and because we want 1 bits in the drawing plane to 
** also be black in order to match the monochrome video, we have to draw in a
** different plane than #0.)  Plane #1 is the normal drawing plane (equivalent
** to the monochrome's only plane).  Planes #2 and #3 are only used by the
** HiliteRect and HiliteClear calls in the library.  
*/

static struct initcolor {	/* Palette initializer. */
    unsigned char index;
    unsigned char red;
    unsigned char green;
    unsigned char blue;
} colors[] = {{1, 13, 13, 13},	/* One is white. */
	      {5, 13, 11, 10},	/* Five is beige. */
	      {9, 11, 13, 9},	/* Nine is green-ish. */
	      {13, 13, 11, 10},	/* Thirteen is beige. */
	      {15, 15, 0, 0},	/* Fifteen is bright red. */
	      {0, 0, 0, 0}	/* The rest are all black. */
};

static struct wp6color {	/* WP6 Palette initializer. */
    unsigned char red;
    unsigned char green;
    unsigned char blue;
} wp6colors[] = {
    {0x3F,0x3F,0x3F},{0x00,0x00,0x00},{0x3F,0x20,0x20},{0x34,0x2C,0x28},
    {0x2C,0x34,0x24},{0x37,0x37,0x37},{0x3F,0x2F,0x00},{0x10,0x3F,0x3F},
    {0x3F,0x2A,0x00},{0x00,0x25,0x00},{0x1d,0x2A,0x3F},{0x2A,0x10,0x3F},
    {0x2e,0x1d,0x04},{0x19,0x19,0x19},{0x21,0x21,0x21},{0x2A,0x2A,0x2A}
};


defaultsetup()
{
    register short i;
    register struct palette *p;
    register struct initcolor *q;
    register struct wp6color *q1;

    setupdata.version = SETUP_VERSION;
    setupdata.clock = CLOCK12HAP; /* Setup default clock setting. */
    setupdata.menuflash = 2;	/* Setup default menu flash number. */
    setupdata.dblclicktime = 30;
    setupdata.mousetype = 2;	/* Parallel mouse. */
    setupdata.tablet.tformA[0][0] = /* 1-1 touch-screen transform matrix. */
	setupdata.tablet.tformA[1][1] = Int2Fix(1);
    setupdata.tablet.tformA[1][0] = setupdata.tablet.tformA[0][1] =
	setupdata.tablet.tformB[0] = setupdata.tablet.tformB[1] = 0;
    p = setupdata.entries;	/* Setup default color palette. */
    q = colors;
    q1 = wp6colors;
    for (i=0; i<16; i++) {
	p->index = i;
	if (wp6) {
	    p->red = q1->red << shiftval;
	    p->green = q1->green << shiftval;
	    p->blue = q1->blue << shiftval;
	    q1++;
	} else {
	    if (i == q->index) {
		p->red = q->red << shiftval;
		p->green = q->green << shiftval;
		p->blue = q->blue << shiftval;
		q++;
	    } else
		p->red = p->green = p->blue = 0;
	}
	p++;
    }
    setupdata.recplay = NORECORDPLAYBACK; /* Default playback mode. */
    setupdata.hibrite = theHdwrRecord.nplanes > 1 ? maxval : maxval - 2;
    setupdata.lobrite = (maxval / 2) - 1;
    setupdata.volume = 250;
    setupdata.dimtime = 60;
    newmenus();
}

setsliders()
{
    redocontrols();
    SetCtlValue(controls[13].cid, setupdata.hibrite);
    SetCtlValue(controls[14].cid, setupdata.lobrite);
    SetCtlValue(controls[15].cid, setupdata.volume);
    SetCtlValue(controls[16].cid, setupdata.dimtime);
    drawdimtime();
}

readsetup()
{
    int fd, i;
    union {
	struct setup new;
	struct oldsetup old;
    } tempsetup;
    struct palette *p, *q;

    if ((fd = open(setupfile, 0)) >= 0) { /* First read the setupfile. */
	lseek(fd, sizeof(setupdata), 0); /* Skip over power-up set. */
	i = read(fd, &tempsetup, sizeof(setupdata)); /* Read working set. */
	close(fd);
	if (i ==sizeof(struct setup) && tempsetup.new.version == SETUP_VERSION)
	    setupdata = tempsetup.new;
	else if ((fd = open(setupfile, 0)) >= 0) { /* Try old version file. */
	    lseek(fd, sizeof(struct oldsetup), 0);
	    i = read(fd, &tempsetup, sizeof(struct oldsetup));
	    close(fd);
	    if (i ==sizeof(struct oldsetup)
		&& tempsetup.old.version == OLD_SETUP_VERSION) {
		setupdata.version = SETUP_VERSION;
		setupdata.clock = tempsetup.old.clock;
		setupdata.menuflash = tempsetup.old.menuflash;
		setupdata.dblclicktime = tempsetup.old.dblclicktime;
		setupdata.hibrite = tempsetup.old.hibrite;
		setupdata.lobrite = tempsetup.old.lobrite;
		setupdata.volume = tempsetup.old.volume;
		setupdata.dimtime = tempsetup.old.dimtime;
		setupdata.mousetype = tempsetup.old.mousetype;
		setupdata.tablet.tformA[0][0] = 
		    setupdata.tablet.tformA[1][1] = Int2Fix(1);
		setupdata.tablet.tformA[1][0] = setupdata.tablet.tformA[0][1] =
		   setupdata.tablet.tformB[0] = setupdata.tablet.tformB[1] = 0;
		p = setupdata.entries;
		q = tempsetup.old.entries;
		for (fd=0; fd < 16; fd++)
		    *p++ = *q++;
	    }
	}
    }
    setupdata.recplay = NORECORDPLAYBACK;
    newmenus();
}

revertsetup()
{
    int fd, i;
    struct setup tempsetup;

    if ((fd = open(setupfile, 0)) >= 0) { /* First read the setupfile. */
	i = read(fd, &tempsetup, sizeof(setupdata)); /* Read power-up set. */
	close(fd);
	if (i == sizeof(struct setup) && tempsetup.version == SETUP_VERSION)
	    setupdata = tempsetup;
    }
    setupdata.recplay = NORECORDPLAYBACK;
    newmenus();
}

firstmenu(flag)
int flag;
{
    DisposeMenu(SETUPMENU);	/* Then build the menus themselves. */
    NewMenu(SETUPMENU, "Setup");
    AppendMenu(SETUPMENU, "/DDefault;Revert To Saved;/SSave;-;/QQuit");
    if (flag) {
	DisableItem(SETUPMENU, 1);
	DisableItem(SETUPMENU, 2);
	DisableItem(SETUPMENU, 3);
    }
}

newmenus()
{				/* Build the menus themselves. */
    firstmenu(0);
    DisposeMenu(CLOCKMENU);
    NewMenu(CLOCKMENU, "Clock");
    AppendMenu(CLOCKMENU, "Disable;-;23:59:59;23:59;12:59:59;12:59;12:59am");
    if (setupdata.clock) {
	CheckItem(CLOCKMENU, setupdata.clock + 2, 1);
    }
    NewMenu(MENUMENU, "Menu");
    AppendMenu(MENUMENU, "Flash 0;Flash 1;Flash 2;Flash 3");
    CheckItem(MENUMENU, setupdata.menuflash + 1, 1);
    NewMenu(CLIKMENU, "Double-Click");
    AppendMenu(CLIKMENU, "Fast;Medium;Slow");
    CheckItem(CLIKMENU, setupdata.dblclicktime / 10, 1);
    NewMenu(MOUSEMENU, "Mouse");
    AppendMenu(MOUSEMENU, "None;-;Parallel;Tablet (on A);Tablet (on B)");
    AppendMenu(MOUSEMENU, "Keyboard;");
    AppendMenu(MOUSEMENU, "ISI Touch Scrn (A);ISI Touch Scrn (B)");
    AppendMenu(MOUSEMENU, "ELO Touch Scrn (A);ELO Touch Scrn (B)");
    AppendMenu(MOUSEMENU, "-;(Touch Scrn Cal\311");
    if (setupdata.mousetype >= 6 && setupdata.mousetype <= 9)
	EnableItem(MOUSEMENU, CALMENUITEM);
    CheckItem(MOUSEMENU, setupdata.mousetype + 1, 1);
    NewMenu(EVENTMENU, "Events");
    AppendMenu(EVENTMENU, "Disable;-;Record;Playback;Playloop");
}

/*
** File is composed of two setup records -- the 'saved' one followed by the
** current one.  The WM itself copies the first over the second at startup.
** Thereafter we play mostly with the second, but a save operation dupes the
** second over the first.
*/

savesetup()
{
    int fd;

    if (testmode || web)
	return;
    if ((fd = creat(setupfile, 0644)) >= 0) {
	write(fd, &setupdata, sizeof(setupdata));
	write(fd, &setupdata, sizeof(setupdata));
	close(fd);
    } else
	Bell();
}

savecurrent()
{
    int fd;

    if (testmode || web)
	return;
    if ((fd = open(setupfile, 2)) >= 0) {
	lseek(fd, sizeof(setupdata), 0); /* Skip over power-up settings. */
	write(fd, &setupdata, sizeof(setupdata));
	close(fd);
    } else
	Bell();
}

/*
** Actually send the setup record to the WM.
*/

dosetup()
{
    register unsigned char *cp;

    if (testmode || web)
	return;
    if (_wmbufptr + 2 * sizeof(short) + sizeof(struct setup) > _wmbufend) {
	FlushWindow();
    }
    cp = _wmbufptr;
    *((short *)cp)++ = sizeof(short) + sizeof(setupdata);
    *((short *)cp)++ = GF_SystemSetup;
    *((struct setup *)cp)++ = setupdata;
    _wmbufptr = cp;
}

draw_icon()
{
    Rect scratch;

    SetRect(&scratch, 0, 0, 32, 32);
    PlotIcon(&scratch, &myicon);
    ValidRect(&scratch);
}

void
updateevent(ep)
EventRecord *ep;
{
    SetPort(ep->who);
    BeginUpdate(ep->who);
    if (calmode)
	drawcal(ep);
    else
	drawnorm(ep);
    EndUpdate(ep->who);
}

drawcal(ep)
EventRecord *ep;
{
    int i, h, v, dv;

    h = 110;
    v = 140;
    dv = 14;
    MoveTo(h, v); v += 2 * dv;
    DrawString("Calibrating the Touch Screen\311");

    MoveTo(h, v); v += dv;
    DrawString("Press each of the three crossmarks on the screen.");
    MoveTo(h, v); v += dv;
    DrawString("Try to be accurate, because errors here will affect");
    MoveTo(h, v); v += dv;
    TextFace(italicText);
    DrawString("all ");
    TextFace(normalText);
    DrawString("subsequent mouse cursor operations.  (You can");
    MoveTo(h, v); v += dv;
    DrawString("always re-run the calibration if you need to.)  To");
    MoveTo(h, v); v += dv;
    DrawString("help eliminate sample jitter, this calibration will");
    MoveTo(h, v); v += dv;
    DrawString("average touch screen samples for so long as you");
    MoveTo(h, v); v += dv;
    DrawString("press on any crossmark.  So, a leisurely pace will");
    MoveTo(h, v); v += 2 * dv;
    DrawString("probably yield the best results.");

    MoveTo(h, v); v += dv;
    DrawString("You will probably want to touch ");
    TextFace(italicText);
    DrawString("below ");
    TextFace(normalText);
    DrawString("the crossmarks");
    MoveTo(h, v); v += dv;
    DrawString("so that the mouse cursor doesn\325t hide under your finger.");
    MoveTo(h, v); v += dv;
    DrawString("If you choose to touch crossmarks off center, make sure");
    MoveTo(h, v); v += 2 * dv;
    DrawString("you use the same offset for all of them.");

    MoveTo(h, v); v += dv;
    DrawString("Press any key to abort the calibration sequence.");

    for (i=0; i < (sizeof(calmarks)/sizeof(calmarks[0])); i++) {
	if (calmarks[i].up == 2)
	    markat(&calmarks[i].s);
    }
}

markat(p)			/* Draw a cross-mark. */
Point *p;
{
    MoveTo(p->h - 10, p->v);
    Line(20, 0);
    MoveTo(p->h, p->v - 10);
    Line(0, 20);
}

killmarkat(p)
Point *p;
{
    Rect eraser;

    SetRect(&eraser, p->h - 10, p->v - 10, p->h + 21, p->v + 21);
    EraseRect(&eraser);
}

drawnorm(ep)
EventRecord *ep;
{
    if (iconified) {
	draw_icon();
    } else {
	if (web) {
	    TextFont(3);
	    TextSize(9);
	    MoveTo(294, 216);
	    DrawString(WMversions);
	    MoveTo(416, 9);
	    DrawString(VERSION);
	    TextFont(0);
	    TextSize(12);
	    MoveTo(22, 100);
	    DrawString("Sorry, this doesn\325t work on a WEB.  "
		       "(It\325s not supposed to.)");
	} else {
	    if (!firstactive)	/* Skip first redraw of controls because */
		DrawControls(ep->who); /* all controls were created drawn. */
	    TextFace(underlineText);
	    MoveTo(12, 12);
	    DrawString("Color");
	    MoveTo(12, 26);
	    DrawString("Model");
	    TextFace(normalText);
	    if (!wp6) {
		MoveTo(175, 101);
		DrawString("Fore-");
		MoveTo(175, 117);
		DrawString("Back-");
		Move(1, -9);
		DrawString("ground");
	    }
	    MoveTo(12, 162);
	    DrawString("Bright Level");
	    MoveTo(12, 192);
	    DrawString("Dim Level");
	    MoveTo(236, 162);
	    DrawString("Bell Volume");
	    MoveTo(236, 192);
	    DrawString("Dim Time");
	    TextFont(3);
	    TextSize(9);
	    MoveTo(55, 17);
	    DrawString("F1");
	    MoveTo(85, 14);
	    DrawString("F2");
	    MoveTo(115, 14);
	    DrawString("F3");
	    MoveTo(145, 14);
	    DrawString("F4");
	    MoveTo(12, 151);
	    DrawString("F5");
	    MoveTo(236, 151);
	    DrawString("F6");
	    MoveTo(12, 181);
	    DrawString("F7");
	    MoveTo(236, 181);
	    DrawString("F8");
	    MoveTo(6, 216);
	    DrawString("Shifted F-key runs slider the other way.");
	    Move(80, 0);
	    DrawString(WMversions);
	    MoveTo(416, 9);
	    DrawString(VERSION);
	    TextFont(0);
	    TextSize(12);
	    drawdimtime();
	    drawlabels();
	    drawswatches();
	}
	DrawGrowIcon(ep->who);
    }
}

/*
** Draw the color swatches.
*/

drawswatches()
{
    Rect scratch;
    short i;
    char *label = "Color Number";

    scratch.edge.top = swatches[0].edge.top - 26;
    scratch.edge.bottom = swatches[0].edge.top - 10;
    scratch.edge.left = swatches[0].edge.left;
    scratch.edge.right = swatches[3].edge.right;
    TextFace(underlineText);
    DrawJustText(label, StringLen(label), &scratch,
		 centerJustify, centerJustify);
    TextFace(normalText);
    if (wp6) {
	for (i=0; i<8; i++) {
	    scratch = swatches[i];
	    BackColor(i);
	    EraseRect(&scratch);
	    FrameRect(&scratch);
	}
	BackColor(0);
    } else {
	for (i=0; i<4; i++) {
	    scratch = swatches[i];
	    FrameRect(&scratch);
	    InsetRect(&scratch, 1, 1);
	    HiliteRect(&scratch, blackPattern, i);
	    FillRect(&scratch, greyPattern);
	    scratch.edge.bottom -= (scratch.edge.bottom - scratch.edge.top)/2;
	    EraseRect(&scratch);
	    DrawJustText("Text", 4, &scratch, centerJustify, centerJustify);
	}
	PenMode(srcBic);	/* Knock out inactive swatches. */
	PenPat(greyPattern);
	PaintRect(&punchout);
	HiliteRect(&punchout, whitePattern, 3);
    }
    PenNormal();
}

/*
** Label the sliders with the right labels for the input color model in use.
*/

drawlabels()
{
    Rect scratch;

    SetRect(&scratch, 79, 110, 166, 122);
    EraseRect(&scratch);
    MoveTo(80, 122);
    DrawText("HueRed", colormodel * 3, 3);
    MoveTo(110, 122);
    DrawText("SatGrn", colormodel * 3, 3);
    MoveTo(141, 122);
    DrawText("BriBlu", colormodel * 3, 3);
}

drawdimtime()
{
    char buffer[16];
    Rect eraser;

    if (setupdata.dimtime == 2)
	strcpy(buffer, "1 minute");
    else
	mysprintf(buffer, (setupdata.dimtime & 1) ? "%d.5 minutes"
		  : "%d minutes", setupdata.dimtime >> 1);
    SetRect(&eraser, 236, 194, 316, 204);
    EraseRect(&eraser);
    MoveTo(236, 203);
    TextFont(1);
    TextSize(12);
    DrawString(buffer);
    TextFont(0);
}

void
windowevent(ep)
EventRecord *ep;
{
    Rect rect;

    SetPort(ep->who);
    switch (ep->message & F_KEYMASK) {
      case windowGoaway:
	ReleaseEvent();
	break;

      case windowMove:
	xpos += ep->where.edge.left;
	ypos += ep->where.edge.top;
	MoveWindow(ep->who, xpos, ypos, 0);
	break;

      case windowResize:
	SetRect(&rect, width - 20, height - 20, width, height);
	EraseRect(&rect);
	InvalRect(&rect);
	width += ep->where.edge.left;
	height += ep->where.edge.top;
	SizeWindow(ep->who, width, height, updateRequest);
	DrawGrowIcon(ep->who);
	break;

      case windowZoom:
	if (iconified) {
	    UnIconifyWindow(ep->who);
	    enablemenus();
	} else {
	    HiliteClear();
	    IconifyWindow(ep->who);
	    draw_icon();
	    disablemenus();
	}
	iconified = !iconified;
	break;
    }
}

disablemenus()
{
    DisableItem(SETUPMENU, 1);
    DisableItem(SETUPMENU, 2);
    DisableItem(SETUPMENU, 3);
    DisableMenu(CLOCKMENU);
    DisableMenu(MENUMENU);
    DisableMenu(CLIKMENU);
    DisableMenu(MOUSEMENU);
    DisableMenu(EVENTMENU);
}

enablemenus()
{
    EnableItem(SETUPMENU, 1);
    EnableItem(SETUPMENU, 2);
    EnableItem(SETUPMENU, 3);
    EnableMenu(CLOCKMENU);
    EnableMenu(MENUMENU);
    EnableMenu(CLIKMENU);
    EnableMenu(MOUSEMENU);
    EnableMenu(EVENTMENU);
}

/*
** Kjell's matrix transformation code.
**
** The transform() routine is actually in the OS touch-screen driver,
** and is used to convert raw device coordinates into the scaled/rotated
** screen-sized range.  It is included here for informational purposes.
**
** The setuptransform() routine makes the necessary transformation matrix
** to convert one set of coordinates to another.
**
** These routines were lifted from the P+ document alignment program.
*/

#if 0
transform(dest, source)
short *dest, *source;
{
    int i;

    for (i = 0; i < 2; i++) {
	dest[i] = Fix2Int(FixMul(Int2Fix(source[0]), a[i][0])
	    + FixMul(Int2Fix(source[1]), a[i][1]) + b[i]);
    }
}
#endif

/*
** Create a transformation matrix out of three points P, Q, and R,
**
** +-----------------------+
** |                       |
** |   * P            * Q  |
** |                       |
** |                       |
** |                       |
** |                       |
** |                       |
** |                       |
** |                       |
** |                       |
** |                  * R  |
** |                       |
** +-----------------------+
**
** with both input (raw mouse) and output (pixel) space coordinates
** given for all three points.  The matrix will then map input space
** to output space by applying the above transform() function.
**
**                        +=     =+
** The matrix is of form: | A B 0 |   so that: X' = AX + BY + E
**                        | C D 0 |            Y' = CX + DY + F
**                        | E F 1 |
**                        +=     =+
**
** This is a standard two-dimensional coordinate transformation matrix that
** can do transformation (shifting), scaling, and rotation.  Coefficients
** A-F are stored in a 2x2 matrix for A-D (matrix A), and a separate 1x2
** matrix for E-F (matrix B) for space efficiency.  Coefficients A-F are
** in fixed-point (16.16) format for speed/space reasons (inside the
** touch-screen driver).
*/

setuptransform(dest, source)
short dest[][2];		/* Pixel, [0][] == P, [1][] == Q, [2][] == R */
short source[][2];		/* Raw mouse, ditto. */
{				/* [][0] == vert, [][1] == horz. */
    int i;
    float a0, a1, b0, d1, tmp;

    d1 = (source[0][0] - source[1][0]);
    d1 /= (source[0][1] - source[1][1]);
    tmp = (source[0][0] - source[2][0]);
    tmp /= (source[0][1] - source[2][1]);
    d1 -= tmp;
    for (i = 0; i < 2; i++) {
	a1 = (dest[0][i] - dest[1][i]);
	a1 /= (source[0][1] - source[1][1]);
	tmp = (dest[0][i] - dest[2][i]);
	tmp /= (source[0][1] - source[2][1]);
	a1 -= tmp;
	a1 /= d1;
	a0 = (dest[0][i] - dest[1][i]) - a1 * (source[0][0] - source[1][0]);
	a0 /= (source[0][1] - source[1][1]);
	b0 = dest[0][i] - a0 * source[0][1] - a1 * source[0][0];
	setupdata.tablet.tformA[i][0] = a1 * 65536; /* To fixed point... */
	setupdata.tablet.tformA[i][1] = a0 * 65536;
	setupdata.tablet.tformB[i] = (b0 + .5) * 65536;
    }
}

/*
** Handles mouse events in calibration mode.
*/

void
calmouseevent(ep)
EventRecord *ep;
{
    int i, done, h, v;
    Point dest[3];
    Point source[3];

    h = ep->where.edge.right;	/* Pick up the raw (absolute) coordinates. */
    v = ep->where.edge.bottom;
    i = -1;
    if (h < 400 && v < 400) { /* Point #1 (upper left) */
	i = 0;
    }
    if (h > 800 && v < 400) { /* Point #2 (upper right) */
	i = 1;
    }
    if (h > 800 && v > 800) { /* Point #3 (lower right) */
	i = 2;
    }
    if (i >= 0) {
	if (((ep->message & F_KEYDOWN) && (calmarks[i].up == 2))
	    || ((ep->message & F_MOUSE1) && (calmarks[i].up == 1))) {
	    calmarks[i].sumx += ep->where.edge.right; /* Average them */
	    calmarks[i].sumy += ep->where.edge.bottom; /* to clean up */
	    calmarks[i].n++;	/* jitter from the touch screen. */
	    if (calmarks[i].up == 2) { /* First hit, take down mark but */
		calmarks[i].up = 1; /* allow more averaging on this mark. */
		killmarkat(&calmarks[i].s);
	    }
	} else if (ep->message & F_KEYUP) {
	    calmarks[i].up = 0;	/* Disallow more hits on this mark. */
	    calmarks[i].t.h = calmarks[i].sumx / calmarks[i].n;
	    calmarks[i].t.v = calmarks[i].sumy / calmarks[i].n;

	    done = 1;
	    for (i=0; i < (sizeof(calmarks)/sizeof(calmarks[0])); i++) {
		source[i] = calmarks[i].t;
		dest[i] = calmarks[i].s;
		if (calmarks[i].up)
		    done = 0;
	    }
	    if (done) {
		setuptransform(dest, source);
		dosetup();
		endcalibrate();
	    }
	}
    }
}

startcalibrate()
{
    int i;
    Rect winrect;

    for (i=0; i<3; i++) {
	calmarks[i].up = 2;
	calmarks[i].sumx = calmarks[i].sumy = calmarks[i].n = 0;
    }
    calmarks[0].s.h = XCALPOS;
    calmarks[0].s.v = YCALPOS;
    calmarks[1].s.h = theHdwrRecord.hdots - XCALPOS;
    calmarks[1].s.v = YCALPOS;
    calmarks[2].s.h = theHdwrRecord.hdots - XCALPOS;
    calmarks[2].s.v = theHdwrRecord.vdots - XCALPOS;
    calmode = 1;
    disablemenus();
    SetRect(&winrect, 0, 0, theHdwrRecord.hdots, theHdwrRecord.vdots);
    calwindow = NewWindow(&winrect, title, showWindow, plainDBox, frontWindow,
		       noOptions, wmfd);
    HideWindow(window);
    SetPort(calwindow);
    HideCursor();
    MouseMode(F_KEYUP | F_KEYDOWN | mouseAll);
    Event(mouseEvt, calmouseevent);
}

endcalibrate()
{
    calmode = 0;
    ShowWindow(window);
    SelectWindow(window);
    DisposeWindow(calwindow);
    SetPort(window);
    enablemenus();
    ShowCursor();
    MouseMode(F_KEYUP | F_KEYDOWN);
    Event(mouseEvt, mouseevent);
}

void
menuevent(ep)
EventRecord *ep;
{
    int menuid, menuitem, devnum;
    Rect rect;

    SetPort(ep->who);
    BeginMenu(ep->message);
    menuid = ep->message >> 16;
    menuitem = ep->message & 0xFFFF;

    switch (menuid) {
      case SETUPMENU:
	switch (menuitem) {
	  case 1:
	    defaultsetup();
	    setsliders();
	    dosetup();
	    break;

	  case 2:
	    revertsetup();
	    setsliders();
	    dosetup();
	    break;

	  case 3:
	    savesetup();
	    break;

	  case 5:
	    ReleaseEvent();
	    break;
	}
	break;

      case CLOCKMENU:
	if (setupdata.clock) {
	    CheckItem(CLOCKMENU, setupdata.clock + 2, 0);
	}
	setupdata.clock = (menuitem == 1) ? 0 : menuitem - 2;
	if (setupdata.clock) {
	    CheckItem(CLOCKMENU, setupdata.clock + 2, 1);
	}
	dosetup();
	break;

      case MENUMENU:
	CheckItem(MENUMENU, setupdata.menuflash + 1, 0);
	setupdata.menuflash = menuitem - 1;
	CheckItem(MENUMENU, menuitem, 1);
	dosetup();
	break;

      case CLIKMENU:
	CheckItem(CLIKMENU, setupdata.dblclicktime / 10, 0);
	setupdata.dblclicktime = menuitem * 10;
	CheckItem(CLIKMENU, menuitem, 1);
	dosetup();
	break;

      case MOUSEMENU:
	if (menuitem == CALMENUITEM) {
	    startcalibrate();
	} else {
	    CheckItem(MOUSEMENU, setupdata.mousetype + 1, 0);
	    setupdata.mousetype = menuitem - 1;
	    CheckItem(MOUSEMENU, menuitem, 1);
	    if (!testmode)
		unlink("/dev/mouse");
	    if (setupdata.mousetype) {
		switch (menuitem) {
		  case 3:	/* Parallel (Apple) mouse. */
		  default:
		    devnum = 0x0700;
		    break;
		  case 4:	/* Tablet (Summagraphics) port A mouse. */
		    devnum = 0x0710;
		    break;
		  case 5:	/* Tablet (Summagraphics) port B mouse. */
		    devnum = 0x0711;
		    break;
		  case 6:	/* Keyboard mouse (emulation). */
		    devnum = 0x0720;
		    break;
		  case 7:	/* ISI Touch screen port A mouse. */
		    devnum = 0x0730;
		    break;
		  case 8:	/* ISI Touch screen port B mouse. */
		    devnum = 0x0731;
		    break;
		  case 9:	/* Elographics Touch screen port A mouse. */
		    devnum = 0x0732;
		    break;
		  case 10:	/* Elographics Touch screen port B mouse. */
		    devnum = 0x0733;
		    break;
		}
		if (!testmode)
		    mknod("/dev/mouse", 020666, devnum);
	    }
	    if (setupdata.mousetype && menuitem >= 7 && menuitem <= 10)
		EnableItem(MOUSEMENU, CALMENUITEM);
	    else
		DisableItem(MOUSEMENU, CALMENUITEM);
	    dosetup();
	}
	break;

      case EVENTMENU:
	setupdata.recplay = (menuitem == 1) ? 0 : menuitem - 2;
	strcpy(setupdata.recordfile, mysystem.nodename);
	dosetup();
	ReleaseEvent();
	break;
    }
    EndMenu(ep->message);
}

void
mouseevent(ep)
EventRecord *ep;
{
    ControlID p;
    short i;

    SetPort(ep->who);
    if (ep->message & F_KEYDOWN) { /* Mouse-down event? */
	if (iconified) {
	    Bell();
	    return;
	}
	if (FindControl(&ep->where.corner.topLeft, ep->who, &p)) {
	    i = TrackControl(p, &ep->where.corner.topLeft, (void (*)()) -1);
	    if (i == inCheckBox) {
		SetCtlValue(p, 1); /* Select Radio button. */
		if (p == controls[0].cid) {	/* Set HSV Model. */
		    SetCtlValue(controls[1].cid, 0);
		    colormodel = 0;
		    redocontrols();
		    drawlabels();
		}
		if (p == controls[1].cid) {	/* Set RGB Model. */
		    SetCtlValue(controls[0].cid, 0);
		    colormodel = 1;
		    redocontrols();
		    drawlabels();
		}
		if (p == controls[5].cid) /* Set Color #0 foreground. */
		    setcolindex(5, 0, 3);
		else if (p == controls[6].cid) /* Set Color #1 foreground. */
		    setcolindex(6, 1, 7);
		else if (p == controls[7].cid) /* Set Color #2 foreground. */
		    setcolindex(7, 2, 11);
		else if (p == controls[8].cid) /* Set Color #3 foreground. */
		    setcolindex(8, 3, 15);
		else if (p == controls[9].cid) /* Set Color #0 background. */
		    setcolindex(9, 4, 1);
		else if (p == controls[10].cid) /* Set Color #1 background. */
		    setcolindex(10, 5, 5);
		else if (p == controls[11].cid) /* Set Color #2 background. */
		    setcolindex(11, 6, 9);
		else if (p == controls[12].cid) /* Set Color #3 background. */
		    setcolindex(12, 7, 13);
	    }
	    if (i == inThumb) {
		readsliders();
		if (p == controls[16].cid)
		    drawdimtime();
		dosetup();
		if (p == controls[15].cid)
		    Bell();
	    }
	} else
	    Bell();		/* Invalid mouse hit. */
    }
}

				/* Call when changing active palette entry. */
setcolindex(cntl, wp6palnum, palnum)
short cntl, wp6palnum, palnum;
{
    short i;
    struct palette *p;

    for (i=5; i<13; i++) {	/* Select the proper button. */
	SetCtlValue(controls[i].cid, cntl == i);
    }
    colornum = wp6 ? wp6palnum : palnum;
    redocontrols();
}

void
keyevent(ep)
EventRecord *ep;
{
    short key, i;

    if (ep->message & F_KEYUP)
	return;
    if (calmode) {
	endcalibrate();
	return;
    }
    SetPort(ep->who);
    key = ep->message & F_KEYMASK;
    i = (ep->message & F_SHIFT) ? inPageDown : inPageUp;
    if (ep->message & F_FUNC) {
	switch (key) {
	  case K_ESC:
	  case K_RESET:
	    ReleaseEvent();
	    break;

	  case K_UP:
	    if (ep->message & F_ALT) {
		ypos -= 18;
		MoveWindow(window, xpos, ypos, 0);
	    }
	    break;

	  case K_DOWN:
	    if (ep->message & F_ALT) {
		ypos += 18;
		MoveWindow(window, xpos, ypos, 0);
	    }
	    break;

	  case K_LEFT:
	    if (ep->message & F_ALT) {
		xpos -= 18;
		MoveWindow(window, xpos, ypos, 0);
	    }
	    break;

	  case K_RIGHT:
	    if (ep->message & F_ALT) {
		xpos += 18;
		MoveWindow(window, xpos, ypos, 0);
	    }
	    break;

	  default:
	    if (iconified) {
		Bell();
		return;
	    }
	    break;
	}
	if (web)
	    key = 0;
	switch (key) {
	  case K_F1:		/* Toggle Color Model. */
	    SetCtlValue(controls[0].cid, colormodel);
	    colormodel ^= 1;
	    SetCtlValue(controls[1].cid, colormodel);
	    redocontrols();
	    drawlabels();
	    break;

	  case K_F2:
	    MyAction(controls[2].cid, i);
	    break;

	  case K_F3:
	    MyAction(controls[3].cid, i);
	    break;

	  case K_F4:
	    MyAction(controls[4].cid, i);
	    break;

	  case K_F5:
	    MyAction(controls[13].cid, i);
	    break;

	  case K_F6:
	    MyAction(controls[15].cid, i);
	    break;

	  case K_F7:
	    MyAction(controls[14].cid, i);
	    break;

	  case K_F8:
	    MyAction(controls[16].cid, i);
	    break;
	}
    } else {			/* Non-function key. */
	if (web)
	    key = 0;
	switch (key) {
	  case '0':		/* Set Color #0 foreground. */
	    setcolindex(5, 0, 3);
	    break;

	  case '1':		/* Set Color #1 foreground. */
	    setcolindex(6, 1, 7);
	    break;

	  case '2':		/* Set Color #2 foreground. */
	    setcolindex(7, 2, 11);
	    break;

	  case '3':		/* Set Color #3 foreground. */
	    setcolindex(8, 3, 15);
	    break;

	  case '4':		/* Set Color #0 background. */
	  case ')':
	    setcolindex(9, 4, 1);
	    break;

	  case '5':		/* Set Color #1 background. */
	  case '!':
	    setcolindex(10, 5, 5);
	    break;

	  case '6':		/* Set Color #2 background. */
	  case '@':
	    setcolindex(11, 6, 9);
	    break;

	  case '7':		/* Set Color #3 background. */
	  case '#':
	    setcolindex(12, 7, 13);
	    break;

	  default:
	    Bell();
	    break;
	}
    }
}

/*
** On an activate event we always appropriately activate/deactivate our
** Controls.  This is primarily done to hide scroll bars on windows that
** can't take events.
*/

void
activatevent(ep)
EventRecord *ep;
{
    short i;

    if (firstactive | web) {	/* Skip first activate event because all the */
	firstactive = 0;	/* controls were created active.  Saves on */
	return;			/* startup time. */
    }
    if (iconified)		/* Iconified window is inactive, do nothing. */
	return;

    if (calmode)		/* Calibrating, nothing to do. */
	return;

    SetPort(ep->who);
    DrawGrowIcon(ep->who);
    if (ep->message & 1) {	/* Activating? */
	i = 0;
	while (controls[i].cid) {
	    HiliteControl(controls[i].cid, controls[i].hilval);
	    i++;		/* Mark all of them active */
	}
    } else {			/* Deactivating. */
	i = 0;
	while (controls[i].cid)
	    HiliteControl(controls[i++].cid, 255); /* Mark all inactive. */
    }
}

/*
** This routine is called repeatedly if the operator presses in the controls
** of scroll bars.  It is optional (installed via SetCtlAction).  Without it
** the scroll bars only accept a single click in any part of them.  Generally
** users will expect scrolling to be continuous, so you want one of these.
**
** This routine is _not_ called when dragging the thumb.
*/

void
MyAction(theControl, partCode)
ControlID theControl;
short partCode;
{
    if (partCode == inUpButton || partCode == inPageUp) {
	if (partCode == inPageUp) {
	    if (theControl == controls[15].cid || 
		theControl == controls[16].cid)
		SetCtlValue(theControl, GetCtlValue(theControl)-10);
	    else
		SetCtlValue(theControl, GetCtlValue(theControl)-1);
	} else
	    SetCtlValue(theControl, GetCtlValue(theControl)-1);
    } else if (partCode == inDownButton || partCode == inPageDown) {
	if (partCode == inPageDown) {
	    if (theControl == controls[15].cid || 
		theControl == controls[16].cid)
		SetCtlValue(theControl, GetCtlValue(theControl)+10);
	    else
		SetCtlValue(theControl, GetCtlValue(theControl)+1);
	} else
	    SetCtlValue(theControl, GetCtlValue(theControl)+1);
    }
    readsliders();
    if (theControl == controls[16].cid)
	drawdimtime();
    dosetup();
    if (theControl == controls[15].cid)
	Bell();
}


redocontrols()			/* Call when switching between HSV and  */
{				/* RGB color models. */
    struct palette *p;
    double h, s, v;
    double R, G, B;

    p = &setupdata.entries[colornum];
    SetCtlMax(controls[2].cid, colormodel ? maxval : hueval);
    SetCtlMax(controls[3].cid, colormodel ? maxval : satval);
    SetCtlMax(controls[4].cid, colormodel ? maxval : satval);
    if (colormodel) {		/* RGB Model? */
	SetCtlValue(controls[2].cid, maxval - (p->red >> shiftval));
	SetCtlValue(controls[3].cid, maxval - (p->green >> shiftval));
	SetCtlValue(controls[4].cid, maxval - (p->blue >> shiftval));
    } else {			/* HSV Model. */
	R = (double) (p->red >> shiftval) / maxval;
	G = (double) (p->green >> shiftval) / maxval;
	B = (double) (p->blue >> shiftval) / maxval;
	RGBxHSV(R, G, B, &h, &s, &v);
	SetCtlValue(controls[2].cid, hueval - (int) (h * hueval));
	SetCtlValue(controls[3].cid, satval - (int) (s * satval));
	SetCtlValue(controls[4].cid, satval - (int) (v * satval));
    }
}


readsliders()			/* Call after changing a slider. */
{
    struct palette *p;
    double h, s, v;
    double R, G, B;

    p = &setupdata.entries[colornum]; /* Assume RGB color model. */
    p->red = (maxval - GetCtlValue(controls[2].cid)) << shiftval;
    p->green = (maxval - GetCtlValue(controls[3].cid)) << shiftval;
    p->blue = (maxval - GetCtlValue(controls[4].cid)) << shiftval;
    if (!colormodel) {		/* HSV Color Model? */
				/* Yes, convert to RGB. */
	h = (double) (hueval - GetCtlValue(controls[2].cid)) / hueval;
	s = (double) (satval - GetCtlValue(controls[3].cid)) / satval;
	v = (double) (satval - GetCtlValue(controls[4].cid)) / satval;
	HSVxRGB(h, s, v, &R, &G, &B);
	p->red = (int) (R * maxval) << shiftval;
	p->green = (int) (G * maxval) << shiftval;
	p->blue = (int) (B * maxval) << shiftval;
    }
    setupdata.hibrite = GetCtlValue(controls[13].cid);
    setupdata.lobrite = GetCtlValue(controls[14].cid);
    setupdata.volume = GetCtlValue(controls[15].cid);
    setupdata.dimtime = GetCtlValue(controls[16].cid);
}

/*
** Given [h,s,v] on [0 - 1.0], return RGB on [0 - 1.0].
*/

void
HSVxRGB(h, s, v, R, G, B)
double h, s, v;
double *R, *G, *B;
{
    double p, q, t;
    int sextant;
    double fract;

    sextant = (int) (fract = (h * 6.0));
    fract = fract - (int) fract;
    p = (1.0 - s) * v;
    q = (1.0 - s * fract) * v;
    t = (1.0 - (s * (1.0 - fract))) * v;
    switch (sextant) {
      case 0:
      case 6: *R = v; *G = t; *B = p; break;
      case 1: *R = q; *G = v; *B = p; break;
      case 2: *R = p; *G = v; *B = t; break;
      case 3: *R = p; *G = q; *B = v; break;
      case 4: *R = t; *G = p; *B = v; break;
      case 5: *R = v; *G = p; *B = q; break;
    }
}

/*
** Given RGB on [0 - 1.0], return [h,s,v] on [0 - 1.0].
**
** For an HSL conversion rather than HSV (HSB), 
** replace V = MAX(R,G,B)
** with    L = (MIN(R,G,B) + MAX(R,G,B) / 2
*/

void
RGBxHSV(R, G, B, H, S, V)
double R, G, B;
double *H, *S, *V;
{
    double MAX, MIN, d, h, r, g, b;

    MAX = MIN = R;
    if (G > MAX) MAX = G;
    if (G < MIN) MIN = G;
    if (B > MAX) MAX = B;
    if (B < MIN) MIN = B;
    d = MAX - MIN;
    *V = MAX;
    if (MAX == 0.0) {
	*S = *H = *V = 0.0;
	return;
    }
    *S = d/MAX;
    if (d == 0.0) {
	*H = 0.0;
	return;
    }
    r = (MAX - R)/d;
    g = (MAX - G)/d;
    b = (MAX - B)/d;
    if (R == MAX) h = b-g;
    if (G == MAX) h = r-b + 2.0;
    if (B == MAX) h = g-r + 4.0;
    if (h < 0.0) *H = h / 6.0 + 1.0;
    if (h >= 0.0) *H = h / 6.0;
}

/*
** Sprintf replacement.
**
** Handles format strings containing:
**
**	 %s, %[0][1..9]d, %[0][1..9]x, %[0][1..9]X
**
** It's much smaller than the standard one.
*/
mysprintf(buffer, fmt, arglist)
char *buffer, *fmt;
char *arglist;
{
    register char *p, *q, *p2;
    short ndigits;
    int leadfill;
    char *itoa();
    union arg {
		long lng;
		char *ptr;
    } *argptr;
	
    argptr = (union arg *) &arglist;
    p = fmt;
    q = buffer;
    while (*p) {		/* While there's some format string. */
	if (*p == '%') {	/* Is it a format control? */
	    p++;		/* Yes, which one (skip over % char)? */
	    if (*p == '0')
		leadfill = *p++;
	    else
		leadfill = ' ';
	    if (*p >= '1' && *p <= '9')	/* Format output length? */
		ndigits = *p++ - '0';
	    else
		ndigits = 0;
	    if (*p == 's') {
		p2 = argptr++->ptr; /* String. */
	    } else if (*p == 'd') {
		p2 = itoa(argptr++->lng, 10, leadfill, ndigits);
	    } else if (*p == 'x' || *p == 'X') {
		p2 = itoa(argptr++->lng, 16, leadfill, ndigits);
	    }
	    p++;		/* Skip over format control. */
	    while (*p2)		/* Copy the inclusion. */
		*q++ = *p2++;
	} else			/* No, it's a literal -- copy it. */
	    *q++ = *p++;
    }
    *q = 0;			/* Null-terminate it. */
    return q - buffer;		/* Return number of characters processed. */
}

char *
itoa(val, radix, leadfill, ndigits)
int val, radix, leadfill, ndigits;
{
    short i;
    char *p, *p2, out;
    static char buf[20];
    unsigned long test, remainder;
	
    p = buf;
    remainder = val;
    i = ndigits;		/* Make sure there're at least the number */
    while (i--)			/* of leading whatevers that we've asked for */
	*p++ = leadfill;	/* as the field width. */
    p2 = p;
    if (radix == 16) {		/* Hex output? */
	test = 0x10000000L;
    } else {
	test = 1000000000L;	/* Decimal output? */
	if (val < 0) {
	    remainder = -val;
	    *p++ = '-';
	}
    }
    while (test > remainder)	/* Scale down to match number. */
	test /= radix;
    if (remainder)
	while (test) {
	    out = remainder / test;
	    *p++ = out + ((out >= 10) ? '7' : '0');
	    remainder %= test;
	    test /= radix;
	}
    else
	*p++ = '0';
    *p = 0;
    if (ndigits) {		/* A field width set? */
	if ((p2 + ndigits) >= p) /* Make sure we never truncate anything. */
	    return p - ndigits;
	else
	    return p2;
    }
    return buf;			/* No field width, return minimum number. */
}
