#include "../clunkdraw.h"
#include "cdbits.h"

// These are for working in offscreen destinations.  Global coordinates.

Boolean
_GetDstBit(const BitMap *bm, int x, int y)
{
   unsigned char mask, *d;

   if (y < 0 || y >= (bm->bounds.edge.bottom - bm->bounds.edge.top) || // Stay on
       x < 0 || x >= (bm->bounds.edge.right - bm->bounds.edge.left))   // BMap!
      return 1; // Special case needed by flood-fill.
   d = bm->baseAddr + (y * bm->rowBytes) + (x / 8);
   mask = 1 << (7 - (x & 7));
   return !!(*d & mask);
}

void
_SetDstBit(const BitMap *bm, int x, int y)
{
   unsigned char mask, *d;

   if (y < 0 || y >= (bm->bounds.edge.bottom - bm->bounds.edge.top) || // Stay on
       x < 0 || x >= (bm->bounds.edge.right - bm->bounds.edge.left))   // BMap!
      return;
   d = bm->baseAddr + (y * bm->rowBytes) + (x / 8);
   mask = 1 << (7 - (x & 7));
   *d |= mask;
}

void
_ClearDstBit(const BitMap *bm, int x, int y)
{
   unsigned char mask, *d;

   if (y < 0 || y >= (bm->bounds.edge.bottom - bm->bounds.edge.top) || // Stay on
       x < 0 || x >= (bm->bounds.edge.right - bm->bounds.edge.left))   // BMap!
      return;
   d = bm->baseAddr + (y * bm->rowBytes) + (x / 8);
   mask = 1 << (7 - (x & 7));
   *d &= ~mask;
}

void
_FlipDstBit(const BitMap *bm, int x, int y)
{
   unsigned char mask, *d;

   if (y < 0 || y >= (bm->bounds.edge.bottom - bm->bounds.edge.top) || // Stay on
       x < 0 || x >= (bm->bounds.edge.right - bm->bounds.edge.left))   // BMap!
      return;
   d = bm->baseAddr + (y * bm->rowBytes) + (x / 8);
   mask = 1 << (7 - (x & 7));
   *d ^= mask;
}

