#include "../clunkdraw.h"
#include "cdbits.h"

/*
** The CalcMask procedure produces a mask showing where paint could
** not flow from any of the outer edges of a rectangle.  You can use
** the resulting mask to paint with CopyBits, or you can convert it to
** a Region with BitMapToRegion, for later clipping or patterned
** painting purposes.
*/

void
CalcMask(BitMap *dstBmap, const Rect *bounds)
{
   const GrafPtr tp = thePort;
   Rect br;
   int ii;
   U8 *p;
   Boolean onScrn;

   if (!tp || tp->fgColor == tp->bkColor)
      return;
   onScrn = tp->portBits.baseAddr == dstBmap->baseAddr;
   // Can't (currently) flood-fill non-monochrome on the screen.
   if (onScrn && tp->portBits.bpp != 1)
      return;
   if (dstBmap->bpp != 1)
      return;
   if (!SectRect(&tp->portRect, bounds, &br) ||
       !SectRect(&br, &dstBmap->bounds, &br))
      return;
   if (onScrn)
      ShieldCursor(&br, &br.corner.topLeft);
   // Now flood inwards from every point on the bounding rectangle.
   // Go vertically first, because the flooder works horizontally.
   // That should cut down on the total work required to flood.
   for (ii=br.edge.top; ii<br.edge.bottom; ii++) {
      _FloodFillScanline(tp, &br, dstBmap, br.edge.left, ii);
      _FloodFillScanline(tp, &br, dstBmap, br.edge.right-1, ii);
   }
   for (ii=br.edge.left; ii<br.edge.right; ii++) {
      _FloodFillScanline(tp, &br, dstBmap, ii, br.edge.top);
      _FloodFillScanline(tp, &br, dstBmap, ii, br.edge.bottom-1);
   }
   // Everything we flooded is NOT the mask, so invert the results.
   p = dstBmap->baseAddr;
   ii = RectHeight(&dstBmap->bounds) * dstBmap->rowBytes;
   while (ii--)
      *p++ ^= 0xFF;
   if (onScrn)
      ShowCursor();
}
