#include "../clunkdraw.h"
#include "cdbits.h"
#include <alloca.h>

// Adapted from the seed fill algorithm by Paul Heckbert.
// Originally published in "Graphics Gems", Academic Press, 1990.
// Only leaks through holes in the cardinal directions, not across
// diagonals.

typedef struct {short xl, xr, y, dy;} Segment; // 8 bytes per drawing segment.

#define MAXDEPTH 512 // 4kB of drawing segment stack.

#define PUSH(Y, XL, XR, DY) \
   do { \
      if (sp < stack+MAXDEPTH) { \
         if (Y+(DY) >= br.edge.left && Y+(DY) <= br.edge.bottom) { \
            sp->y = (Y); sp->xl = (XL); sp->xr = (XR); (sp++)->dy = (DY);} \
      } else return; /* Stack overflow! */ \
   } while (0)

#define POP(Y, XL, XR, DY) \
   do { \
      if (sp >= stack+1) { \
         XL = (--sp)->xl; XR = sp->xr; Y = sp->y + (DY = sp->dy); \
      } else return; /* Stack underflow! */ \
   } while (0)

void
_FloodFillScanline(const GrafPtr tp, const Rect *bounds, BitMap *dbm, int x, int y)
{
   int l, xl, xr, dy, bkPix;
   Rect br;
   Point pt;
   Segment *stack, *sp; // Stack of filled segments.

   br = tp->portRect;
   if (tp->device == 1)
      RotRect(&br, &br);
   if (!SectRect(bounds, &br, &br))
      return;
   SetPt(&pt, x, y);
   if (!PtInRect(&pt, &br))
      return;
   sp = stack = alloca(sizeof(Segment) * MAXDEPTH);
   if ((bkPix = (*tp->_GetPix)(tp, x, y)) != tp->bkColor ||
       _GetDstBit(dbm, x, y))
      return;

   if (x < br.edge.left || x > br.edge.right ||
       y < br.edge.top  || y > br.edge.bottom)
      return;
   PUSH(y, x, x, 1);            /* needed in some cases */
   PUSH(y+1, x, x, -1);         /* seed segment (popped 1st) */
   while (sp > stack) {
      POP(y, xl, xr, dy);       // Pop seg off stack and fill neighbor scanline.
      // Segment of scanline y-dy for xl<=x<=xr was previously filled,
      // now explore adjacent pixels in scan line y.
      for (x = xl; x >= br.edge.left &&
                   (*tp->_GetPix)(tp, x, y) == bkPix && !_GetDstBit(dbm, x, y);
           x--)
         _SetDstBit(dbm, x, y);
      if (x >= xl) goto skip;
      l = x+1;
      if (l < xl) PUSH(y, l, xl-1, -dy); /* leak on left? */
      x = xl+1;
      do {
         for (; x <= br.edge.right &&
                 (*tp->_GetPix)(tp, x, y) == bkPix && !_GetDstBit(dbm, x, y);
              x++)
            _SetDstBit(dbm, x, y);
         PUSH(y, l, x-1, dy);
         if (x > xr+1) PUSH(y, xr+1, x-1, -dy); /* leak on right? */
skip:    for (x++; x <= xr &&
                 (*tp->_GetPix)(tp, x, y) != bkPix && _GetDstBit(dbm, x, y);
              x++);
         l = x;
      } while (x <= xr);
   }
}
