// Naive bitmap rotator.

#include "../clunkdraw.h"
#include <stdlib.h> // For calloc
#define TRUE 1
#define FALSE 0

static int
getBit(const BitMap *srcBMap, int sx, int sy)
{
   unsigned char mask;
   unsigned const char *s;
   int xoff, yoff;

   xoff = sx - srcBMap->bounds.edge.left;
   yoff = sy - srcBMap->bounds.edge.top;
   if ((yoff < 0) || (sy >= srcBMap->bounds.edge.bottom) || // Stay on bitmap!
       (xoff < 0) || (sx >= srcBMap->bounds.edge.right)) return FALSE;
   s = srcBMap->baseAddr + (yoff * srcBMap->rowBytes) + (xoff / 8);
   mask = 1 << (7 - (xoff & 7));
   return *s & mask;
}

static void
setBit(const BitMap *dstBMap, int dx, int dy)
{
   unsigned char mask, *d;
   int xoff, yoff;

   xoff = dx - dstBMap->bounds.edge.left;
   yoff = dy - dstBMap->bounds.edge.top;
   if ((yoff < 0) || (dy >= dstBMap->bounds.edge.bottom) || // Stay on bitmap!
       (xoff < 0) || (dx >= dstBMap->bounds.edge.right)) return;
   d = dstBMap->baseAddr + (yoff * dstBMap->rowBytes) + (xoff / 8);
   mask = 1 << (7 - (xoff & 7));
   *d |= mask;
}

static Boolean
RotateBMap(const BitMap *srcBMap, BitMap *dstBMap, int dir)
{
   int imageSize, sx, sy, dx, dy;

   RotRect(&srcBMap->bounds, &dstBMap->bounds);
   dstBMap->rowBytes = (RectWidth(&dstBMap->bounds) + 7) / 8;
   dstBMap->bpp = 1;
   imageSize = RectHeight(&dstBMap->bounds) * dstBMap->rowBytes;
   dstBMap->baseAddr = calloc(1, imageSize);
   if (!dstBMap->baseAddr) return FALSE;
   if (dir > 0)
      dx = dstBMap->bounds.edge.left;
   else
      dx = dstBMap->bounds.edge.right - 1;
   for (sy = srcBMap->bounds.edge.top; sy < srcBMap->bounds.edge.bottom; sy++) {
      if (dir > 0)
         dy = dstBMap->bounds.edge.bottom - 1;
      else
         dy = dstBMap->bounds.edge.top;
      for (sx=srcBMap->bounds.edge.left; sx < srcBMap->bounds.edge.right; sx++) {
         if (getBit(srcBMap, sx, sy))
            setBit(dstBMap, dx, dy);
         dy -= dir;
      }
      dx += dir;
   }
   return TRUE;
}

// Given a source and destination bitMap, rotate it CCW 90 degrees.
// The destination bitmap will be _created_ by this function.  Returns
// TRUE if the operation succeeded, or FALSE if it failed (allocation error).
// The bitMap can be of any size.
//
// The 'interesting' area of the source bitMap (what was justified against
// the top & left of the source) will end up justified against the bottom
// left of the destination.

Boolean
RotateBMapCCW(const BitMap *srcBMap, BitMap *dstBMap)
{
   return RotateBMap(srcBMap, dstBMap, 1);
}

// Ditto, but CW (clockwise).

Boolean
RotateBMapCW(const BitMap *srcBMap, BitMap *dstBMap)
{
   return RotateBMap(srcBMap, dstBMap, -1);
}
