#include "../clunkdraw.h"
#include "cdbits.h"

/*
** Create, in the destination (monochrome) BitMap, a 'flooded' area
** corresponding to the areas in the current Port that are the
** background color, starting the flood at the given point.
**
** If the source bitmap is monochrome, then source and dest can be the same.
*/

void
SeedFill(BitMap *dstBmap, const Point *floodPt)
{
   const GrafPtr tp = thePort;
   Rect br;
   Boolean onScrn;

   if (!tp || tp->fgColor == tp->bkColor)
      return;
   onScrn = tp->portBits.baseAddr == dstBmap->baseAddr;
   // Can't (currently) flood-fill non-monochrome on the screen.
   if (onScrn && tp->portBits.bpp != 1)
      return;
   if (dstBmap->bpp != 1)
      return;
   br = tp->portRect;
   if (tp->device == 1)
      RotRect(&br, &br);
   if (!SectRect(&dstBmap->bounds, &br, &br))
      return;
   if (!PtInRect(floodPt, &br))
      return;
   if (onScrn)
      ShieldCursor(&br, &br.corner.topLeft);
   _FloodFillScanline(tp, &dstBmap->bounds, dstBmap, floodPt->h, floodPt->v);
   if (onScrn)
      ShowCursor();
}
