// Save the given bitmap to a PGM interchange file.  Bulky, but safe?
// If these are color pixels, you really won't be getting what you want!
// Use SaveToPPMFile() instead.

#include "../clunkdraw.h"
#include <stdio.h>

void
SaveToPGMFile(const BitMap *sbits, const char *fName)
{
   int x, y, digits;
   FILE *fp;

   if (!sbits || (x = sbits->bpp) < 1) return;
   if (x & (x-1)) return;       // Not a power of 2.
   if (!(fp = fopen(fName, "w")))
      return;

   fprintf(fp, "P2\n");         // Portable Gray Map.
   fprintf(fp, "%d %d\n", RectWidth(&sbits->bounds), RectHeight(&sbits->bounds));
   digits = (1 << x) - 1;
   fprintf(fp, "%d\n", digits); // Range per emitted value.
   switch (x) {
      case  1: digits =  1; break;
      case  2: digits =  1; break;
      case  4: digits =  2; break;
      case  8: digits =  3; break;
      case 16: digits =  5; break;
      case 32: digits = 10; break;
      default: digits =  0; break;
   }
   for (y = 0; y < RectHeight(&sbits->bounds); y++) {
      fprintf(fp, "# Line %d\n", y + 1);
      for (x = 0; x < RectWidth(&sbits->bounds); x++)
         fprintf(fp, "%*d ", digits, GetPixel(x, y));
      fprintf(fp, "\n");
   }
   fclose(fp);
}
