// Save the given bitmap to a MacPaint (576x720) file.
// Too much data and it will be trimmed, too little and it will be padded.
// File format taken from Macintosh Technical Note #24.
//
// More modern systems might be able to read these if you use a .mac extension.

#include "../clunkdraw.h"
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#define WIDTH 576
#define HEIGHT 720

void
SaveToPaintFile(const BitMap *sbits, const char *fName)
{
   char *d,
        header[512],
        iline[WIDTH/8],   // 576 pixels' worth.
        oline[WIDTH/8+1]; // 576 pixels' worth, plus 1.
   int x, y, ii, fd;
   const char *s;

   if (!sbits || sbits->bpp != 1) return;
   if ((fd = creat(fName, 0666)) < 0)
      return;

   memset(header, 0, sizeof header);
   (void)write(fd, header, sizeof header); // The ver/pat header.  Empty.
   for (y = 1; y <= HEIGHT; y++) {
#if 1
      memset(iline, 0, sizeof iline);
      s = sbits->baseAddr + ((y-1) * sbits->rowBytes);
      for (x = 0; x < sizeof iline; x++)
         if (y > RectHeight(&sbits->bounds) || (x >= sbits->rowBytes))
            iline[x] = 0;
         else
            iline[x] = *s++;
      s = iline;
      d = oline;
      PackBits(&s, &d, sizeof iline); // Pack it down.  (Worst-case is +1.)
      ii = d - oline;
#else //Uncompressed
      memset(oline, 0, sizeof oline);
      s = sbits->baseAddr + ((y-1) * sbits->rowBytes);
      d = oline;
      *d++ = 71;
      for (x = 0; x < sizeof iline; x++)
         if (y > RectHeight(&sbits->bounds) || (x >= sbits->rowBytes))
            *d++ = 0;
         else
            *d++ = *s++;
      ii = sizeof oline;
#endif
      (void)write(fd, oline, ii);
   }
   close(fd);
}
