/*
** Some cherry-picked code and definitions from original Clunkdraw.
**
** See: Apple's QuickDraw documentation
**      (https://developer.apple.com/library/archive/documentation/mac/pdf/ImagingWithQuickDraw.pdf)
**      and the programs in test/ for how to use this library.  You can also
**      look at the source code.
*/

#include <stdint.h> // For sized integers.
#include <stddef.h> // For size_t.

/* Get modern definitions for old-school sized integers. */
typedef uint8_t   U8;
typedef int8_t    S8;
typedef uint16_t U16;
typedef int16_t  S16;
typedef uint32_t U32;
typedef int32_t  S32;
typedef uint64_t U64;
typedef int64_t  S64;
typedef int  Boolean;

/* Transfer modes, used by font and pattern operations. */
/* Boolean Source Transfer modes.  All systems, mono/color. */

#define srcCopy         0       /* Copy from source into window. */
#define srcOr           1       /* Or from source into window. */
#define srcXor          2       /* Xor from source into window. */
#define srcBic          3       /* Clear bits set in source in window. */
#define notSrcCopy      4       /* Copy inverse of source into window. */
#define notSrcOr        5       /* Or inverse of source into window. */
#define notSrcXor       6       /* Xor inverse of source into window. */
#define notSrcBic       7       /* Clear bits with zero in source in window. */

/* Text styles. */

#define normalText      0       /* Normal, is fastest by far. */
#define boldText        1       /* Emboldened, is wider than normal. */
#define fixedBoldText   2       /* Bold, but character width is same. */
#define underlineText   4       /* Underlined. */
#define italicText      8       /* Italicized. */
#define ghostText       16      /* Drawn in Gray (best w/dbl stroked font). */
#define outlineText     32      /* Outlined, is taller & wider than normal. */
#define strikeoutText   64      /* Strikeout. */
#define shadowText      128     /* With a drop shadow below and to the right. */
#define condenseText    256     /* Unimplemented, use TextExtra instead. */
#define extendText      512     /* Unimplemented, use TextExtra instead. */

/* Text justification flags (for DrawJustText). */

#define leftJustify     0       /* Left justification.                 */
#define rightJustify    1       /*                 Right justification.*/
#define centerJustify   2       /*         Center Justification.       */
#define fullJustify     3       /* F u l l  J u s t i f i c a t i o n .*/

/* For vertical justification, use leftJustify for top, rightJustify for
   bottom, and centerJustify for (of course) center.  (There is no full.) */

/* Window types/messages. */

#define documentProc    0       /* Regular window with all the trimmings. */
#define dBoxProc        1       /* No title bar, inset border. */
#define plainDBox       2       /* No title, no dropshadow. */
#define altDBoxProc     3       /* No title, dropshadow. */
#define noGrowDocProc   4       /* Fixed size regular window. */
#define noBordDBox      5       /* No border/fill at all. */
#define iconWProc       6       /* Iconized window. */

/* Window visible, parameter 'visible' in 'NewWindow' call. */

#define hideWindow      0       /* Window invisible */
#define showWindow      1       /* Window visible */

/*
** Values for NewWindow's goAwayFlag field (window options).  These're for
** documentProc and noGrowDocProc types of windows only.
*/

#define hasGoAway       1       /* Has go-away region (active window only). */
#define hasZoom         2       /* Has zoom region (active window only). */
#define hasSendToBack   4       /* Has send-to-back rgn (active wdo only). */
#define hasGrow         8       /* Has grow zone. */
#define hasDrag         16      /* Has drag area. */
#define isShy           32      /* Is shy, must hit drag region to select. */

/* Window order, parameter 'behindWindow' can be one of the following values
   or a windowid of another window to put this window behind. */

#define frontWindow     -1      /* Make this the frontmost window */
#define secondWindow    -2      /* Make this the second frontmost window */
#define backWindow      -100    /* Make this the backmost window */

/*
** Values returned by FindWindow.
*/

#define inDesk          0       /* In desktop. */
#define inMenuBar       1       /* In Menu bar. */
#define inContent       3       /* In content rgn (except grow, if active). */
#define inDrag          4       /* In drag region. */
#define inGrow          5       /* In grow region (active window only). */
#define inGoAway        6       /* In go-away region (active window only). */
#define inZoom          7       /* In zoom region (active window only). */
#define inSendToBack    8       /* In send-to-back rgn (active window only). */

/*
** Messages to the Window definition function.
*/

#define wDraw           0       /* Draw Window Frame. */
#define wHit            1       /* Tell what region mouse hit was in. */
#define wCalcRgns       2       /* Calculate strucRgn & contRgn. */
#define wNew            3       /* Do any additional window initialization. */
#define wDispose        4       /* Do any additional disposal actions. */
#define wGrow           5       /* Draw window's grow image. */
#define wDrawGIcon      6       /* Draw size box in content region. */

/*
** Returned by Window definition function's wHit routine.
*/

#define wNoHit          0       /* None of the following. */
#define wInContent      1       /* In content rgn (except grow, if active). */
#define wInDrag         2       /* In drag region. */
#define wInGrow         3       /* In grow region (active window only). */
#define wInGoAway       4       /* In go-away region (active window only). */
#define wInZoom         5       /* In zoom box. */
#define wInSendToBack   6       /* In send-to-back box. */

/*
** Arguments to various Track/Drag functions.
*/

#define tdBegin         0       /* Begin tracking/dragging. */
#define tdMiddle        1       /* Track/drag. */
#define tdEnd           2       /* End tracking/dragging. */
#define tdEndSel        3       /* End tracking/dragging, and select. */

/* Notification icons for Notify() function. */

#define noNotify        0       /* Remove notification icon. */
#define genericNotify   0x15    /* Generic notification icon. */
#define errorNotify     0x16    /* System Error notification icon. */
#define mailNotify      0x17    /* Mail notification icon. */
#define phoneNotify     0x18    /* Phone (talk) notification icon. */
#define clockNotify     0x19    /* Clock (alarm) notification icon. */

/* Event code definitions. */

#define MAXEVENT        9

#define mouseEvt        1
#define keyEvt          2
#define activateEvt     3
#define updateEvt       4
#define windowEvt       5
#define menuEvt         6
#define secondWindEvt   7       /* Second-window mouse event. */
#define scrapEvt        8       /* Clipboard updated event. */

/* Window event types, passed in the message field of a window event. */

#define windowGoaway    1
#define windowMove      2
#define windowResize    3
#define windowZoom      4

/* Bitmask definitions to be used with 'GetEvent' and 'CheckEvent'. */

#define noWindow        0
#define anyWindow       ((WindowPtr)-1)
#define noEvent         0
#define anyEvent        (-1)

/* Parameter used in MouseMode command. The mousemode parameter is anded
   with the mouse keycode and flags and shipped to the application if
   the result is nonzero.  Use the key flags definition for specific events.
   Ex: MouseMode(F_KEYDOWN | F_KEYUP); gives all key down and up events.
   When using mouseAll all changes in mouse keys or position inside the
   window are reported.  Combined with HideCursor(); all changes anywhere
   are reported.
*/

#define mouseAll        0x10000000      /* All mouse events inside my window */
#define mouseStream     0x20000000      /* Make stream of mouse events */
#define mouseCrossWinds 0x40000000      /* Let mouse post to many windows. */

/* Keycode flag definitions. */

#define F_KEYMASK	0x0000FFFF /* Mask to get keycode. */
#define F_FLAGMASK	0xFFFF0000 /* Mask to get key flags. */

#define F_FUNC		0x00010000 /* Key is a function key. */
#define F_SHIFT		0x00020000 /* Shift key is pressed with key. */
#define F_CTRL		0x00040000 /* Control key is pressed with key. */
#define F_ALT		0x00080000 /* Alt key is pressed with key. */

#define F_MOUSE1	0x00100000 /* Flags indicating the state of */
#define F_MOUSE2	0x00200000 /* the first 4 mouse buttons. */
#define F_MOUSE3	0x00400000
#define F_MOUSE4	0x00800000

#define F_KEYDOWN	0x01000000 /* Key down flag. */
#define F_KEYUP		0x02000000 /* Key up flag. */

#define	F_KBDIII	0x08000000 /* This keycode from AT / KB-III. */

#define F_KEYLK1	0x80000000 /* Flags indicating the state of */
#define F_KEYLK2	0x40000000 /* the (up to 4) keylocks on the */
#define F_KEYLK3	0x20000000 /* Wincom (4900) keyboard. */
#define F_KEYLK4	0x10000000

#define	K_BCKSP2	0x08	/* KBD-II backspace key. */
#define K_TAB		0x09	/* Tab */
#define K_ENTER		0x0D	/* Enter */
#define K_ESC		0x1B	/* Esc */
#define K_HOME		0x40	/* Cursor home */
#define K_UP		0x41	/* Cursor control keys... */
#define K_DOWN		0x42
#define K_RIGHT		0x43
#define K_LEFT		0x44
#define K_INSERT	0x51	/* Not avail. on kbd-II */ 
#define K_DELETE	0x52	/* Not avail. on kbd-II */
#define K_END		0x53	/* Not avail. on kbd-II */ 
#define K_PGUP		0x54	/* Not avail. on kbd-II */
#define K_PGDN		0x55	/* Not avail. on kbd-II */
#define	K_NPENTER	0x7E
#define	K_BACKSP	0x7F
#define K_F1		0x61	/* Function Keys... */
#define K_RESET		0x57	/* kbd-II only */
#define K_SEND		0x58	/* kbd-II only */

/*
** Low-level structures.
*/

typedef struct {                /* Graphics point structure. */
    short v;                    /* 0: Vertical component. */
    short h;                    /* 2: Horizontal component. */
} Point;

typedef union {                 /* Graphics Rectangle structure. */
    struct {
        short top;              /* 0: Topmost coordinate, etc. */
        short left;             /* 2: */
        short bottom;           /* 4: */
        short right;            /* 6: */
    } edge;
    struct {
        Point topLeft;
        Point botRight;
    } corner;
} Rect;

typedef struct {                /* System Bitmap structure. */
    void *baseAddr;             /* Pointer to bit image. */
    short rowBytes;             /* Number of bytes per row of each plane. */
    short bpp;                  /* Bits per pixel.  (1, 2, 4, 8, 16, 32) */
    Rect bounds;                /* Logical Bounding rectangle of planes. */
} BitMap;

typedef long Fixed;             /* Fixed-point data type. */

/*
** Actually a Fixed is the following structure.  We just call it a long so that
** it may be passed by _value_ rather than reference.
**
** typedef struct {
**      short Integer;          Integral part of fixed-point number.
**      short Fraction;         Fractional part of fixed-point number.
** } Fixed;
*/

typedef struct Region {
    unsigned short rgnSize;     /* Size of total region in bytes (min 10). */
    Rect rgnBBox;               /* Bounding box of region. */
    short rgnData;              /* Optional region data (maybe large). */
} Region;

typedef Region * RgnPtr;
typedef RgnPtr * RgnHandle;

typedef void * Ptr;             /* Generic pointer types. */
typedef Ptr * Handle;

typedef struct RegionHdr {
    unsigned short rgnSize;     /* Size of total region in bytes (min 10). */
    Rect rgnBBox;               /* Bounding box of region. */
} RegionHdr;

typedef struct Polygon {
    unsigned short polySize;    /* Size of total polygon in bytes (min 10). */
    Rect polyBBox;              /* Bounding box of region. */
    Point polyPoints;           /* Polygon data points (maybe large). */
} Polygon, *PolyPtr, **PolyHandle;

typedef struct PolyHdr {
    unsigned short polySize;    /* Size of total polygon in bytes (min 10). */
    Rect polyBBox;              /* Bounding box of region. */
} PolyHdr;

typedef struct Picture {
   unsigned long picSize;       /* Size of total picture in bytes. */
   Rect picFrame;               /* Picture frame. */
                                /* Plus variable-sized picture data. */
} Picture, *PicPtr, **PicHandle;

#define picLParen 0
#define picRParen 1

typedef union Pattern {
    U8  bytes[8];
    U16 words[4];
    U32 quads[2];
} Pattern;

extern const Pattern syspats[60];

typedef struct Cursor {
    unsigned short data[16];    /*  0: Cursor image. */
    unsigned short mask[16];    /* 16: Cursor mask. */
    Point hotSpot;              /* 32: Point aligned with mouse. */
} Cursor;

typedef union Icon {
    U8  bytes[128];
    U16 words[64];
    U32 quads[32];
    unsigned long data[32];
} Icon;

typedef struct FontInfo {
    short ascent;               /* 0: Height above baseline of font. */
    short descent;              /* 2: Height below baseline of font. */
    short widMax;               /* 4: Maximum character width. */
    short leading;              /* 6: Row-to-row spacing of this font. */
} FontInfo;

typedef struct FontRec {
    unsigned short fontType;    /*  0: Font type, 1 = ISC 8-bit index, */
                                /*                2 = ISC 16-bit index, */
                                /*                others = Macintosh. */
    unsigned short firstChar;   /*  2: ASCII code of first character. */
    unsigned short lastChar;    /*  4: ASCII code of last character. */
    short widMax;               /*  6: Maximum character width. */
    short kernMax;              /*  8: Negative of max character kern. */
    short nDescent;             /* 10: Negative of descent value. */
    short fRectWidth;           /* 12: Width of font rectangle. */
    short fRectHeight;          /* 14: Height of font rectangle. */
    unsigned short owTLoc;      /* 16: Offset (byte, from fontType) to */
                                /*     offset/width table. */
    short ascent;               /* 18: Height above baseline of font. */
    short descent;              /* 20: Height below baseline of font. */
    short leading;              /* 22: Row-to-row spacing of this font. */
    unsigned short rowWords;    /* 24: Offset (byte, from fontType) to */
                                /*     location table. */
                                /* 26: The bitmap, and both tables. Actually */
                                /*     is variable size, but we can't */
                                /*     declare that in C. */
} FontRec;

typedef struct FontTableEntry {
   short font;
   short size;
   const FontRec * const frp;
} FontTableEntry;

typedef struct GrafPort {
    short device;               /* Device specific information. */
    BitMap portBits;            /* grafPort's bit map. */
    Rect portRect;              /* grafPort's rectangle. */
    RgnHandle visRgn;           /* Visible region. */
    RgnHandle clipRgn;          /* Clipping region. */
    Pattern bkPat;              /* Background pattern. */
    Pattern fillPat;            /* Fill pattern. */
    Point pnLoc;                /* Pen location. */
    Point pnSize;               /* Pen size. */
    short pnMode;               /* Pen's transfer mode. */
    Pattern pnPat;              /* Pen pattern. */
    short pnVis;                /* Pen visibility. */
    short txFont;               /* Font number for text. */
    const FontRec *fontPtr;     /* Pointer to font table. */
    short txFace;               /* Text's character style. */
    short txMode;               /* Text's transfer mode. */
    short txSize;               /* Font size for text. */
    short txRotate;             /* Text rotation value, degrees CCW. */
    Fixed txExtra;              /* Extra space for text chars. */
    Fixed spExtra;              /* Extra space for ASCII space chars. */
    U16 fgColor;                /* Foreground color. */
    U16 bkColor;                /* Background color. */
    void *picSave;              /* Picture being saved (n/a). */
    RgnHandle rgnSave;          /* Region being saved. */
    PolyHandle polySave;        /* Polygon being saved. */
    FontRec *(*_fontHook)();    /* Hook for getting a font in a size. */
    int  (*_GetPix)();          /* Hook for getting a physical pixel. */
    void (*_SetPix)();          /* Hook for setting a physical pixel. */
    void (*_ClearPix)();        /* Hook for clearing a physical pixel. */
    void (*_FlipPix)();         /* Hook for inverting a physical pixel. */
} GrafPort;

typedef GrafPort * GrafPtr;

typedef struct PenState {       /* A pen state save record. */
    Point pnLoc;                /* Pen's location. */
    Point pnSize;               /* Pen's size. */
    short pnMode;               /* Pen's drawing mode. */
    Pattern pnPat;              /* Pen's pattern. */
} PenState;

/* Event record, used to pass events to the application */

typedef struct EventRecord {    /* An Event Record. */
   GrafPtr who;                 /* Window identifier */
   int what;                    /* Event type. */
   long message;                /* Event message. */
   unsigned long when;          /* Time stamp. */
   Rect where;                  /* Area or rectangle location. */
} EventRecord;

typedef struct FontCtrl {
   unsigned short fontNum;      // Font family number.
   unsigned short fontSize;     // Size of this font in points.
   const FontRec *fontPtr;      // Pointer to the memory image.
   long memSize;                // Size of the memory image.
   struct FontCtrl *next;       // Pointer to next in chain.
   int used;                    // TRUE when font's been used since a purge.
} FontCtrl;

/* State structure passed to StdXXX routines by the ParsePicture command. */

typedef struct GrafState {
    Pattern *bkPat;             /* Background pattern. */
    Pattern *pnPat;             /* Pen pattern. */
    Pattern *fillPat;           /* Fill pattern. */
    Point pnLoc;                /* Pen location. */
    Point pnSize;               /* Pen size. */
    short pnMode;               /* Pen's transfer mode. */
    short txFont;               /* Font number for text. */
    short txFace;               /* Text's character style. */
    short txMode;               /* Text's transfer mode. */
    short txSize;               /* Font size for text. */
    Fixed spExtra;              /* Extra space for ASCII space chars. */
    Fixed txExtra;              /* Extra space for text chars. */
    short txRotate;             /* Text rotation value, degrees CCW. */
    short fgColor;              /* Foreground color. */
    short bkColor;              /* Background color. */
    short colrBit;              /* Color plane(s) in use. */
} GrafState;

/* Flag values for options field of the Standard File I/O structure. */
#define SF_GETFILE 0            /* Open a file. */
#define SF_PUTFILE 1            /* Save As a file. */
#define SF_ALLFILES 2           /* Show all files (.files, like ls -a). */
#define SF_HIDEBAD 4            /* Hide unselectables (instead of gray). */
#define SF_XPOS 8               /* Override X window position. */
#define SF_YPOS 0x10            /* Override Y window position. */
#define SF_UID 0x20             /* Override UID for permission check. */
#define SF_GID 0x40             /* Override GID for permission check. */

/* Standard File (File Picker) I/O structure. */
typedef struct SFDialog {
    short options;              /* Flags from above. */
    short xpos;                 /* X window position override, if enabled. */
    short ypos;                 /* Y window position override, if enabled. */
    short uid;                  /* UID override, if enabled. */
    short gid;                  /* GID override, if enabled. */
    char *initialName;          /* Opt. Initial (default) name to show. */
    char *shellMatch;           /* Opt. Shell-pattern match filter string. */
    char *regexpMatch;          /* Opt. Regular expr match filter string. */
    char *title;                /* Opt. Window Title. */
    char *returnedName;         /* Returned string, malloced for you. */
} SFDialog;

typedef struct WindowRecord {
    GrafPort port;              /* Window's GrafPort. */
    short windowKind;           /* Window class. */
    char visible;       /* TRUE if visible (_not_ same as non-overlapped) */
    char hilited;               /* TRUE if highlighted. */
    char goAwayFlag;            /* Window dingus options.  */
    char spareFlag;             /* not used. */
    RgnHandle strucRgn;         /* Structure Region. */
    RgnHandle contRgn;          /* Content Region. */
    RgnHandle updateRgn;        /* Update Region. */
    RgnHandle saveVisRgn;       /* Saved visRgn during update. */
    Handle titleHandle;         /* Window's title. */
    short titleWidth;           /* Width of title in pixels. */
    void *controlList;          /* Window's control list. */
    struct WindowRecord *nextWindow; /* Next window (behind) in window list. */
    void *refCon;               /* Window's reference value. */
    short pid;                  /* Process ID of window's creator. */
    short windowType;           /* Window type (procID from newWindow). */
} WindowRecord;

typedef GrafPtr WindowPtr;
typedef WindowRecord * WindowPeek;

/* Some globals everybody wants access to. */

GrafPtr thePort;                /* Pointer to current graphics port. */
const Pattern white;
const Pattern black;
const Pattern gray;
const Pattern ltGray;
const Pattern dkGray;

/***********************************************/
/****** Clunkdraw's available functions. *******/
/***********************************************/

// By using this macro, we don't include code for bitmap depths we don't use.
#define BPP(x) _BPP_ ## x
// Use like: InitGraf(X, Y, BPP(Z));
/* Pick only one of these to use.  Returns non-zero if failure. */
int InitGraf(int x, int y, void (*bppInit)(int x, int y));
int InitGraf1(void);        /* Device 1: 3.7" E-Paper monochrome display. */
int InitGraf2(void);        /* Device 2: 2.2" 16-bit LCD display. */
/* Mates to the above. */
void EndGraf(void);
void EndGraf1(void);
void EndGraf2(void);

void FlushGraf(void);       /* Push screen bits to hardware, if necessary. */
void FlushGrafRect(const Rect *r); /* Push screen bits to HW, if necessary. */
void OpenPort(GrafPtr port);
void InitPort(GrafPtr port);
void ClosePort(GrafPtr port);
void SetPort(GrafPtr port);
void GetPort(GrafPtr *port);
void GrafDevice(int device);
void SetPortBits(BitMap *bm);
void PortSize(int width, int height);
void MovePortTo(int leftGlobal, int topGlobal);
void SetOrigin(int horz, int vert);
void SetClip(RgnHandle rgn);
void SetClipSect(RgnHandle rgn);
void SetClipUnion(RgnHandle rgn);
void SetClipDiff(RgnHandle rgn);
void SetClipXor(RgnHandle rgn);
void GetClip(RgnHandle rgn);
void ClipRect(const Rect *r);
void ClipRectSect(const Rect *r);
void ClipRectUnion(const Rect *r);
void ClipRectDiff(const Rect *r);
void ClipRectXor(const Rect *r);
void BackPat(const Pattern *pat);
void InitCursor(void);
void EndCursor(void);
void SetCursor(const Cursor *);
void HideCursor(void);
void ShowCursor(void);
void ObscureCursor(void);
void ShieldCursor(const Rect *shieldRect, const Point *offsetPt);
void _MoveCursor(const Point *screenPt);
void PlotIcon(const Rect *theRect, const Icon *theIcon);
void ScreenRes(int *scrnHRes, int *srcnVRes);
void HidePen(void);
void ShowPen(void);
void GetPen(Point *pt);
void GetPenState(PenState *pnState);
void SetPenState(const PenState *pnState);
void PenSize(int width, int height);
void PenMode(int mode);
void PenPat(const Pattern *pat);
void PenNormal(void);
void MoveTo(int horz, int vert);
void _MoveTo(int horz, int vert);
void Move(int dh, int dv);
void _Move(int dh, int dv);
void LineTo(int horz, int vert);
void Line(int dh, int dv);
int GetPixel(int horz, int vert);
void SetPixel(void);
void TextFont(int font);
void TextFace(int styles);
void TextMode(int mode);
void TextSize(int size);
void TextRotate(int degreesCCW);
void SpaceExtra(Fixed extra);   /* Per-ASCII-space extra spacing. */
void DrawChar(char ch);
void DrawString(const char *string);
void DrawText(const char *textBuf, int firstByte, int byteCount);
void DrawJustText(const char *textBuf, int byteCount, const Rect *inBox, int hflag, int vflag);
// void DrawText16(unsigned short *textBuf, int firstShort, int shortCount,
//                 void *(*charCache)(unsigned short whichChar));
int CharWidth(char ch);
int StringWidth(const char *string);
int TextWidth(const char *textBuf, int firstByte, int byteCount);
// int TextWidth16(unsigned short *textBuf, int firstShort, int shortCount);
void GetFontInfo(FontInfo *info);
void ForeColor(int color);
void BackColor(int color);
// void ColorBit(int planemask);
// void SetRGBColor(unsigned short index, unsigned short red, unsigned short green, unsigned short blue);
// void GetRGBColor(unsigned short index, unsigned short *red, unsigned short *green, unsigned short *blue);
void SetRect(Rect *rect, short left, short top, short right, short bottom);
int RectWidth(const Rect *rect);
int RectHeight(const Rect *rect);
void OffsetRect(Rect *rect, short dh, short dv);
void InsetRect(Rect *rect, short dh, short dv);
Boolean SectRect(const Rect *src1, const Rect *src2, Rect *dstRect);
void UnionRect(const Rect *src1, const Rect *src2, Rect *dstRect);
Boolean PtInRect(const Point *pt, const Rect *rect);
void Pt2Rect(const Point *pt1, const Point *pt2, Rect *rect);
void RotRect(const Rect *sr, Rect *dr);
void PtToAngle(const Rect *rect, const Point *point, short *angle);
Boolean EqualRect(const Rect *rect1, const Rect *rect2);
Boolean EmptyRect(const Rect *rect);
void FrameRect(const Rect *rect);
void PaintRect(const Rect *rect);
void EraseRect(const Rect *rect);
void InvertRect(const Rect *rect);
void FillRect(const Rect *rect, const Pattern *fillPat);
void FrameOval(const Rect *rect);
void PaintOval(const Rect *rect);
void EraseOval(const Rect *rect);
void InvertOval(const Rect *rect);
void FillOval(const Rect *rect, const Pattern *fillpat);
void FrameRoundRect(const Rect *rect, int ovalWidth, int ovalHeight);
void PaintRoundRect(const Rect *rect, int ovalWidth, int ovalHeight);
void EraseRoundRect(const Rect *rect, int ovalWidth, int ovalHeight);
void InvertRoundRect(const Rect *rect, int ovalWidth, int ovalHeight);
void FillRoundRect(const Rect *rect, int ovalWidth, int ovalHeight,
                   const Pattern *fillpat);
void FrameArc(const Rect *rect, short startAngle, short arcAngle);
void FrameWedge(const Rect *rect, short startAngle, short arcAngle);
void PaintArc(const Rect *rect, short startAngle, short arcAngle);
void EraseArc(const Rect *rect, short startAngle, short arcAngle);
void InvertArc(const Rect *rect, short startAngle, short arcAngle);
void FillArc(const Rect *rect, short startAngle, short arcAngle, const Pattern *fillpat);
RgnHandle NewRgn(void);
void OpenRgn(void);
void CloseRgn(RgnHandle dstRgn);
void DisposeRgn(RgnHandle rgn);
void CopyRgn(RgnHandle srcRgn, RgnHandle dstRgn);
void _CopyWorkRgn(RgnHandle srcRgn, short *ps);
void SetEmptyRgn(RgnHandle dstRgn);
void SetRectRgn(RgnHandle dstRgn, short left, short top,
                short right, short bottom);
void RectRgn(RgnHandle dstRgn, const Rect *srcRect);
void OffsetRgn(RgnHandle dstRgn, short dh, short dv);
void InsetRgn(RgnHandle dstRgn, short dh, short dv);
void SectRgn(RgnHandle srcRgnA, RgnHandle srcRgnB, RgnHandle dstRgn);
void UnionRgn(RgnHandle srcRgnA, RgnHandle srcRgnB, RgnHandle dstRgn);
void DiffRgn(RgnHandle srcRgnA, RgnHandle srcRgnB, RgnHandle dstRgn);
void XorRgn(RgnHandle srcRgnA, RgnHandle srcRgnB, RgnHandle dstRgn);
Boolean PtInRgn(const Point *point, RgnHandle region);
Boolean RectInRgn(const Rect *rect, RgnHandle region);
Boolean EqualRgn(RgnHandle rgnA, RgnHandle rgnB);
Boolean EmptyRgn(RgnHandle rgn);
void FrameRgn(RgnHandle rgn);
void PaintRgn(RgnHandle rgn);
void EraseRgn(RgnHandle rgn);
void InvertRgn(RgnHandle rgn);
void FillRgn(RgnHandle rgn, const Pattern *pat);
void ScrollRect(const Rect *r, short dh, short dv, RgnHandle updateRgn);
void CopyBits(const BitMap *srcBits, const BitMap *dstBits,
              const Rect *srcRect, const Rect *dstRect,
              int mode, RgnHandle maskRgn);
Boolean BitMapToRegion(RgnHandle rgn, const BitMap *sbm);
void SeedFill(BitMap *dstBmap, const Point *floodPt);
void CalcMask(BitMap *dstBmap, const Rect *bounds);
/* Picture **OpenPicture(Rect *); */
/* void ClosePicture(void); */
/* void PicComment(int, int, char *); */
/* void DrawPicture(char *, Rect *); */
/* void KillPicture(char *); */
PolyHandle OpenPoly(void);
void ClosePoly(void);
void KillPoly(PolyHandle poly);
void OffsetPoly(PolyHandle poly, short dh, short dv);
void FramePoly(PolyHandle poly);
void PaintPoly(PolyHandle poly);
void ErasePoly(PolyHandle poly);
void InvertPoly(PolyHandle poly);
void FillPoly(PolyHandle poly, const Pattern *pattern);
void AddPt(const Point *srcPt, Point *dstPt);
void SubPt(const Point *srcPt, Point *dstPt);
long DeltaPoint(const Point *, const Point *);
void SetPt(Point *dstPt, int horz, int vert);
void OffsetPt(Point *pt, int deltah, int deltav);
Boolean EqualPt(const Point *pt1, const Point *pt2);
void LocalToGlobal(Point *pt);
void GlobalToLocal(Point *pt);
void ScalePt(Point *, const Rect *, const Rect *);
void MapPt(Point *, const Rect *, const Rect *);
void MapRect(Rect *, const Rect *, const Rect *);
void MapRgn(RgnHandle, const Rect *, const Rect *);
void MapPoly(Polygon **, const Rect *, const Rect *);
void PackBits(const char **srcPtr, char **dstPtr, short srcBytes);
void UnpackBits(const char **srcPtr, char **dstPtr, short dstBytes);
void SaveToPaintFile(const BitMap *sbits, const char *fName);
void SaveToPBMFile(const BitMap *sbits, const char *fName);
void SaveToPGMFile(const BitMap *sbits, const char *fName);
void SaveToPPMFile(const BitMap *sbits, const char *fName);
void SaveToJPGFile(const BitMap *sbits, const char *fName);

/*** Font functions. ***/
void InitFonts();
void PurgeFonts(int flag);

/*** Event functions. ***/

//void (*Event(int type, void (*proc)()))(); /* Bind a procedure to an event. */
//void (*EventWithArg(int type, void (*proc)(), long extraArg))(); /* Bind a procedure with arg to an event. */
//long GetKey(void);
//int ReadKeys(long *bp, int size);		/* Read buffered keys */
Boolean GetNextEvent(int eventMask, EventRecord *); /* Get next event of spec. types */
void ReleaseEvent(void);
//int CheckKey(void); /* Check if key is available */
//int CheckEvent(WindowPtr windowmask, int eventmask); /* Check if event of spec. types from spec. windows available */
//int CheckEventWithArg(WindowPtr windowmask, int eventmask, long extraarg);	/* Check if event of spec. types from spec. windows available */
//void MouseMode(long mode);
void Bell(void);

/*** Following are the Window Manager routine declarations. ***/

void InitWindows(void);
void EndWindows(void);
void RsrvMenuBar(int height);
void GetWMgrPort(GrafPtr *wPort);
WindowPtr NewWindow(WindowPeek wStorage, Rect *boundsRect, char *title, int visible, int procID,
                    WindowPtr behind, int goAwayFlag, void *refCon);
/* WindowPtr GetNewWindow(int, WindowPtr); */
void CloseWindow(WindowPtr theWindow);
void DisposeWindow(WindowPtr theWindow);
void SetWTitle(WindowPtr, char *);
void GetWTitle(WindowPtr, char *);
void SelectWindow(WindowPtr theWindow);
void HideWindow(WindowPtr theWindow);
void ShowWindow(WindowPtr theWindow);
void ShowHide(WindowPtr theWindow, Boolean showFlag);
void HiliteWindow(WindowPtr theWindow, Boolean fHilite);
void BringToFront(WindowPtr theWindow);
void SendBehind(WindowPtr theWindow, WindowPtr behindWindow);
WindowPtr FrontWindow(void);
void DrawGrowIcon(WindowPtr theWindow);
short FindWindow(Point *thePt, WindowPtr *whichWindow);
Boolean TrackBox(WindowPtr theWindow, Point *thePt, int partCode);
void MoveWindow(WindowPtr theWindow, int hGlobal, int vGlobal, Boolean makeFront);
Boolean DragWindow(WindowPtr theWindow, const Point *delta, const Rect *bounds, int trackDrag);
Boolean GrowWindow(WindowPtr theWindow, const Point *delta, const Rect *size, int trackDrag);
void SizeWindow(WindowPtr theWindow, int newWidth, int newHeight, int fUpdate);
void ZoomWindow(WindowPtr theWindow, const Rect *newRect, int makeFront);
void InvalRect(const Rect *badRect);
void InvalRgn(RgnHandle badRgn);
void ValidRect(const Rect *goodRect);
void ValidRgn(RgnHandle goodRgn);
void BeginUpdate(WindowPtr theWindow);
void EndUpdate(WindowPtr theWindow);
void SetWRefCon(WindowPtr theWindow, void *data);
void *GetWRefCon(WindowPtr theWindow);
/* void SetWindowPic(WindowPtr, char *); */
/* char *GetWindowPic(WindowPtr); */
Point PinRect(const Rect *theRect, const Point *thePt);
/* long DragGrayRgn(RgnHandle, Point *, Rect *, Rect *, int, void ()()); */
WindowPtr CheckUpdate(WindowPtr flag);
void ClipAbove(WindowPeek window);
void SaveOld(WindowPeek window);
void DrawNew(WindowPeek window, Boolean update);
void PaintOne(WindowPeek window, RgnHandle clobberedRgn);
void PaintBehind(WindowPeek startWindow, RgnHandle clobberedRgn);
void CalcVis(WindowPeek window);
void CalcVisBehind(WindowPeek startWindow, RgnHandle clobberedRgn);
void ChangeWindow(WindowPeek theWindow, short newtype);

/*** Following are Miscellaneous routine declarations. ***/

Fixed SlopeFromAngle(short angle);
short AngleFromSlope(Fixed slope);
Fixed TextWidHit(const char *string, short byteCount, short hoffset);
Boolean RotateBMapCCW(const BitMap *srcBMap, BitMap *dstBMap);
Boolean RotateBMapCW(const BitMap *srcBMap, BitMap *dstBMap);
void TextExtra(Fixed extra);    /* Per-character extra spacing. */
//FontList *GetFontList(int *nfonts);
//int GetFontSizes(int font, unsigned char *list, int listsize);
Handle NewHandle(size_t size);
void DisposeHandle(Handle hand);
void SetHandleSize(Handle hand, size_t size);
size_t GetHandleSize(Handle hand);

/* Miscellaneous routines that aren't 'shared'. */

long StringLen(const char *string);
long FixRound(Fixed number);
Fixed Int2Fix(int number);
short Fix2Int(Fixed number);
unsigned short Fix2Frac(Fixed number);
unsigned long HiWord(long number);
unsigned long LoWord(long number);
Fixed FixRatio(short numerator, short denominator);
Fixed FixMul(Fixed a, Fixed b);
U32 StrToFixU32(const char *s, char **end, int scale, Boolean trunc);
S32 StrToFixS32(const char *s, char **end, int scale, Boolean trunc);
U64 StrToFixU64(const char *s, char **end, int scale, Boolean trunc);
S64 StrToFixS64(const char *s, char **end, int scale, Boolean trunc);
char *FixU32ToStr(char *str, int len, U32 fixed, U32 scale, int dtrunc, Boolean commas);
char *FixU64ToStr(char *str, int len, U64 fixed, U64 scale, int dtrunc, Boolean commas);
char *FixS32ToStr(char *str, int len, S32 fixed, U32 scale, int dtrunc, Boolean commas);
char *FixS64ToStr(char *str, int len, S64 fixed, U64 scale, int dtrunc, Boolean commas);
long MulS(short multiplier, short multiplicand);
unsigned long MulU(unsigned short multiplier, unsigned short multiplicand);
short DivS(long divisor, short dividend);
unsigned short DivU(unsigned long divisor, unsigned short dividend);
long DivModS(long divisor, short dividend);
unsigned long DivModU(unsigned long divisor, unsigned short dividend);
short ModS(long divisor, short dividend);
unsigned short ModU(unsigned long divisor, unsigned short dividend);
short ScaleS(short a, short b, short c);      /* A * B / C */
unsigned short ScaleU(unsigned short a, unsigned short b, unsigned short c);
short ScaleRS(short a, short b, short c); /* Rounded A*B/C */
unsigned short ScaleRU(unsigned short a, unsigned short b, unsigned short c);
void BlockMove(const void *src, void *dest, long nbytes);

/*** Debugging declarations. ***/
void DumpPt(const Point *p, const char *label);
void DumpRect(const Rect *rp, const char *label);
void DumpPattern(const Pattern *pattern, const char *label);
void DumpMode(int mode);
void DumpFace(int face);
void DumpBMap(const BitMap *bp, const char *label);
void DumpPort(GrafPtr thePort);
void DumpRgn(RgnHandle rgn, const char *label);
void DumpPoly(PolyHandle poly, const char *label);
void DumpWindow(WindowPeek theWindow, char *label);
void DumpWinds(char *label);

/* Used by the BPP macro. */
void _BPP_1(int x, int y);
void _BPP_2(int x, int y);
void _BPP_4(int x, int y);
void _BPP_8(int x, int y);
void _BPP_16(int x, int y);

/*******************************/
/****** Control Manager. *******/
/*******************************/

/*
** Standard Control types.
*/

#define pushButProc 0           /* Simple button. */
#define checkBoxProc 1          /* Check box. */
#define radioButProc 2          /* Radio button. */
#define defButProc 3            /* Outlined simple button, for default. */
#define useWFont 8              /* Add to above to use current font (not 0). */
#define scrollBarProc 16        /* Scroll bar. */

/*
** Part codes of the standard controls.
*/

#define inButton 10             /* Simple button. */
#define inCheckBox 11           /* Check box or radio button. */
#define inUpButton 20           /* Up arrow of a scroll bar. */
#define inDownButton 21         /* Down arrow of a scroll bar. */
#define inPageUp 22             /* Page up region of scroll bar. */
#define inPageDown 23           /* Page down region of scroll bar. */
#define inThumb 129             /* Thumb of a scroll bar. */

/*
** Values for visible argument to NewControl.
*/

#define hideControl 0           /* Create hidden control. */
#define showControl 1           /* Create visible control. */

#ifndef CONTROLS_SOURCE
typedef void *ControlID;
#else
typedef struct ControlRecord *ControlID;
#endif

void InitControls();
void SetCDefProc(void *(*cdefproc)(short varCode, ControlID theControl, short message, void *param));
ControlID NewControl(WindowPtr theWindow, Rect *boundsRect, char *title,
                     Boolean visible, short value, short min, short max,
                     short procID, void *refCon);
void DisposeControl(ControlID theControl);
void DisposeControlNoErase(ControlID theControl);
void KillControls(WindowPtr theWindow);
void SetCTitle(ControlID theControl, char *title);
void HideControl(ControlID theControl);
void HideControlNoInval(ControlID theControl);
void ShowControl(ControlID theControl);
void DrawControls(WindowPtr theWindow);
void DrawControl(ControlID theControl);
void HiliteControl(ControlID theControl, int hiliteState);
int FindControl(Point *thePoint, WindowPtr theWindow, ControlID *whichControl);
int TrackControl(ControlID theControl, Point *thePoint,
                 void (*actionProc)(ControlID theControl, int partCode));
int TestControl(ControlID theControl, Point *thePoint);
void MoveControl(ControlID theControl, short newh, short newv);
void ShiftControl(ControlID theControl, short newh, short newv);
void SizeControl(ControlID theControl, short width, short height);
void SetCtlValue(ControlID theControl, short theValue);
void SetCtlValues(ControlID theControl, short theValue, short theValue2);
void Set4CtlVals(ControlID theControl, short minValue, short maxValue,
                 short theValue, short theValue2);
Boolean IsCtlActive(ControlID theControl);
int GetCtlValue(ControlID theControl);
int GetCtlValue2(ControlID theControl);
void SetCtlMin(ControlID theControl, short minValue);
int GetCtlMin(ControlID theControl);
void SetCtlMax(ControlID theControl, short maxValue);
int GetCtlMax(ControlID theControl);
void SetCtlAction(ControlID theContrl,
                  void (*actionProc)(ControlID theControl, int partCode));
void (*GetCtlAction(ControlID theControl))
                   (ControlID theControl, int partCode);
void SetCtlScrollRate(ControlID theContrl, long ticks);
long GetCtlScrollRate(ControlID theContrl);
void SetCRefCon(ControlID theControl, void *data);
void *GetCRefCon(ControlID theControl);
/* Prototype for a control definition procedure. */
void *_myControl(short varCode, ControlID theControl, short message, void *param);
/*************************/
/****** Text Edit. *******/
/*************************/

#ifndef TEXTEDIT_SOURCE
typedef void TERec;
#else
typedef struct TERec TERec;
#endif

void TEInit();
TERec *TENew(Rect *destRect, Rect *viewRect);
void TEDispose(TERec *pTE);
void TESetText(char *text, int maxLength, TERec *pTE);
int TEGetText(char *text, TERec *pTE);
void TEClick(Point *pt, int extend, TERec *pTE);
void TESetSelect(int selStart, int selEnd, TERec *pTE);
long TEGetSelect(TERec *pTE);
void TEActivate(TERec *pTE);
void TEDeactivate(TERec *pTE);
int TEActivated(TERec *pTE);
void TEKey(long key, TERec *pTE);
void TECut(TERec *pTE);
void TECopy(TERec *pTE);
void TEPaste(TERec *pTE);
void TEDelete(TERec *pTE);
void TEInsert(char *newText, long textSize, TERec *pTE);
void TESetJust(int just, TERec *pTE);
void TEUpdate(Rect *rUpdate, TERec *pTE);
void SetWordBreak(int (*procPtr)(char *text, int charPos), TERec *pTE);
void TEScroll(int h, int v, TERec *pTE);
void TEPinScroll(int h, int v, TERec *pTE);
void TESelView(TERec *pTE);
void TEAutoView(Boolean onOff, TERec *pTE);
void TECROnly(Boolean onOff, TERec *pTE);
void TEAutoWrap(Boolean onOff, TERec *pTE);
void TECalText(TERec *pTE);
void TELeading(Boolean onOff, TERec *pTE);
void TESpillGuts(TERec *pTE);
int TELineNumber(int offset, TERec *pTE);
int TELineStart(int lineNo, TERec *pTE);
long TEGetScrapLen(void);
char *TEGetScrapAddr(void);
void TESetScrapLen(long newSize);
void TEMove(int x, int y, TERec *pTE);
void TESetViewRect(Rect *viewRect, TERec *pTE);
int TEVisLine(char *buf, int maxLen, int lineNo, TERec *pTE);
void TextBox(char *str, int length, Rect *box, int just);

/****************************/
/****** Menu Manager. *******/
/****************************/

#ifndef MENUS_SOURCE
typedef void MenuInfo;
#else
typedef struct MenuInfo MenuInfo;
#endif

void InitMenus();
void EndMenus();
long MenuKey(long key);
void NewMenu(MenuInfo **menuRoot, int menuID, char *menuTitle);
void DisposeMenu(MenuInfo **menuRoot, int menuID);
void DisableMenu(MenuInfo *menuRoot, int menuID);
void EnableMenu(MenuInfo *menuRoot, int menuID);
void AppendMenu(MenuInfo *menuRoot, int menuID, char *data);
void InsertMenu(MenuInfo *menuRoot, int menuID, int beforeID);
void DrawMenuBar();
void FlashMenuBar();
void RefreshMenuBar();
void DeleteMenu(MenuInfo *menuRoot, int menuID);
void ChopMenu(MenuInfo *menuRoot, int menuID, int item);
void OpenMenu(MenuInfo *thisMenu);
void CloseMenu();
void HiliteMenu(int menuID);
int  HiliteMenuItem(int itemID);
void SetItem(MenuInfo *menuRoot, int menuID, int item, char *itemString);
void SetItemMark(MenuInfo *menuRoot, int menuID, int item, int markChar);
void SetItemStyle(MenuInfo *menuRoot, int menuID, int item, int style);
void SetItemKey(MenuInfo *menuRoot, int menuID, int item, long key);
void DisableItem(MenuInfo *menuRoot, int menuID, int item);
void EnableItem(MenuInfo *menuRoot, int menuID, int item);
void BeginMenu(MenuInfo *menu, long menuMessage);
void EndMenu(MenuInfo *menu, long menuMessage);
void SetMenuFlash(int flashCount);
void MenuClock(int mode);
void MenuTick(void); // Call once per second.
