#define CONTROLS_SOURCE
#include "../clunkdraw.h"
#include "cmcontrols.h"

#ifdef DEBUG
#include <stdio.h>
#endif

/*
** Calculate the five rectangles comprising a scroll bar.  Also returns an
** orientation flag.
*/

static short
_calcScrollRects(theControl, upRect, dnRect, pgUpRect, pgDnRect, thumbRect)
ControlID theControl;
Rect *upRect, *dnRect, *pgUpRect, *pgDnRect, *thumbRect;
{
   short horizontal, ii, jj, height;

   *upRect = *dnRect = theControl->contrlRect;
   horizontal = ((upRect->edge.bottom - upRect->edge.top) <
                 (upRect->edge.right - upRect->edge.left)) ? 1 : 0;

   if (horizontal)
      upRect->edge.right = upRect->edge.left + 17;
   else
      upRect->edge.bottom = upRect->edge.top + 17;

   if (horizontal)
      dnRect->edge.left = dnRect->edge.right - 17;
   else
      dnRect->edge.top = dnRect->edge.bottom - 17;

   *pgUpRect = theControl->contrlRect;
   if (horizontal) {
      InsetRect(pgUpRect, 17, 2);
   } else {
      InsetRect(pgUpRect, 2, 17);
   }
   *pgDnRect = *thumbRect = *pgUpRect;
   height = 0;
   if (ii = theControl->contrlMax - theControl->contrlMin) {
      jj = (horizontal) ? thumbRect->edge.right - thumbRect->edge.left
                        : thumbRect->edge.bottom - thumbRect->edge.top;
      height = ScaleS(theControl->contrlValue2-theControl->contrlMin, jj, ii);
   }
   if (height < 16)             /* Minimum elevator box height (pixels). */
      height = 16;
   if (horizontal) {
      thumbRect->edge.right -= height; /* Elevator box itself. */
      jj = thumbRect->edge.right - thumbRect->edge.left;
   } else {
      thumbRect->edge.bottom -= height; /* Elevator box itself. */
      jj = thumbRect->edge.bottom - thumbRect->edge.top;
   }            /* J equals the height of the elevator area in pixels. */
   ii = ii ? ScaleS(theControl->contrlValue - theControl->contrlMin, jj, ii) : 0;
   if (horizontal) {
      thumbRect->edge.left += ii;
      thumbRect->edge.right = thumbRect->edge.left + height;
      pgUpRect->edge.right = thumbRect->edge.left;
      pgDnRect->edge.left = thumbRect->edge.right;
   } else {
      thumbRect->edge.top += ii;
      thumbRect->edge.bottom = thumbRect->edge.top + height;
      pgUpRect->edge.bottom = thumbRect->edge.top;
      pgDnRect->edge.top = thumbRect->edge.bottom;
   }
   return horizontal;
}

void *
_myControl(varCode, theControl, message, param)
short varCode;
ControlID theControl;
short message;
void *param;
{                               /* Working rectangles. */
   Rect scratch, scratch2, pageUpRect, pageDnRect, thumbRect;
   short horizontal;            /* Flag for scroll bar orientation. */
   BitMap arrows;               /* Arrow bitmaps. */
   extern U8 _upArrow[], _downArrow[], _upArrow2[], _downArrow2[];
   extern U8 _leftArrow[], _rightArrow[], _leftArrow2[], _rightArrow2[];
   short ii, jj, height;        /* Scratch variables. */
   EventRecord evt;             /* Used to track the mouse. */
   Point hit;                   /* " */
   short saveFont, saveSize, saveFace, saveMode; /* For saving state. */
   short saveFgColor, saveBkColor;
   FontInfo finfo;
   short penisNormal;           /* For efficiency. */

   switch (message) {
      case drawCntl:            /* Draw all or part of the control. */
                                /* In this case param's part code to hilite. */
         if (!theControl->contrlVis)
            return 0;
         saveFont = thePort->txFont;
         saveSize = thePort->txSize;
         saveFace = thePort->txFace;
         saveMode = thePort->txMode;
         if (theControl->contrlTitle && varCode != scrollBarProc) {
            if (saveFont != theControl->contrlFont)
               TextFont(theControl->contrlFont);
            if (saveSize != theControl->contrlSize)
               TextSize(theControl->contrlSize);
            if (saveFace != normalText)
               TextFace(normalText);
            if (saveMode != srcOr)
               TextMode(srcOr);
         }
         saveBkColor = thePort->bkColor;
         if (saveBkColor != theControl->contrlBColor)
            BackColor(theControl->contrlBColor);
         saveFgColor = thePort->fgColor;
         if (saveFgColor != theControl->contrlFColor)
            ForeColor(theControl->contrlFColor);
         PenNormal();
         penisNormal = 1;
         switch (varCode) {
            case pushButProc:     /* Draw simple button. */
            case defButProc:
               scratch = theControl->contrlRect;
               if (varCode == defButProc) {
                  if (!param)
                     FillRoundRect(&scratch, 16, 16, &white);
                  PenSize(2, 2);        /* Draw heavy outline. */
                  FrameRoundRect(&scratch, 16, 16);
                  InsetRect(&scratch, 4, 4);
                  PenSize(1, 1);
               }
               if (!param)
                  FillRoundRect(&scratch, 8, 8, &white);
               FrameRoundRect(&scratch, 8, 8);
               InsetRect(&scratch, 1, 1);
               FillRoundRect(&scratch, 8, 8, &white);
               if (theControl->contrlTitle) {
                  MoveTo(theControl->contrlText.h, theControl->contrlText.v);
                  DrawText(theControl->contrlTitle, 0, theControl->titleLen);
               }
               if (theControl->contrlHilite == 255) { /* Gray-out the control. */
                  PenMode(srcBic);
                  PenPat(&syspats[7]);
                  PaintRoundRect(&scratch, 8, 8);
                  penisNormal = 0;
               } else if ((int)param == inButton) {
                  InsetRect(&scratch, -1, -1);
                  InvertRoundRect(&scratch, 8, 8);
               }
               break;

            case checkBoxProc:  /* Draw check box. */
               SetRect(&scratch, theControl->contrlRect.edge.left,
                       theControl->contrlRect.edge.top,
                       theControl->contrlRect.edge.left + 12,
                       theControl->contrlRect.edge.top + 12);
               if (!param) {    /* Draw the whole thing. */
                  FillRect(&theControl->contrlRect, &white);
                  if (theControl->contrlTitle) {
                     MoveTo(theControl->contrlText.h, theControl->contrlText.v);
                     DrawString(theControl->contrlTitle);
                  }
               } else
                  FillRect(&scratch, &white);
               FrameRect(&scratch);
               if (theControl->contrlValue) {
                  if (theControl->contrlValue == 2) {
                     InsetRect(&scratch, 1, 1);
                     FillRect(&scratch, &gray);
                     InsetRect(&scratch, -1, -1);
                  } else {
                     MoveTo(scratch.edge.left, scratch.edge.top);
                     LineTo(scratch.edge.right - 1, scratch.edge.bottom - 1);
                     MoveTo(scratch.edge.left, scratch.edge.bottom - 1);
                     LineTo(scratch.edge.right - 1, scratch.edge.top);
                  }
               }
               if (theControl->contrlHilite == inCheckBox) { /* Hilighting. */
                  InsetRect(&scratch, 1, 1);
                  FrameRect(&scratch);
                  InsetRect(&scratch, -1, -1);
               } else if (theControl->contrlHilite == 255) { /* Gray-out? */
                  thumbRect = theControl->contrlRect;
                  thumbRect.edge.left = scratch.edge.right;
                  PenMode(srcBic);
                  PenPat(&syspats[7]);
                  PaintRect(&thumbRect);
                  penisNormal = 0;
               }
               break;

            case radioButProc:  /* Draw radio button. */
               SetRect(&scratch, theControl->contrlRect.edge.left,
                       theControl->contrlRect.edge.top,
                       theControl->contrlRect.edge.left + 12,
                       theControl->contrlRect.edge.top + 12);
               if (!param) {    /* Draw the whole thing. */
                  FillRect(&theControl->contrlRect, &white);
                  if (theControl->contrlTitle) {
                     MoveTo(theControl->contrlText.h, theControl->contrlText.v);
                     DrawString(theControl->contrlTitle);
                  }
               } else
                  FillRect(&scratch, &white);
               FrameRoundRect(&scratch, 8, 8);
               if (theControl->contrlHilite == inCheckBox) { /* Hilighting. */
                  InsetRect(&scratch, 1, 1);
                  FrameRoundRect(&scratch, 8, 8);
                  InsetRect(&scratch, -1, -1);
               } else if (theControl->contrlHilite == 255) { /* Gray-out? */
                  thumbRect = theControl->contrlRect;
                  thumbRect.edge.left = scratch.edge.right;
                  PenMode(srcBic);
                  PenPat(&syspats[7]);
                  PaintRect(&thumbRect);
                  penisNormal = 0;
               }
               if (theControl->contrlValue) {
                  if (!penisNormal)
                     PenNormal();
                  if (theControl->contrlValue == 2) {
                     PenPat(&gray);
                     penisNormal = 0;
                  }
                  InsetRect(&scratch, 3, 3);
                  PaintRoundRect(&scratch, 4, 4);
               }
               break;

            case scrollBarProc: /* Draw scroll bar. */
               horizontal = _calcScrollRects(theControl, &scratch, &scratch2,
                                           &pageUpRect, &pageDnRect, &thumbRect);
               SetRect(&arrows.bounds, 0, 0, 16, 16);
               arrows.rowBytes = 2;
               arrows.bpp = 1;
               if (!param) {    /* Draw the whole thing. */
                  FrameRect(&theControl->contrlRect);
                  if (horizontal) {
                     MoveTo(scratch.edge.left + 16, scratch.edge.top);
                     Line(0, scratch.edge.bottom - scratch.edge.top - 1);
                     PenPat(&white);
                     MoveTo(pageUpRect.edge.left, pageUpRect.edge.top - 1);
                     Line(pageDnRect.edge.right - pageUpRect.edge.left - 1, 0);
                     MoveTo(pageUpRect.edge.left, pageUpRect.edge.bottom);
                     Line(pageDnRect.edge.right - pageUpRect.edge.left - 1, 0);
                  } else {
                     MoveTo(scratch.edge.left, scratch.edge.top + 16);
                     Line(scratch.edge.right - scratch.edge.left - 1, 0);
                     PenPat(&white);
                     MoveTo(pageUpRect.edge.left - 1, pageUpRect.edge.top);
                     Line(0, pageDnRect.edge.bottom - pageUpRect.edge.top - 1);
                     MoveTo(pageUpRect.edge.right, pageUpRect.edge.top);
                     Line(0, pageDnRect.edge.bottom - pageUpRect.edge.top - 1);
                  }
                  PenNormal();
                  penisNormal = 1;
               }
               if (!param || (int)param == inUpButton) {
                  if (horizontal)
                     arrows.baseAddr = (theControl->contrlHilite == inUpButton)
                                     ? _leftArrow2 : _leftArrow;
                  else
                     arrows.baseAddr = (theControl->contrlHilite == inUpButton)
                                     ? _upArrow2 : _upArrow;
                  scratch.edge.bottom--;
                  scratch.edge.right--;
                  CopyBits(&arrows, &thePort->portBits, &arrows.bounds, &scratch,
                           srcCopy, 0); /* Draw the up-arrow. */
                  scratch.edge.bottom++;
                  scratch.edge.right++;
               }
               if (!param || (int)param == inDownButton) {
                  if (horizontal)
                     arrows.baseAddr = (theControl->contrlHilite == inDownButton)
                                     ? _rightArrow2 : _rightArrow;
                  else
                     arrows.baseAddr = (theControl->contrlHilite == inDownButton)
                                     ? _downArrow2 : _downArrow;
                  scratch2.edge.bottom--;
                  scratch2.edge.right--;
                  CopyBits(&arrows, &thePort->portBits, &arrows.bounds,
                           &scratch2, srcCopy, 0); /* Draw the down-arrow. */
                  scratch2.edge.bottom++;
                  scratch2.edge.right++;
               }
               if (theControl->contrlMin == theControl->contrlMax
                   || theControl->contrlHilite == 255) { /* De-activate? */
                  pageUpRect.corner.botRight = pageDnRect.corner.botRight;
                  FillRect(&pageUpRect, &white);
                  break;
               }
               if (!param || (int)param == inPageUp || (int)param == inPageDown
                   || (int)param == inThumb) { /* Draw the elevator area. */
                  FillRect(&pageUpRect, &ltGray); /* Grey the bar. */
                  FillRect(&pageDnRect, &ltGray); /* Grey the bar. */
                  InsetRect(&thumbRect, 1, 1);
                  FillRect(&thumbRect, &white);
                  InsetRect(&thumbRect, -1, -1);
                  FrameRect(&thumbRect);
               }
               break;
         }
         if (saveBkColor != thePort->bkColor)
            BackColor(saveBkColor);
         if (saveFgColor != thePort->fgColor)
            ForeColor(saveFgColor);
         if (theControl->contrlTitle) {
            if (saveFont != thePort->txFont)
               TextFont(saveFont);
            if (saveSize != thePort->txSize)
               TextSize(saveSize);
            if (saveFace != thePort->txFace)
               TextFace(saveFace);
            if (saveMode != thePort->txMode)
               TextMode(saveMode);
         }
         if (!penisNormal)
            PenNormal();
         return 0;
         break;

      case testCntl:            /* Find which part of control hit was in. */
         if (!theControl->contrlVis || theControl->contrlHilite == 255)
            return 0;           /* In this case, param -> the test point. */
         if (PtInRect((Point *) param, &theControl->contrlRect))
            switch (varCode) {
               case pushButProc:        /* Hit in simple button. */
               case defButProc:
                  return (void *)inButton;

               case checkBoxProc: /* Hit in check box. */
               case radioButProc: /* Hit in radio button. */
                  return (void *)inCheckBox;

               case scrollBarProc: /* Hit in scroll bar. */
                  if (theControl->contrlMin == theControl->contrlMax)
                     return 0;  /* Effectively inactive in this case. */
                  _calcScrollRects(theControl, &scratch, &scratch2,
                                   &pageUpRect, &pageDnRect, &thumbRect);
                  if (PtInRect((Point *) param, &scratch))
                     return (void *)inUpButton;

                  if (PtInRect((Point *) param, &scratch2))
                     return (void *)inDownButton;

                  if (PtInRect((Point *) param, &thumbRect))
                     return (void *)inThumb;

                  if (PtInRect((Point *) param, &pageUpRect))
                     return (void *)inPageUp;

                  if (PtInRect((Point *) param, &pageDnRect))
                     return (void *)inPageDown;
                  break;
            }
         else
            return 0;

      case calcCRgns:           /* Calculate control's regions. */
         GetFontInfo(&finfo);
         switch (varCode) {
            case pushButProc:   /* Simple button. */
            case defButProc:
               ii = theControl->contrlTitle ?
                  TextWidth(theControl->contrlTitle, 0,
                            theControl->titleLen) : 0;
               jj = (theControl->contrlRect.edge.right
                  - theControl->contrlRect.edge.left) / 2;
               theControl->contrlText.h = theControl->contrlRect.edge.left
                                        + jj - ii/2;
               ii = finfo.ascent + finfo.descent;
               jj = theControl->contrlRect.edge.bottom
                  - theControl->contrlRect.edge.top;
               theControl->contrlText.v = theControl->contrlRect.edge.top
                                        + finfo.ascent + 1 + (jj - ii) / 2;
               break;

            case checkBoxProc:  /* Check box. */
            case radioButProc:  /* Radio button. */
               theControl->contrlText.h = theControl->contrlRect.edge.left + 20;
               theControl->contrlText.v = theControl->contrlRect.edge.top
                                        + finfo.ascent + 1;
               break;
         }
         break;

      case dragCntl:            /* Drag the control or indicator around. */
         if (!theControl->contrlVis)
            return 0;
         switch (varCode) {
            case scrollBarProc:
               horizontal = _calcScrollRects(theControl, &scratch, &scratch2,
                                           &pageUpRect, &pageDnRect, &thumbRect);
               UnionRect(&pageUpRect, &pageDnRect, &scratch); /* Pinning rect. */
               scratch2 = scratch;      /* Slop rectangle. */
               if (horizontal) {
                  InsetRect(&scratch2, -100, -20);
               } else {
                  InsetRect(&scratch2, -20, -100);
               }
               pageUpRect = thumbRect; /* Elevator's starting point. */
               pageDnRect = thumbRect; /* Displaced thumb. */
               PenNormal();     /* Prepare for elevator highlighting. */
               PenPat(&ltGray);
               PenMode(notSrcXor);
               ii = 0;          /* We need to re-draw the thumb. */
               do {             /* Now track away! */
                  if (!ii) {    /* Draw the thumb if it's changed. */
                     FrameRect(&thumbRect);
                     ii = 1;
                  }
#if 0
                  evt = GetEvent(1<<lastWindow, 1<<mouseEvt);
#else
                  (void)GetNextEvent(1<<mouseEvt, &evt);
#endif
                  hit = evt.where.corner.topLeft;
                  if (PtInRect(&hit, &scratch2)) { /* If within tracking rect. */
                     if (horizontal) {
                        jj = PinRect(&scratch, &hit).h;
                        jj -= ((Point *) param)->h; /* How far to move thumb. */
               /* Make sure thumb doesn't get out of the scroll box area. */
                        if (jj > scratch.edge.right - pageUpRect.edge.right)
                           jj = scratch.edge.right - pageUpRect.edge.right;
                        else if (jj < scratch.edge.left - pageUpRect.edge.left)
                           jj = scratch.edge.left - pageUpRect.edge.left;
                        pageDnRect.edge.left = pageUpRect.edge.left + jj;
                        pageDnRect.edge.right = pageUpRect.edge.right + jj;
                     } else {   /* Vertical. */
                        jj = PinRect(&scratch, &hit).v;
                        jj -= ((Point *) param)->v; /* How far to move thumb. */
               /* Make sure thumb doesn't get out of the scroll box area. */
                        if (jj > scratch.edge.bottom - pageUpRect.edge.bottom)
                           jj = scratch.edge.bottom - pageUpRect.edge.bottom;
                        else if (jj < scratch.edge.top - pageUpRect.edge.top)
                           jj = scratch.edge.top - pageUpRect.edge.top;
                        pageDnRect.edge.top = pageUpRect.edge.top + jj;
                        pageDnRect.edge.bottom = pageUpRect.edge.bottom + jj;
                     }
                     if (!EqualRect(&pageDnRect, &thumbRect)) {
                        if (ii == 1) /* Take down old rectangle? */
                           FrameRect(&thumbRect); /* Yes. */
                        thumbRect = pageDnRect;
                        ii = 0;  /* Force redraw. */
                     }
                  } else {      /* Outside tracking rectangle. */
                     if (ii == 1)
                        FrameRect(&thumbRect); /* Force highlighting off. */
                     ii = 2;     /* Stay off 'till back in the sloprect. */
                     SetRect(&thumbRect, 0, 0, 0, 0);
                     jj = 0;     /* No drag if we released mouse here. */
                  }
               } while (!(evt.message & F_KEYUP)); /* Done tracking mouse. */
               PenNormal();
               return (void *)(int)jj; /* Acknowledge that we dragged. */
         }
         return 0;
         break;

      case posCntl:             /* Position the control's indicator. */
         horizontal = _calcScrollRects(theControl, &scratch, &scratch2,
                                       &pageUpRect, &pageDnRect, &thumbRect);
         UnionRect(&pageUpRect, &pageDnRect, &scratch);
               /* J equals the height of the free elevator area in pixels. */
               /* I equals the new position within this of the elevator. */
         if (horizontal) {
            jj = scratch.edge.right - scratch.edge.left
               - (thumbRect.edge.right - thumbRect.edge.left);
            ii = thumbRect.edge.left + (int)param - scratch.edge.left;
         } else {
            jj = scratch.edge.bottom - scratch.edge.top
               - (thumbRect.edge.bottom - thumbRect.edge.top);
            ii = thumbRect.edge.top + (int)param - scratch.edge.top;
         }
         jj = theControl->contrlMin
            + ScaleRS(ii, theControl->contrlMax - theControl->contrlMin, jj);
         if ((jj == theControl->contrlValue) && param) { /* Force a change of */
            if ((int)param > 0) /* at least 1 in value if the control's been */
               jj++;            /* moved a pixel or more. */
            else
               jj--;
         }
         SetCtlValue(theControl, jj);
         break;
   }
}
