#include "../clunkdraw.h"
#include "cdcurs.h"
#define TRUE 1
#define BLACK 0x0000 // RGB565 colors for the cursor (mouse pointer).
#define WHITE 0xFFFF

/*
** ShowCursor increments the cursor level, which may have been
** decremented by HideCursor, and displays the cursor on the
** screen if the level becomes 0.  A call to ShowCursor should
** balance each previous call to HideCursor.  The level isn't
** incremented beyond 0, so extra calls to ShowCursor have no
** effect.
**
** If the cursor has been changed (with SetCursor) while hidden,
** ShowCursor presents the new cursor.
**
** Currently only works with 16-bit color displays, because that's
** the only one I have that works 'live'.  (Mouse cursors are not
** currently useful on E-Paper displays.)
*/

void
ShowCursor(void)
{
   int ii, jj, x, y;
   U8 mask;
   const U8 *cp, *mp;
   U16 saveBk, saveFg;
   GrafPtr tp;

   if (_CrsrInited) {
      if (++_CrsrState >= 0) {
         _CrsrState = 0;        // Pin it to 0 (visible).
         if (!_CrsrVis) {
            SetRect(&_CrsrRect, _Mouse.h, _Mouse.v,
                                _Mouse.h + 16, _Mouse.v + 16);
            OffsetRect(&_CrsrRect, -_TheCrsr.hotSpot.h, -_TheCrsr.hotSpot.v);
            if ((tp = thePort) && tp->portBits.bpp == 16) {
               saveBk = tp->bkColor; tp->bkColor = WHITE;
               saveFg = tp->fgColor; tp->fgColor = BLACK;
               cp = (const U8 *)_TheCrsr.data;
               mp = (const U8 *)_TheCrsr.mask;
               for (ii = 0, y = _CrsrRect.edge.top;
                            y < _CrsrRect.edge.bottom; y++, ii++) {
                  mask = 0x80;
                  for (jj = 0, x = _CrsrRect.edge.left;
                               x < _CrsrRect.edge.right; x++, jj++) {
                     // Save the under-Cursor screen bits.
                     _CrsrSave[ii][jj] = (*tp->_GetPix)(tp, x, y);
                     // Paint the cursor onto the screen.
                     if (*mp & mask) {
                        if (*cp & mask)
                           (*tp->_SetPix)(tp, x, y);
                        else
                           (*tp->_ClearPix)(tp, x, y);
                     }
                     if (!(mask >>= 1)) {
                        cp++;
                        mp++;
                        mask = 0x80;
                     }
                  }
               }
               tp->bkColor = saveBk;
               tp->fgColor = saveFg;
            }
         }
         _CrsrVis = TRUE;
      }
   }
}
