#include "../clunkdraw.h"
#include "cdtftrpi.h"
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/fb.h>
#include <sys/mman.h>

static const char _initcmds[] = {
   "sudo sh -c '"
   "killall fbcp;"              // Disconnect /dev/fb1 from console image.
// "modprobe -r fb_ili9340;"    // Disconnect SPI from LCD.
// "raspi-gpio set 27 op dl;"   // Reset LCD
   "raspi-gpio set 18 op dh;"   // Turn on LCD backlight
// "raspi-gpio set 27 op dh;"   // Take LCD out of reset.
// "modprobe fb_ili9340;"       // Reconnect SPI to LCD.
   "'"
};

static int fbfd;                /* Frame buffer file descriptor. */
static char *fbp;               /* Frame buffer info, for releasing later. */
static int msize;

/*
** Hook up to our SPI-connected TFT LCD RGB565 screen.  We will ASSUME
** that it is on /dev/fb1, let's hope that doesn't change.
*/

int
TFT_SPI_Init(bmp)
BitMap *bmp;
{
   struct fb_var_screeninfo vinfo;
   struct fb_fix_screeninfo finfo;
   int rowBytes;

   (void)system(_initcmds);     // Get SPI hooked up to display.

   fbfd = open("/dev/fb1", O_RDWR|O_CLOEXEC);
   if (fbfd < 0)
      return 1;
   if (ioctl(fbfd, FBIOGET_FSCREENINFO, &finfo)) {
      close(fbfd);
      fbfd = -1;
      return 2;
   }
   if (ioctl(fbfd, FBIOGET_VSCREENINFO, &vinfo)) {
      close(fbfd);
      fbfd = -1;
      return 3;
   }

   SetRect(&bmp->bounds, 0, 0, vinfo.xres, vinfo.yres);
   bmp->rowBytes = (vinfo.xres * vinfo.bits_per_pixel + 7) / 8;
   bmp->bpp = vinfo.bits_per_pixel;
#if 0
   printf("Display is %d x %d %dbpp\n",
          vinfo.xres, vinfo.yres, vinfo.bits_per_pixel);
#endif
   msize = finfo.smem_len;
   fbp = mmap(0, msize, PROT_READ | PROT_WRITE, MAP_SHARED, fbfd, 0);
   if (fbp <= 0) {
      close(fbfd);
      fbfd = -1;
      return 4;
   }
   // At this point we have access to the framebuffer memory, and can
   // just draw in it.  The kernel (and hardware) will take care of
   // continually copying it out to the display, so it appears 'live'.
   bmp->baseAddr = fbp;
   return 0;
}

// Could re-start the system's fbcp process at this point, restoring
// console information to the LCD screen, but at the moment I choose not to.
void
TFT_SPI_End()
{
   munmap(fbp, msize);
   fbp = 0;
   close(fbfd);
   fbfd = -1;
}
