// Select from the embedded fonts, it only finds a font/size that
// exists pre-built.  The first font in the EmbeddedFonts array is
// used as the default font, if nothing is found.
//
// The scaling system is neither present nor engaged unless InitFonts
// is called.

#include "../clunkdraw.h"
#include "cdfontm.h"

static const FontTableEntry *
_GetFontFromTable(GrafPtr tp)
{
   const FontTableEntry *ftep;

   for (ftep = &EmbeddedFonts; ftep->frp; ftep++) {
      if (tp->txFont == ftep->font && tp->txSize == ftep->size)
         break;
   }
   return ftep;
}

// This one is used to build the scaling system's initial font chain.
// No substitutions!
//
FontRec *
_GetEmbeddedFont(GrafPtr tp)
{
   const FontTableEntry *ftep;

   ftep = _GetFontFromTable(tp);
   return (void *)(ftep ? ftep->frp : 0);
}

// This one is used when _only_ pre-built fonts are allowed.  (When the
// InitFonts function has not been called.)  Will return the default
// (the first-defined) font if nothing matching is found.
//
FontRec *
_GetEmbFontOrDef(GrafPtr tp)
{
   const FontTableEntry *ftep;

   ftep = _GetFontFromTable(tp);
   return (void *)((ftep && ftep->frp) ? ftep->frp : EmbeddedFonts.frp);
}
