#include "../clunkdraw.h"
#include "cdfontm.h"
#include <stdlib.h> // For calloc and free
#include <string.h> // For memset
#include <endian.h> // For be16toh et al.
#define TRUE 1
#define FALSE 0

static int
charbitwid(U16 *charimage, int charheight, U16 *firstbit, U16 *lastbit)
{
   int ii, jj, kk;
   U16 mask, test;

   test = 0;                   // Make composite of all rows of the char.
   while (charheight--)
      test |= be16toh(*charimage++);
   mask = 0x8000;
   ii = 16;
   jj = 0;
   while (ii--) {       // Find left-most one bit.
      if (test & mask)
         break;
      jj++;
      mask >>= 1;
   }
   mask = 1;
   ii = 16;
   kk = 16;
   while (ii--) {       // Find right-most one bit.
      if (test & mask)
         break;
      kk--;
      mask <<= 1;
   }
   ii = kk - jj;
   if (ii < 0) {
      ii = 0;
      jj = 0;
      kk = 0;
   }
   if (firstbit)
      *firstbit = jj;
   if (lastbit)
      *lastbit = kk;
   return ii;
}

#if 0
void
_writeToFile(unsigned char *frp, int size, char *fName)
{
   int fd;

   fd = creat(fName, 0666);
   if (fd < 0) return;
   write(fd, frp, size);
   close(fd);
}
#endif

//
// Convert one font size into another.  ISC fonts are converted to Mac
// fonts when resized.  Won't resize a 16-bit index font.  Characters are
// resized individually -- for ISC fonts we have no choice, so we do the
// same for Mac fonts even though the bitmap could be resized (larger at
// least) in one step.  (A one-step enlargement of a Mac font should be
// _considerably_ faster than one-at-a-time, but beware inter-character
// bleeding effects, possibly due to smoothing.  Maybe someday...)
//
static const FontRec *
_ScaleFont(const FontRec *frp, int font, int fsize, int match)
{
   FontCtrl *fcp;
   FontRec *nfrp;
   const short *owtab;  // Address of src font's offset/width table.
   short *nowtab;       // Address of new font's offset/width table.
   const U16 *loctab;   // Addr of src font's loc table.
   U16 *nloctab;        // Addr of new font's loc table.
   const U16 *cp;
   U16 *wordstore, *p;  // For converting ISC bit image.
   const unsigned char *baseaddr, *p1;
   BitMap sbits, dbits;
   Rect srcrect, dstrect;
   long size, imagesize, offwid;
   int ii, jj, kk;
   int skinny, nchars, nchars2, nwid, noffs;

   if (be16toh(frp->fontType) == 2)
      return 0;                 // Chinese font, can't do anything (yet).
   else {                       // ISC or Macintosh font.
      skinny = 1;
    retry:                      // Retry point for too-large font.
      wordstore = 0;
      ii = 0;                   // New image width.
      nchars = (be16toh(frp->lastChar) - be16toh(frp->firstChar)) + 2;
      nchars2 = nchars;
      if (be16toh(frp->fontType) == 1) {
         cp = &frp->rowWords + 1;
         if (be16toh(frp->widMax) <= 9) { // Narrow ISC font.
            wordstore = malloc(be16toh(frp->fRectHeight) * sizeof(U16));
            cp = &frp->rowWords + 1;
            p1 = (const unsigned char *) cp;
            p = wordstore;
            while (nchars-- > 0) { // Find new bit image width. Source bitmap's
               kk = be16toh(frp->fRectHeight); // only 8-bits wide, must copy to
               while (kk-- > 0)              // a temp buffer as 16-bits wide.
                  *p++ = be16toh(*p1++ << 8);
               p = wordstore;
               jj = charbitwid(p, be16toh(frp->fRectHeight), 0, 0);
               ii += ScaleRS(jj, fsize, match * skinny);
            }
         } else if (be16toh(frp->widMax) <= 17) { // Wider ISC font.
            while (nchars-- > 0) { // Find new bit image width.
               jj = charbitwid(p, be16toh(frp->fRectHeight), 0, 0);
               p += be16toh(frp->fRectHeight) / 2;
               ii += ScaleRS(jj, fsize, match * skinny);
            }
         }
      } else {                  // Macintosh font, has a loctab.
         loctab = (U16 *)&frp[1];
         loctab += be16toh(frp->fRectHeight) * be16toh(frp->rowWords);
         while (nchars-- > 0) {
            ii += ScaleRS(be16toh(loctab[1]) - be16toh(loctab[0]),
                          fsize, match * skinny);
            loctab++;
         }
      }
      ii = ((ii + 15) / 16) * 2; // New rowBytes.  Must be even.
      jj = ScaleRS(be16toh(frp->fRectHeight), fsize, match);
      imagesize = MulS(ii, jj);
      nchars = nchars2;
      size = sizeof(FontRec) + imagesize + (nchars + 1) * 2 * sizeof(U16);
      offwid = 4 + (imagesize / 2) + (nchars + 1) + 1;
      if (offwid > 65535) {     // Will the offset pointer overflow?
         skinny++;              // Yes, font's too big.  Try making it with
         if (wordstore)
            free(wordstore);
         goto retry;            // extra-skinny characters.
      }
      nfrp = calloc(1, size);  // New font image store.
      if (!nfrp) return frp;
      *nfrp = *frp;             // Copy the fields from old to new.
      if (be16toh(frp->fontType) == 1) // Scale EZ ones...
         if (be16toh(frp->owTLoc))
            nfrp->fontType = htobe16(0x9000);
         else
            nfrp->fontType = htobe16(0xB000);
      nfrp->widMax = ScaleRS(be16toh(frp->widMax), fsize, match);
      nfrp->widMax = htobe16(nfrp->widMax);
      nfrp->kernMax = ScaleRS(be16toh(frp->kernMax), fsize, match);
      nfrp->kernMax = htobe16(nfrp->kernMax);
      nfrp->nDescent = ScaleRS(be16toh(frp->nDescent), fsize, match);
      nfrp->nDescent = htobe16(nfrp->nDescent);
      nfrp->fRectWidth = ScaleRS(be16toh(frp->fRectWidth), fsize, match);
      nfrp->fRectWidth = htobe16(nfrp->fRectWidth);
      nfrp->fRectHeight = jj;
      nfrp->fRectHeight = htobe16(nfrp->fRectHeight);
      nfrp->ascent = ScaleRS(be16toh(frp->ascent), fsize, match);
      nfrp->ascent = htobe16(nfrp->ascent);
      nfrp->descent = jj - be16toh(nfrp->ascent);
      nfrp->descent = htobe16(nfrp->descent);
      nfrp->leading = ScaleRS(be16toh(frp->leading), fsize, match);
      nfrp->leading = htobe16(nfrp->leading);
      nfrp->rowWords = ii / 2;
      nfrp->rowWords = htobe16(nfrp->rowWords);
      nfrp->owTLoc = offwid;
      nfrp->owTLoc = htobe16(nfrp->owTLoc);
      sbits.baseAddr = (void *)&frp[1]; // Stretch/shrink the font's bitmap now.
      sbits.bpp = 1;
      baseaddr = sbits.baseAddr;
      dbits.baseAddr = &nfrp[1];
      dbits.bpp = 1;
      if (be16toh(frp->fontType) == 1)
         sbits.rowBytes = sizeof(U16);
      else
         sbits.rowBytes = be16toh(frp->rowWords) * sizeof(U16);
      dbits.rowBytes = ii;
      SetPt(&sbits.bounds.corner.topLeft, 0, 0);
      SetPt(&srcrect.corner.topLeft, 0, 0);
      sbits.bounds.edge.bottom = be16toh(frp->fRectHeight);
      srcrect.edge.bottom = sbits.bounds.edge.bottom;
      if (be16toh(frp->fontType) == 1) {
         sbits.bounds.edge.right = be16toh(frp->fRectWidth);
         owtab = 0;
         if (be16toh(frp->owTLoc))
            owtab = (short *)((unsigned char *)frp + be16toh(frp->owTLoc));
      } else {
         sbits.bounds.edge.right = be16toh(frp->rowWords) * 16;
         owtab = &frp->owTLoc;
         owtab += be16toh(frp->owTLoc);
      }
      SetPt(&dbits.bounds.corner.topLeft, 0, 0);
      SetPt(&dstrect.corner.topLeft, 0, 0);
      dbits.bounds.edge.bottom = be16toh(nfrp->fRectHeight);
      dstrect.edge.bottom = dbits.bounds.edge.bottom;
      dbits.bounds.edge.right = be16toh(nfrp->rowWords) * 16;
      p1 = (unsigned char *) (&nfrp[1]) + imagesize;
      nloctab = (U16 *)p1;
      nowtab = (short *)((U16 *)&nfrp->owTLoc + be16toh(nfrp->owTLoc));
      loctab = (U16 *)&frp[1];
      loctab += MulS(be16toh(frp->fRectHeight), be16toh(frp->rowWords));
      while (nchars-- > 0) {
         sbits.baseAddr = (void *)baseaddr;
         if (be16toh(frp->fontType) == 1) {  // ISC font...
            if (be16toh(frp->widMax) <= 9) { // narrow ISC font.
               p = wordstore;              // Source bitmap's only 8 bits wide,
               kk = be16toh(frp->fRectHeight); // must copy to a temp buffer
               while (kk-- > 0)            // as 16-bits wide (for CopyBits).
                  *p++ = *baseaddr++ << 8;
               sbits.baseAddr = (unsigned char *)wordstore;
            }
            jj = charbitwid((U16 *)sbits.baseAddr, sbits.bounds.edge.bottom,
                            &srcrect.edge.left, &srcrect.edge.right);
         } else {               // Macintosh font, has a loctab.
            srcrect.edge.right = be16toh(loctab[1]);
            srcrect.edge.left = be16toh(loctab[0]);
            jj = srcrect.edge.right - srcrect.edge.left;
         }
         ii = ScaleRS(jj, fsize, match * skinny);
         dstrect.edge.right = dstrect.edge.left + ii; // Stretch/shrink bits!
         if (!EmptyRect(&srcrect))
            CopyBits(&sbits, &dbits, &srcrect, &dstrect, srcOr, 0);
         *nloctab++ = be16toh(dstrect.edge.left);
         loctab++;
                                // Calculate new offset/width table entry.
         if (!owtab)            // Fixed-pitch ISC font.
            nwid = be16toh(nfrp->widMax);
         else                   // Proportional ISC or Mac font.
            nwid = ScaleRS(be16toh(*owtab) & 0xFF, fsize, match);
         if (be16toh(frp->fontType) == 1)
            noffs = srcrect.edge.left; // For new offs table entry.
         else                   // Mac font.
            noffs = (U16)be16toh(*owtab) >> 8;
         nwid = (ScaleRS(noffs, fsize, match) << 8) | (nwid & 0xFF);
         if (owtab) {
            if (*owtab++ != -1)
               *nowtab++ = be16toh(nwid);
            else
               *nowtab++ = be16toh(-1);
         } else {
            *nowtab++ = be16toh(nwid);
         }
         dstrect.edge.left += ii;
         dstrect.edge.right += ii;
         if (be16toh(frp->fontType) == 1) {
            if (be16toh(frp->widMax) > 9)
               baseaddr += be16toh(frp->fRectHeight) * sizeof(U16);
         } else {               // Mac font.
            srcrect.edge.left += jj;
            srcrect.edge.right += jj;
         }
      }
      *nloctab++ = be16toh(dstrect.edge.left);
      *nowtab++ = -1;
      if (wordstore)
         free(wordstore);
   }
#if 0 // For debugging using dumpfont program.
   _writeToFile(PByte(nfrp), size, 'font-' + IntToStr(fsize) + '.fnt');
#endif

   fcp = _fontChain;
   _fontChain = malloc(sizeof *_fontChain);
   _fontChain->fontNum = font;        // Fill in the chained font record.
   _fontChain->fontSize = fsize;
   _fontChain->fontPtr = nfrp;
   _fontChain->memSize = size;
   _fontChain->next = fcp;
   _fontChain->used = TRUE;
   return nfrp;
}

//
// Function to find all sizes of a font that are available in store.
// Returns a filled vector of sizes.
//

static Boolean
_GetFontSizeV(int font, U8 *vector, int sizev)
{
   Boolean ii;
   FontCtrl *fcp;

   ii = FALSE;
   memset(vector, 0, sizev);
   for (fcp = _fontChain; fcp; fcp = fcp->next)
      if (fcp->fontNum == font && fcp->fontSize > 0 && fcp->fontSize < sizev)
         vector[fcp->fontSize] = ii = TRUE;
   return ii;
}

static const FontRec *
_GetExistingFont(int font, int size, Boolean use)
{
   FontCtrl *fcp;

   for (fcp = _fontChain; fcp; fcp = fcp->next)
      if (fcp->fontNum == font && fcp->fontSize == size) {
         if (use)
            fcp->used = TRUE;
         return fcp->fontPtr;
      }
   return 0;
}

//
// Internal function called whenever we actually need a font, either for
// drawing or measuring.  Will fetch (from cache if possible) the font,
// or it will resize a font as necessary.
//

FontRec *
_GetScaledFont(GrafPtr tp)
{
   const FontRec *frp;
   int fsize, font, ii, match;
   U8 available[256];           // Array of available font sizes.

   if (!tp) return 0;           // Protect ourself from bozo calls.
   fsize = tp->txSize;
   if (!fsize)                  // Get font's point size.  Since we're _making_
      fsize = 12;               // fonts, we need a real target size.
   font = tp->txFont;
   frp = _GetExistingFont(font, fsize, TRUE); // Find font if already there.
   if (!frp) {                  // Font already allocated or available?
                                // No, open the font store.
      ii = _GetFontSizeV(font, available, sizeof available);
      if (!ii) {                // Failed, use default (built-in) font.
failed:  TextFont(EmbeddedFonts.font);
         TextSize(EmbeddedFonts.size);
         frp = EmbeddedFonts.frp;
         return (void *)(tp->fontPtr = frp);
      }
      match = 0;                // Now, find best original size to make a
      ii = fsize * 2;           // scaled font out of.  Try 2x first,
      if (ii < sizeof(available) && available[ii])
         match = ii;
      else {                    // then try 1/2x next,
         ii = fsize / 2;
         if (!(fsize & 1) && available[ii]) // Watch out for rounding!
            match = ii;
         else {                 // then look for next largest,
            for (ii = fsize + 1; ii < sizeof available; ii++)
               if (available[ii]) {
                  match = ii;
                  break;
               }
            if (!match) {       // then look (finally) for next smallest.
               for (ii = fsize - 1; ii; ii--)
                  if (available[ii]) {
                     match = ii;
                     break;
                  }
            }
         }
      }
      if (!match)               // Total failure, use the default
         goto failed;           // (built-in) font.
      frp = _GetExistingFont(font, match, FALSE); // Get the raw (unscaled) font
      frp = _ScaleFont(frp, font, fsize, match);  // we'll use, and scale it.
   }
   return (void *)(tp->fontPtr = frp);
}
