/*
** Equates for the Pull-Down Menu package.
*/

#define MENUSINBAR 16           /* Number of menus allowed in menu bar. */
#define MENUFONT 2              /* Font to use for menus, if not overridden. */
#define MENUSIZE 18             /* Size of menu font, if not overridden. */

#define KBDTIME 30              /* 30-second kbd menu navigation timeout. */

#define NOCLOCK 0
#define CLOCK24HS 1
#define CLOCK24H 2
#define CLOCK12HS 3
#define CLOCK12H 4
#define CLOCK12HAP 5

/*
** Low-level structures.
*/

struct MenuItem {
   long keyEquivalent;          /* Keyboard equivalent (if any). */
   unsigned char menuMark;      /* Extra marks (checkmark, etc.) */
   unsigned char itemStyle;     /* Style of item. */
   int itemLen;                 /* Length of item title string. */
   char *itemTitle;             /* The text of the item. */
   struct MenuItem *nextItem;   /* Pointer to next item in chain. */
};

typedef struct MenuItem MenuItem;

struct MenuInfo {               /* Menu Structure. */
   int menuID;                  /* Menu ID. */
   int menuWidth;               /* Menu width in pixels. */
   int menuHeight;              /* Menu height in pixels. */
   int numberItems;             /* Number of items in menu. */
   unsigned long enabled;       /* Bitset of enabled items. */
   char *menuTitle;             /* Pointer to Menu Title. */
   MenuItem *items;             /* Pointer to Menu Item linked list. */
   struct MenuInfo *nextMenu;   /* Pointer to next Menu Info */
   int leftCoord;               /* Coordinate of left of title space. */
   int rightCoord;              /* Coordinate of right of title space. */
};

typedef struct MenuInfo MenuInfo;

#define SystemID        (-1)    /* Menu ID for the system menu */
extern MenuInfo *SystemMenu;    /* Pointer to System Menu structure */
extern int OpenedMenu;          /* Currently opened menu (for kbd work). */

typedef struct MenuVars {
   MenuInfo *SystemMenu;        /* Pointer to System Menu structure */
   int OpenedMenu;              /* Currently opened menu (for kbd work). */
   int kbdNav;                  /* TRUE if keyboard menu navigation used. */
                                /* Also used as a keyboard menu timer. */
   int ExitingMenu;             /* Exiting after select, Eat K_ENTER | F_KEYUP */

   int MenuCount;               /* Number of menus in Menu Bar */
   int menuFlash;               /* Flash count for menus. */
   int TheMenu;                 /* Currently highlighted menu. */
   int TheMenuItem;             /* Currently highlighted menu item. */
   int PendingEndmenu;          /* There's a pending EndMenu somewhere. */
   int NoMenuKeyXlate;          /* Stop translating menu keys 'till EndMenu. */
   Rect menuRect;               /* For working on stuff. */
   BitMap *srcBits, dstBits;    /* Bitmap structures for menu work. */
   int mbarHeight;              /* Height (pixels) of menu bar. */
   int mbarWidth;               /* Width (pixels) occupied by menu bar. */
   int scrnWidth;               /* Width (pixels) of screen. */
   int displayStart;
   int displayEnd;
   int clockStart;
   int clockEnd;
   int clockMode;               /* Menubar clock mode */
   int timerRunning;            /* Timer running flag */
   char *displayText;           /* Pointer to text to display in menu bar */
   FontInfo fontInfo;           /* Font information. */
   int chWidth;
   int menuFont;
   int menuFontSize;
} MenuVars;

// Private menu system variables, collected together.

MenuVars _MV;

///void KillMenuKeyXlate(struct ccb *ccb);
GrafPtr _MenuSetup(void);
void _MenuRestore(GrafPtr savePort);
MenuInfo *_FindMenu(MenuInfo *mp, int menuID);
unsigned char *_MenuKeystr(long keyCode, unsigned char *where);
void _MenuDisplay(int start, int stop, char *text, char *text2, int changing);
void MenuDisplay(char *text);
void DumpMenu(MenuInfo *mp);
void _InvertMenu(int menuID, int flag);
void _CalcMenuSize(MenuInfo *mp);
void _FreeMenuItems(MenuInfo *mp);
int _InvertMitem(int itemID, int flag);
MenuInfo *_GetItemPointer(MenuInfo *mp, int menuID, int item, MenuItem **pip);
