#define MENUS_SOURCE
#include "../clunkdraw.h"
#include "menus.h"

/* Check if a keyboard event belongs to the Menu system.  Returns -1 if key */
/* event is to be passed on unchanged, a menu/item pair if it's a menu */
/* choice (whether by navigation or key-equivalence), or 0 if the event */
/* is to be discarded (was eaten internally by menu navigation). */

long
MenuKey(key)
long key;
{
   MenuInfo *mp;
   MenuItem *ip;
   long matchKey, workingKey;
   int flags;
   int itemNo, keyCode;

   keyCode = key & F_KEYMASK;
   if (!keyCode)                /* We can't possibly match a NULL keycode. */
      return -1;                /* Therefore bag the testing.  We only get */
                                /* these as Sync points from the KES system */
                                /* via Wincom, and these tests all fail when */
                                /* testing a NULL so we short-circuit them. */
   flags = key & F_FLAGMASK;
   matchKey = key & (F_KEYMASK | F_FUNC | F_CTRL | F_ALT);
   if (OpenedMenu)              /* Don't allow FKEY direct matches if a menu */
      matchKey &= ~F_FUNC;      /* is open (it interferes with navigation). */
   mp = _MV.SystemMenu;  /* Check menu-key entries in the menus for a match. */
   while (mp) {                 /* Search each menu... */
      ip = mp->items;
      itemNo = 1;
      while (ip) {              /* Search each item... */
         workingKey = matchKey;
         if (!(ip->keyEquivalent & F_SHIFT)) { /* Case-insensitive? */
            if (!(flags & F_FUNC) && keyCode >= 'a' && keyCode <= 'z')
               workingKey &= ~('a' - 'A'); /* Yes, mask to upper-case. */
         } else                 /* Case-sensitive. */
            workingKey |= F_SHIFT;
         if (ip->keyEquivalent == workingKey) { /* Found? */
            if (!(mp->enabled & 1) || !(mp->enabled & (1<<itemNo))
                || _MV.NoMenuKeyXlate)
               return -1;       /* Skip disabled menus/items, or if told to. */
            if ((flags & F_KEYDOWN) || /* Allow auto-repeat */
                (!(flags & F_KEYUP) && !_MV.PendingEndmenu)) {
               if (OpenedMenu) { /* menu events if not in one already. */
                  _MV.kbdNav = 0; /* Keyboard menu navigation. */
                  CloseMenu();
                  HiliteMenu(0);
               }
               if (!_MV.PendingEndmenu) { /* For when they're stacked up. */
                  HiliteMenu(_MV.PendingEndmenu = mp->menuID);
                  _MV.TheMenu = 0;
               }
               return (mp->menuID << 16) | itemNo;      /* Return it. */
            } else {
               return 0;        /* Key matches but not down code or is still */
            }                   /* processing the last one.  Ignore it. */
         }
         ip = ip->nextItem;
         itemNo++;
      }
      mp = mp->nextMenu;
   }

   key &= F_KEYMASK;            /* Wasn't a key-equivalent for a menu. */
                                /* Are we navigating with the keyboard then? */
   if (_MV.ExitingMenu && (key == K_ESC || key == K_ENTER || key == K_NPENTER
                       || key == K_RESET || key == K_SEND)) {
      if ((flags & F_KEYUP))
         _MV.ExitingMenu = 0;
      if (!(flags & F_KEYDOWN))
         return 0;
   }

   /* Time to open a new menu? */
   if ((flags & (F_FUNC | F_CTRL | F_ALT | F_SHIFT)) == (F_FUNC | F_ALT) &&
       (key == K_ESC || key == K_RESET ||
        (key >= K_F1 && key < K_F1 - 1 + _MV.MenuCount))) {
      if (!(flags & F_KEYDOWN) )
         return 0;              /* Menu function key but not down code. */
      if (key == K_ESC || key == K_RESET) {
         key = 0;               /* Alt-ESC (or RESET), open system menu. */
      } else {
         if (_MV.PendingEndmenu) /* No new application menus if there's still */
            return 0;           /* one pending completion. */
         key -= K_F1 - 1;       /* Open application menu. */
      }

      mp = SystemMenu;          /* Find menu to open (based on keycode). */
      while (mp && key-- > 0) {
         mp = mp->nextMenu;
      }
      if (mp && mp->menuID != OpenedMenu) {
         if (OpenedMenu) {
            CloseMenu();        /* Close any other menu that's already open. */
            HiliteMenu(0);
         }
         OpenMenu(mp);
         _MV.kbdNav = KBDTIME;  /* Keyboard menu navigation timeout. */
      }
      return 0;
   } else if (OpenedMenu) {     /* Menu already open */
      if (flags & F_KEYUP )
         return 0;              /* In the menu system so discard up key. */
      if ((key == K_ENTER || key == K_NPENTER || key == K_SEND)
          && _MV.TheMenuItem) { /* Take this menu choice. */
         key = _MV.OpenedMenu > 0 ? _MV.OpenedMenu : 0;
         key = (key << 16) | _MV.TheMenuItem;
         mp = _FindMenu(SystemMenu, OpenedMenu);
         if (mp != SystemMenu) {
            _MV.PendingEndmenu = _MV.OpenedMenu;
            _MV.TheMenu = 0;
         }
         _MV.kbdNav = 0;        /* Keyboard menu navigation. */
         CloseMenu();
         _MV.ExitingMenu = 1;
         return key;
      } else if (key == K_ESC || key == K_ENTER || key == K_NPENTER ||
                 key == K_SEND || key == K_RESET) {
         _MV.kbdNav = 0;        /* Keyboard menu navigation. */
         CloseMenu();
         HiliteMenu(0);
         _MV.ExitingMenu = 1;
         return 0xFFFF;         /* Nonexistant system menu item. */
                                /* Results in an EndMenu only. */
      } else if ((flags & F_FUNC) && (mp = _FindMenu(SystemMenu, OpenedMenu))) {
         if ((flags & F_ALT) == 0 && key >= K_F1 && key < K_F1 + mp->numberItems) {
            key -= K_F1 - 1;
            if ((mp->enabled & 1) && (mp->enabled & (1<<key))) {
               HiliteMenuItem((int) key);
            }
         } else {
            switch (key) {
               case K_UP:
                  if (mp->enabled & 1) {
                     key = _MV.TheMenuItem;
                     do {
                        if (--key < 0) key = mp->numberItems;
                     } while ((mp->enabled & (1<<key)) == 0 &&
                              key != _MV.TheMenuItem);
                     HiliteMenuItem((int) key);
                  }
                  break;
               case K_DOWN:
                  if (mp->enabled & 1) {
                     key = _MV.TheMenuItem;
                     do {
                        if (++key > mp->numberItems) key = 0;
                     } while ((mp->enabled & (1<<key)) == 0 &&
                              key != _MV.TheMenuItem);
                     HiliteMenuItem((int) key);
                  }
                  break;
               case K_RIGHT:
                  if (_MV.PendingEndmenu)
                     break;
                  if (!(mp = mp->nextMenu))
                     mp = _MV.SystemMenu;
                  if (_MV.OpenedMenu != mp->menuID) {
                     CloseMenu();
                     OpenMenu(mp);
                     _MV.kbdNav = KBDTIME; /* Renew keyboard menu timeout. */
                  }
                  break;
               case K_LEFT:
                  if (_MV.PendingEndmenu)
                     break;
                  mp = _MV.SystemMenu;
                  if (mp->menuID == _MV.OpenedMenu) {
                     while (mp->nextMenu) {
                        mp = mp->nextMenu;
                     }
                  } else {
                     while (mp->nextMenu &&
                            mp->nextMenu->menuID != _MV.OpenedMenu) {
                        mp = mp->nextMenu;
                     }
                  }
                  if (_MV.OpenedMenu != mp->menuID) {
                     CloseMenu();
                     OpenMenu(mp);
                     _MV.kbdNav = KBDTIME; /* Renew keyboard menu timeout. */
                  }
                  break;
            }
         }
      }
      return 0;
   }
   return -1;
}
