#define MENUS_SOURCE
#include "../clunkdraw.h"
#include "menus.h"
#include <unistd.h>

/* Check if a mouse event belongs to the Menu system.  Returns 0 if mouse */
/* event is to be passed on unchanged, a menu/item pair if it's a menu */
/* choice, or -1 if the event is to be discarded (was eaten internally by */
/* menu navigation). */

int
MenuMouse(key, ypos, xpos)
long key;
int ypos, xpos;
{
   int ii;
   MenuInfo *mp;

   if (key & F_MOUSE1) {
      if (ypos < _MV.mbarHeight) { /* Within the menu bar? */
         if (_MV.TheMenuItem) { /* Unselect old item if any. */
            HiliteMenuItem(0);
         }
         mp = _MV.SystemMenu;
         while (mp) {           /* Search for which menu this is. */
            if (xpos >= mp->leftCoord && xpos < mp->rightCoord) {
               break;
            }
            if (_MV.PendingEndmenu) /* If menu event is in progress disallow */
               mp = 0;          /* further ones to the application, but */
            else                /* allow system menu searches. */
               mp = mp->nextMenu;
         }
         if (mp == NULL || mp->menuID != _MV.OpenedMenu) {
            if (_MV.OpenedMenu) {
               CloseMenu();
               HiliteMenu(0);
            }
            if (mp) {
               OpenMenu(mp);
            }
         }
         return 0;
      } else if (_MV.OpenedMenu) {
         if ((mp = _FindMenu(_MV.SystemMenu, _MV.OpenedMenu))
             && xpos >= mp->leftCoord && xpos < mp->leftCoord + mp->menuWidth
             && ypos >= _MV.mbarHeight && ypos < _MV.mbarHeight + mp->menuHeight) {
            key = DivS(ypos - _MV.mbarHeight, _MV.fontInfo.ascent +
                       _MV.fontInfo.descent + _MV.fontInfo.leading + 3)
                + 1;
            if (key != _MV.TheMenuItem && key <= mp->numberItems) {
               HiliteMenuItem((int) key);
            }
         } else {
            HiliteMenuItem(0);
         }
         return 0;
      }
      return -1;
   } else if ((key & F_KEYUP) &&
              (key = _MV.TheMenuItem)) { /* Released in an item? */
      if ((mp = _FindMenu(_MV.SystemMenu, _MV.OpenedMenu))
          && (mp->enabled & 1) && (mp->enabled & (1<<_MV.TheMenuItem))) {
         for (ii = 0; ii < _MV.menuFlash; ii++) {
            HiliteMenuItem(ii & 1 ? (int)key : 0); /* Flash selected item */
            usleep(33000);
         }
         key |= _MV.OpenedMenu > 0 ? _MV.OpenedMenu << 16 : 0;
         if (mp != _MV.SystemMenu) {
            _MV.PendingEndmenu = _MV.OpenedMenu;
            _MV.TheMenu = 0;
         }
         CloseMenu();
         return key;
      }
   }
   if (_MV.OpenedMenu) {
      CloseMenu();
      HiliteMenu(0);
   }
   return 0xFFFF;               /* Non-existing system menu item. */
}
