/*
** FramePoly plays back the line-drawing routine calls that define the given
** polygon, using the current grafPort's pen pattern, mode, and size.  The pen
** will hang below and to the right of each point on the boundary of the
** polygon; thus, the polygon drawn will extend beyond the right and bottom
** edges of *(*poly)->polyBBox by the pen width and height, respectively.  All
** other graphic operations occur strictly within the boundary of the polygon,
** as for other shapes.
**
** If a polygon is open and being formed, FramePoly affects the outline of the
** polygon just as if the line-drawing routines themselves had been called.  If
** a region is open and being formed, the outside outline of the polygon being
** framed is mathematically added to the region's boundary.
*/

#include "../clunkdraw.h"
#include "cdpoly.h"

void
FramePoly(poly)
PolyHandle poly;
{
   PolyPtr pp;
   Point *p;
   int ii;

   if (!poly || !thePort) return;
   pp = *poly;
   if (!pp)                     /* Protect against bozo actions. */
      return;
   if (thePort->rgnSave)
      _PolyIntoRgn(poly);
   p = &pp->polyPoints;
   ii = (pp->polySize - sizeof(PolyHdr)) / sizeof(Point); /* # pts. */
   if (ii--) {                  /* Anything in the polygon to draw? */
      MoveTo(p->h, p->v);       /* First point is the MoveTo, */
      p++;
      while (ii--) {            /* the rest are the line segments. */
         LineTo(p->h, p->v);
         p++;
      }
   }
}
