/*
** OpenPoly returns a handle to a new polygon and tells Clunkdraw to start
** saving the polygon definition as specified by calls to line-drawing
** routines.  While a polygon is open, all calls to Line and LineTo affect the
** outline of the polygon.  Only the line endpoints affect the polygon
** definition; the pen mode, pattern, and size do not affect it.  In fact,
** OpenPoly calls HidePen, so no drawing occurs on the screen while the
** polygon is open (unless you call ShowPen just after OpenPoly, or you called
** ShowPen previously without balancing it with a call to HidePen).
**
** A polygon should consist of a sequence of connected lines.  Even though the
** on-screen presentation of a polygon is clipped, the definition of a polygon
** is not; you can define a polygon anywhere on the coordinate plane.
**
** When a polygon is open, the current grafPort's polySave field contains a
** handle to information related to the polygon definition.  If you want to
** temporarily disable the polygon definition, you can save the current value
** of this field, set the field to NULL, and later restore the saved value to
** resume the polygon definition.
**
** WARNING:  Do not call OpenPoly while a region or another polygon is
**           already open.
**
** WARNING:  The polygon may not be referred to or used until the ClosePoly.
*/

#include "../clunkdraw.h"

PolyHandle
OpenPoly()
{
    GrafPtr tp;
    PolyHandle ph;

    tp = thePort;
    if (tp->polySave)
        KillPoly(tp->polySave);
    HidePen();
    ph = (PolyHandle) NewHandle(sizeof(PolyHdr));
    if (tp->polySave = ph)
        (*ph)->polySize = sizeof(PolyHdr);
    return ph;
}
