/*
** Initialize Clunkdraw for work with offscreen 16-bit RGB565 color pixmaps.
*/
#include "../clunkdraw.h"
#include "cdport.h"
#include <stdlib.h> // For calloc
#include <string.h> // For memset

//#define COMMON

#ifdef COMMON
static U16 *
addr(const GrafPtr tp, int x, int y)
{
   int wid, hei;

   wid = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   hei = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return 0;     // the map!
   return tp->portBits.baseAddr + (y * tp->portBits.rowBytes) + x;
}
#endif

// Plot RGB565 (16-bit) pixels in memory.
// These routines work in 0-based physical coordinates, which
// means the layer above can merge patterns cleanly, etc.
//
// Each of these subroutines clearly has more common code than not,
// but we choose not to factor this out in order to keep the speed up.
// (This naive implementation is bad enough as it is!)
// These are at the very bottom of the call tree,
// runtime inefficiency is magnified.
//
void
_RGB565SetPix(const GrafPtr tp, int x, int y) // RGB565 color pixel.
{
   U16 *d;
#ifdef COMMON
   if (!(d = addr(tp, x, y))) return;
#else
   int wid, hei;

   wid = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   hei = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return;       // the pixmap!
   d = tp->portBits.baseAddr + (y * tp->portBits.rowBytes) + (x * sizeof(U16));
#endif
   *d = tp->fgColor;
}

void
_RGB565ClearPix(const GrafPtr tp, int x, int y) // RGB565 color pixel.
{
   U16 *d;
#ifdef COMMON
   if (!(d = addr(tp, x, y))) return;
#else
   int wid, hei;

   wid = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   hei = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return;       // the pixmap!
   d = tp->portBits.baseAddr + (y * tp->portBits.rowBytes) + (x * sizeof(U16));
#endif
   *d = tp->bkColor;
}

void
_RGB565FlipPix(const GrafPtr tp, int x, int y) // RGB565 color pixel.
{
   U16 *d;
#ifdef COMMON
   if (!(d = addr(tp, x, y))) return;
#else
   int wid, hei;

   wid = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   hei = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return;       // the pixmap!
   d = tp->portBits.baseAddr + (y * tp->portBits.rowBytes) + (x * sizeof(U16));
#endif
   *d ^= tp->fgColor ^ tp->bkColor;
}

int
_RGB565GetPix(const GrafPtr tp, int x, int y) // RGB565 color pixel.
{
   U16 *d;
#ifdef COMMON
   if (!(d = addr(tp, x, y))) return 0;
#else
   int wid, hei;

   wid = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   hei = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return 0;     // the pixmap!
   d = tp->portBits.baseAddr + (y * tp->portBits.rowBytes) + (x * sizeof(U16));
#endif
   return *d;
}

void
_BPP_16(int x, int y)
{
   int rowBytes;

   rowBytes = x * sizeof(U16);          // Bytes to contain row of pixels.
   screenBits.rowBytes = rowBytes;
   screenBits.bpp = 16;
   if (!screenBits.baseAddr)
      screenBits.baseAddr = calloc(y, rowBytes);
   memset(screenBits.baseAddr, 0xFF, rowBytes * y); // Background all white.
   memset(&screenPort, 0, sizeof screenPort);
   InitPort(&screenPort.port);
   screenPort.port.fgColor = 0;      /* Foreground = black. */
   screenPort.port.bkColor = 0xFFFF; /* Background = white. */
   screenPort.port._GetPix = _RGB565GetPix;
   screenPort.port._SetPix = _RGB565SetPix;
   screenPort.port._ClearPix = _RGB565ClearPix;
   screenPort.port._FlipPix = _RGB565FlipPix;
}
