/*
** Initialize Clunkdraw for work with offscreen 4-bit grayscale bitmaps.
*/
#include "../clunkdraw.h"
#include "cdport.h"
#include <stdlib.h> // For calloc
#include <string.h> // For memset

#define M 15

//#define COMMON

#ifdef COMMON
static U8 *
addr(const GrafPtr tp, int x, int y)
{
   int wid, hei;

   wid = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   hei = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return 0;     // the map!
   return tp->portBits.baseAddr + (y * tp->portBits.rowBytes) + (x / 2);
}
#endif

// These routines work in 0-based physical coordinates, which
// means the layer above can merge patterns cleanly, etc.
//
// Each of these subroutines clearly has more common code than not,
// but we choose not to factor this out in order to keep the speed up.
// (This naive implementation is bad enough as it is!)
// These are at the very bottom of the call tree,
// runtime inefficiency is magnified.
//
void
_Set4Bit(const GrafPtr tp, int x, int y)
{
   U8 mask, *d;
   int shift;
#ifdef COMMON
   if (!(d = addr(tp, x, y))) return;
#else
   int wid, hei;

   wid = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   hei = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return;       // the map!
   d = tp->portBits.baseAddr + (y * tp->portBits.rowBytes) + (x / 2);
#endif
   shift = (M - (x & M)) * 4;
   mask = M << shift;
   *d = (*d & ~mask) | (((tp->fgColor & M) << shift) & mask);
}

void
_Clear4Bit(const GrafPtr tp, int x, int y)
{
   U8 mask, *d;
   int shift;
#ifdef COMMON
   if (!(d = addr(tp, x, y))) return;
#else
   int wid, hei;

   wid = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   hei = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return;       // the map!
   d = tp->portBits.baseAddr + (y * tp->portBits.rowBytes) + (x / 2);
#endif
   shift = (M - (x & M)) * 4;
   mask = M << shift;
   *d = (*d & ~mask) | (((tp->bkColor & M) << shift) & mask);
}

void
_Flip4Bit(const GrafPtr tp, int x, int y)
{
   U8 *d;
   int shift;
#ifdef COMMON
   if (!(d = addr(tp, x, y))) return;
#else
   int wid, hei;

   wid = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   hei = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return;       // the map!
   d = tp->portBits.baseAddr + (y * tp->portBits.rowBytes) + (x / 2);
#endif
   shift = (M - (x & M)) * 4;
   *d ^= ((tp->fgColor ^ tp->bkColor) & M) << shift;
}

int
_Get4Bit(const GrafPtr tp, int x, int y)
{
   U8 *d;
   int shift;
#ifdef COMMON
   if (!(d = addr(tp, x, y))) return 0;
#else
   int wid, hei;

   wid = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   hei = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return 0;     // the map!
   d = tp->portBits.baseAddr + (y * tp->portBits.rowBytes) + (x / 2);
#endif
   shift = (M - (x & M)) * 4;
   return (*d >> shift) & M;
}

void
_BPP_4(int x, int y)
{
   int rowBytes;

   rowBytes = ((x + 1) / 2);    // Bytes to contain row of pixels.
   screenBits.rowBytes = rowBytes;
   screenBits.bpp = 4;
   if (!screenBits.baseAddr)
      screenBits.baseAddr = calloc(y, rowBytes);
   memset(screenBits.baseAddr, 0xFF, rowBytes * y); // Background all white.
   memset(&screenPort, 0, sizeof screenPort);
   InitPort(&screenPort.port);
   screenPort.port.fgColor = 0;      /* Foreground = black. */
   screenPort.port.bkColor = M;      /* Background = white. */
   screenPort.port._GetPix = _Get4Bit;
   screenPort.port._SetPix = _Set4Bit;
   screenPort.port._ClearPix = _Clear4Bit;
   screenPort.port._FlipPix = _Flip4Bit;
}
