/*
** Initialize Clunkdraw for work with offscreen 8-bit grayscale bitmaps.
*/
#include "../clunkdraw.h"
#include "cdport.h"
#include <stdlib.h> // For calloc
#include <string.h> // For memset

//#define COMMON

#ifdef COMMON
static U8 *
addr(const GrafPtr tp, int x, int y)
{
   int wid, hei;

   wid = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   hei = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return 0;     // the map!
   return tp->portBits.baseAddr + (y * tp->portBits.rowBytes) + x;
}
#endif

// These routines work in 0-based physical coordinates, which
// means the layer above can merge patterns cleanly, etc.
//
// Each of these subroutines clearly has more common code than not,
// but we choose not to factor this out in order to keep the speed up.
// (This naive implementation is bad enough as it is!)
// These are at the very bottom of the call tree,
// runtime inefficiency is magnified.
//
void
_Set8Bit(const GrafPtr tp, int x, int y)
{
   U8 *d;
#ifdef COMMON
   if (!(d = addr(tp, x, y))) return;
#else
   int wid, hei;

   wid = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   hei = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return;       // the map!
   d = tp->portBits.baseAddr + (y * tp->portBits.rowBytes) + x;
#endif
   *d = tp->fgColor;
}

void
_Clear8Bit(const GrafPtr tp, int x, int y)
{
   U8 *d;
#ifdef COMMON
   if (!(d = addr(tp, x, y))) return;
#else
   int wid, hei;

   wid = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   hei = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return;       // the map!
   d = tp->portBits.baseAddr + (y * tp->portBits.rowBytes) + x;
#endif
   *d = tp->bkColor;
}

void
_Flip8Bit(const GrafPtr tp, int x, int y)
{
   U8 *d;
#ifdef COMMON
   if (!(d = addr(tp, x, y))) return;
#else
   int wid, hei;

   wid = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   hei = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return;       // the map!
   d = tp->portBits.baseAddr + (y * tp->portBits.rowBytes) + x;
#endif
   *d ^= tp->fgColor ^ tp->bkColor;
}

int
_Get8Bit(const GrafPtr tp, int x, int y)
{
   U8 *d;
#ifdef COMMON
   if (!(d = addr(tp, x, y))) return 0;
#else
   int wid, hei;

   wid = tp->portBits.bounds.edge.right - tp->portBits.bounds.edge.left;
   hei = tp->portBits.bounds.edge.bottom - tp->portBits.bounds.edge.top;
   if (y < 0 || y >= hei ||             // Stay on
       x < 0 || x >= wid) return 0;     // the map!
   d = tp->portBits.baseAddr + (y * tp->portBits.rowBytes) + x;
#endif
   return *d;
}

void
_BPP_8(int x, int y)
{
   screenBits.rowBytes = x;    // Bytes to contain row of pixels.
   screenBits.bpp = 8;
   if (!screenBits.baseAddr)
      screenBits.baseAddr = calloc(y, x);
   memset(screenBits.baseAddr, 0xFF, x * y); // Background all white.
   memset(&screenPort, 0, sizeof screenPort);
   InitPort(&screenPort.port);
   screenPort.port.fgColor = 0;      /* Foreground = black. */
   screenPort.port.bkColor = 0xFF;   /* Background = white. */
   screenPort.port._GetPix = _Get8Bit;
   screenPort.port._SetPix = _Set8Bit;
   screenPort.port._ClearPix = _Clear8Bit;
   screenPort.port._FlipPix = _Flip8Bit;
}
