// FrameRect draws an outline just inside the specified rectangle, using the
// current grafPort's pen pattern, mode, and size.  The outline is as wide as
// the pen width and as tall as the pen height.  It's drawn with the pnPat,
// according to the transfer mode specified by pnMode.  The pen location is
// not changed by this procedure.
//
// We do this by using four 'stick' rectangles as a frame.  (With small
// rectangles and fat pens, we might only need one.  Because of the XOR
// transfer modes, we don't want to hit any pixel of the frame more than
// once.)
//
// FillRect's faster than PaintRect and they are the same operation when
// using srcCopy mode, so use that when suitable.


#include "../clunkdraw.h"

void
FrameRect(const Rect *r)
{
    Rect tlline, brline;
    Point pnSize;
    PenState penSave;
    GrafPtr tp;

    if (!(tp = thePort) || (tp->pnVis < 0)) return;
    if ((tp->pnSize.h < 1) || (tp->pnSize.v < 1)) return;
    GetPenState(&penSave);
    tlline = *r;  // Make sure frame lines don't squeeze together.  We use two
    brline = *r;  // rectangles, first top and bottom, then left and right;
    pnSize = tp->pnSize;
    tlline.edge.bottom = tlline.edge.top + pnSize.v;
    brline.edge.top = brline.edge.bottom - pnSize.v;
    if (SectRect(r, &tlline, &tlline)) // Stay within the original bounds.
        if (tp->pnMode == srcCopy)
            FillRect(&tlline, &tp->pnPat);
        else
           PaintRect(&tlline);
    if (brline.edge.top < tlline.edge.bottom) // Squashed together?  Don't strike
        brline.edge.top = tlline.edge.bottom; // any pixels twice.
    if (SectRect(r, &brline, &brline)) // Stay within the original bounds.
        if (tp->pnMode == srcCopy)
            FillRect(&brline, &tp->pnPat);
        else
            PaintRect(&brline);
    if (brline.edge.top > tlline.edge.bottom) { // Any verticals necessary?
        SetRect(&tlline, tlline.edge.left, tlline.edge.bottom,
                tlline.edge.left + pnSize.h, brline.edge.top);
        if (SectRect(r, &tlline, &tlline)) // Stay within the original bounds.
            if (tp->pnMode == srcCopy)
                FillRect(&tlline, &tp->pnPat);
            else
                PaintRect(&tlline);
        SetRect(&brline, brline.edge.right - pnSize.h, tlline.edge.top,
                brline.edge.right, tlline.edge.bottom);
        if (SectRect(r, &brline, &brline)) // Stay within the original bounds.
            if (tp->pnMode == srcCopy)
                FillRect(&brline, &tp->pnPat);
            else
                PaintRect(&brline);
    }
    SetPenState(&penSave);
}
