/*
** PtToAngle calculates an integer angle between a line from the center of the
** rectangle to the given point and a line from the center of the rectangle
** pointing straight up (12 o'clock high).  The angle is in degrees from 0 to
** 359, measured clockwise from 12 o'clock, with 90 degrees at 3 o'clock, 180
** at 6 o'clock, and 270 at 9 o'clock.  Other angles are measured relative to
** the rectangle:  If the line to the given point goes through the top right
** corner of the rectangle, the angle returned is 45 degrees, even if the
** rectangle isn't square; if it goes through the bottom left corner, the
** angle is 135 degrees, and so on.
**
** The angle returned might be used as input to one of the procedures that
** manipulates arcs and wedges.
*/

#include "../clunkdraw.h"

void
PtToAngle(rect, point, angle)
const Rect *rect;
const Point *point;
short *angle;
{
    short deltav, deltah, retval;
    Fixed slope, squashby;

    deltav = point->v - ((short) (rect->edge.top + rect->edge.bottom) >> 1);
    deltah = point->h - ((short) (rect->edge.left + rect->edge.right) >> 1);
    if (!deltah)
        retval = (deltav <= 0) ? 0 : 180;
    else {
        slope = FixRatio(deltah, deltav);
        squashby = FixRatio(rect->edge.bottom - rect->edge.top,
                            rect->edge.right - rect->edge.left);
        slope = FixMul(slope, squashby);
        retval = AngleFromSlope(slope);
        if (deltah < 0) {
            retval += 180;
            if (retval == 360)
                retval = 0;
        }
    }
    *angle = retval;
}
