// The BitMapToRegion function converts a given Bitmap record to a
// region.  You would generally use this region later for drawing
// operations.  The region parameter must be a valid region handle
// created with the NewRgn function.  The old region contents are lost.

#include "../clunkdraw.h"
#include "cdrgn.h"
#include <stdlib.h>
#include <alloca.h>
#define TRUE 1
#define FALSE 0

static Boolean
srcBitIsSet(const BitMap *bm, int x, int y)
{
   U8 mask, *src;

   y -= bm->bounds.edge.top;
   if (y < 0 || y >= bm->bounds.edge.bottom) return FALSE;
   x -= bm->bounds.edge.left;
   if (x < 0 || x >= bm->bounds.edge.right) return FALSE;
   src = bm->baseAddr;   // Base address of source image.
   src += y * bm->rowBytes;
   src += x / 8;
   mask = 1 << (7 - (x & 7));
   return !!(*src & mask);
}

int
BitMapToRegion(RgnHandle rgn, const BitMap *sbm)
{
   int x, sx, scanline;
   BitMap bm;
   GrafPtr tp;
   Boolean freeBm;
   Boolean lastBit, nextBit;
   short *outPtr, *outBuf, *lastOut, *anchor;
   Rect lasso;

   if (!rgn || !*rgn) return 1;
   freeBm = FALSE;
   bm = *sbm;
   if ((tp = thePort) && (sbm == &tp->portBits)) {
      if (tp->device == 1) {
         if (!RotateBMapCW(sbm, &bm)) return 1;
         freeBm = TRUE;
      }
   }
   SetRect(&lasso, 32767, bm.bounds.edge.top,
                  -32768, bm.bounds.edge.bottom); // No first coordinates.

   // Calculate the worst-possible-case region size, for working storage.
   x = RectWidth(&bm.bounds) / 2;       // (Max scanline segments.)
   x *= 2 * sizeof(S16);                // (Size of scanline segment.)
   x += 2 * sizeof(S16);                // (Size of Y markers in scanline.)
   x *= RectHeight(&bm.bounds);         // (Number of scanlines.)
   x += sizeof(U16) + sizeof(RegionHdr); // (Region header.)
   x += 2 * sizeof(S16);                // (Dummy starting scanline.)
   outBuf = alloca(x);                  // Probably very overlarge.

   scanline = bm.bounds.edge.top;
   outPtr = &((RgnPtr)outBuf)->rgnData;
   anchor = outPtr;
   lastOut = &((RgnPtr)outBuf)->rgnBBox.edge.bottom; // (Overwrite 1/2 rgnBBox.)
   lastOut[0] = scanline;       // Put out a dummy empty scanline so we can
   lastOut[1] = END_TAG;        // compare 1st output scanline to something.
   while (scanline < bm.bounds.edge.bottom) {
      lastBit = FALSE;
      *outPtr++ = scanline;
      sx = bm.bounds.edge.left;
      for (x = bm.bounds.edge.left; x <= bm.bounds.edge.right + 1; x++) {
         nextBit = srcBitIsSet(&bm, x, scanline);
         if (nextBit != lastBit) {
            lastBit = nextBit;
            if (!lastBit) {     // This WAS a pen-down segment.
               *outPtr++ = sx;
               *outPtr++ = x;
               if (sx < lasso.edge.left)
                  lasso.edge.left = sx;
               if (x > lasso.edge.right)
                  lasso.edge.right = x;
            }
            sx = x;
         }
      }
      *outPtr++ = END_TAG;      // Mark end of X coordinate pairs.
      outPtr = _CheckDups(&anchor, &lastOut, outPtr);
      scanline++;
   }
   _WrapUp(&lasso, outBuf, &bm.bounds, outPtr, rgn); // Puts on end marker too.
   //   free(outBuf);
   if (freeBm)
      free(bm.baseAddr);
   return 0;
}
