/*
** CopyWorkRgn converts a region into a more verbose form that's better suited
** for use by the region merging routines.  It does it by copying it into a
** new place (although this isn't the most efficient way to do it).
**
** The new form is the same as the original, but it's always in scanline form
** (even if it was originally in rectangular form), and there's an extra null
** scanline on the front (which will be filled in later to make both regions
** involved in region math start on the same scanline) and an extra null
** scanline on the end to force expiry of the region at the end during the
** region math routines.
**
** The region's size field _must_ be accurate for this to work correctly.
*/

#include "../clunkdraw.h"
#include "cdrgn.h"

void
_CopyWorkRgn(RgnHandle srcRgn, short *ps) // Expand a rgn to more verbose form.
{
   short i, j;
   long *p, *q;
   Rect *rp;

   if (srcRgn) {               /* Not if it isn't real. */
      i = (*srcRgn)->rgnSize;
      if (i == sizeof(RegionHdr)) { /* Was it rectangular? */
         *ps++ = 0;          /* Filled in by who called us. */
         *ps++ = END_TAG;

         rp = &(*srcRgn)->rgnBBox;
         *ps++ = rp->edge.top; /* Construct a scanline version of a */
         *ps++ = rp->edge.left; /* simple rectangular region that's the */
         *ps++ = rp->edge.right; /* same as the region's bounding box. */
         *ps++ = END_TAG;
         *ps++ = rp->edge.bottom;
         *ps++ = END_TAG;
         *ps = END_TAG;
      } else {               /* For complex regions, we use up the former */
                             /* region header to make room for additions. */
         *ps++ = 0;          /* Filled in by who called us. */
         *ps++ = END_TAG;

         p = (long *) ps;
         q = (long *) &(*srcRgn)->rgnData;
         i -= sizeof(RegionHdr); /* We don't want the region header. */
         j = i / sizeof(long); /* Convert byte count to long count. */
         while (j--)         /* Copy the region's data. */
            *p++ = *q++;
         ps = (short *) p;
         if (i & 3)          /* Odd number of words long? */
            *ps++ = *((short *)q); /* Yes, do the last one then. */
         ps--;               /* Add a dummy (empty) scanline. */
         *ps++ = (*srcRgn)->rgnBBox.edge.bottom;
         *ps++ = END_TAG;
         *ps++ = END_TAG;
      }
   }
}
