#include "../clunkdraw.h"
#include "../rect/cdrect.h"
#include "../port/cdport.h"
#include "cdrgn.h"

// Common region code shared by PaintRgn and FillRgn.
//
void
_DrawRgn(RgnHandle rgn, void (*bitfn)(GrafPtr tp, int x, int y))
{
   Rect br;          // Bounding rectangle, in global coordinates.
   Rect pbr;         // PortBits bounding rectangle.
   int ro, co, vo, x, y;
   Boolean patBit;
   GrafPtr tp;
   const short *rps, *cps, *vps; // Rgn data pointers; Nil if inactive.
   int rStartX, rEndX, cStartX, cEndX, vStartX, vEndX;

   if (!(tp = thePort)) return;
   ro = 0; rEndX = 0; rps = 0;
   co = 0; cEndX = 0; cps = 0;
   vo = 0; vEndX = 0; vps = 0;
   pbr = tp->portBits.bounds;   // We need the logical portBits.bounds for
   if (tp->device == 1)         // clipping, and moving to global coordinates.
      RotRect(&pbr, &pbr);
   if (!rgn || !*rgn) return;
   if (SectRect(&(*rgn)->rgnBBox, &tp->portRect, &br)) {
      if (SectRect(&br, &pbr, &br)) {
         if (_ShieldCursHook)
            (*_ShieldCursHook)(&br, &pbr.corner.topLeft);
         if ((*rgn)->rgnSize != sizeof(RegionHdr))
            rps = &(*rgn)->rgnData;
         if (tp->clipRgn && *tp->clipRgn) {
            if (!SectRect(&br, &(*tp->clipRgn)->rgnBBox, &br)) return;
            if ((*tp->clipRgn)->rgnSize != sizeof(RegionHdr))
               cps = &(*tp->clipRgn)->rgnData;
         }
         if (tp->visRgn && *tp->visRgn) {
            if (!SectRect(&br, &(*tp->visRgn)->rgnBBox, &br)) return;
            if ((*tp->visRgn)->rgnSize != sizeof(RegionHdr))
               vps = &(*tp->visRgn)->rgnData;
         }
         for (y = br.edge.top - pbr.edge.top;
              y < br.edge.bottom - pbr.edge.top; y++) {
            if (rps) {
               if (!ro)            // Find the first appropriate scanline.
                  _FastForward(&rps, y + pbr.edge.top);
               if (rps[ro] != END_TAG && y >= (rps[ro] - pbr.edge.top))
                  rps += ro;       // Time to move the Y?
               ro = 1;             // For re-traversing X list for each Y.
               rStartX = -32767;   // Coords of active X segment of region;
               rEndX   = -32767;   // Minus infinity to start with.
            }
            if (cps) {
               if (!co)            // Find the first appropriate scanline.
                  _FastForward(&cps, y + pbr.edge.top);
               if (cps[co] != END_TAG && y >= (cps[co] - pbr.edge.top))
                  cps += co;       // Time to move the Y?
               co = 1;             // For re-traversing X list for each Y.
               cStartX = -32767;   // Coords of active X segment of clip region;
               cEndX   = -32767;   // Minus infinity to start with.
            }
            if (vps) {
               if (!vo)            // Find the first appropriate scanline.
                  _FastForward(&vps, y + pbr.edge.top);
               if (vps[vo] != END_TAG && y >= (vps[vo] - pbr.edge.top))
                  vps += vo;       // Time to move the Y?
               vo = 1;             // For re-traversing X list for each Y.
               vStartX = -32767;   // Coords of active X segment of vis region;
               vEndX   = -32767;   // Minus infinity to start with.
            }
            for (x = br.edge.left - pbr.edge.left;
                 x < br.edge.right - pbr.edge.left; x++) {
               while (rps && x >= rEndX && rps[ro] != END_TAG) {
                  rStartX = rps[ro++] - pbr.edge.left; // Get the X segment of
                  rEndX = rps[ro++] - pbr.edge.left;   // rgn for current loc.
               }
               while (cps && x >= cEndX && cps[co] != END_TAG) {
                  cStartX = cps[co++] - pbr.edge.left; // Get the X segment of
                  cEndX = cps[co++] - pbr.edge.left;   // clipRgn for current loc
               }
               while (vps && x >= vEndX && vps[vo] != END_TAG) {
                  vStartX = vps[vo++] - pbr.edge.left; // Get the X segment of
                  vEndX = vps[vo++] - pbr.edge.left;   // visRgn for current loc.
               }
               if ((!rps || (x >= rStartX && x < rEndX)) && // Clip to rgns.
                   (!cps || (x >= cStartX && x < cEndX)) &&
                   (!vps || (x >= vStartX && x < vEndX))) {
                  (*bitfn)(tp, x, y);
               }
            }
            while (rps && rps[ro] != END_TAG) // Get to end of X
               ro += 2;            // segments, if not already there.
            ro++;                  // Point to next Y marker.
            while (cps && cps[co] != END_TAG) // Get to end of X
               co += 2;            // segments, if not already there.
            co++;                  // Point to next Y marker.
            while (vps && vps[vo] != END_TAG) // Get to end of X
               vo += 2;            // segments, if not already there.
            vo++;                  // Point to next Y marker.
         }
         if (_ShowCursHook)
            (*_ShowCursHook)();
      }
   }
}
