//
// PtInRgn checks whether the pixel _below_and_to_the_right_ of the given
// coordinate point is within the specified region, and returns TRUE if so or
// FALSE if not.

#include "../clunkdraw.h"
#include "cdrgn.h"
#define TRUE 1
#define FALSE 0

Boolean
PtInRgn(const Point *pt, const RgnHandle rgn)
{
   const short *p, *q;
   int retval;

   retval = FALSE;    // Assume we did not find the point in the region.
   if (rgn && *rgn) {
      if (PtInRect(pt, &(*rgn)->rgnBBox)) { // In bounding box?
         if ((*rgn)->rgnSize != sizeof(RegionHdr)) { // Complex region?
            p = &(*rgn)->rgnData;
            q = p;
            while (*p != END_TAG) { // Find each scanline of the region.
               q = p++;         // Record a pointer to its scanline record.
               while (*p != END_TAG) // Do each X pair of the scanline.
                  p += 2;       // Bump past X coordinate pairs.
               p++;             // Bump to next Y coordinate.
               if ((pt->v >= *q) && (pt->v < *p)) { // Found right line?
                  q++;          // (Bump to X coordinate pairs.)
                  while (*q != END_TAG) { // Yes, check each X pair in it.
                     if ((pt->h >= q[0]) && (pt->h < q[1]))
                        return TRUE;   // Found it.
                     q += 2;
                  }
                  break;        // Had to be here, wasn't -- we failed.
               }
            }
         } else {
            retval = TRUE;    // Simple region, was in it, passed.
         }
      }
   }
   return retval;
}
