/* Miscellaneous internal values used for region manipulation. */

//
// Regarding getting working space for combining regions, in the worst
// case scenario we could need nearly 4x the space of the largest input
// region.  In the X direction you can end up with nearly 2x the segments,
// if B splits every segment of A.  However, in the Y direction a run can
// be split, and require coming back, for 3x the scanlines, and the middle
// scanline is 2x the size that it was.  Something like:
//
//                          (A: DE
// (A: DE) xor (B: FG)  ==>  B: DFGE
//                           C: DE)
//
// (3 wrds) xor (3 wrds) => (11 words)
//
// So, an expansion factor (XF) of 4 is called for.
#define XF 4

#define A_BIT 2                 /* Used for runnning the inner loops. */
#define B_BIT 1
#define AOFFBOFF 0
#define AOFFBON 1
#define AONBOFF 2
#define AONBON  3
#define END_TAG 0x7FFF

short *_CheckDups(short **anchor, short **lastout, short *outptr);
void _WrapUp(Rect *lasso, short *outbuf, Rect *unionrect, short *outptr, RgnHandle dstRgn);
void _FastForward(const short **ps, int y);
void _DrawRgn(RgnHandle rgn, void (*bitfn)(GrafPtr tp, int x, int y));
