/*
** UnionRgn calculates the union of two regions and places the union in a
** third region.  This does _not_ create the destination region; space must
** already have been allocated for it.  The destination region can be one of
** the source regions, if desired.
**
** If both regions are empty, the destination is set to the empty region
** defined by the rectangle (0,0)(0,0).
*/

#include "../clunkdraw.h"
#include "cdrgn.h"
#include <alloca.h>

void
UnionRgn(RgnHandle srcRgnA, RgnHandle srcRgnB, RgnHandle dstRgn)
{
   Rect urect, lasso;
   short scanline, penstate, start;
   short *pA, *pB, *outptr, *curlineA, *curlineB;
   short *lastout, *anchor, holdingA, holdingB;
   short *outbuf, *workA, *workB;
   Rect *rectA, *rectB;
   short Aval;

   if (EmptyRgn(srcRgnA)) {     // One of the sources empty?
      CopyRgn(srcRgnB, dstRgn); // Yes, result is the other one.
      return;
   }
   if (EmptyRgn(srcRgnB)) {
      CopyRgn(srcRgnA, dstRgn);
      return;
   }
   UnionRect(&(*srcRgnA)->rgnBBox, &(*srcRgnB)->rgnBBox, &urect);
   scanline = (*srcRgnA)->rgnSize; // Temp var usage.
   penstate = (*srcRgnB)->rgnSize; // "
   workA = alloca(scanline + 5 * sizeof(short));
   workB = alloca(penstate + 5 * sizeof(short));
   if (penstate > scanline)     // Buffer thrice+ the size of largest input.
      scanline = penstate;
   outbuf = alloca(XF * scanline + 100); // See cdrgn.h
   _CopyWorkRgn(srcRgnA, workA);
   _CopyWorkRgn(srcRgnB, workB);
   pA = &workA[2];              // Skip over room for dummy scanline.
   pB = &workB[2];
   SetRect(&lasso, 32767, 32767, -32768, -32768); // No first coordinates.

   rectA = &(*srcRgnA)->rgnBBox;
   rectB = &(*srcRgnB)->rgnBBox;
   if (rectA->edge.top > rectB->edge.top) { // They don't start in the same
      *--pA = END_TAG;          // place, add a dummy scanline so they do.
      *--pA = rectB->edge.top;
   }
   if (rectB->edge.top > rectA->edge.top) {
      *--pB = END_TAG;
      *--pB = rectA->edge.top;
   }
   curlineA = pA;               // Drop the current input scanline anchors.
   curlineB = pB;

   scanline = urect.edge.top;
   outptr = anchor = &((RgnPtr)(outbuf))->rgnData;
   lastout = &((RgnPtr)outbuf)->rgnBBox.edge.bottom;
   lastout[0] = scanline;       // Put out a dummy empty scanline so we can
   lastout[1] = END_TAG;        // compare 1st output scanline to something.
   while (scanline < urect.edge.bottom) {
      penstate = AOFFBOFF;      // Neither region is 'ON'.
      if (!(holdingA = (*pA == END_TAG))) // More of this region left?
         pA++;                  // Skip to X coordinate pairs.
      if (!(holdingB = (*pB == END_TAG)))
         pB++;
      *outptr++ = scanline;     // Output a Y coordinate.
      while (*pA != END_TAG || *pB != END_TAG) { // Loop across scanline.
         Aval = *pA;
         if ((Aval = *pA) < *pB) { // A's changing.
            if (penstate == AOFFBOFF) // Result off & going on.
               start = Aval;
            if (penstate == AONBOFF) { // Result on & going off.
               if (start < lasso.edge.left)
                  lasso.edge.left = start;
               if (Aval > lasso.edge.right)
                  lasso.edge.right = Aval;
               *outptr++ = start;
               *outptr++ = Aval;
            }
            penstate ^= AONBOFF;
            pA++;
         } else if (*pB < Aval) { // B's changing.
            if (penstate == AOFFBOFF) // Result off & going on.
               start = *pB;
            if (penstate == AOFFBON) { // Result on & going off.
               if (start < lasso.edge.left)
                  lasso.edge.left = start;
               if (*pB > lasso.edge.right)
                  lasso.edge.right = *pB;
               *outptr++ = start;
               *outptr++ = *pB;
            }
            penstate ^= AOFFBON;
            pB++;
         } else {               // A & B both changing.
            if (penstate == AOFFBOFF) // Result off & going on.
               start = Aval;
            if (penstate == AONBON) { // Result on & going off.
               if (start < lasso.edge.left)
                  lasso.edge.left = start;
               if (Aval > lasso.edge.right)
                  lasso.edge.right = Aval;
               *outptr++ = start;
               *outptr++ = Aval;
            }
            penstate ^= AONBON;
            pA++;
            pB++;
         }
      }
      *outptr++ = END_TAG;      // Mark end of X coordinate pairs.
      outptr = _CheckDups(&anchor, &lastout, outptr); // Check for dup line
      if (!holdingA)
         pA++;                  // Bump to the next Y coordinate in lists.
      if (!holdingB)
         pB++;
      if (*pA <= *pB) {         // Current A line expired?
         curlineA = pA;         // Yes, use next A scanline.
         scanline = *pA;
      }
      if (*pB <= *pA) {         // Current B line expired?
         curlineB = pB;         // Yes, use next B scanline.
         scanline = *pB;
      }
      pA = curlineA;
      pB = curlineB;
   }
   _WrapUp(&lasso, outbuf, &urect, outptr, dstRgn);
}
