//
// FrameOval draws an outline just inside the oval that fits into the specified
// rectangle, using the current grafPort's pen pattern, mode, and size.  The
// The outline is as wide as the pen width and as tall as the pen height.  It's
// drawn with the pnPat, according to the pattern transfer mode specified by
// pnMode.  The pen location is not changed by this procedure.
//
// Note:  For fat frames, We don't do FrameOval the same as the rest of them
//        because FrameRgn won't work well with such a complex region (because
//        of the InsetRgn call).  Therefore we make our own framed region
//        before PaintRgn-ing it.

#include "../clunkdraw.h"
#include "cdoval.h"
#include <stdlib.h> // For malloc and free

void
FrameOval(const Rect *r)
{
   short *p1,   // For creating arc points in.
         *p2,   // For creating dummy region in.
         *p3;   // For creating dummy region in.
   int size1, size2;
   RgnPtr dummyRgn1, dummyRgn2;
   RgnHandle dummyRgn;
   Rect scratch;
   GrafPtr tp;

   if (!(tp = thePort)) return;
   if (EmptyRect(r)) return;
   if ((tp->pnSize.h <= 2) || (tp->pnSize.v <= 2))
      _DrawOval(r, FrameRgn, 0);
   else {
      dummyRgn = NewRgn();
      size1 = (1 + r->edge.bottom - r->edge.top) * sizeof(short);
      p1 = malloc(size1);      // Number of arc points (one per scan).
      _BuildArc(r, p1);
      size2 = (size1 * 4) + sizeof(RegionHdr) + sizeof(short); // Size of
      p2 = malloc(size2);      // dummy regions.
      _BuildORgn(r, p1, p2);
      dummyRgn1 = (RgnPtr) p2;
      p3 = 0;
      if (tp->pnVis >= 0) {
         p3 = malloc(size2);
         scratch = *r;
         InsetRect(&scratch, tp->pnSize.h, tp->pnSize.v);
         _BuildArc(&scratch, p1);
         _BuildORgn(&scratch, p1, p3);
         dummyRgn2 = (RgnPtr) p3;
         DiffRgn(&dummyRgn1, &dummyRgn2, dummyRgn);
         PaintRgn(dummyRgn);
      }
      if (p3) free(p3);
      free(p2);
      free(p1);
      DisposeRgn(dummyRgn);
   }
}
