/*
** DrawRRect draws rounded rects with the specified drawing verb (via regions).
*/

#include "../clunkdraw.h"
#include "../rgn/cdrgn.h"
#include "cdrrect.h"
#include <alloca.h>

/*
** BuildRRgn builds a fake region data structure that describes a rounded-
** corner rectangle.  It's used by the xxRoundRect drawing routines.
*/

static void
BuildRRgn(rect, ovlrect, arc, p)
const Rect *rect;            /* What to make a rounded region from. */
const Rect *ovlrect;         /* The oval's rectangle. */
register short *arc;            /* The arc control points. */
register short *p;              /* Where to make it. */
{
    register short i, j, leftx, rightx;
    register short *lastpt;
    short *save1, *save2, nullpt;

    save1 = p;
    p++;                        /* Skip over the size field. */
    *((Rect *) p) = *rect;      /* Dummy region's bounding box. */
    p += sizeof *rect / sizeof *p;
    save2 = p;
    j = ovlrect->edge.bottom / 2;
    i = (rect->edge.bottom - rect->edge.top)/2;
    if (j > i)                  /* Make sure oval's not bigger than the */
        j = i;                  /* rectangle we're trimming. */
    nullpt = END_TAG;           /* So we get the first region segment. */
    lastpt = &nullpt;
    leftx = ovlrect->edge.right / 2;
    rightx = rect->edge.right - leftx;
    leftx += rect->edge.left;
    for (i=0; i<j; i++) {       /* Top half of rect. */
        if (*arc != *lastpt) {  /* Don't output unnecessary region segments. */
            lastpt = arc;
            *p++ = rect->edge.top + i;
            *p++ = leftx - *arc;
            *p++ = rightx + *arc++;
            *p++ = END_TAG;
        } else
            arc++;
    }

    for (i=0; i<j; i++) {       /* Bottom half of rect. */
        if (*--arc != *lastpt) {
            lastpt = arc;
            *p++ = rect->edge.bottom - j + i;
            *p++ = leftx - *arc;
            *p++ = rightx + *arc;
            *p++ = END_TAG;
        }
    }
    if (save2 == p)             /* No actual data? */
        *save1 = sizeof(RegionHdr); /* Re-size it then. */
    else {
        *p++ = END_TAG;         /* End-of-region marker. */
        *save1 = (char *) p - (char *) save1; /* Plug in real size. */
    }
}

void
_DrawRRect(rect, ovlrect, verb, arg)
const Rect *rect;
const Rect *ovlrect;
void (*verb)();
void * const arg;
{
    short *p1;                  /* For creating arc points in. */
    short *p2;                  /* For creating dummy region in. */
    long size1, size2;
    RgnPtr dummyRgn;

    if (EmptyRect(rect))
        return;
    size1 = (1 + ovlrect->edge.bottom) * sizeof(short); /* Number of arc */
    p1 = alloca(size1);         /* points (maximum one per scanline). */
    _BuildArc(ovlrect, p1);
    size2 = (size1 * 4) + sizeof(RegionHdr) + sizeof(short); /* Size of */
    p2 = alloca(size2);         /* dummy region. */
#ifdef DEBUG
    DumpRect(rect, "Rectangle ");
    DumpRect(ovlrect, "Oval ");
    {short i, *q;
     fprintf(stderr, "Arc data: ");
     for (q=p1, i=0; i<(size1/2); i++)
         fprintf(stderr, "%d, ", *q++);
     fprintf(stderr, "\n");
}
#endif
    BuildRRgn(rect, ovlrect, p1, p2);
    dummyRgn = (RgnPtr) p2;
#ifdef DEBUG
    DumpRgn("Rounded Rect ", &dummyRgn);
#endif
    (*verb)(&dummyRgn, arg);
}
