/*
**
**  A quick hack of the X-windows program 'maze' for ISC's windowman
**
**     ===== H. Marc Lewis, Spokane, March-1992 =====
**
**  V1.2 5/17/93 DH
*/

#include "../clunkdraw.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

#define SPEED           2
#define PAUSE           3
#define OPTIONS         1

#define MIN_MAZE_SIZE   3
#define MAX_MAZE_SIZE_X 205
#define MAX_MAZE_SIZE_Y 205

#define MOVE_LIST_SIZE  (MAX_MAZE_SIZE_X * MAX_MAZE_SIZE_Y)

#define WALL_TOP        0x8000
#define WALL_RIGHT      0x4000
#define WALL_BOTTOM     0x2000
#define WALL_LEFT       0x1000

#define DOOR_IN_TOP     0x800
#define DOOR_IN_RIGHT   0x400
#define DOOR_IN_BOTTOM  0x200
#define DOOR_IN_LEFT    0x100
#define DOOR_IN_ANY     0xF00

#define DOOR_OUT_TOP    0x80
#define DOOR_OUT_RIGHT  0x40
#define DOOR_OUT_BOTTOM 0x20
#define DOOR_OUT_LEFT   0x10

#define START_SQUARE    0x2
#define END_SQUARE      0x1

#define SQ_SIZE_X       10
#define SQ_SIZE_Y       10

#define NUM_RANDOM      100

#define BORDERWIDTH     2
#define border_x        (5)
#define border_y        (5)
#define MIN_W           200
#define MIN_H           200

#define get_random(x)   (rand() % (x))

int wid;
short flashing=1;

static int logo_x, logo_y;

static unsigned short maze[MAX_MAZE_SIZE_X][MAX_MAZE_SIZE_Y];

static struct {
    unsigned char x;
    unsigned char y;
    unsigned char dir;
} move_list[MOVE_LIST_SIZE],
  save_path[MOVE_LIST_SIZE],
  path[MOVE_LIST_SIZE];

static int maze_size_x, maze_size_y;
static int sqnum, cur_sq_x, cur_sq_y, path_length;
static int start_x, start_y, start_dir, end_x, end_y, end_dir;
static short keydelay = 30;
static short pausec = 3;
static short last_speed;
static short show_backtrace = 1;
static short Width;
static short Height;

static int pulldnevent(EventRecord *ep);
static int windowevent(EventRecord *ep);
void draw_solid_square(int i, int j,
                       int dir, const Pattern *pat);
void draw_wall(int i, int j, int dir);
void enter_square(int n);

int x = 0, y = 0, restart = 0, stop = 1, state = 1;

int
check_for_key()
{
   return 0;
}

void
initialize_maze()         /* draw the surrounding wall and start/end squares */
{
    int i, j, wall;

    /* initialize all squares */
    for (i=0; i<maze_size_x; i++) {
        for (j=0; j<maze_size_y; j++) {
            maze[i][j] = 0;
        }
    }

    /* top wall */
    for (i=0; i<maze_size_x; i++) {
        maze[i][0] |= WALL_TOP;
    }

    /* right wall */
    for (j=0; j<maze_size_y; j++) {
        maze[maze_size_x-1][j] |= WALL_RIGHT;
    }

    /* bottom wall */
    for (i=0; i<maze_size_x; i++) {
        maze[i][maze_size_y-1] |= WALL_BOTTOM;
    }

    /* left wall */
    for (j=0; j<maze_size_y; j++) {
        maze[0][j] |= WALL_LEFT;
    }

    /* set start square */
    wall = get_random(4);
    switch (wall) {
      case 0:
        i = get_random(maze_size_x);
        j = 0;
        break;

      case 1:
        i = maze_size_x - 1;
        j = get_random(maze_size_y);
        break;

      case 2:
        i = get_random(maze_size_x);
        j = maze_size_y - 1;
        break;

      case 3:
        i = 0;
        j = get_random(maze_size_y);
        break;
    }
    maze[i][j] |= START_SQUARE;
    maze[i][j] |= (DOOR_IN_TOP >> wall);
    maze[i][j] &= ~(WALL_TOP >> wall);
    cur_sq_x = i;
    cur_sq_y = j;
    start_x = i;
    start_y = j;
    start_dir = wall;
    sqnum = 0;

    /* set end square */
    wall = (wall + 2) % 4;
    switch (wall) {
      case 0:
        i = get_random(maze_size_x);
        j = 0;
        break;
      case 1:
        i = maze_size_x - 1;
        j = get_random(maze_size_y);
        break;
      case 2:
        i = get_random(maze_size_x);
        j = maze_size_y - 1;
        break;
      case 3:
        i = 0;
        j = get_random(maze_size_y);
        break;
    }
    maze[i][j] |= END_SQUARE;
    maze[i][j] |= (DOOR_OUT_TOP >> wall);
    maze[i][j] &= ~(WALL_TOP >> wall);
    end_x = i;
    end_y = j;
    end_dir = wall;
    check_for_key();
}

void
draw_maze_border()           /* draw the maze outline */
{
    short i, j;

    PenNormal();
    for (i = 0; i < maze_size_x; i++) {
        if (maze[i][0] & WALL_TOP) {
            MoveTo(border_x + SQ_SIZE_X * i, border_y);
            Line(SQ_SIZE_X, 0);
        }
        if ((maze[i][maze_size_y - 1] & WALL_BOTTOM)) {
            MoveTo(border_x + SQ_SIZE_X * i,
                   border_y + SQ_SIZE_Y * (maze_size_y));
            Line(SQ_SIZE_X, 0);
        }
    }
    check_for_key();
    for (j = 0; j < maze_size_y; j++) {
        if (maze[maze_size_x - 1][j] & WALL_RIGHT) {
            MoveTo(border_x + SQ_SIZE_X * maze_size_x,
                   border_y + SQ_SIZE_Y * j);
            Line(0, SQ_SIZE_Y);
        }
        if (maze[0][j] & WALL_LEFT) {
            MoveTo(border_x, border_y + SQ_SIZE_Y * j);
            Line(0, SQ_SIZE_Y);
        }
    }
    check_for_key();
    draw_solid_square(start_x, start_y, start_dir, &black);
    draw_solid_square(end_x, end_y, end_dir, &black);
    if (flashing) {
        for (i=0; i<10; ++i) {
            draw_solid_square(start_x, start_y, start_dir,
                              i & 1 ? &black : &white);
            draw_solid_square(end_x, end_y, end_dir,
                              i & 1 ? &black : &white);
            usleep(100*1000);
        }
        check_for_key();
    }
}

void
set_maze_sizes(int width, int height)
{
    if ((maze_size_x = width / SQ_SIZE_X) > MAX_MAZE_SIZE_X)
        exit(4);                /* Screen too big. */
    if ((maze_size_y = height / SQ_SIZE_Y) > MAX_MAZE_SIZE_Y)
        exit(4);                /* Screen too big. */
}

int
backup()                                                  /* back up a move */
{
    sqnum--;
    cur_sq_x = move_list[sqnum].x;
    cur_sq_y = move_list[sqnum].y;
    return sqnum;
}

int
choose_door()
{                               /* pick a new path */
    unsigned short *p, current;
    int num_candidates;
    int candidates[3];

    num_candidates = 0;

    p = &maze[cur_sq_x][cur_sq_y];
    current = *p;
  topwall:
    /* top wall */
    if (current & DOOR_IN_TOP)
        goto rightwall;
    if (current & DOOR_OUT_TOP)
        goto rightwall;
    if (current & WALL_TOP)
        goto rightwall;
    if (maze[cur_sq_x][cur_sq_y - 1] & DOOR_IN_ANY) {
        *p = (current |= WALL_TOP);
        maze[cur_sq_x][cur_sq_y - 1] |= WALL_BOTTOM;
        draw_wall(cur_sq_x, cur_sq_y, 0);
        goto rightwall;
    }
    candidates[num_candidates++] = 0;

  rightwall:
    /* right wall */
    if (current & DOOR_IN_RIGHT)
        goto bottomwall;
    if (current & DOOR_OUT_RIGHT)
        goto bottomwall;
    if (current & WALL_RIGHT)
        goto bottomwall;
    if (maze[cur_sq_x + 1][cur_sq_y] & DOOR_IN_ANY) {
        *p = (current |= WALL_RIGHT);
        maze[cur_sq_x + 1][cur_sq_y] |= WALL_LEFT;
        draw_wall(cur_sq_x, cur_sq_y, 1);
        goto bottomwall;
    }
    candidates[num_candidates++] = 1;

  bottomwall:
    /* bottom wall */
    if (current & DOOR_IN_BOTTOM)
        goto leftwall;
    if (current & DOOR_OUT_BOTTOM)
        goto leftwall;
    if (current & WALL_BOTTOM)
        goto leftwall;
    if (maze[cur_sq_x][cur_sq_y + 1] & DOOR_IN_ANY) {
        *p = (current |= WALL_BOTTOM);
        maze[cur_sq_x][cur_sq_y + 1] |= WALL_TOP;
        draw_wall(cur_sq_x, cur_sq_y, 2);
        goto leftwall;
    }
    candidates[num_candidates++] = 2;

  leftwall:
    /* left wall */
    if (current & DOOR_IN_LEFT)
        goto donewall;
    if (current & DOOR_OUT_LEFT)
        goto donewall;
    if (current & WALL_LEFT)
        goto donewall;
    if (maze[cur_sq_x - 1][cur_sq_y] & DOOR_IN_ANY) {
        *p = (current |= WALL_LEFT);
        maze[cur_sq_x - 1][cur_sq_y] |= WALL_RIGHT;
        draw_wall(cur_sq_x, cur_sq_y, 3);
        goto donewall;
    }
    candidates[num_candidates++] = 3;

  donewall:
    if (num_candidates == 0)
        return -1;
    if (num_candidates == 1)
        return candidates[0];
    return candidates[get_random(num_candidates)];
}

void
create_maze()     /* create a maze layout given the initialized maze */
{
    int i, newdoor;

    for (;;) {
        move_list[sqnum].x = cur_sq_x;
        move_list[sqnum].y = cur_sq_y;
        move_list[sqnum].dir = newdoor;
        while ((newdoor = choose_door()) == -1) { /* pick a door */
            if (backup() == -1) { /* no more doors ... backup */
                return; /* done ... return */
            }
        }

        /* mark the out door */
        maze[cur_sq_x][cur_sq_y] |= (DOOR_OUT_TOP >> newdoor);

        switch (newdoor) {
          case 0:
            cur_sq_y--;
            break;
          case 1:
            cur_sq_x++;
            break;
          case 2:
            cur_sq_y++;
            break;
          case 3:
            cur_sq_x--;
            break;
        }
        sqnum++;

        /* mark the in door */
        maze[cur_sq_x][cur_sq_y] |= (DOOR_IN_TOP >> ((newdoor+2)%4));

        /* if end square set path length and save path */
        if (maze[cur_sq_x][cur_sq_y] & END_SQUARE) {
            path_length = sqnum;
            for (i=0; i<path_length; i++) {
                save_path[i].x = move_list[i].x;
                save_path[i].y = move_list[i].y;
                save_path[i].dir = move_list[i].dir;
            }
        }
    }
}


/* Draw a single wall. */
void
draw_wall(int i, int j, int dir)
{
    switch (dir) {
      case 0:
        MoveTo(border_x + SQ_SIZE_X * i, border_y + SQ_SIZE_Y * j);
        Line(SQ_SIZE_X, 0);
        break;

      case 1:
        MoveTo(border_x + SQ_SIZE_X * (i + 1), border_y + SQ_SIZE_Y * j);
        Line(0, SQ_SIZE_Y);
        break;

      case 2:
        MoveTo(border_x + SQ_SIZE_X * i, border_y + SQ_SIZE_Y * (j + 1));
        Line(SQ_SIZE_X, 0);
        break;

      case 3:
        MoveTo(border_x + SQ_SIZE_X * i, border_y + SQ_SIZE_Y * j);
        Line(0, SQ_SIZE_Y);
        break;
    }
}

void
solve_maze()        /* solve it with graphical feedback */
{
    short i;
    EventRecord *ep;
    int ch;
    short every = 0;

    /* plug up the surrounding wall */
    maze[start_x][start_y] |= (WALL_TOP >> start_dir);
    maze[end_x][end_y] |= (WALL_TOP >> end_dir);

    /* initialize search path */
    i = 0;
    path[i].x = end_x;
    path[i].y = end_y;
    path[i].dir = -1;

    /* do it */
    for (;;) {
        if (++path[i].dir >= 4) {
            i--;
            draw_solid_square((int)(path[i].x), (int)(path[i].y),
                              (int)(path[i].dir),
                              (show_backtrace) ? &gray : &white);
        }
        else if (!(maze[path[i].x][path[i].y] & (WALL_TOP >> path[i].dir))
                 && ((i == 0) || ((path[i].dir !=
                                (int)(path[i-1].dir+2)%4)))) {
            enter_square(i);
            i++;
            if (maze[path[i].x][path[i].y] & START_SQUARE) {
                for (i=0; i<pausec; ++i) {
                    usleep(500*1000);
                    if (check_for_key())
                        return;
                }
                return;
            }
        }
        if (++every >= keydelay) {
            every = 0;
            if (check_for_key())
                return;
        }
    }
}


/* move into a neighboring square */
void
enter_square(int n)
{
    draw_solid_square((int)path[n].x, (int)path[n].y,
                      (int)path[n].dir, &black);

    path[n+1].dir = -1;
    switch (path[n].dir) {
      case 0:
        path[n+1].x = path[n].x;
        path[n+1].y = path[n].y - 1;
        break;
      case 1:
        path[n+1].x = path[n].x + 1;
        path[n+1].y = path[n].y;
        break;
      case 2:
        path[n+1].x = path[n].x;
        path[n+1].y = path[n].y + 1;
        break;
      case 3:
        path[n+1].x = path[n].x - 1;
        path[n+1].y = path[n].y;
        break;
    }
}



void
draw_solid_square(int i, int j, int dir, const Pattern *pat) /* draw a solid square
                                                          in a square */
{
    Rect sq;

    switch (dir) {
      case 0:
        SetRect(&sq,
                border_x + 3 + SQ_SIZE_X * i,
                border_y - 3 + SQ_SIZE_Y * j,
                border_x + 3 + SQ_SIZE_X * i + SQ_SIZE_X - 6,
                border_y - 3 + SQ_SIZE_Y * j + SQ_SIZE_Y);
        break;
      case 1:
        SetRect(&sq,
                border_x + 3 + SQ_SIZE_X * i,
                border_y + 3 + SQ_SIZE_Y * j,
                border_x + 3 + SQ_SIZE_X * i + SQ_SIZE_X,
                border_y + 3 + SQ_SIZE_Y * j + SQ_SIZE_Y - 6);
        break;
      case 2:
        SetRect(&sq,
                border_x + 3 + SQ_SIZE_X * i,
                border_y + 3 + SQ_SIZE_Y * j,
                border_x + 3 + SQ_SIZE_X * i + SQ_SIZE_X - 6,
                border_y + 3 + SQ_SIZE_Y * j + SQ_SIZE_Y);
        break;
      case 3:
        SetRect(&sq,
                border_x - 3 + SQ_SIZE_X * i,
                border_y + 3 + SQ_SIZE_Y * j,
                border_x - 3 + SQ_SIZE_X * i + SQ_SIZE_X,
                border_y + 3 + SQ_SIZE_Y * j + SQ_SIZE_Y - 6);
        break;
    }
    FillRect(&sq, pat);
}

int
main(int argc, char *argv[])
{
    Rect win, erase;

    /* Initialize graphics toolbox */
    InitGraf2();
    win = thePort->portRect;
    Width = win.edge.right - 10;
    Height = win.edge.bottom - 10;
    erase = win;
    OffsetRect(&erase, -erase.edge.left, -erase.edge.top);
    srand(getpid());
    set_maze_sizes(Width, Height);
    for (;;) {
        initialize_maze();
        create_maze();
        draw_maze_border();
        solve_maze();
        EraseRect(&erase);
    }
}
