/*
** Simplest/smallest canonical test program.
**
** We play some games to keep the linker from bringing in anything
** that's not absolutely necessary to the task.  On a RPi, this entire
** program is just over 9,700 bytes of code.
**
** If we don't eliminate the extra unnecessary code by providing these
** dummies, the program weighs in at over 28,000 bytes.  If we don't
** eliminate the unnecessary fonts, the full naive Monty, it's more like
** 1,102,000 bytes!  (All the fonts together are huge.)
**
** So you can get Clunkdraw-using programs onto things like an Arduino,
** if you keep them modest, but you might have to work a little bit at it.
**
** Examining the link map is crucial to make sure you're not getting
** anything large that you're not intending to use.
*/

#include "../clunkdraw.h"
#include "../fonts/cdfont.h"

// We're not using any algorithmic text styling, so provide a dummy subroutine.
void
_DrawIt() /* Saves 2.7 kB */
{
}

// We're not using any Macintosh fonts, so eliminate that code.
void
_DrawMacChar() /* Saves 1.5kB */
{
}

// Plus there are ripple effects, other subroutines these called that
// aren't included either.  It adds up.

// We're only using the one (default) font, so eliminate all others.
const FontTableEntry EmbeddedFonts[2] = {
    {0, 12, &LaCenter12_Font},
    {0, 0, 0} // End marker
};

int
main(int argc, char *argv[])
{
   InitGraf(96, 13, BPP(1));
   _MoveTo(4, 10); // Avoid line-related things like polygons, regions.
   DrawString("Hello, world!");
//   SaveToPBMFile(&thePort->portBits, "junk.pbm");
   EndGraf();

   return 0;
}
