#include "../clunkdraw.h"

/*
** Draw head and shoulders outline on overlay.  We use four ovals,
** judiciously located and/or trimmed since we don't have clipping
** regions in this stripped version of Clunkdraw.  (Only the head
** remains a full oval.)
**
** We don't need to do the punch-outs, because the ovals don't
** actually overlap enough to need cleaning up that way.
*/
void
headshape()
{
   Rect r;
   PenSize(4, 4);               /* We'll use a fairly heavy outline. */

   SetRect(&r, 0, 0, 20, 40);   /* Two ovals for neck. */
   OffsetRect(&r, 120, 184);
   FrameOval(&r);               /* Left side of the neck. */
   OffsetRect(&r, 60, 0);
   FrameOval(&r);               /* Right side of the neck. */
   r.edge.left += 10;
   EraseRect(&r);               /* Remove right half of right neck oval. */
   OffsetRect(&r, -70, 0);
   EraseRect(&r);               /* Remove left half of left neck oval. */

   SetRect(&r, 0, 0, 200, 40);  /* Wide and short (oval) for shoulders. */
   OffsetRect(&r, 60, 220);     /* (Bottom half lies off the screen, unshown.) */
// EraseOval(&r);               /* Punch out the shoulders. */
   FrameOval(&r);               /* Frame the shoulders. */

   SetRect(&r, 0, 0, 2, 3);     /* Human head aspect ratio, 2 wide 3 high. */
   r.edge.right *= 60;          /* Scale it up. */
   r.edge.bottom *= 60;
   OffsetRect(&r, 100, 20);
   r.edge.top += 10;            /* Just a little less egg-like, I think. */
// EraseOval(&r);               /* Punch out the head. */
   FrameOval(&r);               /* Frame the head. */

   PenNormal();
}

int
main(int argc, char *argv[])
{
   int ii;

   InitGraf2();
   ForeColor(0xFA80); // Orange in RGB565
//   for (ii=0; ii<1000; ii++)
      headshape();
   SaveToPPMFile(&thePort->portBits, "junk.ppm");
   EndGraf2();

   return 0;
}
