#include "../clunkdraw.h"
#include "cdtext.h"
#include <stdlib.h> // for malloc, etc.
#define TRUE 1
#define FALSE 0

#ifdef DEBUG
#  include <stdio.h>
static void
BinStr(char *s, int b, int bits)
{
   int ii, mask;

   mask = 1 << (bits - 1);
   for (; mask; mask>>=1)
      *s++ = (b & mask) ? '1' : '0';
   *s++ = 0;
}

// Dump out the styled bitmap, visually.
// (Best if it's not too big!)
// For debugging.  (There are TWO!)
static void
dumpBmap(BitMap *bits, int uline, int sty, int strwidth, int weight, char *lbl)
{
   int x, y, z, u, s;
   U8 *p;
   char bitz[10];

   puts(lbl);
   fputs("       ", stdout);         // Column numbers.
   for (s = 1; s < bits->rowBytes * 8 + 1; s++)
      if (!(s % 10))
         printf("%d", s / 10);
      else
         putchar(' ');
   putchar('\n');
   fputs("       ", stdout);
   for (s = 1; s < bits->rowBytes * 8 + 1; s++)
      printf("%d", s % 10);
   putchar('\n');
   p = bits->baseAddr;
   u = uline;
   s = ((bits->bounds.edge.top + uline) / 2) + 1;
   if ((outlineText|shadowText) & sty)
      s++;
   if (shadowText & sty)
      s++;
   for (y = bits->bounds.edge.bottom; y; y--) {
      printf("  %02X: ", y);
      if (u <= 0 && u > -weight)
         putchar('_');  // Line markers.
      else if (s <= 0 && s > -weight)
         putchar('-');
      else
         putchar(' ');
      for (x = 0; x < bits->rowBytes; x++) {
         BinStr(bitz, *p, 8);
         for (z = 0; z < 8; z++) {
            if (bitz[z] == '1')
               bitz[z] = '*';
            else
               bitz[z] = '.';
            if (strwidth == (x * 8 + z + 1))
               if (bitz[z] == '.')
                  bitz[z] = '|';
         }
         if (bitz[7] == '.' && x != (bits->rowBytes - 1))
            bitz[7] = ' '; // So we can see byte boundaries.
         fputs(bitz, stdout);
         p++;
      }
      if (u <= 0 && u > -weight)
         putchar('_');  // Line markers.
      else if (s <= 0 && s > -weight)
         putchar('-');
      putchar('\n');
      u--;
      s--;
   }
}
#endif

//
// _AddStyle applies the given text styles to an offscreen bitmap that
// already has the plain text imaged into it.  The bitmap is assumed to
// have enough room in it to hold the stylized text.  The order in which
// we apply style effects is somewhat important.
//
void
_AddStyle(BitMap *bits, int uline, int sty, int strwidth)
{
   int ii, jj, kk, ll, width, weight, size;
   U8 *p, *p1, *p2, *q, *q1;
   U8 work, carry, carryin;
   int styl;                    // For style labeling when debugging.

   weight = uline / 16;         // Stroke weight for underlines and strikeouts.
   if (weight <= 0)
      weight = 1;
   styl = 0;
   if (shadowText & sty)
      sty |= outlineText;       // Shadow is an enhancement of outline.
#ifdef DEBUG // Enable for debug printout.
   styl = sty;
   fputs("Applying", stdout);
   if (boldText & styl) fputs(" Bold", stdout);
   if (fixedBoldText & styl) fputs(" FixedBold", stdout);
   if (strikeoutText & styl) fputs(" Strikeout", stdout);
   if (underlineText & styl) fputs(" Underline", stdout);
   if (italicText & styl) fputs(" Italic", stdout);
   if (outlineText & styl) fputs(" Outline", stdout);
   if (shadowText & styl) fputs(" Shadow", stdout);
   if (ghostText & styl) fputs(" Ghosting", stdout);
   puts(" styles...");
   dumpBmap(bits, uline, sty, strwidth, weight, "Before:");
#endif
   if ((boldText|fixedBoldText) & sty) { // Algorithmic bolding?
      p = bits->baseAddr;       // Yes, smear the thing one
      q = p;                    // pixel to the right.
      jj = bits->bounds.edge.bottom; // Number of lines in bitmap to do.
      while (jj > 0) {
         ii = bits->rowBytes;  // Count of bytes across bitmap.
         carry = 0;            // Preload the carry.
         while (ii-- > 0) {
            work = *p;
            kk = work & 1;   // For making carry.
            work |= (work >> 1) | carry;
            *p++ = work;
            carry = kk << 7;
         }
         jj--;
      }
      styl &= ~(boldText|fixedBoldText);
#ifdef DEBUG
      if (styl) dumpBmap(bits, uline, sty, strwidth, weight, "bolded:");
#endif
   }

   if (strikeoutText & sty) {   // Algorithmic strikeout?
                                // Line in bitmap to put strikeout on.
      ii = ((bits->bounds.edge.top + uline) / 2) + 1;
      if (outlineText & sty)
         ii++;
      if (ii > (bits->bounds.edge.bottom - 1)) // (Make sure it's there.)
         ii = ((bits->bounds.edge.bottom + bits->bounds.edge.top) / 2) + 1;
      ll = ii;                  // For reloading ii on heavyweight lines.
      for (kk = 0; kk < weight; kk++) {
         ii = ll + kk;
         p = bits->baseAddr + (bits->rowBytes * ii);
         carry = 0;
         if (outlineText & sty) // We trim off the first pixel for outlined
            carry = 0x80;       // so that it will get an end cap as well.
         ii = strwidth / 8;     // Count of bytes across bitmap.
         while (ii > 0) {
            *p++ |= ~carry;
            carry = 0;
            ii--;
         }
         if (shadowText & sty)  // Do the rightmost end cap.
            ii = (strwidth - 3) & 7;
         else if (outlineText & sty)
            ii = (strwidth - 2) & 7;
         else
            ii = (strwidth - 1) & 7;
         if (ii)
            *p = (*p | ~carry) & -(1 << (8 - ii));
      }
      styl &= ~strikeoutText;
#ifdef DEBUG
      if (styl) dumpBmap(bits, uline, sty, strwidth, weight, "struck out:");
#endif
   }

// Because line weights can be heavier than one line, we have to not modify
// the original source (because of the descender avoidance logic) until
// we've calculated all the underline bits->  Only then apply them.
// (Underlines, unlike strikeouts, are not intended to reduce readability,
// hence the descender avoidance logic.)
   if (underlineText & sty) {   // Algorithmic underline?
      q1 = 0;
      if (weight > 1) {         // Make an off-off-screen buffer.
         ii = bits->rowBytes * weight;
         q1 = calloc(1, ii);
      }
      for (ll = 0; ll < weight; ll++) {
         ii = uline + ll;      // Line in bitmap to put underline on.
         if (ii > (bits->bounds.edge.bottom - 1)) // (Make sure it's
            ii = bits->bounds.edge.bottom - 1; // really there.)
         p = bits->baseAddr + (bits->rowBytes * ii);
         if (!q1)               // Apply thin lines
            q = p;              // in-place,
         else                   // and heavy lines
            q = q1 + (bits->rowBytes * ll); // in the off-off-screen buffer.
         ii = bits->rowBytes;   // Count of bytes across bitmap.
         p1 = &p[-ii];          // When we find something in
         p2 = p1;               // our underline row, we want
         if ((uline + ll) < (bits->bounds.edge.bottom - 1))
            p2 = &p[ii];        // to look in adjacent rows (if possible)
         carry = 0;             // to make a composite hole-knocker.
         if (outlineText & sty) // We trim off the first pixel for outlined
            carry = 1;          // so that it will get an end cap as well.
         ii = strwidth / 8;     // Count of bytes across bitmap.
         while (ii > 0) {
            work = *p++ | *p1++ | *p2++; // Check rows above & below us
            kk = work & 1;      // so's not to touch any bit of text with
            if (work)           // the underline.
               work |= (work << 1) | (work >> 1);
            if (ii)             // There's a word to the right of us, so handle
               if ((*p | *p1 | *p2) & 0x80) // carry from right.
                  work |= 1;
            if (carry)
               work |= 0x80;
            carry = kk;
            *q++ |= ~work;      // Slap up an underline (with holes
            ii--;               // around descenders).
         }
         if (shadowText & sty)  // Do the rightmost end cap.
            ii = (strwidth - 3) & 7;
         else if (outlineText & sty)
            ii = (strwidth - 2) & 7;
         else
            ii = (strwidth - 1) & 7;
         if (ii)
            *q = (*q | ~work) & -(1 << (8 - ii));
      }
      if (q1) {                 // Heavyweight line?  If so, copy it on.
         ii = uline;            // Line in bitmap to put underline on.
         if (ii > (bits->bounds.edge.bottom - 1)) // (Make sure it's
            ii = bits->bounds.edge.bottom - 1; // really there.)
         ii *= bits->rowBytes;
         p = bits->baseAddr + ii;
         ii = weight * bits->rowBytes;
         q = q1;
         while (ii-- > 0)       // OR the line into the bitmap;
            *p++ |= *q++;
         free(q1);              // Now get rid of off-off-screen bits->
      }
      styl &= ~underlineText;
#ifdef DEBUG
      if (styl) dumpBmap(bits, uline, sty, strwidth, weight, "underlined:");
#endif
   }

   if (italicText & sty) {      // Algorithmic italics?
                                // First smear things above the
                                // baseline to the right.
      jj = uline - 3;           // Number of lines in bitmap to do.
      q = bits->baseAddr;
      while (jj > 0) {
         p = q;
         kk = 0;
         while (kk++ < jj) {
            ii = bits->rowBytes; // Count of bytes across bitmap.
            carry = 0;           // Preload carry-in.
            while (ii-- > 0) {
               work = *p;
               *p++ = (work >> 1) | carry; // Shift over 1 bit.
               if (work & 1)
                  carry = 0x80;
               else
                  carry = 0;
            }
         }
         jj -= 2;               // Skip two the next time.
      }
                                // Next smear things below the
      jj = uline - 1;           // baseline to the left.
      while (jj < bits->bounds.edge.bottom) {
         p = bits->baseAddr + MulU(jj, bits->rowBytes);
         q = p;
         kk = jj;
         while (kk++ < bits->bounds.edge.bottom) {
            ii = bits->rowBytes; // Count of bytes across bitmap.
            while (ii > 0) {
               work = *q++;
               if (*q & 0x80)   // Smear sideways.
                  carry = 1;
               else
                  carry = 0;
               if (!ii)
                  carry = 0;
               // Make sure we get a left end cap on underline, if outlining too.
               if (ii == bits->rowBytes && (underlineText & sty) &&
                   (outlineText & sty) &&
                   ((kk + weight) < bits->bounds.edge.bottom))
                  work &= 0x3F;
               *p++ = (work << 1) | carry;
               ii--;
            }
         }
         jj += 2;               // Skip two the next time.
      }
      styl &= ~italicText;
#ifdef DEBUG
      if (styl) dumpBmap(bits, uline, sty, strwidth, weight, "italicized:");
#endif
   }

   if (outlineText & sty) {     // Algorithmic outlining (and shadow)?
                                // Make copy, then smear up & down 1 pixel.
      p = bits->baseAddr;       // Upward dest.
      size = MulU(bits->rowBytes, bits->bounds.edge.bottom);
      p1 = malloc(size);
      BlockMove(p, p1, size);
      width = bits->rowBytes;
      q = &p1[width];           // Source.
      p2 = &p[2*width];         // Downward dest.
      jj = bits->bounds.edge.bottom - 1; // Lines in bitmap to do.
      while (jj-- > 0) {
         ii = width;           // Count of bytes across bitmap.
         while (ii-- > 0) {
            *p++ |= *q;         // Smear upwards.
            *p2++ |= *q++;      // Smear downwards.
         }
      }
      if (shadowText & sty) {   // Smear downwards one more pixel.
         p = bits->baseAddr;
         q = &p1[width];        // Source.
         p2 = &p[3*width];      // Double-downward dest.
         jj = bits->bounds.edge.bottom - 1; // Lines in bitmap to do.
         while (jj-- > 0) {
            ii = width;         // Count of bytes across bitmap.
            while (ii-- > 0)
               *p2++ |= *q++;   // Smear downwards.
         }
      }
                                // Now smear left & right 1 pixel.
      q = bits->baseAddr;       // Source & Dest.
      p2 = q;
      jj = bits->bounds.edge.bottom; // Lines in bitmap to do.
      while (jj-- > 0) {
         carryin = 0;           // Preload carries.
         carry = 0;
         ii = width;            // Count of bytes across bitmap.
         while (ii-- > 0) {
            work = *q++;
            if (*q & 0x80)      // Smear sideways.
               carry = 1;
            else
               carry = 0;
            *p2++ |= (work >> 1) | (work << 1) | carry | carryin;
            if (work & 1)
               carryin = 0x80;
            else
               carryin = 0;
         }
      }

      if (shadowText & sty) {   // Smear rightwards one more pixel.
         q = bits->baseAddr;
         p2 = q;
         jj = bits->bounds.edge.bottom; // Lines in bitmap to do.
         while (jj-- > 0) {
            carry = 0;          // Preload carry.
            ii = width;         // Count of bytes across bitmap.
            while (ii-- > 0) {
               work = *q++;
               *p2++ |= (work >> 1) | carry;
               if (work & 1)
                  carry = 0x80;
               else
                  carry = 0;
            }
         }
      }

      p = bits->baseAddr;       // Now XOR in the original, punch out center.
      q = p1;
      jj = bits->bounds.edge.bottom; // Lines in bitmap to do.
      width = bits->rowBytes;   // Count of words across bitmap.
      while (jj-- > 0) {
         ii = width;
         while (ii-- > 0)
            *p++ ^= *q++;       // Knock out the center piece.
      }
      free(p1);
      styl &= ~outlineText;
#ifdef DEBUG
      if (styl) dumpBmap(bits, uline, sty, strwidth, weight, "outlined:");
#endif
      styl &= ~shadowText;
#ifdef DEBUG
      if (styl) dumpBmap(bits, uline, sty, strwidth, weight, "shadowed:");
#endif
   }

   if (ghostText & sty) {       // Algorithmic ghosting?
      p = bits->baseAddr;
      work = 0xAA;              // Knock out every other bit.
      jj = bits->bounds.edge.bottom; // Lines in bitmap to do.
      while (jj-- > 0) {
         ii = bits->rowBytes;   // Count of bytes across bitmap.
         while (ii-- > 0)
            *p++ &= work;       // Punch out every other bit.
         work = ~work;          // Alternate knockout pattern per row.
      }
   }
#ifdef DEBUG
   dumpBmap(bits, uline, sty, strwidth, weight, "After:");
#endif
}
