#include "../clunkdraw.h"

// Draws a single 8-bit character (CH) at the current pen position
// using the current text attributes.  The left-hand point of the
// baseline of the character is placed at the current pen position.
// The pen is advanced by the width of the character in the appropriate
// direction after the character is drawn.  (To the right normally,
// but it could be in any direction depending on the setting of
// TextRotate.)  The pen is left in an appropriate position to
// draw more characters so that they will merge correctly.  That is:
//
//      DrawChar('H');
//      DrawChar('i');
//
// is more-or-less the same as:
//
//      DrawString('Hi');
//
// except that it's a lot less efficient, and there may be
// discrepencies introduced by SpaceExtra and/or TextExtra as
// mentioned above.
//
// Do *not* call DrawChar if you can call DrawText or DrawString instead,
// as DrawChar is highly inefficient when compared to drawing whole
// strings at a time.
//
// Note that *no* block formatting is done --- there are no control
// characters and every character used in a text-drawing operation
// will be drawn (with a 'missing' symbol substituted if there isn't
// an appropriate character in the font; each font has its own
// 'missing' symbol).  If you're looking for something that acts like
// a terminal you'll have to write it yourself.  These are low-level
// routines that don't even assume that the characters are ASCII
// (excepting the space decisions made by SpaceExtra and the bold
// and outline style logic).  There is nothing preventing a font
// with an EBCDIC (or any other strange) encoding from being used.
//
void
DrawChar(char ch)
{
   if (!thePort) return;
   DrawText(&ch, 0, 1);
}
