#include "../clunkdraw.h"
#include "cdtext.h"
#include <stdlib.h> // For malloc, etc.

//
// Internal routine that takes care of applying style and rotations to
// a pre-made temporary text bitmap and then copying it to the destination.
//
void
_DrawIt(BitMap *tmpbits, GrafPtr tp, int ascent, Rect *dest, int swidth, int frectheight)
{
   Point newpen;              // For moving the pen.
   BitMap tmpbits2;           // For rotated off-screen images.
   Rect src;                  // Source Copybits rect.
   int toleft, totop, deltah, deltav;

   newpen = tp->pnLoc;
   src = tmpbits->bounds;
   src.edge.right = dest->edge.right - dest->edge.left;
   if (swidth > src.edge.right) {
      dest->edge.right += swidth - src.edge.right;
      src.edge.right = swidth;
   }
   if (tp->txFace)
      _AddStyle(tmpbits, ascent + 1, tp->txFace,
                swidth + (newpen.h - dest->edge.left));
   if (tp->txRotate) {         // Rotating text?
      if (!(tp->txRotate % 90)) { // Only 90 degree rotations now.
         deltah = src.edge.right - src.edge.left;
         deltav = src.edge.bottom - src.edge.top;
         toleft = newpen.h - dest->edge.left; // (Newpen is actually
         totop = newpen.v - dest->edge.top;   //  still old pen.)
         switch (tp->txRotate) {
            case 360:
                   newpen.h += swidth;                  // Move the pen.
                   break;

            case 180:
                   RotateBMapCCW(tmpbits, &tmpbits2);   // Rotate.
                   free(tmpbits->baseAddr);
                   *tmpbits = tmpbits2;                 // So it gets freed.
                   RotateBMapCCW(tmpbits, &tmpbits2);   // Rotate.
                   OffsetRect(&tmpbits2.bounds, -tmpbits2.bounds.edge.left,
                              -tmpbits2.bounds.edge.top); // Normalize rect.
                   free(tmpbits->baseAddr);
                   *tmpbits = tmpbits2;                 // So it gets freed.
                   src = tmpbits->bounds;
                   src.edge.left = src.edge.right - deltah;
                   src.edge.top = src.edge.bottom - deltav;
                   dest->edge.right = newpen.h + toleft; // Spin the drect so
                   dest->edge.bottom = newpen.v + totop; // it shows correctly.
                   dest->edge.left = dest->edge.right - deltah;
                   dest->edge.top = dest->edge.bottom - deltav;
                   newpen.h -= swidth;                  // Move the pen.
                   break;

            case 90:
                   RotateBMapCCW(tmpbits, &tmpbits2);   // Rotate.
                   OffsetRect(&tmpbits2.bounds, -tmpbits2.bounds.edge.left,
                              -tmpbits2.bounds.edge.top); // Normalize rect.
                   free(tmpbits->baseAddr);
                   *tmpbits = tmpbits2;                 // So it gets freed.
                   tmpbits = &tmpbits2;
                   src = tmpbits->bounds;
                   src.edge.right = src.edge.left + deltav;
                   src.edge.top = src.edge.bottom - deltah;
                   dest->edge.left = newpen.h - totop;  // Spin the drect so
                   dest->edge.bottom = newpen.v + toleft; // it shows correctly.
                   dest->edge.right = dest->edge.left + deltav;
                   dest->edge.top = dest->edge.bottom - deltah;
                   newpen.v -= swidth;                  // Move the pen.
                   break;

            case 270:
                   RotateBMapCW(tmpbits, &tmpbits2);    // Rotate.
                   OffsetRect(&tmpbits2.bounds, -tmpbits2.bounds.edge.left,
                              -tmpbits2.bounds.edge.top); // Normalize rect.
                   free(tmpbits->baseAddr);
                   *tmpbits = tmpbits2;                 // So it gets freed.
                   tmpbits = &tmpbits2;
                   src = tmpbits->bounds;
                   src.edge.right = src.edge.left + deltav;
                   src.edge.bottom = src.edge.top + deltah;
                   dest->edge.right = newpen.h + totop; // Spin the drect so
                   dest->edge.top = newpen.v - toleft;  // it shows correctly.
                   dest->edge.left = dest->edge.right - deltav;
                   dest->edge.bottom = dest->edge.top + deltah;
                   newpen.v += swidth;                  // Move the pen.
                   break;
         }
      } else {                  // Non-90 degree multiple rotate.
         return;                // Fascist!
      }
   } else                       // Non-rotated text, move in normal direction.
      newpen.h += swidth;
                                // Slam the bits to the video now.
   CopyBits(tmpbits, &tp->portBits, &src, dest, tp->txMode, 0);
   tp->pnLoc = newpen;          // Update the pen position.
}
