#include "../clunkdraw.h"
#include "../port/cdport.h"
#include "cdtext.h"

void
DrawJustText(string, count, r, hflag, vflag)
const char *string;
const Rect *r;
int count; int hflag; int vflag;
{
   int ii, jj, h, v, spaces, spWid, minSpWid, maxSpWid, maxWid;
   const char *p;
   FontInfo finfo;
   Fixed stretchRatio, saveSp, saveTx;

   if (!thePort || !string || !*string || count < 1 || EmptyRect(r)) return;
   /* Horizontal positioning. */
   switch (hflag) {
      case fullJustify:         /* Full justification. */
         if (count == 1) {
            DrawJustText(string, count, r, centerJustify, vflag);
            return;
         }
         for (spaces=0, p=string, ii=0; ii<count; ii++)
            if (*p++ == ' ')
               spaces++;
         saveSp = thePort->spExtra;
         saveTx = thePort->txExtra;
         SpaceExtra(0);
         TextExtra(0);
         spWid = CharWidth(' ');        // Spaces tend to be narrow already,
         minSpWid = ScaleS(spWid, 2, 3); // don't shrink to less than 2/3 normal.
         maxSpWid = ScaleS(CharWidth('M'), 3, 2);
         ii = TextWidth(string, 0, count); // Spread the rest out
         maxWid = RectWidth(r);         // inter-character.
         stretchRatio = FixRatio(maxWid - ii, spaces - 1);
         // Don't grow spaces too wide.
         if (stretchRatio > Int2Fix(maxSpWid - spWid)) {
            SpaceExtra(Int2Fix(maxSpWid - spWid));
         // Important!  We divvy up the extra space amongst all but one
         // of the characters, because we don't want any extra space after
         // the last one.  This will make for a non-zero 'error' term, but
         // the result will look a lot better.
            stretchRatio = FixRatio(maxWid - ii - ((maxSpWid - spWid) * spaces),
                                    count - 1);
            TextExtra(stretchRatio);
         } else
            SpaceExtra(stretchRatio);

         /* FALL-THROUGH! */
      case leftJustify:         /* Left justification. */
         h = r->edge.left;
         break;

      case rightJustify:        /* Right justification. */
         h = r->edge.right - TextWidth(string, 0, count);
         break;

      case centerJustify:       /* Center justification. */
         ii = TextWidth(string, 0, count);
         jj = (r->edge.right - r->edge.left) / 2;
         h = r->edge.left + jj - ii/2;
         break;

      default:                  /* Un-justified! */
         return;                /* (No such thing.) */
         break;
   }

   /* Vertical positioning. */
   GetFontInfo(&finfo);
   switch (vflag) {
      case leftJustify:         /* Top justification. */
         v = r->edge.top + finfo.ascent;
         break;

      case rightJustify:        /* Bottom justification. */
         v = r->edge.bottom - finfo.descent;
         break;

      case centerJustify:       /* Center justification. */
         ii = finfo.ascent + finfo.descent;
         jj = r->edge.bottom - r->edge.top;
         v = r->edge.top + finfo.ascent + 1 + (jj - ii) / 2;
         break;

      default:                  /* Un-justified! */
         v = thePort->pnLoc.v;  /* No move, vertically. */
         break;
   }

   _MoveTo(h, v);               /* Avoid line-drawing code.  (Smaller link?) */
   DrawText(string, 0, count);
   if (hflag == fullJustify) {
      SpaceExtra(saveSp);
      TextExtra(saveTx);
   }
}
