// GetFontInfo returns the following information about the current grafPort's
// character font, taking into consideration the style and size in which the
// characters will be drawn:  the ascent, descent, maximum character width
// (the greatest distance the pen will move when a character is drawn), and
// leading (the vertical distance between the descent line and the ascent line
// below it), all in pixels.
//
// The line height (in pixels) can be determined by adding the ascent, descent,
// and leading.

#include "../clunkdraw.h"
#include <endian.h> // For be16toh et al.

void
GetFontInfo(FontInfo *finfo)
{
   GrafPtr tp;
   const FontRec *frp;

   finfo->ascent = 0;
   finfo->descent = 0;
   finfo->widMax = 0;
   finfo->leading = 0;

   if (!(tp = thePort)) return;
   if (!(frp = tp->fontPtr)) {  // No font currently selected?
      if (!tp->_fontHook ||     // Try to get one.
          !(frp = (*tp->_fontHook)(tp)))
         return;                // Still nothing?  No information then.
      tp->fontPtr = frp;
   }
   finfo->ascent = be16toh(tp->fontPtr->ascent);
   finfo->descent = be16toh(tp->fontPtr->descent);
   finfo->widMax = be16toh(tp->fontPtr->widMax);
   finfo->leading = be16toh(tp->fontPtr->leading);
   if (boldText & tp->txFace)       // Algorithmic bolding?
      finfo->widMax += 1;           // Add 1 pixel per character for smear.
   if (outlineText & tp->txFace) {  // Outlined?
      finfo->widMax += 2;           // Add 2 pixels per character for smear.
      finfo->ascent += 1;           // Add 1 pixel height for smear.
      finfo->descent += 1;          // Add 1 pixel height for smear.
   }
   if (shadowText & tp->txFace) {   // Shadowed?
      finfo->widMax += 1;           // Add 1 pixel per character for smear.
      finfo->descent += 1;          // Add 1 pixel height for smear.
   }
   if (tp->fontPtr->fontType == 2)  // Chinese font?
      finfo->widMax += FixRound(tp->txExtra);
}
