#define TEXTEDIT_SOURCE
#include "../clunkdraw.h"
#include "texted.h"
#include <alloca.h>

/*
* _TECalText -  internal routine to set up lineStarts array, and to determine
*               which lines have been affected since last time the lineStarts
*               array was calculated
*/
void
_TECalText(start, len, pTE)
int start;
int len;
TERec *pTE;
{
   int   totWidth,
         charIndx,
         saveStart,
         saveEnd,
         oldnLines,
         endCount,
        *endStarts,
        *viewStarts,
         thisStart,
         ii,
         addLines,
         end;

   /* save the current start/end locations    (may have been set up */
   /* by delete/insert calls when replacing text)                    */
   saveStart = pTE->redrawStart;
   saveEnd = pTE->redrawEnd;

   /* calculate display width from destination rectangle    */
   /* (remove 1 to ensure there's enough room for a cursor) */
   totWidth = pTE->destRect.edge.right - pTE->destRect.edge.left - 1;
   if (pTE->just == centerJustify) {
      /* Need the extra width of a space at each edge */
      totWidth -= (pTE->spaceWidth * 2);
   }

   /* need extra space for the cursor around leaning characters */
   totWidth -= pTE->leanRight;

   /* set up the number of lines in the current display */
   _TEDispLines(pTE);

   charIndx = 0;
   if (start == -1) {
      /* unknown starting point, assume everything has to be re-drawn */
      viewStarts = 0;
      endStarts = 0;
      endCount = 0;
      *pTE->lineStarts = 0;     /* first line always at 0             */
      pTE->nLines = 1;  /* always a line to put the cursor on */
   } else {
      oldnLines = pTE->nLines;

      /* changed somewhere within the text, start recalculating */
      /* lines at the point where the change occurred             */
      pTE->nLines = _TESelLine(pTE, start);
      if (pTE->nLines > 0) {
         /* back up one, the prev line has changed if a new wrap */
         /* point was caused by the change to this line     */
         charIndx = pTE->lineStarts[pTE->nLines - 1];
      } else {
         pTE->nLines = 1;
      }

      /* save the starting points for current display lines */
      viewStarts = alloca(sizeof(short) * pTE->dispLines);
      BlockMove(&pTE->lineStarts[pTE->firstLine],
                viewStarts, sizeof(short) * pTE->dispLines);

      /* calculate the number of line starts following the changed area */
      if (len < 0) {
         end = start - len;
         if (end > pTE->teLength)
            end = pTE->teLength;
      } else
         end = start;
      endCount = oldnLines - _TESelLine(pTE, end) - 1;

      /* save the linestart following the changed area */
      endStarts = alloca(sizeof(short) * endCount);
      BlockMove(&pTE->lineStarts[oldnLines - endCount + 1], endStarts,
                endCount * sizeof(short));
    }

   /* Add in lines until they're all used up */
   while (charIndx < pTE->teLength) {
      if (pTE->nLines >= pTE->lsCount - 2) {
         /* pointers used up, allocate more (+ extra for last line) */
         if (_TEAddStarts(pTE, PTRCOUNT))
            return;
      }
      charIndx += _TELineLength(pTE, pTE->hText + charIndx, totWidth,
                                pTE->teLength - charIndx);
      pTE->lineStarts[pTE->nLines++] = charIndx;
      if (charIndx == pTE->teLength) {
         if (pTE->hText[pTE->teLength - 1] == '\n') {
            /* special processing for the last line                        */
            /* it has a phoney lineStart that's equal to the prev line */
            pTE->lineStarts[pTE->nLines++] = charIndx;
         }
      } else {
         /* skip the old starts which wrapped at a different point */
         while (endCount && charIndx > endStarts[0] + len) {
            endCount--;
            endStarts++;
         }

         /* see if the remaining text lines up with the */
         /* previous locations (don't re-calculate if   */
         /* everything lines up from here to the end)   */
         if (endCount && charIndx == endStarts[0] + len &&
             charIndx > start && charIndx > start + len) {
            /* breakpoint follows changed area, and this line   */
            /* breaks at the same character as it did before    */
            /* So, don't re-calculate anything from here to the end */

            /* allocate more pointers if they're required */
            addLines = pTE->nLines + endCount + 2 - pTE->lsCount;
            if (addLines > 0) {
               /* round it to the next PTRCOUNT */
               addLines += (PTRCOUNT - 1);
               addLines /= PTRCOUNT;
               addLines *= PTRCOUNT;
               /* pointers used up, allocate more (+ 1 for last line) */
               if (_TEAddStarts(pTE, addLines))
                  return;
            }

            while (--endCount) {
               endStarts++;
               pTE->lineStarts[pTE->nLines++] = endStarts[0] + len;
            }
            if (pTE->nLines > 2 && pTE->lineStarts[pTE->nLines - 1] ==
                /* special case for deletion at end of text */
                pTE->lineStarts[pTE->nLines - 3]) {
               pTE->nLines--;
            }
            break;
         }
      }
   }
   /* back up a bit, the last index is beyond the end of the text */
   if (--pTE->nLines < 1)
      pTE->nLines = 1;

   if (viewStarts) {
      /* decide which characters need to be redrawn */
      pTE->redrawStart = start;
      pTE->redrawEnd = pTE->teLength;
      for (ii = 0; ii < pTE->dispLines && ii + pTE->firstLine < pTE->nLines;
           ii++) {
         thisStart = pTE->lineStarts[ii + pTE->firstLine];
         if (thisStart != viewStarts[ii]) {
            /* the line start location has changed */
            if (thisStart < pTE->redrawStart) {
               pTE->redrawStart = thisStart;
            } else if (viewStarts[ii] < pTE->redrawStart) {
               pTE->redrawStart = viewStarts[ii];
            }
         }
         if (thisStart > pTE->selStart) {
            /* line follows selStart, check it out */
            if (thisStart == viewStarts[ii] + len) {
               pTE->redrawEnd = viewStarts[ii] + len;
               break;
            }
         }
      }
      if (oldnLines > pTE->nLines) {
         /* back up the refresh area if text has been deleted,  */
         /* and the cursor's now on the new last line           */
         if (ii < pTE->dispLines && viewStarts[ii] < pTE->redrawStart)
            pTE->redrawStart = viewStarts[ii];
         /* set up redrawEnd so the end of the screen will be erased    */
         pTE->redrawEnd = pTE->teLength + 1;
      } else if (pTE->nLines > oldnLines) {
         /* possible wrap to new last line, back up redraw area */
         /* so that wrapped chars from previous line will be erased  */
         if (pTE->lineStarts[pTE->firstLine + ii] < pTE->redrawStart)
            pTE->redrawStart = pTE->lineStarts[pTE->firstLine + ii];
      }
   }
   /* set up the maximum area that needs to be redrawn */
   if (pTE->redrawStart != -1 && saveStart != -1) {
      if (saveStart < pTE->redrawStart)
         pTE->redrawStart = saveStart;
      if (saveEnd > pTE->redrawEnd)
         pTE->redrawEnd = saveEnd;
   }

    /* special case for no text */
    if (pTE->teLength == 0)
       pTE->lineStarts[1] = 0;
}
