#define TEXTEDIT_SOURCE
#include "../clunkdraw.h"
#include "texted.h"

static void _TEClickSelect(int start, int end, int mode, int reHilite,
                           int minEnd, TERec *pTE);

void
TEClick(pt, extend, pTE)
Point *pt;
int extend;
TERec *pTE;
{
   EventRecord evt;
   int  start, end;
   long savedMode, newMode;
   int  doClick, mode, lastH, reHilite, round, minEnd;
   static long lastEvent = 0L;

   /* Don't allow clicks in inactive TE or outside of view rectangle */
   if (!pTE->active || !PtInRect(pt, &pTE->viewRect))
      return;

   /* get starting location */
   start = _TEHit(pt, pTE, pTE->autoView, 1, 1);
   end = start;

   if (extend) {
      /* have to un-hilite selection in case cursor line has changed */
      TEDeactivate(pTE);
      reHilite = 1;

      /* add to existing selection, set up start and end points   */
      if (end > pTE->selStart)
         start = pTE->selStart;
      else
         start = pTE->selEnd;
   } else
      reHilite = 0;

#if 0
   /* start out with drag mode and stream mode */
   savedMode = theMouseMode;
   MouseMode(F_KEYDOWN | F_KEYUP | F_MOUSE1 | mouseStream);
#endif

   mode = 0;                    /* single character mode */
   minEnd = -1;

   /* initial selection. */
   _TEClickSelect(start, end, mode, reHilite, -1, pTE);

   do {
      lastH = pTE->hOffset;
      GetNextEvent(1<<mouseEvt, &evt); /* Mouse event. */
      doClick = 0;
      if (lastEvent != -1L) {
         if (evt.when - lastEvent <= _TEDClickTime() && !extend) {
            /* double click, enter word at a time mode */
            mode = 1;
            doClick = 1;
            start = _TEHit(&evt.where.corner.topLeft, pTE,
                           pTE->autoView, 1, 0);
            end = start;
         }
      } else {
         if (!(evt.message & F_KEYUP)) {
            end = _TEHit(&evt.where.corner.topLeft, pTE,
                         pTE->autoView, 1, mode ? 0 : 1);
            if (end < 0)    /* Allow dragging off left to do a select. */
               end = 0;
            doClick = 1;
         }
      }
      if (doClick)
         _TEClickSelect(start, end, mode, 0, minEnd, pTE);

      if (lastEvent != -1L) {
         lastEvent = -1L;
         if (mode) {
            /* first event in word mode */
            if (pTE->selStart < pTE->selEnd)
               minEnd = pTE->selEnd;
            else
               minEnd = pTE->selStart;
         }
      }

      /* turn off stream mode */
      newMode = F_KEYDOWN | F_KEYUP | F_MOUSE1;

      if ((mode && lastH != pTE->hOffset) ||    /* word mode caused a scroll*/
          !PtInRect(&evt.where.corner.topLeft, &pTE->viewRect)) {
         /* mouse cursor outside of view rectangle, allow stream mode */
         newMode |= mouseStream;
      }
#if 0
      MouseMode(newMode);
#endif
   } while (!(evt.message & F_KEYUP));

   lastEvent = evt.when;        /* save the time it happened */
#if 0
   MouseMode(savedMode);        /* Restore old mousies.      */
#endif
}

/*
* _TEClickSelect - set the selection range based on char/word mode
*/
static void
_TEClickSelect(start, end, mode, reHilite, minEnd, pTE)
int start;
int end;
int mode;
int reHilite;
int minEnd;
TERec *pTE;
{
   int newEnd, autoView, temp;
   char *textPtr;

   textPtr = pTE->hText;

   /* save current autoview mode */
   autoView = pTE->autoView;

   if (start > end) {
      /* swap start and end, the mouse went backward */
      temp = start;
      start = end;
      end = temp;

      /* set special autoview mode for reversed start and end */
      if (pTE->autoView)
         pTE->autoView = 2;
   }
   if (mode) {                  /* in word mode, find word breaks */
      if (!(*(pTE->wordBreak))(textPtr, start)) { // back up to beginning of word
         start = _TEScan(0, pTE, start);
         if ((*(pTE->wordBreak))(textPtr, start))
            start++;
      }
      if (!(*(pTE->wordBreak))(textPtr, end)) {
            /* go to end of word if not on a break character */
         newEnd = _TEScan(1, pTE, end);
            /* don't use it if it's on the next line */
         if (_TESelLine(pTE, end) == _TESelLine(pTE, newEnd))
            end = newEnd;
      } else {
         if (textPtr[end] != '\n') {
            if (pTE->lineStarts[_TESelLine(pTE, end + 1)] != (end + 1))
               end++;
         }
      }
      if (end < minEnd && minEnd != -1)
         end = minEnd;
   }
   _TESetSelect(start, end, pTE);

   /* re-align display with cursor */
   _TESelView(pTE->autoView, pTE);
   if (reHilite) {
      TEActivate(pTE);
   }

   pTE->autoView = autoView;
}
