#define TEXTEDIT_SOURCE
#include "../clunkdraw.h"
#include "texted.h"

/*
* _TEHilite - calculate coordinates of current selection area and
*             invert rectangle(s) surrounding that area
*/
void
_TEHilite(pTE, onOff, clip)
TERec *pTE;
int onOff;
int clip;
{
   int offset, startLine, endLine, dispLen;
   Rect oldHilite, scratch;
   RgnHandle oldClip;

   if (clip) {
      /* Further clip output to the intersection of viewRect and destRect */
      GetClip(oldClip);
      SectRect(&pTE->viewRect, &pTE->destRect, &scratch);
      ClipRectSect(&scratch);
   }

   oldHilite = pTE->selRect;
   if (pTE->active) {
      if (onOff == 2) {
         /* changing, take down old cursor if it's there */
         if (pTE->lastselStart == pTE->lastselEnd) {
            _TEInvertRect(pTE, &oldHilite);
            pTE->active = 0;

            /* set up last vars so that everything will be re-drawn */
            pTE->lastselStart = -1;
            pTE->lastselEnd = MAXLINES;
         }
      } else {
         /* Take down old highlighting. */
         _TEInvertRect(pTE, &oldHilite);
         pTE->lastselStart = pTE->lastselEnd = -1;
         pTE->active = 0;
      }
   }

   /* onoff - 0 = off, 1 = on, 2 = change (on) */
   if (onOff) { /* Are we enabling highlighting? */
      /* first, get the lines in the selection range */
      startLine = _TESelLine(pTE, pTE->selStart) - pTE->firstLine;
      endLine = _TESelLine(pTE, pTE->selEnd) - pTE->firstLine;

      if (pTE->selStart != pTE->selEnd) {
         /* text IS selected    */
         if (pTE->selStart == pTE->lineStarts[startLine + 1]) {
            /* first selected char really starts the next line  */
            startLine++;
         }
      }

      /* set up the top and bottom edges of the rectangle */
      pTE->selRect.edge.top = pTE->destRect.edge.top
         + (startLine * pTE->lineHeight) - pTE->vOffset;
      pTE->selRect.edge.bottom = pTE->destRect.edge.top
         + ((endLine + 1) * pTE->lineHeight) - pTE->vOffset;

      /* do left edge */
      if (pTE->selStart < pTE->lineStarts[startLine + pTE->firstLine]) {
         /* start of selection scrolled off the top, start at left edge */
         pTE->selRect.edge.left = 0;
      } else {
         /* calc # of characters on this line in front of cursor */
         dispLen = pTE->selStart - pTE->lineStarts[startLine +
                                                   pTE->firstLine];

         /* calc # of pixels from left of line to start of hilite */
         pTE->selRect.edge.left = pTE->destRect.edge.left +
            _TEPixFromLeft(pTE, pTE->hText +
                           pTE->lineStarts[startLine + pTE->firstLine],
                           _TELineChars(pTE, startLine + pTE->firstLine, 0), dispLen) - 1;
      }

      /* do right edge */
      if (pTE->selEnd == pTE->selStart) {
         /* add one back in for the cursor if there's no text selected */
         pTE->selRect.edge.right = pTE->selRect.edge.left + 1;
      }
      else if (pTE->selEnd > pTE->lineStarts[endLine + 1 + pTE->firstLine]) {
         if (pTE->selEnd == pTE->teLength) {
            /* special handling for last line */
            pTE->selRect.edge.right = pTE->selRect.edge.left;
         } else {
            /* end of selection scrolled off bottom, go to right edge */
            pTE->selRect.edge.right = pTE->destRect.edge.right;
         }
      } else {
         /* calc # of characters on this line in front of cursor */
         dispLen = pTE->selEnd - pTE->lineStarts[endLine + pTE->firstLine];

         /* calc # of pixels from left of line to start of hilite */
         pTE->selRect.edge.right = pTE->destRect.edge.left +
            _TEPixFromLeft(pTE, pTE->hText +
                           pTE->lineStarts[endLine + pTE->firstLine],
                           _TELineChars(pTE, endLine + pTE->firstLine, 0),
                           dispLen) - 1;
      }
      pTE->selRect.edge.left -= pTE->hOffset;
      pTE->selRect.edge.right -= pTE->hOffset;

      pTE->cursorLine = endLine + pTE->firstLine;

      scratch = pTE->selRect;
      if (onOff != 1) {
         if (pTE->selStart == pTE->selEnd) {
            /* hilite only the cursor, turn off the old */
            if (pTE->active) {
               _TEInvertRect(pTE, &oldHilite);
            }
         }
         else if (pTE->selStart < pTE->lastselStart &&
                  pTE->selEnd == pTE->lastselEnd) {
            /* add hilite to beginning of selection */
            scratch.edge.right = oldHilite.edge.left;
            scratch.edge.bottom = oldHilite.edge.top + pTE->lineHeight;
         }
         else if (pTE->selStart > pTE->lastselStart &&
                  pTE->selEnd == pTE->lastselEnd) {

            /* remove hilite from beginning of selection */
            scratch.edge.right = pTE->selRect.edge.left;
            scratch.edge.bottom = pTE->selRect.edge.top +
               pTE->lineHeight;
            scratch.edge.left = oldHilite.edge.left;
            scratch.edge.top = oldHilite.edge.top;
         }
         else if (pTE->selEnd > pTE->lastselEnd &&
                  pTE->selStart == pTE->lastselStart) {
            /* add hilite to end of selection */
            scratch.edge.left = oldHilite.edge.right;
            scratch.edge.top = oldHilite.edge.bottom - pTE->lineHeight;
         }
         else if (pTE->selEnd < pTE->lastselEnd &&
                  pTE->selStart == pTE->lastselStart) {

            /* remove hilite from end of selection */
            scratch.edge.right = oldHilite.edge.right;
            scratch.edge.bottom = oldHilite.edge.bottom;
            scratch.edge.top = pTE->selRect.edge.bottom -
               pTE->lineHeight;
            scratch.edge.left = pTE->selRect.edge.right;
         }
         else {
            /* turn off old, everything's changed */
            if (pTE->active)
               _TEInvertRect(pTE, &oldHilite);
         }
      }
      /* highlite current selection */
      _TEInvertRect(pTE, &scratch);
      pTE->active = 1;
   }
   if (clip) {
      SetClip(oldClip);
   }
   pTE->lastselStart = pTE->selStart;
   pTE->lastselEnd = pTE->selEnd;
}
