#define TEXTEDIT_SOURCE
#include "../clunkdraw.h"
#include "texted.h"

/*
* TEKey - process special keys, add standard keystrokes to text
*/
void
TEKey(key, pTE)
long key;
TERec *pTE;
{
   int rawKey, ii;
   char *p, keyChar;
   int curLine, lines;

   if (key & F_KEYUP)   /* We don't do key-up codes. */
      return;

   rawKey = key & F_KEYMASK;
   if (rawKey == 0x0D || (rawKey == K_NPENTER && (key & F_FUNC)))
      key = rawKey = '\n';

   if (key & F_FUNC) {    /* Handle function keys. */
      switch (rawKey) {
         case K_LEFT:
            _TEGoLeft(pTE);
            break;
         case K_RIGHT:
            _TEGoRight(pTE);
            break;
         case K_HOME:
            /* set vars up for the first line of text */
            TESetSelect(0, 0, pTE);
            break;
         case K_END:
            /* set vars up for the last line of text */
            TESetSelect(pTE->teLength, pTE->teLength, pTE);
            pTE->cursorLine = pTE->nLines - 1;
            break;
         case K_PGUP:
            /* calculate the first line of the new display */
            curLine = _TESelLine(pTE, pTE->selStart);

            if (pTE->dispLines == 1)
               curLine--;
            else
               curLine -= (pTE->dispLines - 1);
            if (curLine < 0)
               curLine = 0;

            /* now, set up selection on new line */
            TESetSelect(pTE->lineStarts[curLine],
                        pTE->lineStarts[curLine], pTE);

            /* SetSelect will leave the cursor on       */
            /* the end of the previous line             */
            pTE->cursorLine++;

            /* scroll the display to the new location */
            _TEScroll(-pTE->hOffset,
                      (pTE->dispLines - 1) * pTE->lineHeight, pTE, 0);
            break;
         case K_PGDN:
            if (pTE->selEnd != pTE->teLength) {
               _TESelView(1, pTE);
               curLine = _TESelLine(pTE, pTE->selEnd);
               if (pTE->dispLines == 1)
                  curLine++;
               else
                  curLine += (pTE->dispLines - 1);
               if (curLine >= pTE->nLines)
                  curLine = pTE->nLines - 1;
               TESetSelect(pTE->lineStarts[curLine],
                           pTE->lineStarts[curLine], pTE);

               /* SetSelect will leave the cursor on    */
               /* the end of the previous line          */
               if (pTE->cursorLine < pTE->nLines - 1)
                  pTE->cursorLine++;

               /* scroll the display to the new location */
               _TEScroll(pTE->hOffset,
                         -((pTE->dispLines - 1)) * pTE->lineHeight, pTE, 0);
            }
            break;
         case K_UP:
         case K_DOWN:
            _TEUpDown(pTE, rawKey);
            break;
         case K_BACKSP:    /* Backspace keys. */
         case K_BCKSP2:
            _TEBackSpace(pTE, key);
            break;

         case K_DELETE:    /* Delete selection/character. */
            _TEDelChar(pTE);
            break;
      }
   } else if (key & F_CTRL) { /* Handle ksh-like line editing. */
      switch (rawKey) {
         case 'a':
         case 'A':      /* Beginning of line. */
            _TEBOL(pTE);
            break;
         case 'b':
         case 'B':      /* Backwards character. */
            _TEGoLeft(pTE);
            break;
         case 'c':
         case 'C':      /* Copy. */
            TECopy(pTE);
            break;
         case 'd':
         case 'D':      /* Delete character. */
            _TEDelChar(pTE);
            break;
         case 'e':
         case 'E':      /* End of line. */
            _TEEOL(pTE);
            break;
         case 'f':
         case 'F':      /* Forwards character. */
            _TEGoRight(pTE);
            break;
         case 'k':
         case 'K':      /* Kill to end of line. */
            /* find beginning of the next line */
            curLine = _TESelLine(pTE, pTE->selEnd);
            ii = pTE->lineStarts[curLine + 1];

            /* leave newlines in place */
            if (ii > pTE->selEnd + 1 && pTE->hText[ii - 1] == '\n')
               ii--;
            TESetSelect(pTE->selEnd, ii, pTE);
            TECut(pTE);
            break;
         case 'n':
         case 'N':      /* Next line. */
            _TEUpDown(pTE, K_DOWN);
            break;

         case 'p':
         case 'P':      /* Previous line. */
            _TEUpDown(pTE, K_UP);
            break;

         case 'v':
         case 'V':      /* Paste. */
         case 'y':
         case 'Y':
            TEPaste(pTE);
            break;
         case 'x':
         case 'X':      /* Cut. */
            TECut(pTE);
            break;
      }
   } else if (key & F_ALT) { /* Handle ksh-like line editing. */
      switch (rawKey) {
         case 'a':
         case 'A':      /* Beginning of line. */
            _TEBOL(pTE);
            break;
         case 'b':
         case 'B':      /* Backward word. */
            ii = _TEScan(0, pTE, pTE->selStart);
            TESetSelect(ii, ii, pTE);
            break;
         case 'd':
         case 'D':      /* Delete word. */
            ii = _TEScan(1, pTE, pTE->selStart);
            TESetSelect(pTE->selStart, ii, pTE);
            TECut(pTE);
            break;
         case 'e':
         case 'E':      /* End of line. */
            _TEEOL(pTE);
            break;
         case 'f':
         case 'F':      /* Forward word. */
            ii = _TEScan(1, pTE, pTE->selEnd);
            TESetSelect(ii, ii, pTE);
            break;
      }
   } else {         /* Normal typing character. */
      p = pTE->hText;
      if (!p)
         Bell();
      else {
         lines = pTE->nLines;
         _TEDelete(pTE);        /* Delete selection first? */
         if (lines > pTE->nLines &&
             pTE->nLines < pTE->firstLine + pTE->dispLines) {
            /* have to update screen to erase remnants of old lines     */
            /* because there are now fewer lines, and the last  */
            /* line from before won't get erased                        */
            _TEUpdate(&pTE->destRect, pTE, 1); /* Re-draw. */
         }
         if (pTE->teLength < pTE->teMaxLen) { /* Insert the character. */
            keyChar = (char)rawKey;
            if (keyChar == '\n') {
               pTE->cursorLine++;
            }
            TEInsert(&keyChar, 1, pTE);
         } else
            Bell();     /* Can't take input. */
      }
   }
   /* re-align the display with the cursor */
   _TESelView(pTE->autoView, pTE);
}
