#define TEXTEDIT_SOURCE
#include "../clunkdraw.h"
#include "texted.h"

/*
* _TELineLength - return the number of characters which will fit in
*                 a given line width
*/
int
_TELineLength(pTE, str, maxWidth, bytesInString)
TERec *pTE;
char *str;
int maxWidth;
int bytesInString;
{
   int charCount, noBreakPoint, crLoc;

   if (pTE->crOnly && pTE->just != centerJustify) {
      /* line breaks only at newline */
      crLoc = 0;
      while (crLoc < bytesInString) {
         if (str[crLoc++] == '\n')
            return crLoc;
      }
      return bytesInString;
   }

   /* do a fast check for a newline                               */
   /* (this check eliminates a LOT of calls to LocalTextWidHit,  */
   /* which in turn eliminates a LOT of calls to LocalTextWIdth) */
   crLoc = maxWidth / pTE->widMax;
   for (charCount = 0; charCount < crLoc && charCount < bytesInString;
        charCount++) {
      if (str[charCount] == '\n') {
         return charCount + 1;
      }
   }
   if (crLoc >= bytesInString)
      return bytesInString;

   /* get the maximum string that'll fit in the area */
   noBreakPoint = HiWord(TextWidHit(str, bytesInString, maxWidth));

   /* find the first newline in the string */
   while (crLoc <= noBreakPoint) {
      if (str[crLoc] == '\n')
         return crLoc + 1;
      crLoc++;
   }
   /* return if all the characters in the string will fit */
   if (noBreakPoint == bytesInString)
      return bytesInString;

   charCount = noBreakPoint;

   /* return if the entire line will fit */
   if (charCount == bytesInString)
      return charCount;

   /* word break char at EOL has to come back out, because */
   /* it's displayed in left justification mode     */
   if (pTE->just  == leftJustify)
      charCount--;

   /* have to have at least 1 char, or the string will never get displayed */
   if (charCount < 1)
      return 1;

   /* back up to last break char */
   do {
      if ((*(pTE->wordBreak))(str, charCount--)) {
         if (pTE->just == rightJustify) {
            /* move break char to beginning of next line in right mode */
            charCount--;
         }
         return charCount + 2;
      }
   } while (charCount > 0);

   /* This line is executed if there aren't any break points in the string */
   return noBreakPoint;
}
