#define TEXTEDIT_SOURCE
#include "../clunkdraw.h"
#include "texted.h"

/*
* _TEPartialUpdate - update only the area affected by an insert or
*                   delete operation
*/
void
_TEPartialUpdate(pTE)
TERec *pTE;
{
   Rect clip;
   int start, end, startLine, endLine, curLine, maxLine;

   /* save off starting and ending locations */
   start = pTE->redrawStart;
   end = pTE->redrawEnd;

   /* reset structure so _TEUpdate will work */
   pTE->redrawStart = -1;

   /* get starting and ending line numbers in update area */
   startLine = _TESelLine(pTE, start);
   if (startLine && pTE->lineStarts[startLine] == start) {
      startLine--;         /* back up to erase end of previous line */
   }
   endLine = _TESelLine(pTE, end);

   /* don't leave it on the beginning of an unchanged line */
   if (pTE->lineStarts[endLine] == end) {
      /* no chars on this line, move back one */
      if (endLine > 0)
         endLine--;
   }

   if (end > pTE->teLength) {
      /* text was deleted and end changed, erase all trailing lines */
      maxLine = pTE->firstLine + pTE->dispLines - 1;
   } else {
      maxLine = endLine;
   }

   for (curLine = startLine; curLine <= maxLine; curLine++) {
      /* set up top and bottom */
      clip.edge.top = pTE->destRect.edge.top - pTE->vOffset +
         (pTE->lineHeight * (curLine - pTE->firstLine));
      clip.edge.bottom = clip.edge.top + pTE->lineHeight;

      /* set up default left and right edges */
      clip.edge.left = pTE->destRect.edge.left;
      clip.edge.right = pTE->destRect.edge.right;

      if (pTE->just == leftJustify) {
         /* redraw optimization for left entry */

         /* set left offset for first line */
         if (curLine == startLine) {
            /* adjust left edge for current offset */
            clip.edge.left -= pTE->hOffset;
            /* skip stuff that doesn't need re-drawn */
            clip.edge.left += _TETextWidth(pTE,
                    pTE->hText + pTE->lineStarts[startLine],
                        start - pTE->lineStarts[startLine]);
         }
      }
      _TEUpdate(&clip, pTE, 1); /* redraw this line */
   }
}
