#define TEXTEDIT_SOURCE
#include "../clunkdraw.h"
#include "texted.h"

/*
* _TEUpdate - routine to update the display area
*/
void
_TEUpdate(rUpdate, pTE, doEraseFirst)
Rect *rUpdate;
TERec *pTE;
int doEraseFirst;
{
   Rect clip, scratch;
   int hilited, lineTop, curLine, firstLine, leadOffset, x;
   RgnHandle oldClip;

   if (pTE->redrawStart != -1) {
      /* turn off hiliting      */
      hilited = pTE->active;
      _TEHilite(pTE, 0, 1);

      /* only have to redraw from pTE->redrawStart to pTE->redrawEnd */
      _TEPartialUpdate(pTE);

      /* re-hilite it */
      if (hilited)
         _TEHilite(pTE, 1, 1);
      return;
   }

   hilited = pTE->active;
   if (SectRect(&pTE->viewRect, rUpdate, &clip)) {
      /* only update visible stuff within dest rectangle        */
      GetClip(oldClip);
      SectRect(&pTE->destRect, &clip, &clip);
      ClipRectSect(&clip);

      if (hilited)
         _TEHilite(pTE, 0, 0);    /* Take down existing highlighting. */

      if (doEraseFirst) /* Set up for redraw. */
         EraseRect(&clip);
      TextFont(pTE->txFont);
      TextSize(pTE->txSize);
      TextFace(pTE->txFace);
      TextMode(srcOr);

      /* keep track of the lines to be drawn    */
      firstLine = pTE->firstLine;
      curLine = firstLine;

      while (curLine < firstLine + pTE->dispLines) {
         /* calculate a rectangle containing the line to be drawn */
         lineTop = pTE->destRect.edge.top - pTE->vOffset +
            (curLine - firstLine) * pTE->lineHeight;
         SetRect(&scratch, pTE->destRect.edge.left,
                 lineTop, pTE->destRect.edge.right, lineTop + pTE->lineHeight);

         /* don't re-draw if the line's outside the clip rect */
         if (curLine < pTE->nLines && SectRect(&scratch, &clip, &scratch)) {
            if (pTE->leadOpt) {
               /* split leading between top and bottom of line  */
               leadOffset = pTE->leading >> 1;
            } else
               leadOffset = 0;
            MoveTo(pTE->destRect.edge.left +
                   _TEPixFromLeft(pTE, pTE->hText +
                                  pTE->lineStarts[curLine],
                                  _TELineChars(pTE, curLine, 0), 0) - pTE->hOffset,
                   leadOffset + pTE->destRect.edge.top +
                   pTE->fontAscent - pTE->vOffset +
                   pTE->lineHeight * (curLine - firstLine));

            DrawText(pTE->hText, pTE->lineStarts[curLine],
                     _TELineChars(pTE, curLine, 1));
         }
         curLine++;
      }
      if (hilited)
         _TEHilite(pTE, 1, 0);    /* Put up proper highlighting. */
      SetClip(oldClip);
   }
}
