#define TEXTEDIT_SOURCE
#include "../clunkdraw.h"
#include "texted.h"

/*
* TEVisLine - copy the visible portion of a TE line to the caller's buffer
*             (used for local printing RT screens)
*/
int
TEVisLine(char *buf, int maxLen, int lineNo, TERec *pTE)
{
   Point pt;
   int len, firstChar, lastChar, pixOffset;

   /* has to be a line number in the current display    */
   if (lineNo < 0 || lineNo >= pTE->dispLines || lineNo >= pTE->nLines)
      return -1;

   /* use _TEHit to find the first visible char */
   pt.h = pTE->viewRect.edge.left;
   pt.v = pTE->viewRect.edge.top;
   pt.v += lineNo * pTE->lineHeight;
   firstChar = _TEHit(&pt, pTE, 0, 1, 0);

   /* use _TEHit to find the last visible char          */
   pt.h = pTE->viewRect.edge.right;
   lastChar = _TEHit(&pt, pTE, 0, 1, 0);

   /* adjust line number for the first visible line     */
   lineNo += pTE->firstLine;

   /* set up the length of the string   */
   if (lastChar == pTE->lineStarts[lineNo + 1])
      lastChar--;
   len = lastChar - firstChar + 1;
   if (len >= maxLen)
      len = maxLen - 1;

   /* copy it to the caller's buffer    */
   BlockMove(pTE->hText + firstChar, buf, len);
   buf[len] = '\0';

   /* trim off the trailing newline     */
   if (len && buf[len - 1] == '\n')
      buf[len - 1] = '\0';

   /* tell the caller how far to offset the first character     */
   pixOffset = _TEPixFromLeft(pTE, pTE->hText + pTE->lineStarts[lineNo],
                              _TELineChars(pTE, lineNo, 1),
                              firstChar - pTE->lineStarts[lineNo]);
   return pixOffset < 0 ? 0 : pixOffset;
}
