/*
** Call BeginUpdate when an update event occurs for theWindow.  BeginUpdate
** replaces the visRgn of the window's grafPort with the intersection of the
** visRgn and the update region and then sets the window's update region to an
** empty region.  The original visRgn is saved away.  You would then usually
** (re)draw the entire content region, although it suffices to draw only the
** visRgn; in either case, only the parts of the window that require updating
** will actually be drawn on the screen.  Every call to BeginUpdate must be
** balanced by a call to EndUpdate.
**
** Unlike in Quickdraw, BeginUpdate and EndUpdate calls may be nested, but not
** if they refer to the same window (that case is protected against so that
** nothing truly evil happens if you do, but it certainly won't work right).
*/

#include "../clunkdraw.h"

void
BeginUpdate(theWindow)
WindowPtr theWindow;
{
   RgnHandle scratch;

   /* Calculate intersection of vis & update. */
   if (!((WindowPeek) theWindow)->saveVisRgn) { /* Dennis-proofing! */
      if (theWindow->device == 1) { // Handle weird e-paper rotated coords.
         OffsetRgn(((WindowPeek) theWindow)->updateRgn, // Move to lcl coords
                   theWindow->portBits.bounds.edge.top,   // (same ones the
                   theWindow->portBits.bounds.edge.left); //  visRgn is in).
      } else {
         OffsetRgn(((WindowPeek) theWindow)->updateRgn, /* Move to lcl coords */
                   theWindow->portBits.bounds.edge.left, /* (same ones the */
                   theWindow->portBits.bounds.edge.top); /*  visRgn is in). */
      }
      scratch = ((WindowPeek) theWindow)->updateRgn;
      ((WindowPeek) theWindow)->updateRgn = NewRgn();
      SectRgn(theWindow->visRgn, scratch, scratch); /* Intersect them. */
      ((WindowPeek) theWindow)->saveVisRgn = theWindow->visRgn;
      theWindow->visRgn = scratch; // Change the visRgn to the intersection
   }                               // of the dirty and drawable areas.
}
