/*
** BeginUpdateRect is like BeginUpdate except that you tell it what rectangle
** you will be updating.  The rectangle is in the local coordinates of the
** window.  This call is used to compensate for pipeline delay in the graphics
** system.  You need only issue BeginUpdateRect if you use the update
** rectangle passed in the update event to limit what you redraw.  If you
** always redraw the whole window you need not use BeginUpdateRect, but may
** use BeginUpdate instead.
**
** The offending scenario:
**
**  1)  Window is made dirty -- update event is placed in the pipeline to be
**      passed to the application.  Update rectangle is for this original
**      dirty area.
**
**  2)  Application issues a ScrollRect (for example) that causes more of the
**      window to become dirty.
**
**  3)  Application receives the original update event, issues a BeginUpdate
**      and sends only the data that falls into the original update rectangle.
**
**  4)  Application issues an EndUpdate, which fully clears out the update
**      region.
**
** The end result:  A dirty window (from step 2) that is marked as clean.
*/

#include "../clunkdraw.h"

void
BeginUpdateRect(theWindow, uprect)
WindowPtr theWindow;
Rect *uprect;
{
   RgnHandle scratch, updating, notupdated;
   int dh, dv;

   /* Calculate intersection of vis & update. */
   if (!((WindowPeek) theWindow)->saveVisRgn) { /* Dennis-proofing! */
      updating = NewRgn();
      notupdated = NewRgn();
      RectRgn(updating, uprect);
      if (theWindow->device == 1) { // Handle weird e-paper rotated coords.
         dh = theWindow->portBits.bounds.edge.top;
         dv = theWindow->portBits.bounds.edge.left;
      } else {
         dh = theWindow->portBits.bounds.edge.left;
         dv = theWindow->portBits.bounds.edge.top;
      }
      OffsetRgn(updating, -dh, -dv); /* Move to global coords. */
      DiffRgn(((WindowPeek) theWindow)->updateRgn, updating, notupdated);
      DisposeRgn(updating);
      OffsetRgn(((WindowPeek) theWindow)->updateRgn, // Move to local
                dh, dv);        // coords (the same ones the visRgn is in).
      scratch = ((WindowPeek) theWindow)->updateRgn;
      ((WindowPeek) theWindow)->updateRgn = notupdated;
      SectRgn(theWindow->visRgn, scratch, scratch); /* Intersect them. */
      ((WindowPeek) theWindow)->saveVisRgn = theWindow->visRgn;
      theWindow->visRgn = scratch; // Change the visRgn to the intersection
   }                               // of the dirty and drawable areas.
}
