/*
** BringToFront brings theWindow to the front of all other windows and redraws
** the window as necessary.  Normally you won't have to call this procedure,
** since you should call SelectWindow to make a window active, and SelectWindow
** takes care of bringing the window to the front.  If you do call
** BringToFront, however, remember to call HiliteWindow to make the necessary
** highlighting changes.
*/

#include "../clunkdraw.h"
#include "cdwmdata.h"

void
BringToFront(theWindow)
WindowPtr theWindow;
{
   WindowPeek p;
   GrafPtr savePort;
   RgnHandle scratch, r;
   int dx, dy;

   DragFlag = 0;                /* Not moving the windows around. */
   p = WindowList;
   if (p != (WindowPeek) theWindow) { /* Already frontmost? */
      GetPort(&savePort);       /* No, gotta go to work then. */
      SetPort(WMgrPort);        /* Work in window manager's port. */
      scratch = NewRgn();

      /* Subtract the structure region of our window from the visible parts
         of all windows that 'were' in front of it.   While we're at it,
         also accumulate a region that's the union of all windows' structure
         regions that were in front of us.  That will be used by PaintOne
         to figure out how much of our window has been exposed.  */

      while (p && p != (WindowPeek) theWindow) {
         if (p->visible) {
            if (p->port.device == 1) { // Handle e-paper rotated coords.
               dx = p->port.portBits.bounds.edge.top;
               dy = p->port.portBits.bounds.edge.left;
            } else {
               dx = p->port.portBits.bounds.edge.left;
               dy = p->port.portBits.bounds.edge.top;
            }
            dx = p->port.portBits.bounds.edge.left;
            dy = p->port.portBits.bounds.edge.top;
            r = p->port.visRgn;
            OffsetRgn(r, -dx, -dy);
            DiffRgn(r, ((WindowPeek) theWindow)->strucRgn, r);
            OffsetRgn(r, dx, dy);
            UnionRgn(p->strucRgn, scratch, scratch);
            if (r = p->saveVisRgn) { /* Window's being updated? */
               OffsetRgn(r, -dx, -dy);
               DiffRgn(r, ((WindowPeek) theWindow)->strucRgn, r);
               OffsetRgn(r, dx, dy);
            }
         }
         p = p->nextWindow;
      }
      _unlink_win(theWindow);   /* Rearrange window list in new order. */
      ((WindowPeek) theWindow)->nextWindow = WindowList;
      WindowList = (WindowPeek) theWindow;

      if (!DragFlag) {
         SectRgn(((WindowPeek) theWindow)->strucRgn, scratch, scratch);
         PaintOne((WindowPeek) theWindow, scratch);
      }
      CalcVis((WindowPeek) theWindow);
      DisposeRgn(scratch);
      SetPort(savePort);
   }
}
