/*
** CloseWindow removes the given window from the screen and deletes it
** from the window list.  It releases the memory occupied by all data
** structures associated with the window, but _not_ the memory taken
** up by the window record itself.  Call this procedure when you're
** done with a window.
**
** Any update events for the window are discarded.  If the window was the
** frontmost window and there was another window behind it, the latter window
** is highlighted and an appropriate activate event is generated.
*/

#include "../clunkdraw.h"
#include "cdwmdata.h"
#include <stdlib.h> // For free

void
CloseWindow(theWindow)
WindowPtr theWindow;
{
    WindowPeek p;
    GrafPtr savePort;

    GetPort(&savePort);
    SetPort(theWindow);
    if (thePort == savePort)
        savePort = 0;
    SetPort(WMgrPort);          /* Work in window manager's port. */
    _unlink_win(theWindow);     /* Remove window from list. */
    PaintBehind(((WindowPeek) theWindow)->nextWindow, /* Re-do the windows */
                ((WindowPeek) theWindow)->strucRgn);  /* behind us. */
    CalcVisBehind(((WindowPeek) theWindow)->nextWindow,
                ((WindowPeek) theWindow)->strucRgn);
    DisposeRgn(((WindowPeek) theWindow)->strucRgn); /* Dispose of regions. */
    DisposeRgn(((WindowPeek) theWindow)->contRgn);
    DisposeRgn(((WindowPeek) theWindow)->updateRgn);
    if (((WindowPeek) theWindow)->saveVisRgn)
        DisposeRgn(((WindowPeek) theWindow)->saveVisRgn);
    DisposeHandle(((WindowPeek) theWindow)->titleHandle); /* Kill title. */
    ClosePort(theWindow);       /* Close its grafport. */
    free(theWindow);            /* Free its space. */
    if (p = (WindowPeek) FrontWindow()) { /* If the front died, highlight */
        HiliteWindow((WindowPtr) p, 1); /* the next in the hierarchy. */
    }
    SetPort(savePort);
}
