/*
** Call DragWindow to move a dotted outline of theWindow around.  Point delta
** tells how far to move it.  When moving, call DragWindow once with the old
** position and then once with the new position.  Call it once more to remove
** the outline altogether.  It returns TRUE if the window is being dragged to
** an acceptable place on the screen.  If TRUE, and the dragging is ended,
** the window is moved and optionally selected if it wasn't already front.
**
** Bounds, if given, is in Global coordinates.  As is delta.
*/

#include "../clunkdraw.h"
#include "cdwmdata.h"
#define TRUE 1
#define FALSE 0

Boolean
DragWindow(theWindow, delta, bounds, trackDrag)
WindowPtr theWindow;
const Point *delta;
const Rect *bounds;
int trackDrag;
{
    int retval;
    GrafPtr savePort;
    Rect scratch, br;

    GetPort(&savePort);         /* Work for now in Window Manager's port. */
    SetPort(WMgrPort);
    SetClip(GrayRgn);
    ClipAbove((WindowPeek) theWindow);
    PenPat(&gray);
    PenMode(notSrcXor);
    scratch = (*(((WindowPeek) theWindow)->strucRgn))->rgnBBox;
    OffsetRect(&scratch, delta->h, delta->v);
    SetRect(&br, WMgrPort->portRect.edge.left + 4,
                 WMgrPort->portRect.edge.top + _menubar - 4,
                 WMgrPort->portRect.edge.right - 4,
                 WMgrPort->portRect.edge.bottom - 4);
    if (bounds)
       SectRect(&br, bounds, &br);
    retval = TRUE;              /* Limit window's possible travel. */
    if (scratch.edge.left > br.edge.right ||
        scratch.edge.right < br.edge.left ||
        scratch.edge.top > br.edge.bottom ||
        scratch.edge.top < br.edge.top)
        retval = FALSE;
    if (retval)                 /* It's OK, draw the frame. */
       FrameRect(&scratch);
    PenNormal();
    SetPort(savePort);
    if (retval && trackDrag >= tdEnd)
       MoveWindow(theWindow, delta->h, delta->v, trackDrag == tdEndSel);
    return retval;
}
