/*
** MoveWindow moves theWindow to another part of the screen, without affecting
** its size or plane.  The top left corner of the portRect of the window's
** grafPort is moved to the screen point indicated by the global coordinates
** hGlobal and vGlobal.  The local coordinates of the top left coordinate
** remain the same.  If the makeFront parameter is TRUE and theWindow isn't
** the active window, MoveWindow makes it the active window by calling
** SelectWindow(theWindow).
*/

#include "../clunkdraw.h"
#include "cdwmdata.h"

void
MoveWindow(theWindow, hGlobal, vGlobal, makeFront)
WindowPtr theWindow;
int hGlobal, vGlobal;
Boolean makeFront;
{
   GrafPtr savePort;
   RgnHandle scratch, scratch2, sr;
   int dh, dv;

   if (theWindow->device == 1) { // Handle weird e-paper rotated coords.
      dv = vGlobal - theWindow->portRect.edge.top
         + theWindow->portBits.bounds.edge.left;
      dh = hGlobal - theWindow->portRect.edge.left
         + theWindow->portBits.bounds.edge.top;
   } else {
      dv = vGlobal - theWindow->portRect.edge.top
         + theWindow->portBits.bounds.edge.top;
      dh = hGlobal - theWindow->portRect.edge.left
         + theWindow->portBits.bounds.edge.left;
   }
   if (!(dh | dv))              /* No movement at all? */
      return;                   /* Quit then. */
   GetPort(&savePort);          /* Work for now in Window Manager's port. */
   SetPort(WMgrPort);
   SetClip(GrayRgn);
   ClipAbove((WindowPeek) theWindow);
   scratch = NewRgn();
   SectRgn(sr = ((WindowPeek) theWindow)->strucRgn, WMgrPort->clipRgn, scratch);
   scratch2 = NewRgn();         /* Make a copy of window's structure region. */
   CopyRgn(sr, scratch2);
   _move_Wcoords(theWindow, hGlobal, vGlobal, scratch);
   if (makeFront)
      SetClip(GrayRgn);
   CopyBits(&WMgrPort->portBits, &WMgrPort->portBits, /* Move window data. */
            &(*scratch2)->rgnBBox,
            &(*sr)->rgnBBox,
            srcCopy, scratch);
   DiffRgn(sr, scratch, scratch);
   PaintBehind(((WindowPeek) theWindow)->nextWindow, scratch2);
   if (makeFront) {             /* If making window frontmost then do so. */
      DragFlag = 1;             /* It moved too. */
      _doSelect(theWindow);
   }
   PaintOne((WindowPeek) theWindow, scratch);
   UnionRgn(scratch2, sr, scratch2);
   CalcVisBehind((WindowPeek) theWindow, scratch2);
   DisposeRgn(scratch);         /* Clean up. */
   DisposeRgn(scratch2);
   SetPort(savePort);
}
