/*
** Window definition function.  Replaceable for custom window appearance.
*/

#define BOZO_WEENIES            /* DEFINED for weenie title bar look. */

#include "../clunkdraw.h"
#include "cdwmdata.h"
#include <endian.h> // For be16toh et al.

#ifdef BOZO_WEENIES
static short weenieflag = 2;    /* Bit 0 = weenie title bar fill, */
                                /* Bit 1 = rectangular control boxes, */
                                /* Bit 2 = all-on-one-side rectangles. */
                                /* Bit 3 = no titles. */
static int titleheit = 18;

void
WindowStyle(arg)
short arg;
{
    weenieflag = arg;
    if (arg > 15)
        titleheit = arg >> 4;
}

/*
static Pattern titlepats[] ={{0x55, 0xFF, 0x55, 0xFF, 0x55, 0xFF, 0x55, 0xFF},
                             {0xFF, 0x55, 0xFF, 0x55, 0xFF, 0x55, 0xFF, 0x55},
                             {0xAA, 0xFF, 0xAA, 0xFF, 0xAA, 0xFF, 0xAA, 0xFF},
                             {0xFF, 0xAA, 0xFF, 0xAA, 0xFF, 0xAA, 0xFF, 0xAA}};
static Pattern titlepats[] ={{0x44, 0xFF, 0x11, 0xFF, 0x44, 0xFF, 0x11, 0xFF},
                             {0xFF, 0x44, 0xFF, 0x11, 0xFF, 0x44, 0xFF, 0x11},
                             {0x22, 0xFF, 0x88, 0xFF, 0x22, 0xFF, 0x88, 0xFF},
                             {0xFF, 0x22, 0xFF, 0x88, 0xFF, 0x22, 0xFF, 0x88}};
*/
static Pattern titlepats[] ={{0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55},
                             {0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA}};
#endif

static void
calcwrects(contbox, closebox, zoombox, backbox, growbox, dragbox)
Rect *contbox, *closebox, *zoombox, *backbox, *growbox, *dragbox;
{
   int ii;
   Boolean flag1, flag2;

#ifdef BOZO_WEENIES
   flag1 = weenieflag & 2;
   flag2 = weenieflag & 4;
#else
   flag1 = FALSE;
   flag2 = FALSE;
#endif
   *dragbox = *contbox;
   ii = dragbox->edge.top;
   dragbox->edge.top = ii - titleheit;
   dragbox->edge.bottom = ii + 1;

   backbox->edge.top = closebox->edge.top = ii - (titleheit - 3);
   backbox->edge.bottom = closebox->edge.bottom = ii - 4;
   closebox->edge.left = dragbox->edge.left + 5;
   closebox->edge.right = closebox->edge.left + titleheit - (flag1 ? 3 : 7);

   if (flag2)
      backbox->edge.right = closebox->edge.right + (titleheit * 2)
                          + (flag1 ? 8 : 0);
   else
      backbox->edge.right = dragbox->edge.right - (flag1 ? 8 : 5);
   backbox->edge.left = backbox->edge.right - titleheit + (flag1 ? 3 : 7);

   *zoombox = *backbox;
   OffsetRect(zoombox, (flag1 ? -(titleheit + 4) : -titleheit), 0);

   *growbox = *contbox;
   growbox->edge.left = growbox->edge.right - 17;
   growbox->edge.top = growbox->edge.bottom - 17;
}

static Boolean
prep(r, param, varcode)
Rect *r;
long param; short varcode;
{
   InsetRect(r, 1, 1);
   if (param)
      PenPat(&black);
   else
      PenPat(&white);
   FrameRect(r);
   PenPat(&black);
   if (!varcode) {
      EraseRect(r);
      InsetRect(r, -1, -1);
      FrameRect(r);
      InsetRect(r, -1, -1);
      PenPat(&white);
      FrameRect(r);
      PenPat(&black);
   }
   return !varcode;
}

void *
MyWindow(varcode, theWindow, message, param) /* Window def function. */
int varcode;                    /* Window variation code (type). */
WindowPeek theWindow;           /* Window to work on. */
int message;                    /* What to do to it. */
void *param;                    /* Optional, as needed (return's the same). */
{
   Rect scratch, scratch2;
   Rect closebox, zoombox, backbox, growbox, dragbox;
   Point hit;
   RgnHandle arect;
   GrafPtr savePort;
   FontInfo finfo;
   int wid;                    /* String width. */
#ifdef BOZO_WEENIES
   Pattern *weeniepat;
#endif

   scratch = theWindow->port.portRect;
   GetPort(&savePort);
   SetPort(&theWindow->port);
   LocalToGlobal(&scratch.corner.topLeft);
   LocalToGlobal(&scratch.corner.botRight);
#ifdef BOZO_WEENIES
#  if 0
   weeniepat = &titlepats[(scratch.edge.top & 1) |
                          ((scratch.edge.left & 1) << 1)];
#  else
   weeniepat = &titlepats[(scratch.edge.top & 1) ^ (scratch.edge.left & 1)];
#  endif
#endif
   SetPort(savePort);
   GlobalToLocal(&scratch.corner.topLeft);
   GlobalToLocal(&scratch.corner.botRight);
   switch (message) {
      case wDraw:               /* Draw the window frame. */
         if (theWindow->visible) {
            switch (((WindowPeek) theWindow)->windowType) {
               case dBoxProc:
                  PenSize(2, 2);
                  InsetRect(&scratch, -2, -2);
                  FrameRect(&scratch); /* Frame it. */
                  PenPat(&white);
                  InsetRect(&scratch, -2, -2);
                  FrameRect(&scratch); /* Frame it. */
                  PenNormal();
                  InsetRect(&scratch, -1, -1);
                  FrameRect(&scratch); /* Frame it. */
                  break;

               case altDBoxProc:
                  scratch2 = scratch; /* Make a copy of globalized portRect. */
                  scratch2.edge.top += 2;
                  scratch2.edge.bottom++;
                  scratch2.edge.left = scratch2.edge.right + 1;
                  scratch2.edge.right += 3;
                  PaintRect(&scratch2); /* Dropshadow the right side. */

                  scratch2 = scratch; /* Make a copy */
                  scratch2.edge.top = scratch2.edge.bottom + 1;
                  scratch2.edge.left += 2;
                  scratch2.edge.right += 3;
                  scratch2.edge.bottom += 3;
                  PaintRect(&scratch2); /* Dropshadow the bottom. */

               case plainDBox:
                  InsetRect(&scratch, -1, -1);
                  FrameRect(&scratch); /* Frame it. */
                  break;

               case documentProc:
                  calcwrects(&scratch, &closebox, &zoombox, &backbox,
                             &growbox, &dragbox);
                  scratch2 = scratch; /* Make copy of globalized portRect. */
                  if (!varcode) {
                     InsetRect(&scratch, -1, -1);
                     scratch.edge.top -= titleheit;
                     FrameRect(&scratch); /* Frame the active region. */
                     scratch.edge.bottom = scratch.edge.top + titleheit + 1;
                     InsetRect(&scratch, 1, 1);
                     MoveTo(scratch.edge.left, scratch.edge.bottom);
                     LineTo(scratch.edge.right, scratch.edge.bottom);
                     EraseRect(&scratch); /* Title bar itself. */
                     if (theWindow->hilited) { /* Put up highlight lines. */
#ifdef BOZO_WEENIES
                        Rect blot;

                        if (weenieflag & 1) {
                           SetRect(&blot, scratch.edge.left + 1,
                                   scratch.edge.top + 5,
                                   scratch.edge.right - 1,
                                   scratch.edge.top + titleheit - 6);
                           FillRect(&blot, weeniepat);
                           wid = scratch.edge.right - scratch.edge.left - 3;
                           MoveTo(blot.edge.left, scratch.edge.top + 3);
                           Line(wid, 0);
                           MoveTo(blot.edge.left,
                                  scratch.edge.top + titleheit - 5);
                           Line(wid, 0);
                        } else {
                           for (wid = 3; wid < (titleheit - 4); wid += 2) {
                              MoveTo(scratch.edge.left + 1,
                                     scratch.edge.top + wid);
                              LineTo(scratch.edge.right - 2,
                                     scratch.edge.top + wid);
                           }
                        }
#else
                        for (wid = 3; wid < (titleheit - 4); wid += 2) {
                           MoveTo(scratch.edge.left + 1,
                                  scratch.edge.top + wid);
                           LineTo(scratch.edge.right - 2,
                                  scratch.edge.top + wid);
                        }
#endif
                     }           /*  */
                     if (!(weenieflag & 8)) {
                        wid = (scratch.edge.left + scratch.edge.right) / 2
                            - theWindow->titleWidth / 2;
                        scratch.edge.left = wid - 7;
                        scratch.edge.right = wid + theWindow->titleWidth + 7;
                        EraseRect(&scratch);
                        TextFont(0);
                        TextSize(0);
                        GetFontInfo(&finfo); // Force a font fetch.
                        MoveTo(wid, scratch.edge.top +
                               be16toh(thePort->fontPtr->ascent) + 4);
                        DrawString((char *) *theWindow->titleHandle);
                     }
                  }
                  if (theWindow->hilited) { /* Only do widgets if active. */
                     if (theWindow->goAwayFlag & hasGoAway) {
                        if (!varcode || varcode == inGoAway) {
                           if (prep(&closebox, (long) param, varcode)) {
                              scratch = closebox;
                              InsetRect(&scratch, 5, 5);
                              FrameRect(&scratch);
                           }
                        }
                     }
                     if (theWindow->goAwayFlag & hasZoom) {
                        if (!varcode || varcode == inZoom) {
                           if (prep(&zoombox, (long) param, varcode)) {
                              InsetRect(&zoombox, 1, 1);
                              scratch = zoombox;
                              scratch.edge.top += 7;
                              scratch.edge.right -= 7;
                              FrameRect(&scratch);
                              scratch.edge.top -= 4;
                              scratch.edge.right += 4;
                              FrameRect(&scratch);
                           }
                        }
                     }
                     if (theWindow->goAwayFlag & hasSendToBack) {
                        if (!varcode || varcode == inSendToBack) {
                           if (prep(&backbox, (long) param, varcode)) {
                              InsetRect(&backbox, 1, 1);
                              scratch = backbox;
                              scratch.edge.bottom -= 3;
                              scratch.edge.left += 3;
                              scratch.edge.top += 2;
                              scratch.edge.right -= 2;
                              FrameRect(&scratch);
                              scratch.edge.bottom += 1;
                              scratch.edge.left -= 1;
                              scratch.edge.top += 2;
                              scratch.edge.right -= 2;
                              EraseRect(&scratch);
                              FrameRect(&scratch);
                           }
                        }
                     }
                  }

                  if (!varcode) { /* Do dropshadows. */
                     scratch = scratch2;
                     scratch.edge.top -= titleheit;
                     scratch.edge.bottom++;
                     scratch.edge.left = scratch.edge.right + 1;
                     scratch.edge.right += 2;
                     PaintRect(&scratch); /* Dropshadow the right side. */

                     scratch2.edge.top = scratch2.edge.bottom + 1;
                     scratch2.edge.left += 1;
                     scratch2.edge.right += 2;
                     scratch2.edge.bottom += 2;
                     PaintRect(&scratch2); /* Dropshadow the bottom. */
                  }
                  break;

               case noBordDBox:
                  break;          /* No frame drawing required. */

               case iconWProc:
                  InsetRect(&scratch, -1, -1);
                  scratch.edge.top -= 10;
                  FrameRect(&scratch); /* Frame it. */
                  zoombox = scratch;
                  zoombox.edge.bottom = zoombox.edge.top + 11;
                  scratch = zoombox;
                  InsetRect(&scratch, 1, 1);
                  zoombox.edge.left = zoombox.edge.right - 11;
                  if (!varcode) { /* Drawing whole frame? */
                     MoveTo(scratch.edge.left, scratch.edge.bottom);
                     LineTo(scratch.edge.right, scratch.edge.bottom);
                     EraseRect(&scratch); /* Title bar itself. */
                     if (theWindow->hilited) { /* Put up highlight lines. */
                        for (wid = 2; wid < 8; wid += 2) {
                           MoveTo(scratch.edge.left + 1,
                                  scratch.edge.top + wid);
                           LineTo(scratch.edge.right-2,
                                  scratch.edge.top + wid);
                        }
                     }
                  }
                  if (theWindow->hilited) { /* Only do widgets if active. */
                     if (theWindow->goAwayFlag & hasZoom) {
                        if (!varcode || varcode == inZoom) {
                           InsetRect(&zoombox, 1, 1);
                           if (param)
                              PenPat(&black);
                           else
                              PenPat(&white);
                           FrameRect(&zoombox);
                           PenPat(&black);
                           if (!varcode) {
                              zoombox.edge.left -= 2;
                              EraseRect(&zoombox);
                              zoombox.edge.left += 2;
                              InsetRect(&zoombox, -1, -1);
                              FrameRect(&zoombox);
                           }
                        }
                     }
                  }
                  break;
            }
         }
         break;

      case wHit:                /* Find hit point within window. */
         if (theWindow->visible) {
            if (!param)
               return (void *)wNoHit;
            hit = *(Point *) param;
            GlobalToLocal(&hit);
            switch (((WindowPeek) theWindow)->windowType) {
            case dBoxProc:
            case altDBoxProc:
            case plainDBox:
            case noBordDBox:
               if (PtInRect(&hit, &scratch))
                  return (void *)wInContent;
               break;

            case documentProc:
               calcwrects(&scratch, &closebox, &zoombox, &backbox,
                          &growbox, &dragbox);
               if (theWindow->hilited) { /* Dinguses only when in front. */
                  if (theWindow->goAwayFlag & hasGoAway)
                     if (PtInRect(&hit, &closebox))
                        return (void *)wInGoAway;
                  if (theWindow->goAwayFlag & hasZoom)
                     if (PtInRect(&hit, &zoombox))
                        return (void *)wInZoom;
                  if (theWindow->goAwayFlag & hasSendToBack)
                     if (PtInRect(&hit, &backbox))
                        return (void *)wInSendToBack;
                  if (theWindow->goAwayFlag & hasGrow)
                     if (PtInRect(&hit, &growbox))
                        return (void *)wInGrow;
               }
               if (PtInRect(&hit, &dragbox))
                  return (void *)wInDrag;
               if (PtInRect(&hit, &scratch))
                  return (void *)wInContent;
               break;

            case iconWProc:
               scratch.edge.top -= 11;
               zoombox = scratch;
               zoombox.edge.left = zoombox.edge.right - 10;
               zoombox.edge.bottom = zoombox.edge.top + 11;
               if (theWindow->hilited) /* Dinguses only when in front. */
                  if (theWindow->goAwayFlag & hasZoom)
                     if (PtInRect(&hit, &zoombox))
                        return (void *)wInZoom;
               if (PtInRect(&hit, &scratch))
                  return (void *)wInDrag;
            }
         }
         return (void *)wNoHit;

      case wCalcRgns:           /* Calculate window regions. */
         arect = NewRgn();
         switch (((WindowPeek) theWindow)->windowType) {
         case dBoxProc:
            scratch2 = scratch; /* Make a copy */
            InsetRect(&scratch2, -5, -5);
            RectRgn(theWindow->strucRgn, &scratch2);
            break;

         case altDBoxProc:
            scratch2 = scratch; /* Make a copy, do body. */
            InsetRect(&scratch2, -1, -1);
            RectRgn(theWindow->strucRgn, &scratch2); /* Body */

            scratch2 = scratch; /* Make a copy, do right side strip. */
            scratch2.edge.top += 2;
            scratch2.edge.bottom++;
            scratch2.edge.left = scratch2.edge.right + 1;
            scratch2.edge.right += 3;
            RectRgn(arect, &scratch2); /* Right side. */
            UnionRgn(arect, theWindow->strucRgn, theWindow->strucRgn);

            scratch2 = scratch; /* Make a copy, do bottom strip. */
            scratch2.edge.top = scratch2.edge.bottom + 1;
            scratch2.edge.left += 2;
            scratch2.edge.right += 3;
            scratch2.edge.bottom += 3;
            RectRgn(arect, &scratch2); /* Bottom side. */
            UnionRgn(arect, theWindow->strucRgn, theWindow->strucRgn);
            break;

         case plainDBox:
            scratch2 = scratch; /* Make a copy, do body. */
            InsetRect(&scratch2, -1, -1);
            RectRgn(theWindow->strucRgn, &scratch2);
            break;

         case documentProc:
            scratch2 = scratch; /* Make a copy, do body. */
            InsetRect(&scratch2, -1, -1);
            scratch2.edge.top -= titleheit; /* For title bar. */
            RectRgn(theWindow->strucRgn, &scratch2); /* Body */
            scratch2 = scratch; /* Make a copy, do right side strip. */
            scratch2.edge.top -= titleheit;
            scratch2.edge.bottom++;
            scratch2.edge.left = scratch2.edge.right + 1;
            scratch2.edge.right += 2;
            RectRgn(arect, &scratch2); /* Right side. */
            UnionRgn(arect, theWindow->strucRgn, theWindow->strucRgn);

            scratch2 = scratch; /* Make a copy, do bottom strip. */
            scratch2.edge.top = scratch2.edge.bottom + 1;
            scratch2.edge.left += 1;
            scratch2.edge.right += 2;
            scratch2.edge.bottom += 2;
            RectRgn(arect, &scratch2); /* Bottom side. */
            UnionRgn(arect, theWindow->strucRgn, theWindow->strucRgn);
            break;

         case noBordDBox:
            RectRgn(theWindow->strucRgn, &scratch);
            break;

         case iconWProc:
            scratch2 = scratch; /* Make a copy, do body. */
            scratch2.edge.top -= 10;
            InsetRect(&scratch2, -1, -1);
            RectRgn(theWindow->strucRgn, &scratch2);
            break;

         }
         RectRgn(theWindow->contRgn, &scratch);
         DisposeRgn(arect);
         break;

      case wNew:
         break;

      case wDispose:
         break;

      case wGrow:
         if (theWindow->visible) {
            switch (((WindowPeek) theWindow)->windowType) {
            case dBoxProc:
            case altDBoxProc:
            case plainDBox:
            case noBordDBox:
            case iconWProc:
               return 0;

            case documentProc:
               scratch = (*(((WindowPeek) theWindow)->strucRgn))->rgnBBox;
               scratch.edge.right += ((Point *)param)->h;
               scratch.edge.bottom += ((Point *)param)->v;
               if (scratch.edge.right < scratch.edge.left + 60 ||
                   scratch.edge.bottom < scratch.edge.top + 40)
                  return 0;   /* Don't let it (window size) get too small! */
               FrameRect(&scratch);
               MoveTo(scratch.edge.left + 1, scratch.edge.top + titleheit);
               LineTo(scratch.edge.right - 1, scratch.edge.top + titleheit);
               if (((WindowPeek) theWindow)->goAwayFlag & hasGrow) {
                  MoveTo(wid = scratch.edge.right - 18,
                         scratch.edge.top + titleheit + 1);
                  LineTo(wid, scratch.edge.bottom - 1);
                  MoveTo(scratch.edge.left + 1,
                         wid = scratch.edge.bottom - 18);
                  LineTo(scratch.edge.right - 1, wid);
               }
               return (void *)1;
            }
         }
         break;

      case wDrawGIcon:
         if (theWindow->visible) {
            switch (((WindowPeek) theWindow)->windowType) {
            case dBoxProc:
            case altDBoxProc:
            case plainDBox:
            case noBordDBox:
            case iconWProc:
               break;

            case documentProc:
               if (theWindow->goAwayFlag & hasGrow) { /* Grow zone. */
                  calcwrects(&scratch, &closebox, &zoombox, &backbox,
                             &growbox, &dragbox);
                  OffsetRect(&growbox, 1, 1); /* Move over the edge. */
                  FrameRect(&growbox);
                  InsetRect(&growbox, 1, 1);
                  EraseRect(&growbox);
                  InsetRect(&growbox, -1, -1);
                  if (theWindow->hilited) { /* Draw icon?. */
                     growbox.edge.bottom = growbox.edge.top + 8;
                     growbox.edge.right = growbox.edge.left + 8;
                     FrameRect(&growbox);
                     growbox.edge.bottom += 4;
                     growbox.edge.right += 4;
                     FrameRect(&growbox);
                  }
               }
               break;
            }
         }
         break;
   }
   return 0;
}
