/*
** PaintOne "paints" the given window, clipped to clobberedRgn and all
** windows above it:  It draws the window frame and, if some content is
** exposed, erases the exposed area (paints it with the background pattern)
** and adds it to the window's update region.  If the window parameter is NIL,
** the window is the desktop and so is painted with the desktop pattern.
**
** The variables SaveUpdate and PaintWhite are flags used by PaintOne.
** Normally both flags are set.  Clearing SaveUpdate prevents clobberedRgn
** from being added to the window's update region.  Clearing PaintWhite
** prevents clobberedRgn from being erased before being added to the update
** region (this is useful, for example, if the background of the window isn't
** the background pattern).  The Window Manager sets both flags periodically,
** so you should clear the appropriate flag just before each situation you
** wish it to apply to.
*/

#include "../clunkdraw.h"
#include "cdwmdata.h"

void
PaintOne(window, clobberedRgn)
WindowPeek window;
RgnHandle clobberedRgn;
{
    GrafPtr savePort;
    RgnHandle scratch;
    Rect cliprect;

    GetPort(&savePort);
    SetPort(WMgrPort);
    if (window) {               /* Window or desktop? */
        if (window->visible) {  /* Window, is it visible? */
            SectRgn(clobberedRgn, window->strucRgn, WMgrPort->clipRgn);
            if (!EmptyRgn(WMgrPort->clipRgn)) { /* If window intersects */
                ClipAbove(window); /* the clobbered region then do work. */
                if (!EmptyRgn(WMgrPort->clipRgn)) {
                    scratch = NewRgn();
                    CopyRgn(WMgrPort->clipRgn, scratch);
                    if (WindowList == window) { /* For speed. */
                        cliprect = (*window->strucRgn)->rgnBBox;
                        SectRect(&cliprect, &(*GrayRgn)->rgnBBox, &cliprect);
                        ClipRect(&cliprect);
                    }           /* Frame it next. */
                    (void) MyWindow(0, window, wDraw, 0);
                    SectRgn(window->contRgn, scratch, scratch);
                    if (PaintWhite)
                        FillRgn(scratch, &window->port.bkPat); /* No content */
                    if (SaveUpdate) {
                        UnionRgn(window->updateRgn, scratch,
                                 window->updateRgn);
                    }
                    DisposeRgn(scratch);
                }
            }
        }
    } else {                    /* Desktop. */
        SetClip(clobberedRgn);
        _paintdesk();
    }
    SetPort(savePort);
}
