/*
** SendBehind sends theWindow behind behindWindow, redrawing any exposed
** windows.  If behindWindow is NULL, it sends theWindow behind all other
** windows.  If theWindow is the active window, it unhighlights theWindow,
** highlights the new active window, and generates the appropriate activate
** events.
**
** WARNING:  Do not use SendBehind to deactivate a previously active window.
** Calling SelectWindow to make a window active takes care of deactivating the
** previously active window.
**
** Note:  Unlike Apple's routine, this one properly handles the case when you
**        send a window farther forward than it was.
*/

#include "../clunkdraw.h"
#include "cdwmdata.h"

void
SendBehind(theWindow, behindWindow)
WindowPtr theWindow, behindWindow;
{
    GrafPtr savePort;
    WindowPeek p;
    RgnHandle scratch, sr;
    short movingup;

    if (behindWindow &&         /* No effective change?  We're done. */
        ((WindowPeek) behindWindow)->nextWindow == (WindowPeek) theWindow)
        return;
    GetPort(&savePort);         /* Work for now in Window Manager's port. */
    SetPort(WMgrPort);
    movingup = 0;               /* By default, window's not moving forward. */
    p = (WindowPeek) behindWindow; /* Find out if window's moving forward. */
    if (p) {                    /* Search behind 'behindwindow' for it. */
        while (p = p->nextWindow) {
            if (p == (WindowPeek) theWindow) { /* It's back there,  */
                movingup = 1;   /* therefore it's moving forward. */
                break;
            }
        }
    }
    p = (WindowPeek) theWindow;
    if ((WindowPeek) FrontWindow() == p) { /* Moving front window? */
        while (p = p->nextWindow) { /* Yes, select the next visible window. */
            if (p->visible) {
                SelectWindow((WindowPtr) p);
                break;
            }
        }
    }
    p = ((WindowPeek) theWindow)->nextWindow;
    _unlink_win(theWindow);     /* Remove theWindow from the window list. */
    _insert_win((WindowPeek) theWindow,
               (WindowPeek) behindWindow); /* Put it in new place. */
    scratch = NewRgn();         /* Calculate how many windows to update... */
    sr = ((WindowPeek) theWindow)->strucRgn;
    CopyRgn(sr, scratch);       /* Find intersection */
    _substructrgns(scratch, (WindowPeek) theWindow); /* of window's region & */
    DiffRgn(sr, scratch, scratch); /* others'. */
    PaintBehind(p, scratch);    /* Now fix all windows that used to be */
    CalcVisBehind(p, sr);       /* behind us who */
    DisposeRgn(scratch);        /* intersect this spot. */
    if (movingup) {
        PaintOne((WindowPeek) theWindow, sr);
        CalcVis((WindowPeek) theWindow);
    }
    SetPort(savePort);
}
